""" Generated from FAST OutListParameters.xlsx files with AeroelasticSE/src/AeroelasticSE/Util/create_output_vars.py """


""" ElastoDyn """
ElastoDyn = {}

# Blade 1 Tip Motions
ElastoDyn['TipDxc1']          = False     # (m); Blade 1 out-of-plane tip deflection (relative to the undeflected position); Directed along the xc1-axis
ElastoDyn['OoPDefl1']         = False     # (m); Blade 1 out-of-plane tip deflection (relative to the undeflected position); Directed along the xc1-axis
ElastoDyn['TipDyc1']          = False     # (m); Blade 1 in-plane tip deflection (relative to the undeflected position); Directed along the yc1-axis
ElastoDyn['IPDefl1']          = False     # (m); Blade 1 in-plane tip deflection (relative to the undeflected position); Directed along the yc1-axis
ElastoDyn['TipDzc1']          = False     # (m); Blade 1 axial tip deflection (relative to the undeflected position); Directed along the zc1- and zb1-axes
ElastoDyn['TipDzb1']          = False     # (m); Blade 1 axial tip deflection (relative to the undeflected position); Directed along the zc1- and zb1-axes
ElastoDyn['TipDxb1']          = False     # (m); Blade 1 flapwise tip deflection (relative to the undeflected position); Directed along the xb1-axis
ElastoDyn['TipDyb1']          = False     # (m); Blade 1 edgewise tip deflection (relative to the undeflected position); Directed along the yb1-axis
ElastoDyn['TipALxb1']         = False     # (m/s^2); Blade 1 local flapwise tip acceleration (absolute); Directed along the local xb1-axis
ElastoDyn['TipALyb1']         = False     # (m/s^2); Blade 1 local edgewise tip acceleration (absolute); Directed along the local yb1-axis
ElastoDyn['TipALzb1']         = False     # (m/s^2); Blade 1 local axial tip acceleration (absolute); Directed along the local zb1-axis
ElastoDyn['TipRDxb1']         = False     # (deg); Blade 1 roll (angular/rotational) tip deflection (relative to the undeflected position). In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the xb1-axis
ElastoDyn['RollDefl1']        = False     # (deg); Blade 1 roll (angular/rotational) tip deflection (relative to the undeflected position). In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the xb1-axis
ElastoDyn['TipRDyb1']         = False     # (deg); Blade 1 pitch (angular/rotational) tip deflection (relative to the undeflected position). In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the yb1-axis
ElastoDyn['PtchDefl1']        = False     # (deg); Blade 1 pitch (angular/rotational) tip deflection (relative to the undeflected position). In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the yb1-axis
ElastoDyn['TipRDzc1']         = False     # (deg); Blade 1 torsional tip deflection (relative to the undeflected position). This output will always be zero for FAST simulation results. Use it for examining blade torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.  Please note that this output uses the opposite of the sign convention used for blade pitch angles.; About the zc1- and zb1-axes
ElastoDyn['TipRDzb1']         = False     # (deg); Blade 1 torsional tip deflection (relative to the undeflected position). This output will always be zero for FAST simulation results. Use it for examining blade torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.  Please note that this output uses the opposite of the sign convention used for blade pitch angles.; About the zc1- and zb1-axes
ElastoDyn['TwstDefl1']        = False     # (deg); Blade 1 torsional tip deflection (relative to the undeflected position). This output will always be zero for FAST simulation results. Use it for examining blade torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.  Please note that this output uses the opposite of the sign convention used for blade pitch angles.; About the zc1- and zb1-axes
ElastoDyn['TipClrnc1']        = False     # (m); Blade 1 tip-to-tower clearance estimate. This is computed as the perpendicular distance from the yaw axis to the tip of blade 1 when the blade tip is below the yaw bearing. When the tip of blade 1 is above the yaw bearing, it is computed as the absolute distance from the yaw bearing to the blade tip. Please note that you should reduce this value by the tower radius to obtain the actual tower clearance.; N/A
ElastoDyn['TwrClrnc1']        = False     # (m); Blade 1 tip-to-tower clearance estimate. This is computed as the perpendicular distance from the yaw axis to the tip of blade 1 when the blade tip is below the yaw bearing. When the tip of blade 1 is above the yaw bearing, it is computed as the absolute distance from the yaw bearing to the blade tip. Please note that you should reduce this value by the tower radius to obtain the actual tower clearance.; N/A
ElastoDyn['Tip2Twr1']         = False     # (m); Blade 1 tip-to-tower clearance estimate. This is computed as the perpendicular distance from the yaw axis to the tip of blade 1 when the blade tip is below the yaw bearing. When the tip of blade 1 is above the yaw bearing, it is computed as the absolute distance from the yaw bearing to the blade tip. Please note that you should reduce this value by the tower radius to obtain the actual tower clearance.; N/A

# Blade 2 Tip Motions
ElastoDyn['TipDxc2']          = False     # (m); Blade 2 out-of-plane tip deflection (relative to the pitch axis); Directed along the xc2-axis
ElastoDyn['OoPDefl2']         = False     # (m); Blade 2 out-of-plane tip deflection (relative to the pitch axis); Directed along the xc2-axis
ElastoDyn['TipDyc2']          = False     # (m); Blade 2 in-plane tip deflection (relative to the pitch axis); Directed along the yc2-axis
ElastoDyn['IPDefl2']          = False     # (m); Blade 2 in-plane tip deflection (relative to the pitch axis); Directed along the yc2-axis
ElastoDyn['TipDzc2']          = False     # (m); Blade 2 axial tip deflection (relative to the pitch axis); Directed along the zc2- and zb2-axes
ElastoDyn['TipDzb2']          = False     # (m); Blade 2 axial tip deflection (relative to the pitch axis); Directed along the zc2- and zb2-axes
ElastoDyn['TipDxb2']          = False     # (m); Blade 2 flapwise tip deflection (relative to the pitch axis); Directed along the xb2-axis
ElastoDyn['TipDyb2']          = False     # (m); Blade 2 edgewise tip deflection (relative to the pitch axis); Directed along the yb2-axis
ElastoDyn['TipALxb2']         = False     # (m/s^2); Blade 2 local flapwise tip acceleration (absolute); Directed along the local xb2-axis
ElastoDyn['TipALyb2']         = False     # (m/s^2); Blade 2 local edgewise tip acceleration (absolute); Directed along the local yb2-axis
ElastoDyn['TipALzb2']         = False     # (m/s^2); Blade 2 local axial tip acceleration (absolute); Directed along the local zb2-axis
ElastoDyn['TipRDxb2']         = False     # (deg); Blade 2 roll (angular/rotational) tip deflection (relative to the undeflected position). In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the xb2-axis
ElastoDyn['RollDefl2']        = False     # (deg); Blade 2 roll (angular/rotational) tip deflection (relative to the undeflected position). In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the xb2-axis
ElastoDyn['TipRDyb2']         = False     # (deg); Blade 2 pitch (angular/rotational) tip deflection (relative to the undeflected position). In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the yb2-axis
ElastoDyn['PtchDefl2']        = False     # (deg); Blade 2 pitch (angular/rotational) tip deflection (relative to the undeflected position). In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the yb2-axis
ElastoDyn['TipRDzc2']         = False     # (deg); Blade 2 torsional (angular/rotational) tip deflection (relative to the undeflected position). This output will always be zero for FAST simulation results. Use it for examining blade torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.  Please note that this output uses the opposite of the sign convention used for blade pitch angles.; About the zc2- and zb2-axes
ElastoDyn['TipRDzb2']         = False     # (deg); Blade 2 torsional (angular/rotational) tip deflection (relative to the undeflected position). This output will always be zero for FAST simulation results. Use it for examining blade torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.  Please note that this output uses the opposite of the sign convention used for blade pitch angles.; About the zc2- and zb2-axes
ElastoDyn['TwstDefl2']        = False     # (deg); Blade 2 torsional (angular/rotational) tip deflection (relative to the undeflected position). This output will always be zero for FAST simulation results. Use it for examining blade torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.  Please note that this output uses the opposite of the sign convention used for blade pitch angles.; About the zc2- and zb2-axes
ElastoDyn['TipClrnc2']        = False     # (m); Blade 2 tip-to-tower clearance estimate. This is computed as the perpendicular distance from the yaw axis to the tip of blade 1 when the blade tip is below the yaw bearing. When the tip of blade 1 is above the yaw bearing, it is computed as the absolute distance from the yaw bearing to the blade tip. Please note that you should reduce this value by the tower radius to obtain the actual tower clearance.; N/A
ElastoDyn['TwrClrnc2']        = False     # (m); Blade 2 tip-to-tower clearance estimate. This is computed as the perpendicular distance from the yaw axis to the tip of blade 1 when the blade tip is below the yaw bearing. When the tip of blade 1 is above the yaw bearing, it is computed as the absolute distance from the yaw bearing to the blade tip. Please note that you should reduce this value by the tower radius to obtain the actual tower clearance.; N/A
ElastoDyn['Tip2Twr2']         = False     # (m); Blade 2 tip-to-tower clearance estimate. This is computed as the perpendicular distance from the yaw axis to the tip of blade 1 when the blade tip is below the yaw bearing. When the tip of blade 1 is above the yaw bearing, it is computed as the absolute distance from the yaw bearing to the blade tip. Please note that you should reduce this value by the tower radius to obtain the actual tower clearance.; N/A

# Blade 3 Tip Motions
ElastoDyn['TipDxc3']          = False     # (m); Blade 3 out-of-plane tip deflection (relative to the pitch axis); Directed along the xc3-axis
ElastoDyn['OoPDefl3']         = False     # (m); Blade 3 out-of-plane tip deflection (relative to the pitch axis); Directed along the xc3-axis
ElastoDyn['TipDyc3']          = False     # (m); Blade 3 in-plane tip deflection (relative to the pitch axis); Directed along the yc3-axis
ElastoDyn['IPDefl3']          = False     # (m); Blade 3 in-plane tip deflection (relative to the pitch axis); Directed along the yc3-axis
ElastoDyn['TipDzc3']          = False     # (m); Blade 3 axial tip deflection (relative to the pitch axis); Directed along the zc3- and zb3-axes
ElastoDyn['TipDzb3']          = False     # (m); Blade 3 axial tip deflection (relative to the pitch axis); Directed along the zc3- and zb3-axes
ElastoDyn['TipDxb3']          = False     # (m); Blade 3 flapwise tip deflection (relative to the pitch axis); Directed along the xb3-axis
ElastoDyn['TipDyb3']          = False     # (m); Blade 3 edgewise tip deflection (relative to the pitch axis); Directed along the yb3-axis
ElastoDyn['TipALxb3']         = False     # (m/s^2); Blade 3 local flapwise tip acceleration (absolute); Directed along the local xb3-axis
ElastoDyn['TipALyb3']         = False     # (m/s^2); Blade 3 local edgewise tip acceleration (absolute); Directed along the local yb3-axis
ElastoDyn['TipALzb3']         = False     # (m/s^2); Blade 3 local axial tip acceleration (absolute); Directed along the local zb3-axis
ElastoDyn['TipRDxb3']         = False     # (deg); Blade 3 roll (angular/rotational) tip deflection (relative to the undeflected position). In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the xb3-axis
ElastoDyn['RollDefl3']        = False     # (deg); Blade 3 roll (angular/rotational) tip deflection (relative to the undeflected position). In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the xb3-axis
ElastoDyn['TipRDyb3']         = False     # (deg); Blade 3 pitch (angular/rotational) tip deflection (relative to the undeflected position). In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the yb3-axis
ElastoDyn['PtchDefl3']        = False     # (deg); Blade 3 pitch (angular/rotational) tip deflection (relative to the undeflected position). In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the yb3-axis
ElastoDyn['TipRDzc3']         = False     # (deg); Blade 3 torsional tip deflection (relative to the undeflected position). This output will always be zero for FAST simulation results. Use it for examining blade torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.  Please note that this output uses the opposite of the sign convention used for blade pitch angles.; About the zc3- and zb3-axes
ElastoDyn['TipRDzb3']         = False     # (deg); Blade 3 torsional tip deflection (relative to the undeflected position). This output will always be zero for FAST simulation results. Use it for examining blade torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.  Please note that this output uses the opposite of the sign convention used for blade pitch angles.; About the zc3- and zb3-axes
ElastoDyn['TwstDefl3']        = False     # (deg); Blade 3 torsional tip deflection (relative to the undeflected position). This output will always be zero for FAST simulation results. Use it for examining blade torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.  Please note that this output uses the opposite of the sign convention used for blade pitch angles.; About the zc3- and zb3-axes
ElastoDyn['TipClrnc3']        = False     # (m); Blade 3 tip-to-tower clearance estimate. This is computed as the perpendicular distance from the yaw axis to the tip of blade 1 when the blade tip is below the yaw bearing. When the tip of blade 1 is above the yaw bearing, it is computed as the absolute distance from the yaw bearing to the blade tip. Please note that you should reduce this value by the tower radius to obtain the actual tower clearance.; N/A
ElastoDyn['TwrClrnc3']        = False     # (m); Blade 3 tip-to-tower clearance estimate. This is computed as the perpendicular distance from the yaw axis to the tip of blade 1 when the blade tip is below the yaw bearing. When the tip of blade 1 is above the yaw bearing, it is computed as the absolute distance from the yaw bearing to the blade tip. Please note that you should reduce this value by the tower radius to obtain the actual tower clearance.; N/A
ElastoDyn['Tip2Twr3']         = False     # (m); Blade 3 tip-to-tower clearance estimate. This is computed as the perpendicular distance from the yaw axis to the tip of blade 1 when the blade tip is below the yaw bearing. When the tip of blade 1 is above the yaw bearing, it is computed as the absolute distance from the yaw bearing to the blade tip. Please note that you should reduce this value by the tower radius to obtain the actual tower clearance.; N/A

# Blade 1 Local Span Motions
ElastoDyn['Spn1ALxb1']        = False     # (m/s^2); Blade 1 local flapwise acceleration (absolute) of span station 1; Directed along the local xb1-axis
ElastoDyn['Spn1ALyb1']        = False     # (m/s^2); Blade 1 local edgewise acceleration (absolute) of span station 1; Directed along the local yb1-axis
ElastoDyn['Spn1ALzb1']        = False     # (m/s^2); Blade 1 local axial acceleration (absolute) of span station 1; Directed along the local zb1-axis
ElastoDyn['Spn2ALxb1']        = False     # (m/s^2); Blade 1 local flapwise acceleration (absolute) of span  station 2; Directed along the local xb1-axis
ElastoDyn['Spn2ALyb1']        = False     # (m/s^2); Blade 1 local edgewise acceleration (absolute) of span station 2; Directed along the local yb1-axis
ElastoDyn['Spn2ALzb1']        = False     # (m/s^2); Blade 1 local axial acceleration (absolute) of span station 2; Directed along the local zb1-axis
ElastoDyn['Spn3ALxb1']        = False     # (m/s^2); Blade 1 local flapwise acceleration (absolute) of span station 3; Directed along the local xb1-axis
ElastoDyn['Spn3ALyb1']        = False     # (m/s^2); Blade 1 local edgewise acceleration (absolute) of span station 3; Directed along the local yb1-axis
ElastoDyn['Spn3ALzb1']        = False     # (m/s^2); Blade 1 local axial acceleration (absolute) of span station 3; Directed along the local zb1-axis
ElastoDyn['Spn4ALxb1']        = False     # (m/s^2); Blade 1 local flapwise acceleration (absolute) of span station 4; Directed along the local xb1-axis
ElastoDyn['Spn4ALyb1']        = False     # (m/s^2); Blade 1 local edgewise acceleration (absolute) of span station 4; Directed along the local yb1-axis
ElastoDyn['Spn4ALzb1']        = False     # (m/s^2); Blade 1 local axial acceleration (absolute) of span station 4; Directed along the local zb1-axis
ElastoDyn['Spn5ALxb1']        = False     # (m/s^2); Blade 1 local flapwise acceleration (absolute) of span station 5; Directed along the local xb1-axis
ElastoDyn['Spn5ALyb1']        = False     # (m/s^2); Blade 1 local edgewise acceleration (absolute) of span station 5; Directed along the local yb1-axis
ElastoDyn['Spn5ALzb1']        = False     # (m/s^2); Blade 1 local axial acceleration (absolute) of span station 5; Directed along the local zb1-axis
ElastoDyn['Spn6ALxb1']        = False     # (m/s^2); Blade 1 local flapwise acceleration (absolute) of span station 6; Directed along the local xb1-axis
ElastoDyn['Spn6ALyb1']        = False     # (m/s^2); Blade 1 local edgewise acceleration (absolute) of span station 6; Directed along the local yb1-axis
ElastoDyn['Spn6ALzb1']        = False     # (m/s^2); Blade 1 local axial acceleration (absolute) of span station 6; Directed along the local zb1-axis
ElastoDyn['Spn7ALxb1']        = False     # (m/s^2); Blade 1 local flapwise acceleration (absolute) of span station 7; Directed along the local xb1-axis
ElastoDyn['Spn7ALyb1']        = False     # (m/s^2); Blade 1 local edgewise acceleration (absolute) of span station 7; Directed along the local yb1-axis
ElastoDyn['Spn7ALzb1']        = False     # (m/s^2); Blade 1 local axial acceleration (absolute) of span station 7; Directed along the local zb1-axis
ElastoDyn['Spn8ALxb1']        = False     # (m/s^2); Blade 1 local flapwise acceleration (absolute) of span station 8; Directed along the local xb1-axis
ElastoDyn['Spn8ALyb1']        = False     # (m/s^2); Blade 1 local edgewise acceleration (absolute) of span station 8; Directed along the local yb1-axis
ElastoDyn['Spn8ALzb1']        = False     # (m/s^2); Blade 1 local axial acceleration (absolute) of span station 8; Directed along the local zb1-axis
ElastoDyn['Spn9ALxb1']        = False     # (m/s^2); Blade 1 local flapwise acceleration (absolute) of span station 9; Directed along the local xb1-axis
ElastoDyn['Spn9ALyb1']        = False     # (m/s^2); Blade 1 local edgewise acceleration (absolute) of span station 9; Directed along the local yb1-axis
ElastoDyn['Spn9ALzb1']        = False     # (m/s^2); Blade 1 local axial acceleration (absolute) of span station 9; Directed along the local zb1-axis
ElastoDyn['Spn1TDxb1']        = False     # (m); Blade 1 local flapwise (translational) deflection (relative to the undeflected position) of span station 1; Directed along the xb1-axis
ElastoDyn['Spn1TDyb1']        = False     # (m); Blade 1 local edgewise (translational) deflection (relative to the undeflected position) of span station 1; Directed along the yb1-axis
ElastoDyn['Spn1TDzb1']        = False     # (m); Blade 1 local axial (translational) deflection (relative to the undeflected position) of span station 1; Directed along the zb1-axis
ElastoDyn['Spn2TDxb1']        = False     # (m); Blade 1 local flapwise (translational) deflection (relative to the undeflected position) of span station 2; Directed along the xb1-axis
ElastoDyn['Spn2TDyb1']        = False     # (m); Blade 1 local edgewise (translational) deflection (relative to the undeflected position) of span station 2; Directed along the yb1-axis
ElastoDyn['Spn2TDzb1']        = False     # (m); Blade 1 local axial (translational) deflection (relative to the undeflected position) of span station 2; Directed along the zb1-axis
ElastoDyn['Spn3TDxb1']        = False     # (m); Blade 1 local flapwise (translational) deflection (relative to the undeflected position) of span station 3; Directed along the xb1-axis
ElastoDyn['Spn3TDyb1']        = False     # (m); Blade 1 local edgewise (translational) deflection (relative to the undeflected position) of span station 3; Directed along the yb1-axis
ElastoDyn['Spn3TDzb1']        = False     # (m); Blade 1 local axial (translational) deflection (relative to the undeflected position) of span station 3; Directed along the zb1-axis
ElastoDyn['Spn4TDxb1']        = False     # (m); Blade 1 local flapwise (translational) deflection (relative to the undeflected position) of span station 4; Directed along the xb1-axis
ElastoDyn['Spn4TDyb1']        = False     # (m); Blade 1 local edgewise (translational) deflection (relative to the undeflected position) of span station 4; Directed along the yb1-axis
ElastoDyn['Spn4TDzb1']        = False     # (m); Blade 1 local axial (translational) deflection (relative to the undeflected position) of span station 4; Directed along the zb1-axis
ElastoDyn['Spn5TDxb1']        = False     # (m); Blade 1 local flapwise (translational) deflection (relative to the undeflected position) of span station 5; Directed along the xb1-axis
ElastoDyn['Spn5TDyb1']        = False     # (m); Blade 1 local edgewise (translational) deflection (relative to the undeflected position) of span station 5; Directed along the yb1-axis
ElastoDyn['Spn5TDzb1']        = False     # (m); Blade 1 local axial (translational) deflection (relative to the undeflected position) of span station 5; Directed along the zb1-axis
ElastoDyn['Spn6TDxb1']        = False     # (m); Blade 1 local flapwise (translational) deflection (relative to the undeflected position) of span station 6; Directed along the xb1-axis
ElastoDyn['Spn6TDyb1']        = False     # (m); Blade 1 local edgewise (translational) deflection (relative to the undeflected position) of span station 6; Directed along the yb1-axis
ElastoDyn['Spn6TDzb1']        = False     # (m); Blade 1 local axial (translational) deflection (relative to the undeflected position) of span station 6; Directed along the zb1-axis
ElastoDyn['Spn7TDxb1']        = False     # (m); Blade 1 local flapwise (translational) deflection (relative to the undeflected position) of span station 7; Directed along the xb1-axis
ElastoDyn['Spn7TDyb1']        = False     # (m); Blade 1 local edgewise (translational) deflection (relative to the undeflected position) of span station 7; Directed along the yb1-axis
ElastoDyn['Spn7TDzb1']        = False     # (m); Blade 1 local axial (translational) deflection (relative to the undeflected position) of span station 7; Directed along the zb1-axis
ElastoDyn['Spn8TDxb1']        = False     # (m); Blade 1 local flapwise (translational) deflection (relative to the undeflected position) of span station 8; Directed along the xb1-axis
ElastoDyn['Spn8TDyb1']        = False     # (m); Blade 1 local edgewise (translational) deflection (relative to the undeflected position) of span station 8; Directed along the yb1-axis
ElastoDyn['Spn8TDzb1']        = False     # (m); Blade 1 local axial (translational) deflection (relative to the undeflected position) of span station 8; Directed along the zb1-axis
ElastoDyn['Spn9TDxb1']        = False     # (m); Blade 1 local flapwise (translational) deflection (relative to the undeflected position) of span station 9; Directed along the xb1-axis
ElastoDyn['Spn9TDyb1']        = False     # (m); Blade 1 local edgewise (translational) deflection (relative to the undeflected position) of span station 9; Directed along the yb1-axis
ElastoDyn['Spn9TDzb1']        = False     # (m); Blade 1 local axial (translational) deflection (relative to the undeflected position) of span station 9; Directed along the zb1-axis
ElastoDyn['Spn1RDxb1']        = False     # (deg); Blade 1 local roll (angular/rotational)  deflection (relative to the undeflected position) of span station 1. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local xb1-axis
ElastoDyn['Spn1RDyb1']        = False     # (deg); Blade 1 local pitch (angular/rotational) deflection (relative to the undeflected position) of span station 1. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local yb1-axis
ElastoDyn['Spn1RDzb1']        = False     # (deg); Blade 1 local torsional (angular/rotational) deflection (relative to the undeflected position) of span station 1. This output will always be zero for FAST simulation results. Use it for examining blade torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.  Please note that this output uses the opposite of the sign convention used for blade pitch angles.; About the local zb1-axis
ElastoDyn['Spn2RDxb1']        = False     # (deg); Blade 1 local roll (angular/rotational)  deflection (relative to the undeflected position) of span station 2. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local xb1-axis
ElastoDyn['Spn2RDyb1']        = False     # (deg); Blade 1 local pitch (angular/rotational) deflection (relative to the undeflected position) of span station 2. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local yb1-axis
ElastoDyn['Spn2RDzb1']        = False     # (deg); Blade 1 local torsional (angular/rotational) deflection (relative to the undeflected position) of span station 2. This output will always be zero for FAST simulation results. Use it for examining blade torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.  Please note that this output uses the opposite of the sign convention used for blade pitch angles.; About the local zb1-axis
ElastoDyn['Spn3RDxb1']        = False     # (deg); Blade 1 local roll (angular/rotational)  deflection (relative to the undeflected position) of span station 3. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local xb1-axis
ElastoDyn['Spn3RDyb1']        = False     # (deg); Blade 1 local pitch (angular/rotational) deflection (relative to the undeflected position) of span station 3. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local yb1-axis
ElastoDyn['Spn3RDzb1']        = False     # (deg); Blade 1 local torsional (angular/rotational) deflection (relative to the undeflected position) of span station 3. This output will always be zero for FAST simulation results. Use it for examining blade torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.  Please note that this output uses the opposite of the sign convention used for blade pitch angles.; About the local zb1-axis
ElastoDyn['Spn4RDxb1']        = False     # (deg); Blade 1 local roll (angular/rotational)  deflection (relative to the undeflected position) of span station 4. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local xb1-axis
ElastoDyn['Spn4RDyb1']        = False     # (deg); Blade 1 local pitch (angular/rotational) deflection (relative to the undeflected position) of span station 4. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local yb1-axis
ElastoDyn['Spn4RDzb1']        = False     # (deg); Blade 1 local torsional (angular/rotational) deflection (relative to the undeflected position) of span station 4. This output will always be zero for FAST simulation results. Use it for examining blade torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.  Please note that this output uses the opposite of the sign convention used for blade pitch angles.; About the local zb1-axis
ElastoDyn['Spn5RDxb1']        = False     # (deg); Blade 1 local roll (angular/rotational)  deflection (relative to the undeflected position) of span station 5. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local xb1-axis
ElastoDyn['Spn5RDyb1']        = False     # (deg); Blade 1 local pitch (angular/rotational) deflection (relative to the undeflected position) of span station 5. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local yb1-axis
ElastoDyn['Spn5RDzb1']        = False     # (deg); Blade 1 local torsional (angular/rotational) deflection (relative to the undeflected position) of span station 5. This output will always be zero for FAST simulation results. Use it for examining blade torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.  Please note that this output uses the opposite of the sign convention used for blade pitch angles.; About the local zb1-axis
ElastoDyn['Spn6RDxb1']        = False     # (deg); Blade 1 local roll (angular/rotational)  deflection (relative to the undeflected position) of span station 6. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local xb1-axis
ElastoDyn['Spn6RDyb1']        = False     # (deg); Blade 1 local pitch (angular/rotational) deflection (relative to the undeflected position) of span station 6. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local yb1-axis
ElastoDyn['Spn6RDzb1']        = False     # (deg); Blade 1 local torsional (angular/rotational) deflection (relative to the undeflected position) of span station 6. This output will always be zero for FAST simulation results. Use it for examining blade torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.  Please note that this output uses the opposite of the sign convention used for blade pitch angles.; About the local zb1-axis
ElastoDyn['Spn7RDxb1']        = False     # (deg); Blade 1 local roll (angular/rotational)  deflection (relative to the undeflected position) of span station 7. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local xb1-axis
ElastoDyn['Spn7RDyb1']        = False     # (deg); Blade 1 local pitch (angular/rotational) deflection (relative to the undeflected position) of span station 7. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local yb1-axis
ElastoDyn['Spn7RDzb1']        = False     # (deg); Blade 1 local torsional (angular/rotational) deflection (relative to the undeflected position) of span station 7. This output will always be zero for FAST simulation results. Use it for examining blade torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.  Please note that this output uses the opposite of the sign convention used for blade pitch angles.; About the local zb1-axis
ElastoDyn['Spn8RDxb1']        = False     # (deg); Blade 1 local roll (angular/rotational)  deflection (relative to the undeflected position) of span station 8. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local xb1-axis
ElastoDyn['Spn8RDyb1']        = False     # (deg); Blade 1 local pitch (angular/rotational) deflection (relative to the undeflected position) of span station 8. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local yb1-axis
ElastoDyn['Spn8RDzb1']        = False     # (deg); Blade 1 local torsional (angular/rotational) deflection (relative to the undeflected position) of span station 8. This output will always be zero for FAST simulation results. Use it for examining blade torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.  Please note that this output uses the opposite of the sign convention used for blade pitch angles.; About the local zb1-axis
ElastoDyn['Spn9RDxb1']        = False     # (deg); Blade 1 local roll (angular/rotational)  deflection (relative to the undeflected position) of span station 9. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local xb1-axis
ElastoDyn['Spn9RDyb1']        = False     # (deg); Blade 1 local pitch (angular/rotational) deflection (relative to the undeflected position) of span station 9. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local yb1-axis
ElastoDyn['Spn9RDzb1']        = False     # (deg); Blade 1 local torsional (angular/rotational) deflection (relative to the undeflected position) of span station 9. This output will always be zero for FAST simulation results. Use it for examining blade torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.  Please note that this output uses the opposite of the sign convention used for blade pitch angles.; About the local zb1-axis

# Blade 2 Local Span Motions
ElastoDyn['Spn1ALxb2']        = False     # (m/s^2); Blade 2 local flapwise acceleration (absolute) of span station 1; Directed along the local xb2-axis
ElastoDyn['Spn1ALyb2']        = False     # (m/s^2); Blade 2 local edgewise acceleration (absolute) of span station 1; Directed along the local yb2-axis
ElastoDyn['Spn1ALzb2']        = False     # (m/s^2); Blade 2 local axial acceleration (absolute) of span station 1; Directed along the local zb2-axis
ElastoDyn['Spn2ALxb2']        = False     # (m/s^2); Blade 2 local flapwise acceleration (absolute) of span station 2; Directed along the local xb2-axis
ElastoDyn['Spn2ALyb2']        = False     # (m/s^2); Blade 2 local edgewise acceleration (absolute) of span station 2; Directed along the local yb2-axis
ElastoDyn['Spn2ALzb2']        = False     # (m/s^2); Blade 2 local axial acceleration (absolute) of span station 2; Directed along the local zb2-axis
ElastoDyn['Spn3ALxb2']        = False     # (m/s^2); Blade 2 local flapwise acceleration (absolute) of span station 3; Directed along the local xb2-axis
ElastoDyn['Spn3ALyb2']        = False     # (m/s^2); Blade 2 local edgewise acceleration (absolute) of span station 3; Directed along the local yb2-axis
ElastoDyn['Spn3ALzb2']        = False     # (m/s^2); Blade 2 local axial acceleration (absolute) of span station 3; Directed along the local zb2-axis
ElastoDyn['Spn4ALxb2']        = False     # (m/s^2); Blade 2 local flapwise acceleration (absolute) of span station 4; Directed along the local xb2-axis
ElastoDyn['Spn4ALyb2']        = False     # (m/s^2); Blade 2 local edgewise acceleration (absolute) of span station 4; Directed along the local yb2-axis
ElastoDyn['Spn4ALzb2']        = False     # (m/s^2); Blade 2 local axial acceleration (absolute) of span station 4; Directed along the local zb2-axis
ElastoDyn['Spn5ALxb2']        = False     # (m/s^2); Blade 2 local flapwise acceleration (absolute) of span station 5; Directed along the local xb2-axis
ElastoDyn['Spn5ALyb2']        = False     # (m/s^2); Blade 2 local edgewise acceleration (absolute) of span station 5; Directed along the local yb2-axis
ElastoDyn['Spn5ALzb2']        = False     # (m/s^2); Blade 2 local axial acceleration (absolute) of span station 5; Directed along the local zb2-axis
ElastoDyn['Spn6ALxb2']        = False     # (m/s^2); Blade 2 local flapwise acceleration (absolute) of span station 6; Directed along the local xb2-axis
ElastoDyn['Spn6ALyb2']        = False     # (m/s^2); Blade 2 local edgewise acceleration (absolute) of span station 6; Directed along the local yb2-axis
ElastoDyn['Spn6ALzb2']        = False     # (m/s^2); Blade 2 local axial acceleration (absolute) of span station 6; Directed along the local zb2-axis
ElastoDyn['Spn7ALxb2']        = False     # (m/s^2); Blade 2 local flapwise acceleration (absolute) of span station 7; Directed along the local xb2-axis
ElastoDyn['Spn7ALyb2']        = False     # (m/s^2); Blade 2 local edgewise acceleration (absolute) of span station 7; Directed along the local yb2-axis
ElastoDyn['Spn7ALzb2']        = False     # (m/s^2); Blade 2 local axial acceleration (absolute) of span station 7; Directed along the local zb2-axis
ElastoDyn['Spn8ALxb2']        = False     # (m/s^2); Blade 2 local flapwise acceleration (absolute) of span station 8; Directed along the local xb2-axis
ElastoDyn['Spn8ALyb2']        = False     # (m/s^2); Blade 2 local edgewise acceleration (absolute) of span station 8; Directed along the local yb2-axis
ElastoDyn['Spn8ALzb2']        = False     # (m/s^2); Blade 2 local axial acceleration (absolute) of span station 8; Directed along the local zb2-axis
ElastoDyn['Spn9ALxb2']        = False     # (m/s^2); Blade 2 local flapwise acceleration (absolute) of span station 9; Directed along the local xb2-axis
ElastoDyn['Spn9ALyb2']        = False     # (m/s^2); Blade 2 local edgewise acceleration (absolute) of span station 9; Directed along the local yb2-axis
ElastoDyn['Spn9ALzb2']        = False     # (m/s^2); Blade 2 local axial acceleration (absolute) of span station 9; Directed along the local zb2-axis
ElastoDyn['Spn1TDxb2']        = False     # (m); Blade 2 local flapwise (translational) deflection (relative to the undeflected position) of span station 1; Directed along the xb2-axis
ElastoDyn['Spn1TDyb2']        = False     # (m); Blade 2 local edgewise (translational) deflection (relative to the undeflected position) of span station 1; Directed along the yb2-axis
ElastoDyn['Spn1TDzb2']        = False     # (m); Blade 2 local axial (translational) deflection (relative to the undeflected position) of span station 1; Directed along the zb2-axis
ElastoDyn['Spn2TDxb2']        = False     # (m); Blade 2 local flapwise (translational) deflection (relative to the undeflected position) of span station 2; Directed along the xb2-axis
ElastoDyn['Spn2TDyb2']        = False     # (m); Blade 2 local edgewise (translational) deflection (relative to the undeflected position) of span station 2; Directed along the yb2-axis
ElastoDyn['Spn2TDzb2']        = False     # (m); Blade 2 local axial (translational) deflection (relative to the undeflected position) of span station 2; Directed along the zb2-axis
ElastoDyn['Spn3TDxb2']        = False     # (m); Blade 2 local flapwise (translational) deflection (relative to the undeflected position) of span station 3; Directed along the xb2-axis
ElastoDyn['Spn3TDyb2']        = False     # (m); Blade 2 local edgewise (translational) deflection (relative to the undeflected position) of span station 3; Directed along the yb2-axis
ElastoDyn['Spn3TDzb2']        = False     # (m); Blade 2 local axial (translational) deflection (relative to the undeflected position) of span station 3; Directed along the zb2-axis
ElastoDyn['Spn4TDxb2']        = False     # (m); Blade 2 local flapwise (translational) deflection (relative to the undeflected position) of span station 4; Directed along the xb2-axis
ElastoDyn['Spn4TDyb2']        = False     # (m); Blade 2 local edgewise (translational) deflection (relative to the undeflected position) of span station 4; Directed along the yb2-axis
ElastoDyn['Spn4TDzb2']        = False     # (m); Blade 2 local axial (translational) deflection (relative to the undeflected position) of span station 4; Directed along the zb2-axis
ElastoDyn['Spn5TDxb2']        = False     # (m); Blade 2 local flapwise (translational) deflection (relative to the undeflected position) of span station 5; Directed along the xb2-axis
ElastoDyn['Spn5TDyb2']        = False     # (m); Blade 2 local edgewise (translational) deflection (relative to the undeflected position) of span station 5; Directed along the yb2-axis
ElastoDyn['Spn5TDzb2']        = False     # (m); Blade 2 local axial (translational) deflection (relative to the undeflected position) of span station 5; Directed along the zb2-axis
ElastoDyn['Spn6TDxb2']        = False     # (m); Blade 2 local flapwise (translational) deflection (relative to the undeflected position) of span station 6; Directed along the xb2-axis
ElastoDyn['Spn6TDyb2']        = False     # (m); Blade 2 local edgewise (translational) deflection (relative to the undeflected position) of span station 6; Directed along the yb2-axis
ElastoDyn['Spn6TDzb2']        = False     # (m); Blade 2 local axial (translational) deflection (relative to the undeflected position) of span station 6; Directed along the zb2-axis
ElastoDyn['Spn7TDxb2']        = False     # (m); Blade 2 local flapwise (translational) deflection (relative to the undeflected position) of span station 7; Directed along the xb2-axis
ElastoDyn['Spn7TDyb2']        = False     # (m); Blade 2 local edgewise (translational) deflection (relative to the undeflected position) of span station 7; Directed along the yb2-axis
ElastoDyn['Spn7TDzb2']        = False     # (m); Blade 2 local axial (translational) deflection (relative to the undeflected position) of span station 7; Directed along the zb2-axis
ElastoDyn['Spn8TDxb2']        = False     # (m); Blade 2 local flapwise (translational) deflection (relative to the undeflected position) of span station 8; Directed along the xb2-axis
ElastoDyn['Spn8TDyb2']        = False     # (m); Blade 2 local edgewise (translational) deflection (relative to the undeflected position) of span station 8; Directed along the yb2-axis
ElastoDyn['Spn8TDzb2']        = False     # (m); Blade 2 local axial (translational) deflection (relative to the undeflected position) of span station 8; Directed along the zb2-axis
ElastoDyn['Spn9TDxb2']        = False     # (m); Blade 2 local flapwise (translational) deflection (relative to the undeflected position) of span station 9; Directed along the xb2-axis
ElastoDyn['Spn9TDyb2']        = False     # (m); Blade 2 local edgewise (translational) deflection (relative to the undeflected position) of span station 9; Directed along the yb2-axis
ElastoDyn['Spn9TDzb2']        = False     # (m); Blade 2 local axial (translational) deflection (relative to the undeflected position) of span station 9; Directed along the zb2-axis
ElastoDyn['Spn1RDxb2']        = False     # (deg); Blade 2 local roll (angular/rotational)  deflection (relative to the undeflected position) of span station 1. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local xb2-axis
ElastoDyn['Spn1RDyb2']        = False     # (deg); Blade 2 local pitch (angular/rotational) deflection (relative to the undeflected position) of span station 1. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local yb2-axis
ElastoDyn['Spn1RDzb2']        = False     # (deg); Blade 2 local torsional (angular/rotational) deflection (relative to the undeflected position) of span station 1. This output will always be zero for FAST simulation results. Use it for examining blade torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.  Please note that this output uses the opposite of the sign convention used for blade pitch angles.; About the local zb2-axis
ElastoDyn['Spn2RDxb2']        = False     # (deg); Blade 2 local roll (angular/rotational)  deflection (relative to the undeflected position) of span station 2. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local xb2-axis
ElastoDyn['Spn2RDyb2']        = False     # (deg); Blade 2 local pitch (angular/rotational) deflection (relative to the undeflected position) of span station 2. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local yb2-axis
ElastoDyn['Spn2RDzb2']        = False     # (deg); Blade 2 local torsional (angular/rotational) deflection (relative to the undeflected position) of span station 2. This output will always be zero for FAST simulation results. Use it for examining blade torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.  Please note that this output uses the opposite of the sign convention used for blade pitch angles.; About the local zb2-axis
ElastoDyn['Spn3RDxb2']        = False     # (deg); Blade 2 local roll (angular/rotational)  deflection (relative to the undeflected position) of span station 3. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local xb2-axis
ElastoDyn['Spn3RDyb2']        = False     # (deg); Blade 2 local pitch (angular/rotational) deflection (relative to the undeflected position) of span station 3. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local yb2-axis
ElastoDyn['Spn3RDzb2']        = False     # (deg); Blade 2 local torsional (angular/rotational) deflection (relative to the undeflected position) of span station 3. This output will always be zero for FAST simulation results. Use it for examining blade torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.  Please note that this output uses the opposite of the sign convention used for blade pitch angles.; About the local zb2-axis
ElastoDyn['Spn4RDxb2']        = False     # (deg); Blade 2 local roll (angular/rotational)  deflection (relative to the undeflected position) of span station 4. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local xb2-axis
ElastoDyn['Spn4RDyb2']        = False     # (deg); Blade 2 local pitch (angular/rotational) deflection (relative to the undeflected position) of span station 4. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local yb2-axis
ElastoDyn['Spn4RDzb2']        = False     # (deg); Blade 2 local torsional (angular/rotational) deflection (relative to the undeflected position) of span station 4. This output will always be zero for FAST simulation results. Use it for examining blade torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.  Please note that this output uses the opposite of the sign convention used for blade pitch angles.; About the local zb2-axis
ElastoDyn['Spn5RDxb2']        = False     # (deg); Blade 2 local roll (angular/rotational)  deflection (relative to the undeflected position) of span station 5. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local xb2-axis
ElastoDyn['Spn5RDyb2']        = False     # (deg); Blade 2 local pitch (angular/rotational) deflection (relative to the undeflected position) of span station 5. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local yb2-axis
ElastoDyn['Spn5RDzb2']        = False     # (deg); Blade 2 local torsional (angular/rotational) deflection (relative to the undeflected position) of span station 5. This output will always be zero for FAST simulation results. Use it for examining blade torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.  Please note that this output uses the opposite of the sign convention used for blade pitch angles.; About the local zb2-axis
ElastoDyn['Spn6RDxb2']        = False     # (deg); Blade 2 local roll (angular/rotational)  deflection (relative to the undeflected position) of span station 6. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local xb2-axis
ElastoDyn['Spn6RDyb2']        = False     # (deg); Blade 2 local pitch (angular/rotational) deflection (relative to the undeflected position) of span station 6. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local yb2-axis
ElastoDyn['Spn6RDzb2']        = False     # (deg); Blade 2 local torsional (angular/rotational) deflection (relative to the undeflected position) of span station 6. This output will always be zero for FAST simulation results. Use it for examining blade torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.  Please note that this output uses the opposite of the sign convention used for blade pitch angles.; About the local zb2-axis
ElastoDyn['Spn7RDxb2']        = False     # (deg); Blade 2 local roll (angular/rotational)  deflection (relative to the undeflected position) of span station 7. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local xb2-axis
ElastoDyn['Spn7RDyb2']        = False     # (deg); Blade 2 local pitch (angular/rotational) deflection (relative to the undeflected position) of span station 7. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local yb2-axis
ElastoDyn['Spn7RDzb2']        = False     # (deg); Blade 2 local torsional (angular/rotational) deflection (relative to the undeflected position) of span station 7. This output will always be zero for FAST simulation results. Use it for examining blade torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.  Please note that this output uses the opposite of the sign convention used for blade pitch angles.; About the local zb2-axis
ElastoDyn['Spn8RDxb2']        = False     # (deg); Blade 2 local roll (angular/rotational)  deflection (relative to the undeflected position) of span station 8. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local xb2-axis
ElastoDyn['Spn8RDyb2']        = False     # (deg); Blade 2 local pitch (angular/rotational) deflection (relative to the undeflected position) of span station 8. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local yb2-axis
ElastoDyn['Spn8RDzb2']        = False     # (deg); Blade 2 local torsional (angular/rotational) deflection (relative to the undeflected position) of span station 8. This output will always be zero for FAST simulation results. Use it for examining blade torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.  Please note that this output uses the opposite of the sign convention used for blade pitch angles.; About the local zb2-axis
ElastoDyn['Spn9RDxb2']        = False     # (deg); Blade 2 local roll (angular/rotational)  deflection (relative to the undeflected position) of span station 9. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local xb2-axis
ElastoDyn['Spn9RDyb2']        = False     # (deg); Blade 2 local pitch (angular/rotational) deflection (relative to the undeflected position) of span station 9. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local yb2-axis
ElastoDyn['Spn9RDzb2']        = False     # (deg); Blade 2 local torsional (angular/rotational) deflection (relative to the undeflected position) of span station 9. This output will always be zero for FAST simulation results. Use it for examining blade torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.  Please note that this output uses the opposite of the sign convention used for blade pitch angles.; About the local zb2-axis

# Blade 3 Local Span Motions
ElastoDyn['Spn1ALxb3']        = False     # (m/s^2); Blade 3 local flapwise acceleration (absolute) of span station 1; Directed along the local xb3-axis
ElastoDyn['Spn1ALyb3']        = False     # (m/s^2); Blade 3 local edgewise acceleration (absolute) of span station 1; Directed along the local yb3-axis
ElastoDyn['Spn1ALzb3']        = False     # (m/s^2); Blade 3 local axial acceleration (absolute) of span station 1; Directed along the local zb3-axis
ElastoDyn['Spn2ALxb3']        = False     # (m/s^2); Blade 3 local flapwise acceleration (absolute) of span station 2; Directed along the local xb3-axis
ElastoDyn['Spn2ALyb3']        = False     # (m/s^2); Blade 3 local edgewise acceleration (absolute) of span station 2; Directed along the local yb3-axis
ElastoDyn['Spn2ALzb3']        = False     # (m/s^2); Blade 3 local axial acceleration (absolute) of span station 2; Directed along the local zb3-axis
ElastoDyn['Spn3ALxb3']        = False     # (m/s^2); Blade 3 local flapwise acceleration (absolute) of span station 3; Directed along the local xb3-axis
ElastoDyn['Spn3ALyb3']        = False     # (m/s^2); Blade 3 local edgewise acceleration (absolute) of span station 3; Directed along the local yb3-axis
ElastoDyn['Spn3ALzb3']        = False     # (m/s^2); Blade 3 local axial acceleration (absolute) of span station 3; Directed along the local zb3-axis
ElastoDyn['Spn4ALxb3']        = False     # (m/s^2); Blade 3 local flapwise acceleration (absolute) of span station 4; Directed along the local xb3-axis
ElastoDyn['Spn4ALyb3']        = False     # (m/s^2); Blade 3 local edgewise acceleration (absolute) of span station 4; Directed along the local yb3-axis
ElastoDyn['Spn4ALzb3']        = False     # (m/s^2); Blade 3 local axial acceleration (absolute) of span station 4; Directed along the local zb3-axis
ElastoDyn['Spn5ALxb3']        = False     # (m/s^2); Blade 3 local flapwise acceleration (absolute) of span station 5; Directed along the local xb3-axis
ElastoDyn['Spn5ALyb3']        = False     # (m/s^2); Blade 3 local edgewise acceleration (absolute) of span station 5; Directed along the local yb3-axis
ElastoDyn['Spn5ALzb3']        = False     # (m/s^2); Blade 3 local axial acceleration (absolute) of span station 5; Directed along the local zb3-axis
ElastoDyn['Spn6ALxb3']        = False     # (m/s^2); Blade 3 local flapwise acceleration (absolute) of span station 6; Directed along the local xb3-axis
ElastoDyn['Spn6ALyb3']        = False     # (m/s^2); Blade 3 local edgewise acceleration (absolute) of span station 6; Directed along the local yb3-axis
ElastoDyn['Spn6ALzb3']        = False     # (m/s^2); Blade 3 local axial acceleration (absolute) of span station 6; Directed along the local zb3-axis
ElastoDyn['Spn7ALxb3']        = False     # (m/s^2); Blade 3 local flapwise acceleration (absolute) of span station 7; Directed along the local xb3-axis
ElastoDyn['Spn7ALyb3']        = False     # (m/s^2); Blade 3 local edgewise acceleration (absolute) of span station 7; Directed along the local yb3-axis
ElastoDyn['Spn7ALzb3']        = False     # (m/s^2); Blade 3 local axial acceleration (absolute) of span station 7; Directed along the local zb3-axis
ElastoDyn['Spn8ALxb3']        = False     # (m/s^2); Blade 3 local flapwise acceleration (absolute) of span station 8; Directed along the local xb3-axis
ElastoDyn['Spn8ALyb3']        = False     # (m/s^2); Blade 3 local edgewise acceleration (absolute) of span station 8; Directed along the local yb3-axis
ElastoDyn['Spn8ALzb3']        = False     # (m/s^2); Blade 3 local axial acceleration (absolute) of span station 8; Directed along the local zb3-axis
ElastoDyn['Spn9ALxb3']        = False     # (m/s^2); Blade 3 local flapwise acceleration (absolute) of span station 9; Directed along the local xb3-axis
ElastoDyn['Spn9ALyb3']        = False     # (m/s^2); Blade 3 local edgewise acceleration (absolute) of span station 9; Directed along the local yb3-axis
ElastoDyn['Spn9ALzb3']        = False     # (m/s^2); Blade 3 local axial acceleration (absolute) of span station 9; Directed along the local zb3-axis
ElastoDyn['Spn1TDxb3']        = False     # (m); Blade 3 local flapwise (translational) deflection (relative to the undeflected position) of span station 1; Directed along the xb3-axis
ElastoDyn['Spn1TDyb3']        = False     # (m); Blade 3 local edgewise (translational) deflection (relative to the undeflected position) of span station 1; Directed along the yb3-axis
ElastoDyn['Spn1TDzb3']        = False     # (m); Blade 3 local axial (translational) deflection (relative to the undeflected position) of span station 1; Directed along the zb3-axis
ElastoDyn['Spn2TDxb3']        = False     # (m); Blade 3 local flapwise (translational) deflection (relative to the undeflected position) of span station 2; Directed along the xb3-axis
ElastoDyn['Spn2TDyb3']        = False     # (m); Blade 3 local edgewise (translational) deflection (relative to the undeflected position) of span station 2; Directed along the yb3-axis
ElastoDyn['Spn2TDzb3']        = False     # (m); Blade 3 local axial (translational) deflection (relative to the undeflected position) of span station 2; Directed along the zb3-axis
ElastoDyn['Spn3TDxb3']        = False     # (m); Blade 3 local flapwise (translational) deflection (relative to the undeflected position) of span station 3; Directed along the xb3-axis
ElastoDyn['Spn3TDyb3']        = False     # (m); Blade 3 local edgewise (translational) deflection (relative to the undeflected position) of span station 3; Directed along the yb3-axis
ElastoDyn['Spn3TDzb3']        = False     # (m); Blade 3 local axial (translational) deflection (relative to the undeflected position) of span station 3; Directed along the zb3-axis
ElastoDyn['Spn4TDxb3']        = False     # (m); Blade 3 local flapwise (translational) deflection (relative to the undeflected position) of span station 4; Directed along the xb3-axis
ElastoDyn['Spn4TDyb3']        = False     # (m); Blade 3 local edgewise (translational) deflection (relative to the undeflected position) of span station 4; Directed along the yb3-axis
ElastoDyn['Spn4TDzb3']        = False     # (m); Blade 3 local axial (translational) deflection (relative to the undeflected position) of span station 4; Directed along the zb3-axis
ElastoDyn['Spn5TDxb3']        = False     # (m); Blade 3 local flapwise (translational) deflection (relative to the undeflected position) of span station 5; Directed along the xb3-axis
ElastoDyn['Spn5TDyb3']        = False     # (m); Blade 3 local edgewise (translational) deflection (relative to the undeflected position) of span station 5; Directed along the yb3-axis
ElastoDyn['Spn5TDzb3']        = False     # (m); Blade 3 local axial (translational) deflection (relative to the undeflected position) of span station 5; Directed along the zb3-axis
ElastoDyn['Spn6TDxb3']        = False     # (m); Blade 3 local flapwise (translational) deflection (relative to the undeflected position) of span station 6; Directed along the xb3-axis
ElastoDyn['Spn6TDyb3']        = False     # (m); Blade 3 local edgewise (translational) deflection (relative to the undeflected position) of span station 6; Directed along the yb3-axis
ElastoDyn['Spn6TDzb3']        = False     # (m); Blade 3 local axial (translational) deflection (relative to the undeflected position) of span station 6; Directed along the zb3-axis
ElastoDyn['Spn7TDxb3']        = False     # (m); Blade 3 local flapwise (translational) deflection (relative to the undeflected position) of span station 7; Directed along the xb3-axis
ElastoDyn['Spn7TDyb3']        = False     # (m); Blade 3 local edgewise (translational) deflection (relative to the undeflected position) of span station 7; Directed along the yb3-axis
ElastoDyn['Spn7TDzb3']        = False     # (m); Blade 3 local axial (translational) deflection (relative to the undeflected position) of span station 7; Directed along the zb3-axis
ElastoDyn['Spn8TDxb3']        = False     # (m); Blade 3 local flapwise (translational) deflection (relative to the undeflected position) of span station 8; Directed along the xb3-axis
ElastoDyn['Spn8TDyb3']        = False     # (m); Blade 3 local edgewise (translational) deflection (relative to the undeflected position) of span station 8; Directed along the yb3-axis
ElastoDyn['Spn8TDzb3']        = False     # (m); Blade 3 local axial (translational) deflection (relative to the undeflected position) of span station 8; Directed along the zb3-axis
ElastoDyn['Spn9TDxb3']        = False     # (m); Blade 3 local flapwise (translational) deflection (relative to the undeflected position) of span station 9; Directed along the xb3-axis
ElastoDyn['Spn9TDyb3']        = False     # (m); Blade 3 local edgewise (translational) deflection (relative to the undeflected position) of span station 9; Directed along the yb3-axis
ElastoDyn['Spn9TDzb3']        = False     # (m); Blade 3 local axial (translational) deflection (relative to the undeflected position) of span station 9; Directed along the zb3-axis
ElastoDyn['Spn1RDxb3']        = False     # (deg); Blade 3 local roll (angular/rotational)  deflection (relative to the undeflected position) of span station 1. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local xb3-axis
ElastoDyn['Spn1RDyb3']        = False     # (deg); Blade 3 local pitch (angular/rotational) deflection (relative to the undeflected position) of span station 1. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local yb3-axis
ElastoDyn['Spn1RDzb3']        = False     # (deg); Blade 3 local torsional (angular/rotational) deflection (relative to the undeflected position) of span station 1. This output will always be zero for FAST simulation results. Use it for examining blade torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.  Please note that this output uses the opposite of the sign convention used for blade pitch angles.; About the local zb3-axis
ElastoDyn['Spn2RDxb3']        = False     # (deg); Blade 3 local roll (angular/rotational)  deflection (relative to the undeflected position) of span station 2. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local xb3-axis
ElastoDyn['Spn2RDyb3']        = False     # (deg); Blade 3 local pitch (angular/rotational) deflection (relative to the undeflected position) of span station 2. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local yb3-axis
ElastoDyn['Spn2RDzb3']        = False     # (deg); Blade 3 local torsional (angular/rotational) deflection (relative to the undeflected position) of span station 2. This output will always be zero for FAST simulation results. Use it for examining blade torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.  Please note that this output uses the opposite of the sign convention used for blade pitch angles.; About the local zb3-axis
ElastoDyn['Spn3RDxb3']        = False     # (deg); Blade 3 local roll (angular/rotational)  deflection (relative to the undeflected position) of span station 3. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local xb3-axis
ElastoDyn['Spn3RDyb3']        = False     # (deg); Blade 3 local pitch (angular/rotational) deflection (relative to the undeflected position) of span station 3. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local yb3-axis
ElastoDyn['Spn3RDzb3']        = False     # (deg); Blade 3 local torsional (angular/rotational) deflection (relative to the undeflected position) of span station 3. This output will always be zero for FAST simulation results. Use it for examining blade torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.  Please note that this output uses the opposite of the sign convention used for blade pitch angles.; About the local zb3-axis
ElastoDyn['Spn4RDxb3']        = False     # (deg); Blade 3 local roll (angular/rotational)  deflection (relative to the undeflected position) of span station 4. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local xb3-axis
ElastoDyn['Spn4RDyb3']        = False     # (deg); Blade 3 local pitch (angular/rotational) deflection (relative to the undeflected position) of span station 4. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local yb3-axis
ElastoDyn['Spn4RDzb3']        = False     # (deg); Blade 3 local torsional (angular/rotational) deflection (relative to the undeflected position) of span station 4. This output will always be zero for FAST simulation results. Use it for examining blade torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.  Please note that this output uses the opposite of the sign convention used for blade pitch angles.; About the local zb3-axis
ElastoDyn['Spn5RDxb3']        = False     # (deg); Blade 3 local roll (angular/rotational)  deflection (relative to the undeflected position) of span station 5. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local xb3-axis
ElastoDyn['Spn5RDyb3']        = False     # (deg); Blade 3 local pitch (angular/rotational) deflection (relative to the undeflected position) of span station 5. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local yb3-axis
ElastoDyn['Spn5RDzb3']        = False     # (deg); Blade 3 local torsional (angular/rotational) deflection (relative to the undeflected position) of span station 5. This output will always be zero for FAST simulation results. Use it for examining blade torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.  Please note that this output uses the opposite of the sign convention used for blade pitch angles.; About the local zb3-axis
ElastoDyn['Spn6RDxb3']        = False     # (deg); Blade 3 local roll (angular/rotational)  deflection (relative to the undeflected position) of span station 6. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local xb3-axis
ElastoDyn['Spn6RDyb3']        = False     # (deg); Blade 3 local pitch (angular/rotational) deflection (relative to the undeflected position) of span station 6. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local yb3-axis
ElastoDyn['Spn6RDzb3']        = False     # (deg); Blade 3 local torsional (angular/rotational) deflection (relative to the undeflected position) of span station 6. This output will always be zero for FAST simulation results. Use it for examining blade torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.  Please note that this output uses the opposite of the sign convention used for blade pitch angles.; About the local zb3-axis
ElastoDyn['Spn7RDxb3']        = False     # (deg); Blade 3 local roll (angular/rotational)  deflection (relative to the undeflected position) of span station 7. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local xb3-axis
ElastoDyn['Spn7RDyb3']        = False     # (deg); Blade 3 local pitch (angular/rotational) deflection (relative to the undeflected position) of span station 7. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local yb3-axis
ElastoDyn['Spn7RDzb3']        = False     # (deg); Blade 3 local torsional (angular/rotational) deflection (relative to the undeflected position) of span station 7. This output will always be zero for FAST simulation results. Use it for examining blade torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.  Please note that this output uses the opposite of the sign convention used for blade pitch angles.; About the local zb3-axis
ElastoDyn['Spn8RDxb3']        = False     # (deg); Blade 3 local roll (angular/rotational)  deflection (relative to the undeflected position) of span station 8. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local xb3-axis
ElastoDyn['Spn8RDyb3']        = False     # (deg); Blade 3 local pitch (angular/rotational) deflection (relative to the undeflected position) of span station 8. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local yb3-axis
ElastoDyn['Spn8RDzb3']        = False     # (deg); Blade 3 local torsional (angular/rotational) deflection (relative to the undeflected position) of span station 8. This output will always be zero for FAST simulation results. Use it for examining blade torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.  Please note that this output uses the opposite of the sign convention used for blade pitch angles.; About the local zb3-axis
ElastoDyn['Spn9RDxb3']        = False     # (deg); Blade 3 local roll (angular/rotational)  deflection (relative to the undeflected position) of span station 9. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local xb3-axis
ElastoDyn['Spn9RDyb3']        = False     # (deg); Blade 3 local pitch (angular/rotational) deflection (relative to the undeflected position) of span station 9. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local yb3-axis
ElastoDyn['Spn9RDzb3']        = False     # (deg); Blade 3 local torsional (angular/rotational) deflection (relative to the undeflected position) of span station 9. This output will always be zero for FAST simulation results. Use it for examining blade torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.  Please note that this output uses the opposite of the sign convention used for blade pitch angles.; About the local zb3-axis

# Blade Pitch Motions
ElastoDyn['PtchPMzc1']        = False     # (deg); Blade 1 pitch angle (position); Positive towards feather about the minus zc1- and minus zb1-axes
ElastoDyn['PtchPMzb1']        = False     # (deg); Blade 1 pitch angle (position); Positive towards feather about the minus zc1- and minus zb1-axes
ElastoDyn['BldPitch1']        = False     # (deg); Blade 1 pitch angle (position); Positive towards feather about the minus zc1- and minus zb1-axes
ElastoDyn['BlPitch1']         = False     # (deg); Blade 1 pitch angle (position); Positive towards feather about the minus zc1- and minus zb1-axes
ElastoDyn['PtchPMzc2']        = False     # (deg); Blade 2 pitch angle (position); Positive towards feather about the minus zc2- and minus zb2-axes
ElastoDyn['PtchPMzb2']        = False     # (deg); Blade 2 pitch angle (position); Positive towards feather about the minus zc2- and minus zb2-axes
ElastoDyn['BldPitch2']        = False     # (deg); Blade 2 pitch angle (position); Positive towards feather about the minus zc2- and minus zb2-axes
ElastoDyn['BlPitch2']         = False     # (deg); Blade 2 pitch angle (position); Positive towards feather about the minus zc2- and minus zb2-axes
ElastoDyn['PtchPMzc3']        = False     # (deg); Blade 3 pitch angle (position); Positive towards feather about the minus zc3- and minus zb3-axes
ElastoDyn['PtchPMzb3']        = False     # (deg); Blade 3 pitch angle (position); Positive towards feather about the minus zc3- and minus zb3-axes
ElastoDyn['BldPitch3']        = False     # (deg); Blade 3 pitch angle (position); Positive towards feather about the minus zc3- and minus zb3-axes
ElastoDyn['BlPitch3']         = False     # (deg); Blade 3 pitch angle (position); Positive towards feather about the minus zc3- and minus zb3-axes

# Teeter Motions
ElastoDyn['TeetPya']          = False     # (deg); Rotor teeter angle (position); About the ya-axis
ElastoDyn['RotTeetP']         = False     # (deg); Rotor teeter angle (position); About the ya-axis
ElastoDyn['TeetDefl']         = False     # (deg); Rotor teeter angle (position); About the ya-axis
ElastoDyn['TeetVya']          = False     # (deg/s); Rotor teeter angular velocity; About the ya-axis
ElastoDyn['RotTeetV']         = False     # (deg/s); Rotor teeter angular velocity; About the ya-axis
ElastoDyn['TeetAya']          = False     # (deg/s^2); Rotor teeter angular acceleration; About the ya-axis
ElastoDyn['RotTeetA']         = False     # (deg/s^2); Rotor teeter angular acceleration; About the ya-axis

# Shaft Motions
ElastoDyn['LSSTipPxa']        = False     # (deg); Rotor azimuth angle (position); About the xa- and xs-axes
ElastoDyn['LSSTipPxs']        = False     # (deg); Rotor azimuth angle (position); About the xa- and xs-axes
ElastoDyn['LSSTipP']          = False     # (deg); Rotor azimuth angle (position); About the xa- and xs-axes
ElastoDyn['Azimuth']          = False     # (deg); Rotor azimuth angle (position); About the xa- and xs-axes
ElastoDyn['LSSTipVxa']        = False     # (rpm); Rotor azimuth angular speed; About the xa- and xs-axes
ElastoDyn['LSSTipVxs']        = False     # (rpm); Rotor azimuth angular speed; About the xa- and xs-axes
ElastoDyn['LSSTipV']          = False     # (rpm); Rotor azimuth angular speed; About the xa- and xs-axes
ElastoDyn['RotSpeed']         = False     # (rpm); Rotor azimuth angular speed; About the xa- and xs-axes
ElastoDyn['LSSTipAxa']        = False     # (deg/s^2); Rotor azimuth angular acceleration; About the xa- and xs-axes
ElastoDyn['LSSTipAxs']        = False     # (deg/s^2); Rotor azimuth angular acceleration; About the xa- and xs-axes
ElastoDyn['LSSTipA']          = False     # (deg/s^2); Rotor azimuth angular acceleration; About the xa- and xs-axes
ElastoDyn['RotAccel']         = False     # (deg/s^2); Rotor azimuth angular acceleration; About the xa- and xs-axes
ElastoDyn['LSSGagPxa']        = False     # (deg); Low-speed shaft strain gage azimuth angle (position) (on the gearbox side of the low-speed shaft); About the xa- and xs-axes
ElastoDyn['LSSGagPxs']        = False     # (deg); Low-speed shaft strain gage azimuth angle (position) (on the gearbox side of the low-speed shaft); About the xa- and xs-axes
ElastoDyn['LSSGagP']          = False     # (deg); Low-speed shaft strain gage azimuth angle (position) (on the gearbox side of the low-speed shaft); About the xa- and xs-axes
ElastoDyn['LSSGagVxa']        = False     # (rpm); Low-speed shaft strain gage angular speed (on the gearbox side of the low-speed shaft); About the xa- and xs-axes
ElastoDyn['LSSGagVxs']        = False     # (rpm); Low-speed shaft strain gage angular speed (on the gearbox side of the low-speed shaft); About the xa- and xs-axes
ElastoDyn['LSSGagV']          = False     # (rpm); Low-speed shaft strain gage angular speed (on the gearbox side of the low-speed shaft); About the xa- and xs-axes
ElastoDyn['LSSGagAxa']        = False     # (deg/s^2); Low-speed shaft strain gage angular acceleration (on the gearbox side of the low-speed shaft); About the xa- and xs-axes
ElastoDyn['LSSGagAxs']        = False     # (deg/s^2); Low-speed shaft strain gage angular acceleration (on the gearbox side of the low-speed shaft); About the xa- and xs-axes
ElastoDyn['LSSGagA']          = False     # (deg/s^2); Low-speed shaft strain gage angular acceleration (on the gearbox side of the low-speed shaft); About the xa- and xs-axes
ElastoDyn['HSShftV']          = False     # (rpm); Angular speed of the high-speed shaft and generator; Same sign as LSSGagVxa / LSSGagVxs / LSSGagV
ElastoDyn['GenSpeed']         = False     # (rpm); Angular speed of the high-speed shaft and generator; Same sign as LSSGagVxa / LSSGagVxs / LSSGagV
ElastoDyn['HSShftA']          = False     # (deg/s^2); Angular acceleration of the high-speed shaft and generator; Same sign as LSSGagAxa / LSSGagAxs / LSSGagA
ElastoDyn['GenAccel']         = False     # (deg/s^2); Angular acceleration of the high-speed shaft and generator; Same sign as LSSGagAxa / LSSGagAxs / LSSGagA

# Nacelle IMU Motions
ElastoDyn['NcIMUTVxs']        = False     # (m/s); Nacelle inertial measurement unit translational velocity (absolute); Directed along the xs-axis
ElastoDyn['NcIMUTVys']        = False     # (m/s); Nacelle inertial measurement unit translational velocity (absolute); Directed along the ys-axis
ElastoDyn['NcIMUTVzs']        = False     # (m/s); Nacelle inertial measurement unit translational velocity (absolute); Directed along the zs-axis
ElastoDyn['NcIMUTAxs']        = False     # (m/s^2); Nacelle inertial measurement unit translational acceleration (absolute); Directed along the xs-axis
ElastoDyn['NcIMUTAys']        = False     # (m/s^2); Nacelle inertial measurement unit translational acceleration (absolute); Directed along the ys-axis
ElastoDyn['NcIMUTAzs']        = False     # (m/s^2); Nacelle inertial measurement unit translational acceleration (absolute); Directed along the zs-axis
ElastoDyn['NcIMURVxs']        = False     # (deg/s); Nacelle inertial measurement unit angular (rotational) velocity (absolute); About the xs-axis
ElastoDyn['NcIMURVys']        = False     # (deg/s); Nacelle inertial measurement unit angular (rotational) velocity (absolute); About the ys-axis
ElastoDyn['NcIMURVzs']        = False     # (deg/s); Nacelle inertial measurement unit angular (rotational) velocity (absolute); About the zs-axis
ElastoDyn['NcIMURAxs']        = False     # (deg/s^2); Nacelle inertial measurement unit angular (rotational) acceleration (absolute); About the xs-axis
ElastoDyn['NcIMURAys']        = False     # (deg/s^2); Nacelle inertial measurement unit angular (rotational) acceleration (absolute); About the ys-axis
ElastoDyn['NcIMURAzs']        = False     # (deg/s^2); Nacelle inertial measurement unit angular (rotational) acceleration (absolute); About the zs-axis

# Rotor-Furl Motions
ElastoDyn['RotFurlP']         = False     # (deg); Rotor-furl angle (position); About the rotor-furl axis
ElastoDyn['RotFurl']          = False     # (deg); Rotor-furl angle (position); About the rotor-furl axis
ElastoDyn['RotFurlV']         = False     # (deg/s); Rotor-furl angular velocity; About the rotor-furl axis
ElastoDyn['RotFurlA']         = False     # (deg/s^2); Rotor-furl angular acceleration; About the rotor-furl axis

# Tail-Furl Motions
ElastoDyn['TailFurlP']        = False     # (deg); Tail-furl angle (position); About the tail-furl axis
ElastoDyn['TailFurl']         = False     # (deg); Tail-furl angle (position); About the tail-furl axis
ElastoDyn['TailFurlV']        = False     # (deg/s); Tail-furl angular velocity; About the tail-furl axis
ElastoDyn['TailFurlA']        = False     # (deg/s^2); Tail-furl angular acceleration; About the tail-furl axis

# Nacelle Yaw Motions
ElastoDyn['YawPzn']           = False     # (deg); Nacelle yaw angle (position); About the zn- and zp-axes
ElastoDyn['YawPzp']           = False     # (deg); Nacelle yaw angle (position); About the zn- and zp-axes
ElastoDyn['NacYawP']          = False     # (deg); Nacelle yaw angle (position); About the zn- and zp-axes
ElastoDyn['NacYaw']           = False     # (deg); Nacelle yaw angle (position); About the zn- and zp-axes
ElastoDyn['YawPos']           = False     # (deg); Nacelle yaw angle (position); About the zn- and zp-axes
ElastoDyn['YawVzn']           = False     # (deg/s); Nacelle yaw angular velocity; About the zn- and zp-axes
ElastoDyn['YawVzp']           = False     # (deg/s); Nacelle yaw angular velocity; About the zn- and zp-axes
ElastoDyn['NacYawV']          = False     # (deg/s); Nacelle yaw angular velocity; About the zn- and zp-axes
ElastoDyn['YawRate']          = False     # (deg/s); Nacelle yaw angular velocity; About the zn- and zp-axes
ElastoDyn['YawAzn']           = False     # (deg/s^2); Nacelle yaw angular acceleration; About the zn- and zp-axes
ElastoDyn['YawAzp']           = False     # (deg/s^2); Nacelle yaw angular acceleration; About the zn- and zp-axes
ElastoDyn['NacYawA']          = False     # (deg/s^2); Nacelle yaw angular acceleration; About the zn- and zp-axes
ElastoDyn['YawAccel']         = False     # (deg/s^2); Nacelle yaw angular acceleration; About the zn- and zp-axes

# Tower-Top / Yaw Bearing Motions
ElastoDyn['YawBrTDxp']        = False     # (m); Tower-top / yaw bearing fore-aft (translational) deflection (relative to the undeflected position); Directed along the xp-axis
ElastoDyn['YawBrTDyp']        = False     # (m); Tower-top / yaw bearing side-to-side (translational) deflection (relative to the undeflected position); Directed along the yp-axis
ElastoDyn['YawBrTDzp']        = False     # (m); Tower-top / yaw bearing axial (translational) deflection (relative to the undeflected position); Directed along the zp-axis
ElastoDyn['YawBrTDxt']        = False     # (m); Tower-top / yaw bearing fore-aft (translational) deflection (relative to the undeflected position); Directed along the xt-axis
ElastoDyn['TTDspFA']          = False     # (m); Tower-top / yaw bearing fore-aft (translational) deflection (relative to the undeflected position); Directed along the xt-axis
ElastoDyn['YawBrTDyt']        = False     # (m); Tower-top / yaw bearing side-to-side (translation) deflection (relative to the undeflected position); Directed along the yt-axis
ElastoDyn['TTDspSS']          = False     # (m); Tower-top / yaw bearing side-to-side (translation) deflection (relative to the undeflected position); Directed along the yt-axis
ElastoDyn['YawBrTDzt']        = False     # (m); Tower-top / yaw bearing axial (translational) deflection (relative to the undeflected position); Directed along the zt-axis
ElastoDyn['TTDspAx']          = False     # (m); Tower-top / yaw bearing axial (translational) deflection (relative to the undeflected position); Directed along the zt-axis
ElastoDyn['YawBrTAxp']        = False     # (m/s^2); Tower-top / yaw bearing fore-aft (translational) acceleration (absolute); Directed along the xp-axis
ElastoDyn['YawBrTAyp']        = False     # (m/s^2); Tower-top / yaw bearing side-to-side (translational) acceleration (absolute); Directed along the yp-axis
ElastoDyn['YawBrTAzp']        = False     # (m/s^2); Tower-top / yaw bearing axial (translational) acceleration (absolute); Directed along the zp-axis
ElastoDyn['YawBrRDxt']        = False     # (deg); Tower-top / yaw bearing angular (rotational) roll deflection (relative to the undeflected position). In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower deflections, so that the rotation sequence does not matter.; About the xt-axis
ElastoDyn['TTDspRoll']        = False     # (deg); Tower-top / yaw bearing angular (rotational) roll deflection (relative to the undeflected position). In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower deflections, so that the rotation sequence does not matter.; About the xt-axis
ElastoDyn['YawBrRDyt']        = False     # (deg); Tower-top / yaw bearing angular (rotational) pitch deflection (relative to the undeflected position). In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower deflections, so that the rotation sequence does not matter.; About the yt-axis
ElastoDyn['TTDspPtch']        = False     # (deg); Tower-top / yaw bearing angular (rotational) pitch deflection (relative to the undeflected position). In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower deflections, so that the rotation sequence does not matter.; About the yt-axis
ElastoDyn['YawBrRDzt']        = False     # (deg); Tower-top / yaw bearing angular (rotational) torsion deflection (relative to the undeflected position). This output will always be zero for FAST simulation results. Use it for examining tower torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.; About the zt-axis
ElastoDyn['TTDspTwst']        = False     # (deg); Tower-top / yaw bearing angular (rotational) torsion deflection (relative to the undeflected position). This output will always be zero for FAST simulation results. Use it for examining tower torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.; About the zt-axis
ElastoDyn['YawBrRVxp']        = False     # (deg/s); Tower-top / yaw bearing angular (rotational) roll velocity (absolute); About the xp-axis
ElastoDyn['YawBrRVyp']        = False     # (deg/s); Tower-top / yaw bearing angular (rotational) pitch velocity (absolute); About the yp-axis
ElastoDyn['YawBrRVzp']        = False     # (deg/s); Tower-top / yaw bearing angular (rotational) torsion velocity. This output will always be very close to zero for FAST simulation results. Use it for examining tower torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. (absolute); About the zp-axis
ElastoDyn['YawBrRAxp']        = False     # (deg/s^2); Tower-top / yaw bearing angular (rotational) roll acceleration (absolute); About the xp-axis
ElastoDyn['YawBrRAyp']        = False     # (deg/s^2); Tower-top / yaw bearing angular (rotational) pitch acceleration (absolute); About the yp-axis
ElastoDyn['YawBrRAzp']        = False     # (deg/s^2); Tower-top / yaw bearing angular (rotational) torsion acceleration. This output will always be very close to zero for FAST simulation results. Use it for examining tower torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. (absolute); About the zp-axis

# Local Tower Motions
ElastoDyn['TwHt1ALxt']        = False     # (m/s^2); Local tower fore-aft (translational) acceleration (absolute) of tower gage 1 ; Directed along the local xt-axis
ElastoDyn['TwHt1ALyt']        = False     # (m/s^2); Local tower side-to-side (translational) acceleration (absolute) of tower gage 1 ; Directed along the local yt-axis
ElastoDyn['TwHt1ALzt']        = False     # (m/s^2); Local tower axial (translational) acceleration (absolute) of tower gage 1 ; Directed along the local zt-axis
ElastoDyn['TwHt2ALxt']        = False     # (m/s^2); Local tower fore-aft (translational) acceleration (absolute) of tower gage 2; Directed along the local xt-axis
ElastoDyn['TwHt2ALyt']        = False     # (m/s^2); Local tower side-to-side (translational) acceleration (absolute) of tower gage 2; Directed along the local yt-axis
ElastoDyn['TwHt2ALzt']        = False     # (m/s^2); Local tower axial (translational) acceleration (absolute) of tower gage 2; Directed along the local zt-axis
ElastoDyn['TwHt3ALxt']        = False     # (m/s^2); Local tower fore-aft (translational) acceleration (absolute) of tower gage 3; Directed along the local xt-axis
ElastoDyn['TwHt3ALyt']        = False     # (m/s^2); Local tower side-to-side (translational) acceleration (absolute) of tower gage 3; Directed along the local yt-axis
ElastoDyn['TwHt3ALzt']        = False     # (m/s^2); Local tower axial (translational) acceleration (absolute) of tower gage 3; Directed along the local zt-axis
ElastoDyn['TwHt4ALxt']        = False     # (m/s^2); Local tower fore-aft (translational) acceleration (absolute) of tower gage 4; Directed along the local xt-axis
ElastoDyn['TwHt4ALyt']        = False     # (m/s^2); Local tower side-to-side (translational) acceleration (absolute) of tower gage 4; Directed along the local yt-axis
ElastoDyn['TwHt4ALzt']        = False     # (m/s^2); Local tower axial (translational) acceleration (absolute) of tower gage 4; Directed along the local zt-axis
ElastoDyn['TwHt5ALxt']        = False     # (m/s^2); Local tower fore-aft (translational) acceleration (absolute) of tower gage 5; Directed along the local xt-axis
ElastoDyn['TwHt5ALyt']        = False     # (m/s^2); Local tower side-to-side (translational) acceleration (absolute) of tower gage 5; Directed along the local yt-axis
ElastoDyn['TwHt5ALzt']        = False     # (m/s^2); Local tower axial (translational) acceleration (absolute) of tower gage 5; Directed along the local zt-axis
ElastoDyn['TwHt6ALxt']        = False     # (m/s^2); Local tower fore-aft (translational) acceleration (absolute) of tower gage 6; Directed along the local xt-axis
ElastoDyn['TwHt6ALyt']        = False     # (m/s^2); Local tower side-to-side (translational) acceleration (absolute) of tower gage 6; Directed along the local yt-axis
ElastoDyn['TwHt6ALzt']        = False     # (m/s^2); Local tower axial (translational) acceleration (absolute) of tower gage 6; Directed along the local zt-axis
ElastoDyn['TwHt7ALxt']        = False     # (m/s^2); Local tower fore-aft (translational) acceleration (absolute) of tower gage 7; Directed along the local xt-axis
ElastoDyn['TwHt7ALyt']        = False     # (m/s^2); Local tower side-to-side (translational) acceleration (absolute) of tower gage 7; Directed along the local yt-axis
ElastoDyn['TwHt7ALzt']        = False     # (m/s^2); Local tower axial (translational) acceleration (absolute) of tower gage 7; Directed along the local zt-axis
ElastoDyn['TwHt8ALxt']        = False     # (m/s^2); Local tower fore-aft (translational) acceleration (absolute) of tower gage 8; Directed along the local xt-axis
ElastoDyn['TwHt8ALyt']        = False     # (m/s^2); Local tower side-to-side (translational) acceleration (absolute) of tower gage 8; Directed along the local yt-axis
ElastoDyn['TwHt8ALzt']        = False     # (m/s^2); Local tower axial (translational) acceleration (absolute) of tower gage 8; Directed along the local zt-axis
ElastoDyn['TwHt9ALxt']        = False     # (m/s^2); Local tower fore-aft (translational) acceleration (absolute) of tower gage 9; Directed along the local xt-axis
ElastoDyn['TwHt9ALyt']        = False     # (m/s^2); Local tower side-to-side (translational) acceleration (absolute) of tower gage 9; Directed along the local yt-axis
ElastoDyn['TwHt9ALzt']        = False     # (m/s^2); Local tower axial (translational) acceleration (absolute) of tower gage 9; Directed along the local zt-axis
ElastoDyn['TwHt1TDxt']        = False     # (m); Local tower fore-aft (translational) deflection (relative to the undeflected position) of tower gage 1; Directed along the local xt-axis
ElastoDyn['TwHt1TDyt']        = False     # (m); Local tower side-to-side (translational) deflection (relative to the undeflected position) of tower gage 1; Directed along the local yt-axis
ElastoDyn['TwHt1TDzt']        = False     # (m); Local tower axial (translational) deflection (relative to the undeflected position) of tower gage 1; Directed along the local zt-axis
ElastoDyn['TwHt2TDxt']        = False     # (m); Local tower fore-aft (translational) deflection (relative to the undeflected position) of tower gage 2; Directed along the local xt-axis
ElastoDyn['TwHt2TDyt']        = False     # (m); Local tower side-to-side (translational) deflection (relative to the undeflected position) of tower gage 2; Directed along the local yt-axis
ElastoDyn['TwHt2TDzt']        = False     # (m); Local tower axial (translational) deflection (relative to the undeflected position) of tower gage 2; Directed along the local zt-axis
ElastoDyn['TwHt3TDxt']        = False     # (m); Local tower fore-aft (translational) deflection (relative to the undeflected position) of tower gage 3; Directed along the local xt-axis
ElastoDyn['TwHt3TDyt']        = False     # (m); Local tower side-to-side (translational) deflection (relative to the undeflected position) of tower gage 3; Directed along the local yt-axis
ElastoDyn['TwHt3TDzt']        = False     # (m); Local tower axial (translational) deflection (relative to the undeflected position) of tower gage 3; Directed along the local zt-axis
ElastoDyn['TwHt4TDxt']        = False     # (m); Local tower fore-aft (translational) deflection (relative to the undeflected position) of tower gage 4; Directed along the local xt-axis
ElastoDyn['TwHt4TDyt']        = False     # (m); Local tower side-to-side (translational) deflection (relative to the undeflected position) of tower gage 4; Directed along the local yt-axis
ElastoDyn['TwHt4TDzt']        = False     # (m); Local tower axial (translational) deflection (relative to the undeflected position) of tower gage 4; Directed along the local zt-axis
ElastoDyn['TwHt5TDxt']        = False     # (m); Local tower fore-aft (translational) deflection (relative to the undeflected position) of tower gage 5; Directed along the local xt-axis
ElastoDyn['TwHt5TDyt']        = False     # (m); Local tower side-to-side (translational) deflection (relative to the undeflected position) of tower gage 5; Directed along the local yt-axis
ElastoDyn['TwHt5TDzt']        = False     # (m); Local tower axial (translational) deflection (relative to the undeflected position) of tower gage 5; Directed along the local zt-axis
ElastoDyn['TwHt6TDxt']        = False     # (m); Local tower fore-aft (translational) deflection (relative to the undeflected position) of tower gage 6; Directed along the local xt-axis
ElastoDyn['TwHt6TDyt']        = False     # (m); Local tower side-to-side (translational) deflection (relative to the undeflected position) of tower gage 6; Directed along the local yt-axis
ElastoDyn['TwHt6TDzt']        = False     # (m); Local tower axial (translational) deflection (relative to the undeflected position) of tower gage 6; Directed along the local zt-axis
ElastoDyn['TwHt7TDxt']        = False     # (m); Local tower fore-aft (translational) deflection (relative to the undeflected position) of tower gage 7; Directed along the local xt-axis
ElastoDyn['TwHt7TDyt']        = False     # (m); Local tower side-to-side (translational) deflection (relative to the undeflected position) of tower gage 7; Directed along the local yt-axis
ElastoDyn['TwHt7TDzt']        = False     # (m); Local tower axial (translational) deflection (relative to the undeflected position) of tower gage 7; Directed along the local zt-axis
ElastoDyn['TwHt8TDxt']        = False     # (m); Local tower fore-aft (translational) deflection (relative to the undeflected position) of tower gage 8; Directed along the local xt-axis
ElastoDyn['TwHt8TDyt']        = False     # (m); Local tower side-to-side (translational) deflection (relative to the undeflected position) of tower gage 8; Directed along the local yt-axis
ElastoDyn['TwHt8TDzt']        = False     # (m); Local tower axial (translational) deflection (relative to the undeflected position) of tower gage 8; Directed along the local zt-axis
ElastoDyn['TwHt9TDxt']        = False     # (m); Local tower fore-aft (translational) deflection (relative to the undeflected position) of tower gage 9; Directed along the local xt-axis
ElastoDyn['TwHt9TDyt']        = False     # (m); Local tower side-to-side (translational) deflection (relative to the undeflected position) of tower gage 9; Directed along the local yt-axis
ElastoDyn['TwHt9TDzt']        = False     # (m); Local tower axial (translational) deflection (relative to the undeflected position) of tower gage 9; Directed along the local zt-axis
ElastoDyn['TwHt1RDxt']        = False     # (deg); Local tower angular (rotational) roll deflection (relative to the undeflected position) of tower gage 1. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower deflections, so that the rotation sequence does not matter.; About the local xt-axis
ElastoDyn['TwHt1RDyt']        = False     # (deg); Local tower angular (rotational) pitch deflection (relative to the undeflected position) of tower gage 1. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower deflections, so that the rotation sequence does not matter.; About the local yt-axis
ElastoDyn['TwHt1RDzt']        = False     # (deg); Local tower angular (rotational) torsion deflection (relative to the undeflected position) of tower gage 1. This output will always be zero for FAST simulation results. Use it for examining tower torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.; About the local zt-axis
ElastoDyn['TwHt2RDxt']        = False     # (deg); Local tower angular (rotational) roll deflection (relative to the undeflected position) of tower gage 2. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower deflections, so that the rotation sequence does not matter.; About the local xt-axis
ElastoDyn['TwHt2RDyt']        = False     # (deg); Local tower angular (rotational) pitch deflection (relative to the undeflected position) of tower gage 2. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower deflections, so that the rotation sequence does not matter.; About the local yt-axis
ElastoDyn['TwHt2RDzt']        = False     # (deg); Local tower angular (rotational) torsion deflection (relative to the undeflected position) of tower gage 2. This output will always be zero for FAST simulation results. Use it for examining tower torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.; About the local zt-axis
ElastoDyn['TwHt3RDxt']        = False     # (deg); Local tower angular (rotational) roll deflection (relative to the undeflected position) of tower gage 3. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower deflections, so that the rotation sequence does not matter.; About the local xt-axis
ElastoDyn['TwHt3RDyt']        = False     # (deg); Local tower angular (rotational) pitch deflection (relative to the undeflected position) of tower gage 3. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower deflections, so that the rotation sequence does not matter.; About the local yt-axis
ElastoDyn['TwHt3RDzt']        = False     # (deg); Local tower angular (rotational) torsion deflection (relative to the undeflected position) of tower gage 3. This output will always be zero for FAST simulation results. Use it for examining tower torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.; About the local zt-axis
ElastoDyn['TwHt4RDxt']        = False     # (deg); Local tower angular (rotational) roll deflection (relative to the undeflected position) of tower gage 4. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower deflections, so that the rotation sequence does not matter.; About the local xt-axis
ElastoDyn['TwHt4RDyt']        = False     # (deg); Local tower angular (rotational) pitch deflection (relative to the undeflected position) of tower gage 4. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower deflections, so that the rotation sequence does not matter.; About the local yt-axis
ElastoDyn['TwHt4RDzt']        = False     # (deg); Local tower angular (rotational) torsion deflection (relative to the undeflected position) of tower gage 4. This output will always be zero for FAST simulation results. Use it for examining tower torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.; About the local zt-axis
ElastoDyn['TwHt5RDxt']        = False     # (deg); Local tower angular (rotational) roll deflection (relative to the undeflected position) of tower gage 5. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower deflections, so that the rotation sequence does not matter.; About the local xt-axis
ElastoDyn['TwHt5RDyt']        = False     # (deg); Local tower angular (rotational) pitch deflection (relative to the undeflected position) of tower gage 5. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower deflections, so that the rotation sequence does not matter.; About the local yt-axis
ElastoDyn['TwHt5RDzt']        = False     # (deg); Local tower angular (rotational) torsion deflection (relative to the undeflected position) of tower gage 5. This output will always be zero for FAST simulation results. Use it for examining tower torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.; About the local zt-axis
ElastoDyn['TwHt6RDxt']        = False     # (deg); Local tower angular (rotational) roll deflection (relative to the undeflected position) of tower gage 6. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower deflections, so that the rotation sequence does not matter.; About the local xt-axis
ElastoDyn['TwHt6RDyt']        = False     # (deg); Local tower angular (rotational) pitch deflection (relative to the undeflected position) of tower gage 6. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower deflections, so that the rotation sequence does not matter.; About the local yt-axis
ElastoDyn['TwHt6RDzt']        = False     # (deg); Local tower angular (rotational) torsion deflection (relative to the undeflected position) of tower gage 6. This output will always be zero for FAST simulation results. Use it for examining tower torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.; About the local zt-axis
ElastoDyn['TwHt7RDxt']        = False     # (deg); Local tower angular (rotational) roll deflection (relative to the undeflected position) of tower gage 7. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower deflections, so that the rotation sequence does not matter.; About the local xt-axis
ElastoDyn['TwHt7RDyt']        = False     # (deg); Local tower angular (rotational) pitch deflection (relative to the undeflected position) of tower gage 7. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower deflections, so that the rotation sequence does not matter.; About the local yt-axis
ElastoDyn['TwHt7RDzt']        = False     # (deg); Local tower angular (rotational) torsion deflection (relative to the undeflected position) of tower gage 7. This output will always be zero for FAST simulation results. Use it for examining tower torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.; About the local zt-axis
ElastoDyn['TwHt8RDxt']        = False     # (deg); Local tower angular (rotational) roll deflection (relative to the undeflected position) of tower gage 8. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower deflections, so that the rotation sequence does not matter.; About the local xt-axis
ElastoDyn['TwHt8RDyt']        = False     # (deg); Local tower angular (rotational) pitch deflection (relative to the undeflected position) of tower gage 8. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower deflections, so that the rotation sequence does not matter.; About the local yt-axis
ElastoDyn['TwHt8RDzt']        = False     # (deg); Local tower angular (rotational) torsion deflection (relative to the undeflected position) of tower gage 8. This output will always be zero for FAST simulation results. Use it for examining tower torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.; About the local zt-axis
ElastoDyn['TwHt9RDxt']        = False     # (deg); Local tower angular (rotational) roll deflection (relative to the undeflected position) of tower gage 9. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower deflections, so that the rotation sequence does not matter.; About the local xt-axis
ElastoDyn['TwHt9RDyt']        = False     # (deg); Local tower angular (rotational) pitch deflection (relative to the undeflected position) of tower gage 9. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower deflections, so that the rotation sequence does not matter.; About the local yt-axis
ElastoDyn['TwHt9RDzt']        = False     # (deg); Local tower angular (rotational) torsion deflection (relative to the undeflected position) of tower gage 9. This output will always be zero for FAST simulation results. Use it for examining tower torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.; About the local zt-axis
ElastoDyn['TwHt1TPxi']        = False     # (m); xi-component of the translational position (relative to the inertia frame) of tower gage 1; Directed along the local xi-axis
ElastoDyn['TwHt1TPyi']        = False     # (m); yi-component of the translational position (relative to the inertia frame) of tower gage 1; Directed along the local yi-axis
ElastoDyn['TwHt1TPzi']        = False     # (m); zi-component of the translational position (relative to ground level [onshore] or MSL [offshore]) of tower gage 1; Directed along the local zi-axis
ElastoDyn['TwHt2TPxi']        = False     # (m); xi-component of the translational position (relative to the inertia frame) of tower gage 2; Directed along the local xi-axis
ElastoDyn['TwHt2TPyi']        = False     # (m); yi-component of the translational position (relative to the inertia frame) of tower gage 2; Directed along the local yi-axis
ElastoDyn['TwHt2TPzi']        = False     # (m); zi-component of the translational position (relative to ground level [onshore] or MSL [offshore]) of tower gage 2; Directed along the local zi-axis
ElastoDyn['TwHt3TPxi']        = False     # (m); xi-component of the translational position (relative to the inertia frame) of tower gage 3; Directed along the local xi-axis
ElastoDyn['TwHt3TPyi']        = False     # (m); yi-component of the translational position (relative to the inertia frame) of tower gage 3; Directed along the local yi-axis
ElastoDyn['TwHt3TPzi']        = False     # (m); zi-component of the translational position (relative to ground level [onshore] or MSL [offshore]) of tower gage 3; Directed along the local zi-axis
ElastoDyn['TwHt4TPxi']        = False     # (m); xi-component of the translational position (relative to the inertia frame) of tower gage 4; Directed along the local xi-axis
ElastoDyn['TwHt4TPyi']        = False     # (m); yi-component of the translational position (relative to the inertia frame) of tower gage 4; Directed along the local yi-axis
ElastoDyn['TwHt4TPzi']        = False     # (m); zi-component of the translational position (relative to ground level [onshore] or MSL [offshore]) of tower gage 4; Directed along the local zi-axis
ElastoDyn['TwHt5TPxi']        = False     # (m); xi-component of the translational position (relative to the inertia frame) of tower gage 5; Directed along the local xi-axis
ElastoDyn['TwHt5TPyi']        = False     # (m); yi-component of the translational position (relative to the inertia frame) of tower gage 5; Directed along the local yi-axis
ElastoDyn['TwHt5TPzi']        = False     # (m); zi-component of the translational position (relative to ground level [onshore] or MSL [offshore]) of tower gage 5; Directed along the local zi-axis
ElastoDyn['TwHt6TPxi']        = False     # (m); xi-component of the translational position (relative to the inertia frame) of tower gage 6; Directed along the local xi-axis
ElastoDyn['TwHt6TPyi']        = False     # (m); yi-component of the translational position (relative to the inertia frame) of tower gage 6; Directed along the local yi-axis
ElastoDyn['TwHt6TPzi']        = False     # (m); zi-component of the translational position (relative to ground level [onshore] or MSL [offshore]) of tower gage 6; Directed along the local zi-axis
ElastoDyn['TwHt7TPxi']        = False     # (m); xi-component of the translational position (relative to the inertia frame) of tower gage 7; Directed along the local xi-axis
ElastoDyn['TwHt7TPyi']        = False     # (m); yi-component of the translational position (relative to the inertia frame) of tower gage 7; Directed along the local yi-axis
ElastoDyn['TwHt7TPzi']        = False     # (m); zi-component of the translational position (relative to ground level [onshore] or MSL [offshore]) of tower gage 7; Directed along the local zi-axis
ElastoDyn['TwHt8TPxi']        = False     # (m); xi-component of the translational position (relative to the inertia frame) of tower gage 8; Directed along the local xi-axis
ElastoDyn['TwHt8TPyi']        = False     # (m); yi-component of the translational position (relative to the inertia frame) of tower gage 8; Directed along the local yi-axis
ElastoDyn['TwHt8TPzi']        = False     # (m); zi-component of the translational position (relative to ground level [onshore] or MSL [offshore]) of tower gage 8; Directed along the local zi-axis
ElastoDyn['TwHt9TPxi']        = False     # (m); xi-component of the translational position (relative to the inertia frame) of tower gage 9; Directed along the local xi-axis
ElastoDyn['TwHt9TPyi']        = False     # (m); yi-component of the translational position (relative to the inertia frame) of tower gage 9; Directed along the local yi-axis
ElastoDyn['TwHt9TPzi']        = False     # (m); zi-component of the translational position (relative to ground level [onshore] or MSL [offshore]) of tower gage 9; Directed along the local zi-axis
ElastoDyn['TwHt1RPxi']        = False     # (deg); xi-component of the rotational position (relative to the inertia frame) of tower gage 1. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower and platform rotational deflections, so that the rotation sequence does not matter.; About the local xi-axis
ElastoDyn['TwHt1RPyi']        = False     # (deg); yi-component of the rotational position (relative to the inertia frame) of tower gage 1. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower and platform rotational deflections, so that the rotation sequence does not matter.; About the local yi-axis
ElastoDyn['TwHt1RPzi']        = False     # (deg); zi-component of the rotational position (relative to the inertia frame) of tower gage 1. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower and platform rotational deflections, so that the rotation sequence does not matter.; About the local zi-axis
ElastoDyn['TwHt2RPxi']        = False     # (deg); xi-component of the rotational position (relative to the inertia frame) of tower gage 2. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower and platform rotational deflections, so that the rotation sequence does not matter.; About the local xi-axis
ElastoDyn['TwHt2RPyi']        = False     # (deg); yi-component of the rotational position (relative to the inertia frame) of tower gage 2. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower and platform rotational deflections, so that the rotation sequence does not matter.; About the local yi-axis
ElastoDyn['TwHt2RPzi']        = False     # (deg); zi-component of the rotational position (relative to the inertia frame) of tower gage 2. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower and platform rotational deflections, so that the rotation sequence does not matter.; About the local zi-axis
ElastoDyn['TwHt3RPxi']        = False     # (deg); xi-component of the rotational position (relative to the inertia frame) of tower gage 3. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower and platform rotational deflections, so that the rotation sequence does not matter.; About the local xi-axis
ElastoDyn['TwHt3RPyi']        = False     # (deg); yi-component of the rotational position (relative to the inertia frame) of tower gage 3. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower and platform rotational deflections, so that the rotation sequence does not matter.; About the local yi-axis
ElastoDyn['TwHt3RPzi']        = False     # (deg); zi-component of the rotational position (relative to the inertia frame) of tower gage 3. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower and platform rotational deflections, so that the rotation sequence does not matter.; About the local zi-axis
ElastoDyn['TwHt4RPxi']        = False     # (deg); xi-component of the rotational position (relative to the inertia frame) of tower gage 4. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower and platform rotational deflections, so that the rotation sequence does not matter.; About the local xi-axis
ElastoDyn['TwHt4RPyi']        = False     # (deg); yi-component of the rotational position (relative to the inertia frame) of tower gage 4. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower and platform rotational deflections, so that the rotation sequence does not matter.; About the local yi-axis
ElastoDyn['TwHt4RPzi']        = False     # (deg); zi-component of the rotational position (relative to the inertia frame) of tower gage 4. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower and platform rotational deflections, so that the rotation sequence does not matter.; About the local zi-axis
ElastoDyn['TwHt5RPxi']        = False     # (deg); xi-component of the rotational position (relative to the inertia frame) of tower gage 5. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower and platform rotational deflections, so that the rotation sequence does not matter.; About the local xi-axis
ElastoDyn['TwHt5RPyi']        = False     # (deg); yi-component of the rotational position (relative to the inertia frame) of tower gage 5. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower and platform rotational deflections, so that the rotation sequence does not matter.; About the local yi-axis
ElastoDyn['TwHt5RPzi']        = False     # (deg); zi-component of the rotational position (relative to the inertia frame) of tower gage 5. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower and platform rotational deflections, so that the rotation sequence does not matter.; About the local zi-axis
ElastoDyn['TwHt6RPxi']        = False     # (deg); xi-component of the rotational position (relative to the inertia frame) of tower gage 6. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower and platform rotational deflections, so that the rotation sequence does not matter.; About the local xi-axis
ElastoDyn['TwHt6RPyi']        = False     # (deg); yi-component of the rotational position (relative to the inertia frame) of tower gage 6. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower and platform rotational deflections, so that the rotation sequence does not matter.; About the local yi-axis
ElastoDyn['TwHt6RPzi']        = False     # (deg); zi-component of the rotational position (relative to the inertia frame) of tower gage 6. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower and platform rotational deflections, so that the rotation sequence does not matter.; About the local zi-axis
ElastoDyn['TwHt7RPxi']        = False     # (deg); xi-component of the rotational position (relative to the inertia frame) of tower gage 7. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower and platform rotational deflections, so that the rotation sequence does not matter.; About the local xi-axis
ElastoDyn['TwHt7RPyi']        = False     # (deg); yi-component of the rotational position (relative to the inertia frame) of tower gage 7. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower and platform rotational deflections, so that the rotation sequence does not matter.; About the local yi-axis
ElastoDyn['TwHt7RPzi']        = False     # (deg); zi-component of the rotational position (relative to the inertia frame) of tower gage 7. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower and platform rotational deflections, so that the rotation sequence does not matter.; About the local zi-axis
ElastoDyn['TwHt8RPxi']        = False     # (deg); xi-component of the rotational position (relative to the inertia frame) of tower gage 8. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower and platform rotational deflections, so that the rotation sequence does not matter.; About the local xi-axis
ElastoDyn['TwHt8RPyi']        = False     # (deg); yi-component of the rotational position (relative to the inertia frame) of tower gage 8. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower and platform rotational deflections, so that the rotation sequence does not matter.; About the local yi-axis
ElastoDyn['TwHt8RPzi']        = False     # (deg); zi-component of the rotational position (relative to the inertia frame) of tower gage 8. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower and platform rotational deflections, so that the rotation sequence does not matter.; About the local zi-axis
ElastoDyn['TwHt9RPxi']        = False     # (deg); xi-component of the rotational position (relative to the inertia frame) of tower gage 9. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower and platform rotational deflections, so that the rotation sequence does not matter.; About the local xi-axis
ElastoDyn['TwHt9RPyi']        = False     # (deg); yi-component of the rotational position (relative to the inertia frame) of tower gage 9. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower and platform rotational deflections, so that the rotation sequence does not matter.; About the local yi-axis
ElastoDyn['TwHt9RPzi']        = False     # (deg); zi-component of the rotational position (relative to the inertia frame) of tower gage 9. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower and platform rotational deflections, so that the rotation sequence does not matter.; About the local zi-axis

# Platform Motions
ElastoDyn['PtfmTDxt']         = False     # (m); Platform horizontal surge (translational) displacement; Directed along the xt-axis
ElastoDyn['PtfmTDyt']         = False     # (m); Platform horizontal sway (translational) displacement; Directed along the yt-axis
ElastoDyn['PtfmTDzt']         = False     # (m); Platform vertical heave (translational) displacement; Directed along the zt-axis
ElastoDyn['PtfmTDxi']         = False     # (m); Platform horizontal surge (translational) displacement; Directed along the xi-axis
ElastoDyn['PtfmSurge']        = False     # (m); Platform horizontal surge (translational) displacement; Directed along the xi-axis
ElastoDyn['PtfmTDyi']         = False     # (m); Platform horizontal sway (translational) displacement; Directed along the yi-axis
ElastoDyn['PtfmSway']         = False     # (m); Platform horizontal sway (translational) displacement; Directed along the yi-axis
ElastoDyn['PtfmTDzi']         = False     # (m); Platform vertical heave (translational) displacement; Directed along the zi-axis
ElastoDyn['PtfmHeave']        = False     # (m); Platform vertical heave (translational) displacement; Directed along the zi-axis
ElastoDyn['PtfmTVxt']         = False     # (m/s); Platform horizontal surge (translational) velocity; Directed along the xt-axis
ElastoDyn['PtfmTVyt']         = False     # (m/s); Platform horizontal sway (translational) velocity; Directed along the yt-axis
ElastoDyn['PtfmTVzt']         = False     # (m/s); Platform vertical heave (translational) velocity; Directed along the zt-axis
ElastoDyn['PtfmTVxi']         = False     # (m/s); Platform horizontal surge (translational) velocity; Directed along the xi-axis
ElastoDyn['PtfmTVyi']         = False     # (m/s); Platform horizontal sway (translational) velocity; Directed along the yi-axis
ElastoDyn['PtfmTVzi']         = False     # (m/s); Platform vertical heave (translational) velocity; Directed along the zi-axis
ElastoDyn['PtfmTAxt']         = False     # (m/s^2); Platform horizontal surge (translational) acceleration; Directed along the xt-axis
ElastoDyn['PtfmTAyt']         = False     # (m/s^2); Platform horizontal sway (translational) acceleration; Directed along the yt-axis
ElastoDyn['PtfmTAzt']         = False     # (m/s^2); Platform vertical heave (translational) acceleration; Directed along the zt-axis
ElastoDyn['PtfmTAxi']         = False     # (m/s^2); Platform horizontal surge (translational) acceleration; Directed along the xi-axis
ElastoDyn['PtfmTAyi']         = False     # (m/s^2); Platform horizontal sway (translational) acceleration; Directed along the yi-axis
ElastoDyn['PtfmTAzi']         = False     # (m/s^2); Platform vertical heave (translational) acceleration; Directed along the zi-axis
ElastoDyn['PtfmRDxi']         = False     # (deg); Platform roll tilt angular (rotational) displacement. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small rotational platform displacements, so that the rotation sequence does not matter.; About the xi-axis
ElastoDyn['PtfmRoll']         = False     # (deg); Platform roll tilt angular (rotational) displacement. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small rotational platform displacements, so that the rotation sequence does not matter.; About the xi-axis
ElastoDyn['PtfmRDyi']         = False     # (deg); Platform pitch tilt angular (rotational) displacement. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small rotational platform displacements, so that the rotation sequence does not matter.; About the yi-axis
ElastoDyn['PtfmPitch']        = False     # (deg); Platform pitch tilt angular (rotational) displacement. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small rotational platform displacements, so that the rotation sequence does not matter.; About the yi-axis
ElastoDyn['PtfmRDzi']         = False     # (deg); Platform yaw angular (rotational) displacement. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small rotational platform displacements, so that the rotation sequence does not matter.; About the zi-axis
ElastoDyn['PtfmYaw']          = False     # (deg); Platform yaw angular (rotational) displacement. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small rotational platform displacements, so that the rotation sequence does not matter.; About the zi-axis
ElastoDyn['PtfmRVxt']         = False     # (deg/s); Platform roll tilt angular (rotational) velocity; About the xt-axis
ElastoDyn['PtfmRVyt']         = False     # (deg/s); Platform pitch tilt angular (rotational) velocity; About the yt-axis
ElastoDyn['PtfmRVzt']         = False     # (deg/s); Platform yaw angular (rotational) velocity; About the zt-axis
ElastoDyn['PtfmRVxi']         = False     # (deg/s); Platform roll tilt angular (rotational) velocity; About the xi-axis
ElastoDyn['PtfmRVyi']         = False     # (deg/s); Platform pitch tilt angular (rotational) velocity; About the yi-axis
ElastoDyn['PtfmRVzi']         = False     # (deg/s); Platform yaw angular (rotational) velocity; About the zi-axis
ElastoDyn['PtfmRAxt']         = False     # (deg/s^2); Platform roll tilt angular (rotational) acceleration; About the xt-axis
ElastoDyn['PtfmRAyt']         = False     # (deg/s^2); Platform pitch tilt angular (rotational) acceleration; About the yt-axis
ElastoDyn['PtfmRAzt']         = False     # (deg/s^2); Platform yaw angular (rotational) acceleration; About the zt-axis
ElastoDyn['PtfmRAxi']         = False     # (deg/s^2); Platform roll tilt angular (rotational) acceleration; About the xi-axis
ElastoDyn['PtfmRAyi']         = False     # (deg/s^2); Platform pitch tilt angular (rotational) acceleration; About the yi-axis
ElastoDyn['PtfmRAzi']         = False     # (deg/s^2); Platform yaw angular (rotational) acceleration; About the zi-axis

# Blade 1 Root Loads
ElastoDyn['RootFxc1']         = False     # (kN); Blade 1 out-of-plane shear force at the blade root; Directed along the xc1-axis
ElastoDyn['RootFyc1']         = False     # (kN); Blade 1 in-plane shear force at the blade root; Directed along the yc1-axis
ElastoDyn['RootFzc1']         = False     # (kN); Blade 1 axial force at the blade root; Directed along the zc1- and zb1-axes
ElastoDyn['RootFzb1']         = False     # (kN); Blade 1 axial force at the blade root; Directed along the zc1- and zb1-axes
ElastoDyn['RootFxb1']         = False     # (kN); Blade 1 flapwise shear force at the blade root; Directed along the xb1-axis
ElastoDyn['RootFyb1']         = False     # (kN); Blade 1 edgewise shear force at the blade root; Directed along the yb1-axis
ElastoDyn['RootMxc1']         = False     # (kN m); Blade 1 in-plane moment (i.e., the moment caused by in-plane forces) at the blade root; About the xc1-axis
ElastoDyn['RootMIP1']         = False     # (kN m); Blade 1 in-plane moment (i.e., the moment caused by in-plane forces) at the blade root; About the xc1-axis
ElastoDyn['RootMyc1']         = False     # (kN m); Blade 1 out-of-plane moment (i.e., the moment caused by out-of-plane forces) at the blade root; About the yc1-axis
ElastoDyn['RootMOoP1']        = False     # (kN m); Blade 1 out-of-plane moment (i.e., the moment caused by out-of-plane forces) at the blade root; About the yc1-axis
ElastoDyn['RootMzc1']         = False     # (kN m); Blade 1 pitching moment at the blade root; About the zc1- and zb1-axes
ElastoDyn['RootMzb1']         = False     # (kN m); Blade 1 pitching moment at the blade root; About the zc1- and zb1-axes
ElastoDyn['RootMxb1']         = False     # (kN m); Blade 1 edgewise moment (i.e., the moment caused by edgewise forces) at the blade root; About the xb1-axis
ElastoDyn['RootMEdg1']        = False     # (kN m); Blade 1 edgewise moment (i.e., the moment caused by edgewise forces) at the blade root; About the xb1-axis
ElastoDyn['RootMyb1']         = False     # (kN m); Blade 1 flapwise moment (i.e., the moment caused by flapwise forces) at the blade root; About the yb1-axis
ElastoDyn['RootMFlp1']        = False     # (kN m); Blade 1 flapwise moment (i.e., the moment caused by flapwise forces) at the blade root; About the yb1-axis

# Blade 2 Root Loads
ElastoDyn['RootFxc2']         = False     # (kN); Blade 2 out-of-plane shear force at the blade root; Directed along the xc2-axis
ElastoDyn['RootFyc2']         = False     # (kN); Blade 2 in-plane shear force at the blade root; Directed along the yc2-axis
ElastoDyn['RootFzc2']         = False     # (kN); Blade 2 axial force at the blade root; Directed along the zc2- and zb2-axes
ElastoDyn['RootFzb2']         = False     # (kN); Blade 2 axial force at the blade root; Directed along the zc2- and zb2-axes
ElastoDyn['RootFxb2']         = False     # (kN); Blade 2 flapwise shear force at the blade root; Directed along the xb2-axis
ElastoDyn['RootFyb2']         = False     # (kN); Blade 2 edgewise shear force at the blade root; Directed along the yb2-axis
ElastoDyn['RootMxc2']         = False     # (kN m); Blade 2 in-plane moment (i.e., the moment caused by in-plane forces) at the blade root; About the xc2-axis
ElastoDyn['RootMIP2']         = False     # (kN m); Blade 2 in-plane moment (i.e., the moment caused by in-plane forces) at the blade root; About the xc2-axis
ElastoDyn['RootMyc2']         = False     # (kN m); Blade 2 out-of-plane moment (i.e., the moment caused by out-of-plane forces) at the blade root; About the yc2-axis
ElastoDyn['RootMOoP2']        = False     # (kN m); Blade 2 out-of-plane moment (i.e., the moment caused by out-of-plane forces) at the blade root; About the yc2-axis
ElastoDyn['RootMzc2']         = False     # (kN m); Blade 2 pitching moment at the blade root; About the zc2- and zb2-axes
ElastoDyn['RootMzb2']         = False     # (kN m); Blade 2 pitching moment at the blade root; About the zc2- and zb2-axes
ElastoDyn['RootMxb2']         = False     # (kN m); Blade 2 edgewise moment (i.e., the moment caused by edgewise forces) at the blade root; About the xb2-axis
ElastoDyn['RootMEdg2']        = False     # (kN m); Blade 2 edgewise moment (i.e., the moment caused by edgewise forces) at the blade root; About the xb2-axis
ElastoDyn['RootMyb2']         = False     # (kN m); Blade 2 flapwise moment (i.e., the moment caused by flapwise forces) at the blade root; About the yb2-axis
ElastoDyn['RootMFlp2']        = False     # (kN m); Blade 2 flapwise moment (i.e., the moment caused by flapwise forces) at the blade root; About the yb2-axis

# Blade 3 Root Loads
ElastoDyn['RootFxc3']         = False     # (kN); Blade 3 out-of-plane shear force at the blade root; Directed along the xc3-axis
ElastoDyn['RootFyc3']         = False     # (kN); Blade 3 in-plane shear force at the blade root; Directed along the yc3-axis
ElastoDyn['RootFzc3']         = False     # (kN); Blade 3 axial force at the blade root; Directed along the zc3- and zb3-axes
ElastoDyn['RootFzb3']         = False     # (kN); Blade 3 axial force at the blade root; Directed along the zc3- and zb3-axes
ElastoDyn['RootFxb3']         = False     # (kN); Blade 3 flapwise shear force at the blade root; Directed along the xb3-axis
ElastoDyn['RootFyb3']         = False     # (kN); Blade 3 edgewise shear force at the blade root; Directed along the yb3-axis
ElastoDyn['RootMxc3']         = False     # (kN m); Blade 3 in-plane moment (i.e., the moment caused by in-plane forces) at the blade root; About the xc3-axis
ElastoDyn['RootMIP3']         = False     # (kN m); Blade 3 in-plane moment (i.e., the moment caused by in-plane forces) at the blade root; About the xc3-axis
ElastoDyn['RootMyc3']         = False     # (kN m); Blade 3 out-of-plane moment (i.e., the moment caused by out-of-plane forces) at the blade root; About the yc3-axis
ElastoDyn['RootMOoP3']        = False     # (kN m); Blade 3 out-of-plane moment (i.e., the moment caused by out-of-plane forces) at the blade root; About the yc3-axis
ElastoDyn['RootMzc3']         = False     # (kN m); Blade 3 pitching moment at the blade root; About the zc3- and zb3-axes
ElastoDyn['RootMzb3']         = False     # (kN m); Blade 3 pitching moment at the blade root; About the zc3- and zb3-axes
ElastoDyn['RootMxb3']         = False     # (kN m); Blade 3 edgewise moment (i.e., the moment caused by edgewise forces) at the blade root; About the xb3-axis
ElastoDyn['RootMEdg3']        = False     # (kN m); Blade 3 edgewise moment (i.e., the moment caused by edgewise forces) at the blade root; About the xb3-axis
ElastoDyn['RootMyb3']         = False     # (kN m); Blade 3 flapwise moment (i.e., the moment caused by flapwise forces) at the blade root; About the yb3-axis
ElastoDyn['RootMFlp3']        = False     # (kN m); Blade 3 flapwise moment (i.e., the moment caused by flapwise forces) at the blade root; About the yb3-axis

# Blade 1 Local Span Loads
ElastoDyn['Spn1MLxb1']        = False     # (kN m); Blade 1 local edgewise moment at span station 1; About the local xb1-axis
ElastoDyn['Spn1MLyb1']        = False     # (kN m); Blade 1 local flapwise moment at span station 1; About the local yb1-axis
ElastoDyn['Spn1MLzb1']        = False     # (kN m); Blade 1 local pitching moment at span station 1; About the local zb1-axis
ElastoDyn['Spn2MLxb1']        = False     # (kN m); Blade 1 local edgewise moment at span station 2; About the local xb1-axis
ElastoDyn['Spn2MLyb1']        = False     # (kN m); Blade 1 local flapwise moment at span station 2; About the local yb1-axis
ElastoDyn['Spn2MLzb1']        = False     # (kN m); Blade 1 local pitching moment at span station 2; About the local zb1-axis
ElastoDyn['Spn3MLxb1']        = False     # (kN m); Blade 1 local edgewise moment at span station 3; About the local xb1-axis
ElastoDyn['Spn3MLyb1']        = False     # (kN m); Blade 1 local flapwise moment at span station 3; About the local yb1-axis
ElastoDyn['Spn3MLzb1']        = False     # (kN m); Blade 1 local pitching moment at span station 3; About the local zb1-axis
ElastoDyn['Spn4MLxb1']        = False     # (kN m); Blade 1 local edgewise moment at span station 4; About the local xb1-axis
ElastoDyn['Spn4MLyb1']        = False     # (kN m); Blade 1 local flapwise moment at span station 4; About the local yb1-axis
ElastoDyn['Spn4MLzb1']        = False     # (kN m); Blade 1 local pitching moment at span station 4; About the local zb1-axis
ElastoDyn['Spn5MLxb1']        = False     # (kN m); Blade 1 local edgewise moment at span station 5; About the local xb1-axis
ElastoDyn['Spn5MLyb1']        = False     # (kN m); Blade 1 local flapwise moment at span station 5; About the local yb1-axis
ElastoDyn['Spn5MLzb1']        = False     # (kN m); Blade 1 local pitching moment at span station 5; About the local zb1-axis
ElastoDyn['Spn6MLxb1']        = False     # (kN m); Blade 1 local edgewise moment at span station 6; About the local xb1-axis
ElastoDyn['Spn6MLyb1']        = False     # (kN m); Blade 1 local flapwise moment at span station 6; About the local yb1-axis
ElastoDyn['Spn6MLzb1']        = False     # (kN m); Blade 1 local pitching moment at span station 6; About the local zb1-axis
ElastoDyn['Spn7MLxb1']        = False     # (kN m); Blade 1 local edgewise moment at span station 7; About the local xb1-axis
ElastoDyn['Spn7MLyb1']        = False     # (kN m); Blade 1 local flapwise moment at span station 7; About the local yb1-axis
ElastoDyn['Spn7MLzb1']        = False     # (kN m); Blade 1 local pitching moment at span station 7; About the local zb1-axis
ElastoDyn['Spn8MLxb1']        = False     # (kN m); Blade 1 local edgewise moment at span station 8; About the local xb1-axis
ElastoDyn['Spn8MLyb1']        = False     # (kN m); Blade 1 local flapwise moment at span station 8; About the local yb1-axis
ElastoDyn['Spn8MLzb1']        = False     # (kN m); Blade 1 local pitching moment at span station 8; About the local zb1-axis
ElastoDyn['Spn9MLxb1']        = False     # (kN m); Blade 1 local edgewise moment at span station 9; About the local xb1-axis
ElastoDyn['Spn9MLyb1']        = False     # (kN m); Blade 1 local flapwise moment at span station 9; About the local yb1-axis
ElastoDyn['Spn9MLzb1']        = False     # (kN m); Blade 1 local pitching moment at span station 9; About the local zb1-axis
ElastoDyn['Spn1FLxb1']        = False     # (kN); Blade 1 local flapwise shear force at span station 1; Directed along the local xb1-axis
ElastoDyn['Spn1FLyb1']        = False     # (kN); Blade 1 local edgewise shear force at span station 1; Directed along the local yb1-axis
ElastoDyn['Spn1FLzb1']        = False     # (kN); Blade 1 local axial force at span station 1; Directed along the local zb1-axis
ElastoDyn['Spn2FLxb1']        = False     # (kN); Blade 1 local flapwise shear force at span station 2; Directed along the local xb1-axis
ElastoDyn['Spn2FLyb1']        = False     # (kN); Blade 1 local edgewise shear force at span station 2; Directed along the local yb1-axis
ElastoDyn['Spn2FLzb1']        = False     # (kN); Blade 1 local axial force at span station 2; Directed along the local zb1-axis
ElastoDyn['Spn3FLxb1']        = False     # (kN); Blade 1 local flapwise shear force at span station 3; Directed along the local xb1-axis
ElastoDyn['Spn3FLyb1']        = False     # (kN); Blade 1 local edgewise shear force at span station 3; Directed along the local yb1-axis
ElastoDyn['Spn3FLzb1']        = False     # (kN); Blade 1 local axial force at span station 3; Directed along the local zb1-axis
ElastoDyn['Spn4FLxb1']        = False     # (kN); Blade 1 local flapwise shear force at span station 4; Directed along the local xb1-axis
ElastoDyn['Spn4FLyb1']        = False     # (kN); Blade 1 local edgewise shear force at span station 4; Directed along the local yb1-axis
ElastoDyn['Spn4FLzb1']        = False     # (kN); Blade 1 local axial force at span station 4; Directed along the local zb1-axis
ElastoDyn['Spn5FLxb1']        = False     # (kN); Blade 1 local flapwise shear force at span station 5; Directed along the local xb1-axis
ElastoDyn['Spn5FLyb1']        = False     # (kN); Blade 1 local edgewise shear force at span station 5; Directed along the local yb1-axis
ElastoDyn['Spn5FLzb1']        = False     # (kN); Blade 1 local axial force at span station 5; Directed along the local zb1-axis
ElastoDyn['Spn6FLxb1']        = False     # (kN); Blade 1 local flapwise shear force at span station 6; Directed along the local xb1-axis
ElastoDyn['Spn6FLyb1']        = False     # (kN); Blade 1 local edgewise shear force at span station 6; Directed along the local yb1-axis
ElastoDyn['Spn6FLzb1']        = False     # (kN); Blade 1 local axial force at span station 6; Directed along the local zb1-axis
ElastoDyn['Spn7FLxb1']        = False     # (kN); Blade 1 local flapwise shear force at span station 7; Directed along the local xb1-axis
ElastoDyn['Spn7FLyb1']        = False     # (kN); Blade 1 local edgewise shear force at span station 7; Directed along the local yb1-axis
ElastoDyn['Spn7FLzb1']        = False     # (kN); Blade 1 local axial force at span station 7; Directed along the local zb1-axis
ElastoDyn['Spn8FLxb1']        = False     # (kN); Blade 1 local flapwise shear force at span station 8; Directed along the local xb1-axis
ElastoDyn['Spn8FLyb1']        = False     # (kN); Blade 1 local edgewise shear force at span station 8; Directed along the local yb1-axis
ElastoDyn['Spn8FLzb1']        = False     # (kN); Blade 1 local axial force at span station 8; Directed along the local zb1-axis
ElastoDyn['Spn9FLxb1']        = False     # (kN); Blade 1 local flapwise shear force at span station 9; Directed along the local xb1-axis
ElastoDyn['Spn9FLyb1']        = False     # (kN); Blade 1 local edgewise shear force at span station 9; Directed along the local yb1-axis
ElastoDyn['Spn9FLzb1']        = False     # (kN); Blade 1 local axial force at span station 9; Directed along the local zb1-axis

# Blade 2 Local Span Loads
ElastoDyn['Spn1MLxb2']        = False     # (kN m); Blade 2 local edgewise moment at span station 1; About the local xb2-axis
ElastoDyn['Spn1MLyb2']        = False     # (kN m); Blade 2 local flapwise moment at span station 1; About the local yb2-axis
ElastoDyn['Spn1MLzb2']        = False     # (kN m); Blade 2 local pitching moment at span station 1; About the local zb2-axis
ElastoDyn['Spn2MLxb2']        = False     # (kN m); Blade 2 local edgewise moment at span station 2; About the local xb2-axis
ElastoDyn['Spn2MLyb2']        = False     # (kN m); Blade 2 local flapwise moment at span station 2; About the local yb2-axis
ElastoDyn['Spn2MLzb2']        = False     # (kN m); Blade 2 local pitching moment at span station 2; About the local zb2-axis
ElastoDyn['Spn3MLxb2']        = False     # (kN m); Blade 2 local edgewise moment at span station 3; About the local xb2-axis
ElastoDyn['Spn3MLyb2']        = False     # (kN m); Blade 2 local flapwise moment at span station 3; About the local yb2-axis
ElastoDyn['Spn3MLzb2']        = False     # (kN m); Blade 2 local pitching moment at span station 3; About the local zb2-axis
ElastoDyn['Spn4MLxb2']        = False     # (kN m); Blade 2 local edgewise moment at span station 4; About the local xb2-axis
ElastoDyn['Spn4MLyb2']        = False     # (kN m); Blade 2 local flapwise moment at span station 4; About the local yb2-axis
ElastoDyn['Spn4MLzb2']        = False     # (kN m); Blade 2 local pitching moment at span station 4; About the local zb2-axis
ElastoDyn['Spn5MLxb2']        = False     # (kN m); Blade 2 local edgewise moment at span station 5; About the local xb2-axis
ElastoDyn['Spn5MLyb2']        = False     # (kN m); Blade 2 local flapwise moment at span station 5; About the local yb2-axis
ElastoDyn['Spn5MLzb2']        = False     # (kN m); Blade 2 local pitching moment at span station 5; About the local zb2-axis
ElastoDyn['Spn6MLxb2']        = False     # (kN m); Blade 2 local edgewise moment at span station 6; About the local xb2-axis
ElastoDyn['Spn6MLyb2']        = False     # (kN m); Blade 2 local flapwise moment at span station 6; About the local yb2-axis
ElastoDyn['Spn6MLzb2']        = False     # (kN m); Blade 2 local pitching moment at span station 6; About the local zb2-axis
ElastoDyn['Spn7MLxb2']        = False     # (kN m); Blade 2 local edgewise moment at span station 7; About the local xb2-axis
ElastoDyn['Spn7MLyb2']        = False     # (kN m); Blade 2 local flapwise moment at span station 7; About the local yb2-axis
ElastoDyn['Spn7MLzb2']        = False     # (kN m); Blade 2 local pitching moment at span station 7; About the local zb2-axis
ElastoDyn['Spn8MLxb2']        = False     # (kN m); Blade 2 local edgewise moment at span station 8; About the local xb2-axis
ElastoDyn['Spn8MLyb2']        = False     # (kN m); Blade 2 local flapwise moment at span station 8; About the local yb2-axis
ElastoDyn['Spn8MLzb2']        = False     # (kN m); Blade 2 local pitching moment at span station 8; About the local zb2-axis
ElastoDyn['Spn9MLxb2']        = False     # (kN m); Blade 2 local edgewise moment at span station 9; About the local xb2-axis
ElastoDyn['Spn9MLyb2']        = False     # (kN m); Blade 2 local flapwise moment at span station 9; About the local yb2-axis
ElastoDyn['Spn9MLzb2']        = False     # (kN m); Blade 2 local pitching moment at span station 9; About the local zb2-axis
ElastoDyn['Spn1FLxb2']        = False     # (kN); Blade 2 local flapwise shear force at span station 1; Directed along the local xb2-axis
ElastoDyn['Spn1FLyb2']        = False     # (kN); Blade 2 local edgewise shear force at span station 1; Directed along the local yb2-axis
ElastoDyn['Spn1FLzb2']        = False     # (kN); Blade 2 local axial force at span station 1; Directed along the local zb2-axis
ElastoDyn['Spn2FLxb2']        = False     # (kN); Blade 2 local flapwise shear force at span station 2; Directed along the local xb2-axis
ElastoDyn['Spn2FLyb2']        = False     # (kN); Blade 2 local edgewise shear force at span station 2; Directed along the local yb2-axis
ElastoDyn['Spn2FLzb2']        = False     # (kN); Blade 2 local axial force at span station 2; Directed along the local zb2-axis
ElastoDyn['Spn3FLxb2']        = False     # (kN); Blade 2 local flapwise shear force at span station 3; Directed along the local xb2-axis
ElastoDyn['Spn3FLyb2']        = False     # (kN); Blade 2 local edgewise shear force at span station 3; Directed along the local yb2-axis
ElastoDyn['Spn3FLzb2']        = False     # (kN); Blade 2 local axial force at span station 3; Directed along the local zb2-axis
ElastoDyn['Spn4FLxb2']        = False     # (kN); Blade 2 local flapwise shear force at span station 4; Directed along the local xb2-axis
ElastoDyn['Spn4FLyb2']        = False     # (kN); Blade 2 local edgewise shear force at span station 4; Directed along the local yb2-axis
ElastoDyn['Spn4FLzb2']        = False     # (kN); Blade 2 local axial force at span station 4; Directed along the local zb2-axis
ElastoDyn['Spn5FLxb2']        = False     # (kN); Blade 2 local flapwise shear force at span station 5; Directed along the local xb2-axis
ElastoDyn['Spn5FLyb2']        = False     # (kN); Blade 2 local edgewise shear force at span station 5; Directed along the local yb2-axis
ElastoDyn['Spn5FLzb2']        = False     # (kN); Blade 2 local axial force at span station 5; Directed along the local zb2-axis
ElastoDyn['Spn6FLxb2']        = False     # (kN); Blade 2 local flapwise shear force at span station 6; Directed along the local xb2-axis
ElastoDyn['Spn6FLyb2']        = False     # (kN); Blade 2 local edgewise shear force at span station 6; Directed along the local yb2-axis
ElastoDyn['Spn6FLzb2']        = False     # (kN); Blade 2 local axial force at span station 6; Directed along the local zb2-axis
ElastoDyn['Spn7FLxb2']        = False     # (kN); Blade 2 local flapwise shear force at span station 7; Directed along the local xb2-axis
ElastoDyn['Spn7FLyb2']        = False     # (kN); Blade 2 local edgewise shear force at span station 7; Directed along the local yb2-axis
ElastoDyn['Spn7FLzb2']        = False     # (kN); Blade 2 local axial force at span station 7; Directed along the local zb2-axis
ElastoDyn['Spn8FLxb2']        = False     # (kN); Blade 2 local flapwise shear force at span station 8; Directed along the local xb2-axis
ElastoDyn['Spn8FLyb2']        = False     # (kN); Blade 2 local edgewise shear force at span station 8; Directed along the local yb2-axis
ElastoDyn['Spn8FLzb2']        = False     # (kN); Blade 2 local axial force at span station 8; Directed along the local zb2-axis
ElastoDyn['Spn9FLxb2']        = False     # (kN); Blade 2 local flapwise shear force at span station 9; Directed along the local xb2-axis
ElastoDyn['Spn9FLyb2']        = False     # (kN); Blade 2 local edgewise shear force at span station 9; Directed along the local yb2-axis
ElastoDyn['Spn9FLzb2']        = False     # (kN); Blade 2 local axial force at span station 9; Directed along the local zb2-axis

# Blade 3 Local Span Loads
ElastoDyn['Spn1MLxb3']        = False     # (kN m); Blade 3 local edgewise moment at span station 1; About the local xb3-axis
ElastoDyn['Spn1MLyb3']        = False     # (kN m); Blade 3 local flapwise moment at span station 1; About the local yb3-axis
ElastoDyn['Spn1MLzb3']        = False     # (kN m); Blade 3 local pitching moment at span station 1; About the local zb3-axis
ElastoDyn['Spn2MLxb3']        = False     # (kN m); Blade 3 local edgewise moment at span station 2; About the local xb3-axis
ElastoDyn['Spn2MLyb3']        = False     # (kN m); Blade 3 local flapwise moment at span station 2; About the local yb3-axis
ElastoDyn['Spn2MLzb3']        = False     # (kN m); Blade 3 local pitching moment at span station 2; About the local zb3-axis
ElastoDyn['Spn3MLxb3']        = False     # (kN m); Blade 3 local edgewise moment at span station 3; About the local xb3-axis
ElastoDyn['Spn3MLyb3']        = False     # (kN m); Blade 3 local flapwise moment at span station 3; About the local yb3-axis
ElastoDyn['Spn3MLzb3']        = False     # (kN m); Blade 3 local pitching moment at span station 3; About the local zb3-axis
ElastoDyn['Spn4MLxb3']        = False     # (kN m); Blade 3 local edgewise moment at span station 4; About the local xb3-axis
ElastoDyn['Spn4MLyb3']        = False     # (kN m); Blade 3 local flapwise moment at span station 4; About the local yb3-axis
ElastoDyn['Spn4MLzb3']        = False     # (kN m); Blade 3 local pitching moment at span station 4; About the local zb3-axis
ElastoDyn['Spn5MLxb3']        = False     # (kN m); Blade 3 local edgewise moment at span station 5; About the local xb3-axis
ElastoDyn['Spn5MLyb3']        = False     # (kN m); Blade 3 local flapwise moment at span station 5; About the local yb3-axis
ElastoDyn['Spn5MLzb3']        = False     # (kN m); Blade 3 local pitching moment at span station 5; About the local zb3-axis
ElastoDyn['Spn6MLxb3']        = False     # (kN m); Blade 3 local edgewise moment at span station 6; About the local xb3-axis
ElastoDyn['Spn6MLyb3']        = False     # (kN m); Blade 3 local flapwise moment at span station 6; About the local yb3-axis
ElastoDyn['Spn6MLzb3']        = False     # (kN m); Blade 3 local pitching moment at span station 6; About the local zb3-axis
ElastoDyn['Spn7MLxb3']        = False     # (kN m); Blade 3 local edgewise moment at span station 7; About the local xb3-axis
ElastoDyn['Spn7MLyb3']        = False     # (kN m); Blade 3 local flapwise moment at span station 7; About the local yb3-axis
ElastoDyn['Spn7MLzb3']        = False     # (kN m); Blade 3 local pitching moment at span station 7; About the local zb3-axis
ElastoDyn['Spn8MLxb3']        = False     # (kN m); Blade 3 local edgewise moment at span station 8; About the local xb3-axis
ElastoDyn['Spn8MLyb3']        = False     # (kN m); Blade 3 local flapwise moment at span station 8; About the local yb3-axis
ElastoDyn['Spn8MLzb3']        = False     # (kN m); Blade 3 local pitching moment at span station 8; About the local zb3-axis
ElastoDyn['Spn9MLxb3']        = False     # (kN m); Blade 3 local edgewise moment at span station 9; About the local xb3-axis
ElastoDyn['Spn9MLyb3']        = False     # (kN m); Blade 3 local flapwise moment at span station 9; About the local yb3-axis
ElastoDyn['Spn9MLzb3']        = False     # (kN m); Blade 3 local pitching moment at span station 9; About the local zb3-axis
ElastoDyn['Spn1FLxb3']        = False     # (kN); Blade 3 local flapwise shear force at span station 1; Directed along the local xb3-axis
ElastoDyn['Spn1FLyb3']        = False     # (kN); Blade 3 local edgewise shear force at span station 1; Directed along the local yb3-axis
ElastoDyn['Spn1FLzb3']        = False     # (kN); Blade 3 local axial force at span station 1; Directed along the local zb3-axis
ElastoDyn['Spn2FLxb3']        = False     # (kN); Blade 3 local flapwise shear force at span station 2; Directed along the local xb3-axis
ElastoDyn['Spn2FLyb3']        = False     # (kN); Blade 3 local edgewise shear force at span station 2; Directed along the local yb3-axis
ElastoDyn['Spn2FLzb3']        = False     # (kN); Blade 3 local axial force at span station 2; Directed along the local zb3-axis
ElastoDyn['Spn3FLxb3']        = False     # (kN); Blade 3 local flapwise shear force at span station 3; Directed along the local xb3-axis
ElastoDyn['Spn3FLyb3']        = False     # (kN); Blade 3 local edgewise shear force at span station 3; Directed along the local yb3-axis
ElastoDyn['Spn3FLzb3']        = False     # (kN); Blade 3 local axial force at span station 3; Directed along the local zb3-axis
ElastoDyn['Spn4FLxb3']        = False     # (kN); Blade 3 local flapwise shear force at span station 4; Directed along the local xb3-axis
ElastoDyn['Spn4FLyb3']        = False     # (kN); Blade 3 local edgewise shear force at span station 4; Directed along the local yb3-axis
ElastoDyn['Spn4FLzb3']        = False     # (kN); Blade 3 local axial force at span station 4; Directed along the local zb3-axis
ElastoDyn['Spn5FLxb3']        = False     # (kN); Blade 3 local flapwise shear force at span station 5; Directed along the local xb3-axis
ElastoDyn['Spn5FLyb3']        = False     # (kN); Blade 3 local edgewise shear force at span station 5; Directed along the local yb3-axis
ElastoDyn['Spn5FLzb3']        = False     # (kN); Blade 3 local axial force at span station 5; Directed along the local zb3-axis
ElastoDyn['Spn6FLxb3']        = False     # (kN); Blade 3 local flapwise shear force at span station 6; Directed along the local xb3-axis
ElastoDyn['Spn6FLyb3']        = False     # (kN); Blade 3 local edgewise shear force at span station 6; Directed along the local yb3-axis
ElastoDyn['Spn6FLzb3']        = False     # (kN); Blade 3 local axial force at span station 6; Directed along the local zb3-axis
ElastoDyn['Spn7FLxb3']        = False     # (kN); Blade 3 local flapwise shear force at span station 7; Directed along the local xb3-axis
ElastoDyn['Spn7FLyb3']        = False     # (kN); Blade 3 local edgewise shear force at span station 7; Directed along the local yb3-axis
ElastoDyn['Spn7FLzb3']        = False     # (kN); Blade 3 local axial force at span station 7; Directed along the local zb3-axis
ElastoDyn['Spn8FLxb3']        = False     # (kN); Blade 3 local flapwise shear force at span station 8; Directed along the local xb3-axis
ElastoDyn['Spn8FLyb3']        = False     # (kN); Blade 3 local edgewise shear force at span station 8; Directed along the local yb3-axis
ElastoDyn['Spn8FLzb3']        = False     # (kN); Blade 3 local axial force at span station 8; Directed along the local zb3-axis
ElastoDyn['Spn9FLxb3']        = False     # (kN); Blade 3 local flapwise shear force at span station 9; Directed along the local xb3-axis
ElastoDyn['Spn9FLyb3']        = False     # (kN); Blade 3 local edgewise shear force at span station 9; Directed along the local yb3-axis
ElastoDyn['Spn9FLzb3']        = False     # (kN); Blade 3 local axial force at span station 9; Directed along the local zb3-axis

# Hub and Rotor Loads
ElastoDyn['LSShftFxa']        = False     # (kN); Low-speed shaft thrust force (this is constant along the shaft and is equivalent to the rotor thrust force); Directed along the xa- and xs-axes
ElastoDyn['LSShftFxs']        = False     # (kN); Low-speed shaft thrust force (this is constant along the shaft and is equivalent to the rotor thrust force); Directed along the xa- and xs-axes
ElastoDyn['LSSGagFxa']        = False     # (kN); Low-speed shaft thrust force (this is constant along the shaft and is equivalent to the rotor thrust force); Directed along the xa- and xs-axes
ElastoDyn['LSSGagFxs']        = False     # (kN); Low-speed shaft thrust force (this is constant along the shaft and is equivalent to the rotor thrust force); Directed along the xa- and xs-axes
ElastoDyn['RotThrust']        = False     # (kN); Low-speed shaft thrust force (this is constant along the shaft and is equivalent to the rotor thrust force); Directed along the xa- and xs-axes
ElastoDyn['LSShftFya']        = False     # (kN); Rotating low-speed shaft shear force (this is constant along the shaft); Directed along the ya-axis
ElastoDyn['LSSGagFya']        = False     # (kN); Rotating low-speed shaft shear force (this is constant along the shaft); Directed along the ya-axis
ElastoDyn['LSShftFza']        = False     # (kN); Rotating low-speed shaft shear force (this is constant along the shaft); Directed along the za-axis
ElastoDyn['LSSGagFza']        = False     # (kN); Rotating low-speed shaft shear force (this is constant along the shaft); Directed along the za-axis
ElastoDyn['LSShftFys']        = False     # (kN); Nonrotating low-speed shaft shear force (this is constant along the shaft); Directed along the ys-axis
ElastoDyn['LSSGagFys']        = False     # (kN); Nonrotating low-speed shaft shear force (this is constant along the shaft); Directed along the ys-axis
ElastoDyn['LSShftFzs']        = False     # (kN); Nonrotating low-speed shaft shear force (this is constant along the shaft); Directed along the zs-axis
ElastoDyn['LSSGagFzs']        = False     # (kN); Nonrotating low-speed shaft shear force (this is constant along the shaft); Directed along the zs-axis
ElastoDyn['LSShftMxa']        = False     # (kN m); Low-speed shaft torque (this is constant along the shaft and is equivalent to the rotor torque); About the xa- and xs-axes
ElastoDyn['LSShftMxs']        = False     # (kN m); Low-speed shaft torque (this is constant along the shaft and is equivalent to the rotor torque); About the xa- and xs-axes
ElastoDyn['LSSGagMxa']        = False     # (kN m); Low-speed shaft torque (this is constant along the shaft and is equivalent to the rotor torque); About the xa- and xs-axes
ElastoDyn['LSSGagMxs']        = False     # (kN m); Low-speed shaft torque (this is constant along the shaft and is equivalent to the rotor torque); About the xa- and xs-axes
ElastoDyn['RotTorq']          = False     # (kN m); Low-speed shaft torque (this is constant along the shaft and is equivalent to the rotor torque); About the xa- and xs-axes
ElastoDyn['LSShftTq']         = False     # (kN m); Low-speed shaft torque (this is constant along the shaft and is equivalent to the rotor torque); About the xa- and xs-axes
ElastoDyn['LSSTipMya']        = False     # (kN m); Rotating low-speed shaft bending moment at the shaft tip (teeter pin for 2-blader, apex of rotation for 3-blader); About the ya-axis
ElastoDyn['LSSTipMza']        = False     # (kN m); Rotating low-speed shaft bending moment at the shaft tip (teeter pin for 2-blader, apex of rotation for 3-blader); About the za-axis
ElastoDyn['LSSTipMys']        = False     # (kN m); Nonrotating low-speed shaft bending moment at the shaft tip (teeter pin for 2-blader, apex of rotation for 3-blader); About the ys-axis
ElastoDyn['LSSTipMzs']        = False     # (kN m); Nonrotating low-speed shaft bending moment at the shaft tip (teeter pin for 2-blader, apex of rotation for 3-blader); About the zs-axis
ElastoDyn['RotPwr']           = False     # (kW); Rotor power (this is equivalent to the low-speed shaft power); N/A
ElastoDyn['LSShftPwr']        = False     # (kW); Rotor power (this is equivalent to the low-speed shaft power); N/A

# Shaft Strain Gage Loads
ElastoDyn['LSSGagMya']        = False     # (kN m); Rotating low-speed shaft bending moment at the shaft's strain gage (shaft strain gage located by input ShftGagL); About the ya-axis
ElastoDyn['LSSGagMza']        = False     # (kN m); Rotating low-speed shaft bending moment at the shaft's strain gage (shaft strain gage located by input ShftGagL); About the za-axis
ElastoDyn['LSSGagMys']        = False     # (kN m); Nonrotating low-speed shaft bending moment at the shaft's strain gage (shaft strain gage located by input ShftGagL); About the ys-axis
ElastoDyn['LSSGagMzs']        = False     # (kN m); Nonrotating low-speed shaft bending moment at the shaft's strain gage (shaft strain gage located by input ShftGagL); About the zs-axis

# High-Speed Shaft Loads
ElastoDyn['HSShftTq']         = False     # (kN m); High-speed shaft torque (this is constant along the shaft); Same sign as LSShftTq / RotTorq / LSShftMxa / LSShftMxs / LSSGagMxa / LSSGagMxs
ElastoDyn['HSSBrTq']          = False     # (kN m); High-speed shaft brake torque (i.e., the actual moment applied to the high-speed shaft by the brake); Always positive (indicating dissipation of power)
ElastoDyn['HSShftPwr']        = False     # (kW); High-speed shaft power; Same sign as HSShftTq

# Rotor-Furl Bearing Loads
ElastoDyn['RFrlBrM']          = False     # (kN m); Rotor-furl bearing moment; About the rotor-furl axis

# Tail-Furl Bearing Loads
ElastoDyn['TFrlBrM']          = False     # (kN m); Tail-furl bearing moment; About the tail-furl axis

# Tower-Top / Yaw Bearing Loads
ElastoDyn['YawBrFxn']         = False     # (kN); Rotating (with nacelle) tower-top / yaw bearing shear force; Directed along the xn-axis
ElastoDyn['YawBrFyn']         = False     # (kN); Rotating (with nacelle) tower-top / yaw bearing shear force; Directed along the yn-axis
ElastoDyn['YawBrFzn']         = False     # (kN); Tower-top / yaw bearing axial force; Directed along the zn- and zp-axes
ElastoDyn['YawBrFzp']         = False     # (kN); Tower-top / yaw bearing axial force; Directed along the zn- and zp-axes
ElastoDyn['YawBrFxp']         = False     # (kN); Tower-top / yaw bearing fore-aft (nonrotating) shear force; Directed along the xp-axis
ElastoDyn['YawBrFyp']         = False     # (kN); Tower-top / yaw bearing side-to-side (nonrotating) shear force; Directed along the yp-axis
ElastoDyn['YawBrMxn']         = False     # (kN m); Rotating (with nacelle) tower-top / yaw bearing roll moment; About the xn-axis
ElastoDyn['YawBrMyn']         = False     # (kN m); Rotating (with nacelle) tower-top / yaw bearing pitch moment; About the yn-axis
ElastoDyn['YawBrMzn']         = False     # (kN m); Tower-top / yaw bearing yaw moment; About the zn- and zp-axes
ElastoDyn['YawBrMzp']         = False     # (kN m); Tower-top / yaw bearing yaw moment; About the zn- and zp-axes
ElastoDyn['YawBrMxp']         = False     # (kN m); Nonrotating tower-top / yaw bearing roll moment; About the xp-axis
ElastoDyn['YawBrMyp']         = False     # (kN m); Nonrotating tower-top / yaw bearing pitch moment; About the yp-axis

# Tower Base Loads
ElastoDyn['TwrBsFxt']         = False     # (kN); Tower base fore-aft shear force; Directed along the xt-axis
ElastoDyn['TwrBsFyt']         = False     # (kN); Tower base side-to-side shear force; Directed along the yt-axis
ElastoDyn['TwrBsFzt']         = False     # (kN); Tower base axial force; Directed along the zt-axis
ElastoDyn['TwrBsMxt']         = False     # (kN m); Tower base roll (or side-to-side) moment (i.e., the moment caused by side-to-side forces); About the xt-axis
ElastoDyn['TwrBsMyt']         = False     # (kN m); Tower base pitching (or fore-aft) moment (i.e., the moment caused by fore-aft forces); About the yt-axis
ElastoDyn['TwrBsMzt']         = False     # (kN m); Tower base yaw (or torsional) moment; About the zt-axis

# Local Tower Loads
ElastoDyn['TwHt1MLxt']        = False     # (kN m); Local tower roll (or side-to-side) moment of tower gage 1; About the local xt-axis
ElastoDyn['TwHt1MLyt']        = False     # (kN m); Local tower pitching (or fore-aft) moment of tower gage 1; About the local yt-axis
ElastoDyn['TwHt1MLzt']        = False     # (kN m); Local tower yaw (or torsional) moment of tower gage 1; About the local zt-axis
ElastoDyn['TwHt2MLxt']        = False     # (kN m); Local tower roll (or side-to-side) moment of tower gage 2; About the local xt-axis
ElastoDyn['TwHt2MLyt']        = False     # (kN m); Local tower pitching (or fore-aft) moment of tower gage 2; About the local yt-axis
ElastoDyn['TwHt2MLzt']        = False     # (kN m); Local tower yaw (or torsional) moment of tower gage 2; About the local zt-axis
ElastoDyn['TwHt3MLxt']        = False     # (kN m); Local tower roll (or side-to-side) moment of tower gage 3; About the local xt-axis
ElastoDyn['TwHt3MLyt']        = False     # (kN m); Local tower pitching (or fore-aft) moment of tower gage 3; About the local yt-axis
ElastoDyn['TwHt3MLzt']        = False     # (kN m); Local tower yaw (or torsional) moment of tower gage 3; About the local zt-axis
ElastoDyn['TwHt4MLxt']        = False     # (kN m); Local tower roll (or side-to-side) moment of tower gage 4; About the local xt-axis
ElastoDyn['TwHt4MLyt']        = False     # (kN m); Local tower pitching (or fore-aft) moment of tower gage 4; About the local yt-axis
ElastoDyn['TwHt4MLzt']        = False     # (kN m); Local tower yaw (or torsional) moment of tower gage 4; About the local zt-axis
ElastoDyn['TwHt5MLxt']        = False     # (kN m); Local tower roll (or side-to-side) moment of tower gage 5; About the local xt-axis
ElastoDyn['TwHt5MLyt']        = False     # (kN m); Local tower pitching (or fore-aft) moment of tower gage 5; About the local yt-axis
ElastoDyn['TwHt5MLzt']        = False     # (kN m); Local tower yaw (or torsional) moment of tower gage 5; About the local zt-axis
ElastoDyn['TwHt6MLxt']        = False     # (kN m); Local tower roll (or side-to-side) moment of tower gage 6; About the local xt-axis
ElastoDyn['TwHt6MLyt']        = False     # (kN m); Local tower pitching (or fore-aft) moment of tower gage 6; About the local yt-axis
ElastoDyn['TwHt6MLzt']        = False     # (kN m); Local tower yaw (or torsional) moment of tower gage 6; About the local zt-axis
ElastoDyn['TwHt7MLxt']        = False     # (kN m); Local tower roll (or side-to-side) moment of tower gage 7; About the local xt-axis
ElastoDyn['TwHt7MLyt']        = False     # (kN m); Local tower pitching (or fore-aft) moment of tower gage 7; About the local yt-axis
ElastoDyn['TwHt7MLzt']        = False     # (kN m); Local tower yaw (or torsional) moment of tower gage 7; About the local zt-axis
ElastoDyn['TwHt8MLxt']        = False     # (kN m); Local tower roll (or side-to-side) moment of tower gage 8; About the local xt-axis
ElastoDyn['TwHt8MLyt']        = False     # (kN m); Local tower pitching (or fore-aft) moment of tower gage 8; About the local yt-axis
ElastoDyn['TwHt8MLzt']        = False     # (kN m); Local tower yaw (or torsional) moment of tower gage 8; About the local zt-axis
ElastoDyn['TwHt9MLxt']        = False     # (kN m); Local tower roll (or side-to-side) moment of tower gage 9; About the local xt-axis
ElastoDyn['TwHt9MLyt']        = False     # (kN m); Local tower pitching (or fore-aft) moment of tower gage 9; About the local yt-axis
ElastoDyn['TwHt9MLzt']        = False     # (kN m); Local tower yaw (or torsional) moment of tower gage 9; About the local zt-axis
ElastoDyn['TwHt1FLxt']        = False     # (kN); Local tower roll (or side-to-side) force of tower gage 1; About the local xt-axis
ElastoDyn['TwHt1FLyt']        = False     # (kN); Local tower pitching (or fore-aft) force of tower gage 1; About the local yt-axis
ElastoDyn['TwHt1FLzt']        = False     # (kN); Local tower yaw (or torsional) force of tower gage 1; About the local zt-axis
ElastoDyn['TwHt2FLxt']        = False     # (kN); Local tower roll (or side-to-side) force of tower gage 2; About the local xt-axis
ElastoDyn['TwHt2FLyt']        = False     # (kN); Local tower pitching (or fore-aft) force of tower gage 2; About the local yt-axis
ElastoDyn['TwHt2FLzt']        = False     # (kN); Local tower yaw (or torsional) force of tower gage 2; About the local zt-axis
ElastoDyn['TwHt3FLxt']        = False     # (kN); Local tower roll (or side-to-side) force of tower gage 3; About the local xt-axis
ElastoDyn['TwHt3FLyt']        = False     # (kN); Local tower pitching (or fore-aft) force of tower gage 3; About the local yt-axis
ElastoDyn['TwHt3FLzt']        = False     # (kN); Local tower yaw (or torsional) force of tower gage 3; About the local zt-axis
ElastoDyn['TwHt4FLxt']        = False     # (kN); Local tower roll (or side-to-side) force of tower gage 4; About the local xt-axis
ElastoDyn['TwHt4FLyt']        = False     # (kN); Local tower pitching (or fore-aft) force of tower gage 4; About the local yt-axis
ElastoDyn['TwHt4FLzt']        = False     # (kN); Local tower yaw (or torsional) force of tower gage 4; About the local zt-axis
ElastoDyn['TwHt5FLxt']        = False     # (kN); Local tower roll (or side-to-side) force of tower gage 5; About the local xt-axis
ElastoDyn['TwHt5FLyt']        = False     # (kN); Local tower pitching (or fore-aft) force of tower gage 5; About the local yt-axis
ElastoDyn['TwHt5FLzt']        = False     # (kN); Local tower yaw (or torsional) force of tower gage 5; About the local zt-axis
ElastoDyn['TwHt6FLxt']        = False     # (kN); Local tower roll (or side-to-side) force of tower gage 6; About the local xt-axis
ElastoDyn['TwHt6FLyt']        = False     # (kN); Local tower pitching (or fore-aft) force of tower gage 6; About the local yt-axis
ElastoDyn['TwHt6FLzt']        = False     # (kN); Local tower yaw (or torsional) force of tower gage 6; About the local zt-axis
ElastoDyn['TwHt7FLxt']        = False     # (kN); Local tower roll (or side-to-side) force of tower gage 7; About the local xt-axis
ElastoDyn['TwHt7FLyt']        = False     # (kN); Local tower pitching (or fore-aft) force of tower gage 7; About the local yt-axis
ElastoDyn['TwHt7FLzt']        = False     # (kN); Local tower yaw (or torsional) force of tower gage 7; About the local zt-axis
ElastoDyn['TwHt8FLxt']        = False     # (kN); Local tower roll (or side-to-side) force of tower gage 8; About the local xt-axis
ElastoDyn['TwHt8FLyt']        = False     # (kN); Local tower pitching (or fore-aft) force of tower gage 8; About the local yt-axis
ElastoDyn['TwHt8FLzt']        = False     # (kN); Local tower yaw (or torsional) force of tower gage 8; About the local zt-axis
ElastoDyn['TwHt9FLxt']        = False     # (kN); Local tower roll (or side-to-side) force of tower gage 9; About the local xt-axis
ElastoDyn['TwHt9FLyt']        = False     # (kN); Local tower pitching (or fore-aft) force of tower gage 9; About the local yt-axis
ElastoDyn['TwHt9FLzt']        = False     # (kN); Local tower yaw (or torsional) force of tower gage 9; About the local zt-axis

# Internal Degrees of Freedom
ElastoDyn['Q_B1E1']           = False     # (m); Displacement of 1st edgewise bending-mode DOF of blade 1; 
ElastoDyn['Q_B2E1']           = False     # (m); Displacement of 1st edgewise bending-mode DOF of blade 2; 
ElastoDyn['Q_B3E1']           = False     # (m); Displacement of 1st edgewise bending-mode DOF of blade 3; 
ElastoDyn['Q_B1F1']           = False     # (m); Displacement of 1st flapwise bending-mode DOF of blade 1; 
ElastoDyn['Q_B2F1']           = False     # (m); Displacement of 1st flapwise bending-mode DOF of blade 2; 
ElastoDyn['Q_B3F1']           = False     # (m); Displacement of 1st flapwise bending-mode DOF of blade 3; 
ElastoDyn['Q_B1F2']           = False     # (m); Displacement of 2nd flapwise bending-mode DOF of blade 1; 
ElastoDyn['Q_B2F2']           = False     # (m); Displacement of 2nd flapwise bending-mode DOF of blade 2; 
ElastoDyn['Q_B3F2']           = False     # (m); Displacement of 2nd flapwise bending-mode DOF of blade 3; 
ElastoDyn['Q_Teet']           = False     # (rad); Displacement of hub teetering DOF; 
ElastoDyn['Q_DrTr']           = False     # (rad); Displacement of drivetrain rotational-flexibility DOF; 
ElastoDyn['Q_GeAz']           = False     # (rad); Displacement of variable speed generator DOF; 
ElastoDyn['Q_RFrl']           = False     # (rad); Displacement of rotor-furl DOF; 
ElastoDyn['Q_TFrl']           = False     # (rad); Displacement of tail-furl DOF; 
ElastoDyn['Q_Yaw']            = False     # (rad); Displacement of nacelle yaw DOF; 
ElastoDyn['Q_TFA1']           = False     # (m); Displacement of 1st tower fore-aft bending mode DOF; 
ElastoDyn['Q_TSS1']           = False     # (m); Displacement of 1st tower side-to-side bending mode DOF; 
ElastoDyn['Q_TFA2']           = False     # (m); Displacement of 2nd tower fore-aft bending mode DOF; 
ElastoDyn['Q_TSS2']           = False     # (m); Displacement of 2nd tower side-to-side bending mode DOF; 
ElastoDyn['Q_Sg']             = False     # (m); Displacement of platform horizontal surge translation DOF; 
ElastoDyn['Q_Sw']             = False     # (m); Displacement of platform horizontal sway translation DOF; 
ElastoDyn['Q_Hv']             = False     # (m); Displacement of platform vertical heave translation DOF; 
ElastoDyn['Q_R']              = False     # (rad); Displacement of platform roll tilt rotation DOF; 
ElastoDyn['Q_P']              = False     # (rad); Displacement of platform pitch tilt rotation DOF; 
ElastoDyn['Q_Y']              = False     # (rad); Displacement of platform yaw rotation DOF; 
ElastoDyn['QD_B1E1']          = False     # (m/s); Velocity of 1st edgewise bending-mode DOF of blade 1; 
ElastoDyn['QD_B2E1']          = False     # (m/s); Velocity of 1st edgewise bending-mode DOF of blade 2; 
ElastoDyn['QD_B3E1']          = False     # (m/s); Velocity of 1st edgewise bending-mode DOF of blade 3; 
ElastoDyn['QD_B1F1']          = False     # (m/s); Velocity of 1st flapwise bending-mode DOF of blade 1; 
ElastoDyn['QD_B2F1']          = False     # (m/s); Velocity of 1st flapwise bending-mode DOF of blade 2; 
ElastoDyn['QD_B3F1']          = False     # (m/s); Velocity of 1st flapwise bending-mode DOF of blade 3; 
ElastoDyn['QD_B1F2']          = False     # (m/s); Velocity of 2nd flapwise bending-mode DOF of blade 1; 
ElastoDyn['QD_B2F2']          = False     # (m/s); Velocity of 2nd flapwise bending-mode DOF of blade 2; 
ElastoDyn['QD_B3F2']          = False     # (m/s); Velocity of 2nd flapwise bending-mode DOF of blade 3; 
ElastoDyn['QD_Teet']          = False     # (rad/s); Velocity of hub teetering DOF; 
ElastoDyn['QD_DrTr']          = False     # (rad/s); Velocity of drivetrain rotational-flexibility DOF; 
ElastoDyn['QD_GeAz']          = False     # (rad/s); Velocity of variable speed generator DOF; 
ElastoDyn['QD_RFrl']          = False     # (rad/s); Velocity of rotor-furl DOF; 
ElastoDyn['QD_TFrl']          = False     # (rad/s); Velocity of tail-furl DOF; 
ElastoDyn['QD_Yaw']           = False     # (rad/s); Velocity of nacelle yaw DOF; 
ElastoDyn['QD_TFA1']          = False     # (m/s); Velocity of 1st tower fore-aft bending mode DOF; 
ElastoDyn['QD_TSS1']          = False     # (m/s); Velocity of 1st tower side-to-side bending mode DOF; 
ElastoDyn['QD_TFA2']          = False     # (m/s); Velocity of 2nd tower fore-aft bending mode DOF; 
ElastoDyn['QD_TSS2']          = False     # (m/s); Velocity of 2nd tower side-to-side bending mode DOF; 
ElastoDyn['QD_Sg']            = False     # (m/s); Velocity of platform horizontal surge translation DOF; 
ElastoDyn['QD_Sw']            = False     # (m/s); Velocity of platform horizontal sway translation DOF; 
ElastoDyn['QD_Hv']            = False     # (m/s); Velocity of platform vertical heave translation DOF; 
ElastoDyn['QD_R']             = False     # (rad/s); Velocity of platform roll tilt rotation DOF; 
ElastoDyn['QD_P']             = False     # (rad/s); Velocity of platform pitch tilt rotation DOF; 
ElastoDyn['QD_Y']             = False     # (rad/s); Velocity of platform yaw rotation DOF; 
ElastoDyn['QD2_B1E1']         = False     # (m/s^2); Acceleration of 1st edgewise bending-mode DOF of blade 1; 
ElastoDyn['QD2_B2E1']         = False     # (m/s^2); Acceleration of 1st edgewise bending-mode DOF of blade 2; 
ElastoDyn['QD2_B3E1']         = False     # (m/s^2); Acceleration of 1st edgewise bending-mode DOF of blade 3; 
ElastoDyn['QD2_B1F1']         = False     # (m/s^2); Acceleration of 1st flapwise bending-mode DOF of blade 1; 
ElastoDyn['QD2_B2F1']         = False     # (m/s^2); Acceleration of 1st flapwise bending-mode DOF of blade 2; 
ElastoDyn['QD2_B3F1']         = False     # (m/s^2); Acceleration of 1st flapwise bending-mode DOF of blade 3; 
ElastoDyn['QD2_B1F2']         = False     # (m/s^2); Acceleration of 2nd flapwise bending-mode DOF of blade 1; 
ElastoDyn['QD2_B2F2']         = False     # (m/s^2); Acceleration of 2nd flapwise bending-mode DOF of blade 2; 
ElastoDyn['QD2_B3F2']         = False     # (m/s^2); Acceleration of 2nd flapwise bending-mode DOF of blade 3; 
ElastoDyn['QD2_Teet']         = False     # (rad/s^2); Acceleration of hub teetering DOF; 
ElastoDyn['QD2_DrTr']         = False     # (rad/s^2); Acceleration of drivetrain rotational-flexibility DOF; 
ElastoDyn['QD2_GeAz']         = False     # (rad/s^2); Acceleration of variable speed generator DOF; 
ElastoDyn['QD2_RFrl']         = False     # (rad/s^2); Acceleration of rotor-furl DOF; 
ElastoDyn['QD2_TFrl']         = False     # (rad/s^2); Acceleration of tail-furl DOF; 
ElastoDyn['QD2_Yaw']          = False     # (rad/s^2); Acceleration of nacelle yaw DOF; 
ElastoDyn['QD2_TFA1']         = False     # (m/s^2); Acceleration of 1st tower fore-aft bending mode DOF; 
ElastoDyn['QD2_TSS1']         = False     # (m/s^2); Acceleration of 1st tower side-to-side bending mode DOF; 
ElastoDyn['QD2_TFA2']         = False     # (m/s^2); Acceleration of 2nd tower fore-aft bending mode DOF; 
ElastoDyn['QD2_TSS2']         = False     # (m/s^2); Acceleration of 2nd tower side-to-side bending mode DOF; 
ElastoDyn['QD2_Sg']           = False     # (m/s^2); Acceleration of platform horizontal surge translation DOF; 
ElastoDyn['QD2_Sw']           = False     # (m/s^2); Acceleration of platform horizontal sway translation DOF; 
ElastoDyn['QD2_Hv']           = False     # (m/s^2); Acceleration of platform vertical heave translation DOF; 
ElastoDyn['QD2_R']            = False     # (rad/s^2); Acceleration of platform roll tilt rotation DOF; 
ElastoDyn['QD2_P']            = False     # (rad/s^2); Acceleration of platform pitch tilt rotation DOF; 
ElastoDyn['QD2_Y']            = False     # (rad/s^2); Acceleration of platform yaw rotation DOF; 


""" BeamDyn """
BeamDyn = {}

# Reaction Loads
BeamDyn['RootFxr']            = False     # (N); x-component of the root reaction force expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['RootFyr']            = False     # (N); y-component of the root reaction force expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['RootFzr']            = False     # (N); z-component of the root reaction force expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['RootMxr']            = False     # (N-m); x-component of the root reaction moment expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['RootMyr']            = False     # (N-m); y-component of the root reaction moment expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['RootMzr']            = False     # (N-m); z-component of the root reaction moment expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system

# Tip Motions
BeamDyn['TipTDxr']            = False     # (m); Tip translational deflection (relative to the undeflected position) expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['TipTDyr']            = False     # (m); Tip translational deflection (relative to the undeflected position) expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['TipTDzr']            = False     # (m); Tip translational deflection (relative to the undeflected position) expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['TipRDxr']            = False     # (-); Tip angular/rotational deflection Wiener-Milenkovi  parameter (relative to the undeflected orientation) expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['TipRDyr']            = False     # (-); Tip angular/rotational deflection Wiener-Milenkovi  parameter (relative to the undeflected orientation) expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['TipRDzr']            = False     # (-); Tip angular/rotational deflection Wiener-Milenkovi  parameter (relative to the undeflected orientation) expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['TipTVXg']            = False     # (m/s); Tip translational velocities (absolute) expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['TipTVYg']            = False     # (m/s); Tip translational velocities (absolute) expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['TipTVZg']            = False     # (m/s); Tip translational velocities (absolute) expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['TipRVXg']            = False     # (deg/s); Tip angular/rotational velocities (absolute) expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['TipRVYg']            = False     # (deg/s); Tip angular/rotational velocities (absolute) expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['TipRVZg']            = False     # (deg/s); Tip angular/rotational velocities (absolute) expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['TipTAXl']            = False     # (m/s^2); Tip translational accelerations (absolute) expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['TipTAYl']            = False     # (m/s^2); Tip translational accelerations (absolute) expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['TipTAZl']            = False     # (m/s^2); Tip translational accelerations (absolute) expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['TipRAXl']            = False     # (deg/s^2); Tip angular/rotational accelerations (absolute) expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['TipRAYl']            = False     # (deg/s^2); Tip angular/rotational accelerations (absolute) expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['TipRAZl']            = False     # (deg/s^2); Tip angular/rotational accelerations (absolute) expressed in l; l: a floating coordinate system local to the deflected beam

# Sectional Loads
BeamDyn['N1Fxl']              = False     # (N); Sectional force resultants at Node 1 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N1Fyl']              = False     # (N); Sectional force resultants at Node 1 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N1Fzl']              = False     # (N); Sectional force resultants at Node 1 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N2Fxl']              = False     # (N); Sectional force resultants at Node 2 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N2Fyl']              = False     # (N); Sectional force resultants at Node 2 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N2Fzl']              = False     # (N); Sectional force resultants at Node 2 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N3Fxl']              = False     # (N); Sectional force resultants at Node 3 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N3Fyl']              = False     # (N); Sectional force resultants at Node 3 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N3Fzl']              = False     # (N); Sectional force resultants at Node 3 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N4Fxl']              = False     # (N); Sectional force resultants at Node 4 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N4Fyl']              = False     # (N); Sectional force resultants at Node 4 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N4Fzl']              = False     # (N); Sectional force resultants at Node 4 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N5Fxl']              = False     # (N); Sectional force resultants at Node 5 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N5Fyl']              = False     # (N); Sectional force resultants at Node 5 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N5Fzl']              = False     # (N); Sectional force resultants at Node 5 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N6Fxl']              = False     # (N); Sectional force resultants at Node 6 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N6Fyl']              = False     # (N); Sectional force resultants at Node 6 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N6Fzl']              = False     # (N); Sectional force resultants at Node 6 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N7Fxl']              = False     # (N); Sectional force resultants at Node 7 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N7Fyl']              = False     # (N); Sectional force resultants at Node 7 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N7Fzl']              = False     # (N); Sectional force resultants at Node 7 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N8Fxl']              = False     # (N); Sectional force resultants at Node 8 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N8Fyl']              = False     # (N); Sectional force resultants at Node 8 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N8Fzl']              = False     # (N); Sectional force resultants at Node 8 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N9Fxl']              = False     # (N); Sectional force resultants at Node 9 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N9Fyl']              = False     # (N); Sectional force resultants at Node 9 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N9Fzl']              = False     # (N); Sectional force resultants at Node 9 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N1Mxl']              = False     # (N-m); Sectional moment resultants at Node 1 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N1Myl']              = False     # (N-m); Sectional moment resultants at Node 1 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N1Mzl']              = False     # (N-m); Sectional moment resultants at Node 1 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N2Mxl']              = False     # (N-m); Sectional moment resultants at Node 2 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N2Myl']              = False     # (N-m); Sectional moment resultants at Node 2 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N2Mzl']              = False     # (N-m); Sectional moment resultants at Node 2 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N3Mxl']              = False     # (N-m); Sectional moment resultants at Node 3 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N3Myl']              = False     # (N-m); Sectional moment resultants at Node 3 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N3Mzl']              = False     # (N-m); Sectional moment resultants at Node 3 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N4Mxl']              = False     # (N-m); Sectional moment resultants at Node 4 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N4Myl']              = False     # (N-m); Sectional moment resultants at Node 4 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N4Mzl']              = False     # (N-m); Sectional moment resultants at Node 4 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N5Mxl']              = False     # (N-m); Sectional moment resultants at Node 5 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N5Myl']              = False     # (N-m); Sectional moment resultants at Node 5 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N5Mzl']              = False     # (N-m); Sectional moment resultants at Node 5 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N6Mxl']              = False     # (N-m); Sectional moment resultants at Node 6 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N6Myl']              = False     # (N-m); Sectional moment resultants at Node 6 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N6Mzl']              = False     # (N-m); Sectional moment resultants at Node 6 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N7Mxl']              = False     # (N-m); Sectional moment resultants at Node 7 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N7Myl']              = False     # (N-m); Sectional moment resultants at Node 7 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N7Mzl']              = False     # (N-m); Sectional moment resultants at Node 7 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N8Mxl']              = False     # (N-m); Sectional moment resultants at Node 8 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N8Myl']              = False     # (N-m); Sectional moment resultants at Node 8 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N8Mzl']              = False     # (N-m); Sectional moment resultants at Node 8 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N9Mxl']              = False     # (N-m); Sectional moment resultants at Node 9 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N9Myl']              = False     # (N-m); Sectional moment resultants at Node 9 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N9Mzl']              = False     # (N-m); Sectional moment resultants at Node 9 expressed in l; l: a floating coordinate system local to the deflected beam

# Sectional Motions
BeamDyn['N1TDxr']             = False     # (m); Sectional translational deflection (relative to the undeflected position) at Node 1 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N1TDyr']             = False     # (m); Sectional translational deflection (relative to the undeflected position) at Node 1 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N1TDzr']             = False     # (m); Sectional translational deflection (relative to the undeflected position) at Node 1 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N2TDxr']             = False     # (m); Sectional translational deflection (relative to the undeflected position) at Node 2 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N2TDyr']             = False     # (m); Sectional translational deflection (relative to the undeflected position) at Node 2 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N2TDzr']             = False     # (m); Sectional translational deflection (relative to the undeflected position) at Node 2 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N3TDxr']             = False     # (m); Sectional translational deflection (relative to the undeflected position) at Node 3 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N3TDyr']             = False     # (m); Sectional translational deflection (relative to the undeflected position) at Node 3 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N3TDzr']             = False     # (m); Sectional translational deflection (relative to the undeflected position) at Node 3 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N4TDxr']             = False     # (m); Sectional translational deflection (relative to the undeflected position) at Node 4 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N4TDyr']             = False     # (m); Sectional translational deflection (relative to the undeflected position) at Node 4 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N4TDzr']             = False     # (m); Sectional translational deflection (relative to the undeflected position) at Node 4 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N5TDxr']             = False     # (m); Sectional translational deflection (relative to the undeflected position) at Node 5 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N5TDyr']             = False     # (m); Sectional translational deflection (relative to the undeflected position) at Node 5 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N5TDzr']             = False     # (m); Sectional translational deflection (relative to the undeflected position) at Node 5 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N6TDxr']             = False     # (m); Sectional translational deflection (relative to the undeflected position) at Node 6 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N6TDyr']             = False     # (m); Sectional translational deflection (relative to the undeflected position) at Node 6 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N6TDzr']             = False     # (m); Sectional translational deflection (relative to the undeflected position) at Node 6 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N7TDxr']             = False     # (m); Sectional translational deflection (relative to the undeflected position) at Node 7 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N7TDyr']             = False     # (m); Sectional translational deflection (relative to the undeflected position) at Node 7 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N7TDzr']             = False     # (m); Sectional translational deflection (relative to the undeflected position) at Node 7 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N8TDxr']             = False     # (m); Sectional translational deflection (relative to the undeflected position) at Node 8 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N8TDyr']             = False     # (m); Sectional translational deflection (relative to the undeflected position) at Node 8 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N8TDzr']             = False     # (m); Sectional translational deflection (relative to the undeflected position) at Node 8 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N9TDxr']             = False     # (m); Sectional translational deflection (relative to the undeflected position) at Node 9 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N9TDyr']             = False     # (m); Sectional translational deflection (relative to the undeflected position) at Node 9 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N9TDzr']             = False     # (m); Sectional translational deflection (relative to the undeflected position) at Node 9 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N1RDxr']             = False     # (-); Sectional angular/rotational deflection Wiener-Milenkovic parameter (relative to the undeflected orientation) at Node 1 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N1RDyr']             = False     # (-); Sectional angular/rotational deflection Wiener-Milenkovic parameter (relative to the undeflected orientation) at Node 1 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N1RDzr']             = False     # (-); Sectional angular/rotational deflection Wiener-Milenkovic parameter (relative to the undeflected orientation) at Node 1 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N2RDxr']             = False     # (-); Sectional angular/rotational deflection Wiener-Milenkovic parameter (relative to the undeflected orientation) at Node 2 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N2RDyr']             = False     # (-); Sectional angular/rotational deflection Wiener-Milenkovic parameter (relative to the undeflected orientation) at Node 2 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N2RDzr']             = False     # (-); Sectional angular/rotational deflection Wiener-Milenkovic parameter (relative to the undeflected orientation) at Node 2 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N3RDxr']             = False     # (-); Sectional angular/rotational deflection Wiener-Milenkovic parameter (relative to the undeflected orientation) at Node 3 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N3RDyr']             = False     # (-); Sectional angular/rotational deflection Wiener-Milenkovic parameter (relative to the undeflected orientation) at Node 3 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N3RDzr']             = False     # (-); Sectional angular/rotational deflection Wiener-Milenkovic parameter (relative to the undeflected orientation) at Node 3 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N4RDxr']             = False     # (-); Sectional angular/rotational deflection Wiener-Milenkovic parameter (relative to the undeflected orientation) at Node 4 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N4RDyr']             = False     # (-); Sectional angular/rotational deflection Wiener-Milenkovic parameter (relative to the undeflected orientation) at Node 4 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N4RDzr']             = False     # (-); Sectional angular/rotational deflection Wiener-Milenkovic parameter (relative to the undeflected orientation) at Node 4 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N5RDxr']             = False     # (-); Sectional angular/rotational deflection Wiener-Milenkovic parameter (relative to the undeflected orientation) at Node 5 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N5RDyr']             = False     # (-); Sectional angular/rotational deflection Wiener-Milenkovic parameter (relative to the undeflected orientation) at Node 5 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N5RDzr']             = False     # (-); Sectional angular/rotational deflection Wiener-Milenkovic parameter (relative to the undeflected orientation) at Node 5 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N6RDxr']             = False     # (-); Sectional angular/rotational deflection Wiener-Milenkovic parameter (relative to the undeflected orientation) at Node 6 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N6RDyr']             = False     # (-); Sectional angular/rotational deflection Wiener-Milenkovic parameter (relative to the undeflected orientation) at Node 6 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N6RDzr']             = False     # (-); Sectional angular/rotational deflection Wiener-Milenkovic parameter (relative to the undeflected orientation) at Node 6 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N7RDxr']             = False     # (-); Sectional angular/rotational deflection Wiener-Milenkovic parameter (relative to the undeflected orientation) at Node 7 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N7RDyr']             = False     # (-); Sectional angular/rotational deflection Wiener-Milenkovic parameter (relative to the undeflected orientation) at Node 7 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N7RDzr']             = False     # (-); Sectional angular/rotational deflection Wiener-Milenkovic parameter (relative to the undeflected orientation) at Node 7 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N8RDxr']             = False     # (-); Sectional angular/rotational deflection Wiener-Milenkovic parameter (relative to the undeflected orientation) at Node 8 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N8RDyr']             = False     # (-); Sectional angular/rotational deflection Wiener-Milenkovic parameter (relative to the undeflected orientation) at Node 8 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N8RDzr']             = False     # (-); Sectional angular/rotational deflection Wiener-Milenkovic parameter (relative to the undeflected orientation) at Node 8 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N9RDxr']             = False     # (-); Sectional angular/rotational deflection Wiener-Milenkovic parameter (relative to the undeflected orientation) at Node 9 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N9RDyr']             = False     # (-); Sectional angular/rotational deflection Wiener-Milenkovic parameter (relative to the undeflected orientation) at Node 9 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N9RDzr']             = False     # (-); Sectional angular/rotational deflection Wiener-Milenkovic parameter (relative to the undeflected orientation) at Node 9 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N1TVXg']             = False     # (m/s); Sectional translational velocities (absolute) at Node 1 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N1TVYg']             = False     # (m/s); Sectional translational velocities (absolute) at Node 1 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N1TVZg']             = False     # (m/s); Sectional translational velocities (absolute) at Node 1 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N2TVXg']             = False     # (m/s); Sectional translational velocities (absolute) at Node 2 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N2TVYg']             = False     # (m/s); Sectional translational velocities (absolute) at Node 2 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N2TVZg']             = False     # (m/s); Sectional translational velocities (absolute) at Node 2 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N3TVXg']             = False     # (m/s); Sectional translational velocities (absolute) at Node 3 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N3TVYg']             = False     # (m/s); Sectional translational velocities (absolute) at Node 3 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N3TVZg']             = False     # (m/s); Sectional translational velocities (absolute) at Node 3 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N4TVXg']             = False     # (m/s); Sectional translational velocities (absolute) at Node 4 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N4TVYg']             = False     # (m/s); Sectional translational velocities (absolute) at Node 4 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N4TVZg']             = False     # (m/s); Sectional translational velocities (absolute) at Node 4 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N5TVXg']             = False     # (m/s); Sectional translational velocities (absolute) at Node 5 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N5TVYg']             = False     # (m/s); Sectional translational velocities (absolute) at Node 5 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N5TVZg']             = False     # (m/s); Sectional translational velocities (absolute) at Node 5 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N6TVXg']             = False     # (m/s); Sectional translational velocities (absolute) at Node 6 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N6TVYg']             = False     # (m/s); Sectional translational velocities (absolute) at Node 6 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N6TVZg']             = False     # (m/s); Sectional translational velocities (absolute) at Node 6 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N7TVXg']             = False     # (m/s); Sectional translational velocities (absolute) at Node 7 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N7TVYg']             = False     # (m/s); Sectional translational velocities (absolute) at Node 7 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N7TVZg']             = False     # (m/s); Sectional translational velocities (absolute) at Node 7 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N8TVXg']             = False     # (m/s); Sectional translational velocities (absolute) at Node 8 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N8TVYg']             = False     # (m/s); Sectional translational velocities (absolute) at Node 8 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N8TVZg']             = False     # (m/s); Sectional translational velocities (absolute) at Node 8 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N9TVXg']             = False     # (m/s); Sectional translational velocities (absolute) at Node 9 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N9TVYg']             = False     # (m/s); Sectional translational velocities (absolute) at Node 9 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N9TVZg']             = False     # (m/s); Sectional translational velocities (absolute) at Node 9 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N1RVXg']             = False     # (deg/s); Sectional angular/rotational velocities (absolute) at Node 1 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N1RVYg']             = False     # (deg/s); Sectional angular/rotational velocities (absolute) at Node 1 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N1RVZg']             = False     # (deg/s); Sectional angular/rotational velocities (absolute) at Node 1 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N2RVXg']             = False     # (deg/s); Sectional angular/rotational velocities (absolute) at Node 2 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N2RVYg']             = False     # (deg/s); Sectional angular/rotational velocities (absolute) at Node 2 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N2RVZg']             = False     # (deg/s); Sectional angular/rotational velocities (absolute) at Node 2 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N3RVXg']             = False     # (deg/s); Sectional angular/rotational velocities (absolute) at Node 3 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N3RVYg']             = False     # (deg/s); Sectional angular/rotational velocities (absolute) at Node 3 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N3RVZg']             = False     # (deg/s); Sectional angular/rotational velocities (absolute) at Node 3 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N4RVXg']             = False     # (deg/s); Sectional angular/rotational velocities (absolute) at Node 4 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N4RVYg']             = False     # (deg/s); Sectional angular/rotational velocities (absolute) at Node 4 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N4RVZg']             = False     # (deg/s); Sectional angular/rotational velocities (absolute) at Node 4 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N5RVXg']             = False     # (deg/s); Sectional angular/rotational velocities (absolute) at Node 5 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N5RVYg']             = False     # (deg/s); Sectional angular/rotational velocities (absolute) at Node 5 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N5RVZg']             = False     # (deg/s); Sectional angular/rotational velocities (absolute) at Node 5 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N6RVXg']             = False     # (deg/s); Sectional angular/rotational velocities (absolute) at Node 6 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N6RVYg']             = False     # (deg/s); Sectional angular/rotational velocities (absolute) at Node 6 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N6RVZg']             = False     # (deg/s); Sectional angular/rotational velocities (absolute) at Node 6 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N7RVXg']             = False     # (deg/s); Sectional angular/rotational velocities (absolute) at Node 7 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N7RVYg']             = False     # (deg/s); Sectional angular/rotational velocities (absolute) at Node 7 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N7RVZg']             = False     # (deg/s); Sectional angular/rotational velocities (absolute) at Node 7 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N8RVXg']             = False     # (deg/s); Sectional angular/rotational velocities (absolute) at Node 8 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N8RVYg']             = False     # (deg/s); Sectional angular/rotational velocities (absolute) at Node 8 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N8RVZg']             = False     # (deg/s); Sectional angular/rotational velocities (absolute) at Node 8 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N9RVXg']             = False     # (deg/s); Sectional angular/rotational velocities (absolute) at Node 9 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N9RVYg']             = False     # (deg/s); Sectional angular/rotational velocities (absolute) at Node 9 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N9RVZg']             = False     # (deg/s); Sectional angular/rotational velocities (absolute) at Node 9 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N1TAXl']             = False     # (m/s^2); Sectional translational accelerations (absolute) at Node 1 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N1TAYl']             = False     # (m/s^2); Sectional translational accelerations (absolute) at Node 1 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N1TAZl']             = False     # (m/s^2); Sectional translational accelerations (absolute) at Node 1 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N2TAXl']             = False     # (m/s^2); Sectional translational accelerations (absolute) at Node 2 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N2TAYl']             = False     # (m/s^2); Sectional translational accelerations (absolute) at Node 2 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N2TAZl']             = False     # (m/s^2); Sectional translational accelerations (absolute) at Node 2 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N3TAXl']             = False     # (m/s^2); Sectional translational accelerations (absolute) at Node 3 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N3TAYl']             = False     # (m/s^2); Sectional translational accelerations (absolute) at Node 3 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N3TAZl']             = False     # (m/s^2); Sectional translational accelerations (absolute) at Node 3 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N4TAXl']             = False     # (m/s^2); Sectional translational accelerations (absolute) at Node 4 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N4TAYl']             = False     # (m/s^2); Sectional translational accelerations (absolute) at Node 4 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N4TAZl']             = False     # (m/s^2); Sectional translational accelerations (absolute) at Node 4 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N5TAXl']             = False     # (m/s^2); Sectional translational accelerations (absolute) at Node 5 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N5TAYl']             = False     # (m/s^2); Sectional translational accelerations (absolute) at Node 5 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N5TAZl']             = False     # (m/s^2); Sectional translational accelerations (absolute) at Node 5 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N6TAXl']             = False     # (m/s^2); Sectional translational accelerations (absolute) at Node 6 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N6TAYl']             = False     # (m/s^2); Sectional translational accelerations (absolute) at Node 6 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N6TAZl']             = False     # (m/s^2); Sectional translational accelerations (absolute) at Node 6 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N7TAXl']             = False     # (m/s^2); Sectional translational accelerations (absolute) at Node 7 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N7TAYl']             = False     # (m/s^2); Sectional translational accelerations (absolute) at Node 7 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N7TAZl']             = False     # (m/s^2); Sectional translational accelerations (absolute) at Node 7 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N8TAXl']             = False     # (m/s^2); Sectional translational accelerations (absolute) at Node 8 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N8TAYl']             = False     # (m/s^2); Sectional translational accelerations (absolute) at Node 8 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N8TAZl']             = False     # (m/s^2); Sectional translational accelerations (absolute) at Node 8 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N9TAXl']             = False     # (m/s^2); Sectional translational accelerations (absolute) at Node 9 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N9TAYl']             = False     # (m/s^2); Sectional translational accelerations (absolute) at Node 9 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N9TAZl']             = False     # (m/s^2); Sectional translational accelerations (absolute) at Node 9 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N1RAXl']             = False     # (deg/s^2); Sectional angular/rotational accelerations (absolute) at Node 1 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N1RAYl']             = False     # (deg/s^2); Sectional angular/rotational accelerations (absolute) at Node 1 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N1RAZl']             = False     # (deg/s^2); Sectional angular/rotational accelerations (absolute) at Node 1 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N2RAXl']             = False     # (deg/s^2); Sectional angular/rotational accelerations (absolute) at Node 2 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N2RAYl']             = False     # (deg/s^2); Sectional angular/rotational accelerations (absolute) at Node 2 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N2RAZl']             = False     # (deg/s^2); Sectional angular/rotational accelerations (absolute) at Node 2 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N3RAXl']             = False     # (deg/s^2); Sectional angular/rotational accelerations (absolute) at Node 3 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N3RAYl']             = False     # (deg/s^2); Sectional angular/rotational accelerations (absolute) at Node 3 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N3RAZl']             = False     # (deg/s^2); Sectional angular/rotational accelerations (absolute) at Node 3 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N4RAXl']             = False     # (deg/s^2); Sectional angular/rotational accelerations (absolute) at Node 4 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N4RAYl']             = False     # (deg/s^2); Sectional angular/rotational accelerations (absolute) at Node 4 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N4RAZl']             = False     # (deg/s^2); Sectional angular/rotational accelerations (absolute) at Node 4 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N5RAXl']             = False     # (deg/s^2); Sectional angular/rotational accelerations (absolute) at Node 5 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N5RAYl']             = False     # (deg/s^2); Sectional angular/rotational accelerations (absolute) at Node 5 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N5RAZl']             = False     # (deg/s^2); Sectional angular/rotational accelerations (absolute) at Node 5 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N6RAXl']             = False     # (deg/s^2); Sectional angular/rotational accelerations (absolute) at Node 6 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N6RAYl']             = False     # (deg/s^2); Sectional angular/rotational accelerations (absolute) at Node 6 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N6RAZl']             = False     # (deg/s^2); Sectional angular/rotational accelerations (absolute) at Node 6 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N7RAXl']             = False     # (deg/s^2); Sectional angular/rotational accelerations (absolute) at Node 7 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N7RAYl']             = False     # (deg/s^2); Sectional angular/rotational accelerations (absolute) at Node 7 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N7RAZl']             = False     # (deg/s^2); Sectional angular/rotational accelerations (absolute) at Node 7 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N8RAXl']             = False     # (deg/s^2); Sectional angular/rotational accelerations (absolute) at Node 8 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N8RAYl']             = False     # (deg/s^2); Sectional angular/rotational accelerations (absolute) at Node 8 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N8RAZl']             = False     # (deg/s^2); Sectional angular/rotational accelerations (absolute) at Node 8 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N9RAXl']             = False     # (deg/s^2); Sectional angular/rotational accelerations (absolute) at Node 9 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N9RAYl']             = False     # (deg/s^2); Sectional angular/rotational accelerations (absolute) at Node 9 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N9RAZl']             = False     # (deg/s^2); Sectional angular/rotational accelerations (absolute) at Node 9 expressed in l; l: a floating coordinate system local to the deflected beam

# Applied Loads
BeamDyn['N1PFxl']             = False     # (N); Applied point forces at Node 1 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N1PFyl']             = False     # (N); Applied point forces at Node 1 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N1PFzl']             = False     # (N); Applied point forces at Node 1 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N2PFxl']             = False     # (N); Applied point forces at Node 2 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N2PFyl']             = False     # (N); Applied point forces at Node 2 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N2PFzl']             = False     # (N); Applied point forces at Node 2 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N3PFxl']             = False     # (N); Applied point forces at Node 3 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N3PFyl']             = False     # (N); Applied point forces at Node 3 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N3PFzl']             = False     # (N); Applied point forces at Node 3 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N4PFxl']             = False     # (N); Applied point forces at Node 4 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N4PFyl']             = False     # (N); Applied point forces at Node 4 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N4PFzl']             = False     # (N); Applied point forces at Node 4 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N5PFxl']             = False     # (N); Applied point forces at Node 5 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N5PFyl']             = False     # (N); Applied point forces at Node 5 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N5PFzl']             = False     # (N); Applied point forces at Node 5 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N6PFxl']             = False     # (N); Applied point forces at Node 6 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N6PFyl']             = False     # (N); Applied point forces at Node 6 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N6PFzl']             = False     # (N); Applied point forces at Node 6 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N7PFxl']             = False     # (N); Applied point forces at Node 7 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N7PFyl']             = False     # (N); Applied point forces at Node 7 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N7PFzl']             = False     # (N); Applied point forces at Node 7 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N8PFxl']             = False     # (N); Applied point forces at Node 8 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N8PFyl']             = False     # (N); Applied point forces at Node 8 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N8PFzl']             = False     # (N); Applied point forces at Node 8 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N9PFxl']             = False     # (N); Applied point forces at Node 9 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N9PFyl']             = False     # (N); Applied point forces at Node 9 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N9PFzl']             = False     # (N); Applied point forces at Node 9 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N1PMxl']             = False     # (N-m); Applied point moments at Node 1 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N1PMyl']             = False     # (N-m); Applied point moments at Node 1 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N1PMzl']             = False     # (N-m); Applied point moments at Node 1 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N2PMxl']             = False     # (N-m); Applied point moments at Node 2 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N2PMyl']             = False     # (N-m); Applied point moments at Node 2 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N2PMzl']             = False     # (N-m); Applied point moments at Node 2 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N3PMxl']             = False     # (N-m); Applied point moments at Node 3 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N3PMyl']             = False     # (N-m); Applied point moments at Node 3 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N3PMzl']             = False     # (N-m); Applied point moments at Node 3 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N4PMxl']             = False     # (N-m); Applied point moments at Node 4 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N4PMyl']             = False     # (N-m); Applied point moments at Node 4 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N4PMzl']             = False     # (N-m); Applied point moments at Node 4 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N5PMxl']             = False     # (N-m); Applied point moments at Node 5 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N5PMyl']             = False     # (N-m); Applied point moments at Node 5 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N5PMzl']             = False     # (N-m); Applied point moments at Node 5 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N6PMxl']             = False     # (N-m); Applied point moments at Node 6 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N6PMyl']             = False     # (N-m); Applied point moments at Node 6 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N6PMzl']             = False     # (N-m); Applied point moments at Node 6 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N7PMxl']             = False     # (N-m); Applied point moments at Node 7 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N7PMyl']             = False     # (N-m); Applied point moments at Node 7 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N7PMzl']             = False     # (N-m); Applied point moments at Node 7 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N8PMxl']             = False     # (N-m); Applied point moments at Node 8 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N8PMyl']             = False     # (N-m); Applied point moments at Node 8 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N8PMzl']             = False     # (N-m); Applied point moments at Node 8 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N9PMxl']             = False     # (N-m); Applied point moments at Node 9 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N9PMyl']             = False     # (N-m); Applied point moments at Node 9 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N9PMzl']             = False     # (N-m); Applied point moments at Node 9 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N1DFxl']             = False     # (N/m); Applied distributed forces at Node 1 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N1DFyl']             = False     # (N/m); Applied distributed forces at Node 1 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N1DFzl']             = False     # (N/m); Applied distributed forces at Node 1 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N2DFxl']             = False     # (N/m); Applied distributed forces at Node 2 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N2DFyl']             = False     # (N/m); Applied distributed forces at Node 2 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N2DFzl']             = False     # (N/m); Applied distributed forces at Node 2 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N3DFxl']             = False     # (N/m); Applied distributed forces at Node 3 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N3DFyl']             = False     # (N/m); Applied distributed forces at Node 3 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N3DFzl']             = False     # (N/m); Applied distributed forces at Node 3 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N4DFxl']             = False     # (N/m); Applied distributed forces at Node 4 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N4DFyl']             = False     # (N/m); Applied distributed forces at Node 4 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N4DFzl']             = False     # (N/m); Applied distributed forces at Node 4 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N5DFxl']             = False     # (N/m); Applied distributed forces at Node 5 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N5DFyl']             = False     # (N/m); Applied distributed forces at Node 5 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N5DFzl']             = False     # (N/m); Applied distributed forces at Node 5 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N6DFxl']             = False     # (N/m); Applied distributed forces at Node 6 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N6DFyl']             = False     # (N/m); Applied distributed forces at Node 6 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N6DFzl']             = False     # (N/m); Applied distributed forces at Node 6 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N7DFxl']             = False     # (N/m); Applied distributed forces at Node 7 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N7DFyl']             = False     # (N/m); Applied distributed forces at Node 7 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N7DFzl']             = False     # (N/m); Applied distributed forces at Node 7 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N8DFxl']             = False     # (N/m); Applied distributed forces at Node 8 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N8DFyl']             = False     # (N/m); Applied distributed forces at Node 8 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N8DFzl']             = False     # (N/m); Applied distributed forces at Node 8 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N9DFxl']             = False     # (N/m); Applied distributed forces at Node 9 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N9DFyl']             = False     # (N/m); Applied distributed forces at Node 9 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N9DFzl']             = False     # (N/m); Applied distributed forces at Node 9 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N1DMxl']             = False     # (N-m/m); Applied distributed moments at Node 1 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N1DMyl']             = False     # (N-m/m); Applied distributed moments at Node 1 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N1DMzl']             = False     # (N-m/m); Applied distributed moments at Node 1 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N2DMxl']             = False     # (N-m/m); Applied distributed moments at Node 2 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N2DMyl']             = False     # (N-m/m); Applied distributed moments at Node 2 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N2DMzl']             = False     # (N-m/m); Applied distributed moments at Node 2 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N3DMxl']             = False     # (N-m/m); Applied distributed moments at Node 3 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N3DMyl']             = False     # (N-m/m); Applied distributed moments at Node 3 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N3DMzl']             = False     # (N-m/m); Applied distributed moments at Node 3 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N4DMxl']             = False     # (N-m/m); Applied distributed moments at Node 4 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N4DMyl']             = False     # (N-m/m); Applied distributed moments at Node 4 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N4DMzl']             = False     # (N-m/m); Applied distributed moments at Node 4 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N5DMxl']             = False     # (N-m/m); Applied distributed moments at Node 5 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N5DMyl']             = False     # (N-m/m); Applied distributed moments at Node 5 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N5DMzl']             = False     # (N-m/m); Applied distributed moments at Node 5 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N6DMxl']             = False     # (N-m/m); Applied distributed moments at Node 6 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N6DMyl']             = False     # (N-m/m); Applied distributed moments at Node 6 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N6DMzl']             = False     # (N-m/m); Applied distributed moments at Node 6 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N7DMxl']             = False     # (N-m/m); Applied distributed moments at Node 7 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N7DMyl']             = False     # (N-m/m); Applied distributed moments at Node 7 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N7DMzl']             = False     # (N-m/m); Applied distributed moments at Node 7 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N8DMxl']             = False     # (N-m/m); Applied distributed moments at Node 8 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N8DMyl']             = False     # (N-m/m); Applied distributed moments at Node 8 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N8DMzl']             = False     # (N-m/m); Applied distributed moments at Node 8 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N9DMxl']             = False     # (N-m/m); Applied distributed moments at Node 9 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N9DMyl']             = False     # (N-m/m); Applied distributed moments at Node 9 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N9DMzl']             = False     # (N-m/m); Applied distributed moments at Node 9 expressed in l; l: a floating coordinate system local to the deflected beam

# Pitch Actuator
BeamDyn['PAngInp']            = False     # (deg); Pitch angle input; 
BeamDyn['PAngAct']            = False     # (deg); Actual pitch angle ; 
BeamDyn['PRatAct']            = False     # (deg/s); Actual pitch rate; 
BeamDyn['PAccAct']            = False     # (deg/s^2); Actual pitch acceleration; 


""" ServoDyn """
ServoDyn = {}

# Pitch Control
ServoDyn['BlPitchC1']         = False     # (deg); Blade 1 pitch angle command; Positive towards feather about the minus zc1- and minus zb1-axes
ServoDyn['BlPitchC2']         = False     # (deg); Blade 2 pitch angle command; Positive towards feather about the minus zc2- and minus zb2-axes
ServoDyn['BlPitchC3']         = False     # (deg); Blade 3 pitch angle command; Positive towards feather about the minus zc3- and minus zb3-axes

# Generator and Torque Control
ServoDyn['GenTq']             = False     # (kN m); Electrical generator torque; Positive reflects power extracted and negative represents a motoring-up situation (power input)
ServoDyn['GenPwr']            = False     # (kW); Electrical generator power; Same sign as GenTq

# High Speed Shaft Brake
ServoDyn['HSSBrTqC']          = False     # (kN m); High-speed shaft brake torque command (i.e., the commanded moment applied to the high-speed shaft by the brake); Always positive (indicating dissipation of power)

# Nacelle Yaw Control
ServoDyn['YawMomCom']         = False     # (kN m); Nacelle yaw moment command; About the zn- and zp-axes
ServoDyn['YawMom']            = False     # (kN m); Nacelle yaw moment command; About the zn- and zp-axes

# Nacelle Tuned Mass Damper (TMD)
ServoDyn['NTMD_XQ']           = False     # (m); Nacelle X TMD position (displacement); Relative to rest position
ServoDyn['NTMD_XQD']          = False     # (m/s); Nacelle X TMD velocity; Relative to nacelle
ServoDyn['NTMD_YQ']           = False     # (m); Nacelle Y TMD position (displacement); Relative to rest position
ServoDyn['NTMD_YQD']          = False     # (m/s); Nacelle Y TMD velocity; Relative to nacelle

# Tower Tuned Mass Damper (TMD)
ServoDyn['TTMD_XQ']           = False     # (m); Tower X TMD position (displacement); Relative to rest position
ServoDyn['TTMD_XQD']          = False     # (m/s); Tower X TMD velocity; Relative to tower
ServoDyn['TTMD_YQ']           = False     # (m); Tower Y TMD position (displacement); Relative to rest position
ServoDyn['TTMD_YQD']          = False     # (m/s); Tower Y TMD velocity; Relative to tower

# Flap outputs
ServoDyn['BLFLAP1']          = False     # (m/s); Tower Y TMD velocity; Relative to tower
ServoDyn['BLFLAP2']          = False     # (m/s); Tower Y TMD velocity; Relative to tower
ServoDyn['BLFLAP3']          = False     # (m/s); Tower Y TMD velocity; Relative to tower

""" AeroDyn """
AeroDyn = {}

# Tower
AeroDyn['TwN1VUndx']          = False     # (m/s); Undisturbed x-component wind velocity at Tw node 1; local tower coordinate system
AeroDyn['TwN1VUndy']          = False     # (m/s); Undisturbed y-component wind velocity at Tw node 1; local tower coordinate system
AeroDyn['TwN1VUndz']          = False     # (m/s); Undisturbed z-component wind velocity at Tw node 1; local tower coordinate system
AeroDyn['TwN2VUndx']          = False     # (m/s); Undisturbed x-component wind velocity at Tw node 2; local tower coordinate system
AeroDyn['TwN2VUndy']          = False     # (m/s); Undisturbed y-component wind velocity at Tw node 2; local tower coordinate system
AeroDyn['TwN2VUndz']          = False     # (m/s); Undisturbed z-component wind velocity at Tw node 2; local tower coordinate system
AeroDyn['TwN3VUndx']          = False     # (m/s); Undisturbed x-component wind velocity at Tw node 3; local tower coordinate system
AeroDyn['TwN3VUndy']          = False     # (m/s); Undisturbed y-component wind velocity at Tw node 3; local tower coordinate system
AeroDyn['TwN3VUndz']          = False     # (m/s); Undisturbed z-component wind velocity at Tw node 3; local tower coordinate system
AeroDyn['TwN4VUndx']          = False     # (m/s); Undisturbed x-component wind velocity at Tw node 4; local tower coordinate system
AeroDyn['TwN4VUndy']          = False     # (m/s); Undisturbed y-component wind velocity at Tw node 4; local tower coordinate system
AeroDyn['TwN4VUndz']          = False     # (m/s); Undisturbed z-component wind velocity at Tw node 4; local tower coordinate system
AeroDyn['TwN5VUndx']          = False     # (m/s); Undisturbed x-component wind velocity at Tw node 5; local tower coordinate system
AeroDyn['TwN5VUndy']          = False     # (m/s); Undisturbed y-component wind velocity at Tw node 5; local tower coordinate system
AeroDyn['TwN5VUndz']          = False     # (m/s); Undisturbed z-component wind velocity at Tw node 5; local tower coordinate system
AeroDyn['TwN6VUndx']          = False     # (m/s); Undisturbed x-component wind velocity at Tw node 6; local tower coordinate system
AeroDyn['TwN6VUndy']          = False     # (m/s); Undisturbed y-component wind velocity at Tw node 6; local tower coordinate system
AeroDyn['TwN6VUndz']          = False     # (m/s); Undisturbed z-component wind velocity at Tw node 6; local tower coordinate system
AeroDyn['TwN7VUndx']          = False     # (m/s); Undisturbed x-component wind velocity at Tw node 7; local tower coordinate system
AeroDyn['TwN7VUndy']          = False     # (m/s); Undisturbed y-component wind velocity at Tw node 7; local tower coordinate system
AeroDyn['TwN7VUndz']          = False     # (m/s); Undisturbed z-component wind velocity at Tw node 7; local tower coordinate system
AeroDyn['TwN8VUndx']          = False     # (m/s); Undisturbed x-component wind velocity at Tw node 8; local tower coordinate system
AeroDyn['TwN8VUndy']          = False     # (m/s); Undisturbed y-component wind velocity at Tw node 8; local tower coordinate system
AeroDyn['TwN8VUndz']          = False     # (m/s); Undisturbed z-component wind velocity at Tw node 8; local tower coordinate system
AeroDyn['TwN9VUndx']          = False     # (m/s); Undisturbed x-component wind velocity at Tw node 9; local tower coordinate system
AeroDyn['TwN9VUndy']          = False     # (m/s); Undisturbed y-component wind velocity at Tw node 9; local tower coordinate system
AeroDyn['TwN9VUndz']          = False     # (m/s); Undisturbed z-component wind velocity at Tw node 9; local tower coordinate system
AeroDyn['TwN1STVx']           = False     # (m/s); Structural translational velocity x-component at Tw node 1; local tower coordinate system
AeroDyn['TwN1STVy']           = False     # (m/s); Structural translational velocity y-component at Tw node 1; local tower coordinate system
AeroDyn['TwN1STVz']           = False     # (m/s); Structural translational velocity z-component at Tw node 1; local tower coordinate system
AeroDyn['TwN2STVx']           = False     # (m/s); Structural translational velocity x-component at Tw node 2; local tower coordinate system
AeroDyn['TwN2STVy']           = False     # (m/s); Structural translational velocity y-component at Tw node 2; local tower coordinate system
AeroDyn['TwN2STVz']           = False     # (m/s); Structural translational velocity z-component at Tw node 2; local tower coordinate system
AeroDyn['TwN3STVx']           = False     # (m/s); Structural translational velocity x-component at Tw node 3; local tower coordinate system
AeroDyn['TwN3STVy']           = False     # (m/s); Structural translational velocity y-component at Tw node 3; local tower coordinate system
AeroDyn['TwN3STVz']           = False     # (m/s); Structural translational velocity z-component at Tw node 3; local tower coordinate system
AeroDyn['TwN4STVx']           = False     # (m/s); Structural translational velocity x-component at Tw node 4; local tower coordinate system
AeroDyn['TwN4STVy']           = False     # (m/s); Structural translational velocity y-component at Tw node 4; local tower coordinate system
AeroDyn['TwN4STVz']           = False     # (m/s); Structural translational velocity z-component at Tw node 4; local tower coordinate system
AeroDyn['TwN5STVx']           = False     # (m/s); Structural translational velocity x-component at Tw node 5; local tower coordinate system
AeroDyn['TwN5STVy']           = False     # (m/s); Structural translational velocity y-component at Tw node 5; local tower coordinate system
AeroDyn['TwN5STVz']           = False     # (m/s); Structural translational velocity z-component at Tw node 5; local tower coordinate system
AeroDyn['TwN6STVx']           = False     # (m/s); Structural translational velocity x-component at Tw node 6; local tower coordinate system
AeroDyn['TwN6STVy']           = False     # (m/s); Structural translational velocity y-component at Tw node 6; local tower coordinate system
AeroDyn['TwN6STVz']           = False     # (m/s); Structural translational velocity z-component at Tw node 6; local tower coordinate system
AeroDyn['TwN7STVx']           = False     # (m/s); Structural translational velocity x-component at Tw node 7; local tower coordinate system
AeroDyn['TwN7STVy']           = False     # (m/s); Structural translational velocity y-component at Tw node 7; local tower coordinate system
AeroDyn['TwN7STVz']           = False     # (m/s); Structural translational velocity z-component at Tw node 7; local tower coordinate system
AeroDyn['TwN8STVx']           = False     # (m/s); Structural translational velocity x-component at Tw node 8; local tower coordinate system
AeroDyn['TwN8STVy']           = False     # (m/s); Structural translational velocity y-component at Tw node 8; local tower coordinate system
AeroDyn['TwN8STVz']           = False     # (m/s); Structural translational velocity z-component at Tw node 8; local tower coordinate system
AeroDyn['TwN9STVx']           = False     # (m/s); Structural translational velocity x-component at Tw node 9; local tower coordinate system
AeroDyn['TwN9STVy']           = False     # (m/s); Structural translational velocity y-component at Tw node 9; local tower coordinate system
AeroDyn['TwN9STVz']           = False     # (m/s); Structural translational velocity z-component at Tw node 9; local tower coordinate system
AeroDyn['TwN1Vrel']           = False     # (m/s); Relative wind speed at Tw node 1; 
AeroDyn['TwN2Vrel']           = False     # (m/s); Relative wind speed at Tw node 2; 
AeroDyn['TwN3Vrel']           = False     # (m/s); Relative wind speed at Tw node 3; 
AeroDyn['TwN4Vrel']           = False     # (m/s); Relative wind speed at Tw node 4; 
AeroDyn['TwN5Vrel']           = False     # (m/s); Relative wind speed at Tw node 5; 
AeroDyn['TwN6Vrel']           = False     # (m/s); Relative wind speed at Tw node 6; 
AeroDyn['TwN7Vrel']           = False     # (m/s); Relative wind speed at Tw node 7; 
AeroDyn['TwN8Vrel']           = False     # (m/s); Relative wind speed at Tw node 8; 
AeroDyn['TwN9Vrel']           = False     # (m/s); Relative wind speed at Tw node 9; 
AeroDyn['TwN1DynP']           = False     # (Pa); Dynamic Pressure at Tw node 1; 
AeroDyn['TwN2DynP']           = False     # (Pa); Dynamic Pressure at Tw node 2; 
AeroDyn['TwN3DynP']           = False     # (Pa); Dynamic Pressure at Tw node 3; 
AeroDyn['TwN4DynP']           = False     # (Pa); Dynamic Pressure at Tw node 4; 
AeroDyn['TwN5DynP']           = False     # (Pa); Dynamic Pressure at Tw node 5; 
AeroDyn['TwN6DynP']           = False     # (Pa); Dynamic Pressure at Tw node 6; 
AeroDyn['TwN7DynP']           = False     # (Pa); Dynamic Pressure at Tw node 7; 
AeroDyn['TwN8DynP']           = False     # (Pa); Dynamic Pressure at Tw node 8; 
AeroDyn['TwN9DynP']           = False     # (Pa); Dynamic Pressure at Tw node 9; 
AeroDyn['TwN1Re']             = False     # (-); Reynolds number (in millions) at Tw node 1; 
AeroDyn['TwN2Re']             = False     # (-); Reynolds number (in millions) at Tw node 2; 
AeroDyn['TwN3Re']             = False     # (-); Reynolds number (in millions) at Tw node 3; 
AeroDyn['TwN4Re']             = False     # (-); Reynolds number (in millions) at Tw node 4; 
AeroDyn['TwN5Re']             = False     # (-); Reynolds number (in millions) at Tw node 5; 
AeroDyn['TwN6Re']             = False     # (-); Reynolds number (in millions) at Tw node 6; 
AeroDyn['TwN7Re']             = False     # (-); Reynolds number (in millions) at Tw node 7; 
AeroDyn['TwN8Re']             = False     # (-); Reynolds number (in millions) at Tw node 8; 
AeroDyn['TwN9Re']             = False     # (-); Reynolds number (in millions) at Tw node 9; 
AeroDyn['TwN1M']              = False     # (-); Mach number at Tw node 1; 
AeroDyn['TwN2M']              = False     # (-); Mach number at Tw node 2; 
AeroDyn['TwN3M']              = False     # (-); Mach number at Tw node 3; 
AeroDyn['TwN4M']              = False     # (-); Mach number at Tw node 4; 
AeroDyn['TwN5M']              = False     # (-); Mach number at Tw node 5; 
AeroDyn['TwN6M']              = False     # (-); Mach number at Tw node 6; 
AeroDyn['TwN7M']              = False     # (-); Mach number at Tw node 7; 
AeroDyn['TwN8M']              = False     # (-); Mach number at Tw node 8; 
AeroDyn['TwN9M']              = False     # (-); Mach number at Tw node 9; 
AeroDyn['TwN1Fdx']            = False     # (N/m); x-component of drag force per unit length at Tw node 1; local tower coordinate system
AeroDyn['TwN2Fdx']            = False     # (N/m); x-component of drag force per unit length at Tw node 2; local tower coordinate system
AeroDyn['TwN3Fdx']            = False     # (N/m); x-component of drag force per unit length at Tw node 3; local tower coordinate system
AeroDyn['TwN4Fdx']            = False     # (N/m); x-component of drag force per unit length at Tw node 4; local tower coordinate system
AeroDyn['TwN5Fdx']            = False     # (N/m); x-component of drag force per unit length at Tw node 5; local tower coordinate system
AeroDyn['TwN6Fdx']            = False     # (N/m); x-component of drag force per unit length at Tw node 6; local tower coordinate system
AeroDyn['TwN7Fdx']            = False     # (N/m); x-component of drag force per unit length at Tw node 7; local tower coordinate system
AeroDyn['TwN8Fdx']            = False     # (N/m); x-component of drag force per unit length at Tw node 8; local tower coordinate system
AeroDyn['TwN9Fdx']            = False     # (N/m); x-component of drag force per unit length at Tw node 9; local tower coordinate system
AeroDyn['TwN1Fdy']            = False     # (N/m); y-component of drag force per unit length at Tw node 1; local tower coordinate system
AeroDyn['TwN2Fdy']            = False     # (N/m); y-component of drag force per unit length at Tw node 2; local tower coordinate system
AeroDyn['TwN3Fdy']            = False     # (N/m); y-component of drag force per unit length at Tw node 3; local tower coordinate system
AeroDyn['TwN4Fdy']            = False     # (N/m); y-component of drag force per unit length at Tw node 4; local tower coordinate system
AeroDyn['TwN5Fdy']            = False     # (N/m); y-component of drag force per unit length at Tw node 5; local tower coordinate system
AeroDyn['TwN6Fdy']            = False     # (N/m); y-component of drag force per unit length at Tw node 6; local tower coordinate system
AeroDyn['TwN7Fdy']            = False     # (N/m); y-component of drag force per unit length at Tw node 7; local tower coordinate system
AeroDyn['TwN8Fdy']            = False     # (N/m); y-component of drag force per unit length at Tw node 8; local tower coordinate system
AeroDyn['TwN9Fdy']            = False     # (N/m); y-component of drag force per unit length at Tw node 9; local tower coordinate system

# Blade
AeroDyn['B1Azimuth']          = False     # (deg); Azimuth angle of blade 1; 
AeroDyn['B2Azimuth']          = False     # (deg); Azimuth angle of blade 2; 
AeroDyn['B3Azimuth']          = False     # (deg); Azimuth angle of blade 3; 
AeroDyn['B1Pitch']            = False     # (deg); Pitch angle of blade 1; 
AeroDyn['B2Pitch']            = False     # (deg); Pitch angle of blade 2; 
AeroDyn['B3Pitch']            = False     # (deg); Pitch angle of blade 3; 
AeroDyn['B1N1VUndx']          = False     # (m/s); x-component of undisturbed wind velocity at Blade 1, Node 1; local blade coordinate system
AeroDyn['B1N2VUndx']          = False     # (m/s); x-component of undisturbed wind velocity at Blade 1, Node 2; local blade coordinate system
AeroDyn['B1N3VUndx']          = False     # (m/s); x-component of undisturbed wind velocity at Blade 1, Node 3; local blade coordinate system
AeroDyn['B1N4VUndx']          = False     # (m/s); x-component of undisturbed wind velocity at Blade 1, Node 4; local blade coordinate system
AeroDyn['B1N5VUndx']          = False     # (m/s); x-component of undisturbed wind velocity at Blade 1, Node 5; local blade coordinate system
AeroDyn['B1N6VUndx']          = False     # (m/s); x-component of undisturbed wind velocity at Blade 1, Node 6; local blade coordinate system
AeroDyn['B1N7VUndx']          = False     # (m/s); x-component of undisturbed wind velocity at Blade 1, Node 7; local blade coordinate system
AeroDyn['B1N8VUndx']          = False     # (m/s); x-component of undisturbed wind velocity at Blade 1, Node 8; local blade coordinate system
AeroDyn['B1N9VUndx']          = False     # (m/s); x-component of undisturbed wind velocity at Blade 1, Node 9; local blade coordinate system
AeroDyn['B1N1VUndy']          = False     # (m/s); y-component of undisturbed wind velocity at Blade 1, Node 1; local blade coordinate system
AeroDyn['B1N2VUndy']          = False     # (m/s); y-component of undisturbed wind velocity at Blade 1, Node 2; local blade coordinate system
AeroDyn['B1N3VUndy']          = False     # (m/s); y-component of undisturbed wind velocity at Blade 1, Node 3; local blade coordinate system
AeroDyn['B1N4VUndy']          = False     # (m/s); y-component of undisturbed wind velocity at Blade 1, Node 4; local blade coordinate system
AeroDyn['B1N5VUndy']          = False     # (m/s); y-component of undisturbed wind velocity at Blade 1, Node 5; local blade coordinate system
AeroDyn['B1N6VUndy']          = False     # (m/s); y-component of undisturbed wind velocity at Blade 1, Node 6; local blade coordinate system
AeroDyn['B1N7VUndy']          = False     # (m/s); y-component of undisturbed wind velocity at Blade 1, Node 7; local blade coordinate system
AeroDyn['B1N8VUndy']          = False     # (m/s); y-component of undisturbed wind velocity at Blade 1, Node 8; local blade coordinate system
AeroDyn['B1N9VUndy']          = False     # (m/s); y-component of undisturbed wind velocity at Blade 1, Node 9; local blade coordinate system
AeroDyn['B1N1VUndz']          = False     # (m/s); z-component of undisturbed wind velocity at Blade 1, Node 1; local blade coordinate system
AeroDyn['B1N2VUndz']          = False     # (m/s); z-component of undisturbed wind velocity at Blade 1, Node 2; local blade coordinate system
AeroDyn['B1N3VUndz']          = False     # (m/s); z-component of undisturbed wind velocity at Blade 1, Node 3; local blade coordinate system
AeroDyn['B1N4VUndz']          = False     # (m/s); z-component of undisturbed wind velocity at Blade 1, Node 4; local blade coordinate system
AeroDyn['B1N5VUndz']          = False     # (m/s); z-component of undisturbed wind velocity at Blade 1, Node 5; local blade coordinate system
AeroDyn['B1N6VUndz']          = False     # (m/s); z-component of undisturbed wind velocity at Blade 1, Node 6; local blade coordinate system
AeroDyn['B1N7VUndz']          = False     # (m/s); z-component of undisturbed wind velocity at Blade 1, Node 7; local blade coordinate system
AeroDyn['B1N8VUndz']          = False     # (m/s); z-component of undisturbed wind velocity at Blade 1, Node 8; local blade coordinate system
AeroDyn['B1N9VUndz']          = False     # (m/s); z-component of undisturbed wind velocity at Blade 1, Node 9; local blade coordinate system
AeroDyn['B2N1VUndx']          = False     # (m/s); x-component of undisturbed wind velocity at Blade 2, Node 1; local blade coordinate system
AeroDyn['B2N2VUndx']          = False     # (m/s); x-component of undisturbed wind velocity at Blade 2, Node 2; local blade coordinate system
AeroDyn['B2N3VUndx']          = False     # (m/s); x-component of undisturbed wind velocity at Blade 2, Node 3; local blade coordinate system
AeroDyn['B2N4VUndx']          = False     # (m/s); x-component of undisturbed wind velocity at Blade 2, Node 4; local blade coordinate system
AeroDyn['B2N5VUndx']          = False     # (m/s); x-component of undisturbed wind velocity at Blade 2, Node 5; local blade coordinate system
AeroDyn['B2N6VUndx']          = False     # (m/s); x-component of undisturbed wind velocity at Blade 2, Node 6; local blade coordinate system
AeroDyn['B2N7VUndx']          = False     # (m/s); x-component of undisturbed wind velocity at Blade 2, Node 7; local blade coordinate system
AeroDyn['B2N8VUndx']          = False     # (m/s); x-component of undisturbed wind velocity at Blade 2, Node 8; local blade coordinate system
AeroDyn['B2N9VUndx']          = False     # (m/s); x-component of undisturbed wind velocity at Blade 2, Node 9; local blade coordinate system
AeroDyn['B2N1VUndy']          = False     # (m/s); y-component of undisturbed wind velocity at Blade 2, Node 1; local blade coordinate system
AeroDyn['B2N2VUndy']          = False     # (m/s); y-component of undisturbed wind velocity at Blade 2, Node 2; local blade coordinate system
AeroDyn['B2N3VUndy']          = False     # (m/s); y-component of undisturbed wind velocity at Blade 2, Node 3; local blade coordinate system
AeroDyn['B2N4VUndy']          = False     # (m/s); y-component of undisturbed wind velocity at Blade 2, Node 4; local blade coordinate system
AeroDyn['B2N5VUndy']          = False     # (m/s); y-component of undisturbed wind velocity at Blade 2, Node 5; local blade coordinate system
AeroDyn['B2N6VUndy']          = False     # (m/s); y-component of undisturbed wind velocity at Blade 2, Node 6; local blade coordinate system
AeroDyn['B2N7VUndy']          = False     # (m/s); y-component of undisturbed wind velocity at Blade 2, Node 7; local blade coordinate system
AeroDyn['B2N8VUndy']          = False     # (m/s); y-component of undisturbed wind velocity at Blade 2, Node 8; local blade coordinate system
AeroDyn['B2N9VUndy']          = False     # (m/s); y-component of undisturbed wind velocity at Blade 2, Node 9; local blade coordinate system
AeroDyn['B2N1VUndz']          = False     # (m/s); z-component of undisturbed wind velocity at Blade 2, Node 1; local blade coordinate system
AeroDyn['B2N2VUndz']          = False     # (m/s); z-component of undisturbed wind velocity at Blade 2, Node 2; local blade coordinate system
AeroDyn['B2N3VUndz']          = False     # (m/s); z-component of undisturbed wind velocity at Blade 2, Node 3; local blade coordinate system
AeroDyn['B2N4VUndz']          = False     # (m/s); z-component of undisturbed wind velocity at Blade 2, Node 4; local blade coordinate system
AeroDyn['B2N5VUndz']          = False     # (m/s); z-component of undisturbed wind velocity at Blade 2, Node 5; local blade coordinate system
AeroDyn['B2N6VUndz']          = False     # (m/s); z-component of undisturbed wind velocity at Blade 2, Node 6; local blade coordinate system
AeroDyn['B2N7VUndz']          = False     # (m/s); z-component of undisturbed wind velocity at Blade 2, Node 7; local blade coordinate system
AeroDyn['B2N8VUndz']          = False     # (m/s); z-component of undisturbed wind velocity at Blade 2, Node 8; local blade coordinate system
AeroDyn['B2N9VUndz']          = False     # (m/s); z-component of undisturbed wind velocity at Blade 2, Node 9; local blade coordinate system
AeroDyn['B3N1VUndx']          = False     # (m/s); x-component of undisturbed wind velocity at Blade 3, Node 1; local blade coordinate system
AeroDyn['B3N2VUndx']          = False     # (m/s); x-component of undisturbed wind velocity at Blade 3, Node 2; local blade coordinate system
AeroDyn['B3N3VUndx']          = False     # (m/s); x-component of undisturbed wind velocity at Blade 3, Node 3; local blade coordinate system
AeroDyn['B3N4VUndx']          = False     # (m/s); x-component of undisturbed wind velocity at Blade 3, Node 4; local blade coordinate system
AeroDyn['B3N5VUndx']          = False     # (m/s); x-component of undisturbed wind velocity at Blade 3, Node 5; local blade coordinate system
AeroDyn['B3N6VUndx']          = False     # (m/s); x-component of undisturbed wind velocity at Blade 3, Node 6; local blade coordinate system
AeroDyn['B3N7VUndx']          = False     # (m/s); x-component of undisturbed wind velocity at Blade 3, Node 7; local blade coordinate system
AeroDyn['B3N8VUndx']          = False     # (m/s); x-component of undisturbed wind velocity at Blade 3, Node 8; local blade coordinate system
AeroDyn['B3N9VUndx']          = False     # (m/s); x-component of undisturbed wind velocity at Blade 3, Node 9; local blade coordinate system
AeroDyn['B3N1VUndy']          = False     # (m/s); y-component of undisturbed wind velocity at Blade 3, Node 1; local blade coordinate system
AeroDyn['B3N2VUndy']          = False     # (m/s); y-component of undisturbed wind velocity at Blade 3, Node 2; local blade coordinate system
AeroDyn['B3N3VUndy']          = False     # (m/s); y-component of undisturbed wind velocity at Blade 3, Node 3; local blade coordinate system
AeroDyn['B3N4VUndy']          = False     # (m/s); y-component of undisturbed wind velocity at Blade 3, Node 4; local blade coordinate system
AeroDyn['B3N5VUndy']          = False     # (m/s); y-component of undisturbed wind velocity at Blade 3, Node 5; local blade coordinate system
AeroDyn['B3N6VUndy']          = False     # (m/s); y-component of undisturbed wind velocity at Blade 3, Node 6; local blade coordinate system
AeroDyn['B3N7VUndy']          = False     # (m/s); y-component of undisturbed wind velocity at Blade 3, Node 7; local blade coordinate system
AeroDyn['B3N8VUndy']          = False     # (m/s); y-component of undisturbed wind velocity at Blade 3, Node 8; local blade coordinate system
AeroDyn['B3N9VUndy']          = False     # (m/s); y-component of undisturbed wind velocity at Blade 3, Node 9; local blade coordinate system
AeroDyn['B3N1VUndz']          = False     # (m/s); z-component of undisturbed wind velocity at Blade 3, Node 1; local blade coordinate system
AeroDyn['B3N2VUndz']          = False     # (m/s); z-component of undisturbed wind velocity at Blade 3, Node 2; local blade coordinate system
AeroDyn['B3N3VUndz']          = False     # (m/s); z-component of undisturbed wind velocity at Blade 3, Node 3; local blade coordinate system
AeroDyn['B3N4VUndz']          = False     # (m/s); z-component of undisturbed wind velocity at Blade 3, Node 4; local blade coordinate system
AeroDyn['B3N5VUndz']          = False     # (m/s); z-component of undisturbed wind velocity at Blade 3, Node 5; local blade coordinate system
AeroDyn['B3N6VUndz']          = False     # (m/s); z-component of undisturbed wind velocity at Blade 3, Node 6; local blade coordinate system
AeroDyn['B3N7VUndz']          = False     # (m/s); z-component of undisturbed wind velocity at Blade 3, Node 7; local blade coordinate system
AeroDyn['B3N8VUndz']          = False     # (m/s); z-component of undisturbed wind velocity at Blade 3, Node 8; local blade coordinate system
AeroDyn['B3N9VUndz']          = False     # (m/s); z-component of undisturbed wind velocity at Blade 3, Node 9; local blade coordinate system
AeroDyn['B1N1VDisx']          = False     # (m/s); x-component of disturbed wind velocity at Blade 1, Node 1; local blade coordinate system
AeroDyn['B1N2VDisx']          = False     # (m/s); x-component of disturbed wind velocity at Blade 1, Node 2; local blade coordinate system
AeroDyn['B1N3VDisx']          = False     # (m/s); x-component of disturbed wind velocity at Blade 1, Node 3; local blade coordinate system
AeroDyn['B1N4VDisx']          = False     # (m/s); x-component of disturbed wind velocity at Blade 1, Node 4; local blade coordinate system
AeroDyn['B1N5VDisx']          = False     # (m/s); x-component of disturbed wind velocity at Blade 1, Node 5; local blade coordinate system
AeroDyn['B1N6VDisx']          = False     # (m/s); x-component of disturbed wind velocity at Blade 1, Node 6; local blade coordinate system
AeroDyn['B1N7VDisx']          = False     # (m/s); x-component of disturbed wind velocity at Blade 1, Node 7; local blade coordinate system
AeroDyn['B1N8VDisx']          = False     # (m/s); x-component of disturbed wind velocity at Blade 1, Node 8; local blade coordinate system
AeroDyn['B1N9VDisx']          = False     # (m/s); x-component of disturbed wind velocity at Blade 1, Node 9; local blade coordinate system
AeroDyn['B1N1VDisy']          = False     # (m/s); y-component of disturbed wind velocity at Blade 1, Node 1; local blade coordinate system
AeroDyn['B1N2VDisy']          = False     # (m/s); y-component of disturbed wind velocity at Blade 1, Node 2; local blade coordinate system
AeroDyn['B1N3VDisy']          = False     # (m/s); y-component of disturbed wind velocity at Blade 1, Node 3; local blade coordinate system
AeroDyn['B1N4VDisy']          = False     # (m/s); y-component of disturbed wind velocity at Blade 1, Node 4; local blade coordinate system
AeroDyn['B1N5VDisy']          = False     # (m/s); y-component of disturbed wind velocity at Blade 1, Node 5; local blade coordinate system
AeroDyn['B1N6VDisy']          = False     # (m/s); y-component of disturbed wind velocity at Blade 1, Node 6; local blade coordinate system
AeroDyn['B1N7VDisy']          = False     # (m/s); y-component of disturbed wind velocity at Blade 1, Node 7; local blade coordinate system
AeroDyn['B1N8VDisy']          = False     # (m/s); y-component of disturbed wind velocity at Blade 1, Node 8; local blade coordinate system
AeroDyn['B1N9VDisy']          = False     # (m/s); y-component of disturbed wind velocity at Blade 1, Node 9; local blade coordinate system
AeroDyn['B1N1VDisz']          = False     # (m/s); z-component of disturbed wind velocity at Blade 1, Node 1; local blade coordinate system
AeroDyn['B1N2VDisz']          = False     # (m/s); z-component of disturbed wind velocity at Blade 1, Node 2; local blade coordinate system
AeroDyn['B1N3VDisz']          = False     # (m/s); z-component of disturbed wind velocity at Blade 1, Node 3; local blade coordinate system
AeroDyn['B1N4VDisz']          = False     # (m/s); z-component of disturbed wind velocity at Blade 1, Node 4; local blade coordinate system
AeroDyn['B1N5VDisz']          = False     # (m/s); z-component of disturbed wind velocity at Blade 1, Node 5; local blade coordinate system
AeroDyn['B1N6VDisz']          = False     # (m/s); z-component of disturbed wind velocity at Blade 1, Node 6; local blade coordinate system
AeroDyn['B1N7VDisz']          = False     # (m/s); z-component of disturbed wind velocity at Blade 1, Node 7; local blade coordinate system
AeroDyn['B1N8VDisz']          = False     # (m/s); z-component of disturbed wind velocity at Blade 1, Node 8; local blade coordinate system
AeroDyn['B1N9VDisz']          = False     # (m/s); z-component of disturbed wind velocity at Blade 1, Node 9; local blade coordinate system
AeroDyn['B2N1VDisx']          = False     # (m/s); x-component of disturbed wind velocity at Blade 2, Node 1; local blade coordinate system
AeroDyn['B2N2VDisx']          = False     # (m/s); x-component of disturbed wind velocity at Blade 2, Node 2; local blade coordinate system
AeroDyn['B2N3VDisx']          = False     # (m/s); x-component of disturbed wind velocity at Blade 2, Node 3; local blade coordinate system
AeroDyn['B2N4VDisx']          = False     # (m/s); x-component of disturbed wind velocity at Blade 2, Node 4; local blade coordinate system
AeroDyn['B2N5VDisx']          = False     # (m/s); x-component of disturbed wind velocity at Blade 2, Node 5; local blade coordinate system
AeroDyn['B2N6VDisx']          = False     # (m/s); x-component of disturbed wind velocity at Blade 2, Node 6; local blade coordinate system
AeroDyn['B2N7VDisx']          = False     # (m/s); x-component of disturbed wind velocity at Blade 2, Node 7; local blade coordinate system
AeroDyn['B2N8VDisx']          = False     # (m/s); x-component of disturbed wind velocity at Blade 2, Node 8; local blade coordinate system
AeroDyn['B2N9VDisx']          = False     # (m/s); x-component of disturbed wind velocity at Blade 2, Node 9; local blade coordinate system
AeroDyn['B2N1VDisy']          = False     # (m/s); y-component of disturbed wind velocity at Blade 2, Node 1; local blade coordinate system
AeroDyn['B2N2VDisy']          = False     # (m/s); y-component of disturbed wind velocity at Blade 2, Node 2; local blade coordinate system
AeroDyn['B2N3VDisy']          = False     # (m/s); y-component of disturbed wind velocity at Blade 2, Node 3; local blade coordinate system
AeroDyn['B2N4VDisy']          = False     # (m/s); y-component of disturbed wind velocity at Blade 2, Node 4; local blade coordinate system
AeroDyn['B2N5VDisy']          = False     # (m/s); y-component of disturbed wind velocity at Blade 2, Node 5; local blade coordinate system
AeroDyn['B2N6VDisy']          = False     # (m/s); y-component of disturbed wind velocity at Blade 2, Node 6; local blade coordinate system
AeroDyn['B2N7VDisy']          = False     # (m/s); y-component of disturbed wind velocity at Blade 2, Node 7; local blade coordinate system
AeroDyn['B2N8VDisy']          = False     # (m/s); y-component of disturbed wind velocity at Blade 2, Node 8; local blade coordinate system
AeroDyn['B2N9VDisy']          = False     # (m/s); y-component of disturbed wind velocity at Blade 2, Node 9; local blade coordinate system
AeroDyn['B2N1VDisz']          = False     # (m/s); z-component of disturbed wind velocity at Blade 2, Node 1; local blade coordinate system
AeroDyn['B2N2VDisz']          = False     # (m/s); z-component of disturbed wind velocity at Blade 2, Node 2; local blade coordinate system
AeroDyn['B2N3VDisz']          = False     # (m/s); z-component of disturbed wind velocity at Blade 2, Node 3; local blade coordinate system
AeroDyn['B2N4VDisz']          = False     # (m/s); z-component of disturbed wind velocity at Blade 2, Node 4; local blade coordinate system
AeroDyn['B2N5VDisz']          = False     # (m/s); z-component of disturbed wind velocity at Blade 2, Node 5; local blade coordinate system
AeroDyn['B2N6VDisz']          = False     # (m/s); z-component of disturbed wind velocity at Blade 2, Node 6; local blade coordinate system
AeroDyn['B2N7VDisz']          = False     # (m/s); z-component of disturbed wind velocity at Blade 2, Node 7; local blade coordinate system
AeroDyn['B2N8VDisz']          = False     # (m/s); z-component of disturbed wind velocity at Blade 2, Node 8; local blade coordinate system
AeroDyn['B2N9VDisz']          = False     # (m/s); z-component of disturbed wind velocity at Blade 2, Node 9; local blade coordinate system
AeroDyn['B3N1VDisx']          = False     # (m/s); x-component of disturbed wind velocity at Blade 3, Node 1; local blade coordinate system
AeroDyn['B3N2VDisx']          = False     # (m/s); x-component of disturbed wind velocity at Blade 3, Node 2; local blade coordinate system
AeroDyn['B3N3VDisx']          = False     # (m/s); x-component of disturbed wind velocity at Blade 3, Node 3; local blade coordinate system
AeroDyn['B3N4VDisx']          = False     # (m/s); x-component of disturbed wind velocity at Blade 3, Node 4; local blade coordinate system
AeroDyn['B3N5VDisx']          = False     # (m/s); x-component of disturbed wind velocity at Blade 3, Node 5; local blade coordinate system
AeroDyn['B3N6VDisx']          = False     # (m/s); x-component of disturbed wind velocity at Blade 3, Node 6; local blade coordinate system
AeroDyn['B3N7VDisx']          = False     # (m/s); x-component of disturbed wind velocity at Blade 3, Node 7; local blade coordinate system
AeroDyn['B3N8VDisx']          = False     # (m/s); x-component of disturbed wind velocity at Blade 3, Node 8; local blade coordinate system
AeroDyn['B3N9VDisx']          = False     # (m/s); x-component of disturbed wind velocity at Blade 3, Node 9; local blade coordinate system
AeroDyn['B3N1VDisy']          = False     # (m/s); y-component of disturbed wind velocity at Blade 3, Node 1; local blade coordinate system
AeroDyn['B3N2VDisy']          = False     # (m/s); y-component of disturbed wind velocity at Blade 3, Node 2; local blade coordinate system
AeroDyn['B3N3VDisy']          = False     # (m/s); y-component of disturbed wind velocity at Blade 3, Node 3; local blade coordinate system
AeroDyn['B3N4VDisy']          = False     # (m/s); y-component of disturbed wind velocity at Blade 3, Node 4; local blade coordinate system
AeroDyn['B3N5VDisy']          = False     # (m/s); y-component of disturbed wind velocity at Blade 3, Node 5; local blade coordinate system
AeroDyn['B3N6VDisy']          = False     # (m/s); y-component of disturbed wind velocity at Blade 3, Node 6; local blade coordinate system
AeroDyn['B3N7VDisy']          = False     # (m/s); y-component of disturbed wind velocity at Blade 3, Node 7; local blade coordinate system
AeroDyn['B3N8VDisy']          = False     # (m/s); y-component of disturbed wind velocity at Blade 3, Node 8; local blade coordinate system
AeroDyn['B3N9VDisy']          = False     # (m/s); y-component of disturbed wind velocity at Blade 3, Node 9; local blade coordinate system
AeroDyn['B3N1VDisz']          = False     # (m/s); z-component of disturbed wind velocity at Blade 3, Node 1; local blade coordinate system
AeroDyn['B3N2VDisz']          = False     # (m/s); z-component of disturbed wind velocity at Blade 3, Node 2; local blade coordinate system
AeroDyn['B3N3VDisz']          = False     # (m/s); z-component of disturbed wind velocity at Blade 3, Node 3; local blade coordinate system
AeroDyn['B3N4VDisz']          = False     # (m/s); z-component of disturbed wind velocity at Blade 3, Node 4; local blade coordinate system
AeroDyn['B3N5VDisz']          = False     # (m/s); z-component of disturbed wind velocity at Blade 3, Node 5; local blade coordinate system
AeroDyn['B3N6VDisz']          = False     # (m/s); z-component of disturbed wind velocity at Blade 3, Node 6; local blade coordinate system
AeroDyn['B3N7VDisz']          = False     # (m/s); z-component of disturbed wind velocity at Blade 3, Node 7; local blade coordinate system
AeroDyn['B3N8VDisz']          = False     # (m/s); z-component of disturbed wind velocity at Blade 3, Node 8; local blade coordinate system
AeroDyn['B3N9VDisz']          = False     # (m/s); z-component of disturbed wind velocity at Blade 3, Node 9; local blade coordinate system
AeroDyn['B1N1STVx']           = False     # (m/s); x-component of structural translational velocity at Blade 1, Node 1; local blade coordinate system
AeroDyn['B1N2STVx']           = False     # (m/s); x-component of structural translational velocity at Blade 1, Node 2; local blade coordinate system
AeroDyn['B1N3STVx']           = False     # (m/s); x-component of structural translational velocity at Blade 1, Node 3; local blade coordinate system
AeroDyn['B1N4STVx']           = False     # (m/s); x-component of structural translational velocity at Blade 1, Node 4; local blade coordinate system
AeroDyn['B1N5STVx']           = False     # (m/s); x-component of structural translational velocity at Blade 1, Node 5; local blade coordinate system
AeroDyn['B1N6STVx']           = False     # (m/s); x-component of structural translational velocity at Blade 1, Node 6; local blade coordinate system
AeroDyn['B1N7STVx']           = False     # (m/s); x-component of structural translational velocity at Blade 1, Node 7; local blade coordinate system
AeroDyn['B1N8STVx']           = False     # (m/s); x-component of structural translational velocity at Blade 1, Node 8; local blade coordinate system
AeroDyn['B1N9STVx']           = False     # (m/s); x-component of structural translational velocity at Blade 1, Node 9; local blade coordinate system
AeroDyn['B1N1STVy']           = False     # (m/s); y-component of structural translational velocity at Blade 1, Node 1; local blade coordinate system
AeroDyn['B1N2STVy']           = False     # (m/s); y-component of structural translational velocity at Blade 1, Node 2; local blade coordinate system
AeroDyn['B1N3STVy']           = False     # (m/s); y-component of structural translational velocity at Blade 1, Node 3; local blade coordinate system
AeroDyn['B1N4STVy']           = False     # (m/s); y-component of structural translational velocity at Blade 1, Node 4; local blade coordinate system
AeroDyn['B1N5STVy']           = False     # (m/s); y-component of structural translational velocity at Blade 1, Node 5; local blade coordinate system
AeroDyn['B1N6STVy']           = False     # (m/s); y-component of structural translational velocity at Blade 1, Node 6; local blade coordinate system
AeroDyn['B1N7STVy']           = False     # (m/s); y-component of structural translational velocity at Blade 1, Node 7; local blade coordinate system
AeroDyn['B1N8STVy']           = False     # (m/s); y-component of structural translational velocity at Blade 1, Node 8; local blade coordinate system
AeroDyn['B1N9STVy']           = False     # (m/s); y-component of structural translational velocity at Blade 1, Node 9; local blade coordinate system
AeroDyn['B1N1STVz']           = False     # (m/s); z-component of structural translational velocity at Blade 1, Node 1; local blade coordinate system
AeroDyn['B1N2STVz']           = False     # (m/s); z-component of structural translational velocity at Blade 1, Node 2; local blade coordinate system
AeroDyn['B1N3STVz']           = False     # (m/s); z-component of structural translational velocity at Blade 1, Node 3; local blade coordinate system
AeroDyn['B1N4STVz']           = False     # (m/s); z-component of structural translational velocity at Blade 1, Node 4; local blade coordinate system
AeroDyn['B1N5STVz']           = False     # (m/s); z-component of structural translational velocity at Blade 1, Node 5; local blade coordinate system
AeroDyn['B1N6STVz']           = False     # (m/s); z-component of structural translational velocity at Blade 1, Node 6; local blade coordinate system
AeroDyn['B1N7STVz']           = False     # (m/s); z-component of structural translational velocity at Blade 1, Node 7; local blade coordinate system
AeroDyn['B1N8STVz']           = False     # (m/s); z-component of structural translational velocity at Blade 1, Node 8; local blade coordinate system
AeroDyn['B1N9STVz']           = False     # (m/s); z-component of structural translational velocity at Blade 1, Node 9; local blade coordinate system
AeroDyn['B2N1STVx']           = False     # (m/s); x-component of structural translational velocity at Blade 2, Node 1; local blade coordinate system
AeroDyn['B2N2STVx']           = False     # (m/s); x-component of structural translational velocity at Blade 2, Node 2; local blade coordinate system
AeroDyn['B2N3STVx']           = False     # (m/s); x-component of structural translational velocity at Blade 2, Node 3; local blade coordinate system
AeroDyn['B2N4STVx']           = False     # (m/s); x-component of structural translational velocity at Blade 2, Node 4; local blade coordinate system
AeroDyn['B2N5STVx']           = False     # (m/s); x-component of structural translational velocity at Blade 2, Node 5; local blade coordinate system
AeroDyn['B2N6STVx']           = False     # (m/s); x-component of structural translational velocity at Blade 2, Node 6; local blade coordinate system
AeroDyn['B2N7STVx']           = False     # (m/s); x-component of structural translational velocity at Blade 2, Node 7; local blade coordinate system
AeroDyn['B2N8STVx']           = False     # (m/s); x-component of structural translational velocity at Blade 2, Node 8; local blade coordinate system
AeroDyn['B2N9STVx']           = False     # (m/s); x-component of structural translational velocity at Blade 2, Node 9; local blade coordinate system
AeroDyn['B2N1STVy']           = False     # (m/s); y-component of structural translational velocity at Blade 2, Node 1; local blade coordinate system
AeroDyn['B2N2STVy']           = False     # (m/s); y-component of structural translational velocity at Blade 2, Node 2; local blade coordinate system
AeroDyn['B2N3STVy']           = False     # (m/s); y-component of structural translational velocity at Blade 2, Node 3; local blade coordinate system
AeroDyn['B2N4STVy']           = False     # (m/s); y-component of structural translational velocity at Blade 2, Node 4; local blade coordinate system
AeroDyn['B2N5STVy']           = False     # (m/s); y-component of structural translational velocity at Blade 2, Node 5; local blade coordinate system
AeroDyn['B2N6STVy']           = False     # (m/s); y-component of structural translational velocity at Blade 2, Node 6; local blade coordinate system
AeroDyn['B2N7STVy']           = False     # (m/s); y-component of structural translational velocity at Blade 2, Node 7; local blade coordinate system
AeroDyn['B2N8STVy']           = False     # (m/s); y-component of structural translational velocity at Blade 2, Node 8; local blade coordinate system
AeroDyn['B2N9STVy']           = False     # (m/s); y-component of structural translational velocity at Blade 2, Node 9; local blade coordinate system
AeroDyn['B2N1STVz']           = False     # (m/s); z-component of structural translational velocity at Blade 2, Node 1; local blade coordinate system
AeroDyn['B2N2STVz']           = False     # (m/s); z-component of structural translational velocity at Blade 2, Node 2; local blade coordinate system
AeroDyn['B2N3STVz']           = False     # (m/s); z-component of structural translational velocity at Blade 2, Node 3; local blade coordinate system
AeroDyn['B2N4STVz']           = False     # (m/s); z-component of structural translational velocity at Blade 2, Node 4; local blade coordinate system
AeroDyn['B2N5STVz']           = False     # (m/s); z-component of structural translational velocity at Blade 2, Node 5; local blade coordinate system
AeroDyn['B2N6STVz']           = False     # (m/s); z-component of structural translational velocity at Blade 2, Node 6; local blade coordinate system
AeroDyn['B2N7STVz']           = False     # (m/s); z-component of structural translational velocity at Blade 2, Node 7; local blade coordinate system
AeroDyn['B2N8STVz']           = False     # (m/s); z-component of structural translational velocity at Blade 2, Node 8; local blade coordinate system
AeroDyn['B2N9STVz']           = False     # (m/s); z-component of structural translational velocity at Blade 2, Node 9; local blade coordinate system
AeroDyn['B3N1STVx']           = False     # (m/s); x-component of structural translational velocity at Blade 3, Node 1; local blade coordinate system
AeroDyn['B3N2STVx']           = False     # (m/s); x-component of structural translational velocity at Blade 3, Node 2; local blade coordinate system
AeroDyn['B3N3STVx']           = False     # (m/s); x-component of structural translational velocity at Blade 3, Node 3; local blade coordinate system
AeroDyn['B3N4STVx']           = False     # (m/s); x-component of structural translational velocity at Blade 3, Node 4; local blade coordinate system
AeroDyn['B3N5STVx']           = False     # (m/s); x-component of structural translational velocity at Blade 3, Node 5; local blade coordinate system
AeroDyn['B3N6STVx']           = False     # (m/s); x-component of structural translational velocity at Blade 3, Node 6; local blade coordinate system
AeroDyn['B3N7STVx']           = False     # (m/s); x-component of structural translational velocity at Blade 3, Node 7; local blade coordinate system
AeroDyn['B3N8STVx']           = False     # (m/s); x-component of structural translational velocity at Blade 3, Node 8; local blade coordinate system
AeroDyn['B3N9STVx']           = False     # (m/s); x-component of structural translational velocity at Blade 3, Node 9; local blade coordinate system
AeroDyn['B3N1STVy']           = False     # (m/s); y-component of structural translational velocity at Blade 3, Node 1; local blade coordinate system
AeroDyn['B3N2STVy']           = False     # (m/s); y-component of structural translational velocity at Blade 3, Node 2; local blade coordinate system
AeroDyn['B3N3STVy']           = False     # (m/s); y-component of structural translational velocity at Blade 3, Node 3; local blade coordinate system
AeroDyn['B3N4STVy']           = False     # (m/s); y-component of structural translational velocity at Blade 3, Node 4; local blade coordinate system
AeroDyn['B3N5STVy']           = False     # (m/s); y-component of structural translational velocity at Blade 3, Node 5; local blade coordinate system
AeroDyn['B3N6STVy']           = False     # (m/s); y-component of structural translational velocity at Blade 3, Node 6; local blade coordinate system
AeroDyn['B3N7STVy']           = False     # (m/s); y-component of structural translational velocity at Blade 3, Node 7; local blade coordinate system
AeroDyn['B3N8STVy']           = False     # (m/s); y-component of structural translational velocity at Blade 3, Node 8; local blade coordinate system
AeroDyn['B3N9STVy']           = False     # (m/s); y-component of structural translational velocity at Blade 3, Node 9; local blade coordinate system
AeroDyn['B3N1STVz']           = False     # (m/s); z-component of structural translational velocity at Blade 3, Node 1; local blade coordinate system
AeroDyn['B3N2STVz']           = False     # (m/s); z-component of structural translational velocity at Blade 3, Node 2; local blade coordinate system
AeroDyn['B3N3STVz']           = False     # (m/s); z-component of structural translational velocity at Blade 3, Node 3; local blade coordinate system
AeroDyn['B3N4STVz']           = False     # (m/s); z-component of structural translational velocity at Blade 3, Node 4; local blade coordinate system
AeroDyn['B3N5STVz']           = False     # (m/s); z-component of structural translational velocity at Blade 3, Node 5; local blade coordinate system
AeroDyn['B3N6STVz']           = False     # (m/s); z-component of structural translational velocity at Blade 3, Node 6; local blade coordinate system
AeroDyn['B3N7STVz']           = False     # (m/s); z-component of structural translational velocity at Blade 3, Node 7; local blade coordinate system
AeroDyn['B3N8STVz']           = False     # (m/s); z-component of structural translational velocity at Blade 3, Node 8; local blade coordinate system
AeroDyn['B3N9STVz']           = False     # (m/s); z-component of structural translational velocity at Blade 3, Node 9; local blade coordinate system
AeroDyn['B1N1VRel']           = False     # (m/s); Relvative wind speed at Blade 1, Node 1; 
AeroDyn['B1N2VRel']           = False     # (m/s); Relvative wind speed at Blade 1, Node 2; 
AeroDyn['B1N3VRel']           = False     # (m/s); Relvative wind speed at Blade 1, Node 3; 
AeroDyn['B1N4VRel']           = False     # (m/s); Relvative wind speed at Blade 1, Node 4; 
AeroDyn['B1N5VRel']           = False     # (m/s); Relvative wind speed at Blade 1, Node 5; 
AeroDyn['B1N6VRel']           = False     # (m/s); Relvative wind speed at Blade 1, Node 6; 
AeroDyn['B1N7VRel']           = False     # (m/s); Relvative wind speed at Blade 1, Node 7; 
AeroDyn['B1N8VRel']           = False     # (m/s); Relvative wind speed at Blade 1, Node 8; 
AeroDyn['B1N9VRel']           = False     # (m/s); Relvative wind speed at Blade 1, Node 9; 
AeroDyn['B2N1VRel']           = False     # (m/s); Relvative wind speed at Blade 2, Node 1; 
AeroDyn['B2N2VRel']           = False     # (m/s); Relvative wind speed at Blade 2, Node 2; 
AeroDyn['B2N3VRel']           = False     # (m/s); Relvative wind speed at Blade 2, Node 3; 
AeroDyn['B2N4VRel']           = False     # (m/s); Relvative wind speed at Blade 2, Node 4; 
AeroDyn['B2N5VRel']           = False     # (m/s); Relvative wind speed at Blade 2, Node 5; 
AeroDyn['B2N6VRel']           = False     # (m/s); Relvative wind speed at Blade 2, Node 6; 
AeroDyn['B2N7VRel']           = False     # (m/s); Relvative wind speed at Blade 2, Node 7; 
AeroDyn['B2N8VRel']           = False     # (m/s); Relvative wind speed at Blade 2, Node 8; 
AeroDyn['B2N9VRel']           = False     # (m/s); Relvative wind speed at Blade 2, Node 9; 
AeroDyn['B3N1VRel']           = False     # (m/s); Relvative wind speed at Blade 3, Node 1; 
AeroDyn['B3N2VRel']           = False     # (m/s); Relvative wind speed at Blade 3, Node 2; 
AeroDyn['B3N3VRel']           = False     # (m/s); Relvative wind speed at Blade 3, Node 3; 
AeroDyn['B3N4VRel']           = False     # (m/s); Relvative wind speed at Blade 3, Node 4; 
AeroDyn['B3N5VRel']           = False     # (m/s); Relvative wind speed at Blade 3, Node 5; 
AeroDyn['B3N6VRel']           = False     # (m/s); Relvative wind speed at Blade 3, Node 6; 
AeroDyn['B3N7VRel']           = False     # (m/s); Relvative wind speed at Blade 3, Node 7; 
AeroDyn['B3N8VRel']           = False     # (m/s); Relvative wind speed at Blade 3, Node 8; 
AeroDyn['B3N9VRel']           = False     # (m/s); Relvative wind speed at Blade 3, Node 9; 
AeroDyn['B1N1DynP']           = False     # (Pa); Dynamic pressure at Blade 1, Node 1; 
AeroDyn['B1N2DynP']           = False     # (Pa); Dynamic pressure at Blade 1, Node 2; 
AeroDyn['B1N3DynP']           = False     # (Pa); Dynamic pressure at Blade 1, Node 3; 
AeroDyn['B1N4DynP']           = False     # (Pa); Dynamic pressure at Blade 1, Node 4; 
AeroDyn['B1N5DynP']           = False     # (Pa); Dynamic pressure at Blade 1, Node 5; 
AeroDyn['B1N6DynP']           = False     # (Pa); Dynamic pressure at Blade 1, Node 6; 
AeroDyn['B1N7DynP']           = False     # (Pa); Dynamic pressure at Blade 1, Node 7; 
AeroDyn['B1N8DynP']           = False     # (Pa); Dynamic pressure at Blade 1, Node 8; 
AeroDyn['B1N9DynP']           = False     # (Pa); Dynamic pressure at Blade 1, Node 9; 
AeroDyn['B2N1DynP']           = False     # (Pa); Dynamic pressure at Blade 2, Node 1; 
AeroDyn['B2N2DynP']           = False     # (Pa); Dynamic pressure at Blade 2, Node 2; 
AeroDyn['B2N3DynP']           = False     # (Pa); Dynamic pressure at Blade 2, Node 3; 
AeroDyn['B2N4DynP']           = False     # (Pa); Dynamic pressure at Blade 2, Node 4; 
AeroDyn['B2N5DynP']           = False     # (Pa); Dynamic pressure at Blade 2, Node 5; 
AeroDyn['B2N6DynP']           = False     # (Pa); Dynamic pressure at Blade 2, Node 6; 
AeroDyn['B2N7DynP']           = False     # (Pa); Dynamic pressure at Blade 2, Node 7; 
AeroDyn['B2N8DynP']           = False     # (Pa); Dynamic pressure at Blade 2, Node 8; 
AeroDyn['B2N9DynP']           = False     # (Pa); Dynamic pressure at Blade 2, Node 9; 
AeroDyn['B3N1DynP']           = False     # (Pa); Dynamic pressure at Blade 3, Node 1; 
AeroDyn['B3N2DynP']           = False     # (Pa); Dynamic pressure at Blade 3, Node 2; 
AeroDyn['B3N3DynP']           = False     # (Pa); Dynamic pressure at Blade 3, Node 3; 
AeroDyn['B3N4DynP']           = False     # (Pa); Dynamic pressure at Blade 3, Node 4; 
AeroDyn['B3N5DynP']           = False     # (Pa); Dynamic pressure at Blade 3, Node 5; 
AeroDyn['B3N6DynP']           = False     # (Pa); Dynamic pressure at Blade 3, Node 6; 
AeroDyn['B3N7DynP']           = False     # (Pa); Dynamic pressure at Blade 3, Node 7; 
AeroDyn['B3N8DynP']           = False     # (Pa); Dynamic pressure at Blade 3, Node 8; 
AeroDyn['B3N9DynP']           = False     # (Pa); Dynamic pressure at Blade 3, Node 9; 
AeroDyn['B1N1Re']             = False     # (-); Reynolds number (in millions) at Blade 1, Node 1; 
AeroDyn['B1N2Re']             = False     # (-); Reynolds number (in millions) at Blade 1, Node 2; 
AeroDyn['B1N3Re']             = False     # (-); Reynolds number (in millions) at Blade 1, Node 3; 
AeroDyn['B1N4Re']             = False     # (-); Reynolds number (in millions) at Blade 1, Node 4; 
AeroDyn['B1N5Re']             = False     # (-); Reynolds number (in millions) at Blade 1, Node 5; 
AeroDyn['B1N6Re']             = False     # (-); Reynolds number (in millions) at Blade 1, Node 6; 
AeroDyn['B1N7Re']             = False     # (-); Reynolds number (in millions) at Blade 1, Node 7; 
AeroDyn['B1N8Re']             = False     # (-); Reynolds number (in millions) at Blade 1, Node 8; 
AeroDyn['B1N9Re']             = False     # (-); Reynolds number (in millions) at Blade 1, Node 9; 
AeroDyn['B2N1Re']             = False     # (-); Reynolds number (in millions) at Blade 2, Node 1; 
AeroDyn['B2N2Re']             = False     # (-); Reynolds number (in millions) at Blade 2, Node 2; 
AeroDyn['B2N3Re']             = False     # (-); Reynolds number (in millions) at Blade 2, Node 3; 
AeroDyn['B2N4Re']             = False     # (-); Reynolds number (in millions) at Blade 2, Node 4; 
AeroDyn['B2N5Re']             = False     # (-); Reynolds number (in millions) at Blade 2, Node 5; 
AeroDyn['B2N6Re']             = False     # (-); Reynolds number (in millions) at Blade 2, Node 6; 
AeroDyn['B2N7Re']             = False     # (-); Reynolds number (in millions) at Blade 2, Node 7; 
AeroDyn['B2N8Re']             = False     # (-); Reynolds number (in millions) at Blade 2, Node 8; 
AeroDyn['B2N9Re']             = False     # (-); Reynolds number (in millions) at Blade 2, Node 9; 
AeroDyn['B3N1Re']             = False     # (-); Reynolds number (in millions) at Blade 3, Node 1; 
AeroDyn['B3N2Re']             = False     # (-); Reynolds number (in millions) at Blade 3, Node 2; 
AeroDyn['B3N3Re']             = False     # (-); Reynolds number (in millions) at Blade 3, Node 3; 
AeroDyn['B3N4Re']             = False     # (-); Reynolds number (in millions) at Blade 3, Node 4; 
AeroDyn['B3N5Re']             = False     # (-); Reynolds number (in millions) at Blade 3, Node 5; 
AeroDyn['B3N6Re']             = False     # (-); Reynolds number (in millions) at Blade 3, Node 6; 
AeroDyn['B3N7Re']             = False     # (-); Reynolds number (in millions) at Blade 3, Node 7; 
AeroDyn['B3N8Re']             = False     # (-); Reynolds number (in millions) at Blade 3, Node 8; 
AeroDyn['B3N9Re']             = False     # (-); Reynolds number (in millions) at Blade 3, Node 9; 
AeroDyn['B1N1M']              = False     # (-); Mach number at Blade 1, Node 1; 
AeroDyn['B1N2M']              = False     # (-); Mach number at Blade 1, Node 2; 
AeroDyn['B1N3M']              = False     # (-); Mach number at Blade 1, Node 3; 
AeroDyn['B1N4M']              = False     # (-); Mach number at Blade 1, Node 4; 
AeroDyn['B1N5M']              = False     # (-); Mach number at Blade 1, Node 5; 
AeroDyn['B1N6M']              = False     # (-); Mach number at Blade 1, Node 6; 
AeroDyn['B1N7M']              = False     # (-); Mach number at Blade 1, Node 7; 
AeroDyn['B1N8M']              = False     # (-); Mach number at Blade 1, Node 8; 
AeroDyn['B1N9M']              = False     # (-); Mach number at Blade 1, Node 9; 
AeroDyn['B2N1M']              = False     # (-); Mach number at Blade 2, Node 1; 
AeroDyn['B2N2M']              = False     # (-); Mach number at Blade 2, Node 2; 
AeroDyn['B2N3M']              = False     # (-); Mach number at Blade 2, Node 3; 
AeroDyn['B2N4M']              = False     # (-); Mach number at Blade 2, Node 4; 
AeroDyn['B2N5M']              = False     # (-); Mach number at Blade 2, Node 5; 
AeroDyn['B2N6M']              = False     # (-); Mach number at Blade 2, Node 6; 
AeroDyn['B2N7M']              = False     # (-); Mach number at Blade 2, Node 7; 
AeroDyn['B2N8M']              = False     # (-); Mach number at Blade 2, Node 8; 
AeroDyn['B2N9M']              = False     # (-); Mach number at Blade 2, Node 9; 
AeroDyn['B3N1M']              = False     # (-); Mach number at Blade 3, Node 1; 
AeroDyn['B3N2M']              = False     # (-); Mach number at Blade 3, Node 2; 
AeroDyn['B3N3M']              = False     # (-); Mach number at Blade 3, Node 3; 
AeroDyn['B3N4M']              = False     # (-); Mach number at Blade 3, Node 4; 
AeroDyn['B3N5M']              = False     # (-); Mach number at Blade 3, Node 5; 
AeroDyn['B3N6M']              = False     # (-); Mach number at Blade 3, Node 6; 
AeroDyn['B3N7M']              = False     # (-); Mach number at Blade 3, Node 7; 
AeroDyn['B3N8M']              = False     # (-); Mach number at Blade 3, Node 8; 
AeroDyn['B3N9M']              = False     # (-); Mach number at Blade 3, Node 9; 
AeroDyn['B1N1Vindx']          = False     # (m/s); Axial induced wind velocity at Blade 1, Node 1; 
AeroDyn['B1N2Vindx']          = False     # (m/s); Axial induced wind velocity at Blade 1, Node 2; 
AeroDyn['B1N3Vindx']          = False     # (m/s); Axial induced wind velocity at Blade 1, Node 3; 
AeroDyn['B1N4Vindx']          = False     # (m/s); Axial induced wind velocity at Blade 1, Node 4; 
AeroDyn['B1N5Vindx']          = False     # (m/s); Axial induced wind velocity at Blade 1, Node 5; 
AeroDyn['B1N6Vindx']          = False     # (m/s); Axial induced wind velocity at Blade 1, Node 6; 
AeroDyn['B1N7Vindx']          = False     # (m/s); Axial induced wind velocity at Blade 1, Node 7; 
AeroDyn['B1N8Vindx']          = False     # (m/s); Axial induced wind velocity at Blade 1, Node 8; 
AeroDyn['B1N9Vindx']          = False     # (m/s); Axial induced wind velocity at Blade 1, Node 9; 
AeroDyn['B2N1Vindx']          = False     # (m/s); Axial induced wind velocity at Blade 2, Node 1; 
AeroDyn['B2N2Vindx']          = False     # (m/s); Axial induced wind velocity at Blade 2, Node 2; 
AeroDyn['B2N3Vindx']          = False     # (m/s); Axial induced wind velocity at Blade 2, Node 3; 
AeroDyn['B2N4Vindx']          = False     # (m/s); Axial induced wind velocity at Blade 2, Node 4; 
AeroDyn['B2N5Vindx']          = False     # (m/s); Axial induced wind velocity at Blade 2, Node 5; 
AeroDyn['B2N6Vindx']          = False     # (m/s); Axial induced wind velocity at Blade 2, Node 6; 
AeroDyn['B2N7Vindx']          = False     # (m/s); Axial induced wind velocity at Blade 2, Node 7; 
AeroDyn['B2N8Vindx']          = False     # (m/s); Axial induced wind velocity at Blade 2, Node 8; 
AeroDyn['B2N9Vindx']          = False     # (m/s); Axial induced wind velocity at Blade 2, Node 9; 
AeroDyn['B3N1Vindx']          = False     # (m/s); Axial induced wind velocity at Blade 3, Node 1; 
AeroDyn['B3N2Vindx']          = False     # (m/s); Axial induced wind velocity at Blade 3, Node 2; 
AeroDyn['B3N3Vindx']          = False     # (m/s); Axial induced wind velocity at Blade 3, Node 3; 
AeroDyn['B3N4Vindx']          = False     # (m/s); Axial induced wind velocity at Blade 3, Node 4; 
AeroDyn['B3N5Vindx']          = False     # (m/s); Axial induced wind velocity at Blade 3, Node 5; 
AeroDyn['B3N6Vindx']          = False     # (m/s); Axial induced wind velocity at Blade 3, Node 6; 
AeroDyn['B3N7Vindx']          = False     # (m/s); Axial induced wind velocity at Blade 3, Node 7; 
AeroDyn['B3N8Vindx']          = False     # (m/s); Axial induced wind velocity at Blade 3, Node 8; 
AeroDyn['B3N9Vindx']          = False     # (m/s); Axial induced wind velocity at Blade 3, Node 9; 
AeroDyn['B1N1Vindy']          = False     # (m/s); Tangential induced wind velocity at Blade 1, Node 1; 
AeroDyn['B1N2Vindy']          = False     # (m/s); Tangential induced wind velocity at Blade 1, Node 2; 
AeroDyn['B1N3Vindy']          = False     # (m/s); Tangential induced wind velocity at Blade 1, Node 3; 
AeroDyn['B1N4Vindy']          = False     # (m/s); Tangential induced wind velocity at Blade 1, Node 4; 
AeroDyn['B1N5Vindy']          = False     # (m/s); Tangential induced wind velocity at Blade 1, Node 5; 
AeroDyn['B1N6Vindy']          = False     # (m/s); Tangential induced wind velocity at Blade 1, Node 6; 
AeroDyn['B1N7Vindy']          = False     # (m/s); Tangential induced wind velocity at Blade 1, Node 7; 
AeroDyn['B1N8Vindy']          = False     # (m/s); Tangential induced wind velocity at Blade 1, Node 8; 
AeroDyn['B1N9Vindy']          = False     # (m/s); Tangential induced wind velocity at Blade 1, Node 9; 
AeroDyn['B2N1Vindy']          = False     # (m/s); Tangential induced wind velocity at Blade 2, Node 1; 
AeroDyn['B2N2Vindy']          = False     # (m/s); Tangential induced wind velocity at Blade 2, Node 2; 
AeroDyn['B2N3Vindy']          = False     # (m/s); Tangential induced wind velocity at Blade 2, Node 3; 
AeroDyn['B2N4Vindy']          = False     # (m/s); Tangential induced wind velocity at Blade 2, Node 4; 
AeroDyn['B2N5Vindy']          = False     # (m/s); Tangential induced wind velocity at Blade 2, Node 5; 
AeroDyn['B2N6Vindy']          = False     # (m/s); Tangential induced wind velocity at Blade 2, Node 6; 
AeroDyn['B2N7Vindy']          = False     # (m/s); Tangential induced wind velocity at Blade 2, Node 7; 
AeroDyn['B2N8Vindy']          = False     # (m/s); Tangential induced wind velocity at Blade 2, Node 8; 
AeroDyn['B2N9Vindy']          = False     # (m/s); Tangential induced wind velocity at Blade 2, Node 9; 
AeroDyn['B3N1Vindy']          = False     # (m/s); Tangential induced wind velocity at Blade 3, Node 1; 
AeroDyn['B3N2Vindy']          = False     # (m/s); Tangential induced wind velocity at Blade 3, Node 2; 
AeroDyn['B3N3Vindy']          = False     # (m/s); Tangential induced wind velocity at Blade 3, Node 3; 
AeroDyn['B3N4Vindy']          = False     # (m/s); Tangential induced wind velocity at Blade 3, Node 4; 
AeroDyn['B3N5Vindy']          = False     # (m/s); Tangential induced wind velocity at Blade 3, Node 5; 
AeroDyn['B3N6Vindy']          = False     # (m/s); Tangential induced wind velocity at Blade 3, Node 6; 
AeroDyn['B3N7Vindy']          = False     # (m/s); Tangential induced wind velocity at Blade 3, Node 7; 
AeroDyn['B3N8Vindy']          = False     # (m/s); Tangential induced wind velocity at Blade 3, Node 8; 
AeroDyn['B3N9Vindy']          = False     # (m/s); Tangential induced wind velocity at Blade 3, Node 9; 
AeroDyn['B1N1AxInd']          = False     # (-); Axial induction factor at Blade 1, Node 1; 
AeroDyn['B1N2AxInd']          = False     # (-); Axial induction factor at Blade 1, Node 2; 
AeroDyn['B1N3AxInd']          = False     # (-); Axial induction factor at Blade 1, Node 3; 
AeroDyn['B1N4AxInd']          = False     # (-); Axial induction factor at Blade 1, Node 4; 
AeroDyn['B1N5AxInd']          = False     # (-); Axial induction factor at Blade 1, Node 5; 
AeroDyn['B1N6AxInd']          = False     # (-); Axial induction factor at Blade 1, Node 6; 
AeroDyn['B1N7AxInd']          = False     # (-); Axial induction factor at Blade 1, Node 7; 
AeroDyn['B1N8AxInd']          = False     # (-); Axial induction factor at Blade 1, Node 8; 
AeroDyn['B1N9AxInd']          = False     # (-); Axial induction factor at Blade 1, Node 9; 
AeroDyn['B2N1AxInd']          = False     # (-); Axial induction factor at Blade 2, Node 1; 
AeroDyn['B2N2AxInd']          = False     # (-); Axial induction factor at Blade 2, Node 2; 
AeroDyn['B2N3AxInd']          = False     # (-); Axial induction factor at Blade 2, Node 3; 
AeroDyn['B2N4AxInd']          = False     # (-); Axial induction factor at Blade 2, Node 4; 
AeroDyn['B2N5AxInd']          = False     # (-); Axial induction factor at Blade 2, Node 5; 
AeroDyn['B2N6AxInd']          = False     # (-); Axial induction factor at Blade 2, Node 6; 
AeroDyn['B2N7AxInd']          = False     # (-); Axial induction factor at Blade 2, Node 7; 
AeroDyn['B2N8AxInd']          = False     # (-); Axial induction factor at Blade 2, Node 8; 
AeroDyn['B2N9AxInd']          = False     # (-); Axial induction factor at Blade 2, Node 9; 
AeroDyn['B3N1AxInd']          = False     # (-); Axial induction factor at Blade 3, Node 1; 
AeroDyn['B3N2AxInd']          = False     # (-); Axial induction factor at Blade 3, Node 2; 
AeroDyn['B3N3AxInd']          = False     # (-); Axial induction factor at Blade 3, Node 3; 
AeroDyn['B3N4AxInd']          = False     # (-); Axial induction factor at Blade 3, Node 4; 
AeroDyn['B3N5AxInd']          = False     # (-); Axial induction factor at Blade 3, Node 5; 
AeroDyn['B3N6AxInd']          = False     # (-); Axial induction factor at Blade 3, Node 6; 
AeroDyn['B3N7AxInd']          = False     # (-); Axial induction factor at Blade 3, Node 7; 
AeroDyn['B3N8AxInd']          = False     # (-); Axial induction factor at Blade 3, Node 8; 
AeroDyn['B3N9AxInd']          = False     # (-); Axial induction factor at Blade 3, Node 9; 
AeroDyn['B1N1TnInd']          = False     # (-); Tangential induction factor at Blade 1, Node 1; 
AeroDyn['B1N2TnInd']          = False     # (-); Tangential induction factor at Blade 1, Node 2; 
AeroDyn['B1N3TnInd']          = False     # (-); Tangential induction factor at Blade 1, Node 3; 
AeroDyn['B1N4TnInd']          = False     # (-); Tangential induction factor at Blade 1, Node 4; 
AeroDyn['B1N5TnInd']          = False     # (-); Tangential induction factor at Blade 1, Node 5; 
AeroDyn['B1N6TnInd']          = False     # (-); Tangential induction factor at Blade 1, Node 6; 
AeroDyn['B1N7TnInd']          = False     # (-); Tangential induction factor at Blade 1, Node 7; 
AeroDyn['B1N8TnInd']          = False     # (-); Tangential induction factor at Blade 1, Node 8; 
AeroDyn['B1N9TnInd']          = False     # (-); Tangential induction factor at Blade 1, Node 9; 
AeroDyn['B2N1TnInd']          = False     # (-); Tangential induction factor at Blade 2, Node 1; 
AeroDyn['B2N2TnInd']          = False     # (-); Tangential induction factor at Blade 2, Node 2; 
AeroDyn['B2N3TnInd']          = False     # (-); Tangential induction factor at Blade 2, Node 3; 
AeroDyn['B2N4TnInd']          = False     # (-); Tangential induction factor at Blade 2, Node 4; 
AeroDyn['B2N5TnInd']          = False     # (-); Tangential induction factor at Blade 2, Node 5; 
AeroDyn['B2N6TnInd']          = False     # (-); Tangential induction factor at Blade 2, Node 6; 
AeroDyn['B2N7TnInd']          = False     # (-); Tangential induction factor at Blade 2, Node 7; 
AeroDyn['B2N8TnInd']          = False     # (-); Tangential induction factor at Blade 2, Node 8; 
AeroDyn['B2N9TnInd']          = False     # (-); Tangential induction factor at Blade 2, Node 9; 
AeroDyn['B3N1TnInd']          = False     # (-); Tangential induction factor at Blade 3, Node 1; 
AeroDyn['B3N2TnInd']          = False     # (-); Tangential induction factor at Blade 3, Node 2; 
AeroDyn['B3N3TnInd']          = False     # (-); Tangential induction factor at Blade 3, Node 3; 
AeroDyn['B3N4TnInd']          = False     # (-); Tangential induction factor at Blade 3, Node 4; 
AeroDyn['B3N5TnInd']          = False     # (-); Tangential induction factor at Blade 3, Node 5; 
AeroDyn['B3N6TnInd']          = False     # (-); Tangential induction factor at Blade 3, Node 6; 
AeroDyn['B3N7TnInd']          = False     # (-); Tangential induction factor at Blade 3, Node 7; 
AeroDyn['B3N8TnInd']          = False     # (-); Tangential induction factor at Blade 3, Node 8; 
AeroDyn['B3N9TnInd']          = False     # (-); Tangential induction factor at Blade 3, Node 9; 
AeroDyn['B1N1Alpha']          = False     # (deg); Angle of attack at Blade 1, Node 1; 
AeroDyn['B1N2Alpha']          = False     # (deg); Angle of attack at Blade 1, Node 2; 
AeroDyn['B1N3Alpha']          = False     # (deg); Angle of attack at Blade 1, Node 3; 
AeroDyn['B1N4Alpha']          = False     # (deg); Angle of attack at Blade 1, Node 4; 
AeroDyn['B1N5Alpha']          = False     # (deg); Angle of attack at Blade 1, Node 5; 
AeroDyn['B1N6Alpha']          = False     # (deg); Angle of attack at Blade 1, Node 6; 
AeroDyn['B1N7Alpha']          = False     # (deg); Angle of attack at Blade 1, Node 7; 
AeroDyn['B1N8Alpha']          = False     # (deg); Angle of attack at Blade 1, Node 8; 
AeroDyn['B1N9Alpha']          = False     # (deg); Angle of attack at Blade 1, Node 9; 
AeroDyn['B2N1Alpha']          = False     # (deg); Angle of attack at Blade 2, Node 1; 
AeroDyn['B2N2Alpha']          = False     # (deg); Angle of attack at Blade 2, Node 2; 
AeroDyn['B2N3Alpha']          = False     # (deg); Angle of attack at Blade 2, Node 3; 
AeroDyn['B2N4Alpha']          = False     # (deg); Angle of attack at Blade 2, Node 4; 
AeroDyn['B2N5Alpha']          = False     # (deg); Angle of attack at Blade 2, Node 5; 
AeroDyn['B2N6Alpha']          = False     # (deg); Angle of attack at Blade 2, Node 6; 
AeroDyn['B2N7Alpha']          = False     # (deg); Angle of attack at Blade 2, Node 7; 
AeroDyn['B2N8Alpha']          = False     # (deg); Angle of attack at Blade 2, Node 8; 
AeroDyn['B2N9Alpha']          = False     # (deg); Angle of attack at Blade 2, Node 9; 
AeroDyn['B3N1Alpha']          = False     # (deg); Angle of attack at Blade 3, Node 1; 
AeroDyn['B3N2Alpha']          = False     # (deg); Angle of attack at Blade 3, Node 2; 
AeroDyn['B3N3Alpha']          = False     # (deg); Angle of attack at Blade 3, Node 3; 
AeroDyn['B3N4Alpha']          = False     # (deg); Angle of attack at Blade 3, Node 4; 
AeroDyn['B3N5Alpha']          = False     # (deg); Angle of attack at Blade 3, Node 5; 
AeroDyn['B3N6Alpha']          = False     # (deg); Angle of attack at Blade 3, Node 6; 
AeroDyn['B3N7Alpha']          = False     # (deg); Angle of attack at Blade 3, Node 7; 
AeroDyn['B3N8Alpha']          = False     # (deg); Angle of attack at Blade 3, Node 8; 
AeroDyn['B3N9Alpha']          = False     # (deg); Angle of attack at Blade 3, Node 9; 
AeroDyn['B1N1Theta']          = False     # (deg); Pitch+Twist angle at Blade 1, Node 1; 
AeroDyn['B1N2Theta']          = False     # (deg); Pitch+Twist angle at Blade 1, Node 2; 
AeroDyn['B1N3Theta']          = False     # (deg); Pitch+Twist angle at Blade 1, Node 3; 
AeroDyn['B1N4Theta']          = False     # (deg); Pitch+Twist angle at Blade 1, Node 4; 
AeroDyn['B1N5Theta']          = False     # (deg); Pitch+Twist angle at Blade 1, Node 5; 
AeroDyn['B1N6Theta']          = False     # (deg); Pitch+Twist angle at Blade 1, Node 6; 
AeroDyn['B1N7Theta']          = False     # (deg); Pitch+Twist angle at Blade 1, Node 7; 
AeroDyn['B1N8Theta']          = False     # (deg); Pitch+Twist angle at Blade 1, Node 8; 
AeroDyn['B1N9Theta']          = False     # (deg); Pitch+Twist angle at Blade 1, Node 9; 
AeroDyn['B2N1Theta']          = False     # (deg); Pitch+Twist angle at Blade 2, Node 1; 
AeroDyn['B2N2Theta']          = False     # (deg); Pitch+Twist angle at Blade 2, Node 2; 
AeroDyn['B2N3Theta']          = False     # (deg); Pitch+Twist angle at Blade 2, Node 3; 
AeroDyn['B2N4Theta']          = False     # (deg); Pitch+Twist angle at Blade 2, Node 4; 
AeroDyn['B2N5Theta']          = False     # (deg); Pitch+Twist angle at Blade 2, Node 5; 
AeroDyn['B2N6Theta']          = False     # (deg); Pitch+Twist angle at Blade 2, Node 6; 
AeroDyn['B2N7Theta']          = False     # (deg); Pitch+Twist angle at Blade 2, Node 7; 
AeroDyn['B2N8Theta']          = False     # (deg); Pitch+Twist angle at Blade 2, Node 8; 
AeroDyn['B2N9Theta']          = False     # (deg); Pitch+Twist angle at Blade 2, Node 9; 
AeroDyn['B3N1Theta']          = False     # (deg); Pitch+Twist angle at Blade 3, Node 1; 
AeroDyn['B3N2Theta']          = False     # (deg); Pitch+Twist angle at Blade 3, Node 2; 
AeroDyn['B3N3Theta']          = False     # (deg); Pitch+Twist angle at Blade 3, Node 3; 
AeroDyn['B3N4Theta']          = False     # (deg); Pitch+Twist angle at Blade 3, Node 4; 
AeroDyn['B3N5Theta']          = False     # (deg); Pitch+Twist angle at Blade 3, Node 5; 
AeroDyn['B3N6Theta']          = False     # (deg); Pitch+Twist angle at Blade 3, Node 6; 
AeroDyn['B3N7Theta']          = False     # (deg); Pitch+Twist angle at Blade 3, Node 7; 
AeroDyn['B3N8Theta']          = False     # (deg); Pitch+Twist angle at Blade 3, Node 8; 
AeroDyn['B3N9Theta']          = False     # (deg); Pitch+Twist angle at Blade 3, Node 9; 
AeroDyn['B1N1Phi']            = False     # (deg); Inflow angle at Blade 1, Node 1; 
AeroDyn['B1N2Phi']            = False     # (deg); Inflow angle at Blade 1, Node 2; 
AeroDyn['B1N3Phi']            = False     # (deg); Inflow angle at Blade 1, Node 3; 
AeroDyn['B1N4Phi']            = False     # (deg); Inflow angle at Blade 1, Node 4; 
AeroDyn['B1N5Phi']            = False     # (deg); Inflow angle at Blade 1, Node 5; 
AeroDyn['B1N6Phi']            = False     # (deg); Inflow angle at Blade 1, Node 6; 
AeroDyn['B1N7Phi']            = False     # (deg); Inflow angle at Blade 1, Node 7; 
AeroDyn['B1N8Phi']            = False     # (deg); Inflow angle at Blade 1, Node 8; 
AeroDyn['B1N9Phi']            = False     # (deg); Inflow angle at Blade 1, Node 9; 
AeroDyn['B2N1Phi']            = False     # (deg); Inflow angle at Blade 2, Node 1; 
AeroDyn['B2N2Phi']            = False     # (deg); Inflow angle at Blade 2, Node 2; 
AeroDyn['B2N3Phi']            = False     # (deg); Inflow angle at Blade 2, Node 3; 
AeroDyn['B2N4Phi']            = False     # (deg); Inflow angle at Blade 2, Node 4; 
AeroDyn['B2N5Phi']            = False     # (deg); Inflow angle at Blade 2, Node 5; 
AeroDyn['B2N6Phi']            = False     # (deg); Inflow angle at Blade 2, Node 6; 
AeroDyn['B2N7Phi']            = False     # (deg); Inflow angle at Blade 2, Node 7; 
AeroDyn['B2N8Phi']            = False     # (deg); Inflow angle at Blade 2, Node 8; 
AeroDyn['B2N9Phi']            = False     # (deg); Inflow angle at Blade 2, Node 9; 
AeroDyn['B3N1Phi']            = False     # (deg); Inflow angle at Blade 3, Node 1; 
AeroDyn['B3N2Phi']            = False     # (deg); Inflow angle at Blade 3, Node 2; 
AeroDyn['B3N3Phi']            = False     # (deg); Inflow angle at Blade 3, Node 3; 
AeroDyn['B3N4Phi']            = False     # (deg); Inflow angle at Blade 3, Node 4; 
AeroDyn['B3N5Phi']            = False     # (deg); Inflow angle at Blade 3, Node 5; 
AeroDyn['B3N6Phi']            = False     # (deg); Inflow angle at Blade 3, Node 6; 
AeroDyn['B3N7Phi']            = False     # (deg); Inflow angle at Blade 3, Node 7; 
AeroDyn['B3N8Phi']            = False     # (deg); Inflow angle at Blade 3, Node 8; 
AeroDyn['B3N9Phi']            = False     # (deg); Inflow angle at Blade 3, Node 9; 
AeroDyn['B1N1Curve']          = False     # (deg); Curvature angle at Blade 1, Node 1; 
AeroDyn['B1N2Curve']          = False     # (deg); Curvature angle at Blade 1, Node 2; 
AeroDyn['B1N3Curve']          = False     # (deg); Curvature angle at Blade 1, Node 3; 
AeroDyn['B1N4Curve']          = False     # (deg); Curvature angle at Blade 1, Node 4; 
AeroDyn['B1N5Curve']          = False     # (deg); Curvature angle at Blade 1, Node 5; 
AeroDyn['B1N6Curve']          = False     # (deg); Curvature angle at Blade 1, Node 6; 
AeroDyn['B1N7Curve']          = False     # (deg); Curvature angle at Blade 1, Node 7; 
AeroDyn['B1N8Curve']          = False     # (deg); Curvature angle at Blade 1, Node 8; 
AeroDyn['B1N9Curve']          = False     # (deg); Curvature angle at Blade 1, Node 9; 
AeroDyn['B2N1Curve']          = False     # (deg); Curvature angle at Blade 2, Node 1; 
AeroDyn['B2N2Curve']          = False     # (deg); Curvature angle at Blade 2, Node 2; 
AeroDyn['B2N3Curve']          = False     # (deg); Curvature angle at Blade 2, Node 3; 
AeroDyn['B2N4Curve']          = False     # (deg); Curvature angle at Blade 2, Node 4; 
AeroDyn['B2N5Curve']          = False     # (deg); Curvature angle at Blade 2, Node 5; 
AeroDyn['B2N6Curve']          = False     # (deg); Curvature angle at Blade 2, Node 6; 
AeroDyn['B2N7Curve']          = False     # (deg); Curvature angle at Blade 2, Node 7; 
AeroDyn['B2N8Curve']          = False     # (deg); Curvature angle at Blade 2, Node 8; 
AeroDyn['B2N9Curve']          = False     # (deg); Curvature angle at Blade 2, Node 9; 
AeroDyn['B3N1Curve']          = False     # (deg); Curvature angle at Blade 3, Node 1; 
AeroDyn['B3N2Curve']          = False     # (deg); Curvature angle at Blade 3, Node 2; 
AeroDyn['B3N3Curve']          = False     # (deg); Curvature angle at Blade 3, Node 3; 
AeroDyn['B3N4Curve']          = False     # (deg); Curvature angle at Blade 3, Node 4; 
AeroDyn['B3N5Curve']          = False     # (deg); Curvature angle at Blade 3, Node 5; 
AeroDyn['B3N6Curve']          = False     # (deg); Curvature angle at Blade 3, Node 6; 
AeroDyn['B3N7Curve']          = False     # (deg); Curvature angle at Blade 3, Node 7; 
AeroDyn['B3N8Curve']          = False     # (deg); Curvature angle at Blade 3, Node 8; 
AeroDyn['B3N9Curve']          = False     # (deg); Curvature angle at Blade 3, Node 9; 
AeroDyn['B1N1Cl']             = False     # (-); Lift force coefficient at Blade 1, Node 1; 
AeroDyn['B1N2Cl']             = False     # (-); Lift force coefficient at Blade 1, Node 2; 
AeroDyn['B1N3Cl']             = False     # (-); Lift force coefficient at Blade 1, Node 3; 
AeroDyn['B1N4Cl']             = False     # (-); Lift force coefficient at Blade 1, Node 4; 
AeroDyn['B1N5Cl']             = False     # (-); Lift force coefficient at Blade 1, Node 5; 
AeroDyn['B1N6Cl']             = False     # (-); Lift force coefficient at Blade 1, Node 6; 
AeroDyn['B1N7Cl']             = False     # (-); Lift force coefficient at Blade 1, Node 7; 
AeroDyn['B1N8Cl']             = False     # (-); Lift force coefficient at Blade 1, Node 8; 
AeroDyn['B1N9Cl']             = False     # (-); Lift force coefficient at Blade 1, Node 9; 
AeroDyn['B2N1Cl']             = False     # (-); Lift force coefficient at Blade 2, Node 1; 
AeroDyn['B2N2Cl']             = False     # (-); Lift force coefficient at Blade 2, Node 2; 
AeroDyn['B2N3Cl']             = False     # (-); Lift force coefficient at Blade 2, Node 3; 
AeroDyn['B2N4Cl']             = False     # (-); Lift force coefficient at Blade 2, Node 4; 
AeroDyn['B2N5Cl']             = False     # (-); Lift force coefficient at Blade 2, Node 5; 
AeroDyn['B2N6Cl']             = False     # (-); Lift force coefficient at Blade 2, Node 6; 
AeroDyn['B2N7Cl']             = False     # (-); Lift force coefficient at Blade 2, Node 7; 
AeroDyn['B2N8Cl']             = False     # (-); Lift force coefficient at Blade 2, Node 8; 
AeroDyn['B2N9Cl']             = False     # (-); Lift force coefficient at Blade 2, Node 9; 
AeroDyn['B3N1Cl']             = False     # (-); Lift force coefficient at Blade 3, Node 1; 
AeroDyn['B3N2Cl']             = False     # (-); Lift force coefficient at Blade 3, Node 2; 
AeroDyn['B3N3Cl']             = False     # (-); Lift force coefficient at Blade 3, Node 3; 
AeroDyn['B3N4Cl']             = False     # (-); Lift force coefficient at Blade 3, Node 4; 
AeroDyn['B3N5Cl']             = False     # (-); Lift force coefficient at Blade 3, Node 5; 
AeroDyn['B3N6Cl']             = False     # (-); Lift force coefficient at Blade 3, Node 6; 
AeroDyn['B3N7Cl']             = False     # (-); Lift force coefficient at Blade 3, Node 7; 
AeroDyn['B3N8Cl']             = False     # (-); Lift force coefficient at Blade 3, Node 8; 
AeroDyn['B3N9Cl']             = False     # (-); Lift force coefficient at Blade 3, Node 9; 
AeroDyn['B1N1Cd']             = False     # (-); Drag force coefficient at Blade 1, Node 1; 
AeroDyn['B1N2Cd']             = False     # (-); Drag force coefficient at Blade 1, Node 2; 
AeroDyn['B1N3Cd']             = False     # (-); Drag force coefficient at Blade 1, Node 3; 
AeroDyn['B1N4Cd']             = False     # (-); Drag force coefficient at Blade 1, Node 4; 
AeroDyn['B1N5Cd']             = False     # (-); Drag force coefficient at Blade 1, Node 5; 
AeroDyn['B1N6Cd']             = False     # (-); Drag force coefficient at Blade 1, Node 6; 
AeroDyn['B1N7Cd']             = False     # (-); Drag force coefficient at Blade 1, Node 7; 
AeroDyn['B1N8Cd']             = False     # (-); Drag force coefficient at Blade 1, Node 8; 
AeroDyn['B1N9Cd']             = False     # (-); Drag force coefficient at Blade 1, Node 9; 
AeroDyn['B2N1Cd']             = False     # (-); Drag force coefficient at Blade 2, Node 1; 
AeroDyn['B2N2Cd']             = False     # (-); Drag force coefficient at Blade 2, Node 2; 
AeroDyn['B2N3Cd']             = False     # (-); Drag force coefficient at Blade 2, Node 3; 
AeroDyn['B2N4Cd']             = False     # (-); Drag force coefficient at Blade 2, Node 4; 
AeroDyn['B2N5Cd']             = False     # (-); Drag force coefficient at Blade 2, Node 5; 
AeroDyn['B2N6Cd']             = False     # (-); Drag force coefficient at Blade 2, Node 6; 
AeroDyn['B2N7Cd']             = False     # (-); Drag force coefficient at Blade 2, Node 7; 
AeroDyn['B2N8Cd']             = False     # (-); Drag force coefficient at Blade 2, Node 8; 
AeroDyn['B2N9Cd']             = False     # (-); Drag force coefficient at Blade 2, Node 9; 
AeroDyn['B3N1Cd']             = False     # (-); Drag force coefficient at Blade 3, Node 1; 
AeroDyn['B3N2Cd']             = False     # (-); Drag force coefficient at Blade 3, Node 2; 
AeroDyn['B3N3Cd']             = False     # (-); Drag force coefficient at Blade 3, Node 3; 
AeroDyn['B3N4Cd']             = False     # (-); Drag force coefficient at Blade 3, Node 4; 
AeroDyn['B3N5Cd']             = False     # (-); Drag force coefficient at Blade 3, Node 5; 
AeroDyn['B3N6Cd']             = False     # (-); Drag force coefficient at Blade 3, Node 6; 
AeroDyn['B3N7Cd']             = False     # (-); Drag force coefficient at Blade 3, Node 7; 
AeroDyn['B3N8Cd']             = False     # (-); Drag force coefficient at Blade 3, Node 8; 
AeroDyn['B3N9Cd']             = False     # (-); Drag force coefficient at Blade 3, Node 9; 
AeroDyn['B1N1Cm']             = False     # (-); Pitching moment coefficient at Blade 1, Node 1; 
AeroDyn['B1N2Cm']             = False     # (-); Pitching moment coefficient at Blade 1, Node 2; 
AeroDyn['B1N3Cm']             = False     # (-); Pitching moment coefficient at Blade 1, Node 3; 
AeroDyn['B1N4Cm']             = False     # (-); Pitching moment coefficient at Blade 1, Node 4; 
AeroDyn['B1N5Cm']             = False     # (-); Pitching moment coefficient at Blade 1, Node 5; 
AeroDyn['B1N6Cm']             = False     # (-); Pitching moment coefficient at Blade 1, Node 6; 
AeroDyn['B1N7Cm']             = False     # (-); Pitching moment coefficient at Blade 1, Node 7; 
AeroDyn['B1N8Cm']             = False     # (-); Pitching moment coefficient at Blade 1, Node 8; 
AeroDyn['B1N9Cm']             = False     # (-); Pitching moment coefficient at Blade 1, Node 9; 
AeroDyn['B2N1Cm']             = False     # (-); Pitching moment coefficient at Blade 2, Node 1; 
AeroDyn['B2N2Cm']             = False     # (-); Pitching moment coefficient at Blade 2, Node 2; 
AeroDyn['B2N3Cm']             = False     # (-); Pitching moment coefficient at Blade 2, Node 3; 
AeroDyn['B2N4Cm']             = False     # (-); Pitching moment coefficient at Blade 2, Node 4; 
AeroDyn['B2N5Cm']             = False     # (-); Pitching moment coefficient at Blade 2, Node 5; 
AeroDyn['B2N6Cm']             = False     # (-); Pitching moment coefficient at Blade 2, Node 6; 
AeroDyn['B2N7Cm']             = False     # (-); Pitching moment coefficient at Blade 2, Node 7; 
AeroDyn['B2N8Cm']             = False     # (-); Pitching moment coefficient at Blade 2, Node 8; 
AeroDyn['B2N9Cm']             = False     # (-); Pitching moment coefficient at Blade 2, Node 9; 
AeroDyn['B3N1Cm']             = False     # (-); Pitching moment coefficient at Blade 3, Node 1; 
AeroDyn['B3N2Cm']             = False     # (-); Pitching moment coefficient at Blade 3, Node 2; 
AeroDyn['B3N3Cm']             = False     # (-); Pitching moment coefficient at Blade 3, Node 3; 
AeroDyn['B3N4Cm']             = False     # (-); Pitching moment coefficient at Blade 3, Node 4; 
AeroDyn['B3N5Cm']             = False     # (-); Pitching moment coefficient at Blade 3, Node 5; 
AeroDyn['B3N6Cm']             = False     # (-); Pitching moment coefficient at Blade 3, Node 6; 
AeroDyn['B3N7Cm']             = False     # (-); Pitching moment coefficient at Blade 3, Node 7; 
AeroDyn['B3N8Cm']             = False     # (-); Pitching moment coefficient at Blade 3, Node 8; 
AeroDyn['B3N9Cm']             = False     # (-); Pitching moment coefficient at Blade 3, Node 9; 
AeroDyn['B1N1Cx']             = False     # (-); Normal force (to plane) coefficient at Blade 1, Node 1; 
AeroDyn['B1N2Cx']             = False     # (-); Normal force (to plane) coefficient at Blade 1, Node 2; 
AeroDyn['B1N3Cx']             = False     # (-); Normal force (to plane) coefficient at Blade 1, Node 3; 
AeroDyn['B1N4Cx']             = False     # (-); Normal force (to plane) coefficient at Blade 1, Node 4; 
AeroDyn['B1N5Cx']             = False     # (-); Normal force (to plane) coefficient at Blade 1, Node 5; 
AeroDyn['B1N6Cx']             = False     # (-); Normal force (to plane) coefficient at Blade 1, Node 6; 
AeroDyn['B1N7Cx']             = False     # (-); Normal force (to plane) coefficient at Blade 1, Node 7; 
AeroDyn['B1N8Cx']             = False     # (-); Normal force (to plane) coefficient at Blade 1, Node 8; 
AeroDyn['B1N9Cx']             = False     # (-); Normal force (to plane) coefficient at Blade 1, Node 9; 
AeroDyn['B2N1Cx']             = False     # (-); Normal force (to plane) coefficient at Blade 2, Node 1; 
AeroDyn['B2N2Cx']             = False     # (-); Normal force (to plane) coefficient at Blade 2, Node 2; 
AeroDyn['B2N3Cx']             = False     # (-); Normal force (to plane) coefficient at Blade 2, Node 3; 
AeroDyn['B2N4Cx']             = False     # (-); Normal force (to plane) coefficient at Blade 2, Node 4; 
AeroDyn['B2N5Cx']             = False     # (-); Normal force (to plane) coefficient at Blade 2, Node 5; 
AeroDyn['B2N6Cx']             = False     # (-); Normal force (to plane) coefficient at Blade 2, Node 6; 
AeroDyn['B2N7Cx']             = False     # (-); Normal force (to plane) coefficient at Blade 2, Node 7; 
AeroDyn['B2N8Cx']             = False     # (-); Normal force (to plane) coefficient at Blade 2, Node 8; 
AeroDyn['B2N9Cx']             = False     # (-); Normal force (to plane) coefficient at Blade 2, Node 9; 
AeroDyn['B3N1Cx']             = False     # (-); Normal force (to plane) coefficient at Blade 3, Node 1; 
AeroDyn['B3N2Cx']             = False     # (-); Normal force (to plane) coefficient at Blade 3, Node 2; 
AeroDyn['B3N3Cx']             = False     # (-); Normal force (to plane) coefficient at Blade 3, Node 3; 
AeroDyn['B3N4Cx']             = False     # (-); Normal force (to plane) coefficient at Blade 3, Node 4; 
AeroDyn['B3N5Cx']             = False     # (-); Normal force (to plane) coefficient at Blade 3, Node 5; 
AeroDyn['B3N6Cx']             = False     # (-); Normal force (to plane) coefficient at Blade 3, Node 6; 
AeroDyn['B3N7Cx']             = False     # (-); Normal force (to plane) coefficient at Blade 3, Node 7; 
AeroDyn['B3N8Cx']             = False     # (-); Normal force (to plane) coefficient at Blade 3, Node 8; 
AeroDyn['B3N9Cx']             = False     # (-); Normal force (to plane) coefficient at Blade 3, Node 9; 
AeroDyn['B1N1Cy']             = False     # (-); Tangential force (to plane) coefficient at Blade 1, Node 1; 
AeroDyn['B1N2Cy']             = False     # (-); Tangential force (to plane) coefficient at Blade 1, Node 2; 
AeroDyn['B1N3Cy']             = False     # (-); Tangential force (to plane) coefficient at Blade 1, Node 3; 
AeroDyn['B1N4Cy']             = False     # (-); Tangential force (to plane) coefficient at Blade 1, Node 4; 
AeroDyn['B1N5Cy']             = False     # (-); Tangential force (to plane) coefficient at Blade 1, Node 5; 
AeroDyn['B1N6Cy']             = False     # (-); Tangential force (to plane) coefficient at Blade 1, Node 6; 
AeroDyn['B1N7Cy']             = False     # (-); Tangential force (to plane) coefficient at Blade 1, Node 7; 
AeroDyn['B1N8Cy']             = False     # (-); Tangential force (to plane) coefficient at Blade 1, Node 8; 
AeroDyn['B1N9Cy']             = False     # (-); Tangential force (to plane) coefficient at Blade 1, Node 9; 
AeroDyn['B2N1Cy']             = False     # (-); Tangential force (to plane) coefficient at Blade 2, Node 1; 
AeroDyn['B2N2Cy']             = False     # (-); Tangential force (to plane) coefficient at Blade 2, Node 2; 
AeroDyn['B2N3Cy']             = False     # (-); Tangential force (to plane) coefficient at Blade 2, Node 3; 
AeroDyn['B2N4Cy']             = False     # (-); Tangential force (to plane) coefficient at Blade 2, Node 4; 
AeroDyn['B2N5Cy']             = False     # (-); Tangential force (to plane) coefficient at Blade 2, Node 5; 
AeroDyn['B2N6Cy']             = False     # (-); Tangential force (to plane) coefficient at Blade 2, Node 6; 
AeroDyn['B2N7Cy']             = False     # (-); Tangential force (to plane) coefficient at Blade 2, Node 7; 
AeroDyn['B2N8Cy']             = False     # (-); Tangential force (to plane) coefficient at Blade 2, Node 8; 
AeroDyn['B2N9Cy']             = False     # (-); Tangential force (to plane) coefficient at Blade 2, Node 9; 
AeroDyn['B3N1Cy']             = False     # (-); Tangential force (to plane) coefficient at Blade 3, Node 1; 
AeroDyn['B3N2Cy']             = False     # (-); Tangential force (to plane) coefficient at Blade 3, Node 2; 
AeroDyn['B3N3Cy']             = False     # (-); Tangential force (to plane) coefficient at Blade 3, Node 3; 
AeroDyn['B3N4Cy']             = False     # (-); Tangential force (to plane) coefficient at Blade 3, Node 4; 
AeroDyn['B3N5Cy']             = False     # (-); Tangential force (to plane) coefficient at Blade 3, Node 5; 
AeroDyn['B3N6Cy']             = False     # (-); Tangential force (to plane) coefficient at Blade 3, Node 6; 
AeroDyn['B3N7Cy']             = False     # (-); Tangential force (to plane) coefficient at Blade 3, Node 7; 
AeroDyn['B3N8Cy']             = False     # (-); Tangential force (to plane) coefficient at Blade 3, Node 8; 
AeroDyn['B3N9Cy']             = False     # (-); Tangential force (to plane) coefficient at Blade 3, Node 9; 
AeroDyn['B1N1Cn']             = False     # (-); Normal force (to chord) coefficient at Blade 1, Node 1; 
AeroDyn['B1N2Cn']             = False     # (-); Normal force (to chord) coefficient at Blade 1, Node 2; 
AeroDyn['B1N3Cn']             = False     # (-); Normal force (to chord) coefficient at Blade 1, Node 3; 
AeroDyn['B1N4Cn']             = False     # (-); Normal force (to chord) coefficient at Blade 1, Node 4; 
AeroDyn['B1N5Cn']             = False     # (-); Normal force (to chord) coefficient at Blade 1, Node 5; 
AeroDyn['B1N6Cn']             = False     # (-); Normal force (to chord) coefficient at Blade 1, Node 6; 
AeroDyn['B1N7Cn']             = False     # (-); Normal force (to chord) coefficient at Blade 1, Node 7; 
AeroDyn['B1N8Cn']             = False     # (-); Normal force (to chord) coefficient at Blade 1, Node 8; 
AeroDyn['B1N9Cn']             = False     # (-); Normal force (to chord) coefficient at Blade 1, Node 9; 
AeroDyn['B2N1Cn']             = False     # (-); Normal force (to chord) coefficient at Blade 2, Node 1; 
AeroDyn['B2N2Cn']             = False     # (-); Normal force (to chord) coefficient at Blade 2, Node 2; 
AeroDyn['B2N3Cn']             = False     # (-); Normal force (to chord) coefficient at Blade 2, Node 3; 
AeroDyn['B2N4Cn']             = False     # (-); Normal force (to chord) coefficient at Blade 2, Node 4; 
AeroDyn['B2N5Cn']             = False     # (-); Normal force (to chord) coefficient at Blade 2, Node 5; 
AeroDyn['B2N6Cn']             = False     # (-); Normal force (to chord) coefficient at Blade 2, Node 6; 
AeroDyn['B2N7Cn']             = False     # (-); Normal force (to chord) coefficient at Blade 2, Node 7; 
AeroDyn['B2N8Cn']             = False     # (-); Normal force (to chord) coefficient at Blade 2, Node 8; 
AeroDyn['B2N9Cn']             = False     # (-); Normal force (to chord) coefficient at Blade 2, Node 9; 
AeroDyn['B3N1Cn']             = False     # (-); Normal force (to chord) coefficient at Blade 3, Node 1; 
AeroDyn['B3N2Cn']             = False     # (-); Normal force (to chord) coefficient at Blade 3, Node 2; 
AeroDyn['B3N3Cn']             = False     # (-); Normal force (to chord) coefficient at Blade 3, Node 3; 
AeroDyn['B3N4Cn']             = False     # (-); Normal force (to chord) coefficient at Blade 3, Node 4; 
AeroDyn['B3N5Cn']             = False     # (-); Normal force (to chord) coefficient at Blade 3, Node 5; 
AeroDyn['B3N6Cn']             = False     # (-); Normal force (to chord) coefficient at Blade 3, Node 6; 
AeroDyn['B3N7Cn']             = False     # (-); Normal force (to chord) coefficient at Blade 3, Node 7; 
AeroDyn['B3N8Cn']             = False     # (-); Normal force (to chord) coefficient at Blade 3, Node 8; 
AeroDyn['B3N9Cn']             = False     # (-); Normal force (to chord) coefficient at Blade 3, Node 9; 
AeroDyn['B1N1Ct']             = False     # (-); Tangential force (to chord) coefficient at Blade 1, Node 1; 
AeroDyn['B1N2Ct']             = False     # (-); Tangential force (to chord) coefficient at Blade 1, Node 2; 
AeroDyn['B1N3Ct']             = False     # (-); Tangential force (to chord) coefficient at Blade 1, Node 3; 
AeroDyn['B1N4Ct']             = False     # (-); Tangential force (to chord) coefficient at Blade 1, Node 4; 
AeroDyn['B1N5Ct']             = False     # (-); Tangential force (to chord) coefficient at Blade 1, Node 5; 
AeroDyn['B1N6Ct']             = False     # (-); Tangential force (to chord) coefficient at Blade 1, Node 6; 
AeroDyn['B1N7Ct']             = False     # (-); Tangential force (to chord) coefficient at Blade 1, Node 7; 
AeroDyn['B1N8Ct']             = False     # (-); Tangential force (to chord) coefficient at Blade 1, Node 8; 
AeroDyn['B1N9Ct']             = False     # (-); Tangential force (to chord) coefficient at Blade 1, Node 9; 
AeroDyn['B2N1Ct']             = False     # (-); Tangential force (to chord) coefficient at Blade 2, Node 1; 
AeroDyn['B2N2Ct']             = False     # (-); Tangential force (to chord) coefficient at Blade 2, Node 2; 
AeroDyn['B2N3Ct']             = False     # (-); Tangential force (to chord) coefficient at Blade 2, Node 3; 
AeroDyn['B2N4Ct']             = False     # (-); Tangential force (to chord) coefficient at Blade 2, Node 4; 
AeroDyn['B2N5Ct']             = False     # (-); Tangential force (to chord) coefficient at Blade 2, Node 5; 
AeroDyn['B2N6Ct']             = False     # (-); Tangential force (to chord) coefficient at Blade 2, Node 6; 
AeroDyn['B2N7Ct']             = False     # (-); Tangential force (to chord) coefficient at Blade 2, Node 7; 
AeroDyn['B2N8Ct']             = False     # (-); Tangential force (to chord) coefficient at Blade 2, Node 8; 
AeroDyn['B2N9Ct']             = False     # (-); Tangential force (to chord) coefficient at Blade 2, Node 9; 
AeroDyn['B3N1Ct']             = False     # (-); Tangential force (to chord) coefficient at Blade 3, Node 1; 
AeroDyn['B3N2Ct']             = False     # (-); Tangential force (to chord) coefficient at Blade 3, Node 2; 
AeroDyn['B3N3Ct']             = False     # (-); Tangential force (to chord) coefficient at Blade 3, Node 3; 
AeroDyn['B3N4Ct']             = False     # (-); Tangential force (to chord) coefficient at Blade 3, Node 4; 
AeroDyn['B3N5Ct']             = False     # (-); Tangential force (to chord) coefficient at Blade 3, Node 5; 
AeroDyn['B3N6Ct']             = False     # (-); Tangential force (to chord) coefficient at Blade 3, Node 6; 
AeroDyn['B3N7Ct']             = False     # (-); Tangential force (to chord) coefficient at Blade 3, Node 7; 
AeroDyn['B3N8Ct']             = False     # (-); Tangential force (to chord) coefficient at Blade 3, Node 8; 
AeroDyn['B3N9Ct']             = False     # (-); Tangential force (to chord) coefficient at Blade 3, Node 9; 
AeroDyn['B1N1Fl']             = False     # (N/m); Lift force per unit length at Blade 1, Node 1; 
AeroDyn['B1N2Fl']             = False     # (N/m); Lift force per unit length at Blade 1, Node 2; 
AeroDyn['B1N3Fl']             = False     # (N/m); Lift force per unit length at Blade 1, Node 3; 
AeroDyn['B1N4Fl']             = False     # (N/m); Lift force per unit length at Blade 1, Node 4; 
AeroDyn['B1N5Fl']             = False     # (N/m); Lift force per unit length at Blade 1, Node 5; 
AeroDyn['B1N6Fl']             = False     # (N/m); Lift force per unit length at Blade 1, Node 6; 
AeroDyn['B1N7Fl']             = False     # (N/m); Lift force per unit length at Blade 1, Node 7; 
AeroDyn['B1N8Fl']             = False     # (N/m); Lift force per unit length at Blade 1, Node 8; 
AeroDyn['B1N9Fl']             = False     # (N/m); Lift force per unit length at Blade 1, Node 9; 
AeroDyn['B2N1Fl']             = False     # (N/m); Lift force per unit length at Blade 2, Node 1; 
AeroDyn['B2N2Fl']             = False     # (N/m); Lift force per unit length at Blade 2, Node 2; 
AeroDyn['B2N3Fl']             = False     # (N/m); Lift force per unit length at Blade 2, Node 3; 
AeroDyn['B2N4Fl']             = False     # (N/m); Lift force per unit length at Blade 2, Node 4; 
AeroDyn['B2N5Fl']             = False     # (N/m); Lift force per unit length at Blade 2, Node 5; 
AeroDyn['B2N6Fl']             = False     # (N/m); Lift force per unit length at Blade 2, Node 6; 
AeroDyn['B2N7Fl']             = False     # (N/m); Lift force per unit length at Blade 2, Node 7; 
AeroDyn['B2N8Fl']             = False     # (N/m); Lift force per unit length at Blade 2, Node 8; 
AeroDyn['B2N9Fl']             = False     # (N/m); Lift force per unit length at Blade 2, Node 9; 
AeroDyn['B3N1Fl']             = False     # (N/m); Lift force per unit length at Blade 3, Node 1; 
AeroDyn['B3N2Fl']             = False     # (N/m); Lift force per unit length at Blade 3, Node 2; 
AeroDyn['B3N3Fl']             = False     # (N/m); Lift force per unit length at Blade 3, Node 3; 
AeroDyn['B3N4Fl']             = False     # (N/m); Lift force per unit length at Blade 3, Node 4; 
AeroDyn['B3N5Fl']             = False     # (N/m); Lift force per unit length at Blade 3, Node 5; 
AeroDyn['B3N6Fl']             = False     # (N/m); Lift force per unit length at Blade 3, Node 6; 
AeroDyn['B3N7Fl']             = False     # (N/m); Lift force per unit length at Blade 3, Node 7; 
AeroDyn['B3N8Fl']             = False     # (N/m); Lift force per unit length at Blade 3, Node 8; 
AeroDyn['B3N9Fl']             = False     # (N/m); Lift force per unit length at Blade 3, Node 9; 
AeroDyn['B1N1Fd']             = False     # (N/m); Drag force per unit length at Blade 1, Node 1; 
AeroDyn['B1N2Fd']             = False     # (N/m); Drag force per unit length at Blade 1, Node 2; 
AeroDyn['B1N3Fd']             = False     # (N/m); Drag force per unit length at Blade 1, Node 3; 
AeroDyn['B1N4Fd']             = False     # (N/m); Drag force per unit length at Blade 1, Node 4; 
AeroDyn['B1N5Fd']             = False     # (N/m); Drag force per unit length at Blade 1, Node 5; 
AeroDyn['B1N6Fd']             = False     # (N/m); Drag force per unit length at Blade 1, Node 6; 
AeroDyn['B1N7Fd']             = False     # (N/m); Drag force per unit length at Blade 1, Node 7; 
AeroDyn['B1N8Fd']             = False     # (N/m); Drag force per unit length at Blade 1, Node 8; 
AeroDyn['B1N9Fd']             = False     # (N/m); Drag force per unit length at Blade 1, Node 9; 
AeroDyn['B2N1Fd']             = False     # (N/m); Drag force per unit length at Blade 2, Node 1; 
AeroDyn['B2N2Fd']             = False     # (N/m); Drag force per unit length at Blade 2, Node 2; 
AeroDyn['B2N3Fd']             = False     # (N/m); Drag force per unit length at Blade 2, Node 3; 
AeroDyn['B2N4Fd']             = False     # (N/m); Drag force per unit length at Blade 2, Node 4; 
AeroDyn['B2N5Fd']             = False     # (N/m); Drag force per unit length at Blade 2, Node 5; 
AeroDyn['B2N6Fd']             = False     # (N/m); Drag force per unit length at Blade 2, Node 6; 
AeroDyn['B2N7Fd']             = False     # (N/m); Drag force per unit length at Blade 2, Node 7; 
AeroDyn['B2N8Fd']             = False     # (N/m); Drag force per unit length at Blade 2, Node 8; 
AeroDyn['B2N9Fd']             = False     # (N/m); Drag force per unit length at Blade 2, Node 9; 
AeroDyn['B3N1Fd']             = False     # (N/m); Drag force per unit length at Blade 3, Node 1; 
AeroDyn['B3N2Fd']             = False     # (N/m); Drag force per unit length at Blade 3, Node 2; 
AeroDyn['B3N3Fd']             = False     # (N/m); Drag force per unit length at Blade 3, Node 3; 
AeroDyn['B3N4Fd']             = False     # (N/m); Drag force per unit length at Blade 3, Node 4; 
AeroDyn['B3N5Fd']             = False     # (N/m); Drag force per unit length at Blade 3, Node 5; 
AeroDyn['B3N6Fd']             = False     # (N/m); Drag force per unit length at Blade 3, Node 6; 
AeroDyn['B3N7Fd']             = False     # (N/m); Drag force per unit length at Blade 3, Node 7; 
AeroDyn['B3N8Fd']             = False     # (N/m); Drag force per unit length at Blade 3, Node 8; 
AeroDyn['B3N9Fd']             = False     # (N/m); Drag force per unit length at Blade 3, Node 9; 
AeroDyn['B1N1Mm']             = False     # (N m/m); Pitching moment per unit length at Blade 1, Node 1; 
AeroDyn['B1N2Mm']             = False     # (N m/m); Pitching moment per unit length at Blade 1, Node 2; 
AeroDyn['B1N3Mm']             = False     # (N m/m); Pitching moment per unit length at Blade 1, Node 3; 
AeroDyn['B1N4Mm']             = False     # (N m/m); Pitching moment per unit length at Blade 1, Node 4; 
AeroDyn['B1N5Mm']             = False     # (N m/m); Pitching moment per unit length at Blade 1, Node 5; 
AeroDyn['B1N6Mm']             = False     # (N m/m); Pitching moment per unit length at Blade 1, Node 6; 
AeroDyn['B1N7Mm']             = False     # (N m/m); Pitching moment per unit length at Blade 1, Node 7; 
AeroDyn['B1N8Mm']             = False     # (N m/m); Pitching moment per unit length at Blade 1, Node 8; 
AeroDyn['B1N9Mm']             = False     # (N m/m); Pitching moment per unit length at Blade 1, Node 9; 
AeroDyn['B2N1Mm']             = False     # (N m/m); Pitching moment per unit length at Blade 2, Node 1; 
AeroDyn['B2N2Mm']             = False     # (N m/m); Pitching moment per unit length at Blade 2, Node 2; 
AeroDyn['B2N3Mm']             = False     # (N m/m); Pitching moment per unit length at Blade 2, Node 3; 
AeroDyn['B2N4Mm']             = False     # (N m/m); Pitching moment per unit length at Blade 2, Node 4; 
AeroDyn['B2N5Mm']             = False     # (N m/m); Pitching moment per unit length at Blade 2, Node 5; 
AeroDyn['B2N6Mm']             = False     # (N m/m); Pitching moment per unit length at Blade 2, Node 6; 
AeroDyn['B2N7Mm']             = False     # (N m/m); Pitching moment per unit length at Blade 2, Node 7; 
AeroDyn['B2N8Mm']             = False     # (N m/m); Pitching moment per unit length at Blade 2, Node 8; 
AeroDyn['B2N9Mm']             = False     # (N m/m); Pitching moment per unit length at Blade 2, Node 9; 
AeroDyn['B3N1Mm']             = False     # (N m/m); Pitching moment per unit length at Blade 3, Node 1; 
AeroDyn['B3N2Mm']             = False     # (N m/m); Pitching moment per unit length at Blade 3, Node 2; 
AeroDyn['B3N3Mm']             = False     # (N m/m); Pitching moment per unit length at Blade 3, Node 3; 
AeroDyn['B3N4Mm']             = False     # (N m/m); Pitching moment per unit length at Blade 3, Node 4; 
AeroDyn['B3N5Mm']             = False     # (N m/m); Pitching moment per unit length at Blade 3, Node 5; 
AeroDyn['B3N6Mm']             = False     # (N m/m); Pitching moment per unit length at Blade 3, Node 6; 
AeroDyn['B3N7Mm']             = False     # (N m/m); Pitching moment per unit length at Blade 3, Node 7; 
AeroDyn['B3N8Mm']             = False     # (N m/m); Pitching moment per unit length at Blade 3, Node 8; 
AeroDyn['B3N9Mm']             = False     # (N m/m); Pitching moment per unit length at Blade 3, Node 9; 
AeroDyn['B1N1Fx']             = False     # (N/m); Normal force (to plane) per unit length at Blade 1, Node 1; 
AeroDyn['B1N2Fx']             = False     # (N/m); Normal force (to plane) per unit length at Blade 1, Node 2; 
AeroDyn['B1N3Fx']             = False     # (N/m); Normal force (to plane) per unit length at Blade 1, Node 3; 
AeroDyn['B1N4Fx']             = False     # (N/m); Normal force (to plane) per unit length at Blade 1, Node 4; 
AeroDyn['B1N5Fx']             = False     # (N/m); Normal force (to plane) per unit length at Blade 1, Node 5; 
AeroDyn['B1N6Fx']             = False     # (N/m); Normal force (to plane) per unit length at Blade 1, Node 6; 
AeroDyn['B1N7Fx']             = False     # (N/m); Normal force (to plane) per unit length at Blade 1, Node 7; 
AeroDyn['B1N8Fx']             = False     # (N/m); Normal force (to plane) per unit length at Blade 1, Node 8; 
AeroDyn['B1N9Fx']             = False     # (N/m); Normal force (to plane) per unit length at Blade 1, Node 9; 
AeroDyn['B2N1Fx']             = False     # (N/m); Normal force (to plane) per unit length at Blade 2, Node 1; 
AeroDyn['B2N2Fx']             = False     # (N/m); Normal force (to plane) per unit length at Blade 2, Node 2; 
AeroDyn['B2N3Fx']             = False     # (N/m); Normal force (to plane) per unit length at Blade 2, Node 3; 
AeroDyn['B2N4Fx']             = False     # (N/m); Normal force (to plane) per unit length at Blade 2, Node 4; 
AeroDyn['B2N5Fx']             = False     # (N/m); Normal force (to plane) per unit length at Blade 2, Node 5; 
AeroDyn['B2N6Fx']             = False     # (N/m); Normal force (to plane) per unit length at Blade 2, Node 6; 
AeroDyn['B2N7Fx']             = False     # (N/m); Normal force (to plane) per unit length at Blade 2, Node 7; 
AeroDyn['B2N8Fx']             = False     # (N/m); Normal force (to plane) per unit length at Blade 2, Node 8; 
AeroDyn['B2N9Fx']             = False     # (N/m); Normal force (to plane) per unit length at Blade 2, Node 9; 
AeroDyn['B3N1Fx']             = False     # (N/m); Normal force (to plane) per unit length at Blade 3, Node 1; 
AeroDyn['B3N2Fx']             = False     # (N/m); Normal force (to plane) per unit length at Blade 3, Node 2; 
AeroDyn['B3N3Fx']             = False     # (N/m); Normal force (to plane) per unit length at Blade 3, Node 3; 
AeroDyn['B3N4Fx']             = False     # (N/m); Normal force (to plane) per unit length at Blade 3, Node 4; 
AeroDyn['B3N5Fx']             = False     # (N/m); Normal force (to plane) per unit length at Blade 3, Node 5; 
AeroDyn['B3N6Fx']             = False     # (N/m); Normal force (to plane) per unit length at Blade 3, Node 6; 
AeroDyn['B3N7Fx']             = False     # (N/m); Normal force (to plane) per unit length at Blade 3, Node 7; 
AeroDyn['B3N8Fx']             = False     # (N/m); Normal force (to plane) per unit length at Blade 3, Node 8; 
AeroDyn['B3N9Fx']             = False     # (N/m); Normal force (to plane) per unit length at Blade 3, Node 9; 
AeroDyn['B1N1Fy']             = False     # (N/m); Tangential force (to plane) per unit length at Blade 1, Node 1; 
AeroDyn['B1N2Fy']             = False     # (N/m); Tangential force (to plane) per unit length at Blade 1, Node 2; 
AeroDyn['B1N3Fy']             = False     # (N/m); Tangential force (to plane) per unit length at Blade 1, Node 3; 
AeroDyn['B1N4Fy']             = False     # (N/m); Tangential force (to plane) per unit length at Blade 1, Node 4; 
AeroDyn['B1N5Fy']             = False     # (N/m); Tangential force (to plane) per unit length at Blade 1, Node 5; 
AeroDyn['B1N6Fy']             = False     # (N/m); Tangential force (to plane) per unit length at Blade 1, Node 6; 
AeroDyn['B1N7Fy']             = False     # (N/m); Tangential force (to plane) per unit length at Blade 1, Node 7; 
AeroDyn['B1N8Fy']             = False     # (N/m); Tangential force (to plane) per unit length at Blade 1, Node 8; 
AeroDyn['B1N9Fy']             = False     # (N/m); Tangential force (to plane) per unit length at Blade 1, Node 9; 
AeroDyn['B2N1Fy']             = False     # (N/m); Tangential force (to plane) per unit length at Blade 2, Node 1; 
AeroDyn['B2N2Fy']             = False     # (N/m); Tangential force (to plane) per unit length at Blade 2, Node 2; 
AeroDyn['B2N3Fy']             = False     # (N/m); Tangential force (to plane) per unit length at Blade 2, Node 3; 
AeroDyn['B2N4Fy']             = False     # (N/m); Tangential force (to plane) per unit length at Blade 2, Node 4; 
AeroDyn['B2N5Fy']             = False     # (N/m); Tangential force (to plane) per unit length at Blade 2, Node 5; 
AeroDyn['B2N6Fy']             = False     # (N/m); Tangential force (to plane) per unit length at Blade 2, Node 6; 
AeroDyn['B2N7Fy']             = False     # (N/m); Tangential force (to plane) per unit length at Blade 2, Node 7; 
AeroDyn['B2N8Fy']             = False     # (N/m); Tangential force (to plane) per unit length at Blade 2, Node 8; 
AeroDyn['B2N9Fy']             = False     # (N/m); Tangential force (to plane) per unit length at Blade 2, Node 9; 
AeroDyn['B3N1Fy']             = False     # (N/m); Tangential force (to plane) per unit length at Blade 3, Node 1; 
AeroDyn['B3N2Fy']             = False     # (N/m); Tangential force (to plane) per unit length at Blade 3, Node 2; 
AeroDyn['B3N3Fy']             = False     # (N/m); Tangential force (to plane) per unit length at Blade 3, Node 3; 
AeroDyn['B3N4Fy']             = False     # (N/m); Tangential force (to plane) per unit length at Blade 3, Node 4; 
AeroDyn['B3N5Fy']             = False     # (N/m); Tangential force (to plane) per unit length at Blade 3, Node 5; 
AeroDyn['B3N6Fy']             = False     # (N/m); Tangential force (to plane) per unit length at Blade 3, Node 6; 
AeroDyn['B3N7Fy']             = False     # (N/m); Tangential force (to plane) per unit length at Blade 3, Node 7; 
AeroDyn['B3N8Fy']             = False     # (N/m); Tangential force (to plane) per unit length at Blade 3, Node 8; 
AeroDyn['B3N9Fy']             = False     # (N/m); Tangential force (to plane) per unit length at Blade 3, Node 9; 
AeroDyn['B1N1Fn']             = False     # (N/m); Normal force (to chord) per unit length at Blade 1, Node 1; 
AeroDyn['B1N2Fn']             = False     # (N/m); Normal force (to chord) per unit length at Blade 1, Node 2; 
AeroDyn['B1N3Fn']             = False     # (N/m); Normal force (to chord) per unit length at Blade 1, Node 3; 
AeroDyn['B1N4Fn']             = False     # (N/m); Normal force (to chord) per unit length at Blade 1, Node 4; 
AeroDyn['B1N5Fn']             = False     # (N/m); Normal force (to chord) per unit length at Blade 1, Node 5; 
AeroDyn['B1N6Fn']             = False     # (N/m); Normal force (to chord) per unit length at Blade 1, Node 6; 
AeroDyn['B1N7Fn']             = False     # (N/m); Normal force (to chord) per unit length at Blade 1, Node 7; 
AeroDyn['B1N8Fn']             = False     # (N/m); Normal force (to chord) per unit length at Blade 1, Node 8; 
AeroDyn['B1N9Fn']             = False     # (N/m); Normal force (to chord) per unit length at Blade 1, Node 9; 
AeroDyn['B2N1Fn']             = False     # (N/m); Normal force (to chord) per unit length at Blade 2, Node 1; 
AeroDyn['B2N2Fn']             = False     # (N/m); Normal force (to chord) per unit length at Blade 2, Node 2; 
AeroDyn['B2N3Fn']             = False     # (N/m); Normal force (to chord) per unit length at Blade 2, Node 3; 
AeroDyn['B2N4Fn']             = False     # (N/m); Normal force (to chord) per unit length at Blade 2, Node 4; 
AeroDyn['B2N5Fn']             = False     # (N/m); Normal force (to chord) per unit length at Blade 2, Node 5; 
AeroDyn['B2N6Fn']             = False     # (N/m); Normal force (to chord) per unit length at Blade 2, Node 6; 
AeroDyn['B2N7Fn']             = False     # (N/m); Normal force (to chord) per unit length at Blade 2, Node 7; 
AeroDyn['B2N8Fn']             = False     # (N/m); Normal force (to chord) per unit length at Blade 2, Node 8; 
AeroDyn['B2N9Fn']             = False     # (N/m); Normal force (to chord) per unit length at Blade 2, Node 9; 
AeroDyn['B3N1Fn']             = False     # (N/m); Normal force (to chord) per unit length at Blade 3, Node 1; 
AeroDyn['B3N2Fn']             = False     # (N/m); Normal force (to chord) per unit length at Blade 3, Node 2; 
AeroDyn['B3N3Fn']             = False     # (N/m); Normal force (to chord) per unit length at Blade 3, Node 3; 
AeroDyn['B3N4Fn']             = False     # (N/m); Normal force (to chord) per unit length at Blade 3, Node 4; 
AeroDyn['B3N5Fn']             = False     # (N/m); Normal force (to chord) per unit length at Blade 3, Node 5; 
AeroDyn['B3N6Fn']             = False     # (N/m); Normal force (to chord) per unit length at Blade 3, Node 6; 
AeroDyn['B3N7Fn']             = False     # (N/m); Normal force (to chord) per unit length at Blade 3, Node 7; 
AeroDyn['B3N8Fn']             = False     # (N/m); Normal force (to chord) per unit length at Blade 3, Node 8; 
AeroDyn['B3N9Fn']             = False     # (N/m); Normal force (to chord) per unit length at Blade 3, Node 9; 
AeroDyn['B1N1Ft']             = False     # (N/m); Tangential force (to chord) per unit length at Blade 1, Node 1; 
AeroDyn['B1N2Ft']             = False     # (N/m); Tangential force (to chord) per unit length at Blade 1, Node 2; 
AeroDyn['B1N3Ft']             = False     # (N/m); Tangential force (to chord) per unit length at Blade 1, Node 3; 
AeroDyn['B1N4Ft']             = False     # (N/m); Tangential force (to chord) per unit length at Blade 1, Node 4; 
AeroDyn['B1N5Ft']             = False     # (N/m); Tangential force (to chord) per unit length at Blade 1, Node 5; 
AeroDyn['B1N6Ft']             = False     # (N/m); Tangential force (to chord) per unit length at Blade 1, Node 6; 
AeroDyn['B1N7Ft']             = False     # (N/m); Tangential force (to chord) per unit length at Blade 1, Node 7; 
AeroDyn['B1N8Ft']             = False     # (N/m); Tangential force (to chord) per unit length at Blade 1, Node 8; 
AeroDyn['B1N9Ft']             = False     # (N/m); Tangential force (to chord) per unit length at Blade 1, Node 9; 
AeroDyn['B2N1Ft']             = False     # (N/m); Tangential force (to chord) per unit length at Blade 2, Node 1; 
AeroDyn['B2N2Ft']             = False     # (N/m); Tangential force (to chord) per unit length at Blade 2, Node 2; 
AeroDyn['B2N3Ft']             = False     # (N/m); Tangential force (to chord) per unit length at Blade 2, Node 3; 
AeroDyn['B2N4Ft']             = False     # (N/m); Tangential force (to chord) per unit length at Blade 2, Node 4; 
AeroDyn['B2N5Ft']             = False     # (N/m); Tangential force (to chord) per unit length at Blade 2, Node 5; 
AeroDyn['B2N6Ft']             = False     # (N/m); Tangential force (to chord) per unit length at Blade 2, Node 6; 
AeroDyn['B2N7Ft']             = False     # (N/m); Tangential force (to chord) per unit length at Blade 2, Node 7; 
AeroDyn['B2N8Ft']             = False     # (N/m); Tangential force (to chord) per unit length at Blade 2, Node 8; 
AeroDyn['B2N9Ft']             = False     # (N/m); Tangential force (to chord) per unit length at Blade 2, Node 9; 
AeroDyn['B3N1Ft']             = False     # (N/m); Tangential force (to chord) per unit length at Blade 3, Node 1; 
AeroDyn['B3N2Ft']             = False     # (N/m); Tangential force (to chord) per unit length at Blade 3, Node 2; 
AeroDyn['B3N3Ft']             = False     # (N/m); Tangential force (to chord) per unit length at Blade 3, Node 3; 
AeroDyn['B3N4Ft']             = False     # (N/m); Tangential force (to chord) per unit length at Blade 3, Node 4; 
AeroDyn['B3N5Ft']             = False     # (N/m); Tangential force (to chord) per unit length at Blade 3, Node 5; 
AeroDyn['B3N6Ft']             = False     # (N/m); Tangential force (to chord) per unit length at Blade 3, Node 6; 
AeroDyn['B3N7Ft']             = False     # (N/m); Tangential force (to chord) per unit length at Blade 3, Node 7; 
AeroDyn['B3N8Ft']             = False     # (N/m); Tangential force (to chord) per unit length at Blade 3, Node 8; 
AeroDyn['B3N9Ft']             = False     # (N/m); Tangential force (to chord) per unit length at Blade 3, Node 9; 
AeroDyn['B1N1Clrnc']          = False     # (m); Tower clearance at Blade 1, Node 1 (based on the absolute distance to the nearest point in the tower from B1N1 minus the local tower radius, in the deflected configuration); please note that this clearance is only approximate because the calculation assumes that the blade is a line with no volume (however, the calculation does use the local tower radius); when B1N1 is above the tower top (or below the tower base), the absolute distance to the tower top (or base) minus the local tower radius, in the deflected configuration, is output; 
AeroDyn['B1N2Clrnc']          = False     # (m); Tower clearance at Blade 1, Node 2 (based on the absolute distance to the nearest point in the tower from B1N2 minus the local tower radius, in the deflected configuration); please note that this clearance is only approximate because the calculation assumes that the blade is a line with no volume (however, the calculation does use the local tower radius); when B1N2 is above the tower top (or below the tower base), the absolute distance to the tower top (or base) minus the local tower radius, in the deflected configuration, is output; 
AeroDyn['B1N3Clrnc']          = False     # (m); Tower clearance at Blade 1, Node 3 (based on the absolute distance to the nearest point in the tower from B1N3 minus the local tower radius, in the deflected configuration); please note that this clearance is only approximate because the calculation assumes that the blade is a line with no volume (however, the calculation does use the local tower radius); when B1N3 is above the tower top (or below the tower base), the absolute distance to the tower top (or base) minus the local tower radius, in the deflected configuration, is output; 
AeroDyn['B1N4Clrnc']          = False     # (m); Tower clearance at Blade 1, Node 4 (based on the absolute distance to the nearest point in the tower from B1N4 minus the local tower radius, in the deflected configuration); please note that this clearance is only approximate because the calculation assumes that the blade is a line with no volume (however, the calculation does use the local tower radius); when B1N4 is above the tower top (or below the tower base), the absolute distance to the tower top (or base) minus the local tower radius, in the deflected configuration, is output; 
AeroDyn['B1N5Clrnc']          = False     # (m); Tower clearance at Blade 1, Node 5 (based on the absolute distance to the nearest point in the tower from B1N5 minus the local tower radius, in the deflected configuration); please note that this clearance is only approximate because the calculation assumes that the blade is a line with no volume (however, the calculation does use the local tower radius); when B1N5 is above the tower top (or below the tower base), the absolute distance to the tower top (or base) minus the local tower radius, in the deflected configuration, is output; 
AeroDyn['B1N6Clrnc']          = False     # (m); Tower clearance at Blade 1, Node 6 (based on the absolute distance to the nearest point in the tower from B1N6 minus the local tower radius, in the deflected configuration); please note that this clearance is only approximate because the calculation assumes that the blade is a line with no volume (however, the calculation does use the local tower radius); when B1N6 is above the tower top (or below the tower base), the absolute distance to the tower top (or base) minus the local tower radius, in the deflected configuration, is output; 
AeroDyn['B1N7Clrnc']          = False     # (m); Tower clearance at Blade 1, Node 7 (based on the absolute distance to the nearest point in the tower from B1N7 minus the local tower radius, in the deflected configuration); please note that this clearance is only approximate because the calculation assumes that the blade is a line with no volume (however, the calculation does use the local tower radius); when B1N7 is above the tower top (or below the tower base), the absolute distance to the tower top (or base) minus the local tower radius, in the deflected configuration, is output; 
AeroDyn['B1N8Clrnc']          = False     # (m); Tower clearance at Blade 1, Node 8 (based on the absolute distance to the nearest point in the tower from B1N8 minus the local tower radius, in the deflected configuration); please note that this clearance is only approximate because the calculation assumes that the blade is a line with no volume (however, the calculation does use the local tower radius); when B1N8 is above the tower top (or below the tower base), the absolute distance to the tower top (or base) minus the local tower radius, in the deflected configuration, is output; 
AeroDyn['B1N9Clrnc']          = False     # (m); Tower clearance at Blade 1, Node 9 (based on the absolute distance to the nearest point in the tower from B1N9 minus the local tower radius, in the deflected configuration); please note that this clearance is only approximate because the calculation assumes that the blade is a line with no volume (however, the calculation does use the local tower radius); when B1N9 is above the tower top (or below the tower base), the absolute distance to the tower top (or base) minus the local tower radius, in the deflected configuration, is output; 
AeroDyn['B2N1Clrnc']          = False     # (m); Tower clearance at Blade 2, Node 1 (based on the absolute distance to the nearest point in the tower from B2N1 minus the local tower radius, in the deflected configuration); please note that this clearance is only approximate because the calculation assumes that the blade is a line with no volume (however, the calculation does use the local tower radius); when B2N1 is above the tower top (or below the tower base), the absolute distance to the tower top (or base) minus the local tower radius, in the deflected configuration, is output; 
AeroDyn['B2N2Clrnc']          = False     # (m); Tower clearance at Blade 2, Node 2 (based on the absolute distance to the nearest point in the tower from B2N2 minus the local tower radius, in the deflected configuration); please note that this clearance is only approximate because the calculation assumes that the blade is a line with no volume (however, the calculation does use the local tower radius); when B2N2 is above the tower top (or below the tower base), the absolute distance to the tower top (or base) minus the local tower radius, in the deflected configuration, is output; 
AeroDyn['B2N3Clrnc']          = False     # (m); Tower clearance at Blade 2, Node 3 (based on the absolute distance to the nearest point in the tower from B2N3 minus the local tower radius, in the deflected configuration); please note that this clearance is only approximate because the calculation assumes that the blade is a line with no volume (however, the calculation does use the local tower radius); when B2N3 is above the tower top (or below the tower base), the absolute distance to the tower top (or base) minus the local tower radius, in the deflected configuration, is output; 
AeroDyn['B2N4Clrnc']          = False     # (m); Tower clearance at Blade 2, Node 4 (based on the absolute distance to the nearest point in the tower from B2N4 minus the local tower radius, in the deflected configuration); please note that this clearance is only approximate because the calculation assumes that the blade is a line with no volume (however, the calculation does use the local tower radius); when B2N4 is above the tower top (or below the tower base), the absolute distance to the tower top (or base) minus the local tower radius, in the deflected configuration, is output; 
AeroDyn['B2N5Clrnc']          = False     # (m); Tower clearance at Blade 2, Node 5 (based on the absolute distance to the nearest point in the tower from B2N5 minus the local tower radius, in the deflected configuration); please note that this clearance is only approximate because the calculation assumes that the blade is a line with no volume (however, the calculation does use the local tower radius); when B2N5 is above the tower top (or below the tower base), the absolute distance to the tower top (or base) minus the local tower radius, in the deflected configuration, is output; 
AeroDyn['B2N6Clrnc']          = False     # (m); Tower clearance at Blade 2, Node 6 (based on the absolute distance to the nearest point in the tower from B2N6 minus the local tower radius, in the deflected configuration); please note that this clearance is only approximate because the calculation assumes that the blade is a line with no volume (however, the calculation does use the local tower radius); when B2N6 is above the tower top (or below the tower base), the absolute distance to the tower top (or base) minus the local tower radius, in the deflected configuration, is output; 
AeroDyn['B2N7Clrnc']          = False     # (m); Tower clearance at Blade 2, Node 7 (based on the absolute distance to the nearest point in the tower from B2N7 minus the local tower radius, in the deflected configuration); please note that this clearance is only approximate because the calculation assumes that the blade is a line with no volume (however, the calculation does use the local tower radius); when B2N7 is above the tower top (or below the tower base), the absolute distance to the tower top (or base) minus the local tower radius, in the deflected configuration, is output; 
AeroDyn['B2N8Clrnc']          = False     # (m); Tower clearance at Blade 2, Node 8 (based on the absolute distance to the nearest point in the tower from B2N8 minus the local tower radius, in the deflected configuration); please note that this clearance is only approximate because the calculation assumes that the blade is a line with no volume (however, the calculation does use the local tower radius); when B2N8 is above the tower top (or below the tower base), the absolute distance to the tower top (or base) minus the local tower radius, in the deflected configuration, is output; 
AeroDyn['B2N9Clrnc']          = False     # (m); Tower clearance at Blade 2, Node 9 (based on the absolute distance to the nearest point in the tower from B2N9 minus the local tower radius, in the deflected configuration); please note that this clearance is only approximate because the calculation assumes that the blade is a line with no volume (however, the calculation does use the local tower radius); when B2N9 is above the tower top (or below the tower base), the absolute distance to the tower top (or base) minus the local tower radius, in the deflected configuration, is output; 
AeroDyn['B3N1Clrnc']          = False     # (m); Tower clearance at Blade 3, Node 1 (based on the absolute distance to the nearest point in the tower from B3N1 minus the local tower radius, in the deflected configuration); please note that this clearance is only approximate because the calculation assumes that the blade is a line with no volume (however, the calculation does use the local tower radius); when B3N1 is above the tower top (or below the tower base), the absolute distance to the tower top (or base) minus the local tower radius, in the deflected configuration, is output; 
AeroDyn['B3N2Clrnc']          = False     # (m); Tower clearance at Blade 3, Node 2 (based on the absolute distance to the nearest point in the tower from B3N2 minus the local tower radius, in the deflected configuration); please note that this clearance is only approximate because the calculation assumes that the blade is a line with no volume (however, the calculation does use the local tower radius); when B3N2 is above the tower top (or below the tower base), the absolute distance to the tower top (or base) minus the local tower radius, in the deflected configuration, is output; 
AeroDyn['B3N3Clrnc']          = False     # (m); Tower clearance at Blade 3, Node 3 (based on the absolute distance to the nearest point in the tower from B3N3 minus the local tower radius, in the deflected configuration); please note that this clearance is only approximate because the calculation assumes that the blade is a line with no volume (however, the calculation does use the local tower radius); when B3N3 is above the tower top (or below the tower base), the absolute distance to the tower top (or base) minus the local tower radius, in the deflected configuration, is output; 
AeroDyn['B3N4Clrnc']          = False     # (m); Tower clearance at Blade 3, Node 4 (based on the absolute distance to the nearest point in the tower from B3N4 minus the local tower radius, in the deflected configuration); please note that this clearance is only approximate because the calculation assumes that the blade is a line with no volume (however, the calculation does use the local tower radius); when B3N4 is above the tower top (or below the tower base), the absolute distance to the tower top (or base) minus the local tower radius, in the deflected configuration, is output; 
AeroDyn['B3N5Clrnc']          = False     # (m); Tower clearance at Blade 3, Node 5 (based on the absolute distance to the nearest point in the tower from B3N5 minus the local tower radius, in the deflected configuration); please note that this clearance is only approximate because the calculation assumes that the blade is a line with no volume (however, the calculation does use the local tower radius); when B3N5 is above the tower top (or below the tower base), the absolute distance to the tower top (or base) minus the local tower radius, in the deflected configuration, is output; 
AeroDyn['B3N6Clrnc']          = False     # (m); Tower clearance at Blade 3, Node 6 (based on the absolute distance to the nearest point in the tower from B3N6 minus the local tower radius, in the deflected configuration); please note that this clearance is only approximate because the calculation assumes that the blade is a line with no volume (however, the calculation does use the local tower radius); when B3N6 is above the tower top (or below the tower base), the absolute distance to the tower top (or base) minus the local tower radius, in the deflected configuration, is output; 
AeroDyn['B3N7Clrnc']          = False     # (m); Tower clearance at Blade 3, Node 7 (based on the absolute distance to the nearest point in the tower from B3N7 minus the local tower radius, in the deflected configuration); please note that this clearance is only approximate because the calculation assumes that the blade is a line with no volume (however, the calculation does use the local tower radius); when B3N7 is above the tower top (or below the tower base), the absolute distance to the tower top (or base) minus the local tower radius, in the deflected configuration, is output; 
AeroDyn['B3N8Clrnc']          = False     # (m); Tower clearance at Blade 3, Node 8 (based on the absolute distance to the nearest point in the tower from B3N8 minus the local tower radius, in the deflected configuration); please note that this clearance is only approximate because the calculation assumes that the blade is a line with no volume (however, the calculation does use the local tower radius); when B3N8 is above the tower top (or below the tower base), the absolute distance to the tower top (or base) minus the local tower radius, in the deflected configuration, is output; 
AeroDyn['B3N9Clrnc']          = False     # (m); Tower clearance at Blade 3, Node 9 (based on the absolute distance to the nearest point in the tower from B3N9 minus the local tower radius, in the deflected configuration); please note that this clearance is only approximate because the calculation assumes that the blade is a line with no volume (however, the calculation does use the local tower radius); when B3N9 is above the tower top (or below the tower base), the absolute distance to the tower top (or base) minus the local tower radius, in the deflected configuration, is output; 

# Rotor
AeroDyn['RtSpeed']            = False     # (rpm); Rotor speed; 
AeroDyn['RtTSR']              = False     # (-); Rotor tip-speed ratio; 
AeroDyn['RtVAvgxh']           = False     # (m/s); Rotor-disk-averaged relative wind velocity (x-component);  the hub coordinate system
AeroDyn['RtVAvgyh']           = False     # (m/s); Rotor-disk-averaged relative wind velocity (y-component);  the hub coordinate system
AeroDyn['RtVAvgzh']           = False     # (m/s); Rotor-disk-averaged relative wind velocity (z-component);  the hub coordinate system
AeroDyn['RtSkew']             = False     # (deg); Rotor inflow-skew angle; 
AeroDyn['RtAeroFxh']          = False     # (N); Total rotor aerodynamic load (force in x direction);  the hub coordinate system
AeroDyn['RtAeroFyh']          = False     # (N); Total rotor aerodynamic load (force in y direction);  the hub coordinate system
AeroDyn['RtAeroFzh']          = False     # (N); Total rotor aerodynamic load (force in z direction);  the hub coordinate system
AeroDyn['RtAeroMxh']          = False     # (N m); Total rotor aerodynamic load (moment in x direction);  the hub coordinate system
AeroDyn['RtAeroMyh']          = False     # (N m); Total rotor aerodynamic load (moment in y direction);  the hub coordinate system
AeroDyn['RtAeroMzh']          = False     # (N m); Total rotor aerodynamic load (moment in z direction);  the hub coordinate system
AeroDyn['RtAeroPwr']          = False     # (W); Rotor aerodynamic power; 
AeroDyn['RtArea']             = False     # (m^2); Rotor swept area; 
AeroDyn['RtAeroCp']           = False     # (-); Rotor aerodynamic power coefficient; 
AeroDyn['RtAeroCq']           = False     # (-); Rotor aerodynamic torque coefficient; 
AeroDyn['RtAeroCt']           = False     # (-); Rotor aerodynamic thrust coefficient; 


""" InflowWind """
InflowWind = {}

# Wind Motions
InflowWind['Wind1VelX']       = False     # (m/s); X component of wind at user selected wind point 1; Directed along the xi-axis                         
InflowWind['Wind1VelY']       = False     # (m/s); Y component of wind at user selected wind point 1; Directed along the yi-axis                         
InflowWind['Wind1VelZ']       = False     # (m/s); Z component of wind at user selected wind point 1; Directed along the zi-axis                         
InflowWind['Wind2VelX']       = False     # (m/s); X component of wind at user selected wind point 2; Directed along the xi-axis                         
InflowWind['Wind2VelY']       = False     # (m/s); Y component of wind at user selected wind point 2; Directed along the yi-axis                         
InflowWind['Wind2VelZ']       = False     # (m/s); Z component of wind at user selected wind point 2; Directed along the zi-axis                         
InflowWind['Wind3VelX']       = False     # (m/s); X component of wind at user selected wind point 3; Directed along the xi-axis                         
InflowWind['Wind3VelY']       = False     # (m/s); Y component of wind at user selected wind point 3; Directed along the yi-axis                         
InflowWind['Wind3VelZ']       = False     # (m/s); Z component of wind at user selected wind point 3; Directed along the zi-axis                         
InflowWind['Wind4VelX']       = False     # (m/s); X component of wind at user selected wind point 4; Directed along the xi-axis                         
InflowWind['Wind4VelY']       = False     # (m/s); Y component of wind at user selected wind point 4; Directed along the yi-axis                         
InflowWind['Wind4VelZ']       = False     # (m/s); Z component of wind at user selected wind point 4; Directed along the zi-axis                         
InflowWind['Wind5VelX']       = False     # (m/s); X component of wind at user selected wind point 5; Directed along the xi-axis                         
InflowWind['Wind5VelY']       = False     # (m/s); Y component of wind at user selected wind point 5; Directed along the yi-axis                         
InflowWind['Wind5VelZ']       = False     # (m/s); Z component of wind at user selected wind point 5; Directed along the zi-axis                         
InflowWind['Wind6VelX']       = False     # (m/s); X component of wind at user selected wind point 6; Directed along the xi-axis                         
InflowWind['Wind6VelY']       = False     # (m/s); Y component of wind at user selected wind point 6; Directed along the yi-axis                         
InflowWind['Wind6VelZ']       = False     # (m/s); Z component of wind at user selected wind point 6; Directed along the zi-axis                         
InflowWind['Wind7VelX']       = False     # (m/s); X component of wind at user selected wind point 7; Directed along the xi-axis                         
InflowWind['Wind7VelY']       = False     # (m/s); Y component of wind at user selected wind point 7; Directed along the yi-axis                         
InflowWind['Wind7VelZ']       = False     # (m/s); Z component of wind at user selected wind point 7; Directed along the zi-axis                         
InflowWind['Wind8VelX']       = False     # (m/s); X component of wind at user selected wind point 8; Directed along the xi-axis                         
InflowWind['Wind8VelY']       = False     # (m/s); Y component of wind at user selected wind point 8; Directed along the yi-axis                         
InflowWind['Wind8VelZ']       = False     # (m/s); Z component of wind at user selected wind point 8; Directed along the zi-axis                         
InflowWind['Wind9VelX']       = False     # (m/s); X component of wind at user selected wind point 9; Directed along the xi-axis                         
InflowWind['Wind9VelY']       = False     # (m/s); Y component of wind at user selected wind point 9; Directed along the yi-axis                         
InflowWind['Wind9VelZ']       = False     # (m/s); Z component of wind at user selected wind point 9; Directed along the zi-axis                         

# Wind Sensor Measurements
InflowWind['WindMeas1']       = False     # (m/s); Wind measurement at sensor 1; Defined by sensor
InflowWind['WindMeas2']       = False     # (m/s); Wind measurement at sensor 2; Defined by sensor
InflowWind['WindMeas3']       = False     # (m/s); Wind measurement at sensor 3; Defined by sensor
InflowWind['WindMeas4']       = False     # (m/s); Wind measurement at sensor 4; Defined by sensor
InflowWind['WindMeas5']       = False     # (m/s); Wind measurement at sensor 5; Defined by sensor


""" WAMIT """
WAMIT = {}

# WAMIT Body Forces
WAMIT['Wave1El2']             = False     # (m); 2nd order wave elevation correction; 
WAMIT['Wave2El2']             = False     # (m); 2nd order wave elevation correction; 
WAMIT['Wave3El2']             = False     # (m); 2nd order wave elevation correction; 
WAMIT['Wave4El2']             = False     # (m); 2nd order wave elevation correction; 
WAMIT['Wave5El2']             = False     # (m); 2nd order wave elevation correction; 
WAMIT['Wave6El2']             = False     # (m); 2nd order wave elevation correction; 
WAMIT['Wave7El2']             = False     # (m); 2nd order wave elevation correction; 
WAMIT['Wave8El2']             = False     # (m); 2nd order wave elevation correction; 
WAMIT['Wave9El2']             = False     # (m); 2nd order wave elevation correction; 

# WAMIT second order Body Forces
WAMIT['WavesF2xi']            = False     # (N); ; 
WAMIT['WavesF2yi']            = False     # (N); ; 
WAMIT['WavesF2zi']            = False     # (N); ; 
WAMIT['WavesM2xi']            = False     # (N m); ; 
WAMIT['WavesM2yi']            = False     # (N m); ; 
WAMIT['WavesM2zi']            = False     # (N m); ; 

# WAMIT Body Forces
WAMIT['WavesFxi']             = False     # (N); ; 
WAMIT['WavesFyi']             = False     # (N); ; 
WAMIT['WavesFzi']             = False     # (N); ; 
WAMIT['WavesMxi']             = False     # (N m); ; 
WAMIT['WavesMyi']             = False     # (N m); ; 
WAMIT['WavesMzi']             = False     # (N m); ; 
WAMIT['HdrStcFxi']            = False     # (N); ; 
WAMIT['HdrStcFyi']            = False     # (N); ; 
WAMIT['HdrStcFzi']            = False     # (N); ; 
WAMIT['HdrStcMxi']            = False     # (N m); ; 
WAMIT['HdrStcMyi']            = False     # (N m); ; 
WAMIT['HdrStcMzi']            = False     # (N m); ; 
WAMIT['RdtnFxi']              = False     # (N); ; 
WAMIT['RdtnFyi']              = False     # (N); ; 
WAMIT['RdtnFzi']              = False     # (N); ; 
WAMIT['RdtnMxi']              = False     # (N m); ; 
WAMIT['RdtnMyi']              = False     # (N m); ; 
WAMIT['RdtnMzi']              = False     # (N m); ; 

""" HydroDyn """
HydroDyn = {}

# Forces due to additional preload, stiffness, and damping
HydroDyn['AddFxi']            = False     # (N); ; 
HydroDyn['AddFyi']            = False     # (N); ; 
HydroDyn['AddFzi']            = False     # (N); ; 
HydroDyn['AddMxi']            = False     # (N m); ; 
HydroDyn['AddMyi']            = False     # (N m); ; 
HydroDyn['AddMzi']            = False     # (N m); ; 

# Integrated hydrodynamic loads at the WAMIT reference point
HydroDyn['HydroFxi']          = False     # (N); ; 
HydroDyn['HydroFyi']          = False     # (N); ; 
HydroDyn['HydroFzi']          = False     # (N); ; 
HydroDyn['HydroMxi']          = False     # (N m); ; 
HydroDyn['HydroMyi']          = False     # (N m); ; 
HydroDyn['HydroMzi']          = False     # (N m); ; 

# Wave Motions
HydroDyn['Wave1Elev']         = False     # (m); ; 
HydroDyn['Wave2Elev']         = False     # (m); ; 
HydroDyn['Wave3Elev']         = False     # (m); ; 
HydroDyn['Wave4Elev']         = False     # (m); ; 
HydroDyn['Wave5Elev']         = False     # (m); ; 
HydroDyn['Wave6Elev']         = False     # (m); ; 
HydroDyn['Wave7Elev']         = False     # (m); ; 
HydroDyn['Wave8Elev']         = False     # (m); ; 
HydroDyn['Wave9Elev']         = False     # (m); ; 

# WRP Motions
HydroDyn['WRPSurge']          = False     # (m); ; 
HydroDyn['WRPSway']           = False     # (m); ; 
HydroDyn['WRPHeave']          = False     # (m); ; 
HydroDyn['WRPRoll']           = False     # (rad); ; 
HydroDyn['WRPPitch']          = False     # (rad); ; 
HydroDyn['WRPYaw']            = False     # (rad); ; 
HydroDyn['WRPTVxi']           = False     # (m/s); ; 
HydroDyn['WRPTVyi']           = False     # (m/s); ; 
HydroDyn['WRPTVzi']           = False     # (m/s); ; 
HydroDyn['WRPRVxi']           = False     # (rad/s); ; 
HydroDyn['WRPRVyi']           = False     # (rad/s); ; 
HydroDyn['WRPRVzi']           = False     # (rad/s); ; 
HydroDyn['WRPTAxi']           = False     # (m/s^2); ; 
HydroDyn['WRPTAyi']           = False     # (m/s^2); ; 
HydroDyn['WRPTAzi']           = False     # (m/s^2); ; 
HydroDyn['WRPRAxi']           = False     # (rad/s^2); ; 
HydroDyn['WRPRAyi']           = False     # (rad/s^2); ; 
HydroDyn['WRPRAzi']           = False     # (rad/s^2); ; 


""" Morison """
Morison = {}

# Member-level Wave Kinematics 
Morison['M1N1Axi']            = False     # (m/s^2); fluid acceleration; 
Morison['M1N2Axi']            = False     # (m/s^2); ; 
Morison['M1N3Axi']            = False     # (m/s^2); ; 
Morison['M1N4Axi']            = False     # (m/s^2); ; 
Morison['M1N5Axi']            = False     # (m/s^2); ; 
Morison['M1N6Axi']            = False     # (m/s^2); ; 
Morison['M1N7Axi']            = False     # (m/s^2); ; 
Morison['M1N8Axi']            = False     # (m/s^2); ; 
Morison['M1N9Axi']            = False     # (m/s^2); ; 
Morison['M2N1Axi']            = False     # (m/s^2); ; 
Morison['M2N2Axi']            = False     # (m/s^2); ; 
Morison['M2N3Axi']            = False     # (m/s^2); ; 
Morison['M2N4Axi']            = False     # (m/s^2); ; 
Morison['M2N5Axi']            = False     # (m/s^2); ; 
Morison['M2N6Axi']            = False     # (m/s^2); ; 
Morison['M2N7Axi']            = False     # (m/s^2); ; 
Morison['M2N8Axi']            = False     # (m/s^2); ; 
Morison['M2N9Axi']            = False     # (m/s^2); ; 
Morison['M3N1Axi']            = False     # (m/s^2); ; 
Morison['M3N2Axi']            = False     # (m/s^2); ; 
Morison['M3N3Axi']            = False     # (m/s^2); ; 
Morison['M3N4Axi']            = False     # (m/s^2); ; 
Morison['M3N5Axi']            = False     # (m/s^2); ; 
Morison['M3N6Axi']            = False     # (m/s^2); ; 
Morison['M3N7Axi']            = False     # (m/s^2); ; 
Morison['M3N8Axi']            = False     # (m/s^2); ; 
Morison['M3N9Axi']            = False     # (m/s^2); ; 
Morison['M4N1Axi']            = False     # (m/s^2); ; 
Morison['M4N2Axi']            = False     # (m/s^2); ; 
Morison['M4N3Axi']            = False     # (m/s^2); ; 
Morison['M4N4Axi']            = False     # (m/s^2); ; 
Morison['M4N5Axi']            = False     # (m/s^2); ; 
Morison['M4N6Axi']            = False     # (m/s^2); ; 
Morison['M4N7Axi']            = False     # (m/s^2); ; 
Morison['M4N8Axi']            = False     # (m/s^2); ; 
Morison['M4N9Axi']            = False     # (m/s^2); ; 
Morison['M5N1Axi']            = False     # (m/s^2); ; 
Morison['M5N2Axi']            = False     # (m/s^2); ; 
Morison['M5N3Axi']            = False     # (m/s^2); ; 
Morison['M5N4Axi']            = False     # (m/s^2); ; 
Morison['M5N5Axi']            = False     # (m/s^2); ; 
Morison['M5N6Axi']            = False     # (m/s^2); ; 
Morison['M5N7Axi']            = False     # (m/s^2); ; 
Morison['M5N8Axi']            = False     # (m/s^2); ; 
Morison['M5N9Axi']            = False     # (m/s^2); ; 
Morison['M6N1Axi']            = False     # (m/s^2); ; 
Morison['M6N2Axi']            = False     # (m/s^2); ; 
Morison['M6N3Axi']            = False     # (m/s^2); ; 
Morison['M6N4Axi']            = False     # (m/s^2); ; 
Morison['M6N5Axi']            = False     # (m/s^2); ; 
Morison['M6N6Axi']            = False     # (m/s^2); ; 
Morison['M6N7Axi']            = False     # (m/s^2); ; 
Morison['M6N8Axi']            = False     # (m/s^2); ; 
Morison['M6N9Axi']            = False     # (m/s^2); ; 
Morison['M7N1Axi']            = False     # (m/s^2); ; 
Morison['M7N2Axi']            = False     # (m/s^2); ; 
Morison['M7N3Axi']            = False     # (m/s^2); ; 
Morison['M7N4Axi']            = False     # (m/s^2); ; 
Morison['M7N5Axi']            = False     # (m/s^2); ; 
Morison['M7N6Axi']            = False     # (m/s^2); ; 
Morison['M7N7Axi']            = False     # (m/s^2); ; 
Morison['M7N8Axi']            = False     # (m/s^2); ; 
Morison['M7N9Axi']            = False     # (m/s^2); ; 
Morison['M8N1Axi']            = False     # (m/s^2); ; 
Morison['M8N2Axi']            = False     # (m/s^2); ; 
Morison['M8N3Axi']            = False     # (m/s^2); ; 
Morison['M8N4Axi']            = False     # (m/s^2); ; 
Morison['M8N5Axi']            = False     # (m/s^2); ; 
Morison['M8N6Axi']            = False     # (m/s^2); ; 
Morison['M8N7Axi']            = False     # (m/s^2); ; 
Morison['M8N8Axi']            = False     # (m/s^2); ; 
Morison['M8N9Axi']            = False     # (m/s^2); ; 
Morison['M9N1Axi']            = False     # (m/s^2); ; 
Morison['M9N2Axi']            = False     # (m/s^2); ; 
Morison['M9N3Axi']            = False     # (m/s^2); ; 
Morison['M9N4Axi']            = False     # (m/s^2); ; 
Morison['M9N5Axi']            = False     # (m/s^2); ; 
Morison['M9N6Axi']            = False     # (m/s^2); ; 
Morison['M9N7Axi']            = False     # (m/s^2); ; 
Morison['M9N8Axi']            = False     # (m/s^2); ; 
Morison['M9N9Axi']            = False     # (m/s^2); ; 
Morison['M1N1Ayi']            = False     # (m/s^2); ; 
Morison['M1N2Ayi']            = False     # (m/s^2); ; 
Morison['M1N3Ayi']            = False     # (m/s^2); ; 
Morison['M1N4Ayi']            = False     # (m/s^2); ; 
Morison['M1N5Ayi']            = False     # (m/s^2); ; 
Morison['M1N6Ayi']            = False     # (m/s^2); ; 
Morison['M1N7Ayi']            = False     # (m/s^2); ; 
Morison['M1N8Ayi']            = False     # (m/s^2); ; 
Morison['M1N9Ayi']            = False     # (m/s^2); ; 
Morison['M2N1Ayi']            = False     # (m/s^2); ; 
Morison['M2N2Ayi']            = False     # (m/s^2); ; 
Morison['M2N3Ayi']            = False     # (m/s^2); ; 
Morison['M2N4Ayi']            = False     # (m/s^2); ; 
Morison['M2N5Ayi']            = False     # (m/s^2); ; 
Morison['M2N6Ayi']            = False     # (m/s^2); ; 
Morison['M2N7Ayi']            = False     # (m/s^2); ; 
Morison['M2N8Ayi']            = False     # (m/s^2); ; 
Morison['M2N9Ayi']            = False     # (m/s^2); ; 
Morison['M3N1Ayi']            = False     # (m/s^2); ; 
Morison['M3N2Ayi']            = False     # (m/s^2); ; 
Morison['M3N3Ayi']            = False     # (m/s^2); ; 
Morison['M3N4Ayi']            = False     # (m/s^2); ; 
Morison['M3N5Ayi']            = False     # (m/s^2); ; 
Morison['M3N6Ayi']            = False     # (m/s^2); ; 
Morison['M3N7Ayi']            = False     # (m/s^2); ; 
Morison['M3N8Ayi']            = False     # (m/s^2); ; 
Morison['M3N9Ayi']            = False     # (m/s^2); ; 
Morison['M4N1Ayi']            = False     # (m/s^2); ; 
Morison['M4N2Ayi']            = False     # (m/s^2); ; 
Morison['M4N3Ayi']            = False     # (m/s^2); ; 
Morison['M4N4Ayi']            = False     # (m/s^2); ; 
Morison['M4N5Ayi']            = False     # (m/s^2); ; 
Morison['M4N6Ayi']            = False     # (m/s^2); ; 
Morison['M4N7Ayi']            = False     # (m/s^2); ; 
Morison['M4N8Ayi']            = False     # (m/s^2); ; 
Morison['M4N9Ayi']            = False     # (m/s^2); ; 
Morison['M5N1Ayi']            = False     # (m/s^2); ; 
Morison['M5N2Ayi']            = False     # (m/s^2); ; 
Morison['M5N3Ayi']            = False     # (m/s^2); ; 
Morison['M5N4Ayi']            = False     # (m/s^2); ; 
Morison['M5N5Ayi']            = False     # (m/s^2); ; 
Morison['M5N6Ayi']            = False     # (m/s^2); ; 
Morison['M5N7Ayi']            = False     # (m/s^2); ; 
Morison['M5N8Ayi']            = False     # (m/s^2); ; 
Morison['M5N9Ayi']            = False     # (m/s^2); ; 
Morison['M6N1Ayi']            = False     # (m/s^2); ; 
Morison['M6N2Ayi']            = False     # (m/s^2); ; 
Morison['M6N3Ayi']            = False     # (m/s^2); ; 
Morison['M6N4Ayi']            = False     # (m/s^2); ; 
Morison['M6N5Ayi']            = False     # (m/s^2); ; 
Morison['M6N6Ayi']            = False     # (m/s^2); ; 
Morison['M6N7Ayi']            = False     # (m/s^2); ; 
Morison['M6N8Ayi']            = False     # (m/s^2); ; 
Morison['M6N9Ayi']            = False     # (m/s^2); ; 
Morison['M7N1Ayi']            = False     # (m/s^2); ; 
Morison['M7N2Ayi']            = False     # (m/s^2); ; 
Morison['M7N3Ayi']            = False     # (m/s^2); ; 
Morison['M7N4Ayi']            = False     # (m/s^2); ; 
Morison['M7N5Ayi']            = False     # (m/s^2); ; 
Morison['M7N6Ayi']            = False     # (m/s^2); ; 
Morison['M7N7Ayi']            = False     # (m/s^2); ; 
Morison['M7N8Ayi']            = False     # (m/s^2); ; 
Morison['M7N9Ayi']            = False     # (m/s^2); ; 
Morison['M8N1Ayi']            = False     # (m/s^2); ; 
Morison['M8N2Ayi']            = False     # (m/s^2); ; 
Morison['M8N3Ayi']            = False     # (m/s^2); ; 
Morison['M8N4Ayi']            = False     # (m/s^2); ; 
Morison['M8N5Ayi']            = False     # (m/s^2); ; 
Morison['M8N6Ayi']            = False     # (m/s^2); ; 
Morison['M8N7Ayi']            = False     # (m/s^2); ; 
Morison['M8N8Ayi']            = False     # (m/s^2); ; 
Morison['M8N9Ayi']            = False     # (m/s^2); ; 
Morison['M9N1Ayi']            = False     # (m/s^2); ; 
Morison['M9N2Ayi']            = False     # (m/s^2); ; 
Morison['M9N3Ayi']            = False     # (m/s^2); ; 
Morison['M9N4Ayi']            = False     # (m/s^2); ; 
Morison['M9N5Ayi']            = False     # (m/s^2); ; 
Morison['M9N6Ayi']            = False     # (m/s^2); ; 
Morison['M9N7Ayi']            = False     # (m/s^2); ; 
Morison['M9N8Ayi']            = False     # (m/s^2); ; 
Morison['M9N9Ayi']            = False     # (m/s^2); ; 
Morison['M1N1Azi']            = False     # (m/s^2); ; 
Morison['M1N2Azi']            = False     # (m/s^2); ; 
Morison['M1N3Azi']            = False     # (m/s^2); ; 
Morison['M1N4Azi']            = False     # (m/s^2); ; 
Morison['M1N5Azi']            = False     # (m/s^2); ; 
Morison['M1N6Azi']            = False     # (m/s^2); ; 
Morison['M1N7Azi']            = False     # (m/s^2); ; 
Morison['M1N8Azi']            = False     # (m/s^2); ; 
Morison['M1N9Azi']            = False     # (m/s^2); ; 
Morison['M2N1Azi']            = False     # (m/s^2); ; 
Morison['M2N2Azi']            = False     # (m/s^2); ; 
Morison['M2N3Azi']            = False     # (m/s^2); ; 
Morison['M2N4Azi']            = False     # (m/s^2); ; 
Morison['M2N5Azi']            = False     # (m/s^2); ; 
Morison['M2N6Azi']            = False     # (m/s^2); ; 
Morison['M2N7Azi']            = False     # (m/s^2); ; 
Morison['M2N8Azi']            = False     # (m/s^2); ; 
Morison['M2N9Azi']            = False     # (m/s^2); ; 
Morison['M3N1Azi']            = False     # (m/s^2); ; 
Morison['M3N2Azi']            = False     # (m/s^2); ; 
Morison['M3N3Azi']            = False     # (m/s^2); ; 
Morison['M3N4Azi']            = False     # (m/s^2); ; 
Morison['M3N5Azi']            = False     # (m/s^2); ; 
Morison['M3N6Azi']            = False     # (m/s^2); ; 
Morison['M3N7Azi']            = False     # (m/s^2); ; 
Morison['M3N8Azi']            = False     # (m/s^2); ; 
Morison['M3N9Azi']            = False     # (m/s^2); ; 
Morison['M4N1Azi']            = False     # (m/s^2); ; 
Morison['M4N2Azi']            = False     # (m/s^2); ; 
Morison['M4N3Azi']            = False     # (m/s^2); ; 
Morison['M4N4Azi']            = False     # (m/s^2); ; 
Morison['M4N5Azi']            = False     # (m/s^2); ; 
Morison['M4N6Azi']            = False     # (m/s^2); ; 
Morison['M4N7Azi']            = False     # (m/s^2); ; 
Morison['M4N8Azi']            = False     # (m/s^2); ; 
Morison['M4N9Azi']            = False     # (m/s^2); ; 
Morison['M5N1Azi']            = False     # (m/s^2); ; 
Morison['M5N2Azi']            = False     # (m/s^2); ; 
Morison['M5N3Azi']            = False     # (m/s^2); ; 
Morison['M5N4Azi']            = False     # (m/s^2); ; 
Morison['M5N5Azi']            = False     # (m/s^2); ; 
Morison['M5N6Azi']            = False     # (m/s^2); ; 
Morison['M5N7Azi']            = False     # (m/s^2); ; 
Morison['M5N8Azi']            = False     # (m/s^2); ; 
Morison['M5N9Azi']            = False     # (m/s^2); ; 
Morison['M6N1Azi']            = False     # (m/s^2); ; 
Morison['M6N2Azi']            = False     # (m/s^2); ; 
Morison['M6N3Azi']            = False     # (m/s^2); ; 
Morison['M6N4Azi']            = False     # (m/s^2); ; 
Morison['M6N5Azi']            = False     # (m/s^2); ; 
Morison['M6N6Azi']            = False     # (m/s^2); ; 
Morison['M6N7Azi']            = False     # (m/s^2); ; 
Morison['M6N8Azi']            = False     # (m/s^2); ; 
Morison['M6N9Azi']            = False     # (m/s^2); ; 
Morison['M7N1Azi']            = False     # (m/s^2); ; 
Morison['M7N2Azi']            = False     # (m/s^2); ; 
Morison['M7N3Azi']            = False     # (m/s^2); ; 
Morison['M7N4Azi']            = False     # (m/s^2); ; 
Morison['M7N5Azi']            = False     # (m/s^2); ; 
Morison['M7N6Azi']            = False     # (m/s^2); ; 
Morison['M7N7Azi']            = False     # (m/s^2); ; 
Morison['M7N8Azi']            = False     # (m/s^2); ; 
Morison['M7N9Azi']            = False     # (m/s^2); ; 
Morison['M8N1Azi']            = False     # (m/s^2); ; 
Morison['M8N2Azi']            = False     # (m/s^2); ; 
Morison['M8N3Azi']            = False     # (m/s^2); ; 
Morison['M8N4Azi']            = False     # (m/s^2); ; 
Morison['M8N5Azi']            = False     # (m/s^2); ; 
Morison['M8N6Azi']            = False     # (m/s^2); ; 
Morison['M8N7Azi']            = False     # (m/s^2); ; 
Morison['M8N8Azi']            = False     # (m/s^2); ; 
Morison['M8N9Azi']            = False     # (m/s^2); ; 
Morison['M9N1Azi']            = False     # (m/s^2); ; 
Morison['M9N2Azi']            = False     # (m/s^2); ; 
Morison['M9N3Azi']            = False     # (m/s^2); ; 
Morison['M9N4Azi']            = False     # (m/s^2); ; 
Morison['M9N5Azi']            = False     # (m/s^2); ; 
Morison['M9N6Azi']            = False     # (m/s^2); ; 
Morison['M9N7Azi']            = False     # (m/s^2); ; 
Morison['M9N8Azi']            = False     # (m/s^2); ; 
Morison['M9N9Azi']            = False     # (m/s^2); ; 
Morison['M1N1Vxi']            = False     # (m/s); fluid velocity; 
Morison['M1N2Vxi']            = False     # (m/s); ; 
Morison['M1N3Vxi']            = False     # (m/s); ; 
Morison['M1N4Vxi']            = False     # (m/s); ; 
Morison['M1N5Vxi']            = False     # (m/s); ; 
Morison['M1N6Vxi']            = False     # (m/s); ; 
Morison['M1N7Vxi']            = False     # (m/s); ; 
Morison['M1N8Vxi']            = False     # (m/s); ; 
Morison['M1N9Vxi']            = False     # (m/s); ; 
Morison['M2N1Vxi']            = False     # (m/s); ; 
Morison['M2N2Vxi']            = False     # (m/s); ; 
Morison['M2N3Vxi']            = False     # (m/s); ; 
Morison['M2N4Vxi']            = False     # (m/s); ; 
Morison['M2N5Vxi']            = False     # (m/s); ; 
Morison['M2N6Vxi']            = False     # (m/s); ; 
Morison['M2N7Vxi']            = False     # (m/s); ; 
Morison['M2N8Vxi']            = False     # (m/s); ; 
Morison['M2N9Vxi']            = False     # (m/s); ; 
Morison['M3N1Vxi']            = False     # (m/s); ; 
Morison['M3N2Vxi']            = False     # (m/s); ; 
Morison['M3N3Vxi']            = False     # (m/s); ; 
Morison['M3N4Vxi']            = False     # (m/s); ; 
Morison['M3N5Vxi']            = False     # (m/s); ; 
Morison['M3N6Vxi']            = False     # (m/s); ; 
Morison['M3N7Vxi']            = False     # (m/s); ; 
Morison['M3N8Vxi']            = False     # (m/s); ; 
Morison['M3N9Vxi']            = False     # (m/s); ; 
Morison['M4N1Vxi']            = False     # (m/s); ; 
Morison['M4N2Vxi']            = False     # (m/s); ; 
Morison['M4N3Vxi']            = False     # (m/s); ; 
Morison['M4N4Vxi']            = False     # (m/s); ; 
Morison['M4N5Vxi']            = False     # (m/s); ; 
Morison['M4N6Vxi']            = False     # (m/s); ; 
Morison['M4N7Vxi']            = False     # (m/s); ; 
Morison['M4N8Vxi']            = False     # (m/s); ; 
Morison['M4N9Vxi']            = False     # (m/s); ; 
Morison['M5N1Vxi']            = False     # (m/s); ; 
Morison['M5N2Vxi']            = False     # (m/s); ; 
Morison['M5N3Vxi']            = False     # (m/s); ; 
Morison['M5N4Vxi']            = False     # (m/s); ; 
Morison['M5N5Vxi']            = False     # (m/s); ; 
Morison['M5N6Vxi']            = False     # (m/s); ; 
Morison['M5N7Vxi']            = False     # (m/s); ; 
Morison['M5N8Vxi']            = False     # (m/s); ; 
Morison['M5N9Vxi']            = False     # (m/s); ; 
Morison['M6N1Vxi']            = False     # (m/s); ; 
Morison['M6N2Vxi']            = False     # (m/s); ; 
Morison['M6N3Vxi']            = False     # (m/s); ; 
Morison['M6N4Vxi']            = False     # (m/s); ; 
Morison['M6N5Vxi']            = False     # (m/s); ; 
Morison['M6N6Vxi']            = False     # (m/s); ; 
Morison['M6N7Vxi']            = False     # (m/s); ; 
Morison['M6N8Vxi']            = False     # (m/s); ; 
Morison['M6N9Vxi']            = False     # (m/s); ; 
Morison['M7N1Vxi']            = False     # (m/s); ; 
Morison['M7N2Vxi']            = False     # (m/s); ; 
Morison['M7N3Vxi']            = False     # (m/s); ; 
Morison['M7N4Vxi']            = False     # (m/s); ; 
Morison['M7N5Vxi']            = False     # (m/s); ; 
Morison['M7N6Vxi']            = False     # (m/s); ; 
Morison['M7N7Vxi']            = False     # (m/s); ; 
Morison['M7N8Vxi']            = False     # (m/s); ; 
Morison['M7N9Vxi']            = False     # (m/s); ; 
Morison['M8N1Vxi']            = False     # (m/s); ; 
Morison['M8N2Vxi']            = False     # (m/s); ; 
Morison['M8N3Vxi']            = False     # (m/s); ; 
Morison['M8N4Vxi']            = False     # (m/s); ; 
Morison['M8N5Vxi']            = False     # (m/s); ; 
Morison['M8N6Vxi']            = False     # (m/s); ; 
Morison['M8N7Vxi']            = False     # (m/s); ; 
Morison['M8N8Vxi']            = False     # (m/s); ; 
Morison['M8N9Vxi']            = False     # (m/s); ; 
Morison['M9N1Vxi']            = False     # (m/s); ; 
Morison['M9N2Vxi']            = False     # (m/s); ; 
Morison['M9N3Vxi']            = False     # (m/s); ; 
Morison['M9N4Vxi']            = False     # (m/s); ; 
Morison['M9N5Vxi']            = False     # (m/s); ; 
Morison['M9N6Vxi']            = False     # (m/s); ; 
Morison['M9N7Vxi']            = False     # (m/s); ; 
Morison['M9N8Vxi']            = False     # (m/s); ; 
Morison['M9N9Vxi']            = False     # (m/s); ; 
Morison['M1N1Vyi']            = False     # (m/s); ; 
Morison['M1N2Vyi']            = False     # (m/s); ; 
Morison['M1N3Vyi']            = False     # (m/s); ; 
Morison['M1N4Vyi']            = False     # (m/s); ; 
Morison['M1N5Vyi']            = False     # (m/s); ; 
Morison['M1N6Vyi']            = False     # (m/s); ; 
Morison['M1N7Vyi']            = False     # (m/s); ; 
Morison['M1N8Vyi']            = False     # (m/s); ; 
Morison['M1N9Vyi']            = False     # (m/s); ; 
Morison['M2N1Vyi']            = False     # (m/s); ; 
Morison['M2N2Vyi']            = False     # (m/s); ; 
Morison['M2N3Vyi']            = False     # (m/s); ; 
Morison['M2N4Vyi']            = False     # (m/s); ; 
Morison['M2N5Vyi']            = False     # (m/s); ; 
Morison['M2N6Vyi']            = False     # (m/s); ; 
Morison['M2N7Vyi']            = False     # (m/s); ; 
Morison['M2N8Vyi']            = False     # (m/s); ; 
Morison['M2N9Vyi']            = False     # (m/s); ; 
Morison['M3N1Vyi']            = False     # (m/s); ; 
Morison['M3N2Vyi']            = False     # (m/s); ; 
Morison['M3N3Vyi']            = False     # (m/s); ; 
Morison['M3N4Vyi']            = False     # (m/s); ; 
Morison['M3N5Vyi']            = False     # (m/s); ; 
Morison['M3N6Vyi']            = False     # (m/s); ; 
Morison['M3N7Vyi']            = False     # (m/s); ; 
Morison['M3N8Vyi']            = False     # (m/s); ; 
Morison['M3N9Vyi']            = False     # (m/s); ; 
Morison['M4N1Vyi']            = False     # (m/s); ; 
Morison['M4N2Vyi']            = False     # (m/s); ; 
Morison['M4N3Vyi']            = False     # (m/s); ; 
Morison['M4N4Vyi']            = False     # (m/s); ; 
Morison['M4N5Vyi']            = False     # (m/s); ; 
Morison['M4N6Vyi']            = False     # (m/s); ; 
Morison['M4N7Vyi']            = False     # (m/s); ; 
Morison['M4N8Vyi']            = False     # (m/s); ; 
Morison['M4N9Vyi']            = False     # (m/s); ; 
Morison['M5N1Vyi']            = False     # (m/s); ; 
Morison['M5N2Vyi']            = False     # (m/s); ; 
Morison['M5N3Vyi']            = False     # (m/s); ; 
Morison['M5N4Vyi']            = False     # (m/s); ; 
Morison['M5N5Vyi']            = False     # (m/s); ; 
Morison['M5N6Vyi']            = False     # (m/s); ; 
Morison['M5N7Vyi']            = False     # (m/s); ; 
Morison['M5N8Vyi']            = False     # (m/s); ; 
Morison['M5N9Vyi']            = False     # (m/s); ; 
Morison['M6N1Vyi']            = False     # (m/s); ; 
Morison['M6N2Vyi']            = False     # (m/s); ; 
Morison['M6N3Vyi']            = False     # (m/s); ; 
Morison['M6N4Vyi']            = False     # (m/s); ; 
Morison['M6N5Vyi']            = False     # (m/s); ; 
Morison['M6N6Vyi']            = False     # (m/s); ; 
Morison['M6N7Vyi']            = False     # (m/s); ; 
Morison['M6N8Vyi']            = False     # (m/s); ; 
Morison['M6N9Vyi']            = False     # (m/s); ; 
Morison['M7N1Vyi']            = False     # (m/s); ; 
Morison['M7N2Vyi']            = False     # (m/s); ; 
Morison['M7N3Vyi']            = False     # (m/s); ; 
Morison['M7N4Vyi']            = False     # (m/s); ; 
Morison['M7N5Vyi']            = False     # (m/s); ; 
Morison['M7N6Vyi']            = False     # (m/s); ; 
Morison['M7N7Vyi']            = False     # (m/s); ; 
Morison['M7N8Vyi']            = False     # (m/s); ; 
Morison['M7N9Vyi']            = False     # (m/s); ; 
Morison['M8N1Vyi']            = False     # (m/s); ; 
Morison['M8N2Vyi']            = False     # (m/s); ; 
Morison['M8N3Vyi']            = False     # (m/s); ; 
Morison['M8N4Vyi']            = False     # (m/s); ; 
Morison['M8N5Vyi']            = False     # (m/s); ; 
Morison['M8N6Vyi']            = False     # (m/s); ; 
Morison['M8N7Vyi']            = False     # (m/s); ; 
Morison['M8N8Vyi']            = False     # (m/s); ; 
Morison['M8N9Vyi']            = False     # (m/s); ; 
Morison['M9N1Vyi']            = False     # (m/s); ; 
Morison['M9N2Vyi']            = False     # (m/s); ; 
Morison['M9N3Vyi']            = False     # (m/s); ; 
Morison['M9N4Vyi']            = False     # (m/s); ; 
Morison['M9N5Vyi']            = False     # (m/s); ; 
Morison['M9N6Vyi']            = False     # (m/s); ; 
Morison['M9N7Vyi']            = False     # (m/s); ; 
Morison['M9N8Vyi']            = False     # (m/s); ; 
Morison['M9N9Vyi']            = False     # (m/s); ; 
Morison['M1N1Vzi']            = False     # (m/s); ; 
Morison['M1N2Vzi']            = False     # (m/s); ; 
Morison['M1N3Vzi']            = False     # (m/s); ; 
Morison['M1N4Vzi']            = False     # (m/s); ; 
Morison['M1N5Vzi']            = False     # (m/s); ; 
Morison['M1N6Vzi']            = False     # (m/s); ; 
Morison['M1N7Vzi']            = False     # (m/s); ; 
Morison['M1N8Vzi']            = False     # (m/s); ; 
Morison['M1N9Vzi']            = False     # (m/s); ; 
Morison['M2N1Vzi']            = False     # (m/s); ; 
Morison['M2N2Vzi']            = False     # (m/s); ; 
Morison['M2N3Vzi']            = False     # (m/s); ; 
Morison['M2N4Vzi']            = False     # (m/s); ; 
Morison['M2N5Vzi']            = False     # (m/s); ; 
Morison['M2N6Vzi']            = False     # (m/s); ; 
Morison['M2N7Vzi']            = False     # (m/s); ; 
Morison['M2N8Vzi']            = False     # (m/s); ; 
Morison['M2N9Vzi']            = False     # (m/s); ; 
Morison['M3N1Vzi']            = False     # (m/s); ; 
Morison['M3N2Vzi']            = False     # (m/s); ; 
Morison['M3N3Vzi']            = False     # (m/s); ; 
Morison['M3N4Vzi']            = False     # (m/s); ; 
Morison['M3N5Vzi']            = False     # (m/s); ; 
Morison['M3N6Vzi']            = False     # (m/s); ; 
Morison['M3N7Vzi']            = False     # (m/s); ; 
Morison['M3N8Vzi']            = False     # (m/s); ; 
Morison['M3N9Vzi']            = False     # (m/s); ; 
Morison['M4N1Vzi']            = False     # (m/s); ; 
Morison['M4N2Vzi']            = False     # (m/s); ; 
Morison['M4N3Vzi']            = False     # (m/s); ; 
Morison['M4N4Vzi']            = False     # (m/s); ; 
Morison['M4N5Vzi']            = False     # (m/s); ; 
Morison['M4N6Vzi']            = False     # (m/s); ; 
Morison['M4N7Vzi']            = False     # (m/s); ; 
Morison['M4N8Vzi']            = False     # (m/s); ; 
Morison['M4N9Vzi']            = False     # (m/s); ; 
Morison['M5N1Vzi']            = False     # (m/s); ; 
Morison['M5N2Vzi']            = False     # (m/s); ; 
Morison['M5N3Vzi']            = False     # (m/s); ; 
Morison['M5N4Vzi']            = False     # (m/s); ; 
Morison['M5N5Vzi']            = False     # (m/s); ; 
Morison['M5N6Vzi']            = False     # (m/s); ; 
Morison['M5N7Vzi']            = False     # (m/s); ; 
Morison['M5N8Vzi']            = False     # (m/s); ; 
Morison['M5N9Vzi']            = False     # (m/s); ; 
Morison['M6N1Vzi']            = False     # (m/s); ; 
Morison['M6N2Vzi']            = False     # (m/s); ; 
Morison['M6N3Vzi']            = False     # (m/s); ; 
Morison['M6N4Vzi']            = False     # (m/s); ; 
Morison['M6N5Vzi']            = False     # (m/s); ; 
Morison['M6N6Vzi']            = False     # (m/s); ; 
Morison['M6N7Vzi']            = False     # (m/s); ; 
Morison['M6N8Vzi']            = False     # (m/s); ; 
Morison['M6N9Vzi']            = False     # (m/s); ; 
Morison['M7N1Vzi']            = False     # (m/s); ; 
Morison['M7N2Vzi']            = False     # (m/s); ; 
Morison['M7N3Vzi']            = False     # (m/s); ; 
Morison['M7N4Vzi']            = False     # (m/s); ; 
Morison['M7N5Vzi']            = False     # (m/s); ; 
Morison['M7N6Vzi']            = False     # (m/s); ; 
Morison['M7N7Vzi']            = False     # (m/s); ; 
Morison['M7N8Vzi']            = False     # (m/s); ; 
Morison['M7N9Vzi']            = False     # (m/s); ; 
Morison['M8N1Vzi']            = False     # (m/s); ; 
Morison['M8N2Vzi']            = False     # (m/s); ; 
Morison['M8N3Vzi']            = False     # (m/s); ; 
Morison['M8N4Vzi']            = False     # (m/s); ; 
Morison['M8N5Vzi']            = False     # (m/s); ; 
Morison['M8N6Vzi']            = False     # (m/s); ; 
Morison['M8N7Vzi']            = False     # (m/s); ; 
Morison['M8N8Vzi']            = False     # (m/s); ; 
Morison['M8N9Vzi']            = False     # (m/s); ; 
Morison['M9N1Vzi']            = False     # (m/s); ; 
Morison['M9N2Vzi']            = False     # (m/s); ; 
Morison['M9N3Vzi']            = False     # (m/s); ; 
Morison['M9N4Vzi']            = False     # (m/s); ; 
Morison['M9N5Vzi']            = False     # (m/s); ; 
Morison['M9N6Vzi']            = False     # (m/s); ; 
Morison['M9N7Vzi']            = False     # (m/s); ; 
Morison['M9N8Vzi']            = False     # (m/s); ; 
Morison['M9N9Vzi']            = False     # (m/s); ; 
Morison['M1N1DynP']           = False     # (Pa); fluid dynamic pressure; 
Morison['M1N2DynP']           = False     # (Pa); ; 
Morison['M1N3DynP']           = False     # (Pa); ; 
Morison['M1N4DynP']           = False     # (Pa); ; 
Morison['M1N5DynP']           = False     # (Pa); ; 
Morison['M1N6DynP']           = False     # (Pa); ; 
Morison['M1N7DynP']           = False     # (Pa); ; 
Morison['M1N8DynP']           = False     # (Pa); ; 
Morison['M1N9DynP']           = False     # (Pa); ; 
Morison['M2N1DynP']           = False     # (Pa); ; 
Morison['M2N2DynP']           = False     # (Pa); ; 
Morison['M2N3DynP']           = False     # (Pa); ; 
Morison['M2N4DynP']           = False     # (Pa); ; 
Morison['M2N5DynP']           = False     # (Pa); ; 
Morison['M2N6DynP']           = False     # (Pa); ; 
Morison['M2N7DynP']           = False     # (Pa); ; 
Morison['M2N8DynP']           = False     # (Pa); ; 
Morison['M2N9DynP']           = False     # (Pa); ; 
Morison['M3N1DynP']           = False     # (Pa); ; 
Morison['M3N2DynP']           = False     # (Pa); ; 
Morison['M3N3DynP']           = False     # (Pa); ; 
Morison['M3N4DynP']           = False     # (Pa); ; 
Morison['M3N5DynP']           = False     # (Pa); ; 
Morison['M3N6DynP']           = False     # (Pa); ; 
Morison['M3N7DynP']           = False     # (Pa); ; 
Morison['M3N8DynP']           = False     # (Pa); ; 
Morison['M3N9DynP']           = False     # (Pa); ; 
Morison['M4N1DynP']           = False     # (Pa); ; 
Morison['M4N2DynP']           = False     # (Pa); ; 
Morison['M4N3DynP']           = False     # (Pa); ; 
Morison['M4N4DynP']           = False     # (Pa); ; 
Morison['M4N5DynP']           = False     # (Pa); ; 
Morison['M4N6DynP']           = False     # (Pa); ; 
Morison['M4N7DynP']           = False     # (Pa); ; 
Morison['M4N8DynP']           = False     # (Pa); ; 
Morison['M4N9DynP']           = False     # (Pa); ; 
Morison['M5N1DynP']           = False     # (Pa); ; 
Morison['M5N2DynP']           = False     # (Pa); ; 
Morison['M5N3DynP']           = False     # (Pa); ; 
Morison['M5N4DynP']           = False     # (Pa); ; 
Morison['M5N5DynP']           = False     # (Pa); ; 
Morison['M5N6DynP']           = False     # (Pa); ; 
Morison['M5N7DynP']           = False     # (Pa); ; 
Morison['M5N8DynP']           = False     # (Pa); ; 
Morison['M5N9DynP']           = False     # (Pa); ; 
Morison['M6N1DynP']           = False     # (Pa); ; 
Morison['M6N2DynP']           = False     # (Pa); ; 
Morison['M6N3DynP']           = False     # (Pa); ; 
Morison['M6N4DynP']           = False     # (Pa); ; 
Morison['M6N5DynP']           = False     # (Pa); ; 
Morison['M6N6DynP']           = False     # (Pa); ; 
Morison['M6N7DynP']           = False     # (Pa); ; 
Morison['M6N8DynP']           = False     # (Pa); ; 
Morison['M6N9DynP']           = False     # (Pa); ; 
Morison['M7N1DynP']           = False     # (Pa); ; 
Morison['M7N2DynP']           = False     # (Pa); ; 
Morison['M7N3DynP']           = False     # (Pa); ; 
Morison['M7N4DynP']           = False     # (Pa); ; 
Morison['M7N5DynP']           = False     # (Pa); ; 
Morison['M7N6DynP']           = False     # (Pa); ; 
Morison['M7N7DynP']           = False     # (Pa); ; 
Morison['M7N8DynP']           = False     # (Pa); ; 
Morison['M7N9DynP']           = False     # (Pa); ; 
Morison['M8N1DynP']           = False     # (Pa); ; 
Morison['M8N2DynP']           = False     # (Pa); ; 
Morison['M8N3DynP']           = False     # (Pa); ; 
Morison['M8N4DynP']           = False     # (Pa); ; 
Morison['M8N5DynP']           = False     # (Pa); ; 
Morison['M8N6DynP']           = False     # (Pa); ; 
Morison['M8N7DynP']           = False     # (Pa); ; 
Morison['M8N8DynP']           = False     # (Pa); ; 
Morison['M8N9DynP']           = False     # (Pa); ; 
Morison['M9N1DynP']           = False     # (Pa); ; 
Morison['M9N2DynP']           = False     # (Pa); ; 
Morison['M9N3DynP']           = False     # (Pa); ; 
Morison['M9N4DynP']           = False     # (Pa); ; 
Morison['M9N5DynP']           = False     # (Pa); ; 
Morison['M9N6DynP']           = False     # (Pa); ; 
Morison['M9N7DynP']           = False     # (Pa); ; 
Morison['M9N8DynP']           = False     # (Pa); ; 
Morison['M9N9DynP']           = False     # (Pa); ; 
Morison['M1N1STVxi']          = False     # (m/s); structure translational velocity; 
Morison['M1N2STVxi']          = False     # (m/s); ; 
Morison['M1N3STVxi']          = False     # (m/s); ; 
Morison['M1N4STVxi']          = False     # (m/s); ; 
Morison['M1N5STVxi']          = False     # (m/s); ; 
Morison['M1N6STVxi']          = False     # (m/s); ; 
Morison['M1N7STVxi']          = False     # (m/s); ; 
Morison['M1N8STVxi']          = False     # (m/s); ; 
Morison['M1N9STVxi']          = False     # (m/s); ; 
Morison['M2N1STVxi']          = False     # (m/s); ; 
Morison['M2N2STVxi']          = False     # (m/s); ; 
Morison['M2N3STVxi']          = False     # (m/s); ; 
Morison['M2N4STVxi']          = False     # (m/s); ; 
Morison['M2N5STVxi']          = False     # (m/s); ; 
Morison['M2N6STVxi']          = False     # (m/s); ; 
Morison['M2N7STVxi']          = False     # (m/s); ; 
Morison['M2N8STVxi']          = False     # (m/s); ; 
Morison['M2N9STVxi']          = False     # (m/s); ; 
Morison['M3N1STVxi']          = False     # (m/s); ; 
Morison['M3N2STVxi']          = False     # (m/s); ; 
Morison['M3N3STVxi']          = False     # (m/s); ; 
Morison['M3N4STVxi']          = False     # (m/s); ; 
Morison['M3N5STVxi']          = False     # (m/s); ; 
Morison['M3N6STVxi']          = False     # (m/s); ; 
Morison['M3N7STVxi']          = False     # (m/s); ; 
Morison['M3N8STVxi']          = False     # (m/s); ; 
Morison['M3N9STVxi']          = False     # (m/s); ; 
Morison['M4N1STVxi']          = False     # (m/s); ; 
Morison['M4N2STVxi']          = False     # (m/s); ; 
Morison['M4N3STVxi']          = False     # (m/s); ; 
Morison['M4N4STVxi']          = False     # (m/s); ; 
Morison['M4N5STVxi']          = False     # (m/s); ; 
Morison['M4N6STVxi']          = False     # (m/s); ; 
Morison['M4N7STVxi']          = False     # (m/s); ; 
Morison['M4N8STVxi']          = False     # (m/s); ; 
Morison['M4N9STVxi']          = False     # (m/s); ; 
Morison['M5N1STVxi']          = False     # (m/s); ; 
Morison['M5N2STVxi']          = False     # (m/s); ; 
Morison['M5N3STVxi']          = False     # (m/s); ; 
Morison['M5N4STVxi']          = False     # (m/s); ; 
Morison['M5N5STVxi']          = False     # (m/s); ; 
Morison['M5N6STVxi']          = False     # (m/s); ; 
Morison['M5N7STVxi']          = False     # (m/s); ; 
Morison['M5N8STVxi']          = False     # (m/s); ; 
Morison['M5N9STVxi']          = False     # (m/s); ; 
Morison['M6N1STVxi']          = False     # (m/s); ; 
Morison['M6N2STVxi']          = False     # (m/s); ; 
Morison['M6N3STVxi']          = False     # (m/s); ; 
Morison['M6N4STVxi']          = False     # (m/s); ; 
Morison['M6N5STVxi']          = False     # (m/s); ; 
Morison['M6N6STVxi']          = False     # (m/s); ; 
Morison['M6N7STVxi']          = False     # (m/s); ; 
Morison['M6N8STVxi']          = False     # (m/s); ; 
Morison['M6N9STVxi']          = False     # (m/s); ; 
Morison['M7N1STVxi']          = False     # (m/s); ; 
Morison['M7N2STVxi']          = False     # (m/s); ; 
Morison['M7N3STVxi']          = False     # (m/s); ; 
Morison['M7N4STVxi']          = False     # (m/s); ; 
Morison['M7N5STVxi']          = False     # (m/s); ; 
Morison['M7N6STVxi']          = False     # (m/s); ; 
Morison['M7N7STVxi']          = False     # (m/s); ; 
Morison['M7N8STVxi']          = False     # (m/s); ; 
Morison['M7N9STVxi']          = False     # (m/s); ; 
Morison['M8N1STVxi']          = False     # (m/s); ; 
Morison['M8N2STVxi']          = False     # (m/s); ; 
Morison['M8N3STVxi']          = False     # (m/s); ; 
Morison['M8N4STVxi']          = False     # (m/s); ; 
Morison['M8N5STVxi']          = False     # (m/s); ; 
Morison['M8N6STVxi']          = False     # (m/s); ; 
Morison['M8N7STVxi']          = False     # (m/s); ; 
Morison['M8N8STVxi']          = False     # (m/s); ; 
Morison['M8N9STVxi']          = False     # (m/s); ; 
Morison['M9N1STVxi']          = False     # (m/s); ; 
Morison['M9N2STVxi']          = False     # (m/s); ; 
Morison['M9N3STVxi']          = False     # (m/s); ; 
Morison['M9N4STVxi']          = False     # (m/s); ; 
Morison['M9N5STVxi']          = False     # (m/s); ; 
Morison['M9N6STVxi']          = False     # (m/s); ; 
Morison['M9N7STVxi']          = False     # (m/s); ; 
Morison['M9N8STVxi']          = False     # (m/s); ; 
Morison['M9N9STVxi']          = False     # (m/s); ; 
Morison['M1N1STVyi']          = False     # (m/s); ; 
Morison['M1N2STVyi']          = False     # (m/s); ; 
Morison['M1N3STVyi']          = False     # (m/s); ; 
Morison['M1N4STVyi']          = False     # (m/s); ; 
Morison['M1N5STVyi']          = False     # (m/s); ; 
Morison['M1N6STVyi']          = False     # (m/s); ; 
Morison['M1N7STVyi']          = False     # (m/s); ; 
Morison['M1N8STVyi']          = False     # (m/s); ; 
Morison['M1N9STVyi']          = False     # (m/s); ; 
Morison['M2N1STVyi']          = False     # (m/s); ; 
Morison['M2N2STVyi']          = False     # (m/s); ; 
Morison['M2N3STVyi']          = False     # (m/s); ; 
Morison['M2N4STVyi']          = False     # (m/s); ; 
Morison['M2N5STVyi']          = False     # (m/s); ; 
Morison['M2N6STVyi']          = False     # (m/s); ; 
Morison['M2N7STVyi']          = False     # (m/s); ; 
Morison['M2N8STVyi']          = False     # (m/s); ; 
Morison['M2N9STVyi']          = False     # (m/s); ; 
Morison['M3N1STVyi']          = False     # (m/s); ; 
Morison['M3N2STVyi']          = False     # (m/s); ; 
Morison['M3N3STVyi']          = False     # (m/s); ; 
Morison['M3N4STVyi']          = False     # (m/s); ; 
Morison['M3N5STVyi']          = False     # (m/s); ; 
Morison['M3N6STVyi']          = False     # (m/s); ; 
Morison['M3N7STVyi']          = False     # (m/s); ; 
Morison['M3N8STVyi']          = False     # (m/s); ; 
Morison['M3N9STVyi']          = False     # (m/s); ; 
Morison['M4N1STVyi']          = False     # (m/s); ; 
Morison['M4N2STVyi']          = False     # (m/s); ; 
Morison['M4N3STVyi']          = False     # (m/s); ; 
Morison['M4N4STVyi']          = False     # (m/s); ; 
Morison['M4N5STVyi']          = False     # (m/s); ; 
Morison['M4N6STVyi']          = False     # (m/s); ; 
Morison['M4N7STVyi']          = False     # (m/s); ; 
Morison['M4N8STVyi']          = False     # (m/s); ; 
Morison['M4N9STVyi']          = False     # (m/s); ; 
Morison['M5N1STVyi']          = False     # (m/s); ; 
Morison['M5N2STVyi']          = False     # (m/s); ; 
Morison['M5N3STVyi']          = False     # (m/s); ; 
Morison['M5N4STVyi']          = False     # (m/s); ; 
Morison['M5N5STVyi']          = False     # (m/s); ; 
Morison['M5N6STVyi']          = False     # (m/s); ; 
Morison['M5N7STVyi']          = False     # (m/s); ; 
Morison['M5N8STVyi']          = False     # (m/s); ; 
Morison['M5N9STVyi']          = False     # (m/s); ; 
Morison['M6N1STVyi']          = False     # (m/s); ; 
Morison['M6N2STVyi']          = False     # (m/s); ; 
Morison['M6N3STVyi']          = False     # (m/s); ; 
Morison['M6N4STVyi']          = False     # (m/s); ; 
Morison['M6N5STVyi']          = False     # (m/s); ; 
Morison['M6N6STVyi']          = False     # (m/s); ; 
Morison['M6N7STVyi']          = False     # (m/s); ; 
Morison['M6N8STVyi']          = False     # (m/s); ; 
Morison['M6N9STVyi']          = False     # (m/s); ; 
Morison['M7N1STVyi']          = False     # (m/s); ; 
Morison['M7N2STVyi']          = False     # (m/s); ; 
Morison['M7N3STVyi']          = False     # (m/s); ; 
Morison['M7N4STVyi']          = False     # (m/s); ; 
Morison['M7N5STVyi']          = False     # (m/s); ; 
Morison['M7N6STVyi']          = False     # (m/s); ; 
Morison['M7N7STVyi']          = False     # (m/s); ; 
Morison['M7N8STVyi']          = False     # (m/s); ; 
Morison['M7N9STVyi']          = False     # (m/s); ; 
Morison['M8N1STVyi']          = False     # (m/s); ; 
Morison['M8N2STVyi']          = False     # (m/s); ; 
Morison['M8N3STVyi']          = False     # (m/s); ; 
Morison['M8N4STVyi']          = False     # (m/s); ; 
Morison['M8N5STVyi']          = False     # (m/s); ; 
Morison['M8N6STVyi']          = False     # (m/s); ; 
Morison['M8N7STVyi']          = False     # (m/s); ; 
Morison['M8N8STVyi']          = False     # (m/s); ; 
Morison['M8N9STVyi']          = False     # (m/s); ; 
Morison['M9N1STVyi']          = False     # (m/s); ; 
Morison['M9N2STVyi']          = False     # (m/s); ; 
Morison['M9N3STVyi']          = False     # (m/s); ; 
Morison['M9N4STVyi']          = False     # (m/s); ; 
Morison['M9N5STVyi']          = False     # (m/s); ; 
Morison['M9N6STVyi']          = False     # (m/s); ; 
Morison['M9N7STVyi']          = False     # (m/s); ; 
Morison['M9N8STVyi']          = False     # (m/s); ; 
Morison['M9N9STVyi']          = False     # (m/s); ; 
Morison['M1N1STVzi']          = False     # (m/s); ; 
Morison['M1N2STVzi']          = False     # (m/s); ; 
Morison['M1N3STVzi']          = False     # (m/s); ; 
Morison['M1N4STVzi']          = False     # (m/s); ; 
Morison['M1N5STVzi']          = False     # (m/s); ; 
Morison['M1N6STVzi']          = False     # (m/s); ; 
Morison['M1N7STVzi']          = False     # (m/s); ; 
Morison['M1N8STVzi']          = False     # (m/s); ; 
Morison['M1N9STVzi']          = False     # (m/s); ; 
Morison['M2N1STVzi']          = False     # (m/s); ; 
Morison['M2N2STVzi']          = False     # (m/s); ; 
Morison['M2N3STVzi']          = False     # (m/s); ; 
Morison['M2N4STVzi']          = False     # (m/s); ; 
Morison['M2N5STVzi']          = False     # (m/s); ; 
Morison['M2N6STVzi']          = False     # (m/s); ; 
Morison['M2N7STVzi']          = False     # (m/s); ; 
Morison['M2N8STVzi']          = False     # (m/s); ; 
Morison['M2N9STVzi']          = False     # (m/s); ; 
Morison['M3N1STVzi']          = False     # (m/s); ; 
Morison['M3N2STVzi']          = False     # (m/s); ; 
Morison['M3N3STVzi']          = False     # (m/s); ; 
Morison['M3N4STVzi']          = False     # (m/s); ; 
Morison['M3N5STVzi']          = False     # (m/s); ; 
Morison['M3N6STVzi']          = False     # (m/s); ; 
Morison['M3N7STVzi']          = False     # (m/s); ; 
Morison['M3N8STVzi']          = False     # (m/s); ; 
Morison['M3N9STVzi']          = False     # (m/s); ; 
Morison['M4N1STVzi']          = False     # (m/s); ; 
Morison['M4N2STVzi']          = False     # (m/s); ; 
Morison['M4N3STVzi']          = False     # (m/s); ; 
Morison['M4N4STVzi']          = False     # (m/s); ; 
Morison['M4N5STVzi']          = False     # (m/s); ; 
Morison['M4N6STVzi']          = False     # (m/s); ; 
Morison['M4N7STVzi']          = False     # (m/s); ; 
Morison['M4N8STVzi']          = False     # (m/s); ; 
Morison['M4N9STVzi']          = False     # (m/s); ; 
Morison['M5N1STVzi']          = False     # (m/s); ; 
Morison['M5N2STVzi']          = False     # (m/s); ; 
Morison['M5N3STVzi']          = False     # (m/s); ; 
Morison['M5N4STVzi']          = False     # (m/s); ; 
Morison['M5N5STVzi']          = False     # (m/s); ; 
Morison['M5N6STVzi']          = False     # (m/s); ; 
Morison['M5N7STVzi']          = False     # (m/s); ; 
Morison['M5N8STVzi']          = False     # (m/s); ; 
Morison['M5N9STVzi']          = False     # (m/s); ; 
Morison['M6N1STVzi']          = False     # (m/s); ; 
Morison['M6N2STVzi']          = False     # (m/s); ; 
Morison['M6N3STVzi']          = False     # (m/s); ; 
Morison['M6N4STVzi']          = False     # (m/s); ; 
Morison['M6N5STVzi']          = False     # (m/s); ; 
Morison['M6N6STVzi']          = False     # (m/s); ; 
Morison['M6N7STVzi']          = False     # (m/s); ; 
Morison['M6N8STVzi']          = False     # (m/s); ; 
Morison['M6N9STVzi']          = False     # (m/s); ; 
Morison['M7N1STVzi']          = False     # (m/s); ; 
Morison['M7N2STVzi']          = False     # (m/s); ; 
Morison['M7N3STVzi']          = False     # (m/s); ; 
Morison['M7N4STVzi']          = False     # (m/s); ; 
Morison['M7N5STVzi']          = False     # (m/s); ; 
Morison['M7N6STVzi']          = False     # (m/s); ; 
Morison['M7N7STVzi']          = False     # (m/s); ; 
Morison['M7N8STVzi']          = False     # (m/s); ; 
Morison['M7N9STVzi']          = False     # (m/s); ; 
Morison['M8N1STVzi']          = False     # (m/s); ; 
Morison['M8N2STVzi']          = False     # (m/s); ; 
Morison['M8N3STVzi']          = False     # (m/s); ; 
Morison['M8N4STVzi']          = False     # (m/s); ; 
Morison['M8N5STVzi']          = False     # (m/s); ; 
Morison['M8N6STVzi']          = False     # (m/s); ; 
Morison['M8N7STVzi']          = False     # (m/s); ; 
Morison['M8N8STVzi']          = False     # (m/s); ; 
Morison['M8N9STVzi']          = False     # (m/s); ; 
Morison['M9N1STVzi']          = False     # (m/s); ; 
Morison['M9N2STVzi']          = False     # (m/s); ; 
Morison['M9N3STVzi']          = False     # (m/s); ; 
Morison['M9N4STVzi']          = False     # (m/s); ; 
Morison['M9N5STVzi']          = False     # (m/s); ; 
Morison['M9N6STVzi']          = False     # (m/s); ; 
Morison['M9N7STVzi']          = False     # (m/s); ; 
Morison['M9N8STVzi']          = False     # (m/s); ; 
Morison['M9N9STVzi']          = False     # (m/s); ; 
Morison['M1N1STAxi']          = False     # (m/s); structure translational Acceleration; 
Morison['M1N2STAxi']          = False     # (m/s); ; 
Morison['M1N3STAxi']          = False     # (m/s); ; 
Morison['M1N4STAxi']          = False     # (m/s); ; 
Morison['M1N5STAxi']          = False     # (m/s); ; 
Morison['M1N6STAxi']          = False     # (m/s); ; 
Morison['M1N7STAxi']          = False     # (m/s); ; 
Morison['M1N8STAxi']          = False     # (m/s); ; 
Morison['M1N9STAxi']          = False     # (m/s); ; 
Morison['M2N1STAxi']          = False     # (m/s); ; 
Morison['M2N2STAxi']          = False     # (m/s); ; 
Morison['M2N3STAxi']          = False     # (m/s); ; 
Morison['M2N4STAxi']          = False     # (m/s); ; 
Morison['M2N5STAxi']          = False     # (m/s); ; 
Morison['M2N6STAxi']          = False     # (m/s); ; 
Morison['M2N7STAxi']          = False     # (m/s); ; 
Morison['M2N8STAxi']          = False     # (m/s); ; 
Morison['M2N9STAxi']          = False     # (m/s); ; 
Morison['M3N1STAxi']          = False     # (m/s); ; 
Morison['M3N2STAxi']          = False     # (m/s); ; 
Morison['M3N3STAxi']          = False     # (m/s); ; 
Morison['M3N4STAxi']          = False     # (m/s); ; 
Morison['M3N5STAxi']          = False     # (m/s); ; 
Morison['M3N6STAxi']          = False     # (m/s); ; 
Morison['M3N7STAxi']          = False     # (m/s); ; 
Morison['M3N8STAxi']          = False     # (m/s); ; 
Morison['M3N9STAxi']          = False     # (m/s); ; 
Morison['M4N1STAxi']          = False     # (m/s); ; 
Morison['M4N2STAxi']          = False     # (m/s); ; 
Morison['M4N3STAxi']          = False     # (m/s); ; 
Morison['M4N4STAxi']          = False     # (m/s); ; 
Morison['M4N5STAxi']          = False     # (m/s); ; 
Morison['M4N6STAxi']          = False     # (m/s); ; 
Morison['M4N7STAxi']          = False     # (m/s); ; 
Morison['M4N8STAxi']          = False     # (m/s); ; 
Morison['M4N9STAxi']          = False     # (m/s); ; 
Morison['M5N1STAxi']          = False     # (m/s); ; 
Morison['M5N2STAxi']          = False     # (m/s); ; 
Morison['M5N3STAxi']          = False     # (m/s); ; 
Morison['M5N4STAxi']          = False     # (m/s); ; 
Morison['M5N5STAxi']          = False     # (m/s); ; 
Morison['M5N6STAxi']          = False     # (m/s); ; 
Morison['M5N7STAxi']          = False     # (m/s); ; 
Morison['M5N8STAxi']          = False     # (m/s); ; 
Morison['M5N9STAxi']          = False     # (m/s); ; 
Morison['M6N1STAxi']          = False     # (m/s); ; 
Morison['M6N2STAxi']          = False     # (m/s); ; 
Morison['M6N3STAxi']          = False     # (m/s); ; 
Morison['M6N4STAxi']          = False     # (m/s); ; 
Morison['M6N5STAxi']          = False     # (m/s); ; 
Morison['M6N6STAxi']          = False     # (m/s); ; 
Morison['M6N7STAxi']          = False     # (m/s); ; 
Morison['M6N8STAxi']          = False     # (m/s); ; 
Morison['M6N9STAxi']          = False     # (m/s); ; 
Morison['M7N1STAxi']          = False     # (m/s); ; 
Morison['M7N2STAxi']          = False     # (m/s); ; 
Morison['M7N3STAxi']          = False     # (m/s); ; 
Morison['M7N4STAxi']          = False     # (m/s); ; 
Morison['M7N5STAxi']          = False     # (m/s); ; 
Morison['M7N6STAxi']          = False     # (m/s); ; 
Morison['M7N7STAxi']          = False     # (m/s); ; 
Morison['M7N8STAxi']          = False     # (m/s); ; 
Morison['M7N9STAxi']          = False     # (m/s); ; 
Morison['M8N1STAxi']          = False     # (m/s); ; 
Morison['M8N2STAxi']          = False     # (m/s); ; 
Morison['M8N3STAxi']          = False     # (m/s); ; 
Morison['M8N4STAxi']          = False     # (m/s); ; 
Morison['M8N5STAxi']          = False     # (m/s); ; 
Morison['M8N6STAxi']          = False     # (m/s); ; 
Morison['M8N7STAxi']          = False     # (m/s); ; 
Morison['M8N8STAxi']          = False     # (m/s); ; 
Morison['M8N9STAxi']          = False     # (m/s); ; 
Morison['M9N1STAxi']          = False     # (m/s); ; 
Morison['M9N2STAxi']          = False     # (m/s); ; 
Morison['M9N3STAxi']          = False     # (m/s); ; 
Morison['M9N4STAxi']          = False     # (m/s); ; 
Morison['M9N5STAxi']          = False     # (m/s); ; 
Morison['M9N6STAxi']          = False     # (m/s); ; 
Morison['M9N7STAxi']          = False     # (m/s); ; 
Morison['M9N8STAxi']          = False     # (m/s); ; 
Morison['M9N9STAxi']          = False     # (m/s); ; 
Morison['M1N1STAyi']          = False     # (m/s); ; 
Morison['M1N2STAyi']          = False     # (m/s); ; 
Morison['M1N3STAyi']          = False     # (m/s); ; 
Morison['M1N4STAyi']          = False     # (m/s); ; 
Morison['M1N5STAyi']          = False     # (m/s); ; 
Morison['M1N6STAyi']          = False     # (m/s); ; 
Morison['M1N7STAyi']          = False     # (m/s); ; 
Morison['M1N8STAyi']          = False     # (m/s); ; 
Morison['M1N9STAyi']          = False     # (m/s); ; 
Morison['M2N1STAyi']          = False     # (m/s); ; 
Morison['M2N2STAyi']          = False     # (m/s); ; 
Morison['M2N3STAyi']          = False     # (m/s); ; 
Morison['M2N4STAyi']          = False     # (m/s); ; 
Morison['M2N5STAyi']          = False     # (m/s); ; 
Morison['M2N6STAyi']          = False     # (m/s); ; 
Morison['M2N7STAyi']          = False     # (m/s); ; 
Morison['M2N8STAyi']          = False     # (m/s); ; 
Morison['M2N9STAyi']          = False     # (m/s); ; 
Morison['M3N1STAyi']          = False     # (m/s); ; 
Morison['M3N2STAyi']          = False     # (m/s); ; 
Morison['M3N3STAyi']          = False     # (m/s); ; 
Morison['M3N4STAyi']          = False     # (m/s); ; 
Morison['M3N5STAyi']          = False     # (m/s); ; 
Morison['M3N6STAyi']          = False     # (m/s); ; 
Morison['M3N7STAyi']          = False     # (m/s); ; 
Morison['M3N8STAyi']          = False     # (m/s); ; 
Morison['M3N9STAyi']          = False     # (m/s); ; 
Morison['M4N1STAyi']          = False     # (m/s); ; 
Morison['M4N2STAyi']          = False     # (m/s); ; 
Morison['M4N3STAyi']          = False     # (m/s); ; 
Morison['M4N4STAyi']          = False     # (m/s); ; 
Morison['M4N5STAyi']          = False     # (m/s); ; 
Morison['M4N6STAyi']          = False     # (m/s); ; 
Morison['M4N7STAyi']          = False     # (m/s); ; 
Morison['M4N8STAyi']          = False     # (m/s); ; 
Morison['M4N9STAyi']          = False     # (m/s); ; 
Morison['M5N1STAyi']          = False     # (m/s); ; 
Morison['M5N2STAyi']          = False     # (m/s); ; 
Morison['M5N3STAyi']          = False     # (m/s); ; 
Morison['M5N4STAyi']          = False     # (m/s); ; 
Morison['M5N5STAyi']          = False     # (m/s); ; 
Morison['M5N6STAyi']          = False     # (m/s); ; 
Morison['M5N7STAyi']          = False     # (m/s); ; 
Morison['M5N8STAyi']          = False     # (m/s); ; 
Morison['M5N9STAyi']          = False     # (m/s); ; 
Morison['M6N1STAyi']          = False     # (m/s); ; 
Morison['M6N2STAyi']          = False     # (m/s); ; 
Morison['M6N3STAyi']          = False     # (m/s); ; 
Morison['M6N4STAyi']          = False     # (m/s); ; 
Morison['M6N5STAyi']          = False     # (m/s); ; 
Morison['M6N6STAyi']          = False     # (m/s); ; 
Morison['M6N7STAyi']          = False     # (m/s); ; 
Morison['M6N8STAyi']          = False     # (m/s); ; 
Morison['M6N9STAyi']          = False     # (m/s); ; 
Morison['M7N1STAyi']          = False     # (m/s); ; 
Morison['M7N2STAyi']          = False     # (m/s); ; 
Morison['M7N3STAyi']          = False     # (m/s); ; 
Morison['M7N4STAyi']          = False     # (m/s); ; 
Morison['M7N5STAyi']          = False     # (m/s); ; 
Morison['M7N6STAyi']          = False     # (m/s); ; 
Morison['M7N7STAyi']          = False     # (m/s); ; 
Morison['M7N8STAyi']          = False     # (m/s); ; 
Morison['M7N9STAyi']          = False     # (m/s); ; 
Morison['M8N1STAyi']          = False     # (m/s); ; 
Morison['M8N2STAyi']          = False     # (m/s); ; 
Morison['M8N3STAyi']          = False     # (m/s); ; 
Morison['M8N4STAyi']          = False     # (m/s); ; 
Morison['M8N5STAyi']          = False     # (m/s); ; 
Morison['M8N6STAyi']          = False     # (m/s); ; 
Morison['M8N7STAyi']          = False     # (m/s); ; 
Morison['M8N8STAyi']          = False     # (m/s); ; 
Morison['M8N9STAyi']          = False     # (m/s); ; 
Morison['M9N1STAyi']          = False     # (m/s); ; 
Morison['M9N2STAyi']          = False     # (m/s); ; 
Morison['M9N3STAyi']          = False     # (m/s); ; 
Morison['M9N4STAyi']          = False     # (m/s); ; 
Morison['M9N5STAyi']          = False     # (m/s); ; 
Morison['M9N6STAyi']          = False     # (m/s); ; 
Morison['M9N7STAyi']          = False     # (m/s); ; 
Morison['M9N8STAyi']          = False     # (m/s); ; 
Morison['M9N9STAyi']          = False     # (m/s); ; 
Morison['M1N1STAzi']          = False     # (m/s); ; 
Morison['M1N2STAzi']          = False     # (m/s); ; 
Morison['M1N3STAzi']          = False     # (m/s); ; 
Morison['M1N4STAzi']          = False     # (m/s); ; 
Morison['M1N5STAzi']          = False     # (m/s); ; 
Morison['M1N6STAzi']          = False     # (m/s); ; 
Morison['M1N7STAzi']          = False     # (m/s); ; 
Morison['M1N8STAzi']          = False     # (m/s); ; 
Morison['M1N9STAzi']          = False     # (m/s); ; 
Morison['M2N1STAzi']          = False     # (m/s); ; 
Morison['M2N2STAzi']          = False     # (m/s); ; 
Morison['M2N3STAzi']          = False     # (m/s); ; 
Morison['M2N4STAzi']          = False     # (m/s); ; 
Morison['M2N5STAzi']          = False     # (m/s); ; 
Morison['M2N6STAzi']          = False     # (m/s); ; 
Morison['M2N7STAzi']          = False     # (m/s); ; 
Morison['M2N8STAzi']          = False     # (m/s); ; 
Morison['M2N9STAzi']          = False     # (m/s); ; 
Morison['M3N1STAzi']          = False     # (m/s); ; 
Morison['M3N2STAzi']          = False     # (m/s); ; 
Morison['M3N3STAzi']          = False     # (m/s); ; 
Morison['M3N4STAzi']          = False     # (m/s); ; 
Morison['M3N5STAzi']          = False     # (m/s); ; 
Morison['M3N6STAzi']          = False     # (m/s); ; 
Morison['M3N7STAzi']          = False     # (m/s); ; 
Morison['M3N8STAzi']          = False     # (m/s); ; 
Morison['M3N9STAzi']          = False     # (m/s); ; 
Morison['M4N1STAzi']          = False     # (m/s); ; 
Morison['M4N2STAzi']          = False     # (m/s); ; 
Morison['M4N3STAzi']          = False     # (m/s); ; 
Morison['M4N4STAzi']          = False     # (m/s); ; 
Morison['M4N5STAzi']          = False     # (m/s); ; 
Morison['M4N6STAzi']          = False     # (m/s); ; 
Morison['M4N7STAzi']          = False     # (m/s); ; 
Morison['M4N8STAzi']          = False     # (m/s); ; 
Morison['M4N9STAzi']          = False     # (m/s); ; 
Morison['M5N1STAzi']          = False     # (m/s); ; 
Morison['M5N2STAzi']          = False     # (m/s); ; 
Morison['M5N3STAzi']          = False     # (m/s); ; 
Morison['M5N4STAzi']          = False     # (m/s); ; 
Morison['M5N5STAzi']          = False     # (m/s); ; 
Morison['M5N6STAzi']          = False     # (m/s); ; 
Morison['M5N7STAzi']          = False     # (m/s); ; 
Morison['M5N8STAzi']          = False     # (m/s); ; 
Morison['M5N9STAzi']          = False     # (m/s); ; 
Morison['M6N1STAzi']          = False     # (m/s); ; 
Morison['M6N2STAzi']          = False     # (m/s); ; 
Morison['M6N3STAzi']          = False     # (m/s); ; 
Morison['M6N4STAzi']          = False     # (m/s); ; 
Morison['M6N5STAzi']          = False     # (m/s); ; 
Morison['M6N6STAzi']          = False     # (m/s); ; 
Morison['M6N7STAzi']          = False     # (m/s); ; 
Morison['M6N8STAzi']          = False     # (m/s); ; 
Morison['M6N9STAzi']          = False     # (m/s); ; 
Morison['M7N1STAzi']          = False     # (m/s); ; 
Morison['M7N2STAzi']          = False     # (m/s); ; 
Morison['M7N3STAzi']          = False     # (m/s); ; 
Morison['M7N4STAzi']          = False     # (m/s); ; 
Morison['M7N5STAzi']          = False     # (m/s); ; 
Morison['M7N6STAzi']          = False     # (m/s); ; 
Morison['M7N7STAzi']          = False     # (m/s); ; 
Morison['M7N8STAzi']          = False     # (m/s); ; 
Morison['M7N9STAzi']          = False     # (m/s); ; 
Morison['M8N1STAzi']          = False     # (m/s); ; 
Morison['M8N2STAzi']          = False     # (m/s); ; 
Morison['M8N3STAzi']          = False     # (m/s); ; 
Morison['M8N4STAzi']          = False     # (m/s); ; 
Morison['M8N5STAzi']          = False     # (m/s); ; 
Morison['M8N6STAzi']          = False     # (m/s); ; 
Morison['M8N7STAzi']          = False     # (m/s); ; 
Morison['M8N8STAzi']          = False     # (m/s); ; 
Morison['M8N9STAzi']          = False     # (m/s); ; 
Morison['M9N1STAzi']          = False     # (m/s); ; 
Morison['M9N2STAzi']          = False     # (m/s); ; 
Morison['M9N3STAzi']          = False     # (m/s); ; 
Morison['M9N4STAzi']          = False     # (m/s); ; 
Morison['M9N5STAzi']          = False     # (m/s); ; 
Morison['M9N6STAzi']          = False     # (m/s); ; 
Morison['M9N7STAzi']          = False     # (m/s); ; 
Morison['M9N8STAzi']          = False     # (m/s); ; 
Morison['M9N9STAzi']          = False     # (m/s); ; 

# Morison Element Loads
Morison['M1N1FDxi']           = False     # (N/m); x-component of the distributed drag force expressed in the inertial coordinate system; 
Morison['M1N2FDxi']           = False     # (N/m); ; 
Morison['M1N3FDxi']           = False     # (N/m); ; 
Morison['M1N4FDxi']           = False     # (N/m); ; 
Morison['M1N5FDxi']           = False     # (N/m); ; 
Morison['M1N6FDxi']           = False     # (N/m); ; 
Morison['M1N7FDxi']           = False     # (N/m); ; 
Morison['M1N8FDxi']           = False     # (N/m); ; 
Morison['M1N9FDxi']           = False     # (N/m); ; 
Morison['M2N1FDxi']           = False     # (N/m); ; 
Morison['M2N2FDxi']           = False     # (N/m); ; 
Morison['M2N3FDxi']           = False     # (N/m); ; 
Morison['M2N4FDxi']           = False     # (N/m); ; 
Morison['M2N5FDxi']           = False     # (N/m); ; 
Morison['M2N6FDxi']           = False     # (N/m); ; 
Morison['M2N7FDxi']           = False     # (N/m); ; 
Morison['M2N8FDxi']           = False     # (N/m); ; 
Morison['M2N9FDxi']           = False     # (N/m); ; 
Morison['M3N1FDxi']           = False     # (N/m); ; 
Morison['M3N2FDxi']           = False     # (N/m); ; 
Morison['M3N3FDxi']           = False     # (N/m); ; 
Morison['M3N4FDxi']           = False     # (N/m); ; 
Morison['M3N5FDxi']           = False     # (N/m); ; 
Morison['M3N6FDxi']           = False     # (N/m); ; 
Morison['M3N7FDxi']           = False     # (N/m); ; 
Morison['M3N8FDxi']           = False     # (N/m); ; 
Morison['M3N9FDxi']           = False     # (N/m); ; 
Morison['M4N1FDxi']           = False     # (N/m); ; 
Morison['M4N2FDxi']           = False     # (N/m); ; 
Morison['M4N3FDxi']           = False     # (N/m); ; 
Morison['M4N4FDxi']           = False     # (N/m); ; 
Morison['M4N5FDxi']           = False     # (N/m); ; 
Morison['M4N6FDxi']           = False     # (N/m); ; 
Morison['M4N7FDxi']           = False     # (N/m); ; 
Morison['M4N8FDxi']           = False     # (N/m); ; 
Morison['M4N9FDxi']           = False     # (N/m); ; 
Morison['M5N1FDxi']           = False     # (N/m); ; 
Morison['M5N2FDxi']           = False     # (N/m); ; 
Morison['M5N3FDxi']           = False     # (N/m); ; 
Morison['M5N4FDxi']           = False     # (N/m); ; 
Morison['M5N5FDxi']           = False     # (N/m); ; 
Morison['M5N6FDxi']           = False     # (N/m); ; 
Morison['M5N7FDxi']           = False     # (N/m); ; 
Morison['M5N8FDxi']           = False     # (N/m); ; 
Morison['M5N9FDxi']           = False     # (N/m); ; 
Morison['M6N1FDxi']           = False     # (N/m); ; 
Morison['M6N2FDxi']           = False     # (N/m); ; 
Morison['M6N3FDxi']           = False     # (N/m); ; 
Morison['M6N4FDxi']           = False     # (N/m); ; 
Morison['M6N5FDxi']           = False     # (N/m); ; 
Morison['M6N6FDxi']           = False     # (N/m); ; 
Morison['M6N7FDxi']           = False     # (N/m); ; 
Morison['M6N8FDxi']           = False     # (N/m); ; 
Morison['M6N9FDxi']           = False     # (N/m); ; 
Morison['M7N1FDxi']           = False     # (N/m); ; 
Morison['M7N2FDxi']           = False     # (N/m); ; 
Morison['M7N3FDxi']           = False     # (N/m); ; 
Morison['M7N4FDxi']           = False     # (N/m); ; 
Morison['M7N5FDxi']           = False     # (N/m); ; 
Morison['M7N6FDxi']           = False     # (N/m); ; 
Morison['M7N7FDxi']           = False     # (N/m); ; 
Morison['M7N8FDxi']           = False     # (N/m); ; 
Morison['M7N9FDxi']           = False     # (N/m); ; 
Morison['M8N1FDxi']           = False     # (N/m); ; 
Morison['M8N2FDxi']           = False     # (N/m); ; 
Morison['M8N3FDxi']           = False     # (N/m); ; 
Morison['M8N4FDxi']           = False     # (N/m); ; 
Morison['M8N5FDxi']           = False     # (N/m); ; 
Morison['M8N6FDxi']           = False     # (N/m); ; 
Morison['M8N7FDxi']           = False     # (N/m); ; 
Morison['M8N8FDxi']           = False     # (N/m); ; 
Morison['M8N9FDxi']           = False     # (N/m); ; 
Morison['M9N1FDxi']           = False     # (N/m); ; 
Morison['M9N2FDxi']           = False     # (N/m); ; 
Morison['M9N3FDxi']           = False     # (N/m); ; 
Morison['M9N4FDxi']           = False     # (N/m); ; 
Morison['M9N5FDxi']           = False     # (N/m); ; 
Morison['M9N6FDxi']           = False     # (N/m); ; 
Morison['M9N7FDxi']           = False     # (N/m); ; 
Morison['M9N8FDxi']           = False     # (N/m); ; 
Morison['M9N9FDxi']           = False     # (N/m); ; 
Morison['M1N1FDyi']           = False     # (N/m); y-component of the distributed drag force expressed in the inertial coordinate system; 
Morison['M1N2FDyi']           = False     # (N/m); ; 
Morison['M1N3FDyi']           = False     # (N/m); ; 
Morison['M1N4FDyi']           = False     # (N/m); ; 
Morison['M1N5FDyi']           = False     # (N/m); ; 
Morison['M1N6FDyi']           = False     # (N/m); ; 
Morison['M1N7FDyi']           = False     # (N/m); ; 
Morison['M1N8FDyi']           = False     # (N/m); ; 
Morison['M1N9FDyi']           = False     # (N/m); ; 
Morison['M2N1FDyi']           = False     # (N/m); ; 
Morison['M2N2FDyi']           = False     # (N/m); ; 
Morison['M2N3FDyi']           = False     # (N/m); ; 
Morison['M2N4FDyi']           = False     # (N/m); ; 
Morison['M2N5FDyi']           = False     # (N/m); ; 
Morison['M2N6FDyi']           = False     # (N/m); ; 
Morison['M2N7FDyi']           = False     # (N/m); ; 
Morison['M2N8FDyi']           = False     # (N/m); ; 
Morison['M2N9FDyi']           = False     # (N/m); ; 
Morison['M3N1FDyi']           = False     # (N/m); ; 
Morison['M3N2FDyi']           = False     # (N/m); ; 
Morison['M3N3FDyi']           = False     # (N/m); ; 
Morison['M3N4FDyi']           = False     # (N/m); ; 
Morison['M3N5FDyi']           = False     # (N/m); ; 
Morison['M3N6FDyi']           = False     # (N/m); ; 
Morison['M3N7FDyi']           = False     # (N/m); ; 
Morison['M3N8FDyi']           = False     # (N/m); ; 
Morison['M3N9FDyi']           = False     # (N/m); ; 
Morison['M4N1FDyi']           = False     # (N/m); ; 
Morison['M4N2FDyi']           = False     # (N/m); ; 
Morison['M4N3FDyi']           = False     # (N/m); ; 
Morison['M4N4FDyi']           = False     # (N/m); ; 
Morison['M4N5FDyi']           = False     # (N/m); ; 
Morison['M4N6FDyi']           = False     # (N/m); ; 
Morison['M4N7FDyi']           = False     # (N/m); ; 
Morison['M4N8FDyi']           = False     # (N/m); ; 
Morison['M4N9FDyi']           = False     # (N/m); ; 
Morison['M5N1FDyi']           = False     # (N/m); ; 
Morison['M5N2FDyi']           = False     # (N/m); ; 
Morison['M5N3FDyi']           = False     # (N/m); ; 
Morison['M5N4FDyi']           = False     # (N/m); ; 
Morison['M5N5FDyi']           = False     # (N/m); ; 
Morison['M5N6FDyi']           = False     # (N/m); ; 
Morison['M5N7FDyi']           = False     # (N/m); ; 
Morison['M5N8FDyi']           = False     # (N/m); ; 
Morison['M5N9FDyi']           = False     # (N/m); ; 
Morison['M6N1FDyi']           = False     # (N/m); ; 
Morison['M6N2FDyi']           = False     # (N/m); ; 
Morison['M6N3FDyi']           = False     # (N/m); ; 
Morison['M6N4FDyi']           = False     # (N/m); ; 
Morison['M6N5FDyi']           = False     # (N/m); ; 
Morison['M6N6FDyi']           = False     # (N/m); ; 
Morison['M6N7FDyi']           = False     # (N/m); ; 
Morison['M6N8FDyi']           = False     # (N/m); ; 
Morison['M6N9FDyi']           = False     # (N/m); ; 
Morison['M7N1FDyi']           = False     # (N/m); ; 
Morison['M7N2FDyi']           = False     # (N/m); ; 
Morison['M7N3FDyi']           = False     # (N/m); ; 
Morison['M7N4FDyi']           = False     # (N/m); ; 
Morison['M7N5FDyi']           = False     # (N/m); ; 
Morison['M7N6FDyi']           = False     # (N/m); ; 
Morison['M7N7FDyi']           = False     # (N/m); ; 
Morison['M7N8FDyi']           = False     # (N/m); ; 
Morison['M7N9FDyi']           = False     # (N/m); ; 
Morison['M8N1FDyi']           = False     # (N/m); ; 
Morison['M8N2FDyi']           = False     # (N/m); ; 
Morison['M8N3FDyi']           = False     # (N/m); ; 
Morison['M8N4FDyi']           = False     # (N/m); ; 
Morison['M8N5FDyi']           = False     # (N/m); ; 
Morison['M8N6FDyi']           = False     # (N/m); ; 
Morison['M8N7FDyi']           = False     # (N/m); ; 
Morison['M8N8FDyi']           = False     # (N/m); ; 
Morison['M8N9FDyi']           = False     # (N/m); ; 
Morison['M9N1FDyi']           = False     # (N/m); ; 
Morison['M9N2FDyi']           = False     # (N/m); ; 
Morison['M9N3FDyi']           = False     # (N/m); ; 
Morison['M9N4FDyi']           = False     # (N/m); ; 
Morison['M9N5FDyi']           = False     # (N/m); ; 
Morison['M9N6FDyi']           = False     # (N/m); ; 
Morison['M9N7FDyi']           = False     # (N/m); ; 
Morison['M9N8FDyi']           = False     # (N/m); ; 
Morison['M9N9FDyi']           = False     # (N/m); ; 
Morison['M1N1FDzi']           = False     # (N/m); z-component of the distributed drag force expressed in the inertial coordinate system; 
Morison['M1N2FDzi']           = False     # (N/m); ; 
Morison['M1N3FDzi']           = False     # (N/m); ; 
Morison['M1N4FDzi']           = False     # (N/m); ; 
Morison['M1N5FDzi']           = False     # (N/m); ; 
Morison['M1N6FDzi']           = False     # (N/m); ; 
Morison['M1N7FDzi']           = False     # (N/m); ; 
Morison['M1N8FDzi']           = False     # (N/m); ; 
Morison['M1N9FDzi']           = False     # (N/m); ; 
Morison['M2N1FDzi']           = False     # (N/m); ; 
Morison['M2N2FDzi']           = False     # (N/m); ; 
Morison['M2N3FDzi']           = False     # (N/m); ; 
Morison['M2N4FDzi']           = False     # (N/m); ; 
Morison['M2N5FDzi']           = False     # (N/m); ; 
Morison['M2N6FDzi']           = False     # (N/m); ; 
Morison['M2N7FDzi']           = False     # (N/m); ; 
Morison['M2N8FDzi']           = False     # (N/m); ; 
Morison['M2N9FDzi']           = False     # (N/m); ; 
Morison['M3N1FDzi']           = False     # (N/m); ; 
Morison['M3N2FDzi']           = False     # (N/m); ; 
Morison['M3N3FDzi']           = False     # (N/m); ; 
Morison['M3N4FDzi']           = False     # (N/m); ; 
Morison['M3N5FDzi']           = False     # (N/m); ; 
Morison['M3N6FDzi']           = False     # (N/m); ; 
Morison['M3N7FDzi']           = False     # (N/m); ; 
Morison['M3N8FDzi']           = False     # (N/m); ; 
Morison['M3N9FDzi']           = False     # (N/m); ; 
Morison['M4N1FDzi']           = False     # (N/m); ; 
Morison['M4N2FDzi']           = False     # (N/m); ; 
Morison['M4N3FDzi']           = False     # (N/m); ; 
Morison['M4N4FDzi']           = False     # (N/m); ; 
Morison['M4N5FDzi']           = False     # (N/m); ; 
Morison['M4N6FDzi']           = False     # (N/m); ; 
Morison['M4N7FDzi']           = False     # (N/m); ; 
Morison['M4N8FDzi']           = False     # (N/m); ; 
Morison['M4N9FDzi']           = False     # (N/m); ; 
Morison['M5N1FDzi']           = False     # (N/m); ; 
Morison['M5N2FDzi']           = False     # (N/m); ; 
Morison['M5N3FDzi']           = False     # (N/m); ; 
Morison['M5N4FDzi']           = False     # (N/m); ; 
Morison['M5N5FDzi']           = False     # (N/m); ; 
Morison['M5N6FDzi']           = False     # (N/m); ; 
Morison['M5N7FDzi']           = False     # (N/m); ; 
Morison['M5N8FDzi']           = False     # (N/m); ; 
Morison['M5N9FDzi']           = False     # (N/m); ; 
Morison['M6N1FDzi']           = False     # (N/m); ; 
Morison['M6N2FDzi']           = False     # (N/m); ; 
Morison['M6N3FDzi']           = False     # (N/m); ; 
Morison['M6N4FDzi']           = False     # (N/m); ; 
Morison['M6N5FDzi']           = False     # (N/m); ; 
Morison['M6N6FDzi']           = False     # (N/m); ; 
Morison['M6N7FDzi']           = False     # (N/m); ; 
Morison['M6N8FDzi']           = False     # (N/m); ; 
Morison['M6N9FDzi']           = False     # (N/m); ; 
Morison['M7N1FDzi']           = False     # (N/m); ; 
Morison['M7N2FDzi']           = False     # (N/m); ; 
Morison['M7N3FDzi']           = False     # (N/m); ; 
Morison['M7N4FDzi']           = False     # (N/m); ; 
Morison['M7N5FDzi']           = False     # (N/m); ; 
Morison['M7N6FDzi']           = False     # (N/m); ; 
Morison['M7N7FDzi']           = False     # (N/m); ; 
Morison['M7N8FDzi']           = False     # (N/m); ; 
Morison['M7N9FDzi']           = False     # (N/m); ; 
Morison['M8N1FDzi']           = False     # (N/m); ; 
Morison['M8N2FDzi']           = False     # (N/m); ; 
Morison['M8N3FDzi']           = False     # (N/m); ; 
Morison['M8N4FDzi']           = False     # (N/m); ; 
Morison['M8N5FDzi']           = False     # (N/m); ; 
Morison['M8N6FDzi']           = False     # (N/m); ; 
Morison['M8N7FDzi']           = False     # (N/m); ; 
Morison['M8N8FDzi']           = False     # (N/m); ; 
Morison['M8N9FDzi']           = False     # (N/m); ; 
Morison['M9N1FDzi']           = False     # (N/m); ; 
Morison['M9N2FDzi']           = False     # (N/m); ; 
Morison['M9N3FDzi']           = False     # (N/m); ; 
Morison['M9N4FDzi']           = False     # (N/m); ; 
Morison['M9N5FDzi']           = False     # (N/m); ; 
Morison['M9N6FDzi']           = False     # (N/m); ; 
Morison['M9N7FDzi']           = False     # (N/m); ; 
Morison['M9N8FDzi']           = False     # (N/m); ; 
Morison['M9N9FDzi']           = False     # (N/m); ; 
Morison['M1N1FIxi']           = False     # (N/m); x-component of the distributed inertial force expressed in the inertial coordinate system; 
Morison['M1N2FIxi']           = False     # (N/m); ; 
Morison['M1N3FIxi']           = False     # (N/m); ; 
Morison['M1N4FIxi']           = False     # (N/m); ; 
Morison['M1N5FIxi']           = False     # (N/m); ; 
Morison['M1N6FIxi']           = False     # (N/m); ; 
Morison['M1N7FIxi']           = False     # (N/m); ; 
Morison['M1N8FIxi']           = False     # (N/m); ; 
Morison['M1N9FIxi']           = False     # (N/m); ; 
Morison['M2N1FIxi']           = False     # (N/m); ; 
Morison['M2N2FIxi']           = False     # (N/m); ; 
Morison['M2N3FIxi']           = False     # (N/m); ; 
Morison['M2N4FIxi']           = False     # (N/m); ; 
Morison['M2N5FIxi']           = False     # (N/m); ; 
Morison['M2N6FIxi']           = False     # (N/m); ; 
Morison['M2N7FIxi']           = False     # (N/m); ; 
Morison['M2N8FIxi']           = False     # (N/m); ; 
Morison['M2N9FIxi']           = False     # (N/m); ; 
Morison['M3N1FIxi']           = False     # (N/m); ; 
Morison['M3N2FIxi']           = False     # (N/m); ; 
Morison['M3N3FIxi']           = False     # (N/m); ; 
Morison['M3N4FIxi']           = False     # (N/m); ; 
Morison['M3N5FIxi']           = False     # (N/m); ; 
Morison['M3N6FIxi']           = False     # (N/m); ; 
Morison['M3N7FIxi']           = False     # (N/m); ; 
Morison['M3N8FIxi']           = False     # (N/m); ; 
Morison['M3N9FIxi']           = False     # (N/m); ; 
Morison['M4N1FIxi']           = False     # (N/m); ; 
Morison['M4N2FIxi']           = False     # (N/m); ; 
Morison['M4N3FIxi']           = False     # (N/m); ; 
Morison['M4N4FIxi']           = False     # (N/m); ; 
Morison['M4N5FIxi']           = False     # (N/m); ; 
Morison['M4N6FIxi']           = False     # (N/m); ; 
Morison['M4N7FIxi']           = False     # (N/m); ; 
Morison['M4N8FIxi']           = False     # (N/m); ; 
Morison['M4N9FIxi']           = False     # (N/m); ; 
Morison['M5N1FIxi']           = False     # (N/m); ; 
Morison['M5N2FIxi']           = False     # (N/m); ; 
Morison['M5N3FIxi']           = False     # (N/m); ; 
Morison['M5N4FIxi']           = False     # (N/m); ; 
Morison['M5N5FIxi']           = False     # (N/m); ; 
Morison['M5N6FIxi']           = False     # (N/m); ; 
Morison['M5N7FIxi']           = False     # (N/m); ; 
Morison['M5N8FIxi']           = False     # (N/m); ; 
Morison['M5N9FIxi']           = False     # (N/m); ; 
Morison['M6N1FIxi']           = False     # (N/m); ; 
Morison['M6N2FIxi']           = False     # (N/m); ; 
Morison['M6N3FIxi']           = False     # (N/m); ; 
Morison['M6N4FIxi']           = False     # (N/m); ; 
Morison['M6N5FIxi']           = False     # (N/m); ; 
Morison['M6N6FIxi']           = False     # (N/m); ; 
Morison['M6N7FIxi']           = False     # (N/m); ; 
Morison['M6N8FIxi']           = False     # (N/m); ; 
Morison['M6N9FIxi']           = False     # (N/m); ; 
Morison['M7N1FIxi']           = False     # (N/m); ; 
Morison['M7N2FIxi']           = False     # (N/m); ; 
Morison['M7N3FIxi']           = False     # (N/m); ; 
Morison['M7N4FIxi']           = False     # (N/m); ; 
Morison['M7N5FIxi']           = False     # (N/m); ; 
Morison['M7N6FIxi']           = False     # (N/m); ; 
Morison['M7N7FIxi']           = False     # (N/m); ; 
Morison['M7N8FIxi']           = False     # (N/m); ; 
Morison['M7N9FIxi']           = False     # (N/m); ; 
Morison['M8N1FIxi']           = False     # (N/m); ; 
Morison['M8N2FIxi']           = False     # (N/m); ; 
Morison['M8N3FIxi']           = False     # (N/m); ; 
Morison['M8N4FIxi']           = False     # (N/m); ; 
Morison['M8N5FIxi']           = False     # (N/m); ; 
Morison['M8N6FIxi']           = False     # (N/m); ; 
Morison['M8N7FIxi']           = False     # (N/m); ; 
Morison['M8N8FIxi']           = False     # (N/m); ; 
Morison['M8N9FIxi']           = False     # (N/m); ; 
Morison['M9N1FIxi']           = False     # (N/m); ; 
Morison['M9N2FIxi']           = False     # (N/m); ; 
Morison['M9N3FIxi']           = False     # (N/m); ; 
Morison['M9N4FIxi']           = False     # (N/m); ; 
Morison['M9N5FIxi']           = False     # (N/m); ; 
Morison['M9N6FIxi']           = False     # (N/m); ; 
Morison['M9N7FIxi']           = False     # (N/m); ; 
Morison['M9N8FIxi']           = False     # (N/m); ; 
Morison['M9N9FIxi']           = False     # (N/m); ; 
Morison['M1N1FIyi']           = False     # (N/m); y-component of the distributed inertial force expressed in the inertial coordinate system; 
Morison['M1N2FIyi']           = False     # (N/m); ; 
Morison['M1N3FIyi']           = False     # (N/m); ; 
Morison['M1N4FIyi']           = False     # (N/m); ; 
Morison['M1N5FIyi']           = False     # (N/m); ; 
Morison['M1N6FIyi']           = False     # (N/m); ; 
Morison['M1N7FIyi']           = False     # (N/m); ; 
Morison['M1N8FIyi']           = False     # (N/m); ; 
Morison['M1N9FIyi']           = False     # (N/m); ; 
Morison['M2N1FIyi']           = False     # (N/m); ; 
Morison['M2N2FIyi']           = False     # (N/m); ; 
Morison['M2N3FIyi']           = False     # (N/m); ; 
Morison['M2N4FIyi']           = False     # (N/m); ; 
Morison['M2N5FIyi']           = False     # (N/m); ; 
Morison['M2N6FIyi']           = False     # (N/m); ; 
Morison['M2N7FIyi']           = False     # (N/m); ; 
Morison['M2N8FIyi']           = False     # (N/m); ; 
Morison['M2N9FIyi']           = False     # (N/m); ; 
Morison['M3N1FIyi']           = False     # (N/m); ; 
Morison['M3N2FIyi']           = False     # (N/m); ; 
Morison['M3N3FIyi']           = False     # (N/m); ; 
Morison['M3N4FIyi']           = False     # (N/m); ; 
Morison['M3N5FIyi']           = False     # (N/m); ; 
Morison['M3N6FIyi']           = False     # (N/m); ; 
Morison['M3N7FIyi']           = False     # (N/m); ; 
Morison['M3N8FIyi']           = False     # (N/m); ; 
Morison['M3N9FIyi']           = False     # (N/m); ; 
Morison['M4N1FIyi']           = False     # (N/m); ; 
Morison['M4N2FIyi']           = False     # (N/m); ; 
Morison['M4N3FIyi']           = False     # (N/m); ; 
Morison['M4N4FIyi']           = False     # (N/m); ; 
Morison['M4N5FIyi']           = False     # (N/m); ; 
Morison['M4N6FIyi']           = False     # (N/m); ; 
Morison['M4N7FIyi']           = False     # (N/m); ; 
Morison['M4N8FIyi']           = False     # (N/m); ; 
Morison['M4N9FIyi']           = False     # (N/m); ; 
Morison['M5N1FIyi']           = False     # (N/m); ; 
Morison['M5N2FIyi']           = False     # (N/m); ; 
Morison['M5N3FIyi']           = False     # (N/m); ; 
Morison['M5N4FIyi']           = False     # (N/m); ; 
Morison['M5N5FIyi']           = False     # (N/m); ; 
Morison['M5N6FIyi']           = False     # (N/m); ; 
Morison['M5N7FIyi']           = False     # (N/m); ; 
Morison['M5N8FIyi']           = False     # (N/m); ; 
Morison['M5N9FIyi']           = False     # (N/m); ; 
Morison['M6N1FIyi']           = False     # (N/m); ; 
Morison['M6N2FIyi']           = False     # (N/m); ; 
Morison['M6N3FIyi']           = False     # (N/m); ; 
Morison['M6N4FIyi']           = False     # (N/m); ; 
Morison['M6N5FIyi']           = False     # (N/m); ; 
Morison['M6N6FIyi']           = False     # (N/m); ; 
Morison['M6N7FIyi']           = False     # (N/m); ; 
Morison['M6N8FIyi']           = False     # (N/m); ; 
Morison['M6N9FIyi']           = False     # (N/m); ; 
Morison['M7N1FIyi']           = False     # (N/m); ; 
Morison['M7N2FIyi']           = False     # (N/m); ; 
Morison['M7N3FIyi']           = False     # (N/m); ; 
Morison['M7N4FIyi']           = False     # (N/m); ; 
Morison['M7N5FIyi']           = False     # (N/m); ; 
Morison['M7N6FIyi']           = False     # (N/m); ; 
Morison['M7N7FIyi']           = False     # (N/m); ; 
Morison['M7N8FIyi']           = False     # (N/m); ; 
Morison['M7N9FIyi']           = False     # (N/m); ; 
Morison['M8N1FIyi']           = False     # (N/m); ; 
Morison['M8N2FIyi']           = False     # (N/m); ; 
Morison['M8N3FIyi']           = False     # (N/m); ; 
Morison['M8N4FIyi']           = False     # (N/m); ; 
Morison['M8N5FIyi']           = False     # (N/m); ; 
Morison['M8N6FIyi']           = False     # (N/m); ; 
Morison['M8N7FIyi']           = False     # (N/m); ; 
Morison['M8N8FIyi']           = False     # (N/m); ; 
Morison['M8N9FIyi']           = False     # (N/m); ; 
Morison['M9N1FIyi']           = False     # (N/m); ; 
Morison['M9N2FIyi']           = False     # (N/m); ; 
Morison['M9N3FIyi']           = False     # (N/m); ; 
Morison['M9N4FIyi']           = False     # (N/m); ; 
Morison['M9N5FIyi']           = False     # (N/m); ; 
Morison['M9N6FIyi']           = False     # (N/m); ; 
Morison['M9N7FIyi']           = False     # (N/m); ; 
Morison['M9N8FIyi']           = False     # (N/m); ; 
Morison['M9N9FIyi']           = False     # (N/m); ; 
Morison['M1N1FIzi']           = False     # (N/m); z-component of the distributed inertial force expressed in the inertial coordinate system; 
Morison['M1N2FIzi']           = False     # (N/m); ; 
Morison['M1N3FIzi']           = False     # (N/m); ; 
Morison['M1N4FIzi']           = False     # (N/m); ; 
Morison['M1N5FIzi']           = False     # (N/m); ; 
Morison['M1N6FIzi']           = False     # (N/m); ; 
Morison['M1N7FIzi']           = False     # (N/m); ; 
Morison['M1N8FIzi']           = False     # (N/m); ; 
Morison['M1N9FIzi']           = False     # (N/m); ; 
Morison['M2N1FIzi']           = False     # (N/m); ; 
Morison['M2N2FIzi']           = False     # (N/m); ; 
Morison['M2N3FIzi']           = False     # (N/m); ; 
Morison['M2N4FIzi']           = False     # (N/m); ; 
Morison['M2N5FIzi']           = False     # (N/m); ; 
Morison['M2N6FIzi']           = False     # (N/m); ; 
Morison['M2N7FIzi']           = False     # (N/m); ; 
Morison['M2N8FIzi']           = False     # (N/m); ; 
Morison['M2N9FIzi']           = False     # (N/m); ; 
Morison['M3N1FIzi']           = False     # (N/m); ; 
Morison['M3N2FIzi']           = False     # (N/m); ; 
Morison['M3N3FIzi']           = False     # (N/m); ; 
Morison['M3N4FIzi']           = False     # (N/m); ; 
Morison['M3N5FIzi']           = False     # (N/m); ; 
Morison['M3N6FIzi']           = False     # (N/m); ; 
Morison['M3N7FIzi']           = False     # (N/m); ; 
Morison['M3N8FIzi']           = False     # (N/m); ; 
Morison['M3N9FIzi']           = False     # (N/m); ; 
Morison['M4N1FIzi']           = False     # (N/m); ; 
Morison['M4N2FIzi']           = False     # (N/m); ; 
Morison['M4N3FIzi']           = False     # (N/m); ; 
Morison['M4N4FIzi']           = False     # (N/m); ; 
Morison['M4N5FIzi']           = False     # (N/m); ; 
Morison['M4N6FIzi']           = False     # (N/m); ; 
Morison['M4N7FIzi']           = False     # (N/m); ; 
Morison['M4N8FIzi']           = False     # (N/m); ; 
Morison['M4N9FIzi']           = False     # (N/m); ; 
Morison['M5N1FIzi']           = False     # (N/m); ; 
Morison['M5N2FIzi']           = False     # (N/m); ; 
Morison['M5N3FIzi']           = False     # (N/m); ; 
Morison['M5N4FIzi']           = False     # (N/m); ; 
Morison['M5N5FIzi']           = False     # (N/m); ; 
Morison['M5N6FIzi']           = False     # (N/m); ; 
Morison['M5N7FIzi']           = False     # (N/m); ; 
Morison['M5N8FIzi']           = False     # (N/m); ; 
Morison['M5N9FIzi']           = False     # (N/m); ; 
Morison['M6N1FIzi']           = False     # (N/m); ; 
Morison['M6N2FIzi']           = False     # (N/m); ; 
Morison['M6N3FIzi']           = False     # (N/m); ; 
Morison['M6N4FIzi']           = False     # (N/m); ; 
Morison['M6N5FIzi']           = False     # (N/m); ; 
Morison['M6N6FIzi']           = False     # (N/m); ; 
Morison['M6N7FIzi']           = False     # (N/m); ; 
Morison['M6N8FIzi']           = False     # (N/m); ; 
Morison['M6N9FIzi']           = False     # (N/m); ; 
Morison['M7N1FIzi']           = False     # (N/m); ; 
Morison['M7N2FIzi']           = False     # (N/m); ; 
Morison['M7N3FIzi']           = False     # (N/m); ; 
Morison['M7N4FIzi']           = False     # (N/m); ; 
Morison['M7N5FIzi']           = False     # (N/m); ; 
Morison['M7N6FIzi']           = False     # (N/m); ; 
Morison['M7N7FIzi']           = False     # (N/m); ; 
Morison['M7N8FIzi']           = False     # (N/m); ; 
Morison['M7N9FIzi']           = False     # (N/m); ; 
Morison['M8N1FIzi']           = False     # (N/m); ; 
Morison['M8N2FIzi']           = False     # (N/m); ; 
Morison['M8N3FIzi']           = False     # (N/m); ; 
Morison['M8N4FIzi']           = False     # (N/m); ; 
Morison['M8N5FIzi']           = False     # (N/m); ; 
Morison['M8N6FIzi']           = False     # (N/m); ; 
Morison['M8N7FIzi']           = False     # (N/m); ; 
Morison['M8N8FIzi']           = False     # (N/m); ; 
Morison['M8N9FIzi']           = False     # (N/m); ; 
Morison['M9N1FIzi']           = False     # (N/m); ; 
Morison['M9N2FIzi']           = False     # (N/m); ; 
Morison['M9N3FIzi']           = False     # (N/m); ; 
Morison['M9N4FIzi']           = False     # (N/m); ; 
Morison['M9N5FIzi']           = False     # (N/m); ; 
Morison['M9N6FIzi']           = False     # (N/m); ; 
Morison['M9N7FIzi']           = False     # (N/m); ; 
Morison['M9N8FIzi']           = False     # (N/m); ; 
Morison['M9N9FIzi']           = False     # (N/m); ; 
Morison['M1N1FBxi']           = False     # (N/m); x-component of the distributed bouyancy force expressed in the inertial coordinate system; 
Morison['M1N2FBxi']           = False     # (N/m); ; 
Morison['M1N3FBxi']           = False     # (N/m); ; 
Morison['M1N4FBxi']           = False     # (N/m); ; 
Morison['M1N5FBxi']           = False     # (N/m); ; 
Morison['M1N6FBxi']           = False     # (N/m); ; 
Morison['M1N7FBxi']           = False     # (N/m); ; 
Morison['M1N8FBxi']           = False     # (N/m); ; 
Morison['M1N9FBxi']           = False     # (N/m); ; 
Morison['M2N1FBxi']           = False     # (N/m); ; 
Morison['M2N2FBxi']           = False     # (N/m); ; 
Morison['M2N3FBxi']           = False     # (N/m); ; 
Morison['M2N4FBxi']           = False     # (N/m); ; 
Morison['M2N5FBxi']           = False     # (N/m); ; 
Morison['M2N6FBxi']           = False     # (N/m); ; 
Morison['M2N7FBxi']           = False     # (N/m); ; 
Morison['M2N8FBxi']           = False     # (N/m); ; 
Morison['M2N9FBxi']           = False     # (N/m); ; 
Morison['M3N1FBxi']           = False     # (N/m); ; 
Morison['M3N2FBxi']           = False     # (N/m); ; 
Morison['M3N3FBxi']           = False     # (N/m); ; 
Morison['M3N4FBxi']           = False     # (N/m); ; 
Morison['M3N5FBxi']           = False     # (N/m); ; 
Morison['M3N6FBxi']           = False     # (N/m); ; 
Morison['M3N7FBxi']           = False     # (N/m); ; 
Morison['M3N8FBxi']           = False     # (N/m); ; 
Morison['M3N9FBxi']           = False     # (N/m); ; 
Morison['M4N1FBxi']           = False     # (N/m); ; 
Morison['M4N2FBxi']           = False     # (N/m); ; 
Morison['M4N3FBxi']           = False     # (N/m); ; 
Morison['M4N4FBxi']           = False     # (N/m); ; 
Morison['M4N5FBxi']           = False     # (N/m); ; 
Morison['M4N6FBxi']           = False     # (N/m); ; 
Morison['M4N7FBxi']           = False     # (N/m); ; 
Morison['M4N8FBxi']           = False     # (N/m); ; 
Morison['M4N9FBxi']           = False     # (N/m); ; 
Morison['M5N1FBxi']           = False     # (N/m); ; 
Morison['M5N2FBxi']           = False     # (N/m); ; 
Morison['M5N3FBxi']           = False     # (N/m); ; 
Morison['M5N4FBxi']           = False     # (N/m); ; 
Morison['M5N5FBxi']           = False     # (N/m); ; 
Morison['M5N6FBxi']           = False     # (N/m); ; 
Morison['M5N7FBxi']           = False     # (N/m); ; 
Morison['M5N8FBxi']           = False     # (N/m); ; 
Morison['M5N9FBxi']           = False     # (N/m); ; 
Morison['M6N1FBxi']           = False     # (N/m); ; 
Morison['M6N2FBxi']           = False     # (N/m); ; 
Morison['M6N3FBxi']           = False     # (N/m); ; 
Morison['M6N4FBxi']           = False     # (N/m); ; 
Morison['M6N5FBxi']           = False     # (N/m); ; 
Morison['M6N6FBxi']           = False     # (N/m); ; 
Morison['M6N7FBxi']           = False     # (N/m); ; 
Morison['M6N8FBxi']           = False     # (N/m); ; 
Morison['M6N9FBxi']           = False     # (N/m); ; 
Morison['M7N1FBxi']           = False     # (N/m); ; 
Morison['M7N2FBxi']           = False     # (N/m); ; 
Morison['M7N3FBxi']           = False     # (N/m); ; 
Morison['M7N4FBxi']           = False     # (N/m); ; 
Morison['M7N5FBxi']           = False     # (N/m); ; 
Morison['M7N6FBxi']           = False     # (N/m); ; 
Morison['M7N7FBxi']           = False     # (N/m); ; 
Morison['M7N8FBxi']           = False     # (N/m); ; 
Morison['M7N9FBxi']           = False     # (N/m); ; 
Morison['M8N1FBxi']           = False     # (N/m); ; 
Morison['M8N2FBxi']           = False     # (N/m); ; 
Morison['M8N3FBxi']           = False     # (N/m); ; 
Morison['M8N4FBxi']           = False     # (N/m); ; 
Morison['M8N5FBxi']           = False     # (N/m); ; 
Morison['M8N6FBxi']           = False     # (N/m); ; 
Morison['M8N7FBxi']           = False     # (N/m); ; 
Morison['M8N8FBxi']           = False     # (N/m); ; 
Morison['M8N9FBxi']           = False     # (N/m); ; 
Morison['M9N1FBxi']           = False     # (N/m); ; 
Morison['M9N2FBxi']           = False     # (N/m); ; 
Morison['M9N3FBxi']           = False     # (N/m); ; 
Morison['M9N4FBxi']           = False     # (N/m); ; 
Morison['M9N5FBxi']           = False     # (N/m); ; 
Morison['M9N6FBxi']           = False     # (N/m); ; 
Morison['M9N7FBxi']           = False     # (N/m); ; 
Morison['M9N8FBxi']           = False     # (N/m); ; 
Morison['M9N9FBxi']           = False     # (N/m); ; 
Morison['M1N1FByi']           = False     # (N/m); y-component of the distributed bouyancy force expressed in the inertial coordinate system; 
Morison['M1N2FByi']           = False     # (N/m); ; 
Morison['M1N3FByi']           = False     # (N/m); ; 
Morison['M1N4FByi']           = False     # (N/m); ; 
Morison['M1N5FByi']           = False     # (N/m); ; 
Morison['M1N6FByi']           = False     # (N/m); ; 
Morison['M1N7FByi']           = False     # (N/m); ; 
Morison['M1N8FByi']           = False     # (N/m); ; 
Morison['M1N9FByi']           = False     # (N/m); ; 
Morison['M2N1FByi']           = False     # (N/m); ; 
Morison['M2N2FByi']           = False     # (N/m); ; 
Morison['M2N3FByi']           = False     # (N/m); ; 
Morison['M2N4FByi']           = False     # (N/m); ; 
Morison['M2N5FByi']           = False     # (N/m); ; 
Morison['M2N6FByi']           = False     # (N/m); ; 
Morison['M2N7FByi']           = False     # (N/m); ; 
Morison['M2N8FByi']           = False     # (N/m); ; 
Morison['M2N9FByi']           = False     # (N/m); ; 
Morison['M3N1FByi']           = False     # (N/m); ; 
Morison['M3N2FByi']           = False     # (N/m); ; 
Morison['M3N3FByi']           = False     # (N/m); ; 
Morison['M3N4FByi']           = False     # (N/m); ; 
Morison['M3N5FByi']           = False     # (N/m); ; 
Morison['M3N6FByi']           = False     # (N/m); ; 
Morison['M3N7FByi']           = False     # (N/m); ; 
Morison['M3N8FByi']           = False     # (N/m); ; 
Morison['M3N9FByi']           = False     # (N/m); ; 
Morison['M4N1FByi']           = False     # (N/m); ; 
Morison['M4N2FByi']           = False     # (N/m); ; 
Morison['M4N3FByi']           = False     # (N/m); ; 
Morison['M4N4FByi']           = False     # (N/m); ; 
Morison['M4N5FByi']           = False     # (N/m); ; 
Morison['M4N6FByi']           = False     # (N/m); ; 
Morison['M4N7FByi']           = False     # (N/m); ; 
Morison['M4N8FByi']           = False     # (N/m); ; 
Morison['M4N9FByi']           = False     # (N/m); ; 
Morison['M5N1FByi']           = False     # (N/m); ; 
Morison['M5N2FByi']           = False     # (N/m); ; 
Morison['M5N3FByi']           = False     # (N/m); ; 
Morison['M5N4FByi']           = False     # (N/m); ; 
Morison['M5N5FByi']           = False     # (N/m); ; 
Morison['M5N6FByi']           = False     # (N/m); ; 
Morison['M5N7FByi']           = False     # (N/m); ; 
Morison['M5N8FByi']           = False     # (N/m); ; 
Morison['M5N9FByi']           = False     # (N/m); ; 
Morison['M6N1FByi']           = False     # (N/m); ; 
Morison['M6N2FByi']           = False     # (N/m); ; 
Morison['M6N3FByi']           = False     # (N/m); ; 
Morison['M6N4FByi']           = False     # (N/m); ; 
Morison['M6N5FByi']           = False     # (N/m); ; 
Morison['M6N6FByi']           = False     # (N/m); ; 
Morison['M6N7FByi']           = False     # (N/m); ; 
Morison['M6N8FByi']           = False     # (N/m); ; 
Morison['M6N9FByi']           = False     # (N/m); ; 
Morison['M7N1FByi']           = False     # (N/m); ; 
Morison['M7N2FByi']           = False     # (N/m); ; 
Morison['M7N3FByi']           = False     # (N/m); ; 
Morison['M7N4FByi']           = False     # (N/m); ; 
Morison['M7N5FByi']           = False     # (N/m); ; 
Morison['M7N6FByi']           = False     # (N/m); ; 
Morison['M7N7FByi']           = False     # (N/m); ; 
Morison['M7N8FByi']           = False     # (N/m); ; 
Morison['M7N9FByi']           = False     # (N/m); ; 
Morison['M8N1FByi']           = False     # (N/m); ; 
Morison['M8N2FByi']           = False     # (N/m); ; 
Morison['M8N3FByi']           = False     # (N/m); ; 
Morison['M8N4FByi']           = False     # (N/m); ; 
Morison['M8N5FByi']           = False     # (N/m); ; 
Morison['M8N6FByi']           = False     # (N/m); ; 
Morison['M8N7FByi']           = False     # (N/m); ; 
Morison['M8N8FByi']           = False     # (N/m); ; 
Morison['M8N9FByi']           = False     # (N/m); ; 
Morison['M9N1FByi']           = False     # (N/m); ; 
Morison['M9N2FByi']           = False     # (N/m); ; 
Morison['M9N3FByi']           = False     # (N/m); ; 
Morison['M9N4FByi']           = False     # (N/m); ; 
Morison['M9N5FByi']           = False     # (N/m); ; 
Morison['M9N6FByi']           = False     # (N/m); ; 
Morison['M9N7FByi']           = False     # (N/m); ; 
Morison['M9N8FByi']           = False     # (N/m); ; 
Morison['M9N9FByi']           = False     # (N/m); ; 
Morison['M1N1FBzi']           = False     # (N/m); z-component of the distributed bouyancy force expressed in the inertial coordinate system; 
Morison['M1N2FBzi']           = False     # (N/m); ; 
Morison['M1N3FBzi']           = False     # (N/m); ; 
Morison['M1N4FBzi']           = False     # (N/m); ; 
Morison['M1N5FBzi']           = False     # (N/m); ; 
Morison['M1N6FBzi']           = False     # (N/m); ; 
Morison['M1N7FBzi']           = False     # (N/m); ; 
Morison['M1N8FBzi']           = False     # (N/m); ; 
Morison['M1N9FBzi']           = False     # (N/m); ; 
Morison['M2N1FBzi']           = False     # (N/m); ; 
Morison['M2N2FBzi']           = False     # (N/m); ; 
Morison['M2N3FBzi']           = False     # (N/m); ; 
Morison['M2N4FBzi']           = False     # (N/m); ; 
Morison['M2N5FBzi']           = False     # (N/m); ; 
Morison['M2N6FBzi']           = False     # (N/m); ; 
Morison['M2N7FBzi']           = False     # (N/m); ; 
Morison['M2N8FBzi']           = False     # (N/m); ; 
Morison['M2N9FBzi']           = False     # (N/m); ; 
Morison['M3N1FBzi']           = False     # (N/m); ; 
Morison['M3N2FBzi']           = False     # (N/m); ; 
Morison['M3N3FBzi']           = False     # (N/m); ; 
Morison['M3N4FBzi']           = False     # (N/m); ; 
Morison['M3N5FBzi']           = False     # (N/m); ; 
Morison['M3N6FBzi']           = False     # (N/m); ; 
Morison['M3N7FBzi']           = False     # (N/m); ; 
Morison['M3N8FBzi']           = False     # (N/m); ; 
Morison['M3N9FBzi']           = False     # (N/m); ; 
Morison['M4N1FBzi']           = False     # (N/m); ; 
Morison['M4N2FBzi']           = False     # (N/m); ; 
Morison['M4N3FBzi']           = False     # (N/m); ; 
Morison['M4N4FBzi']           = False     # (N/m); ; 
Morison['M4N5FBzi']           = False     # (N/m); ; 
Morison['M4N6FBzi']           = False     # (N/m); ; 
Morison['M4N7FBzi']           = False     # (N/m); ; 
Morison['M4N8FBzi']           = False     # (N/m); ; 
Morison['M4N9FBzi']           = False     # (N/m); ; 
Morison['M5N1FBzi']           = False     # (N/m); ; 
Morison['M5N2FBzi']           = False     # (N/m); ; 
Morison['M5N3FBzi']           = False     # (N/m); ; 
Morison['M5N4FBzi']           = False     # (N/m); ; 
Morison['M5N5FBzi']           = False     # (N/m); ; 
Morison['M5N6FBzi']           = False     # (N/m); ; 
Morison['M5N7FBzi']           = False     # (N/m); ; 
Morison['M5N8FBzi']           = False     # (N/m); ; 
Morison['M5N9FBzi']           = False     # (N/m); ; 
Morison['M6N1FBzi']           = False     # (N/m); ; 
Morison['M6N2FBzi']           = False     # (N/m); ; 
Morison['M6N3FBzi']           = False     # (N/m); ; 
Morison['M6N4FBzi']           = False     # (N/m); ; 
Morison['M6N5FBzi']           = False     # (N/m); ; 
Morison['M6N6FBzi']           = False     # (N/m); ; 
Morison['M6N7FBzi']           = False     # (N/m); ; 
Morison['M6N8FBzi']           = False     # (N/m); ; 
Morison['M6N9FBzi']           = False     # (N/m); ; 
Morison['M7N1FBzi']           = False     # (N/m); ; 
Morison['M7N2FBzi']           = False     # (N/m); ; 
Morison['M7N3FBzi']           = False     # (N/m); ; 
Morison['M7N4FBzi']           = False     # (N/m); ; 
Morison['M7N5FBzi']           = False     # (N/m); ; 
Morison['M7N6FBzi']           = False     # (N/m); ; 
Morison['M7N7FBzi']           = False     # (N/m); ; 
Morison['M7N8FBzi']           = False     # (N/m); ; 
Morison['M7N9FBzi']           = False     # (N/m); ; 
Morison['M8N1FBzi']           = False     # (N/m); ; 
Morison['M8N2FBzi']           = False     # (N/m); ; 
Morison['M8N3FBzi']           = False     # (N/m); ; 
Morison['M8N4FBzi']           = False     # (N/m); ; 
Morison['M8N5FBzi']           = False     # (N/m); ; 
Morison['M8N6FBzi']           = False     # (N/m); ; 
Morison['M8N7FBzi']           = False     # (N/m); ; 
Morison['M8N8FBzi']           = False     # (N/m); ; 
Morison['M8N9FBzi']           = False     # (N/m); ; 
Morison['M9N1FBzi']           = False     # (N/m); ; 
Morison['M9N2FBzi']           = False     # (N/m); ; 
Morison['M9N3FBzi']           = False     # (N/m); ; 
Morison['M9N4FBzi']           = False     # (N/m); ; 
Morison['M9N5FBzi']           = False     # (N/m); ; 
Morison['M9N6FBzi']           = False     # (N/m); ; 
Morison['M9N7FBzi']           = False     # (N/m); ; 
Morison['M9N8FBzi']           = False     # (N/m); ; 
Morison['M9N9FBzi']           = False     # (N/m); ; 
Morison['M1N1MBxi']           = False     # (N); x-component of the distributed bouyancy moment expressed in the inertial coordinate system; 
Morison['M1N2MBxi']           = False     # (N); ; 
Morison['M1N3MBxi']           = False     # (N); ; 
Morison['M1N4MBxi']           = False     # (N); ; 
Morison['M1N5MBxi']           = False     # (N); ; 
Morison['M1N6MBxi']           = False     # (N); ; 
Morison['M1N7MBxi']           = False     # (N); ; 
Morison['M1N8MBxi']           = False     # (N); ; 
Morison['M1N9MBxi']           = False     # (N); ; 
Morison['M2N1MBxi']           = False     # (N); ; 
Morison['M2N2MBxi']           = False     # (N); ; 
Morison['M2N3MBxi']           = False     # (N); ; 
Morison['M2N4MBxi']           = False     # (N); ; 
Morison['M2N5MBxi']           = False     # (N); ; 
Morison['M2N6MBxi']           = False     # (N); ; 
Morison['M2N7MBxi']           = False     # (N); ; 
Morison['M2N8MBxi']           = False     # (N); ; 
Morison['M2N9MBxi']           = False     # (N); ; 
Morison['M3N1MBxi']           = False     # (N); ; 
Morison['M3N2MBxi']           = False     # (N); ; 
Morison['M3N3MBxi']           = False     # (N); ; 
Morison['M3N4MBxi']           = False     # (N); ; 
Morison['M3N5MBxi']           = False     # (N); ; 
Morison['M3N6MBxi']           = False     # (N); ; 
Morison['M3N7MBxi']           = False     # (N); ; 
Morison['M3N8MBxi']           = False     # (N); ; 
Morison['M3N9MBxi']           = False     # (N); ; 
Morison['M4N1MBxi']           = False     # (N); ; 
Morison['M4N2MBxi']           = False     # (N); ; 
Morison['M4N3MBxi']           = False     # (N); ; 
Morison['M4N4MBxi']           = False     # (N); ; 
Morison['M4N5MBxi']           = False     # (N); ; 
Morison['M4N6MBxi']           = False     # (N); ; 
Morison['M4N7MBxi']           = False     # (N); ; 
Morison['M4N8MBxi']           = False     # (N); ; 
Morison['M4N9MBxi']           = False     # (N); ; 
Morison['M5N1MBxi']           = False     # (N); ; 
Morison['M5N2MBxi']           = False     # (N); ; 
Morison['M5N3MBxi']           = False     # (N); ; 
Morison['M5N4MBxi']           = False     # (N); ; 
Morison['M5N5MBxi']           = False     # (N); ; 
Morison['M5N6MBxi']           = False     # (N); ; 
Morison['M5N7MBxi']           = False     # (N); ; 
Morison['M5N8MBxi']           = False     # (N); ; 
Morison['M5N9MBxi']           = False     # (N); ; 
Morison['M6N1MBxi']           = False     # (N); ; 
Morison['M6N2MBxi']           = False     # (N); ; 
Morison['M6N3MBxi']           = False     # (N); ; 
Morison['M6N4MBxi']           = False     # (N); ; 
Morison['M6N5MBxi']           = False     # (N); ; 
Morison['M6N6MBxi']           = False     # (N); ; 
Morison['M6N7MBxi']           = False     # (N); ; 
Morison['M6N8MBxi']           = False     # (N); ; 
Morison['M6N9MBxi']           = False     # (N); ; 
Morison['M7N1MBxi']           = False     # (N); ; 
Morison['M7N2MBxi']           = False     # (N); ; 
Morison['M7N3MBxi']           = False     # (N); ; 
Morison['M7N4MBxi']           = False     # (N); ; 
Morison['M7N5MBxi']           = False     # (N); ; 
Morison['M7N6MBxi']           = False     # (N); ; 
Morison['M7N7MBxi']           = False     # (N); ; 
Morison['M7N8MBxi']           = False     # (N); ; 
Morison['M7N9MBxi']           = False     # (N); ; 
Morison['M8N1MBxi']           = False     # (N); ; 
Morison['M8N2MBxi']           = False     # (N); ; 
Morison['M8N3MBxi']           = False     # (N); ; 
Morison['M8N4MBxi']           = False     # (N); ; 
Morison['M8N5MBxi']           = False     # (N); ; 
Morison['M8N6MBxi']           = False     # (N); ; 
Morison['M8N7MBxi']           = False     # (N); ; 
Morison['M8N8MBxi']           = False     # (N); ; 
Morison['M8N9MBxi']           = False     # (N); ; 
Morison['M9N1MBxi']           = False     # (N); ; 
Morison['M9N2MBxi']           = False     # (N); ; 
Morison['M9N3MBxi']           = False     # (N); ; 
Morison['M9N4MBxi']           = False     # (N); ; 
Morison['M9N5MBxi']           = False     # (N); ; 
Morison['M9N6MBxi']           = False     # (N); ; 
Morison['M9N7MBxi']           = False     # (N); ; 
Morison['M9N8MBxi']           = False     # (N); ; 
Morison['M9N9MBxi']           = False     # (N); ; 
Morison['M1N1MByi']           = False     # (N); y-component of the distributed bouyancy moment expressed in the inertial coordinate system; 
Morison['M1N2MByi']           = False     # (N); ; 
Morison['M1N3MByi']           = False     # (N); ; 
Morison['M1N4MByi']           = False     # (N); ; 
Morison['M1N5MByi']           = False     # (N); ; 
Morison['M1N6MByi']           = False     # (N); ; 
Morison['M1N7MByi']           = False     # (N); ; 
Morison['M1N8MByi']           = False     # (N); ; 
Morison['M1N9MByi']           = False     # (N); ; 
Morison['M2N1MByi']           = False     # (N); ; 
Morison['M2N2MByi']           = False     # (N); ; 
Morison['M2N3MByi']           = False     # (N); ; 
Morison['M2N4MByi']           = False     # (N); ; 
Morison['M2N5MByi']           = False     # (N); ; 
Morison['M2N6MByi']           = False     # (N); ; 
Morison['M2N7MByi']           = False     # (N); ; 
Morison['M2N8MByi']           = False     # (N); ; 
Morison['M2N9MByi']           = False     # (N); ; 
Morison['M3N1MByi']           = False     # (N); ; 
Morison['M3N2MByi']           = False     # (N); ; 
Morison['M3N3MByi']           = False     # (N); ; 
Morison['M3N4MByi']           = False     # (N); ; 
Morison['M3N5MByi']           = False     # (N); ; 
Morison['M3N6MByi']           = False     # (N); ; 
Morison['M3N7MByi']           = False     # (N); ; 
Morison['M3N8MByi']           = False     # (N); ; 
Morison['M3N9MByi']           = False     # (N); ; 
Morison['M4N1MByi']           = False     # (N); ; 
Morison['M4N2MByi']           = False     # (N); ; 
Morison['M4N3MByi']           = False     # (N); ; 
Morison['M4N4MByi']           = False     # (N); ; 
Morison['M4N5MByi']           = False     # (N); ; 
Morison['M4N6MByi']           = False     # (N); ; 
Morison['M4N7MByi']           = False     # (N); ; 
Morison['M4N8MByi']           = False     # (N); ; 
Morison['M4N9MByi']           = False     # (N); ; 
Morison['M5N1MByi']           = False     # (N); ; 
Morison['M5N2MByi']           = False     # (N); ; 
Morison['M5N3MByi']           = False     # (N); ; 
Morison['M5N4MByi']           = False     # (N); ; 
Morison['M5N5MByi']           = False     # (N); ; 
Morison['M5N6MByi']           = False     # (N); ; 
Morison['M5N7MByi']           = False     # (N); ; 
Morison['M5N8MByi']           = False     # (N); ; 
Morison['M5N9MByi']           = False     # (N); ; 
Morison['M6N1MByi']           = False     # (N); ; 
Morison['M6N2MByi']           = False     # (N); ; 
Morison['M6N3MByi']           = False     # (N); ; 
Morison['M6N4MByi']           = False     # (N); ; 
Morison['M6N5MByi']           = False     # (N); ; 
Morison['M6N6MByi']           = False     # (N); ; 
Morison['M6N7MByi']           = False     # (N); ; 
Morison['M6N8MByi']           = False     # (N); ; 
Morison['M6N9MByi']           = False     # (N); ; 
Morison['M7N1MByi']           = False     # (N); ; 
Morison['M7N2MByi']           = False     # (N); ; 
Morison['M7N3MByi']           = False     # (N); ; 
Morison['M7N4MByi']           = False     # (N); ; 
Morison['M7N5MByi']           = False     # (N); ; 
Morison['M7N6MByi']           = False     # (N); ; 
Morison['M7N7MByi']           = False     # (N); ; 
Morison['M7N8MByi']           = False     # (N); ; 
Morison['M7N9MByi']           = False     # (N); ; 
Morison['M8N1MByi']           = False     # (N); ; 
Morison['M8N2MByi']           = False     # (N); ; 
Morison['M8N3MByi']           = False     # (N); ; 
Morison['M8N4MByi']           = False     # (N); ; 
Morison['M8N5MByi']           = False     # (N); ; 
Morison['M8N6MByi']           = False     # (N); ; 
Morison['M8N7MByi']           = False     # (N); ; 
Morison['M8N8MByi']           = False     # (N); ; 
Morison['M8N9MByi']           = False     # (N); ; 
Morison['M9N1MByi']           = False     # (N); ; 
Morison['M9N2MByi']           = False     # (N); ; 
Morison['M9N3MByi']           = False     # (N); ; 
Morison['M9N4MByi']           = False     # (N); ; 
Morison['M9N5MByi']           = False     # (N); ; 
Morison['M9N6MByi']           = False     # (N); ; 
Morison['M9N7MByi']           = False     # (N); ; 
Morison['M9N8MByi']           = False     # (N); ; 
Morison['M9N9MByi']           = False     # (N); ; 
Morison['M1N1MBzi']           = False     # (N); z-component of the distributed bouyancy moment expressed in the inertial coordinate system; 
Morison['M1N2MBzi']           = False     # (N); ; 
Morison['M1N3MBzi']           = False     # (N); ; 
Morison['M1N4MBzi']           = False     # (N); ; 
Morison['M1N5MBzi']           = False     # (N); ; 
Morison['M1N6MBzi']           = False     # (N); ; 
Morison['M1N7MBzi']           = False     # (N); ; 
Morison['M1N8MBzi']           = False     # (N); ; 
Morison['M1N9MBzi']           = False     # (N); ; 
Morison['M2N1MBzi']           = False     # (N); ; 
Morison['M2N2MBzi']           = False     # (N); ; 
Morison['M2N3MBzi']           = False     # (N); ; 
Morison['M2N4MBzi']           = False     # (N); ; 
Morison['M2N5MBzi']           = False     # (N); ; 
Morison['M2N6MBzi']           = False     # (N); ; 
Morison['M2N7MBzi']           = False     # (N); ; 
Morison['M2N8MBzi']           = False     # (N); ; 
Morison['M2N9MBzi']           = False     # (N); ; 
Morison['M3N1MBzi']           = False     # (N); ; 
Morison['M3N2MBzi']           = False     # (N); ; 
Morison['M3N3MBzi']           = False     # (N); ; 
Morison['M3N4MBzi']           = False     # (N); ; 
Morison['M3N5MBzi']           = False     # (N); ; 
Morison['M3N6MBzi']           = False     # (N); ; 
Morison['M3N7MBzi']           = False     # (N); ; 
Morison['M3N8MBzi']           = False     # (N); ; 
Morison['M3N9MBzi']           = False     # (N); ; 
Morison['M4N1MBzi']           = False     # (N); ; 
Morison['M4N2MBzi']           = False     # (N); ; 
Morison['M4N3MBzi']           = False     # (N); ; 
Morison['M4N4MBzi']           = False     # (N); ; 
Morison['M4N5MBzi']           = False     # (N); ; 
Morison['M4N6MBzi']           = False     # (N); ; 
Morison['M4N7MBzi']           = False     # (N); ; 
Morison['M4N8MBzi']           = False     # (N); ; 
Morison['M4N9MBzi']           = False     # (N); ; 
Morison['M5N1MBzi']           = False     # (N); ; 
Morison['M5N2MBzi']           = False     # (N); ; 
Morison['M5N3MBzi']           = False     # (N); ; 
Morison['M5N4MBzi']           = False     # (N); ; 
Morison['M5N5MBzi']           = False     # (N); ; 
Morison['M5N6MBzi']           = False     # (N); ; 
Morison['M5N7MBzi']           = False     # (N); ; 
Morison['M5N8MBzi']           = False     # (N); ; 
Morison['M5N9MBzi']           = False     # (N); ; 
Morison['M6N1MBzi']           = False     # (N); ; 
Morison['M6N2MBzi']           = False     # (N); ; 
Morison['M6N3MBzi']           = False     # (N); ; 
Morison['M6N4MBzi']           = False     # (N); ; 
Morison['M6N5MBzi']           = False     # (N); ; 
Morison['M6N6MBzi']           = False     # (N); ; 
Morison['M6N7MBzi']           = False     # (N); ; 
Morison['M6N8MBzi']           = False     # (N); ; 
Morison['M6N9MBzi']           = False     # (N); ; 
Morison['M7N1MBzi']           = False     # (N); ; 
Morison['M7N2MBzi']           = False     # (N); ; 
Morison['M7N3MBzi']           = False     # (N); ; 
Morison['M7N4MBzi']           = False     # (N); ; 
Morison['M7N5MBzi']           = False     # (N); ; 
Morison['M7N6MBzi']           = False     # (N); ; 
Morison['M7N7MBzi']           = False     # (N); ; 
Morison['M7N8MBzi']           = False     # (N); ; 
Morison['M7N9MBzi']           = False     # (N); ; 
Morison['M8N1MBzi']           = False     # (N); ; 
Morison['M8N2MBzi']           = False     # (N); ; 
Morison['M8N3MBzi']           = False     # (N); ; 
Morison['M8N4MBzi']           = False     # (N); ; 
Morison['M8N5MBzi']           = False     # (N); ; 
Morison['M8N6MBzi']           = False     # (N); ; 
Morison['M8N7MBzi']           = False     # (N); ; 
Morison['M8N8MBzi']           = False     # (N); ; 
Morison['M8N9MBzi']           = False     # (N); ; 
Morison['M9N1MBzi']           = False     # (N); ; 
Morison['M9N2MBzi']           = False     # (N); ; 
Morison['M9N3MBzi']           = False     # (N); ; 
Morison['M9N4MBzi']           = False     # (N); ; 
Morison['M9N5MBzi']           = False     # (N); ; 
Morison['M9N6MBzi']           = False     # (N); ; 
Morison['M9N7MBzi']           = False     # (N); ; 
Morison['M9N8MBzi']           = False     # (N); ; 
Morison['M9N9MBzi']           = False     # (N); ; 
Morison['M1N1FBFxi']          = False     # (N/m); x-component of the distributed filled fluid bouyancy force expressed in the inertial coordinate system; 
Morison['M1N2FBFxi']          = False     # (N/m); ; 
Morison['M1N3FBFxi']          = False     # (N/m); ; 
Morison['M1N4FBFxi']          = False     # (N/m); ; 
Morison['M1N5FBFxi']          = False     # (N/m); ; 
Morison['M1N6FBFxi']          = False     # (N/m); ; 
Morison['M1N7FBFxi']          = False     # (N/m); ; 
Morison['M1N8FBFxi']          = False     # (N/m); ; 
Morison['M1N9FBFxi']          = False     # (N/m); ; 
Morison['M2N1FBFxi']          = False     # (N/m); ; 
Morison['M2N2FBFxi']          = False     # (N/m); ; 
Morison['M2N3FBFxi']          = False     # (N/m); ; 
Morison['M2N4FBFxi']          = False     # (N/m); ; 
Morison['M2N5FBFxi']          = False     # (N/m); ; 
Morison['M2N6FBFxi']          = False     # (N/m); ; 
Morison['M2N7FBFxi']          = False     # (N/m); ; 
Morison['M2N8FBFxi']          = False     # (N/m); ; 
Morison['M2N9FBFxi']          = False     # (N/m); ; 
Morison['M3N1FBFxi']          = False     # (N/m); ; 
Morison['M3N2FBFxi']          = False     # (N/m); ; 
Morison['M3N3FBFxi']          = False     # (N/m); ; 
Morison['M3N4FBFxi']          = False     # (N/m); ; 
Morison['M3N5FBFxi']          = False     # (N/m); ; 
Morison['M3N6FBFxi']          = False     # (N/m); ; 
Morison['M3N7FBFxi']          = False     # (N/m); ; 
Morison['M3N8FBFxi']          = False     # (N/m); ; 
Morison['M3N9FBFxi']          = False     # (N/m); ; 
Morison['M4N1FBFxi']          = False     # (N/m); ; 
Morison['M4N2FBFxi']          = False     # (N/m); ; 
Morison['M4N3FBFxi']          = False     # (N/m); ; 
Morison['M4N4FBFxi']          = False     # (N/m); ; 
Morison['M4N5FBFxi']          = False     # (N/m); ; 
Morison['M4N6FBFxi']          = False     # (N/m); ; 
Morison['M4N7FBFxi']          = False     # (N/m); ; 
Morison['M4N8FBFxi']          = False     # (N/m); ; 
Morison['M4N9FBFxi']          = False     # (N/m); ; 
Morison['M5N1FBFxi']          = False     # (N/m); ; 
Morison['M5N2FBFxi']          = False     # (N/m); ; 
Morison['M5N3FBFxi']          = False     # (N/m); ; 
Morison['M5N4FBFxi']          = False     # (N/m); ; 
Morison['M5N5FBFxi']          = False     # (N/m); ; 
Morison['M5N6FBFxi']          = False     # (N/m); ; 
Morison['M5N7FBFxi']          = False     # (N/m); ; 
Morison['M5N8FBFxi']          = False     # (N/m); ; 
Morison['M5N9FBFxi']          = False     # (N/m); ; 
Morison['M6N1FBFxi']          = False     # (N/m); ; 
Morison['M6N2FBFxi']          = False     # (N/m); ; 
Morison['M6N3FBFxi']          = False     # (N/m); ; 
Morison['M6N4FBFxi']          = False     # (N/m); ; 
Morison['M6N5FBFxi']          = False     # (N/m); ; 
Morison['M6N6FBFxi']          = False     # (N/m); ; 
Morison['M6N7FBFxi']          = False     # (N/m); ; 
Morison['M6N8FBFxi']          = False     # (N/m); ; 
Morison['M6N9FBFxi']          = False     # (N/m); ; 
Morison['M7N1FBFxi']          = False     # (N/m); ; 
Morison['M7N2FBFxi']          = False     # (N/m); ; 
Morison['M7N3FBFxi']          = False     # (N/m); ; 
Morison['M7N4FBFxi']          = False     # (N/m); ; 
Morison['M7N5FBFxi']          = False     # (N/m); ; 
Morison['M7N6FBFxi']          = False     # (N/m); ; 
Morison['M7N7FBFxi']          = False     # (N/m); ; 
Morison['M7N8FBFxi']          = False     # (N/m); ; 
Morison['M7N9FBFxi']          = False     # (N/m); ; 
Morison['M8N1FBFxi']          = False     # (N/m); ; 
Morison['M8N2FBFxi']          = False     # (N/m); ; 
Morison['M8N3FBFxi']          = False     # (N/m); ; 
Morison['M8N4FBFxi']          = False     # (N/m); ; 
Morison['M8N5FBFxi']          = False     # (N/m); ; 
Morison['M8N6FBFxi']          = False     # (N/m); ; 
Morison['M8N7FBFxi']          = False     # (N/m); ; 
Morison['M8N8FBFxi']          = False     # (N/m); ; 
Morison['M8N9FBFxi']          = False     # (N/m); ; 
Morison['M9N1FBFxi']          = False     # (N/m); ; 
Morison['M9N2FBFxi']          = False     # (N/m); ; 
Morison['M9N3FBFxi']          = False     # (N/m); ; 
Morison['M9N4FBFxi']          = False     # (N/m); ; 
Morison['M9N5FBFxi']          = False     # (N/m); ; 
Morison['M9N6FBFxi']          = False     # (N/m); ; 
Morison['M9N7FBFxi']          = False     # (N/m); ; 
Morison['M9N8FBFxi']          = False     # (N/m); ; 
Morison['M9N9FBFxi']          = False     # (N/m); ; 
Morison['M1N1FBFyi']          = False     # (N/m); y-component of the distributed filled fluid bouyancy force expressed in the inertial coordinate system; 
Morison['M1N2FBFyi']          = False     # (N/m); ; 
Morison['M1N3FBFyi']          = False     # (N/m); ; 
Morison['M1N4FBFyi']          = False     # (N/m); ; 
Morison['M1N5FBFyi']          = False     # (N/m); ; 
Morison['M1N6FBFyi']          = False     # (N/m); ; 
Morison['M1N7FBFyi']          = False     # (N/m); ; 
Morison['M1N8FBFyi']          = False     # (N/m); ; 
Morison['M1N9FBFyi']          = False     # (N/m); ; 
Morison['M2N1FBFyi']          = False     # (N/m); ; 
Morison['M2N2FBFyi']          = False     # (N/m); ; 
Morison['M2N3FBFyi']          = False     # (N/m); ; 
Morison['M2N4FBFyi']          = False     # (N/m); ; 
Morison['M2N5FBFyi']          = False     # (N/m); ; 
Morison['M2N6FBFyi']          = False     # (N/m); ; 
Morison['M2N7FBFyi']          = False     # (N/m); ; 
Morison['M2N8FBFyi']          = False     # (N/m); ; 
Morison['M2N9FBFyi']          = False     # (N/m); ; 
Morison['M3N1FBFyi']          = False     # (N/m); ; 
Morison['M3N2FBFyi']          = False     # (N/m); ; 
Morison['M3N3FBFyi']          = False     # (N/m); ; 
Morison['M3N4FBFyi']          = False     # (N/m); ; 
Morison['M3N5FBFyi']          = False     # (N/m); ; 
Morison['M3N6FBFyi']          = False     # (N/m); ; 
Morison['M3N7FBFyi']          = False     # (N/m); ; 
Morison['M3N8FBFyi']          = False     # (N/m); ; 
Morison['M3N9FBFyi']          = False     # (N/m); ; 
Morison['M4N1FBFyi']          = False     # (N/m); ; 
Morison['M4N2FBFyi']          = False     # (N/m); ; 
Morison['M4N3FBFyi']          = False     # (N/m); ; 
Morison['M4N4FBFyi']          = False     # (N/m); ; 
Morison['M4N5FBFyi']          = False     # (N/m); ; 
Morison['M4N6FBFyi']          = False     # (N/m); ; 
Morison['M4N7FBFyi']          = False     # (N/m); ; 
Morison['M4N8FBFyi']          = False     # (N/m); ; 
Morison['M4N9FBFyi']          = False     # (N/m); ; 
Morison['M5N1FBFyi']          = False     # (N/m); ; 
Morison['M5N2FBFyi']          = False     # (N/m); ; 
Morison['M5N3FBFyi']          = False     # (N/m); ; 
Morison['M5N4FBFyi']          = False     # (N/m); ; 
Morison['M5N5FBFyi']          = False     # (N/m); ; 
Morison['M5N6FBFyi']          = False     # (N/m); ; 
Morison['M5N7FBFyi']          = False     # (N/m); ; 
Morison['M5N8FBFyi']          = False     # (N/m); ; 
Morison['M5N9FBFyi']          = False     # (N/m); ; 
Morison['M6N1FBFyi']          = False     # (N/m); ; 
Morison['M6N2FBFyi']          = False     # (N/m); ; 
Morison['M6N3FBFyi']          = False     # (N/m); ; 
Morison['M6N4FBFyi']          = False     # (N/m); ; 
Morison['M6N5FBFyi']          = False     # (N/m); ; 
Morison['M6N6FBFyi']          = False     # (N/m); ; 
Morison['M6N7FBFyi']          = False     # (N/m); ; 
Morison['M6N8FBFyi']          = False     # (N/m); ; 
Morison['M6N9FBFyi']          = False     # (N/m); ; 
Morison['M7N1FBFyi']          = False     # (N/m); ; 
Morison['M7N2FBFyi']          = False     # (N/m); ; 
Morison['M7N3FBFyi']          = False     # (N/m); ; 
Morison['M7N4FBFyi']          = False     # (N/m); ; 
Morison['M7N5FBFyi']          = False     # (N/m); ; 
Morison['M7N6FBFyi']          = False     # (N/m); ; 
Morison['M7N7FBFyi']          = False     # (N/m); ; 
Morison['M7N8FBFyi']          = False     # (N/m); ; 
Morison['M7N9FBFyi']          = False     # (N/m); ; 
Morison['M8N1FBFyi']          = False     # (N/m); ; 
Morison['M8N2FBFyi']          = False     # (N/m); ; 
Morison['M8N3FBFyi']          = False     # (N/m); ; 
Morison['M8N4FBFyi']          = False     # (N/m); ; 
Morison['M8N5FBFyi']          = False     # (N/m); ; 
Morison['M8N6FBFyi']          = False     # (N/m); ; 
Morison['M8N7FBFyi']          = False     # (N/m); ; 
Morison['M8N8FBFyi']          = False     # (N/m); ; 
Morison['M8N9FBFyi']          = False     # (N/m); ; 
Morison['M9N1FBFyi']          = False     # (N/m); ; 
Morison['M9N2FBFyi']          = False     # (N/m); ; 
Morison['M9N3FBFyi']          = False     # (N/m); ; 
Morison['M9N4FBFyi']          = False     # (N/m); ; 
Morison['M9N5FBFyi']          = False     # (N/m); ; 
Morison['M9N6FBFyi']          = False     # (N/m); ; 
Morison['M9N7FBFyi']          = False     # (N/m); ; 
Morison['M9N8FBFyi']          = False     # (N/m); ; 
Morison['M9N9FBFyi']          = False     # (N/m); ; 
Morison['M1N1FBFzi']          = False     # (N/m); z-component of the distributed filled fluid bouyancy force expressed in the inertial coordinate system; 
Morison['M1N2FBFzi']          = False     # (N/m); ; 
Morison['M1N3FBFzi']          = False     # (N/m); ; 
Morison['M1N4FBFzi']          = False     # (N/m); ; 
Morison['M1N5FBFzi']          = False     # (N/m); ; 
Morison['M1N6FBFzi']          = False     # (N/m); ; 
Morison['M1N7FBFzi']          = False     # (N/m); ; 
Morison['M1N8FBFzi']          = False     # (N/m); ; 
Morison['M1N9FBFzi']          = False     # (N/m); ; 
Morison['M2N1FBFzi']          = False     # (N/m); ; 
Morison['M2N2FBFzi']          = False     # (N/m); ; 
Morison['M2N3FBFzi']          = False     # (N/m); ; 
Morison['M2N4FBFzi']          = False     # (N/m); ; 
Morison['M2N5FBFzi']          = False     # (N/m); ; 
Morison['M2N6FBFzi']          = False     # (N/m); ; 
Morison['M2N7FBFzi']          = False     # (N/m); ; 
Morison['M2N8FBFzi']          = False     # (N/m); ; 
Morison['M2N9FBFzi']          = False     # (N/m); ; 
Morison['M3N1FBFzi']          = False     # (N/m); ; 
Morison['M3N2FBFzi']          = False     # (N/m); ; 
Morison['M3N3FBFzi']          = False     # (N/m); ; 
Morison['M3N4FBFzi']          = False     # (N/m); ; 
Morison['M3N5FBFzi']          = False     # (N/m); ; 
Morison['M3N6FBFzi']          = False     # (N/m); ; 
Morison['M3N7FBFzi']          = False     # (N/m); ; 
Morison['M3N8FBFzi']          = False     # (N/m); ; 
Morison['M3N9FBFzi']          = False     # (N/m); ; 
Morison['M4N1FBFzi']          = False     # (N/m); ; 
Morison['M4N2FBFzi']          = False     # (N/m); ; 
Morison['M4N3FBFzi']          = False     # (N/m); ; 
Morison['M4N4FBFzi']          = False     # (N/m); ; 
Morison['M4N5FBFzi']          = False     # (N/m); ; 
Morison['M4N6FBFzi']          = False     # (N/m); ; 
Morison['M4N7FBFzi']          = False     # (N/m); ; 
Morison['M4N8FBFzi']          = False     # (N/m); ; 
Morison['M4N9FBFzi']          = False     # (N/m); ; 
Morison['M5N1FBFzi']          = False     # (N/m); ; 
Morison['M5N2FBFzi']          = False     # (N/m); ; 
Morison['M5N3FBFzi']          = False     # (N/m); ; 
Morison['M5N4FBFzi']          = False     # (N/m); ; 
Morison['M5N5FBFzi']          = False     # (N/m); ; 
Morison['M5N6FBFzi']          = False     # (N/m); ; 
Morison['M5N7FBFzi']          = False     # (N/m); ; 
Morison['M5N8FBFzi']          = False     # (N/m); ; 
Morison['M5N9FBFzi']          = False     # (N/m); ; 
Morison['M6N1FBFzi']          = False     # (N/m); ; 
Morison['M6N2FBFzi']          = False     # (N/m); ; 
Morison['M6N3FBFzi']          = False     # (N/m); ; 
Morison['M6N4FBFzi']          = False     # (N/m); ; 
Morison['M6N5FBFzi']          = False     # (N/m); ; 
Morison['M6N6FBFzi']          = False     # (N/m); ; 
Morison['M6N7FBFzi']          = False     # (N/m); ; 
Morison['M6N8FBFzi']          = False     # (N/m); ; 
Morison['M6N9FBFzi']          = False     # (N/m); ; 
Morison['M7N1FBFzi']          = False     # (N/m); ; 
Morison['M7N2FBFzi']          = False     # (N/m); ; 
Morison['M7N3FBFzi']          = False     # (N/m); ; 
Morison['M7N4FBFzi']          = False     # (N/m); ; 
Morison['M7N5FBFzi']          = False     # (N/m); ; 
Morison['M7N6FBFzi']          = False     # (N/m); ; 
Morison['M7N7FBFzi']          = False     # (N/m); ; 
Morison['M7N8FBFzi']          = False     # (N/m); ; 
Morison['M7N9FBFzi']          = False     # (N/m); ; 
Morison['M8N1FBFzi']          = False     # (N/m); ; 
Morison['M8N2FBFzi']          = False     # (N/m); ; 
Morison['M8N3FBFzi']          = False     # (N/m); ; 
Morison['M8N4FBFzi']          = False     # (N/m); ; 
Morison['M8N5FBFzi']          = False     # (N/m); ; 
Morison['M8N6FBFzi']          = False     # (N/m); ; 
Morison['M8N7FBFzi']          = False     # (N/m); ; 
Morison['M8N8FBFzi']          = False     # (N/m); ; 
Morison['M8N9FBFzi']          = False     # (N/m); ; 
Morison['M9N1FBFzi']          = False     # (N/m); ; 
Morison['M9N2FBFzi']          = False     # (N/m); ; 
Morison['M9N3FBFzi']          = False     # (N/m); ; 
Morison['M9N4FBFzi']          = False     # (N/m); ; 
Morison['M9N5FBFzi']          = False     # (N/m); ; 
Morison['M9N6FBFzi']          = False     # (N/m); ; 
Morison['M9N7FBFzi']          = False     # (N/m); ; 
Morison['M9N8FBFzi']          = False     # (N/m); ; 
Morison['M9N9FBFzi']          = False     # (N/m); ; 
Morison['M1N1MBFxi']          = False     # (N); x-component of the distributed filled fluid bouyancy moment expressed in the inertial coordinate system; 
Morison['M1N2MBFxi']          = False     # (N); ; 
Morison['M1N3MBFxi']          = False     # (N); ; 
Morison['M1N4MBFxi']          = False     # (N); ; 
Morison['M1N5MBFxi']          = False     # (N); ; 
Morison['M1N6MBFxi']          = False     # (N); ; 
Morison['M1N7MBFxi']          = False     # (N); ; 
Morison['M1N8MBFxi']          = False     # (N); ; 
Morison['M1N9MBFxi']          = False     # (N); ; 
Morison['M2N1MBFxi']          = False     # (N); ; 
Morison['M2N2MBFxi']          = False     # (N); ; 
Morison['M2N3MBFxi']          = False     # (N); ; 
Morison['M2N4MBFxi']          = False     # (N); ; 
Morison['M2N5MBFxi']          = False     # (N); ; 
Morison['M2N6MBFxi']          = False     # (N); ; 
Morison['M2N7MBFxi']          = False     # (N); ; 
Morison['M2N8MBFxi']          = False     # (N); ; 
Morison['M2N9MBFxi']          = False     # (N); ; 
Morison['M3N1MBFxi']          = False     # (N); ; 
Morison['M3N2MBFxi']          = False     # (N); ; 
Morison['M3N3MBFxi']          = False     # (N); ; 
Morison['M3N4MBFxi']          = False     # (N); ; 
Morison['M3N5MBFxi']          = False     # (N); ; 
Morison['M3N6MBFxi']          = False     # (N); ; 
Morison['M3N7MBFxi']          = False     # (N); ; 
Morison['M3N8MBFxi']          = False     # (N); ; 
Morison['M3N9MBFxi']          = False     # (N); ; 
Morison['M4N1MBFxi']          = False     # (N); ; 
Morison['M4N2MBFxi']          = False     # (N); ; 
Morison['M4N3MBFxi']          = False     # (N); ; 
Morison['M4N4MBFxi']          = False     # (N); ; 
Morison['M4N5MBFxi']          = False     # (N); ; 
Morison['M4N6MBFxi']          = False     # (N); ; 
Morison['M4N7MBFxi']          = False     # (N); ; 
Morison['M4N8MBFxi']          = False     # (N); ; 
Morison['M4N9MBFxi']          = False     # (N); ; 
Morison['M5N1MBFxi']          = False     # (N); ; 
Morison['M5N2MBFxi']          = False     # (N); ; 
Morison['M5N3MBFxi']          = False     # (N); ; 
Morison['M5N4MBFxi']          = False     # (N); ; 
Morison['M5N5MBFxi']          = False     # (N); ; 
Morison['M5N6MBFxi']          = False     # (N); ; 
Morison['M5N7MBFxi']          = False     # (N); ; 
Morison['M5N8MBFxi']          = False     # (N); ; 
Morison['M5N9MBFxi']          = False     # (N); ; 
Morison['M6N1MBFxi']          = False     # (N); ; 
Morison['M6N2MBFxi']          = False     # (N); ; 
Morison['M6N3MBFxi']          = False     # (N); ; 
Morison['M6N4MBFxi']          = False     # (N); ; 
Morison['M6N5MBFxi']          = False     # (N); ; 
Morison['M6N6MBFxi']          = False     # (N); ; 
Morison['M6N7MBFxi']          = False     # (N); ; 
Morison['M6N8MBFxi']          = False     # (N); ; 
Morison['M6N9MBFxi']          = False     # (N); ; 
Morison['M7N1MBFxi']          = False     # (N); ; 
Morison['M7N2MBFxi']          = False     # (N); ; 
Morison['M7N3MBFxi']          = False     # (N); ; 
Morison['M7N4MBFxi']          = False     # (N); ; 
Morison['M7N5MBFxi']          = False     # (N); ; 
Morison['M7N6MBFxi']          = False     # (N); ; 
Morison['M7N7MBFxi']          = False     # (N); ; 
Morison['M7N8MBFxi']          = False     # (N); ; 
Morison['M7N9MBFxi']          = False     # (N); ; 
Morison['M8N1MBFxi']          = False     # (N); ; 
Morison['M8N2MBFxi']          = False     # (N); ; 
Morison['M8N3MBFxi']          = False     # (N); ; 
Morison['M8N4MBFxi']          = False     # (N); ; 
Morison['M8N5MBFxi']          = False     # (N); ; 
Morison['M8N6MBFxi']          = False     # (N); ; 
Morison['M8N7MBFxi']          = False     # (N); ; 
Morison['M8N8MBFxi']          = False     # (N); ; 
Morison['M8N9MBFxi']          = False     # (N); ; 
Morison['M9N1MBFxi']          = False     # (N); ; 
Morison['M9N2MBFxi']          = False     # (N); ; 
Morison['M9N3MBFxi']          = False     # (N); ; 
Morison['M9N4MBFxi']          = False     # (N); ; 
Morison['M9N5MBFxi']          = False     # (N); ; 
Morison['M9N6MBFxi']          = False     # (N); ; 
Morison['M9N7MBFxi']          = False     # (N); ; 
Morison['M9N8MBFxi']          = False     # (N); ; 
Morison['M9N9MBFxi']          = False     # (N); ; 
Morison['M1N1MBFyi']          = False     # (N); y-component of the distributed filled fluid bouyancy moment expressed in the inertial coordinate system; 
Morison['M1N2MBFyi']          = False     # (N); ; 
Morison['M1N3MBFyi']          = False     # (N); ; 
Morison['M1N4MBFyi']          = False     # (N); ; 
Morison['M1N5MBFyi']          = False     # (N); ; 
Morison['M1N6MBFyi']          = False     # (N); ; 
Morison['M1N7MBFyi']          = False     # (N); ; 
Morison['M1N8MBFyi']          = False     # (N); ; 
Morison['M1N9MBFyi']          = False     # (N); ; 
Morison['M2N1MBFyi']          = False     # (N); ; 
Morison['M2N2MBFyi']          = False     # (N); ; 
Morison['M2N3MBFyi']          = False     # (N); ; 
Morison['M2N4MBFyi']          = False     # (N); ; 
Morison['M2N5MBFyi']          = False     # (N); ; 
Morison['M2N6MBFyi']          = False     # (N); ; 
Morison['M2N7MBFyi']          = False     # (N); ; 
Morison['M2N8MBFyi']          = False     # (N); ; 
Morison['M2N9MBFyi']          = False     # (N); ; 
Morison['M3N1MBFyi']          = False     # (N); ; 
Morison['M3N2MBFyi']          = False     # (N); ; 
Morison['M3N3MBFyi']          = False     # (N); ; 
Morison['M3N4MBFyi']          = False     # (N); ; 
Morison['M3N5MBFyi']          = False     # (N); ; 
Morison['M3N6MBFyi']          = False     # (N); ; 
Morison['M3N7MBFyi']          = False     # (N); ; 
Morison['M3N8MBFyi']          = False     # (N); ; 
Morison['M3N9MBFyi']          = False     # (N); ; 
Morison['M4N1MBFyi']          = False     # (N); ; 
Morison['M4N2MBFyi']          = False     # (N); ; 
Morison['M4N3MBFyi']          = False     # (N); ; 
Morison['M4N4MBFyi']          = False     # (N); ; 
Morison['M4N5MBFyi']          = False     # (N); ; 
Morison['M4N6MBFyi']          = False     # (N); ; 
Morison['M4N7MBFyi']          = False     # (N); ; 
Morison['M4N8MBFyi']          = False     # (N); ; 
Morison['M4N9MBFyi']          = False     # (N); ; 
Morison['M5N1MBFyi']          = False     # (N); ; 
Morison['M5N2MBFyi']          = False     # (N); ; 
Morison['M5N3MBFyi']          = False     # (N); ; 
Morison['M5N4MBFyi']          = False     # (N); ; 
Morison['M5N5MBFyi']          = False     # (N); ; 
Morison['M5N6MBFyi']          = False     # (N); ; 
Morison['M5N7MBFyi']          = False     # (N); ; 
Morison['M5N8MBFyi']          = False     # (N); ; 
Morison['M5N9MBFyi']          = False     # (N); ; 
Morison['M6N1MBFyi']          = False     # (N); ; 
Morison['M6N2MBFyi']          = False     # (N); ; 
Morison['M6N3MBFyi']          = False     # (N); ; 
Morison['M6N4MBFyi']          = False     # (N); ; 
Morison['M6N5MBFyi']          = False     # (N); ; 
Morison['M6N6MBFyi']          = False     # (N); ; 
Morison['M6N7MBFyi']          = False     # (N); ; 
Morison['M6N8MBFyi']          = False     # (N); ; 
Morison['M6N9MBFyi']          = False     # (N); ; 
Morison['M7N1MBFyi']          = False     # (N); ; 
Morison['M7N2MBFyi']          = False     # (N); ; 
Morison['M7N3MBFyi']          = False     # (N); ; 
Morison['M7N4MBFyi']          = False     # (N); ; 
Morison['M7N5MBFyi']          = False     # (N); ; 
Morison['M7N6MBFyi']          = False     # (N); ; 
Morison['M7N7MBFyi']          = False     # (N); ; 
Morison['M7N8MBFyi']          = False     # (N); ; 
Morison['M7N9MBFyi']          = False     # (N); ; 
Morison['M8N1MBFyi']          = False     # (N); ; 
Morison['M8N2MBFyi']          = False     # (N); ; 
Morison['M8N3MBFyi']          = False     # (N); ; 
Morison['M8N4MBFyi']          = False     # (N); ; 
Morison['M8N5MBFyi']          = False     # (N); ; 
Morison['M8N6MBFyi']          = False     # (N); ; 
Morison['M8N7MBFyi']          = False     # (N); ; 
Morison['M8N8MBFyi']          = False     # (N); ; 
Morison['M8N9MBFyi']          = False     # (N); ; 
Morison['M9N1MBFyi']          = False     # (N); ; 
Morison['M9N2MBFyi']          = False     # (N); ; 
Morison['M9N3MBFyi']          = False     # (N); ; 
Morison['M9N4MBFyi']          = False     # (N); ; 
Morison['M9N5MBFyi']          = False     # (N); ; 
Morison['M9N6MBFyi']          = False     # (N); ; 
Morison['M9N7MBFyi']          = False     # (N); ; 
Morison['M9N8MBFyi']          = False     # (N); ; 
Morison['M9N9MBFyi']          = False     # (N); ; 
Morison['M1N1MBFzi']          = False     # (N); z-component of the distributed filled fluid bouyancy moment expressed in the inertial coordinate system; 
Morison['M1N2MBFzi']          = False     # (N); ; 
Morison['M1N3MBFzi']          = False     # (N); ; 
Morison['M1N4MBFzi']          = False     # (N); ; 
Morison['M1N5MBFzi']          = False     # (N); ; 
Morison['M1N6MBFzi']          = False     # (N); ; 
Morison['M1N7MBFzi']          = False     # (N); ; 
Morison['M1N8MBFzi']          = False     # (N); ; 
Morison['M1N9MBFzi']          = False     # (N); ; 
Morison['M2N1MBFzi']          = False     # (N); ; 
Morison['M2N2MBFzi']          = False     # (N); ; 
Morison['M2N3MBFzi']          = False     # (N); ; 
Morison['M2N4MBFzi']          = False     # (N); ; 
Morison['M2N5MBFzi']          = False     # (N); ; 
Morison['M2N6MBFzi']          = False     # (N); ; 
Morison['M2N7MBFzi']          = False     # (N); ; 
Morison['M2N8MBFzi']          = False     # (N); ; 
Morison['M2N9MBFzi']          = False     # (N); ; 
Morison['M3N1MBFzi']          = False     # (N); ; 
Morison['M3N2MBFzi']          = False     # (N); ; 
Morison['M3N3MBFzi']          = False     # (N); ; 
Morison['M3N4MBFzi']          = False     # (N); ; 
Morison['M3N5MBFzi']          = False     # (N); ; 
Morison['M3N6MBFzi']          = False     # (N); ; 
Morison['M3N7MBFzi']          = False     # (N); ; 
Morison['M3N8MBFzi']          = False     # (N); ; 
Morison['M3N9MBFzi']          = False     # (N); ; 
Morison['M4N1MBFzi']          = False     # (N); ; 
Morison['M4N2MBFzi']          = False     # (N); ; 
Morison['M4N3MBFzi']          = False     # (N); ; 
Morison['M4N4MBFzi']          = False     # (N); ; 
Morison['M4N5MBFzi']          = False     # (N); ; 
Morison['M4N6MBFzi']          = False     # (N); ; 
Morison['M4N7MBFzi']          = False     # (N); ; 
Morison['M4N8MBFzi']          = False     # (N); ; 
Morison['M4N9MBFzi']          = False     # (N); ; 
Morison['M5N1MBFzi']          = False     # (N); ; 
Morison['M5N2MBFzi']          = False     # (N); ; 
Morison['M5N3MBFzi']          = False     # (N); ; 
Morison['M5N4MBFzi']          = False     # (N); ; 
Morison['M5N5MBFzi']          = False     # (N); ; 
Morison['M5N6MBFzi']          = False     # (N); ; 
Morison['M5N7MBFzi']          = False     # (N); ; 
Morison['M5N8MBFzi']          = False     # (N); ; 
Morison['M5N9MBFzi']          = False     # (N); ; 
Morison['M6N1MBFzi']          = False     # (N); ; 
Morison['M6N2MBFzi']          = False     # (N); ; 
Morison['M6N3MBFzi']          = False     # (N); ; 
Morison['M6N4MBFzi']          = False     # (N); ; 
Morison['M6N5MBFzi']          = False     # (N); ; 
Morison['M6N6MBFzi']          = False     # (N); ; 
Morison['M6N7MBFzi']          = False     # (N); ; 
Morison['M6N8MBFzi']          = False     # (N); ; 
Morison['M6N9MBFzi']          = False     # (N); ; 
Morison['M7N1MBFzi']          = False     # (N); ; 
Morison['M7N2MBFzi']          = False     # (N); ; 
Morison['M7N3MBFzi']          = False     # (N); ; 
Morison['M7N4MBFzi']          = False     # (N); ; 
Morison['M7N5MBFzi']          = False     # (N); ; 
Morison['M7N6MBFzi']          = False     # (N); ; 
Morison['M7N7MBFzi']          = False     # (N); ; 
Morison['M7N8MBFzi']          = False     # (N); ; 
Morison['M7N9MBFzi']          = False     # (N); ; 
Morison['M8N1MBFzi']          = False     # (N); ; 
Morison['M8N2MBFzi']          = False     # (N); ; 
Morison['M8N3MBFzi']          = False     # (N); ; 
Morison['M8N4MBFzi']          = False     # (N); ; 
Morison['M8N5MBFzi']          = False     # (N); ; 
Morison['M8N6MBFzi']          = False     # (N); ; 
Morison['M8N7MBFzi']          = False     # (N); ; 
Morison['M8N8MBFzi']          = False     # (N); ; 
Morison['M8N9MBFzi']          = False     # (N); ; 
Morison['M9N1MBFzi']          = False     # (N); ; 
Morison['M9N2MBFzi']          = False     # (N); ; 
Morison['M9N3MBFzi']          = False     # (N); ; 
Morison['M9N4MBFzi']          = False     # (N); ; 
Morison['M9N5MBFzi']          = False     # (N); ; 
Morison['M9N6MBFzi']          = False     # (N); ; 
Morison['M9N7MBFzi']          = False     # (N); ; 
Morison['M9N8MBFzi']          = False     # (N); ; 
Morison['M9N9MBFzi']          = False     # (N); ; 
Morison['M1N1FMGxi']          = False     # (N/m); x-component of the distributed marine growth weight expressed in the inertial coordinate system; 
Morison['M1N2FMGxi']          = False     # (N/m); ; 
Morison['M1N3FMGxi']          = False     # (N/m); ; 
Morison['M1N4FMGxi']          = False     # (N/m); ; 
Morison['M1N5FMGxi']          = False     # (N/m); ; 
Morison['M1N6FMGxi']          = False     # (N/m); ; 
Morison['M1N7FMGxi']          = False     # (N/m); ; 
Morison['M1N8FMGxi']          = False     # (N/m); ; 
Morison['M1N9FMGxi']          = False     # (N/m); ; 
Morison['M2N1FMGxi']          = False     # (N/m); ; 
Morison['M2N2FMGxi']          = False     # (N/m); ; 
Morison['M2N3FMGxi']          = False     # (N/m); ; 
Morison['M2N4FMGxi']          = False     # (N/m); ; 
Morison['M2N5FMGxi']          = False     # (N/m); ; 
Morison['M2N6FMGxi']          = False     # (N/m); ; 
Morison['M2N7FMGxi']          = False     # (N/m); ; 
Morison['M2N8FMGxi']          = False     # (N/m); ; 
Morison['M2N9FMGxi']          = False     # (N/m); ; 
Morison['M3N1FMGxi']          = False     # (N/m); ; 
Morison['M3N2FMGxi']          = False     # (N/m); ; 
Morison['M3N3FMGxi']          = False     # (N/m); ; 
Morison['M3N4FMGxi']          = False     # (N/m); ; 
Morison['M3N5FMGxi']          = False     # (N/m); ; 
Morison['M3N6FMGxi']          = False     # (N/m); ; 
Morison['M3N7FMGxi']          = False     # (N/m); ; 
Morison['M3N8FMGxi']          = False     # (N/m); ; 
Morison['M3N9FMGxi']          = False     # (N/m); ; 
Morison['M4N1FMGxi']          = False     # (N/m); ; 
Morison['M4N2FMGxi']          = False     # (N/m); ; 
Morison['M4N3FMGxi']          = False     # (N/m); ; 
Morison['M4N4FMGxi']          = False     # (N/m); ; 
Morison['M4N5FMGxi']          = False     # (N/m); ; 
Morison['M4N6FMGxi']          = False     # (N/m); ; 
Morison['M4N7FMGxi']          = False     # (N/m); ; 
Morison['M4N8FMGxi']          = False     # (N/m); ; 
Morison['M4N9FMGxi']          = False     # (N/m); ; 
Morison['M5N1FMGxi']          = False     # (N/m); ; 
Morison['M5N2FMGxi']          = False     # (N/m); ; 
Morison['M5N3FMGxi']          = False     # (N/m); ; 
Morison['M5N4FMGxi']          = False     # (N/m); ; 
Morison['M5N5FMGxi']          = False     # (N/m); ; 
Morison['M5N6FMGxi']          = False     # (N/m); ; 
Morison['M5N7FMGxi']          = False     # (N/m); ; 
Morison['M5N8FMGxi']          = False     # (N/m); ; 
Morison['M5N9FMGxi']          = False     # (N/m); ; 
Morison['M6N1FMGxi']          = False     # (N/m); ; 
Morison['M6N2FMGxi']          = False     # (N/m); ; 
Morison['M6N3FMGxi']          = False     # (N/m); ; 
Morison['M6N4FMGxi']          = False     # (N/m); ; 
Morison['M6N5FMGxi']          = False     # (N/m); ; 
Morison['M6N6FMGxi']          = False     # (N/m); ; 
Morison['M6N7FMGxi']          = False     # (N/m); ; 
Morison['M6N8FMGxi']          = False     # (N/m); ; 
Morison['M6N9FMGxi']          = False     # (N/m); ; 
Morison['M7N1FMGxi']          = False     # (N/m); ; 
Morison['M7N2FMGxi']          = False     # (N/m); ; 
Morison['M7N3FMGxi']          = False     # (N/m); ; 
Morison['M7N4FMGxi']          = False     # (N/m); ; 
Morison['M7N5FMGxi']          = False     # (N/m); ; 
Morison['M7N6FMGxi']          = False     # (N/m); ; 
Morison['M7N7FMGxi']          = False     # (N/m); ; 
Morison['M7N8FMGxi']          = False     # (N/m); ; 
Morison['M7N9FMGxi']          = False     # (N/m); ; 
Morison['M8N1FMGxi']          = False     # (N/m); ; 
Morison['M8N2FMGxi']          = False     # (N/m); ; 
Morison['M8N3FMGxi']          = False     # (N/m); ; 
Morison['M8N4FMGxi']          = False     # (N/m); ; 
Morison['M8N5FMGxi']          = False     # (N/m); ; 
Morison['M8N6FMGxi']          = False     # (N/m); ; 
Morison['M8N7FMGxi']          = False     # (N/m); ; 
Morison['M8N8FMGxi']          = False     # (N/m); ; 
Morison['M8N9FMGxi']          = False     # (N/m); ; 
Morison['M9N1FMGxi']          = False     # (N/m); ; 
Morison['M9N2FMGxi']          = False     # (N/m); ; 
Morison['M9N3FMGxi']          = False     # (N/m); ; 
Morison['M9N4FMGxi']          = False     # (N/m); ; 
Morison['M9N5FMGxi']          = False     # (N/m); ; 
Morison['M9N6FMGxi']          = False     # (N/m); ; 
Morison['M9N7FMGxi']          = False     # (N/m); ; 
Morison['M9N8FMGxi']          = False     # (N/m); ; 
Morison['M9N9FMGxi']          = False     # (N/m); ; 
Morison['M1N1FMGyi']          = False     # (N/m); y-component of the distributed marine growth weight  expressed in the inertial coordinate system; 
Morison['M1N2FMGyi']          = False     # (N/m); ; 
Morison['M1N3FMGyi']          = False     # (N/m); ; 
Morison['M1N4FMGyi']          = False     # (N/m); ; 
Morison['M1N5FMGyi']          = False     # (N/m); ; 
Morison['M1N6FMGyi']          = False     # (N/m); ; 
Morison['M1N7FMGyi']          = False     # (N/m); ; 
Morison['M1N8FMGyi']          = False     # (N/m); ; 
Morison['M1N9FMGyi']          = False     # (N/m); ; 
Morison['M2N1FMGyi']          = False     # (N/m); ; 
Morison['M2N2FMGyi']          = False     # (N/m); ; 
Morison['M2N3FMGyi']          = False     # (N/m); ; 
Morison['M2N4FMGyi']          = False     # (N/m); ; 
Morison['M2N5FMGyi']          = False     # (N/m); ; 
Morison['M2N6FMGyi']          = False     # (N/m); ; 
Morison['M2N7FMGyi']          = False     # (N/m); ; 
Morison['M2N8FMGyi']          = False     # (N/m); ; 
Morison['M2N9FMGyi']          = False     # (N/m); ; 
Morison['M3N1FMGyi']          = False     # (N/m); ; 
Morison['M3N2FMGyi']          = False     # (N/m); ; 
Morison['M3N3FMGyi']          = False     # (N/m); ; 
Morison['M3N4FMGyi']          = False     # (N/m); ; 
Morison['M3N5FMGyi']          = False     # (N/m); ; 
Morison['M3N6FMGyi']          = False     # (N/m); ; 
Morison['M3N7FMGyi']          = False     # (N/m); ; 
Morison['M3N8FMGyi']          = False     # (N/m); ; 
Morison['M3N9FMGyi']          = False     # (N/m); ; 
Morison['M4N1FMGyi']          = False     # (N/m); ; 
Morison['M4N2FMGyi']          = False     # (N/m); ; 
Morison['M4N3FMGyi']          = False     # (N/m); ; 
Morison['M4N4FMGyi']          = False     # (N/m); ; 
Morison['M4N5FMGyi']          = False     # (N/m); ; 
Morison['M4N6FMGyi']          = False     # (N/m); ; 
Morison['M4N7FMGyi']          = False     # (N/m); ; 
Morison['M4N8FMGyi']          = False     # (N/m); ; 
Morison['M4N9FMGyi']          = False     # (N/m); ; 
Morison['M5N1FMGyi']          = False     # (N/m); ; 
Morison['M5N2FMGyi']          = False     # (N/m); ; 
Morison['M5N3FMGyi']          = False     # (N/m); ; 
Morison['M5N4FMGyi']          = False     # (N/m); ; 
Morison['M5N5FMGyi']          = False     # (N/m); ; 
Morison['M5N6FMGyi']          = False     # (N/m); ; 
Morison['M5N7FMGyi']          = False     # (N/m); ; 
Morison['M5N8FMGyi']          = False     # (N/m); ; 
Morison['M5N9FMGyi']          = False     # (N/m); ; 
Morison['M6N1FMGyi']          = False     # (N/m); ; 
Morison['M6N2FMGyi']          = False     # (N/m); ; 
Morison['M6N3FMGyi']          = False     # (N/m); ; 
Morison['M6N4FMGyi']          = False     # (N/m); ; 
Morison['M6N5FMGyi']          = False     # (N/m); ; 
Morison['M6N6FMGyi']          = False     # (N/m); ; 
Morison['M6N7FMGyi']          = False     # (N/m); ; 
Morison['M6N8FMGyi']          = False     # (N/m); ; 
Morison['M6N9FMGyi']          = False     # (N/m); ; 
Morison['M7N1FMGyi']          = False     # (N/m); ; 
Morison['M7N2FMGyi']          = False     # (N/m); ; 
Morison['M7N3FMGyi']          = False     # (N/m); ; 
Morison['M7N4FMGyi']          = False     # (N/m); ; 
Morison['M7N5FMGyi']          = False     # (N/m); ; 
Morison['M7N6FMGyi']          = False     # (N/m); ; 
Morison['M7N7FMGyi']          = False     # (N/m); ; 
Morison['M7N8FMGyi']          = False     # (N/m); ; 
Morison['M7N9FMGyi']          = False     # (N/m); ; 
Morison['M8N1FMGyi']          = False     # (N/m); ; 
Morison['M8N2FMGyi']          = False     # (N/m); ; 
Morison['M8N3FMGyi']          = False     # (N/m); ; 
Morison['M8N4FMGyi']          = False     # (N/m); ; 
Morison['M8N5FMGyi']          = False     # (N/m); ; 
Morison['M8N6FMGyi']          = False     # (N/m); ; 
Morison['M8N7FMGyi']          = False     # (N/m); ; 
Morison['M8N8FMGyi']          = False     # (N/m); ; 
Morison['M8N9FMGyi']          = False     # (N/m); ; 
Morison['M9N1FMGyi']          = False     # (N/m); ; 
Morison['M9N2FMGyi']          = False     # (N/m); ; 
Morison['M9N3FMGyi']          = False     # (N/m); ; 
Morison['M9N4FMGyi']          = False     # (N/m); ; 
Morison['M9N5FMGyi']          = False     # (N/m); ; 
Morison['M9N6FMGyi']          = False     # (N/m); ; 
Morison['M9N7FMGyi']          = False     # (N/m); ; 
Morison['M9N8FMGyi']          = False     # (N/m); ; 
Morison['M9N9FMGyi']          = False     # (N/m); ; 
Morison['M1N1FMGzi']          = False     # (N/m); z-component of the distributed marine growth weight expressed in the inertial coordinate system; 
Morison['M1N2FMGzi']          = False     # (N/m); ; 
Morison['M1N3FMGzi']          = False     # (N/m); ; 
Morison['M1N4FMGzi']          = False     # (N/m); ; 
Morison['M1N5FMGzi']          = False     # (N/m); ; 
Morison['M1N6FMGzi']          = False     # (N/m); ; 
Morison['M1N7FMGzi']          = False     # (N/m); ; 
Morison['M1N8FMGzi']          = False     # (N/m); ; 
Morison['M1N9FMGzi']          = False     # (N/m); ; 
Morison['M2N1FMGzi']          = False     # (N/m); ; 
Morison['M2N2FMGzi']          = False     # (N/m); ; 
Morison['M2N3FMGzi']          = False     # (N/m); ; 
Morison['M2N4FMGzi']          = False     # (N/m); ; 
Morison['M2N5FMGzi']          = False     # (N/m); ; 
Morison['M2N6FMGzi']          = False     # (N/m); ; 
Morison['M2N7FMGzi']          = False     # (N/m); ; 
Morison['M2N8FMGzi']          = False     # (N/m); ; 
Morison['M2N9FMGzi']          = False     # (N/m); ; 
Morison['M3N1FMGzi']          = False     # (N/m); ; 
Morison['M3N2FMGzi']          = False     # (N/m); ; 
Morison['M3N3FMGzi']          = False     # (N/m); ; 
Morison['M3N4FMGzi']          = False     # (N/m); ; 
Morison['M3N5FMGzi']          = False     # (N/m); ; 
Morison['M3N6FMGzi']          = False     # (N/m); ; 
Morison['M3N7FMGzi']          = False     # (N/m); ; 
Morison['M3N8FMGzi']          = False     # (N/m); ; 
Morison['M3N9FMGzi']          = False     # (N/m); ; 
Morison['M4N1FMGzi']          = False     # (N/m); ; 
Morison['M4N2FMGzi']          = False     # (N/m); ; 
Morison['M4N3FMGzi']          = False     # (N/m); ; 
Morison['M4N4FMGzi']          = False     # (N/m); ; 
Morison['M4N5FMGzi']          = False     # (N/m); ; 
Morison['M4N6FMGzi']          = False     # (N/m); ; 
Morison['M4N7FMGzi']          = False     # (N/m); ; 
Morison['M4N8FMGzi']          = False     # (N/m); ; 
Morison['M4N9FMGzi']          = False     # (N/m); ; 
Morison['M5N1FMGzi']          = False     # (N/m); ; 
Morison['M5N2FMGzi']          = False     # (N/m); ; 
Morison['M5N3FMGzi']          = False     # (N/m); ; 
Morison['M5N4FMGzi']          = False     # (N/m); ; 
Morison['M5N5FMGzi']          = False     # (N/m); ; 
Morison['M5N6FMGzi']          = False     # (N/m); ; 
Morison['M5N7FMGzi']          = False     # (N/m); ; 
Morison['M5N8FMGzi']          = False     # (N/m); ; 
Morison['M5N9FMGzi']          = False     # (N/m); ; 
Morison['M6N1FMGzi']          = False     # (N/m); ; 
Morison['M6N2FMGzi']          = False     # (N/m); ; 
Morison['M6N3FMGzi']          = False     # (N/m); ; 
Morison['M6N4FMGzi']          = False     # (N/m); ; 
Morison['M6N5FMGzi']          = False     # (N/m); ; 
Morison['M6N6FMGzi']          = False     # (N/m); ; 
Morison['M6N7FMGzi']          = False     # (N/m); ; 
Morison['M6N8FMGzi']          = False     # (N/m); ; 
Morison['M6N9FMGzi']          = False     # (N/m); ; 
Morison['M7N1FMGzi']          = False     # (N/m); ; 
Morison['M7N2FMGzi']          = False     # (N/m); ; 
Morison['M7N3FMGzi']          = False     # (N/m); ; 
Morison['M7N4FMGzi']          = False     # (N/m); ; 
Morison['M7N5FMGzi']          = False     # (N/m); ; 
Morison['M7N6FMGzi']          = False     # (N/m); ; 
Morison['M7N7FMGzi']          = False     # (N/m); ; 
Morison['M7N8FMGzi']          = False     # (N/m); ; 
Morison['M7N9FMGzi']          = False     # (N/m); ; 
Morison['M8N1FMGzi']          = False     # (N/m); ; 
Morison['M8N2FMGzi']          = False     # (N/m); ; 
Morison['M8N3FMGzi']          = False     # (N/m); ; 
Morison['M8N4FMGzi']          = False     # (N/m); ; 
Morison['M8N5FMGzi']          = False     # (N/m); ; 
Morison['M8N6FMGzi']          = False     # (N/m); ; 
Morison['M8N7FMGzi']          = False     # (N/m); ; 
Morison['M8N8FMGzi']          = False     # (N/m); ; 
Morison['M8N9FMGzi']          = False     # (N/m); ; 
Morison['M9N1FMGzi']          = False     # (N/m); ; 
Morison['M9N2FMGzi']          = False     # (N/m); ; 
Morison['M9N3FMGzi']          = False     # (N/m); ; 
Morison['M9N4FMGzi']          = False     # (N/m); ; 
Morison['M9N5FMGzi']          = False     # (N/m); ; 
Morison['M9N6FMGzi']          = False     # (N/m); ; 
Morison['M9N7FMGzi']          = False     # (N/m); ; 
Morison['M9N8FMGzi']          = False     # (N/m); ; 
Morison['M9N9FMGzi']          = False     # (N/m); ; 
Morison['M1N1FAMxi']          = False     # (N/m); x-component of the distributed added mass force due to the member's displacement of the external fluid, expressed in the inertial coordinate system; 
Morison['M1N2FAMxi']          = False     # (N/m); ; 
Morison['M1N3FAMxi']          = False     # (N/m); ; 
Morison['M1N4FAMxi']          = False     # (N/m); ; 
Morison['M1N5FAMxi']          = False     # (N/m); ; 
Morison['M1N6FAMxi']          = False     # (N/m); ; 
Morison['M1N7FAMxi']          = False     # (N/m); ; 
Morison['M1N8FAMxi']          = False     # (N/m); ; 
Morison['M1N9FAMxi']          = False     # (N/m); ; 
Morison['M2N1FAMxi']          = False     # (N/m); ; 
Morison['M2N2FAMxi']          = False     # (N/m); ; 
Morison['M2N3FAMxi']          = False     # (N/m); ; 
Morison['M2N4FAMxi']          = False     # (N/m); ; 
Morison['M2N5FAMxi']          = False     # (N/m); ; 
Morison['M2N6FAMxi']          = False     # (N/m); ; 
Morison['M2N7FAMxi']          = False     # (N/m); ; 
Morison['M2N8FAMxi']          = False     # (N/m); ; 
Morison['M2N9FAMxi']          = False     # (N/m); ; 
Morison['M3N1FAMxi']          = False     # (N/m); ; 
Morison['M3N2FAMxi']          = False     # (N/m); ; 
Morison['M3N3FAMxi']          = False     # (N/m); ; 
Morison['M3N4FAMxi']          = False     # (N/m); ; 
Morison['M3N5FAMxi']          = False     # (N/m); ; 
Morison['M3N6FAMxi']          = False     # (N/m); ; 
Morison['M3N7FAMxi']          = False     # (N/m); ; 
Morison['M3N8FAMxi']          = False     # (N/m); ; 
Morison['M3N9FAMxi']          = False     # (N/m); ; 
Morison['M4N1FAMxi']          = False     # (N/m); ; 
Morison['M4N2FAMxi']          = False     # (N/m); ; 
Morison['M4N3FAMxi']          = False     # (N/m); ; 
Morison['M4N4FAMxi']          = False     # (N/m); ; 
Morison['M4N5FAMxi']          = False     # (N/m); ; 
Morison['M4N6FAMxi']          = False     # (N/m); ; 
Morison['M4N7FAMxi']          = False     # (N/m); ; 
Morison['M4N8FAMxi']          = False     # (N/m); ; 
Morison['M4N9FAMxi']          = False     # (N/m); ; 
Morison['M5N1FAMxi']          = False     # (N/m); ; 
Morison['M5N2FAMxi']          = False     # (N/m); ; 
Morison['M5N3FAMxi']          = False     # (N/m); ; 
Morison['M5N4FAMxi']          = False     # (N/m); ; 
Morison['M5N5FAMxi']          = False     # (N/m); ; 
Morison['M5N6FAMxi']          = False     # (N/m); ; 
Morison['M5N7FAMxi']          = False     # (N/m); ; 
Morison['M5N8FAMxi']          = False     # (N/m); ; 
Morison['M5N9FAMxi']          = False     # (N/m); ; 
Morison['M6N1FAMxi']          = False     # (N/m); ; 
Morison['M6N2FAMxi']          = False     # (N/m); ; 
Morison['M6N3FAMxi']          = False     # (N/m); ; 
Morison['M6N4FAMxi']          = False     # (N/m); ; 
Morison['M6N5FAMxi']          = False     # (N/m); ; 
Morison['M6N6FAMxi']          = False     # (N/m); ; 
Morison['M6N7FAMxi']          = False     # (N/m); ; 
Morison['M6N8FAMxi']          = False     # (N/m); ; 
Morison['M6N9FAMxi']          = False     # (N/m); ; 
Morison['M7N1FAMxi']          = False     # (N/m); ; 
Morison['M7N2FAMxi']          = False     # (N/m); ; 
Morison['M7N3FAMxi']          = False     # (N/m); ; 
Morison['M7N4FAMxi']          = False     # (N/m); ; 
Morison['M7N5FAMxi']          = False     # (N/m); ; 
Morison['M7N6FAMxi']          = False     # (N/m); ; 
Morison['M7N7FAMxi']          = False     # (N/m); ; 
Morison['M7N8FAMxi']          = False     # (N/m); ; 
Morison['M7N9FAMxi']          = False     # (N/m); ; 
Morison['M8N1FAMxi']          = False     # (N/m); ; 
Morison['M8N2FAMxi']          = False     # (N/m); ; 
Morison['M8N3FAMxi']          = False     # (N/m); ; 
Morison['M8N4FAMxi']          = False     # (N/m); ; 
Morison['M8N5FAMxi']          = False     # (N/m); ; 
Morison['M8N6FAMxi']          = False     # (N/m); ; 
Morison['M8N7FAMxi']          = False     # (N/m); ; 
Morison['M8N8FAMxi']          = False     # (N/m); ; 
Morison['M8N9FAMxi']          = False     # (N/m); ; 
Morison['M9N1FAMxi']          = False     # (N/m); ; 
Morison['M9N2FAMxi']          = False     # (N/m); ; 
Morison['M9N3FAMxi']          = False     # (N/m); ; 
Morison['M9N4FAMxi']          = False     # (N/m); ; 
Morison['M9N5FAMxi']          = False     # (N/m); ; 
Morison['M9N6FAMxi']          = False     # (N/m); ; 
Morison['M9N7FAMxi']          = False     # (N/m); ; 
Morison['M9N8FAMxi']          = False     # (N/m); ; 
Morison['M9N9FAMxi']          = False     # (N/m); ; 
Morison['M1N1FAMyi']          = False     # (N/m); y-component of the distributed added mass force due to the member's displacement of the external fluid, expressed in the inertial coordinate system; 
Morison['M1N2FAMyi']          = False     # (N/m); ; 
Morison['M1N3FAMyi']          = False     # (N/m); ; 
Morison['M1N4FAMyi']          = False     # (N/m); ; 
Morison['M1N5FAMyi']          = False     # (N/m); ; 
Morison['M1N6FAMyi']          = False     # (N/m); ; 
Morison['M1N7FAMyi']          = False     # (N/m); ; 
Morison['M1N8FAMyi']          = False     # (N/m); ; 
Morison['M1N9FAMyi']          = False     # (N/m); ; 
Morison['M2N1FAMyi']          = False     # (N/m); ; 
Morison['M2N2FAMyi']          = False     # (N/m); ; 
Morison['M2N3FAMyi']          = False     # (N/m); ; 
Morison['M2N4FAMyi']          = False     # (N/m); ; 
Morison['M2N5FAMyi']          = False     # (N/m); ; 
Morison['M2N6FAMyi']          = False     # (N/m); ; 
Morison['M2N7FAMyi']          = False     # (N/m); ; 
Morison['M2N8FAMyi']          = False     # (N/m); ; 
Morison['M2N9FAMyi']          = False     # (N/m); ; 
Morison['M3N1FAMyi']          = False     # (N/m); ; 
Morison['M3N2FAMyi']          = False     # (N/m); ; 
Morison['M3N3FAMyi']          = False     # (N/m); ; 
Morison['M3N4FAMyi']          = False     # (N/m); ; 
Morison['M3N5FAMyi']          = False     # (N/m); ; 
Morison['M3N6FAMyi']          = False     # (N/m); ; 
Morison['M3N7FAMyi']          = False     # (N/m); ; 
Morison['M3N8FAMyi']          = False     # (N/m); ; 
Morison['M3N9FAMyi']          = False     # (N/m); ; 
Morison['M4N1FAMyi']          = False     # (N/m); ; 
Morison['M4N2FAMyi']          = False     # (N/m); ; 
Morison['M4N3FAMyi']          = False     # (N/m); ; 
Morison['M4N4FAMyi']          = False     # (N/m); ; 
Morison['M4N5FAMyi']          = False     # (N/m); ; 
Morison['M4N6FAMyi']          = False     # (N/m); ; 
Morison['M4N7FAMyi']          = False     # (N/m); ; 
Morison['M4N8FAMyi']          = False     # (N/m); ; 
Morison['M4N9FAMyi']          = False     # (N/m); ; 
Morison['M5N1FAMyi']          = False     # (N/m); ; 
Morison['M5N2FAMyi']          = False     # (N/m); ; 
Morison['M5N3FAMyi']          = False     # (N/m); ; 
Morison['M5N4FAMyi']          = False     # (N/m); ; 
Morison['M5N5FAMyi']          = False     # (N/m); ; 
Morison['M5N6FAMyi']          = False     # (N/m); ; 
Morison['M5N7FAMyi']          = False     # (N/m); ; 
Morison['M5N8FAMyi']          = False     # (N/m); ; 
Morison['M5N9FAMyi']          = False     # (N/m); ; 
Morison['M6N1FAMyi']          = False     # (N/m); ; 
Morison['M6N2FAMyi']          = False     # (N/m); ; 
Morison['M6N3FAMyi']          = False     # (N/m); ; 
Morison['M6N4FAMyi']          = False     # (N/m); ; 
Morison['M6N5FAMyi']          = False     # (N/m); ; 
Morison['M6N6FAMyi']          = False     # (N/m); ; 
Morison['M6N7FAMyi']          = False     # (N/m); ; 
Morison['M6N8FAMyi']          = False     # (N/m); ; 
Morison['M6N9FAMyi']          = False     # (N/m); ; 
Morison['M7N1FAMyi']          = False     # (N/m); ; 
Morison['M7N2FAMyi']          = False     # (N/m); ; 
Morison['M7N3FAMyi']          = False     # (N/m); ; 
Morison['M7N4FAMyi']          = False     # (N/m); ; 
Morison['M7N5FAMyi']          = False     # (N/m); ; 
Morison['M7N6FAMyi']          = False     # (N/m); ; 
Morison['M7N7FAMyi']          = False     # (N/m); ; 
Morison['M7N8FAMyi']          = False     # (N/m); ; 
Morison['M7N9FAMyi']          = False     # (N/m); ; 
Morison['M8N1FAMyi']          = False     # (N/m); ; 
Morison['M8N2FAMyi']          = False     # (N/m); ; 
Morison['M8N3FAMyi']          = False     # (N/m); ; 
Morison['M8N4FAMyi']          = False     # (N/m); ; 
Morison['M8N5FAMyi']          = False     # (N/m); ; 
Morison['M8N6FAMyi']          = False     # (N/m); ; 
Morison['M8N7FAMyi']          = False     # (N/m); ; 
Morison['M8N8FAMyi']          = False     # (N/m); ; 
Morison['M8N9FAMyi']          = False     # (N/m); ; 
Morison['M9N1FAMyi']          = False     # (N/m); ; 
Morison['M9N2FAMyi']          = False     # (N/m); ; 
Morison['M9N3FAMyi']          = False     # (N/m); ; 
Morison['M9N4FAMyi']          = False     # (N/m); ; 
Morison['M9N5FAMyi']          = False     # (N/m); ; 
Morison['M9N6FAMyi']          = False     # (N/m); ; 
Morison['M9N7FAMyi']          = False     # (N/m); ; 
Morison['M9N8FAMyi']          = False     # (N/m); ; 
Morison['M9N9FAMyi']          = False     # (N/m); ; 
Morison['M1N1FAMzi']          = False     # (N/m); z-component of the distributed added mass force due to the member's displacement of the external fluid, expressed in the inertial coordinate system; 
Morison['M1N2FAMzi']          = False     # (N/m); ; 
Morison['M1N3FAMzi']          = False     # (N/m); ; 
Morison['M1N4FAMzi']          = False     # (N/m); ; 
Morison['M1N5FAMzi']          = False     # (N/m); ; 
Morison['M1N6FAMzi']          = False     # (N/m); ; 
Morison['M1N7FAMzi']          = False     # (N/m); ; 
Morison['M1N8FAMzi']          = False     # (N/m); ; 
Morison['M1N9FAMzi']          = False     # (N/m); ; 
Morison['M2N1FAMzi']          = False     # (N/m); ; 
Morison['M2N2FAMzi']          = False     # (N/m); ; 
Morison['M2N3FAMzi']          = False     # (N/m); ; 
Morison['M2N4FAMzi']          = False     # (N/m); ; 
Morison['M2N5FAMzi']          = False     # (N/m); ; 
Morison['M2N6FAMzi']          = False     # (N/m); ; 
Morison['M2N7FAMzi']          = False     # (N/m); ; 
Morison['M2N8FAMzi']          = False     # (N/m); ; 
Morison['M2N9FAMzi']          = False     # (N/m); ; 
Morison['M3N1FAMzi']          = False     # (N/m); ; 
Morison['M3N2FAMzi']          = False     # (N/m); ; 
Morison['M3N3FAMzi']          = False     # (N/m); ; 
Morison['M3N4FAMzi']          = False     # (N/m); ; 
Morison['M3N5FAMzi']          = False     # (N/m); ; 
Morison['M3N6FAMzi']          = False     # (N/m); ; 
Morison['M3N7FAMzi']          = False     # (N/m); ; 
Morison['M3N8FAMzi']          = False     # (N/m); ; 
Morison['M3N9FAMzi']          = False     # (N/m); ; 
Morison['M4N1FAMzi']          = False     # (N/m); ; 
Morison['M4N2FAMzi']          = False     # (N/m); ; 
Morison['M4N3FAMzi']          = False     # (N/m); ; 
Morison['M4N4FAMzi']          = False     # (N/m); ; 
Morison['M4N5FAMzi']          = False     # (N/m); ; 
Morison['M4N6FAMzi']          = False     # (N/m); ; 
Morison['M4N7FAMzi']          = False     # (N/m); ; 
Morison['M4N8FAMzi']          = False     # (N/m); ; 
Morison['M4N9FAMzi']          = False     # (N/m); ; 
Morison['M5N1FAMzi']          = False     # (N/m); ; 
Morison['M5N2FAMzi']          = False     # (N/m); ; 
Morison['M5N3FAMzi']          = False     # (N/m); ; 
Morison['M5N4FAMzi']          = False     # (N/m); ; 
Morison['M5N5FAMzi']          = False     # (N/m); ; 
Morison['M5N6FAMzi']          = False     # (N/m); ; 
Morison['M5N7FAMzi']          = False     # (N/m); ; 
Morison['M5N8FAMzi']          = False     # (N/m); ; 
Morison['M5N9FAMzi']          = False     # (N/m); ; 
Morison['M6N1FAMzi']          = False     # (N/m); ; 
Morison['M6N2FAMzi']          = False     # (N/m); ; 
Morison['M6N3FAMzi']          = False     # (N/m); ; 
Morison['M6N4FAMzi']          = False     # (N/m); ; 
Morison['M6N5FAMzi']          = False     # (N/m); ; 
Morison['M6N6FAMzi']          = False     # (N/m); ; 
Morison['M6N7FAMzi']          = False     # (N/m); ; 
Morison['M6N8FAMzi']          = False     # (N/m); ; 
Morison['M6N9FAMzi']          = False     # (N/m); ; 
Morison['M7N1FAMzi']          = False     # (N/m); ; 
Morison['M7N2FAMzi']          = False     # (N/m); ; 
Morison['M7N3FAMzi']          = False     # (N/m); ; 
Morison['M7N4FAMzi']          = False     # (N/m); ; 
Morison['M7N5FAMzi']          = False     # (N/m); ; 
Morison['M7N6FAMzi']          = False     # (N/m); ; 
Morison['M7N7FAMzi']          = False     # (N/m); ; 
Morison['M7N8FAMzi']          = False     # (N/m); ; 
Morison['M7N9FAMzi']          = False     # (N/m); ; 
Morison['M8N1FAMzi']          = False     # (N/m); ; 
Morison['M8N2FAMzi']          = False     # (N/m); ; 
Morison['M8N3FAMzi']          = False     # (N/m); ; 
Morison['M8N4FAMzi']          = False     # (N/m); ; 
Morison['M8N5FAMzi']          = False     # (N/m); ; 
Morison['M8N6FAMzi']          = False     # (N/m); ; 
Morison['M8N7FAMzi']          = False     # (N/m); ; 
Morison['M8N8FAMzi']          = False     # (N/m); ; 
Morison['M8N9FAMzi']          = False     # (N/m); ; 
Morison['M9N1FAMzi']          = False     # (N/m); ; 
Morison['M9N2FAMzi']          = False     # (N/m); ; 
Morison['M9N3FAMzi']          = False     # (N/m); ; 
Morison['M9N4FAMzi']          = False     # (N/m); ; 
Morison['M9N5FAMzi']          = False     # (N/m); ; 
Morison['M9N6FAMzi']          = False     # (N/m); ; 
Morison['M9N7FAMzi']          = False     # (N/m); ; 
Morison['M9N8FAMzi']          = False     # (N/m); ; 
Morison['M9N9FAMzi']          = False     # (N/m); ; 
Morison['M1N1FAGxi']          = False     # (N/m); x-component of the distributed added mass force due to the marine growth, expressed in the inertial coordinate system; 
Morison['M1N2FAGxi']          = False     # (N/m); ; 
Morison['M1N3FAGxi']          = False     # (N/m); ; 
Morison['M1N4FAGxi']          = False     # (N/m); ; 
Morison['M1N5FAGxi']          = False     # (N/m); ; 
Morison['M1N6FAGxi']          = False     # (N/m); ; 
Morison['M1N7FAGxi']          = False     # (N/m); ; 
Morison['M1N8FAGxi']          = False     # (N/m); ; 
Morison['M1N9FAGxi']          = False     # (N/m); ; 
Morison['M2N1FAGxi']          = False     # (N/m); ; 
Morison['M2N2FAGxi']          = False     # (N/m); ; 
Morison['M2N3FAGxi']          = False     # (N/m); ; 
Morison['M2N4FAGxi']          = False     # (N/m); ; 
Morison['M2N5FAGxi']          = False     # (N/m); ; 
Morison['M2N6FAGxi']          = False     # (N/m); ; 
Morison['M2N7FAGxi']          = False     # (N/m); ; 
Morison['M2N8FAGxi']          = False     # (N/m); ; 
Morison['M2N9FAGxi']          = False     # (N/m); ; 
Morison['M3N1FAGxi']          = False     # (N/m); ; 
Morison['M3N2FAGxi']          = False     # (N/m); ; 
Morison['M3N3FAGxi']          = False     # (N/m); ; 
Morison['M3N4FAGxi']          = False     # (N/m); ; 
Morison['M3N5FAGxi']          = False     # (N/m); ; 
Morison['M3N6FAGxi']          = False     # (N/m); ; 
Morison['M3N7FAGxi']          = False     # (N/m); ; 
Morison['M3N8FAGxi']          = False     # (N/m); ; 
Morison['M3N9FAGxi']          = False     # (N/m); ; 
Morison['M4N1FAGxi']          = False     # (N/m); ; 
Morison['M4N2FAGxi']          = False     # (N/m); ; 
Morison['M4N3FAGxi']          = False     # (N/m); ; 
Morison['M4N4FAGxi']          = False     # (N/m); ; 
Morison['M4N5FAGxi']          = False     # (N/m); ; 
Morison['M4N6FAGxi']          = False     # (N/m); ; 
Morison['M4N7FAGxi']          = False     # (N/m); ; 
Morison['M4N8FAGxi']          = False     # (N/m); ; 
Morison['M4N9FAGxi']          = False     # (N/m); ; 
Morison['M5N1FAGxi']          = False     # (N/m); ; 
Morison['M5N2FAGxi']          = False     # (N/m); ; 
Morison['M5N3FAGxi']          = False     # (N/m); ; 
Morison['M5N4FAGxi']          = False     # (N/m); ; 
Morison['M5N5FAGxi']          = False     # (N/m); ; 
Morison['M5N6FAGxi']          = False     # (N/m); ; 
Morison['M5N7FAGxi']          = False     # (N/m); ; 
Morison['M5N8FAGxi']          = False     # (N/m); ; 
Morison['M5N9FAGxi']          = False     # (N/m); ; 
Morison['M6N1FAGxi']          = False     # (N/m); ; 
Morison['M6N2FAGxi']          = False     # (N/m); ; 
Morison['M6N3FAGxi']          = False     # (N/m); ; 
Morison['M6N4FAGxi']          = False     # (N/m); ; 
Morison['M6N5FAGxi']          = False     # (N/m); ; 
Morison['M6N6FAGxi']          = False     # (N/m); ; 
Morison['M6N7FAGxi']          = False     # (N/m); ; 
Morison['M6N8FAGxi']          = False     # (N/m); ; 
Morison['M6N9FAGxi']          = False     # (N/m); ; 
Morison['M7N1FAGxi']          = False     # (N/m); ; 
Morison['M7N2FAGxi']          = False     # (N/m); ; 
Morison['M7N3FAGxi']          = False     # (N/m); ; 
Morison['M7N4FAGxi']          = False     # (N/m); ; 
Morison['M7N5FAGxi']          = False     # (N/m); ; 
Morison['M7N6FAGxi']          = False     # (N/m); ; 
Morison['M7N7FAGxi']          = False     # (N/m); ; 
Morison['M7N8FAGxi']          = False     # (N/m); ; 
Morison['M7N9FAGxi']          = False     # (N/m); ; 
Morison['M8N1FAGxi']          = False     # (N/m); ; 
Morison['M8N2FAGxi']          = False     # (N/m); ; 
Morison['M8N3FAGxi']          = False     # (N/m); ; 
Morison['M8N4FAGxi']          = False     # (N/m); ; 
Morison['M8N5FAGxi']          = False     # (N/m); ; 
Morison['M8N6FAGxi']          = False     # (N/m); ; 
Morison['M8N7FAGxi']          = False     # (N/m); ; 
Morison['M8N8FAGxi']          = False     # (N/m); ; 
Morison['M8N9FAGxi']          = False     # (N/m); ; 
Morison['M9N1FAGxi']          = False     # (N/m); ; 
Morison['M9N2FAGxi']          = False     # (N/m); ; 
Morison['M9N3FAGxi']          = False     # (N/m); ; 
Morison['M9N4FAGxi']          = False     # (N/m); ; 
Morison['M9N5FAGxi']          = False     # (N/m); ; 
Morison['M9N6FAGxi']          = False     # (N/m); ; 
Morison['M9N7FAGxi']          = False     # (N/m); ; 
Morison['M9N8FAGxi']          = False     # (N/m); ; 
Morison['M9N9FAGxi']          = False     # (N/m); ; 
Morison['M1N1FAGyi']          = False     # (N/m); y-component of the distributed added mass force due to the marine growth, expressed in the inertial coordinate system; 
Morison['M1N2FAGyi']          = False     # (N/m); ; 
Morison['M1N3FAGyi']          = False     # (N/m); ; 
Morison['M1N4FAGyi']          = False     # (N/m); ; 
Morison['M1N5FAGyi']          = False     # (N/m); ; 
Morison['M1N6FAGyi']          = False     # (N/m); ; 
Morison['M1N7FAGyi']          = False     # (N/m); ; 
Morison['M1N8FAGyi']          = False     # (N/m); ; 
Morison['M1N9FAGyi']          = False     # (N/m); ; 
Morison['M2N1FAGyi']          = False     # (N/m); ; 
Morison['M2N2FAGyi']          = False     # (N/m); ; 
Morison['M2N3FAGyi']          = False     # (N/m); ; 
Morison['M2N4FAGyi']          = False     # (N/m); ; 
Morison['M2N5FAGyi']          = False     # (N/m); ; 
Morison['M2N6FAGyi']          = False     # (N/m); ; 
Morison['M2N7FAGyi']          = False     # (N/m); ; 
Morison['M2N8FAGyi']          = False     # (N/m); ; 
Morison['M2N9FAGyi']          = False     # (N/m); ; 
Morison['M3N1FAGyi']          = False     # (N/m); ; 
Morison['M3N2FAGyi']          = False     # (N/m); ; 
Morison['M3N3FAGyi']          = False     # (N/m); ; 
Morison['M3N4FAGyi']          = False     # (N/m); ; 
Morison['M3N5FAGyi']          = False     # (N/m); ; 
Morison['M3N6FAGyi']          = False     # (N/m); ; 
Morison['M3N7FAGyi']          = False     # (N/m); ; 
Morison['M3N8FAGyi']          = False     # (N/m); ; 
Morison['M3N9FAGyi']          = False     # (N/m); ; 
Morison['M4N1FAGyi']          = False     # (N/m); ; 
Morison['M4N2FAGyi']          = False     # (N/m); ; 
Morison['M4N3FAGyi']          = False     # (N/m); ; 
Morison['M4N4FAGyi']          = False     # (N/m); ; 
Morison['M4N5FAGyi']          = False     # (N/m); ; 
Morison['M4N6FAGyi']          = False     # (N/m); ; 
Morison['M4N7FAGyi']          = False     # (N/m); ; 
Morison['M4N8FAGyi']          = False     # (N/m); ; 
Morison['M4N9FAGyi']          = False     # (N/m); ; 
Morison['M5N1FAGyi']          = False     # (N/m); ; 
Morison['M5N2FAGyi']          = False     # (N/m); ; 
Morison['M5N3FAGyi']          = False     # (N/m); ; 
Morison['M5N4FAGyi']          = False     # (N/m); ; 
Morison['M5N5FAGyi']          = False     # (N/m); ; 
Morison['M5N6FAGyi']          = False     # (N/m); ; 
Morison['M5N7FAGyi']          = False     # (N/m); ; 
Morison['M5N8FAGyi']          = False     # (N/m); ; 
Morison['M5N9FAGyi']          = False     # (N/m); ; 
Morison['M6N1FAGyi']          = False     # (N/m); ; 
Morison['M6N2FAGyi']          = False     # (N/m); ; 
Morison['M6N3FAGyi']          = False     # (N/m); ; 
Morison['M6N4FAGyi']          = False     # (N/m); ; 
Morison['M6N5FAGyi']          = False     # (N/m); ; 
Morison['M6N6FAGyi']          = False     # (N/m); ; 
Morison['M6N7FAGyi']          = False     # (N/m); ; 
Morison['M6N8FAGyi']          = False     # (N/m); ; 
Morison['M6N9FAGyi']          = False     # (N/m); ; 
Morison['M7N1FAGyi']          = False     # (N/m); ; 
Morison['M7N2FAGyi']          = False     # (N/m); ; 
Morison['M7N3FAGyi']          = False     # (N/m); ; 
Morison['M7N4FAGyi']          = False     # (N/m); ; 
Morison['M7N5FAGyi']          = False     # (N/m); ; 
Morison['M7N6FAGyi']          = False     # (N/m); ; 
Morison['M7N7FAGyi']          = False     # (N/m); ; 
Morison['M7N8FAGyi']          = False     # (N/m); ; 
Morison['M7N9FAGyi']          = False     # (N/m); ; 
Morison['M8N1FAGyi']          = False     # (N/m); ; 
Morison['M8N2FAGyi']          = False     # (N/m); ; 
Morison['M8N3FAGyi']          = False     # (N/m); ; 
Morison['M8N4FAGyi']          = False     # (N/m); ; 
Morison['M8N5FAGyi']          = False     # (N/m); ; 
Morison['M8N6FAGyi']          = False     # (N/m); ; 
Morison['M8N7FAGyi']          = False     # (N/m); ; 
Morison['M8N8FAGyi']          = False     # (N/m); ; 
Morison['M8N9FAGyi']          = False     # (N/m); ; 
Morison['M9N1FAGyi']          = False     # (N/m); ; 
Morison['M9N2FAGyi']          = False     # (N/m); ; 
Morison['M9N3FAGyi']          = False     # (N/m); ; 
Morison['M9N4FAGyi']          = False     # (N/m); ; 
Morison['M9N5FAGyi']          = False     # (N/m); ; 
Morison['M9N6FAGyi']          = False     # (N/m); ; 
Morison['M9N7FAGyi']          = False     # (N/m); ; 
Morison['M9N8FAGyi']          = False     # (N/m); ; 
Morison['M9N9FAGyi']          = False     # (N/m); ; 
Morison['M1N1FAGzi']          = False     # (N/m); z-component of the distributed added mass force due to the marine growth, expressed in the inertial coordinate system; 
Morison['M1N2FAGzi']          = False     # (N/m); ; 
Morison['M1N3FAGzi']          = False     # (N/m); ; 
Morison['M1N4FAGzi']          = False     # (N/m); ; 
Morison['M1N5FAGzi']          = False     # (N/m); ; 
Morison['M1N6FAGzi']          = False     # (N/m); ; 
Morison['M1N7FAGzi']          = False     # (N/m); ; 
Morison['M1N8FAGzi']          = False     # (N/m); ; 
Morison['M1N9FAGzi']          = False     # (N/m); ; 
Morison['M2N1FAGzi']          = False     # (N/m); ; 
Morison['M2N2FAGzi']          = False     # (N/m); ; 
Morison['M2N3FAGzi']          = False     # (N/m); ; 
Morison['M2N4FAGzi']          = False     # (N/m); ; 
Morison['M2N5FAGzi']          = False     # (N/m); ; 
Morison['M2N6FAGzi']          = False     # (N/m); ; 
Morison['M2N7FAGzi']          = False     # (N/m); ; 
Morison['M2N8FAGzi']          = False     # (N/m); ; 
Morison['M2N9FAGzi']          = False     # (N/m); ; 
Morison['M3N1FAGzi']          = False     # (N/m); ; 
Morison['M3N2FAGzi']          = False     # (N/m); ; 
Morison['M3N3FAGzi']          = False     # (N/m); ; 
Morison['M3N4FAGzi']          = False     # (N/m); ; 
Morison['M3N5FAGzi']          = False     # (N/m); ; 
Morison['M3N6FAGzi']          = False     # (N/m); ; 
Morison['M3N7FAGzi']          = False     # (N/m); ; 
Morison['M3N8FAGzi']          = False     # (N/m); ; 
Morison['M3N9FAGzi']          = False     # (N/m); ; 
Morison['M4N1FAGzi']          = False     # (N/m); ; 
Morison['M4N2FAGzi']          = False     # (N/m); ; 
Morison['M4N3FAGzi']          = False     # (N/m); ; 
Morison['M4N4FAGzi']          = False     # (N/m); ; 
Morison['M4N5FAGzi']          = False     # (N/m); ; 
Morison['M4N6FAGzi']          = False     # (N/m); ; 
Morison['M4N7FAGzi']          = False     # (N/m); ; 
Morison['M4N8FAGzi']          = False     # (N/m); ; 
Morison['M4N9FAGzi']          = False     # (N/m); ; 
Morison['M5N1FAGzi']          = False     # (N/m); ; 
Morison['M5N2FAGzi']          = False     # (N/m); ; 
Morison['M5N3FAGzi']          = False     # (N/m); ; 
Morison['M5N4FAGzi']          = False     # (N/m); ; 
Morison['M5N5FAGzi']          = False     # (N/m); ; 
Morison['M5N6FAGzi']          = False     # (N/m); ; 
Morison['M5N7FAGzi']          = False     # (N/m); ; 
Morison['M5N8FAGzi']          = False     # (N/m); ; 
Morison['M5N9FAGzi']          = False     # (N/m); ; 
Morison['M6N1FAGzi']          = False     # (N/m); ; 
Morison['M6N2FAGzi']          = False     # (N/m); ; 
Morison['M6N3FAGzi']          = False     # (N/m); ; 
Morison['M6N4FAGzi']          = False     # (N/m); ; 
Morison['M6N5FAGzi']          = False     # (N/m); ; 
Morison['M6N6FAGzi']          = False     # (N/m); ; 
Morison['M6N7FAGzi']          = False     # (N/m); ; 
Morison['M6N8FAGzi']          = False     # (N/m); ; 
Morison['M6N9FAGzi']          = False     # (N/m); ; 
Morison['M7N1FAGzi']          = False     # (N/m); ; 
Morison['M7N2FAGzi']          = False     # (N/m); ; 
Morison['M7N3FAGzi']          = False     # (N/m); ; 
Morison['M7N4FAGzi']          = False     # (N/m); ; 
Morison['M7N5FAGzi']          = False     # (N/m); ; 
Morison['M7N6FAGzi']          = False     # (N/m); ; 
Morison['M7N7FAGzi']          = False     # (N/m); ; 
Morison['M7N8FAGzi']          = False     # (N/m); ; 
Morison['M7N9FAGzi']          = False     # (N/m); ; 
Morison['M8N1FAGzi']          = False     # (N/m); ; 
Morison['M8N2FAGzi']          = False     # (N/m); ; 
Morison['M8N3FAGzi']          = False     # (N/m); ; 
Morison['M8N4FAGzi']          = False     # (N/m); ; 
Morison['M8N5FAGzi']          = False     # (N/m); ; 
Morison['M8N6FAGzi']          = False     # (N/m); ; 
Morison['M8N7FAGzi']          = False     # (N/m); ; 
Morison['M8N8FAGzi']          = False     # (N/m); ; 
Morison['M8N9FAGzi']          = False     # (N/m); ; 
Morison['M9N1FAGzi']          = False     # (N/m); ; 
Morison['M9N2FAGzi']          = False     # (N/m); ; 
Morison['M9N3FAGzi']          = False     # (N/m); ; 
Morison['M9N4FAGzi']          = False     # (N/m); ; 
Morison['M9N5FAGzi']          = False     # (N/m); ; 
Morison['M9N6FAGzi']          = False     # (N/m); ; 
Morison['M9N7FAGzi']          = False     # (N/m); ; 
Morison['M9N8FAGzi']          = False     # (N/m); ; 
Morison['M9N9FAGzi']          = False     # (N/m); ; 
Morison['M1N1FAFxi']          = False     # (N/m); x-component of the distributed added mass force due to the ballast, expressed in the inertial coordinate system; 
Morison['M1N2FAFxi']          = False     # (N/m); ; 
Morison['M1N3FAFxi']          = False     # (N/m); ; 
Morison['M1N4FAFxi']          = False     # (N/m); ; 
Morison['M1N5FAFxi']          = False     # (N/m); ; 
Morison['M1N6FAFxi']          = False     # (N/m); ; 
Morison['M1N7FAFxi']          = False     # (N/m); ; 
Morison['M1N8FAFxi']          = False     # (N/m); ; 
Morison['M1N9FAFxi']          = False     # (N/m); ; 
Morison['M2N1FAFxi']          = False     # (N/m); ; 
Morison['M2N2FAFxi']          = False     # (N/m); ; 
Morison['M2N3FAFxi']          = False     # (N/m); ; 
Morison['M2N4FAFxi']          = False     # (N/m); ; 
Morison['M2N5FAFxi']          = False     # (N/m); ; 
Morison['M2N6FAFxi']          = False     # (N/m); ; 
Morison['M2N7FAFxi']          = False     # (N/m); ; 
Morison['M2N8FAFxi']          = False     # (N/m); ; 
Morison['M2N9FAFxi']          = False     # (N/m); ; 
Morison['M3N1FAFxi']          = False     # (N/m); ; 
Morison['M3N2FAFxi']          = False     # (N/m); ; 
Morison['M3N3FAFxi']          = False     # (N/m); ; 
Morison['M3N4FAFxi']          = False     # (N/m); ; 
Morison['M3N5FAFxi']          = False     # (N/m); ; 
Morison['M3N6FAFxi']          = False     # (N/m); ; 
Morison['M3N7FAFxi']          = False     # (N/m); ; 
Morison['M3N8FAFxi']          = False     # (N/m); ; 
Morison['M3N9FAFxi']          = False     # (N/m); ; 
Morison['M4N1FAFxi']          = False     # (N/m); ; 
Morison['M4N2FAFxi']          = False     # (N/m); ; 
Morison['M4N3FAFxi']          = False     # (N/m); ; 
Morison['M4N4FAFxi']          = False     # (N/m); ; 
Morison['M4N5FAFxi']          = False     # (N/m); ; 
Morison['M4N6FAFxi']          = False     # (N/m); ; 
Morison['M4N7FAFxi']          = False     # (N/m); ; 
Morison['M4N8FAFxi']          = False     # (N/m); ; 
Morison['M4N9FAFxi']          = False     # (N/m); ; 
Morison['M5N1FAFxi']          = False     # (N/m); ; 
Morison['M5N2FAFxi']          = False     # (N/m); ; 
Morison['M5N3FAFxi']          = False     # (N/m); ; 
Morison['M5N4FAFxi']          = False     # (N/m); ; 
Morison['M5N5FAFxi']          = False     # (N/m); ; 
Morison['M5N6FAFxi']          = False     # (N/m); ; 
Morison['M5N7FAFxi']          = False     # (N/m); ; 
Morison['M5N8FAFxi']          = False     # (N/m); ; 
Morison['M5N9FAFxi']          = False     # (N/m); ; 
Morison['M6N1FAFxi']          = False     # (N/m); ; 
Morison['M6N2FAFxi']          = False     # (N/m); ; 
Morison['M6N3FAFxi']          = False     # (N/m); ; 
Morison['M6N4FAFxi']          = False     # (N/m); ; 
Morison['M6N5FAFxi']          = False     # (N/m); ; 
Morison['M6N6FAFxi']          = False     # (N/m); ; 
Morison['M6N7FAFxi']          = False     # (N/m); ; 
Morison['M6N8FAFxi']          = False     # (N/m); ; 
Morison['M6N9FAFxi']          = False     # (N/m); ; 
Morison['M7N1FAFxi']          = False     # (N/m); ; 
Morison['M7N2FAFxi']          = False     # (N/m); ; 
Morison['M7N3FAFxi']          = False     # (N/m); ; 
Morison['M7N4FAFxi']          = False     # (N/m); ; 
Morison['M7N5FAFxi']          = False     # (N/m); ; 
Morison['M7N6FAFxi']          = False     # (N/m); ; 
Morison['M7N7FAFxi']          = False     # (N/m); ; 
Morison['M7N8FAFxi']          = False     # (N/m); ; 
Morison['M7N9FAFxi']          = False     # (N/m); ; 
Morison['M8N1FAFxi']          = False     # (N/m); ; 
Morison['M8N2FAFxi']          = False     # (N/m); ; 
Morison['M8N3FAFxi']          = False     # (N/m); ; 
Morison['M8N4FAFxi']          = False     # (N/m); ; 
Morison['M8N5FAFxi']          = False     # (N/m); ; 
Morison['M8N6FAFxi']          = False     # (N/m); ; 
Morison['M8N7FAFxi']          = False     # (N/m); ; 
Morison['M8N8FAFxi']          = False     # (N/m); ; 
Morison['M8N9FAFxi']          = False     # (N/m); ; 
Morison['M9N1FAFxi']          = False     # (N/m); ; 
Morison['M9N2FAFxi']          = False     # (N/m); ; 
Morison['M9N3FAFxi']          = False     # (N/m); ; 
Morison['M9N4FAFxi']          = False     # (N/m); ; 
Morison['M9N5FAFxi']          = False     # (N/m); ; 
Morison['M9N6FAFxi']          = False     # (N/m); ; 
Morison['M9N7FAFxi']          = False     # (N/m); ; 
Morison['M9N8FAFxi']          = False     # (N/m); ; 
Morison['M9N9FAFxi']          = False     # (N/m); ; 
Morison['M1N1FAFyi']          = False     # (N/m); y-component of the distributed added mass force due to the ballast, expressed in the inertial coordinate system; 
Morison['M1N2FAFyi']          = False     # (N/m); ; 
Morison['M1N3FAFyi']          = False     # (N/m); ; 
Morison['M1N4FAFyi']          = False     # (N/m); ; 
Morison['M1N5FAFyi']          = False     # (N/m); ; 
Morison['M1N6FAFyi']          = False     # (N/m); ; 
Morison['M1N7FAFyi']          = False     # (N/m); ; 
Morison['M1N8FAFyi']          = False     # (N/m); ; 
Morison['M1N9FAFyi']          = False     # (N/m); ; 
Morison['M2N1FAFyi']          = False     # (N/m); ; 
Morison['M2N2FAFyi']          = False     # (N/m); ; 
Morison['M2N3FAFyi']          = False     # (N/m); ; 
Morison['M2N4FAFyi']          = False     # (N/m); ; 
Morison['M2N5FAFyi']          = False     # (N/m); ; 
Morison['M2N6FAFyi']          = False     # (N/m); ; 
Morison['M2N7FAFyi']          = False     # (N/m); ; 
Morison['M2N8FAFyi']          = False     # (N/m); ; 
Morison['M2N9FAFyi']          = False     # (N/m); ; 
Morison['M3N1FAFyi']          = False     # (N/m); ; 
Morison['M3N2FAFyi']          = False     # (N/m); ; 
Morison['M3N3FAFyi']          = False     # (N/m); ; 
Morison['M3N4FAFyi']          = False     # (N/m); ; 
Morison['M3N5FAFyi']          = False     # (N/m); ; 
Morison['M3N6FAFyi']          = False     # (N/m); ; 
Morison['M3N7FAFyi']          = False     # (N/m); ; 
Morison['M3N8FAFyi']          = False     # (N/m); ; 
Morison['M3N9FAFyi']          = False     # (N/m); ; 
Morison['M4N1FAFyi']          = False     # (N/m); ; 
Morison['M4N2FAFyi']          = False     # (N/m); ; 
Morison['M4N3FAFyi']          = False     # (N/m); ; 
Morison['M4N4FAFyi']          = False     # (N/m); ; 
Morison['M4N5FAFyi']          = False     # (N/m); ; 
Morison['M4N6FAFyi']          = False     # (N/m); ; 
Morison['M4N7FAFyi']          = False     # (N/m); ; 
Morison['M4N8FAFyi']          = False     # (N/m); ; 
Morison['M4N9FAFyi']          = False     # (N/m); ; 
Morison['M5N1FAFyi']          = False     # (N/m); ; 
Morison['M5N2FAFyi']          = False     # (N/m); ; 
Morison['M5N3FAFyi']          = False     # (N/m); ; 
Morison['M5N4FAFyi']          = False     # (N/m); ; 
Morison['M5N5FAFyi']          = False     # (N/m); ; 
Morison['M5N6FAFyi']          = False     # (N/m); ; 
Morison['M5N7FAFyi']          = False     # (N/m); ; 
Morison['M5N8FAFyi']          = False     # (N/m); ; 
Morison['M5N9FAFyi']          = False     # (N/m); ; 
Morison['M6N1FAFyi']          = False     # (N/m); ; 
Morison['M6N2FAFyi']          = False     # (N/m); ; 
Morison['M6N3FAFyi']          = False     # (N/m); ; 
Morison['M6N4FAFyi']          = False     # (N/m); ; 
Morison['M6N5FAFyi']          = False     # (N/m); ; 
Morison['M6N6FAFyi']          = False     # (N/m); ; 
Morison['M6N7FAFyi']          = False     # (N/m); ; 
Morison['M6N8FAFyi']          = False     # (N/m); ; 
Morison['M6N9FAFyi']          = False     # (N/m); ; 
Morison['M7N1FAFyi']          = False     # (N/m); ; 
Morison['M7N2FAFyi']          = False     # (N/m); ; 
Morison['M7N3FAFyi']          = False     # (N/m); ; 
Morison['M7N4FAFyi']          = False     # (N/m); ; 
Morison['M7N5FAFyi']          = False     # (N/m); ; 
Morison['M7N6FAFyi']          = False     # (N/m); ; 
Morison['M7N7FAFyi']          = False     # (N/m); ; 
Morison['M7N8FAFyi']          = False     # (N/m); ; 
Morison['M7N9FAFyi']          = False     # (N/m); ; 
Morison['M8N1FAFyi']          = False     # (N/m); ; 
Morison['M8N2FAFyi']          = False     # (N/m); ; 
Morison['M8N3FAFyi']          = False     # (N/m); ; 
Morison['M8N4FAFyi']          = False     # (N/m); ; 
Morison['M8N5FAFyi']          = False     # (N/m); ; 
Morison['M8N6FAFyi']          = False     # (N/m); ; 
Morison['M8N7FAFyi']          = False     # (N/m); ; 
Morison['M8N8FAFyi']          = False     # (N/m); ; 
Morison['M8N9FAFyi']          = False     # (N/m); ; 
Morison['M9N1FAFyi']          = False     # (N/m); ; 
Morison['M9N2FAFyi']          = False     # (N/m); ; 
Morison['M9N3FAFyi']          = False     # (N/m); ; 
Morison['M9N4FAFyi']          = False     # (N/m); ; 
Morison['M9N5FAFyi']          = False     # (N/m); ; 
Morison['M9N6FAFyi']          = False     # (N/m); ; 
Morison['M9N7FAFyi']          = False     # (N/m); ; 
Morison['M9N8FAFyi']          = False     # (N/m); ; 
Morison['M9N9FAFyi']          = False     # (N/m); ; 
Morison['M1N1FAFzi']          = False     # (N/m); z-component of the distributed added mass force due to the ballast, expressed in the inertial coordinate system; 
Morison['M1N2FAFzi']          = False     # (N/m); ; 
Morison['M1N3FAFzi']          = False     # (N/m); ; 
Morison['M1N4FAFzi']          = False     # (N/m); ; 
Morison['M1N5FAFzi']          = False     # (N/m); ; 
Morison['M1N6FAFzi']          = False     # (N/m); ; 
Morison['M1N7FAFzi']          = False     # (N/m); ; 
Morison['M1N8FAFzi']          = False     # (N/m); ; 
Morison['M1N9FAFzi']          = False     # (N/m); ; 
Morison['M2N1FAFzi']          = False     # (N/m); ; 
Morison['M2N2FAFzi']          = False     # (N/m); ; 
Morison['M2N3FAFzi']          = False     # (N/m); ; 
Morison['M2N4FAFzi']          = False     # (N/m); ; 
Morison['M2N5FAFzi']          = False     # (N/m); ; 
Morison['M2N6FAFzi']          = False     # (N/m); ; 
Morison['M2N7FAFzi']          = False     # (N/m); ; 
Morison['M2N8FAFzi']          = False     # (N/m); ; 
Morison['M2N9FAFzi']          = False     # (N/m); ; 
Morison['M3N1FAFzi']          = False     # (N/m); ; 
Morison['M3N2FAFzi']          = False     # (N/m); ; 
Morison['M3N3FAFzi']          = False     # (N/m); ; 
Morison['M3N4FAFzi']          = False     # (N/m); ; 
Morison['M3N5FAFzi']          = False     # (N/m); ; 
Morison['M3N6FAFzi']          = False     # (N/m); ; 
Morison['M3N7FAFzi']          = False     # (N/m); ; 
Morison['M3N8FAFzi']          = False     # (N/m); ; 
Morison['M3N9FAFzi']          = False     # (N/m); ; 
Morison['M4N1FAFzi']          = False     # (N/m); ; 
Morison['M4N2FAFzi']          = False     # (N/m); ; 
Morison['M4N3FAFzi']          = False     # (N/m); ; 
Morison['M4N4FAFzi']          = False     # (N/m); ; 
Morison['M4N5FAFzi']          = False     # (N/m); ; 
Morison['M4N6FAFzi']          = False     # (N/m); ; 
Morison['M4N7FAFzi']          = False     # (N/m); ; 
Morison['M4N8FAFzi']          = False     # (N/m); ; 
Morison['M4N9FAFzi']          = False     # (N/m); ; 
Morison['M5N1FAFzi']          = False     # (N/m); ; 
Morison['M5N2FAFzi']          = False     # (N/m); ; 
Morison['M5N3FAFzi']          = False     # (N/m); ; 
Morison['M5N4FAFzi']          = False     # (N/m); ; 
Morison['M5N5FAFzi']          = False     # (N/m); ; 
Morison['M5N6FAFzi']          = False     # (N/m); ; 
Morison['M5N7FAFzi']          = False     # (N/m); ; 
Morison['M5N8FAFzi']          = False     # (N/m); ; 
Morison['M5N9FAFzi']          = False     # (N/m); ; 
Morison['M6N1FAFzi']          = False     # (N/m); ; 
Morison['M6N2FAFzi']          = False     # (N/m); ; 
Morison['M6N3FAFzi']          = False     # (N/m); ; 
Morison['M6N4FAFzi']          = False     # (N/m); ; 
Morison['M6N5FAFzi']          = False     # (N/m); ; 
Morison['M6N6FAFzi']          = False     # (N/m); ; 
Morison['M6N7FAFzi']          = False     # (N/m); ; 
Morison['M6N8FAFzi']          = False     # (N/m); ; 
Morison['M6N9FAFzi']          = False     # (N/m); ; 
Morison['M7N1FAFzi']          = False     # (N/m); ; 
Morison['M7N2FAFzi']          = False     # (N/m); ; 
Morison['M7N3FAFzi']          = False     # (N/m); ; 
Morison['M7N4FAFzi']          = False     # (N/m); ; 
Morison['M7N5FAFzi']          = False     # (N/m); ; 
Morison['M7N6FAFzi']          = False     # (N/m); ; 
Morison['M7N7FAFzi']          = False     # (N/m); ; 
Morison['M7N8FAFzi']          = False     # (N/m); ; 
Morison['M7N9FAFzi']          = False     # (N/m); ; 
Morison['M8N1FAFzi']          = False     # (N/m); ; 
Morison['M8N2FAFzi']          = False     # (N/m); ; 
Morison['M8N3FAFzi']          = False     # (N/m); ; 
Morison['M8N4FAFzi']          = False     # (N/m); ; 
Morison['M8N5FAFzi']          = False     # (N/m); ; 
Morison['M8N6FAFzi']          = False     # (N/m); ; 
Morison['M8N7FAFzi']          = False     # (N/m); ; 
Morison['M8N8FAFzi']          = False     # (N/m); ; 
Morison['M8N9FAFzi']          = False     # (N/m); ; 
Morison['M9N1FAFzi']          = False     # (N/m); ; 
Morison['M9N2FAFzi']          = False     # (N/m); ; 
Morison['M9N3FAFzi']          = False     # (N/m); ; 
Morison['M9N4FAFzi']          = False     # (N/m); ; 
Morison['M9N5FAFzi']          = False     # (N/m); ; 
Morison['M9N6FAFzi']          = False     # (N/m); ; 
Morison['M9N7FAFzi']          = False     # (N/m); ; 
Morison['M9N8FAFzi']          = False     # (N/m); ; 
Morison['M9N9FAFzi']          = False     # (N/m); ; 
Morison['M1N1FAxi']           = False     # (N/m); x-component of the combined distributed added mass force, expressed in the inertial coordinate system; 
Morison['M1N2FAxi']           = False     # (N/m); ; 
Morison['M1N3FAxi']           = False     # (N/m); ; 
Morison['M1N4FAxi']           = False     # (N/m); ; 
Morison['M1N5FAxi']           = False     # (N/m); ; 
Morison['M1N6FAxi']           = False     # (N/m); ; 
Morison['M1N7FAxi']           = False     # (N/m); ; 
Morison['M1N8FAxi']           = False     # (N/m); ; 
Morison['M1N9FAxi']           = False     # (N/m); ; 
Morison['M2N1FAxi']           = False     # (N/m); ; 
Morison['M2N2FAxi']           = False     # (N/m); ; 
Morison['M2N3FAxi']           = False     # (N/m); ; 
Morison['M2N4FAxi']           = False     # (N/m); ; 
Morison['M2N5FAxi']           = False     # (N/m); ; 
Morison['M2N6FAxi']           = False     # (N/m); ; 
Morison['M2N7FAxi']           = False     # (N/m); ; 
Morison['M2N8FAxi']           = False     # (N/m); ; 
Morison['M2N9FAxi']           = False     # (N/m); ; 
Morison['M3N1FAxi']           = False     # (N/m); ; 
Morison['M3N2FAxi']           = False     # (N/m); ; 
Morison['M3N3FAxi']           = False     # (N/m); ; 
Morison['M3N4FAxi']           = False     # (N/m); ; 
Morison['M3N5FAxi']           = False     # (N/m); ; 
Morison['M3N6FAxi']           = False     # (N/m); ; 
Morison['M3N7FAxi']           = False     # (N/m); ; 
Morison['M3N8FAxi']           = False     # (N/m); ; 
Morison['M3N9FAxi']           = False     # (N/m); ; 
Morison['M4N1FAxi']           = False     # (N/m); ; 
Morison['M4N2FAxi']           = False     # (N/m); ; 
Morison['M4N3FAxi']           = False     # (N/m); ; 
Morison['M4N4FAxi']           = False     # (N/m); ; 
Morison['M4N5FAxi']           = False     # (N/m); ; 
Morison['M4N6FAxi']           = False     # (N/m); ; 
Morison['M4N7FAxi']           = False     # (N/m); ; 
Morison['M4N8FAxi']           = False     # (N/m); ; 
Morison['M4N9FAxi']           = False     # (N/m); ; 
Morison['M5N1FAxi']           = False     # (N/m); ; 
Morison['M5N2FAxi']           = False     # (N/m); ; 
Morison['M5N3FAxi']           = False     # (N/m); ; 
Morison['M5N4FAxi']           = False     # (N/m); ; 
Morison['M5N5FAxi']           = False     # (N/m); ; 
Morison['M5N6FAxi']           = False     # (N/m); ; 
Morison['M5N7FAxi']           = False     # (N/m); ; 
Morison['M5N8FAxi']           = False     # (N/m); ; 
Morison['M5N9FAxi']           = False     # (N/m); ; 
Morison['M6N1FAxi']           = False     # (N/m); ; 
Morison['M6N2FAxi']           = False     # (N/m); ; 
Morison['M6N3FAxi']           = False     # (N/m); ; 
Morison['M6N4FAxi']           = False     # (N/m); ; 
Morison['M6N5FAxi']           = False     # (N/m); ; 
Morison['M6N6FAxi']           = False     # (N/m); ; 
Morison['M6N7FAxi']           = False     # (N/m); ; 
Morison['M6N8FAxi']           = False     # (N/m); ; 
Morison['M6N9FAxi']           = False     # (N/m); ; 
Morison['M7N1FAxi']           = False     # (N/m); ; 
Morison['M7N2FAxi']           = False     # (N/m); ; 
Morison['M7N3FAxi']           = False     # (N/m); ; 
Morison['M7N4FAxi']           = False     # (N/m); ; 
Morison['M7N5FAxi']           = False     # (N/m); ; 
Morison['M7N6FAxi']           = False     # (N/m); ; 
Morison['M7N7FAxi']           = False     # (N/m); ; 
Morison['M7N8FAxi']           = False     # (N/m); ; 
Morison['M7N9FAxi']           = False     # (N/m); ; 
Morison['M8N1FAxi']           = False     # (N/m); ; 
Morison['M8N2FAxi']           = False     # (N/m); ; 
Morison['M8N3FAxi']           = False     # (N/m); ; 
Morison['M8N4FAxi']           = False     # (N/m); ; 
Morison['M8N5FAxi']           = False     # (N/m); ; 
Morison['M8N6FAxi']           = False     # (N/m); ; 
Morison['M8N7FAxi']           = False     # (N/m); ; 
Morison['M8N8FAxi']           = False     # (N/m); ; 
Morison['M8N9FAxi']           = False     # (N/m); ; 
Morison['M9N1FAxi']           = False     # (N/m); ; 
Morison['M9N2FAxi']           = False     # (N/m); ; 
Morison['M9N3FAxi']           = False     # (N/m); ; 
Morison['M9N4FAxi']           = False     # (N/m); ; 
Morison['M9N5FAxi']           = False     # (N/m); ; 
Morison['M9N6FAxi']           = False     # (N/m); ; 
Morison['M9N7FAxi']           = False     # (N/m); ; 
Morison['M9N8FAxi']           = False     # (N/m); ; 
Morison['M9N9FAxi']           = False     # (N/m); ; 
Morison['M1N1FAyi']           = False     # (N/m); y-component of the combined distributed added mass force, expressed in the inertial coordinate system; 
Morison['M1N2FAyi']           = False     # (N/m); ; 
Morison['M1N3FAyi']           = False     # (N/m); ; 
Morison['M1N4FAyi']           = False     # (N/m); ; 
Morison['M1N5FAyi']           = False     # (N/m); ; 
Morison['M1N6FAyi']           = False     # (N/m); ; 
Morison['M1N7FAyi']           = False     # (N/m); ; 
Morison['M1N8FAyi']           = False     # (N/m); ; 
Morison['M1N9FAyi']           = False     # (N/m); ; 
Morison['M2N1FAyi']           = False     # (N/m); ; 
Morison['M2N2FAyi']           = False     # (N/m); ; 
Morison['M2N3FAyi']           = False     # (N/m); ; 
Morison['M2N4FAyi']           = False     # (N/m); ; 
Morison['M2N5FAyi']           = False     # (N/m); ; 
Morison['M2N6FAyi']           = False     # (N/m); ; 
Morison['M2N7FAyi']           = False     # (N/m); ; 
Morison['M2N8FAyi']           = False     # (N/m); ; 
Morison['M2N9FAyi']           = False     # (N/m); ; 
Morison['M3N1FAyi']           = False     # (N/m); ; 
Morison['M3N2FAyi']           = False     # (N/m); ; 
Morison['M3N3FAyi']           = False     # (N/m); ; 
Morison['M3N4FAyi']           = False     # (N/m); ; 
Morison['M3N5FAyi']           = False     # (N/m); ; 
Morison['M3N6FAyi']           = False     # (N/m); ; 
Morison['M3N7FAyi']           = False     # (N/m); ; 
Morison['M3N8FAyi']           = False     # (N/m); ; 
Morison['M3N9FAyi']           = False     # (N/m); ; 
Morison['M4N1FAyi']           = False     # (N/m); ; 
Morison['M4N2FAyi']           = False     # (N/m); ; 
Morison['M4N3FAyi']           = False     # (N/m); ; 
Morison['M4N4FAyi']           = False     # (N/m); ; 
Morison['M4N5FAyi']           = False     # (N/m); ; 
Morison['M4N6FAyi']           = False     # (N/m); ; 
Morison['M4N7FAyi']           = False     # (N/m); ; 
Morison['M4N8FAyi']           = False     # (N/m); ; 
Morison['M4N9FAyi']           = False     # (N/m); ; 
Morison['M5N1FAyi']           = False     # (N/m); ; 
Morison['M5N2FAyi']           = False     # (N/m); ; 
Morison['M5N3FAyi']           = False     # (N/m); ; 
Morison['M5N4FAyi']           = False     # (N/m); ; 
Morison['M5N5FAyi']           = False     # (N/m); ; 
Morison['M5N6FAyi']           = False     # (N/m); ; 
Morison['M5N7FAyi']           = False     # (N/m); ; 
Morison['M5N8FAyi']           = False     # (N/m); ; 
Morison['M5N9FAyi']           = False     # (N/m); ; 
Morison['M6N1FAyi']           = False     # (N/m); ; 
Morison['M6N2FAyi']           = False     # (N/m); ; 
Morison['M6N3FAyi']           = False     # (N/m); ; 
Morison['M6N4FAyi']           = False     # (N/m); ; 
Morison['M6N5FAyi']           = False     # (N/m); ; 
Morison['M6N6FAyi']           = False     # (N/m); ; 
Morison['M6N7FAyi']           = False     # (N/m); ; 
Morison['M6N8FAyi']           = False     # (N/m); ; 
Morison['M6N9FAyi']           = False     # (N/m); ; 
Morison['M7N1FAyi']           = False     # (N/m); ; 
Morison['M7N2FAyi']           = False     # (N/m); ; 
Morison['M7N3FAyi']           = False     # (N/m); ; 
Morison['M7N4FAyi']           = False     # (N/m); ; 
Morison['M7N5FAyi']           = False     # (N/m); ; 
Morison['M7N6FAyi']           = False     # (N/m); ; 
Morison['M7N7FAyi']           = False     # (N/m); ; 
Morison['M7N8FAyi']           = False     # (N/m); ; 
Morison['M7N9FAyi']           = False     # (N/m); ; 
Morison['M8N1FAyi']           = False     # (N/m); ; 
Morison['M8N2FAyi']           = False     # (N/m); ; 
Morison['M8N3FAyi']           = False     # (N/m); ; 
Morison['M8N4FAyi']           = False     # (N/m); ; 
Morison['M8N5FAyi']           = False     # (N/m); ; 
Morison['M8N6FAyi']           = False     # (N/m); ; 
Morison['M8N7FAyi']           = False     # (N/m); ; 
Morison['M8N8FAyi']           = False     # (N/m); ; 
Morison['M8N9FAyi']           = False     # (N/m); ; 
Morison['M9N1FAyi']           = False     # (N/m); ; 
Morison['M9N2FAyi']           = False     # (N/m); ; 
Morison['M9N3FAyi']           = False     # (N/m); ; 
Morison['M9N4FAyi']           = False     # (N/m); ; 
Morison['M9N5FAyi']           = False     # (N/m); ; 
Morison['M9N6FAyi']           = False     # (N/m); ; 
Morison['M9N7FAyi']           = False     # (N/m); ; 
Morison['M9N8FAyi']           = False     # (N/m); ; 
Morison['M9N9FAyi']           = False     # (N/m); ; 
Morison['M1N1FAzi']           = False     # (N/m); z-component of the combined distributed added mass force, expressed in the inertial coordinate system; 
Morison['M1N2FAzi']           = False     # (N/m); ; 
Morison['M1N3FAzi']           = False     # (N/m); ; 
Morison['M1N4FAzi']           = False     # (N/m); ; 
Morison['M1N5FAzi']           = False     # (N/m); ; 
Morison['M1N6FAzi']           = False     # (N/m); ; 
Morison['M1N7FAzi']           = False     # (N/m); ; 
Morison['M1N8FAzi']           = False     # (N/m); ; 
Morison['M1N9FAzi']           = False     # (N/m); ; 
Morison['M2N1FAzi']           = False     # (N/m); ; 
Morison['M2N2FAzi']           = False     # (N/m); ; 
Morison['M2N3FAzi']           = False     # (N/m); ; 
Morison['M2N4FAzi']           = False     # (N/m); ; 
Morison['M2N5FAzi']           = False     # (N/m); ; 
Morison['M2N6FAzi']           = False     # (N/m); ; 
Morison['M2N7FAzi']           = False     # (N/m); ; 
Morison['M2N8FAzi']           = False     # (N/m); ; 
Morison['M2N9FAzi']           = False     # (N/m); ; 
Morison['M3N1FAzi']           = False     # (N/m); ; 
Morison['M3N2FAzi']           = False     # (N/m); ; 
Morison['M3N3FAzi']           = False     # (N/m); ; 
Morison['M3N4FAzi']           = False     # (N/m); ; 
Morison['M3N5FAzi']           = False     # (N/m); ; 
Morison['M3N6FAzi']           = False     # (N/m); ; 
Morison['M3N7FAzi']           = False     # (N/m); ; 
Morison['M3N8FAzi']           = False     # (N/m); ; 
Morison['M3N9FAzi']           = False     # (N/m); ; 
Morison['M4N1FAzi']           = False     # (N/m); ; 
Morison['M4N2FAzi']           = False     # (N/m); ; 
Morison['M4N3FAzi']           = False     # (N/m); ; 
Morison['M4N4FAzi']           = False     # (N/m); ; 
Morison['M4N5FAzi']           = False     # (N/m); ; 
Morison['M4N6FAzi']           = False     # (N/m); ; 
Morison['M4N7FAzi']           = False     # (N/m); ; 
Morison['M4N8FAzi']           = False     # (N/m); ; 
Morison['M4N9FAzi']           = False     # (N/m); ; 
Morison['M5N1FAzi']           = False     # (N/m); ; 
Morison['M5N2FAzi']           = False     # (N/m); ; 
Morison['M5N3FAzi']           = False     # (N/m); ; 
Morison['M5N4FAzi']           = False     # (N/m); ; 
Morison['M5N5FAzi']           = False     # (N/m); ; 
Morison['M5N6FAzi']           = False     # (N/m); ; 
Morison['M5N7FAzi']           = False     # (N/m); ; 
Morison['M5N8FAzi']           = False     # (N/m); ; 
Morison['M5N9FAzi']           = False     # (N/m); ; 
Morison['M6N1FAzi']           = False     # (N/m); ; 
Morison['M6N2FAzi']           = False     # (N/m); ; 
Morison['M6N3FAzi']           = False     # (N/m); ; 
Morison['M6N4FAzi']           = False     # (N/m); ; 
Morison['M6N5FAzi']           = False     # (N/m); ; 
Morison['M6N6FAzi']           = False     # (N/m); ; 
Morison['M6N7FAzi']           = False     # (N/m); ; 
Morison['M6N8FAzi']           = False     # (N/m); ; 
Morison['M6N9FAzi']           = False     # (N/m); ; 
Morison['M7N1FAzi']           = False     # (N/m); ; 
Morison['M7N2FAzi']           = False     # (N/m); ; 
Morison['M7N3FAzi']           = False     # (N/m); ; 
Morison['M7N4FAzi']           = False     # (N/m); ; 
Morison['M7N5FAzi']           = False     # (N/m); ; 
Morison['M7N6FAzi']           = False     # (N/m); ; 
Morison['M7N7FAzi']           = False     # (N/m); ; 
Morison['M7N8FAzi']           = False     # (N/m); ; 
Morison['M7N9FAzi']           = False     # (N/m); ; 
Morison['M8N1FAzi']           = False     # (N/m); ; 
Morison['M8N2FAzi']           = False     # (N/m); ; 
Morison['M8N3FAzi']           = False     # (N/m); ; 
Morison['M8N4FAzi']           = False     # (N/m); ; 
Morison['M8N5FAzi']           = False     # (N/m); ; 
Morison['M8N6FAzi']           = False     # (N/m); ; 
Morison['M8N7FAzi']           = False     # (N/m); ; 
Morison['M8N8FAzi']           = False     # (N/m); ; 
Morison['M8N9FAzi']           = False     # (N/m); ; 
Morison['M9N1FAzi']           = False     # (N/m); ; 
Morison['M9N2FAzi']           = False     # (N/m); ; 
Morison['M9N3FAzi']           = False     # (N/m); ; 
Morison['M9N4FAzi']           = False     # (N/m); ; 
Morison['M9N5FAzi']           = False     # (N/m); ; 
Morison['M9N6FAzi']           = False     # (N/m); ; 
Morison['M9N7FAzi']           = False     # (N/m); ; 
Morison['M9N8FAzi']           = False     # (N/m); ; 
Morison['M9N9FAzi']           = False     # (N/m); ; 

# Joint-level Wave Kinematics 
Morison['J1Vxi']              = False     # (m/s); fluid velocity at the joint; 
Morison['J2Vxi']              = False     # (m/s); ; 
Morison['J3Vxi']              = False     # (m/s); ; 
Morison['J4Vxi']              = False     # (m/s); ; 
Morison['J5Vxi']              = False     # (m/s); ; 
Morison['J6Vxi']              = False     # (m/s); ; 
Morison['J7Vxi']              = False     # (m/s); ; 
Morison['J8Vxi']              = False     # (m/s); ; 
Morison['J9Vxi']              = False     # (m/s); ; 
Morison['J1Vyi']              = False     # (m/s); ; 
Morison['J2Vyi']              = False     # (m/s); ; 
Morison['J3Vyi']              = False     # (m/s); ; 
Morison['J4Vyi']              = False     # (m/s); ; 
Morison['J5Vyi']              = False     # (m/s); ; 
Morison['J6Vyi']              = False     # (m/s); ; 
Morison['J7Vyi']              = False     # (m/s); ; 
Morison['J8Vyi']              = False     # (m/s); ; 
Morison['J9Vyi']              = False     # (m/s); ; 
Morison['J1Vzi']              = False     # (m/s); ; 
Morison['J2Vzi']              = False     # (m/s); ; 
Morison['J3Vzi']              = False     # (m/s); ; 
Morison['J4Vzi']              = False     # (m/s); ; 
Morison['J5Vzi']              = False     # (m/s); ; 
Morison['J6Vzi']              = False     # (m/s); ; 
Morison['J7Vzi']              = False     # (m/s); ; 
Morison['J8Vzi']              = False     # (m/s); ; 
Morison['J9Vzi']              = False     # (m/s); ; 
Morison['J1Axi']              = False     # (m/s^2); fluid acceleration at the joint; 
Morison['J2Axi']              = False     # (m/s^2); ; 
Morison['J3Axi']              = False     # (m/s^2); ; 
Morison['J4Axi']              = False     # (m/s^2); ; 
Morison['J5Axi']              = False     # (m/s^2); ; 
Morison['J6Axi']              = False     # (m/s^2); ; 
Morison['J7Axi']              = False     # (m/s^2); ; 
Morison['J8Axi']              = False     # (m/s^2); ; 
Morison['J9Axi']              = False     # (m/s^2); ; 
Morison['J1Ayi']              = False     # (m/s^2); ; 
Morison['J2Ayi']              = False     # (m/s^2); ; 
Morison['J3Ayi']              = False     # (m/s^2); ; 
Morison['J4Ayi']              = False     # (m/s^2); ; 
Morison['J5Ayi']              = False     # (m/s^2); ; 
Morison['J6Ayi']              = False     # (m/s^2); ; 
Morison['J7Ayi']              = False     # (m/s^2); ; 
Morison['J8Ayi']              = False     # (m/s^2); ; 
Morison['J9Ayi']              = False     # (m/s^2); ; 
Morison['J1Azi']              = False     # (m/s^2); ; 
Morison['J2Azi']              = False     # (m/s^2); ; 
Morison['J3Azi']              = False     # (m/s^2); ; 
Morison['J4Azi']              = False     # (m/s^2); ; 
Morison['J5Azi']              = False     # (m/s^2); ; 
Morison['J6Azi']              = False     # (m/s^2); ; 
Morison['J7Azi']              = False     # (m/s^2); ; 
Morison['J8Azi']              = False     # (m/s^2); ; 
Morison['J9Azi']              = False     # (m/s^2); ; 
Morison['J1DynP']             = False     # (Pa); fluid dynamic pressure at the joint; 
Morison['J2DynP']             = False     # (Pa); ; 
Morison['J3DynP']             = False     # (Pa); ; 
Morison['J4DynP']             = False     # (Pa); ; 
Morison['J5DynP']             = False     # (Pa); ; 
Morison['J6DynP']             = False     # (Pa); ; 
Morison['J7DynP']             = False     # (Pa); ; 
Morison['J8DynP']             = False     # (Pa); ; 
Morison['J9DynP']             = False     # (Pa); ; 
Morison['J1STVxi']            = False     # (m/s); structural translational velocity at the joint; 
Morison['J2STVxi']            = False     # (m/s); ; 
Morison['J3STVxi']            = False     # (m/s); ; 
Morison['J4STVxi']            = False     # (m/s); ; 
Morison['J5STVxi']            = False     # (m/s); ; 
Morison['J6STVxi']            = False     # (m/s); ; 
Morison['J7STVxi']            = False     # (m/s); ; 
Morison['J8STVxi']            = False     # (m/s); ; 
Morison['J9STVxi']            = False     # (m/s); ; 
Morison['J1STVyi']            = False     # (m/s); ; 
Morison['J2STVyi']            = False     # (m/s); ; 
Morison['J3STVyi']            = False     # (m/s); ; 
Morison['J4STVyi']            = False     # (m/s); ; 
Morison['J5STVyi']            = False     # (m/s); ; 
Morison['J6STVyi']            = False     # (m/s); ; 
Morison['J7STVyi']            = False     # (m/s); ; 
Morison['J8STVyi']            = False     # (m/s); ; 
Morison['J9STVyi']            = False     # (m/s); ; 
Morison['J1STVzi']            = False     # (m/s); ; 
Morison['J2STVzi']            = False     # (m/s); ; 
Morison['J3STVzi']            = False     # (m/s); ; 
Morison['J4STVzi']            = False     # (m/s); ; 
Morison['J5STVzi']            = False     # (m/s); ; 
Morison['J6STVzi']            = False     # (m/s); ; 
Morison['J7STVzi']            = False     # (m/s); ; 
Morison['J8STVzi']            = False     # (m/s); ; 
Morison['J9STVzi']            = False     # (m/s); ; 
Morison['J1STAxi']            = False     # (m/s^2); structural translational acceleration at the joint; 
Morison['J2STAxi']            = False     # (m/s^2); ; 
Morison['J3STAxi']            = False     # (m/s^2); ; 
Morison['J4STAxi']            = False     # (m/s^2); ; 
Morison['J5STAxi']            = False     # (m/s^2); ; 
Morison['J6STAxi']            = False     # (m/s^2); ; 
Morison['J7STAxi']            = False     # (m/s^2); ; 
Morison['J8STAxi']            = False     # (m/s^2); ; 
Morison['J9STAxi']            = False     # (m/s^2); ; 
Morison['J1STAyi']            = False     # (m/s^2); ; 
Morison['J2STAyi']            = False     # (m/s^2); ; 
Morison['J3STAyi']            = False     # (m/s^2); ; 
Morison['J4STAyi']            = False     # (m/s^2); ; 
Morison['J5STAyi']            = False     # (m/s^2); ; 
Morison['J6STAyi']            = False     # (m/s^2); ; 
Morison['J7STAyi']            = False     # (m/s^2); ; 
Morison['J8STAyi']            = False     # (m/s^2); ; 
Morison['J9STAyi']            = False     # (m/s^2); ; 
Morison['J1STAzi']            = False     # (m/s^2); ; 
Morison['J2STAzi']            = False     # (m/s^2); ; 
Morison['J3STAzi']            = False     # (m/s^2); ; 
Morison['J4STAzi']            = False     # (m/s^2); ; 
Morison['J5STAzi']            = False     # (m/s^2); ; 
Morison['J6STAzi']            = False     # (m/s^2); ; 
Morison['J7STAzi']            = False     # (m/s^2); ; 
Morison['J8STAzi']            = False     # (m/s^2); ; 
Morison['J9STAzi']            = False     # (m/s^2); ; 

# Joint Loads
Morison['J1FDxi']             = False     # (N); axial drag forces; 
Morison['J2FDxi']             = False     # (N); ; 
Morison['J3FDxi']             = False     # (N); ; 
Morison['J4FDxi']             = False     # (N); ; 
Morison['J5FDxi']             = False     # (N); ; 
Morison['J6FDxi']             = False     # (N); ; 
Morison['J7FDxi']             = False     # (N); ; 
Morison['J8FDxi']             = False     # (N); ; 
Morison['J9FDxi']             = False     # (N); ; 
Morison['J1FDyi']             = False     # (N); ; 
Morison['J2FDyi']             = False     # (N); ; 
Morison['J3FDyi']             = False     # (N); ; 
Morison['J4FDyi']             = False     # (N); ; 
Morison['J5FDyi']             = False     # (N); ; 
Morison['J6FDyi']             = False     # (N); ; 
Morison['J7FDyi']             = False     # (N); ; 
Morison['J8FDyi']             = False     # (N); ; 
Morison['J9FDyi']             = False     # (N); ; 
Morison['J1FDzi']             = False     # (N); ; 
Morison['J2FDzi']             = False     # (N); ; 
Morison['J3FDzi']             = False     # (N); ; 
Morison['J4FDzi']             = False     # (N); ; 
Morison['J5FDzi']             = False     # (N); ; 
Morison['J6FDzi']             = False     # (N); ; 
Morison['J7FDzi']             = False     # (N); ; 
Morison['J8FDzi']             = False     # (N); ; 
Morison['J9FDzi']             = False     # (N); ; 
Morison['J1FBxi']             = False     # (N); end-effect buoyancy forces; 
Morison['J2FBxi']             = False     # (N); ; 
Morison['J3FBxi']             = False     # (N); ; 
Morison['J4FBxi']             = False     # (N); ; 
Morison['J5FBxi']             = False     # (N); ; 
Morison['J6FBxi']             = False     # (N); ; 
Morison['J7FBxi']             = False     # (N); ; 
Morison['J8FBxi']             = False     # (N); ; 
Morison['J9FBxi']             = False     # (N); ; 
Morison['J1FByi']             = False     # (N); ; 
Morison['J2FByi']             = False     # (N); ; 
Morison['J3FByi']             = False     # (N); ; 
Morison['J4FByi']             = False     # (N); ; 
Morison['J5FByi']             = False     # (N); ; 
Morison['J6FByi']             = False     # (N); ; 
Morison['J7FByi']             = False     # (N); ; 
Morison['J8FByi']             = False     # (N); ; 
Morison['J9FByi']             = False     # (N); ; 
Morison['J1FBzi']             = False     # (N); ; 
Morison['J2FBzi']             = False     # (N); ; 
Morison['J3FBzi']             = False     # (N); ; 
Morison['J4FBzi']             = False     # (N); ; 
Morison['J5FBzi']             = False     # (N); ; 
Morison['J6FBzi']             = False     # (N); ; 
Morison['J7FBzi']             = False     # (N); ; 
Morison['J8FBzi']             = False     # (N); ; 
Morison['J9FBzi']             = False     # (N); ; 
Morison['J1MBxi']             = False     # (N-m); end-effect buoyancy moments; 
Morison['J2MBxi']             = False     # (N-m); ; 
Morison['J3MBxi']             = False     # (N-m); ; 
Morison['J4MBxi']             = False     # (N-m); ; 
Morison['J5MBxi']             = False     # (N-m); ; 
Morison['J6MBxi']             = False     # (N-m); ; 
Morison['J7MBxi']             = False     # (N-m); ; 
Morison['J8MBxi']             = False     # (N-m); ; 
Morison['J9MBxi']             = False     # (N-m); ; 
Morison['J1MByi']             = False     # (N-m); ; 
Morison['J2MByi']             = False     # (N-m); ; 
Morison['J3MByi']             = False     # (N-m); ; 
Morison['J4MByi']             = False     # (N-m); ; 
Morison['J5MByi']             = False     # (N-m); ; 
Morison['J6MByi']             = False     # (N-m); ; 
Morison['J7MByi']             = False     # (N-m); ; 
Morison['J8MByi']             = False     # (N-m); ; 
Morison['J9MByi']             = False     # (N-m); ; 
Morison['J1MBzi']             = False     # (N-m); ; 
Morison['J2MBzi']             = False     # (N-m); ; 
Morison['J3MBzi']             = False     # (N-m); ; 
Morison['J4MBzi']             = False     # (N-m); ; 
Morison['J5MBzi']             = False     # (N-m); ; 
Morison['J6MBzi']             = False     # (N-m); ; 
Morison['J7MBzi']             = False     # (N-m); ; 
Morison['J8MBzi']             = False     # (N-m); ; 
Morison['J9MBzi']             = False     # (N-m); ; 
Morison['J1FBFxi']            = False     # (N); end-effect buoyancy forces due to ballasting/flooding; 
Morison['J2FBFxi']            = False     # (N); ; 
Morison['J3FBFxi']            = False     # (N); ; 
Morison['J4FBFxi']            = False     # (N); ; 
Morison['J5FBFxi']            = False     # (N); ; 
Morison['J6FBFxi']            = False     # (N); ; 
Morison['J7FBFxi']            = False     # (N); ; 
Morison['J8FBFxi']            = False     # (N); ; 
Morison['J9FBFxi']            = False     # (N); ; 
Morison['J1FBFyi']            = False     # (N); ; 
Morison['J2FBFyi']            = False     # (N); ; 
Morison['J3FBFyi']            = False     # (N); ; 
Morison['J4FBFyi']            = False     # (N); ; 
Morison['J5FBFyi']            = False     # (N); ; 
Morison['J6FBFyi']            = False     # (N); ; 
Morison['J7FBFyi']            = False     # (N); ; 
Morison['J8FBFyi']            = False     # (N); ; 
Morison['J9FBFyi']            = False     # (N); ; 
Morison['J1FBFzi']            = False     # (N); ; 
Morison['J2FBFzi']            = False     # (N); ; 
Morison['J3FBFzi']            = False     # (N); ; 
Morison['J4FBFzi']            = False     # (N); ; 
Morison['J5FBFzi']            = False     # (N); ; 
Morison['J6FBFzi']            = False     # (N); ; 
Morison['J7FBFzi']            = False     # (N); ; 
Morison['J8FBFzi']            = False     # (N); ; 
Morison['J9FBFzi']            = False     # (N); ; 
Morison['J1MBFxi']            = False     # (N-m); end-effect buoyancy moments due to ballasting/flooding; 
Morison['J2MBFxi']            = False     # (N-m); ; 
Morison['J3MBFxi']            = False     # (N-m); ; 
Morison['J4MBFxi']            = False     # (N-m); ; 
Morison['J5MBFxi']            = False     # (N-m); ; 
Morison['J6MBFxi']            = False     # (N-m); ; 
Morison['J7MBFxi']            = False     # (N-m); ; 
Morison['J8MBFxi']            = False     # (N-m); ; 
Morison['J9MBFxi']            = False     # (N-m); ; 
Morison['J1MBFyi']            = False     # (N-m); ; 
Morison['J2MBFyi']            = False     # (N-m); ; 
Morison['J3MBFyi']            = False     # (N-m); ; 
Morison['J4MBFyi']            = False     # (N-m); ; 
Morison['J5MBFyi']            = False     # (N-m); ; 
Morison['J6MBFyi']            = False     # (N-m); ; 
Morison['J7MBFyi']            = False     # (N-m); ; 
Morison['J8MBFyi']            = False     # (N-m); ; 
Morison['J9MBFyi']            = False     # (N-m); ; 
Morison['J1MBFzi']            = False     # (N-m); ; 
Morison['J2MBFzi']            = False     # (N-m); ; 
Morison['J3MBFzi']            = False     # (N-m); ; 
Morison['J4MBFzi']            = False     # (N-m); ; 
Morison['J5MBFzi']            = False     # (N-m); ; 
Morison['J6MBFzi']            = False     # (N-m); ; 
Morison['J7MBFzi']            = False     # (N-m); ; 
Morison['J8MBFzi']            = False     # (N-m); ; 
Morison['J9MBFzi']            = False     # (N-m); ; 
Morison['J1FIxi']             = False     # (N); end-effect inertial forces ; 
Morison['J2FIxi']             = False     # (N); ; 
Morison['J3FIxi']             = False     # (N); ; 
Morison['J4FIxi']             = False     # (N); ; 
Morison['J5FIxi']             = False     # (N); ; 
Morison['J6FIxi']             = False     # (N); ; 
Morison['J7FIxi']             = False     # (N); ; 
Morison['J8FIxi']             = False     # (N); ; 
Morison['J9FIxi']             = False     # (N); ; 
Morison['J1FIyi']             = False     # (N); ; 
Morison['J2FIyi']             = False     # (N); ; 
Morison['J3FIyi']             = False     # (N); ; 
Morison['J4FIyi']             = False     # (N); ; 
Morison['J5FIyi']             = False     # (N); ; 
Morison['J6FIyi']             = False     # (N); ; 
Morison['J7FIyi']             = False     # (N); ; 
Morison['J8FIyi']             = False     # (N); ; 
Morison['J9FIyi']             = False     # (N); ; 
Morison['J1FIzi']             = False     # (N); ; 
Morison['J2FIzi']             = False     # (N); ; 
Morison['J3FIzi']             = False     # (N); ; 
Morison['J4FIzi']             = False     # (N); ; 
Morison['J5FIzi']             = False     # (N); ; 
Morison['J6FIzi']             = False     # (N); ; 
Morison['J7FIzi']             = False     # (N); ; 
Morison['J8FIzi']             = False     # (N); ; 
Morison['J9FIzi']             = False     # (N); ; 
Morison['J1FAMxi']            = False     # (N); added mass forces; 
Morison['J2FAMxi']            = False     # (N); ; 
Morison['J3FAMxi']            = False     # (N); ; 
Morison['J4FAMxi']            = False     # (N); ; 
Morison['J5FAMxi']            = False     # (N); ; 
Morison['J6FAMxi']            = False     # (N); ; 
Morison['J7FAMxi']            = False     # (N); ; 
Morison['J8FAMxi']            = False     # (N); ; 
Morison['J9FAMxi']            = False     # (N); ; 
Morison['J1FAMyi']            = False     # (N); ; 
Morison['J2FAMyi']            = False     # (N); ; 
Morison['J3FAMyi']            = False     # (N); ; 
Morison['J4FAMyi']            = False     # (N); ; 
Morison['J5FAMyi']            = False     # (N); ; 
Morison['J6FAMyi']            = False     # (N); ; 
Morison['J7FAMyi']            = False     # (N); ; 
Morison['J8FAMyi']            = False     # (N); ; 
Morison['J9FAMyi']            = False     # (N); ; 
Morison['J1FAMzi']            = False     # (N); ; 
Morison['J2FAMzi']            = False     # (N); ; 
Morison['J3FAMzi']            = False     # (N); ; 
Morison['J4FAMzi']            = False     # (N); ; 
Morison['J5FAMzi']            = False     # (N); ; 
Morison['J6FAMzi']            = False     # (N); ; 
Morison['J7FAMzi']            = False     # (N); ; 
Morison['J8FAMzi']            = False     # (N); ; 
Morison['J9FAMzi']            = False     # (N); ; 

""" SubDyn """
SubDyn = {}

# Member Forces
SubDyn['M1N1FKxe']            = False     # (N); xe component of the shear at Node Nj of member Mi- Local Reference System- Static Component; 
SubDyn['M1N2FKxe']            = False     # (N); ; 
SubDyn['M1N3FKxe']            = False     # (N); ; 
SubDyn['M1N4FKxe']            = False     # (N); ; 
SubDyn['M1N5FKxe']            = False     # (N); ; 
SubDyn['M1N6FKxe']            = False     # (N); ; 
SubDyn['M1N7FKxe']            = False     # (N); ; 
SubDyn['M1N8FKxe']            = False     # (N); ; 
SubDyn['M1N9FKxe']            = False     # (N); ; 
SubDyn['M2N1FKxe']            = False     # (N); ; 
SubDyn['M2N2FKxe']            = False     # (N); ; 
SubDyn['M2N3FKxe']            = False     # (N); ; 
SubDyn['M2N4FKxe']            = False     # (N); ; 
SubDyn['M2N5FKxe']            = False     # (N); ; 
SubDyn['M2N6FKxe']            = False     # (N); ; 
SubDyn['M2N7FKxe']            = False     # (N); ; 
SubDyn['M2N8FKxe']            = False     # (N); ; 
SubDyn['M2N9FKxe']            = False     # (N); ; 
SubDyn['M3N1FKxe']            = False     # (N); ; 
SubDyn['M3N2FKxe']            = False     # (N); ; 
SubDyn['M3N3FKxe']            = False     # (N); ; 
SubDyn['M3N4FKxe']            = False     # (N); ; 
SubDyn['M3N5FKxe']            = False     # (N); ; 
SubDyn['M3N6FKxe']            = False     # (N); ; 
SubDyn['M3N7FKxe']            = False     # (N); ; 
SubDyn['M3N8FKxe']            = False     # (N); ; 
SubDyn['M3N9FKxe']            = False     # (N); ; 
SubDyn['M4N1FKxe']            = False     # (N); ; 
SubDyn['M4N2FKxe']            = False     # (N); ; 
SubDyn['M4N3FKxe']            = False     # (N); ; 
SubDyn['M4N4FKxe']            = False     # (N); ; 
SubDyn['M4N5FKxe']            = False     # (N); ; 
SubDyn['M4N6FKxe']            = False     # (N); ; 
SubDyn['M4N7FKxe']            = False     # (N); ; 
SubDyn['M4N8FKxe']            = False     # (N); ; 
SubDyn['M4N9FKxe']            = False     # (N); ; 
SubDyn['M5N1FKxe']            = False     # (N); ; 
SubDyn['M5N2FKxe']            = False     # (N); ; 
SubDyn['M5N3FKxe']            = False     # (N); ; 
SubDyn['M5N4FKxe']            = False     # (N); ; 
SubDyn['M5N5FKxe']            = False     # (N); ; 
SubDyn['M5N6FKxe']            = False     # (N); ; 
SubDyn['M5N7FKxe']            = False     # (N); ; 
SubDyn['M5N8FKxe']            = False     # (N); ; 
SubDyn['M5N9FKxe']            = False     # (N); ; 
SubDyn['M6N1FKxe']            = False     # (N); ; 
SubDyn['M6N2FKxe']            = False     # (N); ; 
SubDyn['M6N3FKxe']            = False     # (N); ; 
SubDyn['M6N4FKxe']            = False     # (N); ; 
SubDyn['M6N5FKxe']            = False     # (N); ; 
SubDyn['M6N6FKxe']            = False     # (N); ; 
SubDyn['M6N7FKxe']            = False     # (N); ; 
SubDyn['M6N8FKxe']            = False     # (N); ; 
SubDyn['M6N9FKxe']            = False     # (N); ; 
SubDyn['M7N1FKxe']            = False     # (N); ; 
SubDyn['M7N2FKxe']            = False     # (N); ; 
SubDyn['M7N3FKxe']            = False     # (N); ; 
SubDyn['M7N4FKxe']            = False     # (N); ; 
SubDyn['M7N5FKxe']            = False     # (N); ; 
SubDyn['M7N6FKxe']            = False     # (N); ; 
SubDyn['M7N7FKxe']            = False     # (N); ; 
SubDyn['M7N8FKxe']            = False     # (N); ; 
SubDyn['M7N9FKxe']            = False     # (N); ; 
SubDyn['M8N1FKxe']            = False     # (N); ; 
SubDyn['M8N2FKxe']            = False     # (N); ; 
SubDyn['M8N3FKxe']            = False     # (N); ; 
SubDyn['M8N4FKxe']            = False     # (N); ; 
SubDyn['M8N5FKxe']            = False     # (N); ; 
SubDyn['M8N6FKxe']            = False     # (N); ; 
SubDyn['M8N7FKxe']            = False     # (N); ; 
SubDyn['M8N8FKxe']            = False     # (N); ; 
SubDyn['M8N9FKxe']            = False     # (N); ; 
SubDyn['M9N1FKxe']            = False     # (N); ; 
SubDyn['M9N2FKxe']            = False     # (N); ; 
SubDyn['M9N3FKxe']            = False     # (N); ; 
SubDyn['M9N4FKxe']            = False     # (N); ; 
SubDyn['M9N5FKxe']            = False     # (N); ; 
SubDyn['M9N6FKxe']            = False     # (N); ; 
SubDyn['M9N7FKxe']            = False     # (N); ; 
SubDyn['M9N8FKxe']            = False     # (N); ; 
SubDyn['M9N9FKxe']            = False     # (N); ; 
SubDyn['M1N1FKye']            = False     # (N); ye component of the shear at Node Nj of member Mi- Local Reference System- Static Component; 
SubDyn['M1N2FKye']            = False     # (N); ; 
SubDyn['M1N3FKye']            = False     # (N); ; 
SubDyn['M1N4FKye']            = False     # (N); ; 
SubDyn['M1N5FKye']            = False     # (N); ; 
SubDyn['M1N6FKye']            = False     # (N); ; 
SubDyn['M1N7FKye']            = False     # (N); ; 
SubDyn['M1N8FKye']            = False     # (N); ; 
SubDyn['M1N9FKye']            = False     # (N); ; 
SubDyn['M2N1FKye']            = False     # (N); ; 
SubDyn['M2N2FKye']            = False     # (N); ; 
SubDyn['M2N3FKye']            = False     # (N); ; 
SubDyn['M2N4FKye']            = False     # (N); ; 
SubDyn['M2N5FKye']            = False     # (N); ; 
SubDyn['M2N6FKye']            = False     # (N); ; 
SubDyn['M2N7FKye']            = False     # (N); ; 
SubDyn['M2N8FKye']            = False     # (N); ; 
SubDyn['M2N9FKye']            = False     # (N); ; 
SubDyn['M3N1FKye']            = False     # (N); ; 
SubDyn['M3N2FKye']            = False     # (N); ; 
SubDyn['M3N3FKye']            = False     # (N); ; 
SubDyn['M3N4FKye']            = False     # (N); ; 
SubDyn['M3N5FKye']            = False     # (N); ; 
SubDyn['M3N6FKye']            = False     # (N); ; 
SubDyn['M3N7FKye']            = False     # (N); ; 
SubDyn['M3N8FKye']            = False     # (N); ; 
SubDyn['M3N9FKye']            = False     # (N); ; 
SubDyn['M4N1FKye']            = False     # (N); ; 
SubDyn['M4N2FKye']            = False     # (N); ; 
SubDyn['M4N3FKye']            = False     # (N); ; 
SubDyn['M4N4FKye']            = False     # (N); ; 
SubDyn['M4N5FKye']            = False     # (N); ; 
SubDyn['M4N6FKye']            = False     # (N); ; 
SubDyn['M4N7FKye']            = False     # (N); ; 
SubDyn['M4N8FKye']            = False     # (N); ; 
SubDyn['M4N9FKye']            = False     # (N); ; 
SubDyn['M5N1FKye']            = False     # (N); ; 
SubDyn['M5N2FKye']            = False     # (N); ; 
SubDyn['M5N3FKye']            = False     # (N); ; 
SubDyn['M5N4FKye']            = False     # (N); ; 
SubDyn['M5N5FKye']            = False     # (N); ; 
SubDyn['M5N6FKye']            = False     # (N); ; 
SubDyn['M5N7FKye']            = False     # (N); ; 
SubDyn['M5N8FKye']            = False     # (N); ; 
SubDyn['M5N9FKye']            = False     # (N); ; 
SubDyn['M6N1FKye']            = False     # (N); ; 
SubDyn['M6N2FKye']            = False     # (N); ; 
SubDyn['M6N3FKye']            = False     # (N); ; 
SubDyn['M6N4FKye']            = False     # (N); ; 
SubDyn['M6N5FKye']            = False     # (N); ; 
SubDyn['M6N6FKye']            = False     # (N); ; 
SubDyn['M6N7FKye']            = False     # (N); ; 
SubDyn['M6N8FKye']            = False     # (N); ; 
SubDyn['M6N9FKye']            = False     # (N); ; 
SubDyn['M7N1FKye']            = False     # (N); ; 
SubDyn['M7N2FKye']            = False     # (N); ; 
SubDyn['M7N3FKye']            = False     # (N); ; 
SubDyn['M7N4FKye']            = False     # (N); ; 
SubDyn['M7N5FKye']            = False     # (N); ; 
SubDyn['M7N6FKye']            = False     # (N); ; 
SubDyn['M7N7FKye']            = False     # (N); ; 
SubDyn['M7N8FKye']            = False     # (N); ; 
SubDyn['M7N9FKye']            = False     # (N); ; 
SubDyn['M8N1FKye']            = False     # (N); ; 
SubDyn['M8N2FKye']            = False     # (N); ; 
SubDyn['M8N3FKye']            = False     # (N); ; 
SubDyn['M8N4FKye']            = False     # (N); ; 
SubDyn['M8N5FKye']            = False     # (N); ; 
SubDyn['M8N6FKye']            = False     # (N); ; 
SubDyn['M8N7FKye']            = False     # (N); ; 
SubDyn['M8N8FKye']            = False     # (N); ; 
SubDyn['M8N9FKye']            = False     # (N); ; 
SubDyn['M9N1FKye']            = False     # (N); ; 
SubDyn['M9N2FKye']            = False     # (N); ; 
SubDyn['M9N3FKye']            = False     # (N); ; 
SubDyn['M9N4FKye']            = False     # (N); ; 
SubDyn['M9N5FKye']            = False     # (N); ; 
SubDyn['M9N6FKye']            = False     # (N); ; 
SubDyn['M9N7FKye']            = False     # (N); ; 
SubDyn['M9N8FKye']            = False     # (N); ; 
SubDyn['M9N9FKye']            = False     # (N); ; 
SubDyn['M1N1FKze']            = False     # (N); Axial Force at Node Nj of member Mi- Local Reference System- Static Component; 
SubDyn['M1N2FKze']            = False     # (N); ; 
SubDyn['M1N3FKze']            = False     # (N); ; 
SubDyn['M1N4FKze']            = False     # (N); ; 
SubDyn['M1N5FKze']            = False     # (N); ; 
SubDyn['M1N6FKze']            = False     # (N); ; 
SubDyn['M1N7FKze']            = False     # (N); ; 
SubDyn['M1N8FKze']            = False     # (N); ; 
SubDyn['M1N9FKze']            = False     # (N); ; 
SubDyn['M2N1FKze']            = False     # (N); ; 
SubDyn['M2N2FKze']            = False     # (N); ; 
SubDyn['M2N3FKze']            = False     # (N); ; 
SubDyn['M2N4FKze']            = False     # (N); ; 
SubDyn['M2N5FKze']            = False     # (N); ; 
SubDyn['M2N6FKze']            = False     # (N); ; 
SubDyn['M2N7FKze']            = False     # (N); ; 
SubDyn['M2N8FKze']            = False     # (N); ; 
SubDyn['M2N9FKze']            = False     # (N); ; 
SubDyn['M3N1FKze']            = False     # (N); ; 
SubDyn['M3N2FKze']            = False     # (N); ; 
SubDyn['M3N3FKze']            = False     # (N); ; 
SubDyn['M3N4FKze']            = False     # (N); ; 
SubDyn['M3N5FKze']            = False     # (N); ; 
SubDyn['M3N6FKze']            = False     # (N); ; 
SubDyn['M3N7FKze']            = False     # (N); ; 
SubDyn['M3N8FKze']            = False     # (N); ; 
SubDyn['M3N9FKze']            = False     # (N); ; 
SubDyn['M4N1FKze']            = False     # (N); ; 
SubDyn['M4N2FKze']            = False     # (N); ; 
SubDyn['M4N3FKze']            = False     # (N); ; 
SubDyn['M4N4FKze']            = False     # (N); ; 
SubDyn['M4N5FKze']            = False     # (N); ; 
SubDyn['M4N6FKze']            = False     # (N); ; 
SubDyn['M4N7FKze']            = False     # (N); ; 
SubDyn['M4N8FKze']            = False     # (N); ; 
SubDyn['M4N9FKze']            = False     # (N); ; 
SubDyn['M5N1FKze']            = False     # (N); ; 
SubDyn['M5N2FKze']            = False     # (N); ; 
SubDyn['M5N3FKze']            = False     # (N); ; 
SubDyn['M5N4FKze']            = False     # (N); ; 
SubDyn['M5N5FKze']            = False     # (N); ; 
SubDyn['M5N6FKze']            = False     # (N); ; 
SubDyn['M5N7FKze']            = False     # (N); ; 
SubDyn['M5N8FKze']            = False     # (N); ; 
SubDyn['M5N9FKze']            = False     # (N); ; 
SubDyn['M6N1FKze']            = False     # (N); ; 
SubDyn['M6N2FKze']            = False     # (N); ; 
SubDyn['M6N3FKze']            = False     # (N); ; 
SubDyn['M6N4FKze']            = False     # (N); ; 
SubDyn['M6N5FKze']            = False     # (N); ; 
SubDyn['M6N6FKze']            = False     # (N); ; 
SubDyn['M6N7FKze']            = False     # (N); ; 
SubDyn['M6N8FKze']            = False     # (N); ; 
SubDyn['M6N9FKze']            = False     # (N); ; 
SubDyn['M7N1FKze']            = False     # (N); ; 
SubDyn['M7N2FKze']            = False     # (N); ; 
SubDyn['M7N3FKze']            = False     # (N); ; 
SubDyn['M7N4FKze']            = False     # (N); ; 
SubDyn['M7N5FKze']            = False     # (N); ; 
SubDyn['M7N6FKze']            = False     # (N); ; 
SubDyn['M7N7FKze']            = False     # (N); ; 
SubDyn['M7N8FKze']            = False     # (N); ; 
SubDyn['M7N9FKze']            = False     # (N); ; 
SubDyn['M8N1FKze']            = False     # (N); ; 
SubDyn['M8N2FKze']            = False     # (N); ; 
SubDyn['M8N3FKze']            = False     # (N); ; 
SubDyn['M8N4FKze']            = False     # (N); ; 
SubDyn['M8N5FKze']            = False     # (N); ; 
SubDyn['M8N6FKze']            = False     # (N); ; 
SubDyn['M8N7FKze']            = False     # (N); ; 
SubDyn['M8N8FKze']            = False     # (N); ; 
SubDyn['M8N9FKze']            = False     # (N); ; 
SubDyn['M9N1FKze']            = False     # (N); ; 
SubDyn['M9N2FKze']            = False     # (N); ; 
SubDyn['M9N3FKze']            = False     # (N); ; 
SubDyn['M9N4FKze']            = False     # (N); ; 
SubDyn['M9N5FKze']            = False     # (N); ; 
SubDyn['M9N6FKze']            = False     # (N); ; 
SubDyn['M9N7FKze']            = False     # (N); ; 
SubDyn['M9N8FKze']            = False     # (N); ; 
SubDyn['M9N9FKze']            = False     # (N); ; 
SubDyn['M1N1FMxe']            = False     # (N); xe component of the shear at Node Nj of member Mi- Local Reference System- Dynamic Component; 
SubDyn['M1N2FMxe']            = False     # (N); ; 
SubDyn['M1N3FMxe']            = False     # (N); ; 
SubDyn['M1N4FMxe']            = False     # (N); ; 
SubDyn['M1N5FMxe']            = False     # (N); ; 
SubDyn['M1N6FMxe']            = False     # (N); ; 
SubDyn['M1N7FMxe']            = False     # (N); ; 
SubDyn['M1N8FMxe']            = False     # (N); ; 
SubDyn['M1N9FMxe']            = False     # (N); ; 
SubDyn['M2N1FMxe']            = False     # (N); ; 
SubDyn['M2N2FMxe']            = False     # (N); ; 
SubDyn['M2N3FMxe']            = False     # (N); ; 
SubDyn['M2N4FMxe']            = False     # (N); ; 
SubDyn['M2N5FMxe']            = False     # (N); ; 
SubDyn['M2N6FMxe']            = False     # (N); ; 
SubDyn['M2N7FMxe']            = False     # (N); ; 
SubDyn['M2N8FMxe']            = False     # (N); ; 
SubDyn['M2N9FMxe']            = False     # (N); ; 
SubDyn['M3N1FMxe']            = False     # (N); ; 
SubDyn['M3N2FMxe']            = False     # (N); ; 
SubDyn['M3N3FMxe']            = False     # (N); ; 
SubDyn['M3N4FMxe']            = False     # (N); ; 
SubDyn['M3N5FMxe']            = False     # (N); ; 
SubDyn['M3N6FMxe']            = False     # (N); ; 
SubDyn['M3N7FMxe']            = False     # (N); ; 
SubDyn['M3N8FMxe']            = False     # (N); ; 
SubDyn['M3N9FMxe']            = False     # (N); ; 
SubDyn['M4N1FMxe']            = False     # (N); ; 
SubDyn['M4N2FMxe']            = False     # (N); ; 
SubDyn['M4N3FMxe']            = False     # (N); ; 
SubDyn['M4N4FMxe']            = False     # (N); ; 
SubDyn['M4N5FMxe']            = False     # (N); ; 
SubDyn['M4N6FMxe']            = False     # (N); ; 
SubDyn['M4N7FMxe']            = False     # (N); ; 
SubDyn['M4N8FMxe']            = False     # (N); ; 
SubDyn['M4N9FMxe']            = False     # (N); ; 
SubDyn['M5N1FMxe']            = False     # (N); ; 
SubDyn['M5N2FMxe']            = False     # (N); ; 
SubDyn['M5N3FMxe']            = False     # (N); ; 
SubDyn['M5N4FMxe']            = False     # (N); ; 
SubDyn['M5N5FMxe']            = False     # (N); ; 
SubDyn['M5N6FMxe']            = False     # (N); ; 
SubDyn['M5N7FMxe']            = False     # (N); ; 
SubDyn['M5N8FMxe']            = False     # (N); ; 
SubDyn['M5N9FMxe']            = False     # (N); ; 
SubDyn['M6N1FMxe']            = False     # (N); ; 
SubDyn['M6N2FMxe']            = False     # (N); ; 
SubDyn['M6N3FMxe']            = False     # (N); ; 
SubDyn['M6N4FMxe']            = False     # (N); ; 
SubDyn['M6N5FMxe']            = False     # (N); ; 
SubDyn['M6N6FMxe']            = False     # (N); ; 
SubDyn['M6N7FMxe']            = False     # (N); ; 
SubDyn['M6N8FMxe']            = False     # (N); ; 
SubDyn['M6N9FMxe']            = False     # (N); ; 
SubDyn['M7N1FMxe']            = False     # (N); ; 
SubDyn['M7N2FMxe']            = False     # (N); ; 
SubDyn['M7N3FMxe']            = False     # (N); ; 
SubDyn['M7N4FMxe']            = False     # (N); ; 
SubDyn['M7N5FMxe']            = False     # (N); ; 
SubDyn['M7N6FMxe']            = False     # (N); ; 
SubDyn['M7N7FMxe']            = False     # (N); ; 
SubDyn['M7N8FMxe']            = False     # (N); ; 
SubDyn['M7N9FMxe']            = False     # (N); ; 
SubDyn['M8N1FMxe']            = False     # (N); ; 
SubDyn['M8N2FMxe']            = False     # (N); ; 
SubDyn['M8N3FMxe']            = False     # (N); ; 
SubDyn['M8N4FMxe']            = False     # (N); ; 
SubDyn['M8N5FMxe']            = False     # (N); ; 
SubDyn['M8N6FMxe']            = False     # (N); ; 
SubDyn['M8N7FMxe']            = False     # (N); ; 
SubDyn['M8N8FMxe']            = False     # (N); ; 
SubDyn['M8N9FMxe']            = False     # (N); ; 
SubDyn['M9N1FMxe']            = False     # (N); ; 
SubDyn['M9N2FMxe']            = False     # (N); ; 
SubDyn['M9N3FMxe']            = False     # (N); ; 
SubDyn['M9N4FMxe']            = False     # (N); ; 
SubDyn['M9N5FMxe']            = False     # (N); ; 
SubDyn['M9N6FMxe']            = False     # (N); ; 
SubDyn['M9N7FMxe']            = False     # (N); ; 
SubDyn['M9N8FMxe']            = False     # (N); ; 
SubDyn['M9N9FMxe']            = False     # (N); ; 
SubDyn['M1N1FMye']            = False     # (N); ye component of the shear at Node Nj of member Mi- Local Reference System- Dynamic Component; 
SubDyn['M1N2FMye']            = False     # (N); ; 
SubDyn['M1N3FMye']            = False     # (N); ; 
SubDyn['M1N4FMye']            = False     # (N); ; 
SubDyn['M1N5FMye']            = False     # (N); ; 
SubDyn['M1N6FMye']            = False     # (N); ; 
SubDyn['M1N7FMye']            = False     # (N); ; 
SubDyn['M1N8FMye']            = False     # (N); ; 
SubDyn['M1N9FMye']            = False     # (N); ; 
SubDyn['M2N1FMye']            = False     # (N); ; 
SubDyn['M2N2FMye']            = False     # (N); ; 
SubDyn['M2N3FMye']            = False     # (N); ; 
SubDyn['M2N4FMye']            = False     # (N); ; 
SubDyn['M2N5FMye']            = False     # (N); ; 
SubDyn['M2N6FMye']            = False     # (N); ; 
SubDyn['M2N7FMye']            = False     # (N); ; 
SubDyn['M2N8FMye']            = False     # (N); ; 
SubDyn['M2N9FMye']            = False     # (N); ; 
SubDyn['M3N1FMye']            = False     # (N); ; 
SubDyn['M3N2FMye']            = False     # (N); ; 
SubDyn['M3N3FMye']            = False     # (N); ; 
SubDyn['M3N4FMye']            = False     # (N); ; 
SubDyn['M3N5FMye']            = False     # (N); ; 
SubDyn['M3N6FMye']            = False     # (N); ; 
SubDyn['M3N7FMye']            = False     # (N); ; 
SubDyn['M3N8FMye']            = False     # (N); ; 
SubDyn['M3N9FMye']            = False     # (N); ; 
SubDyn['M4N1FMye']            = False     # (N); ; 
SubDyn['M4N2FMye']            = False     # (N); ; 
SubDyn['M4N3FMye']            = False     # (N); ; 
SubDyn['M4N4FMye']            = False     # (N); ; 
SubDyn['M4N5FMye']            = False     # (N); ; 
SubDyn['M4N6FMye']            = False     # (N); ; 
SubDyn['M4N7FMye']            = False     # (N); ; 
SubDyn['M4N8FMye']            = False     # (N); ; 
SubDyn['M4N9FMye']            = False     # (N); ; 
SubDyn['M5N1FMye']            = False     # (N); ; 
SubDyn['M5N2FMye']            = False     # (N); ; 
SubDyn['M5N3FMye']            = False     # (N); ; 
SubDyn['M5N4FMye']            = False     # (N); ; 
SubDyn['M5N5FMye']            = False     # (N); ; 
SubDyn['M5N6FMye']            = False     # (N); ; 
SubDyn['M5N7FMye']            = False     # (N); ; 
SubDyn['M5N8FMye']            = False     # (N); ; 
SubDyn['M5N9FMye']            = False     # (N); ; 
SubDyn['M6N1FMye']            = False     # (N); ; 
SubDyn['M6N2FMye']            = False     # (N); ; 
SubDyn['M6N3FMye']            = False     # (N); ; 
SubDyn['M6N4FMye']            = False     # (N); ; 
SubDyn['M6N5FMye']            = False     # (N); ; 
SubDyn['M6N6FMye']            = False     # (N); ; 
SubDyn['M6N7FMye']            = False     # (N); ; 
SubDyn['M6N8FMye']            = False     # (N); ; 
SubDyn['M6N9FMye']            = False     # (N); ; 
SubDyn['M7N1FMye']            = False     # (N); ; 
SubDyn['M7N2FMye']            = False     # (N); ; 
SubDyn['M7N3FMye']            = False     # (N); ; 
SubDyn['M7N4FMye']            = False     # (N); ; 
SubDyn['M7N5FMye']            = False     # (N); ; 
SubDyn['M7N6FMye']            = False     # (N); ; 
SubDyn['M7N7FMye']            = False     # (N); ; 
SubDyn['M7N8FMye']            = False     # (N); ; 
SubDyn['M7N9FMye']            = False     # (N); ; 
SubDyn['M8N1FMye']            = False     # (N); ; 
SubDyn['M8N2FMye']            = False     # (N); ; 
SubDyn['M8N3FMye']            = False     # (N); ; 
SubDyn['M8N4FMye']            = False     # (N); ; 
SubDyn['M8N5FMye']            = False     # (N); ; 
SubDyn['M8N6FMye']            = False     # (N); ; 
SubDyn['M8N7FMye']            = False     # (N); ; 
SubDyn['M8N8FMye']            = False     # (N); ; 
SubDyn['M8N9FMye']            = False     # (N); ; 
SubDyn['M9N1FMye']            = False     # (N); ; 
SubDyn['M9N2FMye']            = False     # (N); ; 
SubDyn['M9N3FMye']            = False     # (N); ; 
SubDyn['M9N4FMye']            = False     # (N); ; 
SubDyn['M9N5FMye']            = False     # (N); ; 
SubDyn['M9N6FMye']            = False     # (N); ; 
SubDyn['M9N7FMye']            = False     # (N); ; 
SubDyn['M9N8FMye']            = False     # (N); ; 
SubDyn['M9N9FMye']            = False     # (N); ; 
SubDyn['M1N1FMze']            = False     # (N); axial force at Node Nj of member Mi- Local Reference System- Dynamic Component; 
SubDyn['M1N2FMze']            = False     # (N); ; 
SubDyn['M1N3FMze']            = False     # (N); ; 
SubDyn['M1N4FMze']            = False     # (N); ; 
SubDyn['M1N5FMze']            = False     # (N); ; 
SubDyn['M1N6FMze']            = False     # (N); ; 
SubDyn['M1N7FMze']            = False     # (N); ; 
SubDyn['M1N8FMze']            = False     # (N); ; 
SubDyn['M1N9FMze']            = False     # (N); ; 
SubDyn['M2N1FMze']            = False     # (N); ; 
SubDyn['M2N2FMze']            = False     # (N); ; 
SubDyn['M2N3FMze']            = False     # (N); ; 
SubDyn['M2N4FMze']            = False     # (N); ; 
SubDyn['M2N5FMze']            = False     # (N); ; 
SubDyn['M2N6FMze']            = False     # (N); ; 
SubDyn['M2N7FMze']            = False     # (N); ; 
SubDyn['M2N8FMze']            = False     # (N); ; 
SubDyn['M2N9FMze']            = False     # (N); ; 
SubDyn['M3N1FMze']            = False     # (N); ; 
SubDyn['M3N2FMze']            = False     # (N); ; 
SubDyn['M3N3FMze']            = False     # (N); ; 
SubDyn['M3N4FMze']            = False     # (N); ; 
SubDyn['M3N5FMze']            = False     # (N); ; 
SubDyn['M3N6FMze']            = False     # (N); ; 
SubDyn['M3N7FMze']            = False     # (N); ; 
SubDyn['M3N8FMze']            = False     # (N); ; 
SubDyn['M3N9FMze']            = False     # (N); ; 
SubDyn['M4N1FMze']            = False     # (N); ; 
SubDyn['M4N2FMze']            = False     # (N); ; 
SubDyn['M4N3FMze']            = False     # (N); ; 
SubDyn['M4N4FMze']            = False     # (N); ; 
SubDyn['M4N5FMze']            = False     # (N); ; 
SubDyn['M4N6FMze']            = False     # (N); ; 
SubDyn['M4N7FMze']            = False     # (N); ; 
SubDyn['M4N8FMze']            = False     # (N); ; 
SubDyn['M4N9FMze']            = False     # (N); ; 
SubDyn['M5N1FMze']            = False     # (N); ; 
SubDyn['M5N2FMze']            = False     # (N); ; 
SubDyn['M5N3FMze']            = False     # (N); ; 
SubDyn['M5N4FMze']            = False     # (N); ; 
SubDyn['M5N5FMze']            = False     # (N); ; 
SubDyn['M5N6FMze']            = False     # (N); ; 
SubDyn['M5N7FMze']            = False     # (N); ; 
SubDyn['M5N8FMze']            = False     # (N); ; 
SubDyn['M5N9FMze']            = False     # (N); ; 
SubDyn['M6N1FMze']            = False     # (N); ; 
SubDyn['M6N2FMze']            = False     # (N); ; 
SubDyn['M6N3FMze']            = False     # (N); ; 
SubDyn['M6N4FMze']            = False     # (N); ; 
SubDyn['M6N5FMze']            = False     # (N); ; 
SubDyn['M6N6FMze']            = False     # (N); ; 
SubDyn['M6N7FMze']            = False     # (N); ; 
SubDyn['M6N8FMze']            = False     # (N); ; 
SubDyn['M6N9FMze']            = False     # (N); ; 
SubDyn['M7N1FMze']            = False     # (N); ; 
SubDyn['M7N2FMze']            = False     # (N); ; 
SubDyn['M7N3FMze']            = False     # (N); ; 
SubDyn['M7N4FMze']            = False     # (N); ; 
SubDyn['M7N5FMze']            = False     # (N); ; 
SubDyn['M7N6FMze']            = False     # (N); ; 
SubDyn['M7N7FMze']            = False     # (N); ; 
SubDyn['M7N8FMze']            = False     # (N); ; 
SubDyn['M7N9FMze']            = False     # (N); ; 
SubDyn['M8N1FMze']            = False     # (N); ; 
SubDyn['M8N2FMze']            = False     # (N); ; 
SubDyn['M8N3FMze']            = False     # (N); ; 
SubDyn['M8N4FMze']            = False     # (N); ; 
SubDyn['M8N5FMze']            = False     # (N); ; 
SubDyn['M8N6FMze']            = False     # (N); ; 
SubDyn['M8N7FMze']            = False     # (N); ; 
SubDyn['M8N8FMze']            = False     # (N); ; 
SubDyn['M8N9FMze']            = False     # (N); ; 
SubDyn['M9N1FMze']            = False     # (N); ; 
SubDyn['M9N2FMze']            = False     # (N); ; 
SubDyn['M9N3FMze']            = False     # (N); ; 
SubDyn['M9N4FMze']            = False     # (N); ; 
SubDyn['M9N5FMze']            = False     # (N); ; 
SubDyn['M9N6FMze']            = False     # (N); ; 
SubDyn['M9N7FMze']            = False     # (N); ; 
SubDyn['M9N8FMze']            = False     # (N); ; 
SubDyn['M9N9FMze']            = False     # (N); ; 
SubDyn['M1N1MKxe']            = False     # (N*m); xe component of the bending moment at Node Nj of member Mi- Local Reference System- Static Component; 
SubDyn['M1N2MKxe']            = False     # (N*m); ; 
SubDyn['M1N3MKxe']            = False     # (N*m); ; 
SubDyn['M1N4MKxe']            = False     # (N*m); ; 
SubDyn['M1N5MKxe']            = False     # (N*m); ; 
SubDyn['M1N6MKxe']            = False     # (N*m); ; 
SubDyn['M1N7MKxe']            = False     # (N*m); ; 
SubDyn['M1N8MKxe']            = False     # (N*m); ; 
SubDyn['M1N9MKxe']            = False     # (N*m); ; 
SubDyn['M2N1MKxe']            = False     # (N*m); ; 
SubDyn['M2N2MKxe']            = False     # (N*m); ; 
SubDyn['M2N3MKxe']            = False     # (N*m); ; 
SubDyn['M2N4MKxe']            = False     # (N*m); ; 
SubDyn['M2N5MKxe']            = False     # (N*m); ; 
SubDyn['M2N6MKxe']            = False     # (N*m); ; 
SubDyn['M2N7MKxe']            = False     # (N*m); ; 
SubDyn['M2N8MKxe']            = False     # (N*m); ; 
SubDyn['M2N9MKxe']            = False     # (N*m); ; 
SubDyn['M3N1MKxe']            = False     # (N*m); ; 
SubDyn['M3N2MKxe']            = False     # (N*m); ; 
SubDyn['M3N3MKxe']            = False     # (N*m); ; 
SubDyn['M3N4MKxe']            = False     # (N*m); ; 
SubDyn['M3N5MKxe']            = False     # (N*m); ; 
SubDyn['M3N6MKxe']            = False     # (N*m); ; 
SubDyn['M3N7MKxe']            = False     # (N*m); ; 
SubDyn['M3N8MKxe']            = False     # (N*m); ; 
SubDyn['M3N9MKxe']            = False     # (N*m); ; 
SubDyn['M4N1MKxe']            = False     # (N*m); ; 
SubDyn['M4N2MKxe']            = False     # (N*m); ; 
SubDyn['M4N3MKxe']            = False     # (N*m); ; 
SubDyn['M4N4MKxe']            = False     # (N*m); ; 
SubDyn['M4N5MKxe']            = False     # (N*m); ; 
SubDyn['M4N6MKxe']            = False     # (N*m); ; 
SubDyn['M4N7MKxe']            = False     # (N*m); ; 
SubDyn['M4N8MKxe']            = False     # (N*m); ; 
SubDyn['M4N9MKxe']            = False     # (N*m); ; 
SubDyn['M5N1MKxe']            = False     # (N*m); ; 
SubDyn['M5N2MKxe']            = False     # (N*m); ; 
SubDyn['M5N3MKxe']            = False     # (N*m); ; 
SubDyn['M5N4MKxe']            = False     # (N*m); ; 
SubDyn['M5N5MKxe']            = False     # (N*m); ; 
SubDyn['M5N6MKxe']            = False     # (N*m); ; 
SubDyn['M5N7MKxe']            = False     # (N*m); ; 
SubDyn['M5N8MKxe']            = False     # (N*m); ; 
SubDyn['M5N9MKxe']            = False     # (N*m); ; 
SubDyn['M6N1MKxe']            = False     # (N*m); ; 
SubDyn['M6N2MKxe']            = False     # (N*m); ; 
SubDyn['M6N3MKxe']            = False     # (N*m); ; 
SubDyn['M6N4MKxe']            = False     # (N*m); ; 
SubDyn['M6N5MKxe']            = False     # (N*m); ; 
SubDyn['M6N6MKxe']            = False     # (N*m); ; 
SubDyn['M6N7MKxe']            = False     # (N*m); ; 
SubDyn['M6N8MKxe']            = False     # (N*m); ; 
SubDyn['M6N9MKxe']            = False     # (N*m); ; 
SubDyn['M7N1MKxe']            = False     # (N*m); ; 
SubDyn['M7N2MKxe']            = False     # (N*m); ; 
SubDyn['M7N3MKxe']            = False     # (N*m); ; 
SubDyn['M7N4MKxe']            = False     # (N*m); ; 
SubDyn['M7N5MKxe']            = False     # (N*m); ; 
SubDyn['M7N6MKxe']            = False     # (N*m); ; 
SubDyn['M7N7MKxe']            = False     # (N*m); ; 
SubDyn['M7N8MKxe']            = False     # (N*m); ; 
SubDyn['M7N9MKxe']            = False     # (N*m); ; 
SubDyn['M8N1MKxe']            = False     # (N*m); ; 
SubDyn['M8N2MKxe']            = False     # (N*m); ; 
SubDyn['M8N3MKxe']            = False     # (N*m); ; 
SubDyn['M8N4MKxe']            = False     # (N*m); ; 
SubDyn['M8N5MKxe']            = False     # (N*m); ; 
SubDyn['M8N6MKxe']            = False     # (N*m); ; 
SubDyn['M8N7MKxe']            = False     # (N*m); ; 
SubDyn['M8N8MKxe']            = False     # (N*m); ; 
SubDyn['M8N9MKxe']            = False     # (N*m); ; 
SubDyn['M9N1MKxe']            = False     # (N*m); ; 
SubDyn['M9N2MKxe']            = False     # (N*m); ; 
SubDyn['M9N3MKxe']            = False     # (N*m); ; 
SubDyn['M9N4MKxe']            = False     # (N*m); ; 
SubDyn['M9N5MKxe']            = False     # (N*m); ; 
SubDyn['M9N6MKxe']            = False     # (N*m); ; 
SubDyn['M9N7MKxe']            = False     # (N*m); ; 
SubDyn['M9N8MKxe']            = False     # (N*m); ; 
SubDyn['M9N9MKxe']            = False     # (N*m); ; 
SubDyn['M1N1MKye']            = False     # (N*m); ye component of the bending moment at Node Nj of member Mi- Local Reference System- Static Component; 
SubDyn['M1N2MKye']            = False     # (N*m); ; 
SubDyn['M1N3MKye']            = False     # (N*m); ; 
SubDyn['M1N4MKye']            = False     # (N*m); ; 
SubDyn['M1N5MKye']            = False     # (N*m); ; 
SubDyn['M1N6MKye']            = False     # (N*m); ; 
SubDyn['M1N7MKye']            = False     # (N*m); ; 
SubDyn['M1N8MKye']            = False     # (N*m); ; 
SubDyn['M1N9MKye']            = False     # (N*m); ; 
SubDyn['M2N1MKye']            = False     # (N*m); ; 
SubDyn['M2N2MKye']            = False     # (N*m); ; 
SubDyn['M2N3MKye']            = False     # (N*m); ; 
SubDyn['M2N4MKye']            = False     # (N*m); ; 
SubDyn['M2N5MKye']            = False     # (N*m); ; 
SubDyn['M2N6MKye']            = False     # (N*m); ; 
SubDyn['M2N7MKye']            = False     # (N*m); ; 
SubDyn['M2N8MKye']            = False     # (N*m); ; 
SubDyn['M2N9MKye']            = False     # (N*m); ; 
SubDyn['M3N1MKye']            = False     # (N*m); ; 
SubDyn['M3N2MKye']            = False     # (N*m); ; 
SubDyn['M3N3MKye']            = False     # (N*m); ; 
SubDyn['M3N4MKye']            = False     # (N*m); ; 
SubDyn['M3N5MKye']            = False     # (N*m); ; 
SubDyn['M3N6MKye']            = False     # (N*m); ; 
SubDyn['M3N7MKye']            = False     # (N*m); ; 
SubDyn['M3N8MKye']            = False     # (N*m); ; 
SubDyn['M3N9MKye']            = False     # (N*m); ; 
SubDyn['M4N1MKye']            = False     # (N*m); ; 
SubDyn['M4N2MKye']            = False     # (N*m); ; 
SubDyn['M4N3MKye']            = False     # (N*m); ; 
SubDyn['M4N4MKye']            = False     # (N*m); ; 
SubDyn['M4N5MKye']            = False     # (N*m); ; 
SubDyn['M4N6MKye']            = False     # (N*m); ; 
SubDyn['M4N7MKye']            = False     # (N*m); ; 
SubDyn['M4N8MKye']            = False     # (N*m); ; 
SubDyn['M4N9MKye']            = False     # (N*m); ; 
SubDyn['M5N1MKye']            = False     # (N*m); ; 
SubDyn['M5N2MKye']            = False     # (N*m); ; 
SubDyn['M5N3MKye']            = False     # (N*m); ; 
SubDyn['M5N4MKye']            = False     # (N*m); ; 
SubDyn['M5N5MKye']            = False     # (N*m); ; 
SubDyn['M5N6MKye']            = False     # (N*m); ; 
SubDyn['M5N7MKye']            = False     # (N*m); ; 
SubDyn['M5N8MKye']            = False     # (N*m); ; 
SubDyn['M5N9MKye']            = False     # (N*m); ; 
SubDyn['M6N1MKye']            = False     # (N*m); ; 
SubDyn['M6N2MKye']            = False     # (N*m); ; 
SubDyn['M6N3MKye']            = False     # (N*m); ; 
SubDyn['M6N4MKye']            = False     # (N*m); ; 
SubDyn['M6N5MKye']            = False     # (N*m); ; 
SubDyn['M6N6MKye']            = False     # (N*m); ; 
SubDyn['M6N7MKye']            = False     # (N*m); ; 
SubDyn['M6N8MKye']            = False     # (N*m); ; 
SubDyn['M6N9MKye']            = False     # (N*m); ; 
SubDyn['M7N1MKye']            = False     # (N*m); ; 
SubDyn['M7N2MKye']            = False     # (N*m); ; 
SubDyn['M7N3MKye']            = False     # (N*m); ; 
SubDyn['M7N4MKye']            = False     # (N*m); ; 
SubDyn['M7N5MKye']            = False     # (N*m); ; 
SubDyn['M7N6MKye']            = False     # (N*m); ; 
SubDyn['M7N7MKye']            = False     # (N*m); ; 
SubDyn['M7N8MKye']            = False     # (N*m); ; 
SubDyn['M7N9MKye']            = False     # (N*m); ; 
SubDyn['M8N1MKye']            = False     # (N*m); ; 
SubDyn['M8N2MKye']            = False     # (N*m); ; 
SubDyn['M8N3MKye']            = False     # (N*m); ; 
SubDyn['M8N4MKye']            = False     # (N*m); ; 
SubDyn['M8N5MKye']            = False     # (N*m); ; 
SubDyn['M8N6MKye']            = False     # (N*m); ; 
SubDyn['M8N7MKye']            = False     # (N*m); ; 
SubDyn['M8N8MKye']            = False     # (N*m); ; 
SubDyn['M8N9MKye']            = False     # (N*m); ; 
SubDyn['M9N1MKye']            = False     # (N*m); ; 
SubDyn['M9N2MKye']            = False     # (N*m); ; 
SubDyn['M9N3MKye']            = False     # (N*m); ; 
SubDyn['M9N4MKye']            = False     # (N*m); ; 
SubDyn['M9N5MKye']            = False     # (N*m); ; 
SubDyn['M9N6MKye']            = False     # (N*m); ; 
SubDyn['M9N7MKye']            = False     # (N*m); ; 
SubDyn['M9N8MKye']            = False     # (N*m); ; 
SubDyn['M9N9MKye']            = False     # (N*m); ; 
SubDyn['M1N1MKze']            = False     # (N*m); Torsion moment at Node Nj of member Mi- Local Reference System- Static Component; 
SubDyn['M1N2MKze']            = False     # (N*m); ; 
SubDyn['M1N3MKze']            = False     # (N*m); ; 
SubDyn['M1N4MKze']            = False     # (N*m); ; 
SubDyn['M1N5MKze']            = False     # (N*m); ; 
SubDyn['M1N6MKze']            = False     # (N*m); ; 
SubDyn['M1N7MKze']            = False     # (N*m); ; 
SubDyn['M1N8MKze']            = False     # (N*m); ; 
SubDyn['M1N9MKze']            = False     # (N*m); ; 
SubDyn['M2N1MKze']            = False     # (N*m); ; 
SubDyn['M2N2MKze']            = False     # (N*m); ; 
SubDyn['M2N3MKze']            = False     # (N*m); ; 
SubDyn['M2N4MKze']            = False     # (N*m); ; 
SubDyn['M2N5MKze']            = False     # (N*m); ; 
SubDyn['M2N6MKze']            = False     # (N*m); ; 
SubDyn['M2N7MKze']            = False     # (N*m); ; 
SubDyn['M2N8MKze']            = False     # (N*m); ; 
SubDyn['M2N9MKze']            = False     # (N*m); ; 
SubDyn['M3N1MKze']            = False     # (N*m); ; 
SubDyn['M3N2MKze']            = False     # (N*m); ; 
SubDyn['M3N3MKze']            = False     # (N*m); ; 
SubDyn['M3N4MKze']            = False     # (N*m); ; 
SubDyn['M3N5MKze']            = False     # (N*m); ; 
SubDyn['M3N6MKze']            = False     # (N*m); ; 
SubDyn['M3N7MKze']            = False     # (N*m); ; 
SubDyn['M3N8MKze']            = False     # (N*m); ; 
SubDyn['M3N9MKze']            = False     # (N*m); ; 
SubDyn['M4N1MKze']            = False     # (N*m); ; 
SubDyn['M4N2MKze']            = False     # (N*m); ; 
SubDyn['M4N3MKze']            = False     # (N*m); ; 
SubDyn['M4N4MKze']            = False     # (N*m); ; 
SubDyn['M4N5MKze']            = False     # (N*m); ; 
SubDyn['M4N6MKze']            = False     # (N*m); ; 
SubDyn['M4N7MKze']            = False     # (N*m); ; 
SubDyn['M4N8MKze']            = False     # (N*m); ; 
SubDyn['M4N9MKze']            = False     # (N*m); ; 
SubDyn['M5N1MKze']            = False     # (N*m); ; 
SubDyn['M5N2MKze']            = False     # (N*m); ; 
SubDyn['M5N3MKze']            = False     # (N*m); ; 
SubDyn['M5N4MKze']            = False     # (N*m); ; 
SubDyn['M5N5MKze']            = False     # (N*m); ; 
SubDyn['M5N6MKze']            = False     # (N*m); ; 
SubDyn['M5N7MKze']            = False     # (N*m); ; 
SubDyn['M5N8MKze']            = False     # (N*m); ; 
SubDyn['M5N9MKze']            = False     # (N*m); ; 
SubDyn['M6N1MKze']            = False     # (N*m); ; 
SubDyn['M6N2MKze']            = False     # (N*m); ; 
SubDyn['M6N3MKze']            = False     # (N*m); ; 
SubDyn['M6N4MKze']            = False     # (N*m); ; 
SubDyn['M6N5MKze']            = False     # (N*m); ; 
SubDyn['M6N6MKze']            = False     # (N*m); ; 
SubDyn['M6N7MKze']            = False     # (N*m); ; 
SubDyn['M6N8MKze']            = False     # (N*m); ; 
SubDyn['M6N9MKze']            = False     # (N*m); ; 
SubDyn['M7N1MKze']            = False     # (N*m); ; 
SubDyn['M7N2MKze']            = False     # (N*m); ; 
SubDyn['M7N3MKze']            = False     # (N*m); ; 
SubDyn['M7N4MKze']            = False     # (N*m); ; 
SubDyn['M7N5MKze']            = False     # (N*m); ; 
SubDyn['M7N6MKze']            = False     # (N*m); ; 
SubDyn['M7N7MKze']            = False     # (N*m); ; 
SubDyn['M7N8MKze']            = False     # (N*m); ; 
SubDyn['M7N9MKze']            = False     # (N*m); ; 
SubDyn['M8N1MKze']            = False     # (N*m); ; 
SubDyn['M8N2MKze']            = False     # (N*m); ; 
SubDyn['M8N3MKze']            = False     # (N*m); ; 
SubDyn['M8N4MKze']            = False     # (N*m); ; 
SubDyn['M8N5MKze']            = False     # (N*m); ; 
SubDyn['M8N6MKze']            = False     # (N*m); ; 
SubDyn['M8N7MKze']            = False     # (N*m); ; 
SubDyn['M8N8MKze']            = False     # (N*m); ; 
SubDyn['M8N9MKze']            = False     # (N*m); ; 
SubDyn['M9N1MKze']            = False     # (N*m); ; 
SubDyn['M9N2MKze']            = False     # (N*m); ; 
SubDyn['M9N3MKze']            = False     # (N*m); ; 
SubDyn['M9N4MKze']            = False     # (N*m); ; 
SubDyn['M9N5MKze']            = False     # (N*m); ; 
SubDyn['M9N6MKze']            = False     # (N*m); ; 
SubDyn['M9N7MKze']            = False     # (N*m); ; 
SubDyn['M9N8MKze']            = False     # (N*m); ; 
SubDyn['M9N9MKze']            = False     # (N*m); ; 
SubDyn['M1N1MMxe']            = False     # (N*m); xe component of the bending moment at Node Nj of member Mi- Local Reference System- Dynamic Component; 
SubDyn['M1N2MMxe']            = False     # (N*m); ; 
SubDyn['M1N3MMxe']            = False     # (N*m); ; 
SubDyn['M1N4MMxe']            = False     # (N*m); ; 
SubDyn['M1N5MMxe']            = False     # (N*m); ; 
SubDyn['M1N6MMxe']            = False     # (N*m); ; 
SubDyn['M1N7MMxe']            = False     # (N*m); ; 
SubDyn['M1N8MMxe']            = False     # (N*m); ; 
SubDyn['M1N9MMxe']            = False     # (N*m); ; 
SubDyn['M2N1MMxe']            = False     # (N*m); ; 
SubDyn['M2N2MMxe']            = False     # (N*m); ; 
SubDyn['M2N3MMxe']            = False     # (N*m); ; 
SubDyn['M2N4MMxe']            = False     # (N*m); ; 
SubDyn['M2N5MMxe']            = False     # (N*m); ; 
SubDyn['M2N6MMxe']            = False     # (N*m); ; 
SubDyn['M2N7MMxe']            = False     # (N*m); ; 
SubDyn['M2N8MMxe']            = False     # (N*m); ; 
SubDyn['M2N9MMxe']            = False     # (N*m); ; 
SubDyn['M3N1MMxe']            = False     # (N*m); ; 
SubDyn['M3N2MMxe']            = False     # (N*m); ; 
SubDyn['M3N3MMxe']            = False     # (N*m); ; 
SubDyn['M3N4MMxe']            = False     # (N*m); ; 
SubDyn['M3N5MMxe']            = False     # (N*m); ; 
SubDyn['M3N6MMxe']            = False     # (N*m); ; 
SubDyn['M3N7MMxe']            = False     # (N*m); ; 
SubDyn['M3N8MMxe']            = False     # (N*m); ; 
SubDyn['M3N9MMxe']            = False     # (N*m); ; 
SubDyn['M4N1MMxe']            = False     # (N*m); ; 
SubDyn['M4N2MMxe']            = False     # (N*m); ; 
SubDyn['M4N3MMxe']            = False     # (N*m); ; 
SubDyn['M4N4MMxe']            = False     # (N*m); ; 
SubDyn['M4N5MMxe']            = False     # (N*m); ; 
SubDyn['M4N6MMxe']            = False     # (N*m); ; 
SubDyn['M4N7MMxe']            = False     # (N*m); ; 
SubDyn['M4N8MMxe']            = False     # (N*m); ; 
SubDyn['M4N9MMxe']            = False     # (N*m); ; 
SubDyn['M5N1MMxe']            = False     # (N*m); ; 
SubDyn['M5N2MMxe']            = False     # (N*m); ; 
SubDyn['M5N3MMxe']            = False     # (N*m); ; 
SubDyn['M5N4MMxe']            = False     # (N*m); ; 
SubDyn['M5N5MMxe']            = False     # (N*m); ; 
SubDyn['M5N6MMxe']            = False     # (N*m); ; 
SubDyn['M5N7MMxe']            = False     # (N*m); ; 
SubDyn['M5N8MMxe']            = False     # (N*m); ; 
SubDyn['M5N9MMxe']            = False     # (N*m); ; 
SubDyn['M6N1MMxe']            = False     # (N*m); ; 
SubDyn['M6N2MMxe']            = False     # (N*m); ; 
SubDyn['M6N3MMxe']            = False     # (N*m); ; 
SubDyn['M6N4MMxe']            = False     # (N*m); ; 
SubDyn['M6N5MMxe']            = False     # (N*m); ; 
SubDyn['M6N6MMxe']            = False     # (N*m); ; 
SubDyn['M6N7MMxe']            = False     # (N*m); ; 
SubDyn['M6N8MMxe']            = False     # (N*m); ; 
SubDyn['M6N9MMxe']            = False     # (N*m); ; 
SubDyn['M7N1MMxe']            = False     # (N*m); ; 
SubDyn['M7N2MMxe']            = False     # (N*m); ; 
SubDyn['M7N3MMxe']            = False     # (N*m); ; 
SubDyn['M7N4MMxe']            = False     # (N*m); ; 
SubDyn['M7N5MMxe']            = False     # (N*m); ; 
SubDyn['M7N6MMxe']            = False     # (N*m); ; 
SubDyn['M7N7MMxe']            = False     # (N*m); ; 
SubDyn['M7N8MMxe']            = False     # (N*m); ; 
SubDyn['M7N9MMxe']            = False     # (N*m); ; 
SubDyn['M8N1MMxe']            = False     # (N*m); ; 
SubDyn['M8N2MMxe']            = False     # (N*m); ; 
SubDyn['M8N3MMxe']            = False     # (N*m); ; 
SubDyn['M8N4MMxe']            = False     # (N*m); ; 
SubDyn['M8N5MMxe']            = False     # (N*m); ; 
SubDyn['M8N6MMxe']            = False     # (N*m); ; 
SubDyn['M8N7MMxe']            = False     # (N*m); ; 
SubDyn['M8N8MMxe']            = False     # (N*m); ; 
SubDyn['M8N9MMxe']            = False     # (N*m); ; 
SubDyn['M9N1MMxe']            = False     # (N*m); ; 
SubDyn['M9N2MMxe']            = False     # (N*m); ; 
SubDyn['M9N3MMxe']            = False     # (N*m); ; 
SubDyn['M9N4MMxe']            = False     # (N*m); ; 
SubDyn['M9N5MMxe']            = False     # (N*m); ; 
SubDyn['M9N6MMxe']            = False     # (N*m); ; 
SubDyn['M9N7MMxe']            = False     # (N*m); ; 
SubDyn['M9N8MMxe']            = False     # (N*m); ; 
SubDyn['M9N9MMxe']            = False     # (N*m); ; 
SubDyn['M1N1MMye']            = False     # (N*m); ye component of the bending moment at Node Nj of member Mi- Local Reference System- Dynamic Component; 
SubDyn['M1N2MMye']            = False     # (N*m); ; 
SubDyn['M1N3MMye']            = False     # (N*m); ; 
SubDyn['M1N4MMye']            = False     # (N*m); ; 
SubDyn['M1N5MMye']            = False     # (N*m); ; 
SubDyn['M1N6MMye']            = False     # (N*m); ; 
SubDyn['M1N7MMye']            = False     # (N*m); ; 
SubDyn['M1N8MMye']            = False     # (N*m); ; 
SubDyn['M1N9MMye']            = False     # (N*m); ; 
SubDyn['M2N1MMye']            = False     # (N*m); ; 
SubDyn['M2N2MMye']            = False     # (N*m); ; 
SubDyn['M2N3MMye']            = False     # (N*m); ; 
SubDyn['M2N4MMye']            = False     # (N*m); ; 
SubDyn['M2N5MMye']            = False     # (N*m); ; 
SubDyn['M2N6MMye']            = False     # (N*m); ; 
SubDyn['M2N7MMye']            = False     # (N*m); ; 
SubDyn['M2N8MMye']            = False     # (N*m); ; 
SubDyn['M2N9MMye']            = False     # (N*m); ; 
SubDyn['M3N1MMye']            = False     # (N*m); ; 
SubDyn['M3N2MMye']            = False     # (N*m); ; 
SubDyn['M3N3MMye']            = False     # (N*m); ; 
SubDyn['M3N4MMye']            = False     # (N*m); ; 
SubDyn['M3N5MMye']            = False     # (N*m); ; 
SubDyn['M3N6MMye']            = False     # (N*m); ; 
SubDyn['M3N7MMye']            = False     # (N*m); ; 
SubDyn['M3N8MMye']            = False     # (N*m); ; 
SubDyn['M3N9MMye']            = False     # (N*m); ; 
SubDyn['M4N1MMye']            = False     # (N*m); ; 
SubDyn['M4N2MMye']            = False     # (N*m); ; 
SubDyn['M4N3MMye']            = False     # (N*m); ; 
SubDyn['M4N4MMye']            = False     # (N*m); ; 
SubDyn['M4N5MMye']            = False     # (N*m); ; 
SubDyn['M4N6MMye']            = False     # (N*m); ; 
SubDyn['M4N7MMye']            = False     # (N*m); ; 
SubDyn['M4N8MMye']            = False     # (N*m); ; 
SubDyn['M4N9MMye']            = False     # (N*m); ; 
SubDyn['M5N1MMye']            = False     # (N*m); ; 
SubDyn['M5N2MMye']            = False     # (N*m); ; 
SubDyn['M5N3MMye']            = False     # (N*m); ; 
SubDyn['M5N4MMye']            = False     # (N*m); ; 
SubDyn['M5N5MMye']            = False     # (N*m); ; 
SubDyn['M5N6MMye']            = False     # (N*m); ; 
SubDyn['M5N7MMye']            = False     # (N*m); ; 
SubDyn['M5N8MMye']            = False     # (N*m); ; 
SubDyn['M5N9MMye']            = False     # (N*m); ; 
SubDyn['M6N1MMye']            = False     # (N*m); ; 
SubDyn['M6N2MMye']            = False     # (N*m); ; 
SubDyn['M6N3MMye']            = False     # (N*m); ; 
SubDyn['M6N4MMye']            = False     # (N*m); ; 
SubDyn['M6N5MMye']            = False     # (N*m); ; 
SubDyn['M6N6MMye']            = False     # (N*m); ; 
SubDyn['M6N7MMye']            = False     # (N*m); ; 
SubDyn['M6N8MMye']            = False     # (N*m); ; 
SubDyn['M6N9MMye']            = False     # (N*m); ; 
SubDyn['M7N1MMye']            = False     # (N*m); ; 
SubDyn['M7N2MMye']            = False     # (N*m); ; 
SubDyn['M7N3MMye']            = False     # (N*m); ; 
SubDyn['M7N4MMye']            = False     # (N*m); ; 
SubDyn['M7N5MMye']            = False     # (N*m); ; 
SubDyn['M7N6MMye']            = False     # (N*m); ; 
SubDyn['M7N7MMye']            = False     # (N*m); ; 
SubDyn['M7N8MMye']            = False     # (N*m); ; 
SubDyn['M7N9MMye']            = False     # (N*m); ; 
SubDyn['M8N1MMye']            = False     # (N*m); ; 
SubDyn['M8N2MMye']            = False     # (N*m); ; 
SubDyn['M8N3MMye']            = False     # (N*m); ; 
SubDyn['M8N4MMye']            = False     # (N*m); ; 
SubDyn['M8N5MMye']            = False     # (N*m); ; 
SubDyn['M8N6MMye']            = False     # (N*m); ; 
SubDyn['M8N7MMye']            = False     # (N*m); ; 
SubDyn['M8N8MMye']            = False     # (N*m); ; 
SubDyn['M8N9MMye']            = False     # (N*m); ; 
SubDyn['M9N1MMye']            = False     # (N*m); ; 
SubDyn['M9N2MMye']            = False     # (N*m); ; 
SubDyn['M9N3MMye']            = False     # (N*m); ; 
SubDyn['M9N4MMye']            = False     # (N*m); ; 
SubDyn['M9N5MMye']            = False     # (N*m); ; 
SubDyn['M9N6MMye']            = False     # (N*m); ; 
SubDyn['M9N7MMye']            = False     # (N*m); ; 
SubDyn['M9N8MMye']            = False     # (N*m); ; 
SubDyn['M9N9MMye']            = False     # (N*m); ; 
SubDyn['M1N1MMze']            = False     # (N*m); Torsion moment at Node Nj of member Mi- Local Reference System- Dynamic Component; 
SubDyn['M1N2MMze']            = False     # (N*m); ; 
SubDyn['M1N3MMze']            = False     # (N*m); ; 
SubDyn['M1N4MMze']            = False     # (N*m); ; 
SubDyn['M1N5MMze']            = False     # (N*m); ; 
SubDyn['M1N6MMze']            = False     # (N*m); ; 
SubDyn['M1N7MMze']            = False     # (N*m); ; 
SubDyn['M1N8MMze']            = False     # (N*m); ; 
SubDyn['M1N9MMze']            = False     # (N*m); ; 
SubDyn['M2N1MMze']            = False     # (N*m); ; 
SubDyn['M2N2MMze']            = False     # (N*m); ; 
SubDyn['M2N3MMze']            = False     # (N*m); ; 
SubDyn['M2N4MMze']            = False     # (N*m); ; 
SubDyn['M2N5MMze']            = False     # (N*m); ; 
SubDyn['M2N6MMze']            = False     # (N*m); ; 
SubDyn['M2N7MMze']            = False     # (N*m); ; 
SubDyn['M2N8MMze']            = False     # (N*m); ; 
SubDyn['M2N9MMze']            = False     # (N*m); ; 
SubDyn['M3N1MMze']            = False     # (N*m); ; 
SubDyn['M3N2MMze']            = False     # (N*m); ; 
SubDyn['M3N3MMze']            = False     # (N*m); ; 
SubDyn['M3N4MMze']            = False     # (N*m); ; 
SubDyn['M3N5MMze']            = False     # (N*m); ; 
SubDyn['M3N6MMze']            = False     # (N*m); ; 
SubDyn['M3N7MMze']            = False     # (N*m); ; 
SubDyn['M3N8MMze']            = False     # (N*m); ; 
SubDyn['M3N9MMze']            = False     # (N*m); ; 
SubDyn['M4N1MMze']            = False     # (N*m); ; 
SubDyn['M4N2MMze']            = False     # (N*m); ; 
SubDyn['M4N3MMze']            = False     # (N*m); ; 
SubDyn['M4N4MMze']            = False     # (N*m); ; 
SubDyn['M4N5MMze']            = False     # (N*m); ; 
SubDyn['M4N6MMze']            = False     # (N*m); ; 
SubDyn['M4N7MMze']            = False     # (N*m); ; 
SubDyn['M4N8MMze']            = False     # (N*m); ; 
SubDyn['M4N9MMze']            = False     # (N*m); ; 
SubDyn['M5N1MMze']            = False     # (N*m); ; 
SubDyn['M5N2MMze']            = False     # (N*m); ; 
SubDyn['M5N3MMze']            = False     # (N*m); ; 
SubDyn['M5N4MMze']            = False     # (N*m); ; 
SubDyn['M5N5MMze']            = False     # (N*m); ; 
SubDyn['M5N6MMze']            = False     # (N*m); ; 
SubDyn['M5N7MMze']            = False     # (N*m); ; 
SubDyn['M5N8MMze']            = False     # (N*m); ; 
SubDyn['M5N9MMze']            = False     # (N*m); ; 
SubDyn['M6N1MMze']            = False     # (N*m); ; 
SubDyn['M6N2MMze']            = False     # (N*m); ; 
SubDyn['M6N3MMze']            = False     # (N*m); ; 
SubDyn['M6N4MMze']            = False     # (N*m); ; 
SubDyn['M6N5MMze']            = False     # (N*m); ; 
SubDyn['M6N6MMze']            = False     # (N*m); ; 
SubDyn['M6N7MMze']            = False     # (N*m); ; 
SubDyn['M6N8MMze']            = False     # (N*m); ; 
SubDyn['M6N9MMze']            = False     # (N*m); ; 
SubDyn['M7N1MMze']            = False     # (N*m); ; 
SubDyn['M7N2MMze']            = False     # (N*m); ; 
SubDyn['M7N3MMze']            = False     # (N*m); ; 
SubDyn['M7N4MMze']            = False     # (N*m); ; 
SubDyn['M7N5MMze']            = False     # (N*m); ; 
SubDyn['M7N6MMze']            = False     # (N*m); ; 
SubDyn['M7N7MMze']            = False     # (N*m); ; 
SubDyn['M7N8MMze']            = False     # (N*m); ; 
SubDyn['M7N9MMze']            = False     # (N*m); ; 
SubDyn['M8N1MMze']            = False     # (N*m); ; 
SubDyn['M8N2MMze']            = False     # (N*m); ; 
SubDyn['M8N3MMze']            = False     # (N*m); ; 
SubDyn['M8N4MMze']            = False     # (N*m); ; 
SubDyn['M8N5MMze']            = False     # (N*m); ; 
SubDyn['M8N6MMze']            = False     # (N*m); ; 
SubDyn['M8N7MMze']            = False     # (N*m); ; 
SubDyn['M8N8MMze']            = False     # (N*m); ; 
SubDyn['M8N9MMze']            = False     # (N*m); ; 
SubDyn['M9N1MMze']            = False     # (N*m); ; 
SubDyn['M9N2MMze']            = False     # (N*m); ; 
SubDyn['M9N3MMze']            = False     # (N*m); ; 
SubDyn['M9N4MMze']            = False     # (N*m); ; 
SubDyn['M9N5MMze']            = False     # (N*m); ; 
SubDyn['M9N6MMze']            = False     # (N*m); ; 
SubDyn['M9N7MMze']            = False     # (N*m); ; 
SubDyn['M9N8MMze']            = False     # (N*m); ; 
SubDyn['M9N9MMze']            = False     # (N*m); ; 

# Displacements
SubDyn['M1N1TDxss']           = False     # (m); xss component of the displacement at Node Nj of member Mi- SS Reference System; 
SubDyn['M1N2TDxss']           = False     # (m); ; 
SubDyn['M1N3TDxss']           = False     # (m); ; 
SubDyn['M1N4TDxss']           = False     # (m); ; 
SubDyn['M1N5TDxss']           = False     # (m); ; 
SubDyn['M1N6TDxss']           = False     # (m); ; 
SubDyn['M1N7TDxss']           = False     # (m); ; 
SubDyn['M1N8TDxss']           = False     # (m); ; 
SubDyn['M1N9TDxss']           = False     # (m); ; 
SubDyn['M2N1TDxss']           = False     # (m); ; 
SubDyn['M2N2TDxss']           = False     # (m); ; 
SubDyn['M2N3TDxss']           = False     # (m); ; 
SubDyn['M2N4TDxss']           = False     # (m); ; 
SubDyn['M2N5TDxss']           = False     # (m); ; 
SubDyn['M2N6TDxss']           = False     # (m); ; 
SubDyn['M2N7TDxss']           = False     # (m); ; 
SubDyn['M2N8TDxss']           = False     # (m); ; 
SubDyn['M2N9TDxss']           = False     # (m); ; 
SubDyn['M3N1TDxss']           = False     # (m); ; 
SubDyn['M3N2TDxss']           = False     # (m); ; 
SubDyn['M3N3TDxss']           = False     # (m); ; 
SubDyn['M3N4TDxss']           = False     # (m); ; 
SubDyn['M3N5TDxss']           = False     # (m); ; 
SubDyn['M3N6TDxss']           = False     # (m); ; 
SubDyn['M3N7TDxss']           = False     # (m); ; 
SubDyn['M3N8TDxss']           = False     # (m); ; 
SubDyn['M3N9TDxss']           = False     # (m); ; 
SubDyn['M4N1TDxss']           = False     # (m); ; 
SubDyn['M4N2TDxss']           = False     # (m); ; 
SubDyn['M4N3TDxss']           = False     # (m); ; 
SubDyn['M4N4TDxss']           = False     # (m); ; 
SubDyn['M4N5TDxss']           = False     # (m); ; 
SubDyn['M4N6TDxss']           = False     # (m); ; 
SubDyn['M4N7TDxss']           = False     # (m); ; 
SubDyn['M4N8TDxss']           = False     # (m); ; 
SubDyn['M4N9TDxss']           = False     # (m); ; 
SubDyn['M5N1TDxss']           = False     # (m); ; 
SubDyn['M5N2TDxss']           = False     # (m); ; 
SubDyn['M5N3TDxss']           = False     # (m); ; 
SubDyn['M5N4TDxss']           = False     # (m); ; 
SubDyn['M5N5TDxss']           = False     # (m); ; 
SubDyn['M5N6TDxss']           = False     # (m); ; 
SubDyn['M5N7TDxss']           = False     # (m); ; 
SubDyn['M5N8TDxss']           = False     # (m); ; 
SubDyn['M5N9TDxss']           = False     # (m); ; 
SubDyn['M6N1TDxss']           = False     # (m); ; 
SubDyn['M6N2TDxss']           = False     # (m); ; 
SubDyn['M6N3TDxss']           = False     # (m); ; 
SubDyn['M6N4TDxss']           = False     # (m); ; 
SubDyn['M6N5TDxss']           = False     # (m); ; 
SubDyn['M6N6TDxss']           = False     # (m); ; 
SubDyn['M6N7TDxss']           = False     # (m); ; 
SubDyn['M6N8TDxss']           = False     # (m); ; 
SubDyn['M6N9TDxss']           = False     # (m); ; 
SubDyn['M7N1TDxss']           = False     # (m); ; 
SubDyn['M7N2TDxss']           = False     # (m); ; 
SubDyn['M7N3TDxss']           = False     # (m); ; 
SubDyn['M7N4TDxss']           = False     # (m); ; 
SubDyn['M7N5TDxss']           = False     # (m); ; 
SubDyn['M7N6TDxss']           = False     # (m); ; 
SubDyn['M7N7TDxss']           = False     # (m); ; 
SubDyn['M7N8TDxss']           = False     # (m); ; 
SubDyn['M7N9TDxss']           = False     # (m); ; 
SubDyn['M8N1TDxss']           = False     # (m); ; 
SubDyn['M8N2TDxss']           = False     # (m); ; 
SubDyn['M8N3TDxss']           = False     # (m); ; 
SubDyn['M8N4TDxss']           = False     # (m); ; 
SubDyn['M8N5TDxss']           = False     # (m); ; 
SubDyn['M8N6TDxss']           = False     # (m); ; 
SubDyn['M8N7TDxss']           = False     # (m); ; 
SubDyn['M8N8TDxss']           = False     # (m); ; 
SubDyn['M8N9TDxss']           = False     # (m); ; 
SubDyn['M9N1TDxss']           = False     # (m); ; 
SubDyn['M9N2TDxss']           = False     # (m); ; 
SubDyn['M9N3TDxss']           = False     # (m); ; 
SubDyn['M9N4TDxss']           = False     # (m); ; 
SubDyn['M9N5TDxss']           = False     # (m); ; 
SubDyn['M9N6TDxss']           = False     # (m); ; 
SubDyn['M9N7TDxss']           = False     # (m); ; 
SubDyn['M9N8TDxss']           = False     # (m); ; 
SubDyn['M9N9TDxss']           = False     # (m); ; 
SubDyn['M1N1TDyss']           = False     # (m); yss component of the displacement at Node Nj of member Mi- SS Reference System; 
SubDyn['M1N2TDyss']           = False     # (m); ; 
SubDyn['M1N3TDyss']           = False     # (m); ; 
SubDyn['M1N4TDyss']           = False     # (m); ; 
SubDyn['M1N5TDyss']           = False     # (m); ; 
SubDyn['M1N6TDyss']           = False     # (m); ; 
SubDyn['M1N7TDyss']           = False     # (m); ; 
SubDyn['M1N8TDyss']           = False     # (m); ; 
SubDyn['M1N9TDyss']           = False     # (m); ; 
SubDyn['M2N1TDyss']           = False     # (m); ; 
SubDyn['M2N2TDyss']           = False     # (m); ; 
SubDyn['M2N3TDyss']           = False     # (m); ; 
SubDyn['M2N4TDyss']           = False     # (m); ; 
SubDyn['M2N5TDyss']           = False     # (m); ; 
SubDyn['M2N6TDyss']           = False     # (m); ; 
SubDyn['M2N7TDyss']           = False     # (m); ; 
SubDyn['M2N8TDyss']           = False     # (m); ; 
SubDyn['M2N9TDyss']           = False     # (m); ; 
SubDyn['M3N1TDyss']           = False     # (m); ; 
SubDyn['M3N2TDyss']           = False     # (m); ; 
SubDyn['M3N3TDyss']           = False     # (m); ; 
SubDyn['M3N4TDyss']           = False     # (m); ; 
SubDyn['M3N5TDyss']           = False     # (m); ; 
SubDyn['M3N6TDyss']           = False     # (m); ; 
SubDyn['M3N7TDyss']           = False     # (m); ; 
SubDyn['M3N8TDyss']           = False     # (m); ; 
SubDyn['M3N9TDyss']           = False     # (m); ; 
SubDyn['M4N1TDyss']           = False     # (m); ; 
SubDyn['M4N2TDyss']           = False     # (m); ; 
SubDyn['M4N3TDyss']           = False     # (m); ; 
SubDyn['M4N4TDyss']           = False     # (m); ; 
SubDyn['M4N5TDyss']           = False     # (m); ; 
SubDyn['M4N6TDyss']           = False     # (m); ; 
SubDyn['M4N7TDyss']           = False     # (m); ; 
SubDyn['M4N8TDyss']           = False     # (m); ; 
SubDyn['M4N9TDyss']           = False     # (m); ; 
SubDyn['M5N1TDyss']           = False     # (m); ; 
SubDyn['M5N2TDyss']           = False     # (m); ; 
SubDyn['M5N3TDyss']           = False     # (m); ; 
SubDyn['M5N4TDyss']           = False     # (m); ; 
SubDyn['M5N5TDyss']           = False     # (m); ; 
SubDyn['M5N6TDyss']           = False     # (m); ; 
SubDyn['M5N7TDyss']           = False     # (m); ; 
SubDyn['M5N8TDyss']           = False     # (m); ; 
SubDyn['M5N9TDyss']           = False     # (m); ; 
SubDyn['M6N1TDyss']           = False     # (m); ; 
SubDyn['M6N2TDyss']           = False     # (m); ; 
SubDyn['M6N3TDyss']           = False     # (m); ; 
SubDyn['M6N4TDyss']           = False     # (m); ; 
SubDyn['M6N5TDyss']           = False     # (m); ; 
SubDyn['M6N6TDyss']           = False     # (m); ; 
SubDyn['M6N7TDyss']           = False     # (m); ; 
SubDyn['M6N8TDyss']           = False     # (m); ; 
SubDyn['M6N9TDyss']           = False     # (m); ; 
SubDyn['M7N1TDyss']           = False     # (m); ; 
SubDyn['M7N2TDyss']           = False     # (m); ; 
SubDyn['M7N3TDyss']           = False     # (m); ; 
SubDyn['M7N4TDyss']           = False     # (m); ; 
SubDyn['M7N5TDyss']           = False     # (m); ; 
SubDyn['M7N6TDyss']           = False     # (m); ; 
SubDyn['M7N7TDyss']           = False     # (m); ; 
SubDyn['M7N8TDyss']           = False     # (m); ; 
SubDyn['M7N9TDyss']           = False     # (m); ; 
SubDyn['M8N1TDyss']           = False     # (m); ; 
SubDyn['M8N2TDyss']           = False     # (m); ; 
SubDyn['M8N3TDyss']           = False     # (m); ; 
SubDyn['M8N4TDyss']           = False     # (m); ; 
SubDyn['M8N5TDyss']           = False     # (m); ; 
SubDyn['M8N6TDyss']           = False     # (m); ; 
SubDyn['M8N7TDyss']           = False     # (m); ; 
SubDyn['M8N8TDyss']           = False     # (m); ; 
SubDyn['M8N9TDyss']           = False     # (m); ; 
SubDyn['M9N1TDyss']           = False     # (m); ; 
SubDyn['M9N2TDyss']           = False     # (m); ; 
SubDyn['M9N3TDyss']           = False     # (m); ; 
SubDyn['M9N4TDyss']           = False     # (m); ; 
SubDyn['M9N5TDyss']           = False     # (m); ; 
SubDyn['M9N6TDyss']           = False     # (m); ; 
SubDyn['M9N7TDyss']           = False     # (m); ; 
SubDyn['M9N8TDyss']           = False     # (m); ; 
SubDyn['M9N9TDyss']           = False     # (m); ; 
SubDyn['M1N1TDzss']           = False     # (m); zss component of the displacement at Node Nj of member Mi- SS Reference System; 
SubDyn['M1N2TDzss']           = False     # (m); ; 
SubDyn['M1N3TDzss']           = False     # (m); ; 
SubDyn['M1N4TDzss']           = False     # (m); ; 
SubDyn['M1N5TDzss']           = False     # (m); ; 
SubDyn['M1N6TDzss']           = False     # (m); ; 
SubDyn['M1N7TDzss']           = False     # (m); ; 
SubDyn['M1N8TDzss']           = False     # (m); ; 
SubDyn['M1N9TDzss']           = False     # (m); ; 
SubDyn['M2N1TDzss']           = False     # (m); ; 
SubDyn['M2N2TDzss']           = False     # (m); ; 
SubDyn['M2N3TDzss']           = False     # (m); ; 
SubDyn['M2N4TDzss']           = False     # (m); ; 
SubDyn['M2N5TDzss']           = False     # (m); ; 
SubDyn['M2N6TDzss']           = False     # (m); ; 
SubDyn['M2N7TDzss']           = False     # (m); ; 
SubDyn['M2N8TDzss']           = False     # (m); ; 
SubDyn['M2N9TDzss']           = False     # (m); ; 
SubDyn['M3N1TDzss']           = False     # (m); ; 
SubDyn['M3N2TDzss']           = False     # (m); ; 
SubDyn['M3N3TDzss']           = False     # (m); ; 
SubDyn['M3N4TDzss']           = False     # (m); ; 
SubDyn['M3N5TDzss']           = False     # (m); ; 
SubDyn['M3N6TDzss']           = False     # (m); ; 
SubDyn['M3N7TDzss']           = False     # (m); ; 
SubDyn['M3N8TDzss']           = False     # (m); ; 
SubDyn['M3N9TDzss']           = False     # (m); ; 
SubDyn['M4N1TDzss']           = False     # (m); ; 
SubDyn['M4N2TDzss']           = False     # (m); ; 
SubDyn['M4N3TDzss']           = False     # (m); ; 
SubDyn['M4N4TDzss']           = False     # (m); ; 
SubDyn['M4N5TDzss']           = False     # (m); ; 
SubDyn['M4N6TDzss']           = False     # (m); ; 
SubDyn['M4N7TDzss']           = False     # (m); ; 
SubDyn['M4N8TDzss']           = False     # (m); ; 
SubDyn['M4N9TDzss']           = False     # (m); ; 
SubDyn['M5N1TDzss']           = False     # (m); ; 
SubDyn['M5N2TDzss']           = False     # (m); ; 
SubDyn['M5N3TDzss']           = False     # (m); ; 
SubDyn['M5N4TDzss']           = False     # (m); ; 
SubDyn['M5N5TDzss']           = False     # (m); ; 
SubDyn['M5N6TDzss']           = False     # (m); ; 
SubDyn['M5N7TDzss']           = False     # (m); ; 
SubDyn['M5N8TDzss']           = False     # (m); ; 
SubDyn['M5N9TDzss']           = False     # (m); ; 
SubDyn['M6N1TDzss']           = False     # (m); ; 
SubDyn['M6N2TDzss']           = False     # (m); ; 
SubDyn['M6N3TDzss']           = False     # (m); ; 
SubDyn['M6N4TDzss']           = False     # (m); ; 
SubDyn['M6N5TDzss']           = False     # (m); ; 
SubDyn['M6N6TDzss']           = False     # (m); ; 
SubDyn['M6N7TDzss']           = False     # (m); ; 
SubDyn['M6N8TDzss']           = False     # (m); ; 
SubDyn['M6N9TDzss']           = False     # (m); ; 
SubDyn['M7N1TDzss']           = False     # (m); ; 
SubDyn['M7N2TDzss']           = False     # (m); ; 
SubDyn['M7N3TDzss']           = False     # (m); ; 
SubDyn['M7N4TDzss']           = False     # (m); ; 
SubDyn['M7N5TDzss']           = False     # (m); ; 
SubDyn['M7N6TDzss']           = False     # (m); ; 
SubDyn['M7N7TDzss']           = False     # (m); ; 
SubDyn['M7N8TDzss']           = False     # (m); ; 
SubDyn['M7N9TDzss']           = False     # (m); ; 
SubDyn['M8N1TDzss']           = False     # (m); ; 
SubDyn['M8N2TDzss']           = False     # (m); ; 
SubDyn['M8N3TDzss']           = False     # (m); ; 
SubDyn['M8N4TDzss']           = False     # (m); ; 
SubDyn['M8N5TDzss']           = False     # (m); ; 
SubDyn['M8N6TDzss']           = False     # (m); ; 
SubDyn['M8N7TDzss']           = False     # (m); ; 
SubDyn['M8N8TDzss']           = False     # (m); ; 
SubDyn['M8N9TDzss']           = False     # (m); ; 
SubDyn['M9N1TDzss']           = False     # (m); ; 
SubDyn['M9N2TDzss']           = False     # (m); ; 
SubDyn['M9N3TDzss']           = False     # (m); ; 
SubDyn['M9N4TDzss']           = False     # (m); ; 
SubDyn['M9N5TDzss']           = False     # (m); ; 
SubDyn['M9N6TDzss']           = False     # (m); ; 
SubDyn['M9N7TDzss']           = False     # (m); ; 
SubDyn['M9N8TDzss']           = False     # (m); ; 
SubDyn['M9N9TDzss']           = False     # (m); ; 
SubDyn['M1N1RDxe']            = False     # (rad); xe component of the rotational displacement at Node Nj of member Mi-Element Reference System; 
SubDyn['M1N2RDxe']            = False     # (rad); ; 
SubDyn['M1N3RDxe']            = False     # (rad); ; 
SubDyn['M1N4RDxe']            = False     # (rad); ; 
SubDyn['M1N5RDxe']            = False     # (rad); ; 
SubDyn['M1N6RDxe']            = False     # (rad); ; 
SubDyn['M1N7RDxe']            = False     # (rad); ; 
SubDyn['M1N8RDxe']            = False     # (rad); ; 
SubDyn['M1N9RDxe']            = False     # (rad); ; 
SubDyn['M2N1RDxe']            = False     # (rad); ; 
SubDyn['M2N2RDxe']            = False     # (rad); ; 
SubDyn['M2N3RDxe']            = False     # (rad); ; 
SubDyn['M2N4RDxe']            = False     # (rad); ; 
SubDyn['M2N5RDxe']            = False     # (rad); ; 
SubDyn['M2N6RDxe']            = False     # (rad); ; 
SubDyn['M2N7RDxe']            = False     # (rad); ; 
SubDyn['M2N8RDxe']            = False     # (rad); ; 
SubDyn['M2N9RDxe']            = False     # (rad); ; 
SubDyn['M3N1RDxe']            = False     # (rad); ; 
SubDyn['M3N2RDxe']            = False     # (rad); ; 
SubDyn['M3N3RDxe']            = False     # (rad); ; 
SubDyn['M3N4RDxe']            = False     # (rad); ; 
SubDyn['M3N5RDxe']            = False     # (rad); ; 
SubDyn['M3N6RDxe']            = False     # (rad); ; 
SubDyn['M3N7RDxe']            = False     # (rad); ; 
SubDyn['M3N8RDxe']            = False     # (rad); ; 
SubDyn['M3N9RDxe']            = False     # (rad); ; 
SubDyn['M4N1RDxe']            = False     # (rad); ; 
SubDyn['M4N2RDxe']            = False     # (rad); ; 
SubDyn['M4N3RDxe']            = False     # (rad); ; 
SubDyn['M4N4RDxe']            = False     # (rad); ; 
SubDyn['M4N5RDxe']            = False     # (rad); ; 
SubDyn['M4N6RDxe']            = False     # (rad); ; 
SubDyn['M4N7RDxe']            = False     # (rad); ; 
SubDyn['M4N8RDxe']            = False     # (rad); ; 
SubDyn['M4N9RDxe']            = False     # (rad); ; 
SubDyn['M5N1RDxe']            = False     # (rad); ; 
SubDyn['M5N2RDxe']            = False     # (rad); ; 
SubDyn['M5N3RDxe']            = False     # (rad); ; 
SubDyn['M5N4RDxe']            = False     # (rad); ; 
SubDyn['M5N5RDxe']            = False     # (rad); ; 
SubDyn['M5N6RDxe']            = False     # (rad); ; 
SubDyn['M5N7RDxe']            = False     # (rad); ; 
SubDyn['M5N8RDxe']            = False     # (rad); ; 
SubDyn['M5N9RDxe']            = False     # (rad); ; 
SubDyn['M6N1RDxe']            = False     # (rad); ; 
SubDyn['M6N2RDxe']            = False     # (rad); ; 
SubDyn['M6N3RDxe']            = False     # (rad); ; 
SubDyn['M6N4RDxe']            = False     # (rad); ; 
SubDyn['M6N5RDxe']            = False     # (rad); ; 
SubDyn['M6N6RDxe']            = False     # (rad); ; 
SubDyn['M6N7RDxe']            = False     # (rad); ; 
SubDyn['M6N8RDxe']            = False     # (rad); ; 
SubDyn['M6N9RDxe']            = False     # (rad); ; 
SubDyn['M7N1RDxe']            = False     # (rad); ; 
SubDyn['M7N2RDxe']            = False     # (rad); ; 
SubDyn['M7N3RDxe']            = False     # (rad); ; 
SubDyn['M7N4RDxe']            = False     # (rad); ; 
SubDyn['M7N5RDxe']            = False     # (rad); ; 
SubDyn['M7N6RDxe']            = False     # (rad); ; 
SubDyn['M7N7RDxe']            = False     # (rad); ; 
SubDyn['M7N8RDxe']            = False     # (rad); ; 
SubDyn['M7N9RDxe']            = False     # (rad); ; 
SubDyn['M8N1RDxe']            = False     # (rad); ; 
SubDyn['M8N2RDxe']            = False     # (rad); ; 
SubDyn['M8N3RDxe']            = False     # (rad); ; 
SubDyn['M8N4RDxe']            = False     # (rad); ; 
SubDyn['M8N5RDxe']            = False     # (rad); ; 
SubDyn['M8N6RDxe']            = False     # (rad); ; 
SubDyn['M8N7RDxe']            = False     # (rad); ; 
SubDyn['M8N8RDxe']            = False     # (rad); ; 
SubDyn['M8N9RDxe']            = False     # (rad); ; 
SubDyn['M9N1RDxe']            = False     # (rad); ; 
SubDyn['M9N2RDxe']            = False     # (rad); ; 
SubDyn['M9N3RDxe']            = False     # (rad); ; 
SubDyn['M9N4RDxe']            = False     # (rad); ; 
SubDyn['M9N5RDxe']            = False     # (rad); ; 
SubDyn['M9N6RDxe']            = False     # (rad); ; 
SubDyn['M9N7RDxe']            = False     # (rad); ; 
SubDyn['M9N8RDxe']            = False     # (rad); ; 
SubDyn['M9N9RDxe']            = False     # (rad); ; 
SubDyn['M1N1RDye']            = False     # (rad); ye component of the rotational displacement at Node Nj of member Mi-Element Reference System; 
SubDyn['M1N2RDye']            = False     # (rad); ; 
SubDyn['M1N3RDye']            = False     # (rad); ; 
SubDyn['M1N4RDye']            = False     # (rad); ; 
SubDyn['M1N5RDye']            = False     # (rad); ; 
SubDyn['M1N6RDye']            = False     # (rad); ; 
SubDyn['M1N7RDye']            = False     # (rad); ; 
SubDyn['M1N8RDye']            = False     # (rad); ; 
SubDyn['M1N9RDye']            = False     # (rad); ; 
SubDyn['M2N1RDye']            = False     # (rad); ; 
SubDyn['M2N2RDye']            = False     # (rad); ; 
SubDyn['M2N3RDye']            = False     # (rad); ; 
SubDyn['M2N4RDye']            = False     # (rad); ; 
SubDyn['M2N5RDye']            = False     # (rad); ; 
SubDyn['M2N6RDye']            = False     # (rad); ; 
SubDyn['M2N7RDye']            = False     # (rad); ; 
SubDyn['M2N8RDye']            = False     # (rad); ; 
SubDyn['M2N9RDye']            = False     # (rad); ; 
SubDyn['M3N1RDye']            = False     # (rad); ; 
SubDyn['M3N2RDye']            = False     # (rad); ; 
SubDyn['M3N3RDye']            = False     # (rad); ; 
SubDyn['M3N4RDye']            = False     # (rad); ; 
SubDyn['M3N5RDye']            = False     # (rad); ; 
SubDyn['M3N6RDye']            = False     # (rad); ; 
SubDyn['M3N7RDye']            = False     # (rad); ; 
SubDyn['M3N8RDye']            = False     # (rad); ; 
SubDyn['M3N9RDye']            = False     # (rad); ; 
SubDyn['M4N1RDye']            = False     # (rad); ; 
SubDyn['M4N2RDye']            = False     # (rad); ; 
SubDyn['M4N3RDye']            = False     # (rad); ; 
SubDyn['M4N4RDye']            = False     # (rad); ; 
SubDyn['M4N5RDye']            = False     # (rad); ; 
SubDyn['M4N6RDye']            = False     # (rad); ; 
SubDyn['M4N7RDye']            = False     # (rad); ; 
SubDyn['M4N8RDye']            = False     # (rad); ; 
SubDyn['M4N9RDye']            = False     # (rad); ; 
SubDyn['M5N1RDye']            = False     # (rad); ; 
SubDyn['M5N2RDye']            = False     # (rad); ; 
SubDyn['M5N3RDye']            = False     # (rad); ; 
SubDyn['M5N4RDye']            = False     # (rad); ; 
SubDyn['M5N5RDye']            = False     # (rad); ; 
SubDyn['M5N6RDye']            = False     # (rad); ; 
SubDyn['M5N7RDye']            = False     # (rad); ; 
SubDyn['M5N8RDye']            = False     # (rad); ; 
SubDyn['M5N9RDye']            = False     # (rad); ; 
SubDyn['M6N1RDye']            = False     # (rad); ; 
SubDyn['M6N2RDye']            = False     # (rad); ; 
SubDyn['M6N3RDye']            = False     # (rad); ; 
SubDyn['M6N4RDye']            = False     # (rad); ; 
SubDyn['M6N5RDye']            = False     # (rad); ; 
SubDyn['M6N6RDye']            = False     # (rad); ; 
SubDyn['M6N7RDye']            = False     # (rad); ; 
SubDyn['M6N8RDye']            = False     # (rad); ; 
SubDyn['M6N9RDye']            = False     # (rad); ; 
SubDyn['M7N1RDye']            = False     # (rad); ; 
SubDyn['M7N2RDye']            = False     # (rad); ; 
SubDyn['M7N3RDye']            = False     # (rad); ; 
SubDyn['M7N4RDye']            = False     # (rad); ; 
SubDyn['M7N5RDye']            = False     # (rad); ; 
SubDyn['M7N6RDye']            = False     # (rad); ; 
SubDyn['M7N7RDye']            = False     # (rad); ; 
SubDyn['M7N8RDye']            = False     # (rad); ; 
SubDyn['M7N9RDye']            = False     # (rad); ; 
SubDyn['M8N1RDye']            = False     # (rad); ; 
SubDyn['M8N2RDye']            = False     # (rad); ; 
SubDyn['M8N3RDye']            = False     # (rad); ; 
SubDyn['M8N4RDye']            = False     # (rad); ; 
SubDyn['M8N5RDye']            = False     # (rad); ; 
SubDyn['M8N6RDye']            = False     # (rad); ; 
SubDyn['M8N7RDye']            = False     # (rad); ; 
SubDyn['M8N8RDye']            = False     # (rad); ; 
SubDyn['M8N9RDye']            = False     # (rad); ; 
SubDyn['M9N1RDye']            = False     # (rad); ; 
SubDyn['M9N2RDye']            = False     # (rad); ; 
SubDyn['M9N3RDye']            = False     # (rad); ; 
SubDyn['M9N4RDye']            = False     # (rad); ; 
SubDyn['M9N5RDye']            = False     # (rad); ; 
SubDyn['M9N6RDye']            = False     # (rad); ; 
SubDyn['M9N7RDye']            = False     # (rad); ; 
SubDyn['M9N8RDye']            = False     # (rad); ; 
SubDyn['M9N9RDye']            = False     # (rad); ; 
SubDyn['M1N1RDze']            = False     # (rad); ze component of the rotational displacement at Node Nj of member Mi-Element Reference System; 
SubDyn['M1N2RDze']            = False     # (rad); ; 
SubDyn['M1N3RDze']            = False     # (rad); ; 
SubDyn['M1N4RDze']            = False     # (rad); ; 
SubDyn['M1N5RDze']            = False     # (rad); ; 
SubDyn['M1N6RDze']            = False     # (rad); ; 
SubDyn['M1N7RDze']            = False     # (rad); ; 
SubDyn['M1N8RDze']            = False     # (rad); ; 
SubDyn['M1N9RDze']            = False     # (rad); ; 
SubDyn['M2N1RDze']            = False     # (rad); ; 
SubDyn['M2N2RDze']            = False     # (rad); ; 
SubDyn['M2N3RDze']            = False     # (rad); ; 
SubDyn['M2N4RDze']            = False     # (rad); ; 
SubDyn['M2N5RDze']            = False     # (rad); ; 
SubDyn['M2N6RDze']            = False     # (rad); ; 
SubDyn['M2N7RDze']            = False     # (rad); ; 
SubDyn['M2N8RDze']            = False     # (rad); ; 
SubDyn['M2N9RDze']            = False     # (rad); ; 
SubDyn['M3N1RDze']            = False     # (rad); ; 
SubDyn['M3N2RDze']            = False     # (rad); ; 
SubDyn['M3N3RDze']            = False     # (rad); ; 
SubDyn['M3N4RDze']            = False     # (rad); ; 
SubDyn['M3N5RDze']            = False     # (rad); ; 
SubDyn['M3N6RDze']            = False     # (rad); ; 
SubDyn['M3N7RDze']            = False     # (rad); ; 
SubDyn['M3N8RDze']            = False     # (rad); ; 
SubDyn['M3N9RDze']            = False     # (rad); ; 
SubDyn['M4N1RDze']            = False     # (rad); ; 
SubDyn['M4N2RDze']            = False     # (rad); ; 
SubDyn['M4N3RDze']            = False     # (rad); ; 
SubDyn['M4N4RDze']            = False     # (rad); ; 
SubDyn['M4N5RDze']            = False     # (rad); ; 
SubDyn['M4N6RDze']            = False     # (rad); ; 
SubDyn['M4N7RDze']            = False     # (rad); ; 
SubDyn['M4N8RDze']            = False     # (rad); ; 
SubDyn['M4N9RDze']            = False     # (rad); ; 
SubDyn['M5N1RDze']            = False     # (rad); ; 
SubDyn['M5N2RDze']            = False     # (rad); ; 
SubDyn['M5N3RDze']            = False     # (rad); ; 
SubDyn['M5N4RDze']            = False     # (rad); ; 
SubDyn['M5N5RDze']            = False     # (rad); ; 
SubDyn['M5N6RDze']            = False     # (rad); ; 
SubDyn['M5N7RDze']            = False     # (rad); ; 
SubDyn['M5N8RDze']            = False     # (rad); ; 
SubDyn['M5N9RDze']            = False     # (rad); ; 
SubDyn['M6N1RDze']            = False     # (rad); ; 
SubDyn['M6N2RDze']            = False     # (rad); ; 
SubDyn['M6N3RDze']            = False     # (rad); ; 
SubDyn['M6N4RDze']            = False     # (rad); ; 
SubDyn['M6N5RDze']            = False     # (rad); ; 
SubDyn['M6N6RDze']            = False     # (rad); ; 
SubDyn['M6N7RDze']            = False     # (rad); ; 
SubDyn['M6N8RDze']            = False     # (rad); ; 
SubDyn['M6N9RDze']            = False     # (rad); ; 
SubDyn['M7N1RDze']            = False     # (rad); ; 
SubDyn['M7N2RDze']            = False     # (rad); ; 
SubDyn['M7N3RDze']            = False     # (rad); ; 
SubDyn['M7N4RDze']            = False     # (rad); ; 
SubDyn['M7N5RDze']            = False     # (rad); ; 
SubDyn['M7N6RDze']            = False     # (rad); ; 
SubDyn['M7N7RDze']            = False     # (rad); ; 
SubDyn['M7N8RDze']            = False     # (rad); ; 
SubDyn['M7N9RDze']            = False     # (rad); ; 
SubDyn['M8N1RDze']            = False     # (rad); ; 
SubDyn['M8N2RDze']            = False     # (rad); ; 
SubDyn['M8N3RDze']            = False     # (rad); ; 
SubDyn['M8N4RDze']            = False     # (rad); ; 
SubDyn['M8N5RDze']            = False     # (rad); ; 
SubDyn['M8N6RDze']            = False     # (rad); ; 
SubDyn['M8N7RDze']            = False     # (rad); ; 
SubDyn['M8N8RDze']            = False     # (rad); ; 
SubDyn['M8N9RDze']            = False     # (rad); ; 
SubDyn['M9N1RDze']            = False     # (rad); ; 
SubDyn['M9N2RDze']            = False     # (rad); ; 
SubDyn['M9N3RDze']            = False     # (rad); ; 
SubDyn['M9N4RDze']            = False     # (rad); ; 
SubDyn['M9N5RDze']            = False     # (rad); ; 
SubDyn['M9N6RDze']            = False     # (rad); ; 
SubDyn['M9N7RDze']            = False     # (rad); ; 
SubDyn['M9N8RDze']            = False     # (rad); ; 
SubDyn['M9N9RDze']            = False     # (rad); ; 

# Accelerations
SubDyn['M1N1TAxe']            = False     # (m/s^2); xe component of the translational acceleration displacement at Node Nj of member Mi-Element Reference System; 
SubDyn['M1N2TAxe']            = False     # (m/s^2); ; 
SubDyn['M1N3TAxe']            = False     # (m/s^2); ; 
SubDyn['M1N4TAxe']            = False     # (m/s^2); ; 
SubDyn['M1N5TAxe']            = False     # (m/s^2); ; 
SubDyn['M1N6TAxe']            = False     # (m/s^2); ; 
SubDyn['M1N7TAxe']            = False     # (m/s^2); ; 
SubDyn['M1N8TAxe']            = False     # (m/s^2); ; 
SubDyn['M1N9TAxe']            = False     # (m/s^2); ; 
SubDyn['M2N1TAxe']            = False     # (m/s^2); ; 
SubDyn['M2N2TAxe']            = False     # (m/s^2); ; 
SubDyn['M2N3TAxe']            = False     # (m/s^2); ; 
SubDyn['M2N4TAxe']            = False     # (m/s^2); ; 
SubDyn['M2N5TAxe']            = False     # (m/s^2); ; 
SubDyn['M2N6TAxe']            = False     # (m/s^2); ; 
SubDyn['M2N7TAxe']            = False     # (m/s^2); ; 
SubDyn['M2N8TAxe']            = False     # (m/s^2); ; 
SubDyn['M2N9TAxe']            = False     # (m/s^2); ; 
SubDyn['M3N1TAxe']            = False     # (m/s^2); ; 
SubDyn['M3N2TAxe']            = False     # (m/s^2); ; 
SubDyn['M3N3TAxe']            = False     # (m/s^2); ; 
SubDyn['M3N4TAxe']            = False     # (m/s^2); ; 
SubDyn['M3N5TAxe']            = False     # (m/s^2); ; 
SubDyn['M3N6TAxe']            = False     # (m/s^2); ; 
SubDyn['M3N7TAxe']            = False     # (m/s^2); ; 
SubDyn['M3N8TAxe']            = False     # (m/s^2); ; 
SubDyn['M3N9TAxe']            = False     # (m/s^2); ; 
SubDyn['M4N1TAxe']            = False     # (m/s^2); ; 
SubDyn['M4N2TAxe']            = False     # (m/s^2); ; 
SubDyn['M4N3TAxe']            = False     # (m/s^2); ; 
SubDyn['M4N4TAxe']            = False     # (m/s^2); ; 
SubDyn['M4N5TAxe']            = False     # (m/s^2); ; 
SubDyn['M4N6TAxe']            = False     # (m/s^2); ; 
SubDyn['M4N7TAxe']            = False     # (m/s^2); ; 
SubDyn['M4N8TAxe']            = False     # (m/s^2); ; 
SubDyn['M4N9TAxe']            = False     # (m/s^2); ; 
SubDyn['M5N1TAxe']            = False     # (m/s^2); ; 
SubDyn['M5N2TAxe']            = False     # (m/s^2); ; 
SubDyn['M5N3TAxe']            = False     # (m/s^2); ; 
SubDyn['M5N4TAxe']            = False     # (m/s^2); ; 
SubDyn['M5N5TAxe']            = False     # (m/s^2); ; 
SubDyn['M5N6TAxe']            = False     # (m/s^2); ; 
SubDyn['M5N7TAxe']            = False     # (m/s^2); ; 
SubDyn['M5N8TAxe']            = False     # (m/s^2); ; 
SubDyn['M5N9TAxe']            = False     # (m/s^2); ; 
SubDyn['M6N1TAxe']            = False     # (m/s^2); ; 
SubDyn['M6N2TAxe']            = False     # (m/s^2); ; 
SubDyn['M6N3TAxe']            = False     # (m/s^2); ; 
SubDyn['M6N4TAxe']            = False     # (m/s^2); ; 
SubDyn['M6N5TAxe']            = False     # (m/s^2); ; 
SubDyn['M6N6TAxe']            = False     # (m/s^2); ; 
SubDyn['M6N7TAxe']            = False     # (m/s^2); ; 
SubDyn['M6N8TAxe']            = False     # (m/s^2); ; 
SubDyn['M6N9TAxe']            = False     # (m/s^2); ; 
SubDyn['M7N1TAxe']            = False     # (m/s^2); ; 
SubDyn['M7N2TAxe']            = False     # (m/s^2); ; 
SubDyn['M7N3TAxe']            = False     # (m/s^2); ; 
SubDyn['M7N4TAxe']            = False     # (m/s^2); ; 
SubDyn['M7N5TAxe']            = False     # (m/s^2); ; 
SubDyn['M7N6TAxe']            = False     # (m/s^2); ; 
SubDyn['M7N7TAxe']            = False     # (m/s^2); ; 
SubDyn['M7N8TAxe']            = False     # (m/s^2); ; 
SubDyn['M7N9TAxe']            = False     # (m/s^2); ; 
SubDyn['M8N1TAxe']            = False     # (m/s^2); ; 
SubDyn['M8N2TAxe']            = False     # (m/s^2); ; 
SubDyn['M8N3TAxe']            = False     # (m/s^2); ; 
SubDyn['M8N4TAxe']            = False     # (m/s^2); ; 
SubDyn['M8N5TAxe']            = False     # (m/s^2); ; 
SubDyn['M8N6TAxe']            = False     # (m/s^2); ; 
SubDyn['M8N7TAxe']            = False     # (m/s^2); ; 
SubDyn['M8N8TAxe']            = False     # (m/s^2); ; 
SubDyn['M8N9TAxe']            = False     # (m/s^2); ; 
SubDyn['M9N1TAxe']            = False     # (m/s^2); ; 
SubDyn['M9N2TAxe']            = False     # (m/s^2); ; 
SubDyn['M9N3TAxe']            = False     # (m/s^2); ; 
SubDyn['M9N4TAxe']            = False     # (m/s^2); ; 
SubDyn['M9N5TAxe']            = False     # (m/s^2); ; 
SubDyn['M9N6TAxe']            = False     # (m/s^2); ; 
SubDyn['M9N7TAxe']            = False     # (m/s^2); ; 
SubDyn['M9N8TAxe']            = False     # (m/s^2); ; 
SubDyn['M9N9TAxe']            = False     # (m/s^2); ; 
SubDyn['M1N1TAye']            = False     # (m/s^2); ye component of the translational acceleration displacement at Node Nj of member Mi-Element Reference System; 
SubDyn['M1N2TAye']            = False     # (m/s^2); ; 
SubDyn['M1N3TAye']            = False     # (m/s^2); ; 
SubDyn['M1N4TAye']            = False     # (m/s^2); ; 
SubDyn['M1N5TAye']            = False     # (m/s^2); ; 
SubDyn['M1N6TAye']            = False     # (m/s^2); ; 
SubDyn['M1N7TAye']            = False     # (m/s^2); ; 
SubDyn['M1N8TAye']            = False     # (m/s^2); ; 
SubDyn['M1N9TAye']            = False     # (m/s^2); ; 
SubDyn['M2N1TAye']            = False     # (m/s^2); ; 
SubDyn['M2N2TAye']            = False     # (m/s^2); ; 
SubDyn['M2N3TAye']            = False     # (m/s^2); ; 
SubDyn['M2N4TAye']            = False     # (m/s^2); ; 
SubDyn['M2N5TAye']            = False     # (m/s^2); ; 
SubDyn['M2N6TAye']            = False     # (m/s^2); ; 
SubDyn['M2N7TAye']            = False     # (m/s^2); ; 
SubDyn['M2N8TAye']            = False     # (m/s^2); ; 
SubDyn['M2N9TAye']            = False     # (m/s^2); ; 
SubDyn['M3N1TAye']            = False     # (m/s^2); ; 
SubDyn['M3N2TAye']            = False     # (m/s^2); ; 
SubDyn['M3N3TAye']            = False     # (m/s^2); ; 
SubDyn['M3N4TAye']            = False     # (m/s^2); ; 
SubDyn['M3N5TAye']            = False     # (m/s^2); ; 
SubDyn['M3N6TAye']            = False     # (m/s^2); ; 
SubDyn['M3N7TAye']            = False     # (m/s^2); ; 
SubDyn['M3N8TAye']            = False     # (m/s^2); ; 
SubDyn['M3N9TAye']            = False     # (m/s^2); ; 
SubDyn['M4N1TAye']            = False     # (m/s^2); ; 
SubDyn['M4N2TAye']            = False     # (m/s^2); ; 
SubDyn['M4N3TAye']            = False     # (m/s^2); ; 
SubDyn['M4N4TAye']            = False     # (m/s^2); ; 
SubDyn['M4N5TAye']            = False     # (m/s^2); ; 
SubDyn['M4N6TAye']            = False     # (m/s^2); ; 
SubDyn['M4N7TAye']            = False     # (m/s^2); ; 
SubDyn['M4N8TAye']            = False     # (m/s^2); ; 
SubDyn['M4N9TAye']            = False     # (m/s^2); ; 
SubDyn['M5N1TAye']            = False     # (m/s^2); ; 
SubDyn['M5N2TAye']            = False     # (m/s^2); ; 
SubDyn['M5N3TAye']            = False     # (m/s^2); ; 
SubDyn['M5N4TAye']            = False     # (m/s^2); ; 
SubDyn['M5N5TAye']            = False     # (m/s^2); ; 
SubDyn['M5N6TAye']            = False     # (m/s^2); ; 
SubDyn['M5N7TAye']            = False     # (m/s^2); ; 
SubDyn['M5N8TAye']            = False     # (m/s^2); ; 
SubDyn['M5N9TAye']            = False     # (m/s^2); ; 
SubDyn['M6N1TAye']            = False     # (m/s^2); ; 
SubDyn['M6N2TAye']            = False     # (m/s^2); ; 
SubDyn['M6N3TAye']            = False     # (m/s^2); ; 
SubDyn['M6N4TAye']            = False     # (m/s^2); ; 
SubDyn['M6N5TAye']            = False     # (m/s^2); ; 
SubDyn['M6N6TAye']            = False     # (m/s^2); ; 
SubDyn['M6N7TAye']            = False     # (m/s^2); ; 
SubDyn['M6N8TAye']            = False     # (m/s^2); ; 
SubDyn['M6N9TAye']            = False     # (m/s^2); ; 
SubDyn['M7N1TAye']            = False     # (m/s^2); ; 
SubDyn['M7N2TAye']            = False     # (m/s^2); ; 
SubDyn['M7N3TAye']            = False     # (m/s^2); ; 
SubDyn['M7N4TAye']            = False     # (m/s^2); ; 
SubDyn['M7N5TAye']            = False     # (m/s^2); ; 
SubDyn['M7N6TAye']            = False     # (m/s^2); ; 
SubDyn['M7N7TAye']            = False     # (m/s^2); ; 
SubDyn['M7N8TAye']            = False     # (m/s^2); ; 
SubDyn['M7N9TAye']            = False     # (m/s^2); ; 
SubDyn['M8N1TAye']            = False     # (m/s^2); ; 
SubDyn['M8N2TAye']            = False     # (m/s^2); ; 
SubDyn['M8N3TAye']            = False     # (m/s^2); ; 
SubDyn['M8N4TAye']            = False     # (m/s^2); ; 
SubDyn['M8N5TAye']            = False     # (m/s^2); ; 
SubDyn['M8N6TAye']            = False     # (m/s^2); ; 
SubDyn['M8N7TAye']            = False     # (m/s^2); ; 
SubDyn['M8N8TAye']            = False     # (m/s^2); ; 
SubDyn['M8N9TAye']            = False     # (m/s^2); ; 
SubDyn['M9N1TAye']            = False     # (m/s^2); ; 
SubDyn['M9N2TAye']            = False     # (m/s^2); ; 
SubDyn['M9N3TAye']            = False     # (m/s^2); ; 
SubDyn['M9N4TAye']            = False     # (m/s^2); ; 
SubDyn['M9N5TAye']            = False     # (m/s^2); ; 
SubDyn['M9N6TAye']            = False     # (m/s^2); ; 
SubDyn['M9N7TAye']            = False     # (m/s^2); ; 
SubDyn['M9N8TAye']            = False     # (m/s^2); ; 
SubDyn['M9N9TAye']            = False     # (m/s^2); ; 
SubDyn['M1N1TAze']            = False     # (m/s^2); ze component of the translational acceleration displacement at Node Nj of member Mi-Element Reference System; 
SubDyn['M1N2TAze']            = False     # (m/s^2); ; 
SubDyn['M1N3TAze']            = False     # (m/s^2); ; 
SubDyn['M1N4TAze']            = False     # (m/s^2); ; 
SubDyn['M1N5TAze']            = False     # (m/s^2); ; 
SubDyn['M1N6TAze']            = False     # (m/s^2); ; 
SubDyn['M1N7TAze']            = False     # (m/s^2); ; 
SubDyn['M1N8TAze']            = False     # (m/s^2); ; 
SubDyn['M1N9TAze']            = False     # (m/s^2); ; 
SubDyn['M2N1TAze']            = False     # (m/s^2); ; 
SubDyn['M2N2TAze']            = False     # (m/s^2); ; 
SubDyn['M2N3TAze']            = False     # (m/s^2); ; 
SubDyn['M2N4TAze']            = False     # (m/s^2); ; 
SubDyn['M2N5TAze']            = False     # (m/s^2); ; 
SubDyn['M2N6TAze']            = False     # (m/s^2); ; 
SubDyn['M2N7TAze']            = False     # (m/s^2); ; 
SubDyn['M2N8TAze']            = False     # (m/s^2); ; 
SubDyn['M2N9TAze']            = False     # (m/s^2); ; 
SubDyn['M3N1TAze']            = False     # (m/s^2); ; 
SubDyn['M3N2TAze']            = False     # (m/s^2); ; 
SubDyn['M3N3TAze']            = False     # (m/s^2); ; 
SubDyn['M3N4TAze']            = False     # (m/s^2); ; 
SubDyn['M3N5TAze']            = False     # (m/s^2); ; 
SubDyn['M3N6TAze']            = False     # (m/s^2); ; 
SubDyn['M3N7TAze']            = False     # (m/s^2); ; 
SubDyn['M3N8TAze']            = False     # (m/s^2); ; 
SubDyn['M3N9TAze']            = False     # (m/s^2); ; 
SubDyn['M4N1TAze']            = False     # (m/s^2); ; 
SubDyn['M4N2TAze']            = False     # (m/s^2); ; 
SubDyn['M4N3TAze']            = False     # (m/s^2); ; 
SubDyn['M4N4TAze']            = False     # (m/s^2); ; 
SubDyn['M4N5TAze']            = False     # (m/s^2); ; 
SubDyn['M4N6TAze']            = False     # (m/s^2); ; 
SubDyn['M4N7TAze']            = False     # (m/s^2); ; 
SubDyn['M4N8TAze']            = False     # (m/s^2); ; 
SubDyn['M4N9TAze']            = False     # (m/s^2); ; 
SubDyn['M5N1TAze']            = False     # (m/s^2); ; 
SubDyn['M5N2TAze']            = False     # (m/s^2); ; 
SubDyn['M5N3TAze']            = False     # (m/s^2); ; 
SubDyn['M5N4TAze']            = False     # (m/s^2); ; 
SubDyn['M5N5TAze']            = False     # (m/s^2); ; 
SubDyn['M5N6TAze']            = False     # (m/s^2); ; 
SubDyn['M5N7TAze']            = False     # (m/s^2); ; 
SubDyn['M5N8TAze']            = False     # (m/s^2); ; 
SubDyn['M5N9TAze']            = False     # (m/s^2); ; 
SubDyn['M6N1TAze']            = False     # (m/s^2); ; 
SubDyn['M6N2TAze']            = False     # (m/s^2); ; 
SubDyn['M6N3TAze']            = False     # (m/s^2); ; 
SubDyn['M6N4TAze']            = False     # (m/s^2); ; 
SubDyn['M6N5TAze']            = False     # (m/s^2); ; 
SubDyn['M6N6TAze']            = False     # (m/s^2); ; 
SubDyn['M6N7TAze']            = False     # (m/s^2); ; 
SubDyn['M6N8TAze']            = False     # (m/s^2); ; 
SubDyn['M6N9TAze']            = False     # (m/s^2); ; 
SubDyn['M7N1TAze']            = False     # (m/s^2); ; 
SubDyn['M7N2TAze']            = False     # (m/s^2); ; 
SubDyn['M7N3TAze']            = False     # (m/s^2); ; 
SubDyn['M7N4TAze']            = False     # (m/s^2); ; 
SubDyn['M7N5TAze']            = False     # (m/s^2); ; 
SubDyn['M7N6TAze']            = False     # (m/s^2); ; 
SubDyn['M7N7TAze']            = False     # (m/s^2); ; 
SubDyn['M7N8TAze']            = False     # (m/s^2); ; 
SubDyn['M7N9TAze']            = False     # (m/s^2); ; 
SubDyn['M8N1TAze']            = False     # (m/s^2); ; 
SubDyn['M8N2TAze']            = False     # (m/s^2); ; 
SubDyn['M8N3TAze']            = False     # (m/s^2); ; 
SubDyn['M8N4TAze']            = False     # (m/s^2); ; 
SubDyn['M8N5TAze']            = False     # (m/s^2); ; 
SubDyn['M8N6TAze']            = False     # (m/s^2); ; 
SubDyn['M8N7TAze']            = False     # (m/s^2); ; 
SubDyn['M8N8TAze']            = False     # (m/s^2); ; 
SubDyn['M8N9TAze']            = False     # (m/s^2); ; 
SubDyn['M9N1TAze']            = False     # (m/s^2); ; 
SubDyn['M9N2TAze']            = False     # (m/s^2); ; 
SubDyn['M9N3TAze']            = False     # (m/s^2); ; 
SubDyn['M9N4TAze']            = False     # (m/s^2); ; 
SubDyn['M9N5TAze']            = False     # (m/s^2); ; 
SubDyn['M9N6TAze']            = False     # (m/s^2); ; 
SubDyn['M9N7TAze']            = False     # (m/s^2); ; 
SubDyn['M9N8TAze']            = False     # (m/s^2); ; 
SubDyn['M9N9TAze']            = False     # (m/s^2); ; 
SubDyn['M1N1RAxe']            = False     # (rad/s^2); xe component of the rotational acceleration displacement at Node Nj of member Mi-Element Reference System; 
SubDyn['M1N2RAxe']            = False     # (rad/s^2); ; 
SubDyn['M1N3RAxe']            = False     # (rad/s^2); ; 
SubDyn['M1N4RAxe']            = False     # (rad/s^2); ; 
SubDyn['M1N5RAxe']            = False     # (rad/s^2); ; 
SubDyn['M1N6RAxe']            = False     # (rad/s^2); ; 
SubDyn['M1N7RAxe']            = False     # (rad/s^2); ; 
SubDyn['M1N8RAxe']            = False     # (rad/s^2); ; 
SubDyn['M1N9RAxe']            = False     # (rad/s^2); ; 
SubDyn['M2N1RAxe']            = False     # (rad/s^2); ; 
SubDyn['M2N2RAxe']            = False     # (rad/s^2); ; 
SubDyn['M2N3RAxe']            = False     # (rad/s^2); ; 
SubDyn['M2N4RAxe']            = False     # (rad/s^2); ; 
SubDyn['M2N5RAxe']            = False     # (rad/s^2); ; 
SubDyn['M2N6RAxe']            = False     # (rad/s^2); ; 
SubDyn['M2N7RAxe']            = False     # (rad/s^2); ; 
SubDyn['M2N8RAxe']            = False     # (rad/s^2); ; 
SubDyn['M2N9RAxe']            = False     # (rad/s^2); ; 
SubDyn['M3N1RAxe']            = False     # (rad/s^2); ; 
SubDyn['M3N2RAxe']            = False     # (rad/s^2); ; 
SubDyn['M3N3RAxe']            = False     # (rad/s^2); ; 
SubDyn['M3N4RAxe']            = False     # (rad/s^2); ; 
SubDyn['M3N5RAxe']            = False     # (rad/s^2); ; 
SubDyn['M3N6RAxe']            = False     # (rad/s^2); ; 
SubDyn['M3N7RAxe']            = False     # (rad/s^2); ; 
SubDyn['M3N8RAxe']            = False     # (rad/s^2); ; 
SubDyn['M3N9RAxe']            = False     # (rad/s^2); ; 
SubDyn['M4N1RAxe']            = False     # (rad/s^2); ; 
SubDyn['M4N2RAxe']            = False     # (rad/s^2); ; 
SubDyn['M4N3RAxe']            = False     # (rad/s^2); ; 
SubDyn['M4N4RAxe']            = False     # (rad/s^2); ; 
SubDyn['M4N5RAxe']            = False     # (rad/s^2); ; 
SubDyn['M4N6RAxe']            = False     # (rad/s^2); ; 
SubDyn['M4N7RAxe']            = False     # (rad/s^2); ; 
SubDyn['M4N8RAxe']            = False     # (rad/s^2); ; 
SubDyn['M4N9RAxe']            = False     # (rad/s^2); ; 
SubDyn['M5N1RAxe']            = False     # (rad/s^2); ; 
SubDyn['M5N2RAxe']            = False     # (rad/s^2); ; 
SubDyn['M5N3RAxe']            = False     # (rad/s^2); ; 
SubDyn['M5N4RAxe']            = False     # (rad/s^2); ; 
SubDyn['M5N5RAxe']            = False     # (rad/s^2); ; 
SubDyn['M5N6RAxe']            = False     # (rad/s^2); ; 
SubDyn['M5N7RAxe']            = False     # (rad/s^2); ; 
SubDyn['M5N8RAxe']            = False     # (rad/s^2); ; 
SubDyn['M5N9RAxe']            = False     # (rad/s^2); ; 
SubDyn['M6N1RAxe']            = False     # (rad/s^2); ; 
SubDyn['M6N2RAxe']            = False     # (rad/s^2); ; 
SubDyn['M6N3RAxe']            = False     # (rad/s^2); ; 
SubDyn['M6N4RAxe']            = False     # (rad/s^2); ; 
SubDyn['M6N5RAxe']            = False     # (rad/s^2); ; 
SubDyn['M6N6RAxe']            = False     # (rad/s^2); ; 
SubDyn['M6N7RAxe']            = False     # (rad/s^2); ; 
SubDyn['M6N8RAxe']            = False     # (rad/s^2); ; 
SubDyn['M6N9RAxe']            = False     # (rad/s^2); ; 
SubDyn['M7N1RAxe']            = False     # (rad/s^2); ; 
SubDyn['M7N2RAxe']            = False     # (rad/s^2); ; 
SubDyn['M7N3RAxe']            = False     # (rad/s^2); ; 
SubDyn['M7N4RAxe']            = False     # (rad/s^2); ; 
SubDyn['M7N5RAxe']            = False     # (rad/s^2); ; 
SubDyn['M7N6RAxe']            = False     # (rad/s^2); ; 
SubDyn['M7N7RAxe']            = False     # (rad/s^2); ; 
SubDyn['M7N8RAxe']            = False     # (rad/s^2); ; 
SubDyn['M7N9RAxe']            = False     # (rad/s^2); ; 
SubDyn['M8N1RAxe']            = False     # (rad/s^2); ; 
SubDyn['M8N2RAxe']            = False     # (rad/s^2); ; 
SubDyn['M8N3RAxe']            = False     # (rad/s^2); ; 
SubDyn['M8N4RAxe']            = False     # (rad/s^2); ; 
SubDyn['M8N5RAxe']            = False     # (rad/s^2); ; 
SubDyn['M8N6RAxe']            = False     # (rad/s^2); ; 
SubDyn['M8N7RAxe']            = False     # (rad/s^2); ; 
SubDyn['M8N8RAxe']            = False     # (rad/s^2); ; 
SubDyn['M8N9RAxe']            = False     # (rad/s^2); ; 
SubDyn['M9N1RAxe']            = False     # (rad/s^2); ; 
SubDyn['M9N2RAxe']            = False     # (rad/s^2); ; 
SubDyn['M9N3RAxe']            = False     # (rad/s^2); ; 
SubDyn['M9N4RAxe']            = False     # (rad/s^2); ; 
SubDyn['M9N5RAxe']            = False     # (rad/s^2); ; 
SubDyn['M9N6RAxe']            = False     # (rad/s^2); ; 
SubDyn['M9N7RAxe']            = False     # (rad/s^2); ; 
SubDyn['M9N8RAxe']            = False     # (rad/s^2); ; 
SubDyn['M9N9RAxe']            = False     # (rad/s^2); ; 
SubDyn['M1N1RAye']            = False     # (rad/s^2); ye component of the rotational acceleration displacement at Node Nj of member Mi-Element Reference System; 
SubDyn['M1N2RAye']            = False     # (rad/s^2); ; 
SubDyn['M1N3RAye']            = False     # (rad/s^2); ; 
SubDyn['M1N4RAye']            = False     # (rad/s^2); ; 
SubDyn['M1N5RAye']            = False     # (rad/s^2); ; 
SubDyn['M1N6RAye']            = False     # (rad/s^2); ; 
SubDyn['M1N7RAye']            = False     # (rad/s^2); ; 
SubDyn['M1N8RAye']            = False     # (rad/s^2); ; 
SubDyn['M1N9RAye']            = False     # (rad/s^2); ; 
SubDyn['M2N1RAye']            = False     # (rad/s^2); ; 
SubDyn['M2N2RAye']            = False     # (rad/s^2); ; 
SubDyn['M2N3RAye']            = False     # (rad/s^2); ; 
SubDyn['M2N4RAye']            = False     # (rad/s^2); ; 
SubDyn['M2N5RAye']            = False     # (rad/s^2); ; 
SubDyn['M2N6RAye']            = False     # (rad/s^2); ; 
SubDyn['M2N7RAye']            = False     # (rad/s^2); ; 
SubDyn['M2N8RAye']            = False     # (rad/s^2); ; 
SubDyn['M2N9RAye']            = False     # (rad/s^2); ; 
SubDyn['M3N1RAye']            = False     # (rad/s^2); ; 
SubDyn['M3N2RAye']            = False     # (rad/s^2); ; 
SubDyn['M3N3RAye']            = False     # (rad/s^2); ; 
SubDyn['M3N4RAye']            = False     # (rad/s^2); ; 
SubDyn['M3N5RAye']            = False     # (rad/s^2); ; 
SubDyn['M3N6RAye']            = False     # (rad/s^2); ; 
SubDyn['M3N7RAye']            = False     # (rad/s^2); ; 
SubDyn['M3N8RAye']            = False     # (rad/s^2); ; 
SubDyn['M3N9RAye']            = False     # (rad/s^2); ; 
SubDyn['M4N1RAye']            = False     # (rad/s^2); ; 
SubDyn['M4N2RAye']            = False     # (rad/s^2); ; 
SubDyn['M4N3RAye']            = False     # (rad/s^2); ; 
SubDyn['M4N4RAye']            = False     # (rad/s^2); ; 
SubDyn['M4N5RAye']            = False     # (rad/s^2); ; 
SubDyn['M4N6RAye']            = False     # (rad/s^2); ; 
SubDyn['M4N7RAye']            = False     # (rad/s^2); ; 
SubDyn['M4N8RAye']            = False     # (rad/s^2); ; 
SubDyn['M4N9RAye']            = False     # (rad/s^2); ; 
SubDyn['M5N1RAye']            = False     # (rad/s^2); ; 
SubDyn['M5N2RAye']            = False     # (rad/s^2); ; 
SubDyn['M5N3RAye']            = False     # (rad/s^2); ; 
SubDyn['M5N4RAye']            = False     # (rad/s^2); ; 
SubDyn['M5N5RAye']            = False     # (rad/s^2); ; 
SubDyn['M5N6RAye']            = False     # (rad/s^2); ; 
SubDyn['M5N7RAye']            = False     # (rad/s^2); ; 
SubDyn['M5N8RAye']            = False     # (rad/s^2); ; 
SubDyn['M5N9RAye']            = False     # (rad/s^2); ; 
SubDyn['M6N1RAye']            = False     # (rad/s^2); ; 
SubDyn['M6N2RAye']            = False     # (rad/s^2); ; 
SubDyn['M6N3RAye']            = False     # (rad/s^2); ; 
SubDyn['M6N4RAye']            = False     # (rad/s^2); ; 
SubDyn['M6N5RAye']            = False     # (rad/s^2); ; 
SubDyn['M6N6RAye']            = False     # (rad/s^2); ; 
SubDyn['M6N7RAye']            = False     # (rad/s^2); ; 
SubDyn['M6N8RAye']            = False     # (rad/s^2); ; 
SubDyn['M6N9RAye']            = False     # (rad/s^2); ; 
SubDyn['M7N1RAye']            = False     # (rad/s^2); ; 
SubDyn['M7N2RAye']            = False     # (rad/s^2); ; 
SubDyn['M7N3RAye']            = False     # (rad/s^2); ; 
SubDyn['M7N4RAye']            = False     # (rad/s^2); ; 
SubDyn['M7N5RAye']            = False     # (rad/s^2); ; 
SubDyn['M7N6RAye']            = False     # (rad/s^2); ; 
SubDyn['M7N7RAye']            = False     # (rad/s^2); ; 
SubDyn['M7N8RAye']            = False     # (rad/s^2); ; 
SubDyn['M7N9RAye']            = False     # (rad/s^2); ; 
SubDyn['M8N1RAye']            = False     # (rad/s^2); ; 
SubDyn['M8N2RAye']            = False     # (rad/s^2); ; 
SubDyn['M8N3RAye']            = False     # (rad/s^2); ; 
SubDyn['M8N4RAye']            = False     # (rad/s^2); ; 
SubDyn['M8N5RAye']            = False     # (rad/s^2); ; 
SubDyn['M8N6RAye']            = False     # (rad/s^2); ; 
SubDyn['M8N7RAye']            = False     # (rad/s^2); ; 
SubDyn['M8N8RAye']            = False     # (rad/s^2); ; 
SubDyn['M8N9RAye']            = False     # (rad/s^2); ; 
SubDyn['M9N1RAye']            = False     # (rad/s^2); ; 
SubDyn['M9N2RAye']            = False     # (rad/s^2); ; 
SubDyn['M9N3RAye']            = False     # (rad/s^2); ; 
SubDyn['M9N4RAye']            = False     # (rad/s^2); ; 
SubDyn['M9N5RAye']            = False     # (rad/s^2); ; 
SubDyn['M9N6RAye']            = False     # (rad/s^2); ; 
SubDyn['M9N7RAye']            = False     # (rad/s^2); ; 
SubDyn['M9N8RAye']            = False     # (rad/s^2); ; 
SubDyn['M9N9RAye']            = False     # (rad/s^2); ; 
SubDyn['M1N1RAze']            = False     # (rad/s^2); ze component of the rotational acceleration displacement at Node Nj of member Mi-Element Reference System; 
SubDyn['M1N2RAze']            = False     # (rad/s^2); ; 
SubDyn['M1N3RAze']            = False     # (rad/s^2); ; 
SubDyn['M1N4RAze']            = False     # (rad/s^2); ; 
SubDyn['M1N5RAze']            = False     # (rad/s^2); ; 
SubDyn['M1N6RAze']            = False     # (rad/s^2); ; 
SubDyn['M1N7RAze']            = False     # (rad/s^2); ; 
SubDyn['M1N8RAze']            = False     # (rad/s^2); ; 
SubDyn['M1N9RAze']            = False     # (rad/s^2); ; 
SubDyn['M2N1RAze']            = False     # (rad/s^2); ; 
SubDyn['M2N2RAze']            = False     # (rad/s^2); ; 
SubDyn['M2N3RAze']            = False     # (rad/s^2); ; 
SubDyn['M2N4RAze']            = False     # (rad/s^2); ; 
SubDyn['M2N5RAze']            = False     # (rad/s^2); ; 
SubDyn['M2N6RAze']            = False     # (rad/s^2); ; 
SubDyn['M2N7RAze']            = False     # (rad/s^2); ; 
SubDyn['M2N8RAze']            = False     # (rad/s^2); ; 
SubDyn['M2N9RAze']            = False     # (rad/s^2); ; 
SubDyn['M3N1RAze']            = False     # (rad/s^2); ; 
SubDyn['M3N2RAze']            = False     # (rad/s^2); ; 
SubDyn['M3N3RAze']            = False     # (rad/s^2); ; 
SubDyn['M3N4RAze']            = False     # (rad/s^2); ; 
SubDyn['M3N5RAze']            = False     # (rad/s^2); ; 
SubDyn['M3N6RAze']            = False     # (rad/s^2); ; 
SubDyn['M3N7RAze']            = False     # (rad/s^2); ; 
SubDyn['M3N8RAze']            = False     # (rad/s^2); ; 
SubDyn['M3N9RAze']            = False     # (rad/s^2); ; 
SubDyn['M4N1RAze']            = False     # (rad/s^2); ; 
SubDyn['M4N2RAze']            = False     # (rad/s^2); ; 
SubDyn['M4N3RAze']            = False     # (rad/s^2); ; 
SubDyn['M4N4RAze']            = False     # (rad/s^2); ; 
SubDyn['M4N5RAze']            = False     # (rad/s^2); ; 
SubDyn['M4N6RAze']            = False     # (rad/s^2); ; 
SubDyn['M4N7RAze']            = False     # (rad/s^2); ; 
SubDyn['M4N8RAze']            = False     # (rad/s^2); ; 
SubDyn['M4N9RAze']            = False     # (rad/s^2); ; 
SubDyn['M5N1RAze']            = False     # (rad/s^2); ; 
SubDyn['M5N2RAze']            = False     # (rad/s^2); ; 
SubDyn['M5N3RAze']            = False     # (rad/s^2); ; 
SubDyn['M5N4RAze']            = False     # (rad/s^2); ; 
SubDyn['M5N5RAze']            = False     # (rad/s^2); ; 
SubDyn['M5N6RAze']            = False     # (rad/s^2); ; 
SubDyn['M5N7RAze']            = False     # (rad/s^2); ; 
SubDyn['M5N8RAze']            = False     # (rad/s^2); ; 
SubDyn['M5N9RAze']            = False     # (rad/s^2); ; 
SubDyn['M6N1RAze']            = False     # (rad/s^2); ; 
SubDyn['M6N2RAze']            = False     # (rad/s^2); ; 
SubDyn['M6N3RAze']            = False     # (rad/s^2); ; 
SubDyn['M6N4RAze']            = False     # (rad/s^2); ; 
SubDyn['M6N5RAze']            = False     # (rad/s^2); ; 
SubDyn['M6N6RAze']            = False     # (rad/s^2); ; 
SubDyn['M6N7RAze']            = False     # (rad/s^2); ; 
SubDyn['M6N8RAze']            = False     # (rad/s^2); ; 
SubDyn['M6N9RAze']            = False     # (rad/s^2); ; 
SubDyn['M7N1RAze']            = False     # (rad/s^2); ; 
SubDyn['M7N2RAze']            = False     # (rad/s^2); ; 
SubDyn['M7N3RAze']            = False     # (rad/s^2); ; 
SubDyn['M7N4RAze']            = False     # (rad/s^2); ; 
SubDyn['M7N5RAze']            = False     # (rad/s^2); ; 
SubDyn['M7N6RAze']            = False     # (rad/s^2); ; 
SubDyn['M7N7RAze']            = False     # (rad/s^2); ; 
SubDyn['M7N8RAze']            = False     # (rad/s^2); ; 
SubDyn['M7N9RAze']            = False     # (rad/s^2); ; 
SubDyn['M8N1RAze']            = False     # (rad/s^2); ; 
SubDyn['M8N2RAze']            = False     # (rad/s^2); ; 
SubDyn['M8N3RAze']            = False     # (rad/s^2); ; 
SubDyn['M8N4RAze']            = False     # (rad/s^2); ; 
SubDyn['M8N5RAze']            = False     # (rad/s^2); ; 
SubDyn['M8N6RAze']            = False     # (rad/s^2); ; 
SubDyn['M8N7RAze']            = False     # (rad/s^2); ; 
SubDyn['M8N8RAze']            = False     # (rad/s^2); ; 
SubDyn['M8N9RAze']            = False     # (rad/s^2); ; 
SubDyn['M9N1RAze']            = False     # (rad/s^2); ; 
SubDyn['M9N2RAze']            = False     # (rad/s^2); ; 
SubDyn['M9N3RAze']            = False     # (rad/s^2); ; 
SubDyn['M9N4RAze']            = False     # (rad/s^2); ; 
SubDyn['M9N5RAze']            = False     # (rad/s^2); ; 
SubDyn['M9N6RAze']            = False     # (rad/s^2); ; 
SubDyn['M9N7RAze']            = False     # (rad/s^2); ; 
SubDyn['M9N8RAze']            = False     # (rad/s^2); ; 
SubDyn['M9N9RAze']            = False     # (rad/s^2); ; 

# Reactions
SubDyn['ReactFXss']           = False     # (N); Base Reaction Force along Xss; 
SubDyn['ReactFYss']           = False     # (N); Base Reaction Force along Yss; 
SubDyn['ReactFZss']           = False     # (N); Base Reaction Force along Zss; 
SubDyn['ReactMXss']           = False     # (Nm); Base Reaction Moment along Xss; 
SubDyn['ReactMYss']           = False     # (Nm); Base Reaction Moment along Yss; 
SubDyn['ReactMZss']           = False     # (Nm); Base Reaction Moment along Zss; 
SubDyn['IntfFXss']            = False     # (N); Interface Reaction Force along Xss; 
SubDyn['IntfFYss']            = False     # (N); Interface Reaction Force along Yss; 
SubDyn['IntfFZss']            = False     # (N); Interface Reaction Force along Zss; 
SubDyn['IntfMXss']            = False     # (Nm); Interface Reaction Moment along Xss; 
SubDyn['IntfMYss']            = False     # (Nm); Interface Reaction Moment along Yss; 
SubDyn['IntfMZss']            = False     # (Nm); Interface Reaction Moment along Zss; 

# Interface Deflections
SubDyn['IntfTDXss']           = False     # (m); Interface Deflection along Xss; 
SubDyn['IntfTDYss']           = False     # (m); Interface Deflection along Yss; 
SubDyn['IntfTDZss']           = False     # (m); Interface Deflection along Zss; 
SubDyn['IntfRDXss']           = False     # (rad); Interface Angular Deflection along Xss; 
SubDyn['IntfRDYss']           = False     # (rad); Interface Angular Deflection along Yss; 
SubDyn['IntfRDZss']           = False     # (rad); Interface Angular Deflection along Zss; 

# Interface Accelerations
SubDyn['IntfTAXss']           = False     # (m/s^2); Interface Acceleration along Xss; 
SubDyn['IntfTAYss']           = False     # (m/s^2); Interface Acceleration along Yss; 
SubDyn['IntfTAZss']           = False     # (m/s^2); Interface Acceleration along Zss; 
SubDyn['IntfRAXss']           = False     # (rad/s^2); Interface Angular Acceleration along Xss; 
SubDyn['IntfRAYss']           = False     # (rad/s^2); Interface Angular Acceleration along Yss; 
SubDyn['IntfRAZss']           = False     # (rad/s^2); Interface Angular Acceleration along Zss; 

# Modal Parameters
SubDyn['SSqm01']              = False     # (--); Modal Parameter (01-99) values; 
SubDyn['SSqm02']              = False     # (--); ; 
SubDyn['SSqm03']              = False     # (--); ; 
SubDyn['SSqm04']              = False     # (--); ; 
SubDyn['SSqm05']              = False     # (--); ; 
SubDyn['SSqm06']              = False     # (--); ; 
SubDyn['SSqm07']              = False     # (--); ; 
SubDyn['SSqm08']              = False     # (--); ; 
SubDyn['SSqm09']              = False     # (--); ; 
SubDyn['SSqm10']              = False     # (--); ; 
SubDyn['SSqm11']              = False     # (--); ; 
SubDyn['SSqm12']              = False     # (--); ; 
SubDyn['SSqm13']              = False     # (--); ; 
SubDyn['SSqm14']              = False     # (--); ; 
SubDyn['SSqm15']              = False     # (--); ; 
SubDyn['SSqm16']              = False     # (--); ; 
SubDyn['SSqm17']              = False     # (--); ; 
SubDyn['SSqm18']              = False     # (--); ; 
SubDyn['SSqm19']              = False     # (--); ; 
SubDyn['SSqm20']              = False     # (--); ; 
SubDyn['SSqm21']              = False     # (--); ; 
SubDyn['SSqm22']              = False     # (--); ; 
SubDyn['SSqm23']              = False     # (--); ; 
SubDyn['SSqm24']              = False     # (--); ; 
SubDyn['SSqm25']              = False     # (--); ; 
SubDyn['SSqm26']              = False     # (--); ; 
SubDyn['SSqm27']              = False     # (--); ; 
SubDyn['SSqm28']              = False     # (--); ; 
SubDyn['SSqm29']              = False     # (--); ; 
SubDyn['SSqm30']              = False     # (--); ; 
SubDyn['SSqm31']              = False     # (--); ; 
SubDyn['SSqm32']              = False     # (--); ; 
SubDyn['SSqm33']              = False     # (--); ; 
SubDyn['SSqm34']              = False     # (--); ; 
SubDyn['SSqm35']              = False     # (--); ; 
SubDyn['SSqm36']              = False     # (--); ; 
SubDyn['SSqm37']              = False     # (--); ; 
SubDyn['SSqm38']              = False     # (--); ; 
SubDyn['SSqm39']              = False     # (--); ; 
SubDyn['SSqm40']              = False     # (--); ; 
SubDyn['SSqm41']              = False     # (--); ; 
SubDyn['SSqm42']              = False     # (--); ; 
SubDyn['SSqm43']              = False     # (--); ; 
SubDyn['SSqm44']              = False     # (--); ; 
SubDyn['SSqm45']              = False     # (--); ; 
SubDyn['SSqm46']              = False     # (--); ; 
SubDyn['SSqm47']              = False     # (--); ; 
SubDyn['SSqm48']              = False     # (--); ; 
SubDyn['SSqm49']              = False     # (--); ; 
SubDyn['SSqm50']              = False     # (--); ; 
SubDyn['SSqm51']              = False     # (--); ; 
SubDyn['SSqm52']              = False     # (--); ; 
SubDyn['SSqm53']              = False     # (--); ; 
SubDyn['SSqm54']              = False     # (--); ; 
SubDyn['SSqm55']              = False     # (--); ; 
SubDyn['SSqm56']              = False     # (--); ; 
SubDyn['SSqm57']              = False     # (--); ; 
SubDyn['SSqm58']              = False     # (--); ; 
SubDyn['SSqm59']              = False     # (--); ; 
SubDyn['SSqm60']              = False     # (--); ; 
SubDyn['SSqm61']              = False     # (--); ; 
SubDyn['SSqm62']              = False     # (--); ; 
SubDyn['SSqm63']              = False     # (--); ; 
SubDyn['SSqm64']              = False     # (--); ; 
SubDyn['SSqm65']              = False     # (--); ; 
SubDyn['SSqm66']              = False     # (--); ; 
SubDyn['SSqm67']              = False     # (--); ; 
SubDyn['SSqm68']              = False     # (--); ; 
SubDyn['SSqm69']              = False     # (--); ; 
SubDyn['SSqm70']              = False     # (--); ; 
SubDyn['SSqm71']              = False     # (--); ; 
SubDyn['SSqm72']              = False     # (--); ; 
SubDyn['SSqm73']              = False     # (--); ; 
SubDyn['SSqm74']              = False     # (--); ; 
SubDyn['SSqm75']              = False     # (--); ; 
SubDyn['SSqm76']              = False     # (--); ; 
SubDyn['SSqm77']              = False     # (--); ; 
SubDyn['SSqm78']              = False     # (--); ; 
SubDyn['SSqm79']              = False     # (--); ; 
SubDyn['SSqm80']              = False     # (--); ; 
SubDyn['SSqm81']              = False     # (--); ; 
SubDyn['SSqm82']              = False     # (--); ; 
SubDyn['SSqm83']              = False     # (--); ; 
SubDyn['SSqm84']              = False     # (--); ; 
SubDyn['SSqm85']              = False     # (--); ; 
SubDyn['SSqm86']              = False     # (--); ; 
SubDyn['SSqm87']              = False     # (--); ; 
SubDyn['SSqm88']              = False     # (--); ; 
SubDyn['SSqm89']              = False     # (--); ; 
SubDyn['SSqm90']              = False     # (--); ; 
SubDyn['SSqm91']              = False     # (--); ; 
SubDyn['SSqm92']              = False     # (--); ; 
SubDyn['SSqm93']              = False     # (--); ; 
SubDyn['SSqm94']              = False     # (--); ; 
SubDyn['SSqm95']              = False     # (--); ; 
SubDyn['SSqm96']              = False     # (--); ; 
SubDyn['SSqm97']              = False     # (--); ; 
SubDyn['SSqm98']              = False     # (--); ; 
SubDyn['SSqm99']              = False     # (--); ; 
SubDyn['SSqmd01']             = False     # (1/s); Modal Parameter (01-99) time derivatives; 
SubDyn['SSqmd02']             = False     # (1/s); ; 
SubDyn['SSqmd03']             = False     # (1/s); ; 
SubDyn['SSqmd04']             = False     # (1/s); ; 
SubDyn['SSqmd05']             = False     # (1/s); ; 
SubDyn['SSqmd06']             = False     # (1/s); ; 
SubDyn['SSqmd07']             = False     # (1/s); ; 
SubDyn['SSqmd08']             = False     # (1/s); ; 
SubDyn['SSqmd09']             = False     # (1/s); ; 
SubDyn['SSqmd10']             = False     # (1/s); ; 
SubDyn['SSqmd11']             = False     # (1/s); ; 
SubDyn['SSqmd12']             = False     # (1/s); ; 
SubDyn['SSqmd13']             = False     # (1/s); ; 
SubDyn['SSqmd14']             = False     # (1/s); ; 
SubDyn['SSqmd15']             = False     # (1/s); ; 
SubDyn['SSqmd16']             = False     # (1/s); ; 
SubDyn['SSqmd17']             = False     # (1/s); ; 
SubDyn['SSqmd18']             = False     # (1/s); ; 
SubDyn['SSqmd19']             = False     # (1/s); ; 
SubDyn['SSqmd20']             = False     # (1/s); ; 
SubDyn['SSqmd21']             = False     # (1/s); ; 
SubDyn['SSqmd22']             = False     # (1/s); ; 
SubDyn['SSqmd23']             = False     # (1/s); ; 
SubDyn['SSqmd24']             = False     # (1/s); ; 
SubDyn['SSqmd25']             = False     # (1/s); ; 
SubDyn['SSqmd26']             = False     # (1/s); ; 
SubDyn['SSqmd27']             = False     # (1/s); ; 
SubDyn['SSqmd28']             = False     # (1/s); ; 
SubDyn['SSqmd29']             = False     # (1/s); ; 
SubDyn['SSqmd30']             = False     # (1/s); ; 
SubDyn['SSqmd31']             = False     # (1/s); ; 
SubDyn['SSqmd32']             = False     # (1/s); ; 
SubDyn['SSqmd33']             = False     # (1/s); ; 
SubDyn['SSqmd34']             = False     # (1/s); ; 
SubDyn['SSqmd35']             = False     # (1/s); ; 
SubDyn['SSqmd36']             = False     # (1/s); ; 
SubDyn['SSqmd37']             = False     # (1/s); ; 
SubDyn['SSqmd38']             = False     # (1/s); ; 
SubDyn['SSqmd39']             = False     # (1/s); ; 
SubDyn['SSqmd40']             = False     # (1/s); ; 
SubDyn['SSqmd41']             = False     # (1/s); ; 
SubDyn['SSqmd42']             = False     # (1/s); ; 
SubDyn['SSqmd43']             = False     # (1/s); ; 
SubDyn['SSqmd44']             = False     # (1/s); ; 
SubDyn['SSqmd45']             = False     # (1/s); ; 
SubDyn['SSqmd46']             = False     # (1/s); ; 
SubDyn['SSqmd47']             = False     # (1/s); ; 
SubDyn['SSqmd48']             = False     # (1/s); ; 
SubDyn['SSqmd49']             = False     # (1/s); ; 
SubDyn['SSqmd50']             = False     # (1/s); ; 
SubDyn['SSqmd51']             = False     # (1/s); ; 
SubDyn['SSqmd52']             = False     # (1/s); ; 
SubDyn['SSqmd53']             = False     # (1/s); ; 
SubDyn['SSqmd54']             = False     # (1/s); ; 
SubDyn['SSqmd55']             = False     # (1/s); ; 
SubDyn['SSqmd56']             = False     # (1/s); ; 
SubDyn['SSqmd57']             = False     # (1/s); ; 
SubDyn['SSqmd58']             = False     # (1/s); ; 
SubDyn['SSqmd59']             = False     # (1/s); ; 
SubDyn['SSqmd60']             = False     # (1/s); ; 
SubDyn['SSqmd61']             = False     # (1/s); ; 
SubDyn['SSqmd62']             = False     # (1/s); ; 
SubDyn['SSqmd63']             = False     # (1/s); ; 
SubDyn['SSqmd64']             = False     # (1/s); ; 
SubDyn['SSqmd65']             = False     # (1/s); ; 
SubDyn['SSqmd66']             = False     # (1/s); ; 
SubDyn['SSqmd67']             = False     # (1/s); ; 
SubDyn['SSqmd68']             = False     # (1/s); ; 
SubDyn['SSqmd69']             = False     # (1/s); ; 
SubDyn['SSqmd70']             = False     # (1/s); ; 
SubDyn['SSqmd71']             = False     # (1/s); ; 
SubDyn['SSqmd72']             = False     # (1/s); ; 
SubDyn['SSqmd73']             = False     # (1/s); ; 
SubDyn['SSqmd74']             = False     # (1/s); ; 
SubDyn['SSqmd75']             = False     # (1/s); ; 
SubDyn['SSqmd76']             = False     # (1/s); ; 
SubDyn['SSqmd77']             = False     # (1/s); ; 
SubDyn['SSqmd78']             = False     # (1/s); ; 
SubDyn['SSqmd79']             = False     # (1/s); ; 
SubDyn['SSqmd80']             = False     # (1/s); ; 
SubDyn['SSqmd81']             = False     # (1/s); ; 
SubDyn['SSqmd82']             = False     # (1/s); ; 
SubDyn['SSqmd83']             = False     # (1/s); ; 
SubDyn['SSqmd84']             = False     # (1/s); ; 
SubDyn['SSqmd85']             = False     # (1/s); ; 
SubDyn['SSqmd86']             = False     # (1/s); ; 
SubDyn['SSqmd87']             = False     # (1/s); ; 
SubDyn['SSqmd88']             = False     # (1/s); ; 
SubDyn['SSqmd89']             = False     # (1/s); ; 
SubDyn['SSqmd90']             = False     # (1/s); ; 
SubDyn['SSqmd91']             = False     # (1/s); ; 
SubDyn['SSqmd92']             = False     # (1/s); ; 
SubDyn['SSqmd93']             = False     # (1/s); ; 
SubDyn['SSqmd94']             = False     # (1/s); ; 
SubDyn['SSqmd95']             = False     # (1/s); ; 
SubDyn['SSqmd96']             = False     # (1/s); ; 
SubDyn['SSqmd97']             = False     # (1/s); ; 
SubDyn['SSqmd98']             = False     # (1/s); ; 
SubDyn['SSqmd99']             = False     # (1/s); ; 
SubDyn['SSqmdd01']            = False     # (1/s^2); Modal Parameter (01-99) 2nd time derivatives; 
SubDyn['SSqmdd02']            = False     # (1/s^2); ; 
SubDyn['SSqmdd03']            = False     # (1/s^2); ; 
SubDyn['SSqmdd04']            = False     # (1/s^2); ; 
SubDyn['SSqmdd05']            = False     # (1/s^2); ; 
SubDyn['SSqmdd06']            = False     # (1/s^2); ; 
SubDyn['SSqmdd07']            = False     # (1/s^2); ; 
SubDyn['SSqmdd08']            = False     # (1/s^2); ; 
SubDyn['SSqmdd09']            = False     # (1/s^2); ; 
SubDyn['SSqmdd10']            = False     # (1/s^2); ; 
SubDyn['SSqmdd11']            = False     # (1/s^2); ; 
SubDyn['SSqmdd12']            = False     # (1/s^2); ; 
SubDyn['SSqmdd13']            = False     # (1/s^2); ; 
SubDyn['SSqmdd14']            = False     # (1/s^2); ; 
SubDyn['SSqmdd15']            = False     # (1/s^2); ; 
SubDyn['SSqmdd16']            = False     # (1/s^2); ; 
SubDyn['SSqmdd17']            = False     # (1/s^2); ; 
SubDyn['SSqmdd18']            = False     # (1/s^2); ; 
SubDyn['SSqmdd19']            = False     # (1/s^2); ; 
SubDyn['SSqmdd20']            = False     # (1/s^2); ; 
SubDyn['SSqmdd21']            = False     # (1/s^2); ; 
SubDyn['SSqmdd22']            = False     # (1/s^2); ; 
SubDyn['SSqmdd23']            = False     # (1/s^2); ; 
SubDyn['SSqmdd24']            = False     # (1/s^2); ; 
SubDyn['SSqmdd25']            = False     # (1/s^2); ; 
SubDyn['SSqmdd26']            = False     # (1/s^2); ; 
SubDyn['SSqmdd27']            = False     # (1/s^2); ; 
SubDyn['SSqmdd28']            = False     # (1/s^2); ; 
SubDyn['SSqmdd29']            = False     # (1/s^2); ; 
SubDyn['SSqmdd30']            = False     # (1/s^2); ; 
SubDyn['SSqmdd31']            = False     # (1/s^2); ; 
SubDyn['SSqmdd32']            = False     # (1/s^2); ; 
SubDyn['SSqmdd33']            = False     # (1/s^2); ; 
SubDyn['SSqmdd34']            = False     # (1/s^2); ; 
SubDyn['SSqmdd35']            = False     # (1/s^2); ; 
SubDyn['SSqmdd36']            = False     # (1/s^2); ; 
SubDyn['SSqmdd37']            = False     # (1/s^2); ; 
SubDyn['SSqmdd38']            = False     # (1/s^2); ; 
SubDyn['SSqmdd39']            = False     # (1/s^2); ; 
SubDyn['SSqmdd40']            = False     # (1/s^2); ; 
SubDyn['SSqmdd41']            = False     # (1/s^2); ; 
SubDyn['SSqmdd42']            = False     # (1/s^2); ; 
SubDyn['SSqmdd43']            = False     # (1/s^2); ; 
SubDyn['SSqmdd44']            = False     # (1/s^2); ; 
SubDyn['SSqmdd45']            = False     # (1/s^2); ; 
SubDyn['SSqmdd46']            = False     # (1/s^2); ; 
SubDyn['SSqmdd47']            = False     # (1/s^2); ; 
SubDyn['SSqmdd48']            = False     # (1/s^2); ; 
SubDyn['SSqmdd49']            = False     # (1/s^2); ; 
SubDyn['SSqmdd50']            = False     # (1/s^2); ; 
SubDyn['SSqmdd51']            = False     # (1/s^2); ; 
SubDyn['SSqmdd52']            = False     # (1/s^2); ; 
SubDyn['SSqmdd53']            = False     # (1/s^2); ; 
SubDyn['SSqmdd54']            = False     # (1/s^2); ; 
SubDyn['SSqmdd55']            = False     # (1/s^2); ; 
SubDyn['SSqmdd56']            = False     # (1/s^2); ; 
SubDyn['SSqmdd57']            = False     # (1/s^2); ; 
SubDyn['SSqmdd58']            = False     # (1/s^2); ; 
SubDyn['SSqmdd59']            = False     # (1/s^2); ; 
SubDyn['SSqmdd60']            = False     # (1/s^2); ; 
SubDyn['SSqmdd61']            = False     # (1/s^2); ; 
SubDyn['SSqmdd62']            = False     # (1/s^2); ; 
SubDyn['SSqmdd63']            = False     # (1/s^2); ; 
SubDyn['SSqmdd64']            = False     # (1/s^2); ; 
SubDyn['SSqmdd65']            = False     # (1/s^2); ; 
SubDyn['SSqmdd66']            = False     # (1/s^2); ; 
SubDyn['SSqmdd67']            = False     # (1/s^2); ; 
SubDyn['SSqmdd68']            = False     # (1/s^2); ; 
SubDyn['SSqmdd69']            = False     # (1/s^2); ; 
SubDyn['SSqmdd70']            = False     # (1/s^2); ; 
SubDyn['SSqmdd71']            = False     # (1/s^2); ; 
SubDyn['SSqmdd72']            = False     # (1/s^2); ; 
SubDyn['SSqmdd73']            = False     # (1/s^2); ; 
SubDyn['SSqmdd74']            = False     # (1/s^2); ; 
SubDyn['SSqmdd75']            = False     # (1/s^2); ; 
SubDyn['SSqmdd76']            = False     # (1/s^2); ; 
SubDyn['SSqmdd77']            = False     # (1/s^2); ; 
SubDyn['SSqmdd78']            = False     # (1/s^2); ; 
SubDyn['SSqmdd79']            = False     # (1/s^2); ; 
SubDyn['SSqmdd80']            = False     # (1/s^2); ; 
SubDyn['SSqmdd81']            = False     # (1/s^2); ; 
SubDyn['SSqmdd82']            = False     # (1/s^2); ; 
SubDyn['SSqmdd83']            = False     # (1/s^2); ; 
SubDyn['SSqmdd84']            = False     # (1/s^2); ; 
SubDyn['SSqmdd85']            = False     # (1/s^2); ; 
SubDyn['SSqmdd86']            = False     # (1/s^2); ; 
SubDyn['SSqmdd87']            = False     # (1/s^2); ; 
SubDyn['SSqmdd88']            = False     # (1/s^2); ; 
SubDyn['SSqmdd89']            = False     # (1/s^2); ; 
SubDyn['SSqmdd90']            = False     # (1/s^2); ; 
SubDyn['SSqmdd91']            = False     # (1/s^2); ; 
SubDyn['SSqmdd92']            = False     # (1/s^2); ; 
SubDyn['SSqmdd93']            = False     # (1/s^2); ; 
SubDyn['SSqmdd94']            = False     # (1/s^2); ; 
SubDyn['SSqmdd95']            = False     # (1/s^2); ; 
SubDyn['SSqmdd96']            = False     # (1/s^2); ; 
SubDyn['SSqmdd97']            = False     # (1/s^2); ; 
SubDyn['SSqmdd98']            = False     # (1/s^2); ; 
SubDyn['SSqmdd99']            = False     # (1/s^2); ; 

""" MoorDyn """
# THIS IS NOT A COMPLETE LIST!
# the "flexible naming system" discussed on page 7-8 of the documentation is not included
# http://www.matt-hall.ca/files/MoorDyn-Users-Guide-2017-08-16.pdf

# also assuming that like other OpenFAST variables, it is limited to 9 output locations per veriable, i.e. FairTen1-FairTen9

MoorDyn = {}
MoorDyn['FairTen1']            = False     # (); ; 
MoorDyn['FairTen2']            = False     # (); ; 
MoorDyn['FairTen3']            = False     # (); ; 
MoorDyn['FairTen4']            = False     # (); ; 
MoorDyn['FairTen5']            = False     # (); ; 
MoorDyn['FairTen6']            = False     # (); ; 
MoorDyn['FairTen7']            = False     # (); ; 
MoorDyn['FairTen8']            = False     # (); ; 
MoorDyn['FairTen9']            = False     # (); ; 
MoorDyn['AnchTen1']            = False     # (); ; 
MoorDyn['AnchTen2']            = False     # (); ; 
MoorDyn['AnchTen3']            = False     # (); ; 
MoorDyn['AnchTen4']            = False     # (); ; 
MoorDyn['AnchTen5']            = False     # (); ; 
MoorDyn['AnchTen6']            = False     # (); ; 
MoorDyn['AnchTen7']            = False     # (); ; 
MoorDyn['AnchTen8']            = False     # (); ; 
MoorDyn['AnchTen9']            = False     # (); ; 

""" BeamDyn """
BeamDyn = {}

# Reaction Loads
BeamDyn['RootFxr']            = False     # (N); x-component of the root reaction force expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['RootFyr']            = False     # (N); y-component of the root reaction force expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['RootFzr']            = False     # (N); z-component of the root reaction force expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['RootMxr']            = False     # (N-m); x-component of the root reaction moment expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['RootMyr']            = False     # (N-m); y-component of the root reaction moment expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['RootMzr']            = False     # (N-m); z-component of the root reaction moment expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system

# Tip Motions
BeamDyn['TipTDxr']            = False     # (m); Tip translational deflection (relative to the undeflected position) expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['TipTDyr']            = False     # (m); Tip translational deflection (relative to the undeflected position) expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['TipTDzr']            = False     # (m); Tip translational deflection (relative to the undeflected position) expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['TipRDxr']            = False     # (-); Tip angular/rotational deflection Wiener-Milenkovi  parameter (relative to the undeflected orientation) expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['TipRDyr']            = False     # (-); Tip angular/rotational deflection Wiener-Milenkovi  parameter (relative to the undeflected orientation) expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['TipRDzr']            = False     # (-); Tip angular/rotational deflection Wiener-Milenkovi  parameter (relative to the undeflected orientation) expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['TipTVXg']            = False     # (m/s); Tip translational velocities (absolute) expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['TipTVYg']            = False     # (m/s); Tip translational velocities (absolute) expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['TipTVZg']            = False     # (m/s); Tip translational velocities (absolute) expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['TipRVXg']            = False     # (deg/s); Tip angular/rotational velocities (absolute) expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['TipRVYg']            = False     # (deg/s); Tip angular/rotational velocities (absolute) expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['TipRVZg']            = False     # (deg/s); Tip angular/rotational velocities (absolute) expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['TipTAXl']            = False     # (m/s^2); Tip translational accelerations (absolute) expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['TipTAYl']            = False     # (m/s^2); Tip translational accelerations (absolute) expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['TipTAZl']            = False     # (m/s^2); Tip translational accelerations (absolute) expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['TipRAXl']            = False     # (deg/s^2); Tip angular/rotational accelerations (absolute) expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['TipRAYl']            = False     # (deg/s^2); Tip angular/rotational accelerations (absolute) expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['TipRAZl']            = False     # (deg/s^2); Tip angular/rotational accelerations (absolute) expressed in l; l: a floating coordinate system local to the deflected beam

# Sectional Loads
BeamDyn['N1Fxl']              = False     # (N); Sectional force resultants at Node 1 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N1Fyl']              = False     # (N); Sectional force resultants at Node 1 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N1Fzl']              = False     # (N); Sectional force resultants at Node 1 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N2Fxl']              = False     # (N); Sectional force resultants at Node 2 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N2Fyl']              = False     # (N); Sectional force resultants at Node 2 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N2Fzl']              = False     # (N); Sectional force resultants at Node 2 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N3Fxl']              = False     # (N); Sectional force resultants at Node 3 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N3Fyl']              = False     # (N); Sectional force resultants at Node 3 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N3Fzl']              = False     # (N); Sectional force resultants at Node 3 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N4Fxl']              = False     # (N); Sectional force resultants at Node 4 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N4Fyl']              = False     # (N); Sectional force resultants at Node 4 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N4Fzl']              = False     # (N); Sectional force resultants at Node 4 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N5Fxl']              = False     # (N); Sectional force resultants at Node 5 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N5Fyl']              = False     # (N); Sectional force resultants at Node 5 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N5Fzl']              = False     # (N); Sectional force resultants at Node 5 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N6Fxl']              = False     # (N); Sectional force resultants at Node 6 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N6Fyl']              = False     # (N); Sectional force resultants at Node 6 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N6Fzl']              = False     # (N); Sectional force resultants at Node 6 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N7Fxl']              = False     # (N); Sectional force resultants at Node 7 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N7Fyl']              = False     # (N); Sectional force resultants at Node 7 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N7Fzl']              = False     # (N); Sectional force resultants at Node 7 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N8Fxl']              = False     # (N); Sectional force resultants at Node 8 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N8Fyl']              = False     # (N); Sectional force resultants at Node 8 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N8Fzl']              = False     # (N); Sectional force resultants at Node 8 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N9Fxl']              = False     # (N); Sectional force resultants at Node 9 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N9Fyl']              = False     # (N); Sectional force resultants at Node 9 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N9Fzl']              = False     # (N); Sectional force resultants at Node 9 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N1Mxl']              = False     # (N-m); Sectional moment resultants at Node 1 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N1Myl']              = False     # (N-m); Sectional moment resultants at Node 1 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N1Mzl']              = False     # (N-m); Sectional moment resultants at Node 1 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N2Mxl']              = False     # (N-m); Sectional moment resultants at Node 2 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N2Myl']              = False     # (N-m); Sectional moment resultants at Node 2 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N2Mzl']              = False     # (N-m); Sectional moment resultants at Node 2 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N3Mxl']              = False     # (N-m); Sectional moment resultants at Node 3 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N3Myl']              = False     # (N-m); Sectional moment resultants at Node 3 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N3Mzl']              = False     # (N-m); Sectional moment resultants at Node 3 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N4Mxl']              = False     # (N-m); Sectional moment resultants at Node 4 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N4Myl']              = False     # (N-m); Sectional moment resultants at Node 4 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N4Mzl']              = False     # (N-m); Sectional moment resultants at Node 4 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N5Mxl']              = False     # (N-m); Sectional moment resultants at Node 5 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N5Myl']              = False     # (N-m); Sectional moment resultants at Node 5 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N5Mzl']              = False     # (N-m); Sectional moment resultants at Node 5 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N6Mxl']              = False     # (N-m); Sectional moment resultants at Node 6 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N6Myl']              = False     # (N-m); Sectional moment resultants at Node 6 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N6Mzl']              = False     # (N-m); Sectional moment resultants at Node 6 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N7Mxl']              = False     # (N-m); Sectional moment resultants at Node 7 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N7Myl']              = False     # (N-m); Sectional moment resultants at Node 7 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N7Mzl']              = False     # (N-m); Sectional moment resultants at Node 7 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N8Mxl']              = False     # (N-m); Sectional moment resultants at Node 8 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N8Myl']              = False     # (N-m); Sectional moment resultants at Node 8 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N8Mzl']              = False     # (N-m); Sectional moment resultants at Node 8 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N9Mxl']              = False     # (N-m); Sectional moment resultants at Node 9 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N9Myl']              = False     # (N-m); Sectional moment resultants at Node 9 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N9Mzl']              = False     # (N-m); Sectional moment resultants at Node 9 expressed in l; l: a floating coordinate system local to the deflected beam

# Sectional Motions
BeamDyn['N1TDxr']             = False     # (m); Sectional translational deflection (relative to the undeflected position) at Node 1 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N1TDyr']             = False     # (m); Sectional translational deflection (relative to the undeflected position) at Node 1 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N1TDzr']             = False     # (m); Sectional translational deflection (relative to the undeflected position) at Node 1 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N2TDxr']             = False     # (m); Sectional translational deflection (relative to the undeflected position) at Node 2 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N2TDyr']             = False     # (m); Sectional translational deflection (relative to the undeflected position) at Node 2 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N2TDzr']             = False     # (m); Sectional translational deflection (relative to the undeflected position) at Node 2 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N3TDxr']             = False     # (m); Sectional translational deflection (relative to the undeflected position) at Node 3 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N3TDyr']             = False     # (m); Sectional translational deflection (relative to the undeflected position) at Node 3 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N3TDzr']             = False     # (m); Sectional translational deflection (relative to the undeflected position) at Node 3 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N4TDxr']             = False     # (m); Sectional translational deflection (relative to the undeflected position) at Node 4 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N4TDyr']             = False     # (m); Sectional translational deflection (relative to the undeflected position) at Node 4 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N4TDzr']             = False     # (m); Sectional translational deflection (relative to the undeflected position) at Node 4 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N5TDxr']             = False     # (m); Sectional translational deflection (relative to the undeflected position) at Node 5 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N5TDyr']             = False     # (m); Sectional translational deflection (relative to the undeflected position) at Node 5 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N5TDzr']             = False     # (m); Sectional translational deflection (relative to the undeflected position) at Node 5 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N6TDxr']             = False     # (m); Sectional translational deflection (relative to the undeflected position) at Node 6 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N6TDyr']             = False     # (m); Sectional translational deflection (relative to the undeflected position) at Node 6 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N6TDzr']             = False     # (m); Sectional translational deflection (relative to the undeflected position) at Node 6 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N7TDxr']             = False     # (m); Sectional translational deflection (relative to the undeflected position) at Node 7 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N7TDyr']             = False     # (m); Sectional translational deflection (relative to the undeflected position) at Node 7 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N7TDzr']             = False     # (m); Sectional translational deflection (relative to the undeflected position) at Node 7 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N8TDxr']             = False     # (m); Sectional translational deflection (relative to the undeflected position) at Node 8 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N8TDyr']             = False     # (m); Sectional translational deflection (relative to the undeflected position) at Node 8 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N8TDzr']             = False     # (m); Sectional translational deflection (relative to the undeflected position) at Node 8 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N9TDxr']             = False     # (m); Sectional translational deflection (relative to the undeflected position) at Node 9 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N9TDyr']             = False     # (m); Sectional translational deflection (relative to the undeflected position) at Node 9 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N9TDzr']             = False     # (m); Sectional translational deflection (relative to the undeflected position) at Node 9 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N1RDxr']             = False     # (-); Sectional angular/rotational deflection Wiener-Milenkovic parameter (relative to the undeflected orientation) at Node 1 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N1RDyr']             = False     # (-); Sectional angular/rotational deflection Wiener-Milenkovic parameter (relative to the undeflected orientation) at Node 1 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N1RDzr']             = False     # (-); Sectional angular/rotational deflection Wiener-Milenkovic parameter (relative to the undeflected orientation) at Node 1 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N2RDxr']             = False     # (-); Sectional angular/rotational deflection Wiener-Milenkovic parameter (relative to the undeflected orientation) at Node 2 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N2RDyr']             = False     # (-); Sectional angular/rotational deflection Wiener-Milenkovic parameter (relative to the undeflected orientation) at Node 2 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N2RDzr']             = False     # (-); Sectional angular/rotational deflection Wiener-Milenkovic parameter (relative to the undeflected orientation) at Node 2 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N3RDxr']             = False     # (-); Sectional angular/rotational deflection Wiener-Milenkovic parameter (relative to the undeflected orientation) at Node 3 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N3RDyr']             = False     # (-); Sectional angular/rotational deflection Wiener-Milenkovic parameter (relative to the undeflected orientation) at Node 3 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N3RDzr']             = False     # (-); Sectional angular/rotational deflection Wiener-Milenkovic parameter (relative to the undeflected orientation) at Node 3 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N4RDxr']             = False     # (-); Sectional angular/rotational deflection Wiener-Milenkovic parameter (relative to the undeflected orientation) at Node 4 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N4RDyr']             = False     # (-); Sectional angular/rotational deflection Wiener-Milenkovic parameter (relative to the undeflected orientation) at Node 4 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N4RDzr']             = False     # (-); Sectional angular/rotational deflection Wiener-Milenkovic parameter (relative to the undeflected orientation) at Node 4 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N5RDxr']             = False     # (-); Sectional angular/rotational deflection Wiener-Milenkovic parameter (relative to the undeflected orientation) at Node 5 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N5RDyr']             = False     # (-); Sectional angular/rotational deflection Wiener-Milenkovic parameter (relative to the undeflected orientation) at Node 5 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N5RDzr']             = False     # (-); Sectional angular/rotational deflection Wiener-Milenkovic parameter (relative to the undeflected orientation) at Node 5 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N6RDxr']             = False     # (-); Sectional angular/rotational deflection Wiener-Milenkovic parameter (relative to the undeflected orientation) at Node 6 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N6RDyr']             = False     # (-); Sectional angular/rotational deflection Wiener-Milenkovic parameter (relative to the undeflected orientation) at Node 6 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N6RDzr']             = False     # (-); Sectional angular/rotational deflection Wiener-Milenkovic parameter (relative to the undeflected orientation) at Node 6 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N7RDxr']             = False     # (-); Sectional angular/rotational deflection Wiener-Milenkovic parameter (relative to the undeflected orientation) at Node 7 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N7RDyr']             = False     # (-); Sectional angular/rotational deflection Wiener-Milenkovic parameter (relative to the undeflected orientation) at Node 7 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N7RDzr']             = False     # (-); Sectional angular/rotational deflection Wiener-Milenkovic parameter (relative to the undeflected orientation) at Node 7 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N8RDxr']             = False     # (-); Sectional angular/rotational deflection Wiener-Milenkovic parameter (relative to the undeflected orientation) at Node 8 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N8RDyr']             = False     # (-); Sectional angular/rotational deflection Wiener-Milenkovic parameter (relative to the undeflected orientation) at Node 8 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N8RDzr']             = False     # (-); Sectional angular/rotational deflection Wiener-Milenkovic parameter (relative to the undeflected orientation) at Node 8 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N9RDxr']             = False     # (-); Sectional angular/rotational deflection Wiener-Milenkovic parameter (relative to the undeflected orientation) at Node 9 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N9RDyr']             = False     # (-); Sectional angular/rotational deflection Wiener-Milenkovic parameter (relative to the undeflected orientation) at Node 9 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N9RDzr']             = False     # (-); Sectional angular/rotational deflection Wiener-Milenkovic parameter (relative to the undeflected orientation) at Node 9 expressed in r; r: a floating reference coordinate system fixed to the root of the moving beam; when coupled to FAST for blades, this is equivalent to the IEC blade (b) coordinate system
BeamDyn['N1TVXg']             = False     # (m/s); Sectional translational velocities (absolute) at Node 1 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N1TVYg']             = False     # (m/s); Sectional translational velocities (absolute) at Node 1 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N1TVZg']             = False     # (m/s); Sectional translational velocities (absolute) at Node 1 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N2TVXg']             = False     # (m/s); Sectional translational velocities (absolute) at Node 2 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N2TVYg']             = False     # (m/s); Sectional translational velocities (absolute) at Node 2 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N2TVZg']             = False     # (m/s); Sectional translational velocities (absolute) at Node 2 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N3TVXg']             = False     # (m/s); Sectional translational velocities (absolute) at Node 3 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N3TVYg']             = False     # (m/s); Sectional translational velocities (absolute) at Node 3 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N3TVZg']             = False     # (m/s); Sectional translational velocities (absolute) at Node 3 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N4TVXg']             = False     # (m/s); Sectional translational velocities (absolute) at Node 4 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N4TVYg']             = False     # (m/s); Sectional translational velocities (absolute) at Node 4 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N4TVZg']             = False     # (m/s); Sectional translational velocities (absolute) at Node 4 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N5TVXg']             = False     # (m/s); Sectional translational velocities (absolute) at Node 5 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N5TVYg']             = False     # (m/s); Sectional translational velocities (absolute) at Node 5 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N5TVZg']             = False     # (m/s); Sectional translational velocities (absolute) at Node 5 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N6TVXg']             = False     # (m/s); Sectional translational velocities (absolute) at Node 6 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N6TVYg']             = False     # (m/s); Sectional translational velocities (absolute) at Node 6 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N6TVZg']             = False     # (m/s); Sectional translational velocities (absolute) at Node 6 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N7TVXg']             = False     # (m/s); Sectional translational velocities (absolute) at Node 7 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N7TVYg']             = False     # (m/s); Sectional translational velocities (absolute) at Node 7 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N7TVZg']             = False     # (m/s); Sectional translational velocities (absolute) at Node 7 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N8TVXg']             = False     # (m/s); Sectional translational velocities (absolute) at Node 8 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N8TVYg']             = False     # (m/s); Sectional translational velocities (absolute) at Node 8 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N8TVZg']             = False     # (m/s); Sectional translational velocities (absolute) at Node 8 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N9TVXg']             = False     # (m/s); Sectional translational velocities (absolute) at Node 9 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N9TVYg']             = False     # (m/s); Sectional translational velocities (absolute) at Node 9 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N9TVZg']             = False     # (m/s); Sectional translational velocities (absolute) at Node 9 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N1RVXg']             = False     # (deg/s); Sectional angular/rotational velocities (absolute) at Node 1 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N1RVYg']             = False     # (deg/s); Sectional angular/rotational velocities (absolute) at Node 1 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N1RVZg']             = False     # (deg/s); Sectional angular/rotational velocities (absolute) at Node 1 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N2RVXg']             = False     # (deg/s); Sectional angular/rotational velocities (absolute) at Node 2 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N2RVYg']             = False     # (deg/s); Sectional angular/rotational velocities (absolute) at Node 2 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N2RVZg']             = False     # (deg/s); Sectional angular/rotational velocities (absolute) at Node 2 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N3RVXg']             = False     # (deg/s); Sectional angular/rotational velocities (absolute) at Node 3 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N3RVYg']             = False     # (deg/s); Sectional angular/rotational velocities (absolute) at Node 3 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N3RVZg']             = False     # (deg/s); Sectional angular/rotational velocities (absolute) at Node 3 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N4RVXg']             = False     # (deg/s); Sectional angular/rotational velocities (absolute) at Node 4 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N4RVYg']             = False     # (deg/s); Sectional angular/rotational velocities (absolute) at Node 4 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N4RVZg']             = False     # (deg/s); Sectional angular/rotational velocities (absolute) at Node 4 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N5RVXg']             = False     # (deg/s); Sectional angular/rotational velocities (absolute) at Node 5 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N5RVYg']             = False     # (deg/s); Sectional angular/rotational velocities (absolute) at Node 5 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N5RVZg']             = False     # (deg/s); Sectional angular/rotational velocities (absolute) at Node 5 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N6RVXg']             = False     # (deg/s); Sectional angular/rotational velocities (absolute) at Node 6 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N6RVYg']             = False     # (deg/s); Sectional angular/rotational velocities (absolute) at Node 6 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N6RVZg']             = False     # (deg/s); Sectional angular/rotational velocities (absolute) at Node 6 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N7RVXg']             = False     # (deg/s); Sectional angular/rotational velocities (absolute) at Node 7 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N7RVYg']             = False     # (deg/s); Sectional angular/rotational velocities (absolute) at Node 7 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N7RVZg']             = False     # (deg/s); Sectional angular/rotational velocities (absolute) at Node 7 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N8RVXg']             = False     # (deg/s); Sectional angular/rotational velocities (absolute) at Node 8 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N8RVYg']             = False     # (deg/s); Sectional angular/rotational velocities (absolute) at Node 8 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N8RVZg']             = False     # (deg/s); Sectional angular/rotational velocities (absolute) at Node 8 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N9RVXg']             = False     # (deg/s); Sectional angular/rotational velocities (absolute) at Node 9 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N9RVYg']             = False     # (deg/s); Sectional angular/rotational velocities (absolute) at Node 9 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N9RVZg']             = False     # (deg/s); Sectional angular/rotational velocities (absolute) at Node 9 expressed in g; g: the global inertial frame coordinate system; when coupled to FAST, this is equivalent to FAST s global inertial frame (i) coordinate system
BeamDyn['N1TAXl']             = False     # (m/s^2); Sectional translational accelerations (absolute) at Node 1 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N1TAYl']             = False     # (m/s^2); Sectional translational accelerations (absolute) at Node 1 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N1TAZl']             = False     # (m/s^2); Sectional translational accelerations (absolute) at Node 1 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N2TAXl']             = False     # (m/s^2); Sectional translational accelerations (absolute) at Node 2 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N2TAYl']             = False     # (m/s^2); Sectional translational accelerations (absolute) at Node 2 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N2TAZl']             = False     # (m/s^2); Sectional translational accelerations (absolute) at Node 2 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N3TAXl']             = False     # (m/s^2); Sectional translational accelerations (absolute) at Node 3 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N3TAYl']             = False     # (m/s^2); Sectional translational accelerations (absolute) at Node 3 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N3TAZl']             = False     # (m/s^2); Sectional translational accelerations (absolute) at Node 3 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N4TAXl']             = False     # (m/s^2); Sectional translational accelerations (absolute) at Node 4 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N4TAYl']             = False     # (m/s^2); Sectional translational accelerations (absolute) at Node 4 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N4TAZl']             = False     # (m/s^2); Sectional translational accelerations (absolute) at Node 4 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N5TAXl']             = False     # (m/s^2); Sectional translational accelerations (absolute) at Node 5 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N5TAYl']             = False     # (m/s^2); Sectional translational accelerations (absolute) at Node 5 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N5TAZl']             = False     # (m/s^2); Sectional translational accelerations (absolute) at Node 5 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N6TAXl']             = False     # (m/s^2); Sectional translational accelerations (absolute) at Node 6 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N6TAYl']             = False     # (m/s^2); Sectional translational accelerations (absolute) at Node 6 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N6TAZl']             = False     # (m/s^2); Sectional translational accelerations (absolute) at Node 6 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N7TAXl']             = False     # (m/s^2); Sectional translational accelerations (absolute) at Node 7 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N7TAYl']             = False     # (m/s^2); Sectional translational accelerations (absolute) at Node 7 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N7TAZl']             = False     # (m/s^2); Sectional translational accelerations (absolute) at Node 7 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N8TAXl']             = False     # (m/s^2); Sectional translational accelerations (absolute) at Node 8 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N8TAYl']             = False     # (m/s^2); Sectional translational accelerations (absolute) at Node 8 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N8TAZl']             = False     # (m/s^2); Sectional translational accelerations (absolute) at Node 8 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N9TAXl']             = False     # (m/s^2); Sectional translational accelerations (absolute) at Node 9 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N9TAYl']             = False     # (m/s^2); Sectional translational accelerations (absolute) at Node 9 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N9TAZl']             = False     # (m/s^2); Sectional translational accelerations (absolute) at Node 9 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N1RAXl']             = False     # (deg/s^2); Sectional angular/rotational accelerations (absolute) at Node 1 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N1RAYl']             = False     # (deg/s^2); Sectional angular/rotational accelerations (absolute) at Node 1 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N1RAZl']             = False     # (deg/s^2); Sectional angular/rotational accelerations (absolute) at Node 1 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N2RAXl']             = False     # (deg/s^2); Sectional angular/rotational accelerations (absolute) at Node 2 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N2RAYl']             = False     # (deg/s^2); Sectional angular/rotational accelerations (absolute) at Node 2 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N2RAZl']             = False     # (deg/s^2); Sectional angular/rotational accelerations (absolute) at Node 2 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N3RAXl']             = False     # (deg/s^2); Sectional angular/rotational accelerations (absolute) at Node 3 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N3RAYl']             = False     # (deg/s^2); Sectional angular/rotational accelerations (absolute) at Node 3 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N3RAZl']             = False     # (deg/s^2); Sectional angular/rotational accelerations (absolute) at Node 3 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N4RAXl']             = False     # (deg/s^2); Sectional angular/rotational accelerations (absolute) at Node 4 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N4RAYl']             = False     # (deg/s^2); Sectional angular/rotational accelerations (absolute) at Node 4 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N4RAZl']             = False     # (deg/s^2); Sectional angular/rotational accelerations (absolute) at Node 4 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N5RAXl']             = False     # (deg/s^2); Sectional angular/rotational accelerations (absolute) at Node 5 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N5RAYl']             = False     # (deg/s^2); Sectional angular/rotational accelerations (absolute) at Node 5 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N5RAZl']             = False     # (deg/s^2); Sectional angular/rotational accelerations (absolute) at Node 5 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N6RAXl']             = False     # (deg/s^2); Sectional angular/rotational accelerations (absolute) at Node 6 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N6RAYl']             = False     # (deg/s^2); Sectional angular/rotational accelerations (absolute) at Node 6 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N6RAZl']             = False     # (deg/s^2); Sectional angular/rotational accelerations (absolute) at Node 6 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N7RAXl']             = False     # (deg/s^2); Sectional angular/rotational accelerations (absolute) at Node 7 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N7RAYl']             = False     # (deg/s^2); Sectional angular/rotational accelerations (absolute) at Node 7 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N7RAZl']             = False     # (deg/s^2); Sectional angular/rotational accelerations (absolute) at Node 7 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N8RAXl']             = False     # (deg/s^2); Sectional angular/rotational accelerations (absolute) at Node 8 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N8RAYl']             = False     # (deg/s^2); Sectional angular/rotational accelerations (absolute) at Node 8 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N8RAZl']             = False     # (deg/s^2); Sectional angular/rotational accelerations (absolute) at Node 8 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N9RAXl']             = False     # (deg/s^2); Sectional angular/rotational accelerations (absolute) at Node 9 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N9RAYl']             = False     # (deg/s^2); Sectional angular/rotational accelerations (absolute) at Node 9 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N9RAZl']             = False     # (deg/s^2); Sectional angular/rotational accelerations (absolute) at Node 9 expressed in l; l: a floating coordinate system local to the deflected beam

# Applied Loads
BeamDyn['N1PFxl']             = False     # (N); Applied point forces at Node 1 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N1PFyl']             = False     # (N); Applied point forces at Node 1 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N1PFzl']             = False     # (N); Applied point forces at Node 1 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N2PFxl']             = False     # (N); Applied point forces at Node 2 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N2PFyl']             = False     # (N); Applied point forces at Node 2 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N2PFzl']             = False     # (N); Applied point forces at Node 2 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N3PFxl']             = False     # (N); Applied point forces at Node 3 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N3PFyl']             = False     # (N); Applied point forces at Node 3 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N3PFzl']             = False     # (N); Applied point forces at Node 3 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N4PFxl']             = False     # (N); Applied point forces at Node 4 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N4PFyl']             = False     # (N); Applied point forces at Node 4 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N4PFzl']             = False     # (N); Applied point forces at Node 4 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N5PFxl']             = False     # (N); Applied point forces at Node 5 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N5PFyl']             = False     # (N); Applied point forces at Node 5 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N5PFzl']             = False     # (N); Applied point forces at Node 5 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N6PFxl']             = False     # (N); Applied point forces at Node 6 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N6PFyl']             = False     # (N); Applied point forces at Node 6 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N6PFzl']             = False     # (N); Applied point forces at Node 6 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N7PFxl']             = False     # (N); Applied point forces at Node 7 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N7PFyl']             = False     # (N); Applied point forces at Node 7 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N7PFzl']             = False     # (N); Applied point forces at Node 7 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N8PFxl']             = False     # (N); Applied point forces at Node 8 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N8PFyl']             = False     # (N); Applied point forces at Node 8 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N8PFzl']             = False     # (N); Applied point forces at Node 8 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N9PFxl']             = False     # (N); Applied point forces at Node 9 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N9PFyl']             = False     # (N); Applied point forces at Node 9 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N9PFzl']             = False     # (N); Applied point forces at Node 9 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N1PMxl']             = False     # (N-m); Applied point moments at Node 1 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N1PMyl']             = False     # (N-m); Applied point moments at Node 1 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N1PMzl']             = False     # (N-m); Applied point moments at Node 1 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N2PMxl']             = False     # (N-m); Applied point moments at Node 2 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N2PMyl']             = False     # (N-m); Applied point moments at Node 2 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N2PMzl']             = False     # (N-m); Applied point moments at Node 2 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N3PMxl']             = False     # (N-m); Applied point moments at Node 3 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N3PMyl']             = False     # (N-m); Applied point moments at Node 3 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N3PMzl']             = False     # (N-m); Applied point moments at Node 3 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N4PMxl']             = False     # (N-m); Applied point moments at Node 4 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N4PMyl']             = False     # (N-m); Applied point moments at Node 4 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N4PMzl']             = False     # (N-m); Applied point moments at Node 4 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N5PMxl']             = False     # (N-m); Applied point moments at Node 5 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N5PMyl']             = False     # (N-m); Applied point moments at Node 5 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N5PMzl']             = False     # (N-m); Applied point moments at Node 5 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N6PMxl']             = False     # (N-m); Applied point moments at Node 6 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N6PMyl']             = False     # (N-m); Applied point moments at Node 6 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N6PMzl']             = False     # (N-m); Applied point moments at Node 6 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N7PMxl']             = False     # (N-m); Applied point moments at Node 7 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N7PMyl']             = False     # (N-m); Applied point moments at Node 7 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N7PMzl']             = False     # (N-m); Applied point moments at Node 7 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N8PMxl']             = False     # (N-m); Applied point moments at Node 8 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N8PMyl']             = False     # (N-m); Applied point moments at Node 8 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N8PMzl']             = False     # (N-m); Applied point moments at Node 8 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N9PMxl']             = False     # (N-m); Applied point moments at Node 9 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N9PMyl']             = False     # (N-m); Applied point moments at Node 9 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N9PMzl']             = False     # (N-m); Applied point moments at Node 9 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N1DFxl']             = False     # (N/m); Applied distributed forces at Node 1 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N1DFyl']             = False     # (N/m); Applied distributed forces at Node 1 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N1DFzl']             = False     # (N/m); Applied distributed forces at Node 1 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N2DFxl']             = False     # (N/m); Applied distributed forces at Node 2 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N2DFyl']             = False     # (N/m); Applied distributed forces at Node 2 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N2DFzl']             = False     # (N/m); Applied distributed forces at Node 2 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N3DFxl']             = False     # (N/m); Applied distributed forces at Node 3 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N3DFyl']             = False     # (N/m); Applied distributed forces at Node 3 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N3DFzl']             = False     # (N/m); Applied distributed forces at Node 3 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N4DFxl']             = False     # (N/m); Applied distributed forces at Node 4 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N4DFyl']             = False     # (N/m); Applied distributed forces at Node 4 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N4DFzl']             = False     # (N/m); Applied distributed forces at Node 4 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N5DFxl']             = False     # (N/m); Applied distributed forces at Node 5 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N5DFyl']             = False     # (N/m); Applied distributed forces at Node 5 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N5DFzl']             = False     # (N/m); Applied distributed forces at Node 5 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N6DFxl']             = False     # (N/m); Applied distributed forces at Node 6 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N6DFyl']             = False     # (N/m); Applied distributed forces at Node 6 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N6DFzl']             = False     # (N/m); Applied distributed forces at Node 6 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N7DFxl']             = False     # (N/m); Applied distributed forces at Node 7 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N7DFyl']             = False     # (N/m); Applied distributed forces at Node 7 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N7DFzl']             = False     # (N/m); Applied distributed forces at Node 7 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N8DFxl']             = False     # (N/m); Applied distributed forces at Node 8 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N8DFyl']             = False     # (N/m); Applied distributed forces at Node 8 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N8DFzl']             = False     # (N/m); Applied distributed forces at Node 8 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N9DFxl']             = False     # (N/m); Applied distributed forces at Node 9 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N9DFyl']             = False     # (N/m); Applied distributed forces at Node 9 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N9DFzl']             = False     # (N/m); Applied distributed forces at Node 9 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N1DMxl']             = False     # (N-m/m); Applied distributed moments at Node 1 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N1DMyl']             = False     # (N-m/m); Applied distributed moments at Node 1 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N1DMzl']             = False     # (N-m/m); Applied distributed moments at Node 1 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N2DMxl']             = False     # (N-m/m); Applied distributed moments at Node 2 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N2DMyl']             = False     # (N-m/m); Applied distributed moments at Node 2 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N2DMzl']             = False     # (N-m/m); Applied distributed moments at Node 2 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N3DMxl']             = False     # (N-m/m); Applied distributed moments at Node 3 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N3DMyl']             = False     # (N-m/m); Applied distributed moments at Node 3 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N3DMzl']             = False     # (N-m/m); Applied distributed moments at Node 3 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N4DMxl']             = False     # (N-m/m); Applied distributed moments at Node 4 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N4DMyl']             = False     # (N-m/m); Applied distributed moments at Node 4 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N4DMzl']             = False     # (N-m/m); Applied distributed moments at Node 4 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N5DMxl']             = False     # (N-m/m); Applied distributed moments at Node 5 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N5DMyl']             = False     # (N-m/m); Applied distributed moments at Node 5 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N5DMzl']             = False     # (N-m/m); Applied distributed moments at Node 5 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N6DMxl']             = False     # (N-m/m); Applied distributed moments at Node 6 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N6DMyl']             = False     # (N-m/m); Applied distributed moments at Node 6 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N6DMzl']             = False     # (N-m/m); Applied distributed moments at Node 6 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N7DMxl']             = False     # (N-m/m); Applied distributed moments at Node 7 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N7DMyl']             = False     # (N-m/m); Applied distributed moments at Node 7 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N7DMzl']             = False     # (N-m/m); Applied distributed moments at Node 7 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N8DMxl']             = False     # (N-m/m); Applied distributed moments at Node 8 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N8DMyl']             = False     # (N-m/m); Applied distributed moments at Node 8 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N8DMzl']             = False     # (N-m/m); Applied distributed moments at Node 8 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N9DMxl']             = False     # (N-m/m); Applied distributed moments at Node 9 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N9DMyl']             = False     # (N-m/m); Applied distributed moments at Node 9 expressed in l; l: a floating coordinate system local to the deflected beam
BeamDyn['N9DMzl']             = False     # (N-m/m); Applied distributed moments at Node 9 expressed in l; l: a floating coordinate system local to the deflected beam

# Pitch Actuator
BeamDyn['PAngInp']            = False     # (deg); Pitch angle input; 
BeamDyn['PAngAct']            = False     # (deg); Actual pitch angle ; 
BeamDyn['PRatAct']            = False     # (deg/s); Actual pitch rate; 
BeamDyn['PAccAct']            = False     # (deg/s^2); Actual pitch acceleration; 

""" Final Output Dictionary """
FstOutput = {}
FstOutput['ElastoDyn']        = ElastoDyn
FstOutput['BeamDyn']          = BeamDyn
FstOutput['ServoDyn']         = ServoDyn
FstOutput['AeroDyn']          = AeroDyn
FstOutput['InflowWind']       = InflowWind
FstOutput['WAMIT']            = WAMIT
FstOutput['HydroDyn']         = HydroDyn
FstOutput['Morison']          = Morison
FstOutput['SubDyn']           = SubDyn
FstOutput['BeamDyn']          = BeamDyn
FstOutput['MoorDyn']          = MoorDyn

""" Generated from FAST OutListParameters.xlsx files with AeroelasticSE/src/AeroelasticSE/Util/create_output_vars.py """


""" OutList """
OutList = {}

# Wind Motions
OutList['WindVxi']            = False     # (m/s); Nominally downwind component of the hub-height wind velocity; Directed along the xi-axis                         
OutList['uWind']              = False     # (m/s); Nominally downwind component of the hub-height wind velocity; Directed along the xi-axis                         
OutList['WindVyi']            = False     # (m/s); Cross-wind component of the hub-height wind velocity; Directed along the yi-axis                         
OutList['vWind']              = False     # (m/s); Cross-wind component of the hub-height wind velocity; Directed along the yi-axis                         
OutList['WindVzi']            = False     # (m/s); Vertical component of the hub-height wind velocity; Directed along the zi-axis                         
OutList['wWind']              = False     # (m/s); Vertical component of the hub-height wind velocity; Directed along the zi-axis                         
OutList['TotWindV']           = False     # (m/s); Total hub-height wind velocity magnitude; N/A                                                
OutList['HorWindV']           = False     # (m/s); Horizontal hub-height wind velocity magnitude; In the xi- and yi-plane                            
OutList['HorWndDir']          = False     # (deg); Horizontal hub-height wind direction.  Please note that FAST  uses the opposite sign convention that AeroDyn uses.  Put a "-", "_", "m", or "M" character in front of this variable name if you want to use the AeroDyn convention.; About the zi-axis                                  
OutList['VerWndDir']          = False     # (deg); Vertical hub-height wind direction; About an axis orthogonal to the zi-axis and the HorWindV-vector

# Blade 1 Tip Motions
OutList['TipDxc1']            = False     # (m); Blade 1 out-of-plane tip deflection (relative to the undeflected position); Directed along the xc1-axis
OutList['OoPDefl1']           = False     # (m); Blade 1 out-of-plane tip deflection (relative to the undeflected position); Directed along the xc1-axis
OutList['TipDyc1']            = False     # (m); Blade 1 in-plane tip deflection (relative to the undeflected position); Directed along the yc1-axis
OutList['IPDefl1']            = False     # (m); Blade 1 in-plane tip deflection (relative to the undeflected position); Directed along the yc1-axis
OutList['TipDzc1']            = False     # (m); Blade 1 axial tip deflection (relative to the undeflected position); Directed along the zc1- and zb1-axes
OutList['TipDzb1']            = False     # (m); Blade 1 axial tip deflection (relative to the undeflected position); Directed along the zc1- and zb1-axes
OutList['TipDxb1']            = False     # (m); Blade 1 flapwise tip deflection (relative to the undeflected position); Directed along the xb1-axis
OutList['TipDyb1']            = False     # (m); Blade 1 edgewise tip deflection (relative to the undeflected position); Directed along the yb1-axis
OutList['TipALxb1']           = False     # (m/s^2); Blade 1 local flapwise tip acceleration (absolute); Directed along the local xb1-axis
OutList['TipALyb1']           = False     # (m/s^2); Blade 1 local edgewise tip acceleration (absolute); Directed along the local yb1-axis
OutList['TipALzb1']           = False     # (m/s^2); Blade 1 local axial tip acceleration (absolute); Directed along the local zb1-axis
OutList['TipRDxb1']           = False     # (deg); Blade 1 roll (angular/rotational) tip deflection (relative to the undeflected position). In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the xb1-axis
OutList['RollDefl1']          = False     # (deg); Blade 1 roll (angular/rotational) tip deflection (relative to the undeflected position). In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the xb1-axis
OutList['TipRDyb1']           = False     # (deg); Blade 1 pitch (angular/rotational) tip deflection (relative to the undeflected position). In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the yb1-axis
OutList['PtchDefl1']          = False     # (deg); Blade 1 pitch (angular/rotational) tip deflection (relative to the undeflected position). In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the yb1-axis
OutList['TipRDzc1']           = False     # (deg); Blade 1 torsional tip deflection (relative to the undeflected position). This output will always be zero for FAST simulation results. Use it for examining blade torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.  Please note that this output uses the opposite of the sign convention used for blade pitch angles.; About the zc1- and zb1-axes
OutList['TipRDzb1']           = False     # (deg); Blade 1 torsional tip deflection (relative to the undeflected position). This output will always be zero for FAST simulation results. Use it for examining blade torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.  Please note that this output uses the opposite of the sign convention used for blade pitch angles.; About the zc1- and zb1-axes
OutList['TwstDefl1']          = False     # (deg); Blade 1 torsional tip deflection (relative to the undeflected position). This output will always be zero for FAST simulation results. Use it for examining blade torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.  Please note that this output uses the opposite of the sign convention used for blade pitch angles.; About the zc1- and zb1-axes
OutList['TipClrnc1']          = False     # (m); Blade 1 tip-to-tower clearance estimate. This is computed as the perpendicular distance from the yaw axis to the tip of blade 1 when the blade tip is below the yaw bearing. When the tip of blade 1 is above the yaw bearing, it is computed as the absolute distance from the yaw bearing to the blade tip. Please note that you should reduce this value by the tower radius to obtain the actual tower clearance.; N/A
OutList['TwrClrnc1']          = False     # (m); Blade 1 tip-to-tower clearance estimate. This is computed as the perpendicular distance from the yaw axis to the tip of blade 1 when the blade tip is below the yaw bearing. When the tip of blade 1 is above the yaw bearing, it is computed as the absolute distance from the yaw bearing to the blade tip. Please note that you should reduce this value by the tower radius to obtain the actual tower clearance.; N/A
OutList['Tip2Twr1']           = False     # (m); Blade 1 tip-to-tower clearance estimate. This is computed as the perpendicular distance from the yaw axis to the tip of blade 1 when the blade tip is below the yaw bearing. When the tip of blade 1 is above the yaw bearing, it is computed as the absolute distance from the yaw bearing to the blade tip. Please note that you should reduce this value by the tower radius to obtain the actual tower clearance.; N/A

# Blade 2 Tip Motions
OutList['TipDxc2']            = False     # (m); Blade 2 out-of-plane tip deflection (relative to the pitch axis); Directed along the xc2-axis
OutList['OoPDefl2']           = False     # (m); Blade 2 out-of-plane tip deflection (relative to the pitch axis); Directed along the xc2-axis
OutList['TipDyc2']            = False     # (m); Blade 2 in-plane tip deflection (relative to the pitch axis); Directed along the yc2-axis
OutList['IPDefl2']            = False     # (m); Blade 2 in-plane tip deflection (relative to the pitch axis); Directed along the yc2-axis
OutList['TipDzc2']            = False     # (m); Blade 2 axial tip deflection (relative to the pitch axis); Directed along the zc2- and zb2-axes
OutList['TipDzb2']            = False     # (m); Blade 2 axial tip deflection (relative to the pitch axis); Directed along the zc2- and zb2-axes
OutList['TipDxb2']            = False     # (m); Blade 2 flapwise tip deflection (relative to the pitch axis); Directed along the xb2-axis
OutList['TipDyb2']            = False     # (m); Blade 2 edgewise tip deflection (relative to the pitch axis); Directed along the yb2-axis
OutList['TipALxb2']           = False     # (m/s^2); Blade 2 local flapwise tip acceleration (absolute); Directed along the local xb2-axis
OutList['TipALyb2']           = False     # (m/s^2); Blade 2 local edgewise tip acceleration (absolute); Directed along the local yb2-axis
OutList['TipALzb2']           = False     # (m/s^2); Blade 2 local axial tip acceleration (absolute); Directed along the local zb2-axis
OutList['TipRDxb2']           = False     # (deg); Blade 2 roll (angular/rotational) tip deflection (relative to the undeflected position). In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the xb2-axis
OutList['RollDefl2']          = False     # (deg); Blade 2 roll (angular/rotational) tip deflection (relative to the undeflected position). In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the xb2-axis
OutList['TipRDyb2']           = False     # (deg); Blade 2 pitch (angular/rotational) tip deflection (relative to the undeflected position). In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the yb2-axis
OutList['PtchDefl2']          = False     # (deg); Blade 2 pitch (angular/rotational) tip deflection (relative to the undeflected position). In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the yb2-axis
OutList['TipRDzc2']           = False     # (deg); Blade 2 torsional (angular/rotational) tip deflection (relative to the undeflected position). This output will always be zero for FAST simulation results. Use it for examining blade torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.  Please note that this output uses the opposite of the sign convention used for blade pitch angles.; About the zc2- and zb2-axes
OutList['TipRDzb2']           = False     # (deg); Blade 2 torsional (angular/rotational) tip deflection (relative to the undeflected position). This output will always be zero for FAST simulation results. Use it for examining blade torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.  Please note that this output uses the opposite of the sign convention used for blade pitch angles.; About the zc2- and zb2-axes
OutList['TwstDefl2']          = False     # (deg); Blade 2 torsional (angular/rotational) tip deflection (relative to the undeflected position). This output will always be zero for FAST simulation results. Use it for examining blade torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.  Please note that this output uses the opposite of the sign convention used for blade pitch angles.; About the zc2- and zb2-axes
OutList['TipClrnc2']          = False     # (m); Blade 2 tip-to-tower clearance estimate. This is computed as the perpendicular distance from the yaw axis to the tip of blade 1 when the blade tip is below the yaw bearing. When the tip of blade 1 is above the yaw bearing, it is computed as the absolute distance from the yaw bearing to the blade tip. Please note that you should reduce this value by the tower radius to obtain the actual tower clearance.; N/A
OutList['TwrClrnc2']          = False     # (m); Blade 2 tip-to-tower clearance estimate. This is computed as the perpendicular distance from the yaw axis to the tip of blade 1 when the blade tip is below the yaw bearing. When the tip of blade 1 is above the yaw bearing, it is computed as the absolute distance from the yaw bearing to the blade tip. Please note that you should reduce this value by the tower radius to obtain the actual tower clearance.; N/A
OutList['Tip2Twr2']           = False     # (m); Blade 2 tip-to-tower clearance estimate. This is computed as the perpendicular distance from the yaw axis to the tip of blade 1 when the blade tip is below the yaw bearing. When the tip of blade 1 is above the yaw bearing, it is computed as the absolute distance from the yaw bearing to the blade tip. Please note that you should reduce this value by the tower radius to obtain the actual tower clearance.; N/A

# Blade 3 Tip Motions
OutList['TipDxc3']            = False     # (m); Blade 3 out-of-plane tip deflection (relative to the pitch axis); Directed along the xc3-axis
OutList['OoPDefl3']           = False     # (m); Blade 3 out-of-plane tip deflection (relative to the pitch axis); Directed along the xc3-axis
OutList['TipDyc3']            = False     # (m); Blade 3 in-plane tip deflection (relative to the pitch axis); Directed along the yc3-axis
OutList['IPDefl3']            = False     # (m); Blade 3 in-plane tip deflection (relative to the pitch axis); Directed along the yc3-axis
OutList['TipDzc3']            = False     # (m); Blade 3 axial tip deflection (relative to the pitch axis); Directed along the zc3- and zb3-axes
OutList['TipDzb3']            = False     # (m); Blade 3 axial tip deflection (relative to the pitch axis); Directed along the zc3- and zb3-axes
OutList['TipDxb3']            = False     # (m); Blade 3 flapwise tip deflection (relative to the pitch axis); Directed along the xb3-axis
OutList['TipDyb3']            = False     # (m); Blade 3 edgewise tip deflection (relative to the pitch axis); Directed along the yb3-axis
OutList['TipALxb3']           = False     # (m/s^2); Blade 3 local flapwise tip acceleration (absolute); Directed along the local xb3-axis
OutList['TipALyb3']           = False     # (m/s^2); Blade 3 local edgewise tip acceleration (absolute); Directed along the local yb3-axis
OutList['TipALzb3']           = False     # (m/s^2); Blade 3 local axial tip acceleration (absolute); Directed along the local zb3-axis
OutList['TipRDxb3']           = False     # (deg); Blade 3 roll (angular/rotational) tip deflection (relative to the undeflected position). In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the xb3-axis
OutList['RollDefl3']          = False     # (deg); Blade 3 roll (angular/rotational) tip deflection (relative to the undeflected position). In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the xb3-axis
OutList['TipRDyb3']           = False     # (deg); Blade 3 pitch (angular/rotational) tip deflection (relative to the undeflected position). In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the yb3-axis
OutList['PtchDefl3']          = False     # (deg); Blade 3 pitch (angular/rotational) tip deflection (relative to the undeflected position). In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the yb3-axis
OutList['TipRDzc3']           = False     # (deg); Blade 3 torsional tip deflection (relative to the undeflected position). This output will always be zero for FAST simulation results. Use it for examining blade torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.  Please note that this output uses the opposite of the sign convention used for blade pitch angles.; About the zc3- and zb3-axes
OutList['TipRDzb3']           = False     # (deg); Blade 3 torsional tip deflection (relative to the undeflected position). This output will always be zero for FAST simulation results. Use it for examining blade torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.  Please note that this output uses the opposite of the sign convention used for blade pitch angles.; About the zc3- and zb3-axes
OutList['TwstDefl3']          = False     # (deg); Blade 3 torsional tip deflection (relative to the undeflected position). This output will always be zero for FAST simulation results. Use it for examining blade torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.  Please note that this output uses the opposite of the sign convention used for blade pitch angles.; About the zc3- and zb3-axes
OutList['TipClrnc3']          = False     # (m); Blade 3 tip-to-tower clearance estimate. This is computed as the perpendicular distance from the yaw axis to the tip of blade 1 when the blade tip is below the yaw bearing. When the tip of blade 1 is above the yaw bearing, it is computed as the absolute distance from the yaw bearing to the blade tip. Please note that you should reduce this value by the tower radius to obtain the actual tower clearance.; N/A
OutList['TwrClrnc3']          = False     # (m); Blade 3 tip-to-tower clearance estimate. This is computed as the perpendicular distance from the yaw axis to the tip of blade 1 when the blade tip is below the yaw bearing. When the tip of blade 1 is above the yaw bearing, it is computed as the absolute distance from the yaw bearing to the blade tip. Please note that you should reduce this value by the tower radius to obtain the actual tower clearance.; N/A
OutList['Tip2Twr3']           = False     # (m); Blade 3 tip-to-tower clearance estimate. This is computed as the perpendicular distance from the yaw axis to the tip of blade 1 when the blade tip is below the yaw bearing. When the tip of blade 1 is above the yaw bearing, it is computed as the absolute distance from the yaw bearing to the blade tip. Please note that you should reduce this value by the tower radius to obtain the actual tower clearance.; N/A

# Blade 1 Local Span Motions
OutList['Spn1ALxb1']          = False     # (m/s^2); Blade 1 local flapwise acceleration (absolute) of span station 1; Directed along the local xb1-axis
OutList['Spn1ALyb1']          = False     # (m/s^2); Blade 1 local edgewise acceleration (absolute) of span station 1; Directed along the local yb1-axis
OutList['Spn1ALzb1']          = False     # (m/s^2); Blade 1 local axial acceleration (absolute) of span station 1; Directed along the local zb1-axis
OutList['Spn2ALxb1']          = False     # (m/s^2); Blade 1 local flapwise acceleration (absolute) of span  station 2; Directed along the local xb1-axis
OutList['Spn2ALyb1']          = False     # (m/s^2); Blade 1 local edgewise acceleration (absolute) of span station 2; Directed along the local yb1-axis
OutList['Spn2ALzb1']          = False     # (m/s^2); Blade 1 local axial acceleration (absolute) of span station 2; Directed along the local zb1-axis
OutList['Spn3ALxb1']          = False     # (m/s^2); Blade 1 local flapwise acceleration (absolute) of span station 3; Directed along the local xb1-axis
OutList['Spn3ALyb1']          = False     # (m/s^2); Blade 1 local edgewise acceleration (absolute) of span station 3; Directed along the local yb1-axis
OutList['Spn3ALzb1']          = False     # (m/s^2); Blade 1 local axial acceleration (absolute) of span station 3; Directed along the local zb1-axis
OutList['Spn4ALxb1']          = False     # (m/s^2); Blade 1 local flapwise acceleration (absolute) of span station 4; Directed along the local xb1-axis
OutList['Spn4ALyb1']          = False     # (m/s^2); Blade 1 local edgewise acceleration (absolute) of span station 4; Directed along the local yb1-axis
OutList['Spn4ALzb1']          = False     # (m/s^2); Blade 1 local axial acceleration (absolute) of span station 4; Directed along the local zb1-axis
OutList['Spn5ALxb1']          = False     # (m/s^2); Blade 1 local flapwise acceleration (absolute) of span station 5; Directed along the local xb1-axis
OutList['Spn5ALyb1']          = False     # (m/s^2); Blade 1 local edgewise acceleration (absolute) of span station 5; Directed along the local yb1-axis
OutList['Spn5ALzb1']          = False     # (m/s^2); Blade 1 local axial acceleration (absolute) of span station 5; Directed along the local zb1-axis
OutList['Spn6ALxb1']          = False     # (m/s^2); Blade 1 local flapwise acceleration (absolute) of span station 6; Directed along the local xb1-axis
OutList['Spn6ALyb1']          = False     # (m/s^2); Blade 1 local edgewise acceleration (absolute) of span station 6; Directed along the local yb1-axis
OutList['Spn6ALzb1']          = False     # (m/s^2); Blade 1 local axial acceleration (absolute) of span station 6; Directed along the local zb1-axis
OutList['Spn7ALxb1']          = False     # (m/s^2); Blade 1 local flapwise acceleration (absolute) of span station 7; Directed along the local xb1-axis
OutList['Spn7ALyb1']          = False     # (m/s^2); Blade 1 local edgewise acceleration (absolute) of span station 7; Directed along the local yb1-axis
OutList['Spn7ALzb1']          = False     # (m/s^2); Blade 1 local axial acceleration (absolute) of span station 7; Directed along the local zb1-axis
OutList['Spn8ALxb1']          = False     # (m/s^2); Blade 1 local flapwise acceleration (absolute) of span station 8; Directed along the local xb1-axis
OutList['Spn8ALyb1']          = False     # (m/s^2); Blade 1 local edgewise acceleration (absolute) of span station 8; Directed along the local yb1-axis
OutList['Spn8ALzb1']          = False     # (m/s^2); Blade 1 local axial acceleration (absolute) of span station 8; Directed along the local zb1-axis
OutList['Spn9ALxb1']          = False     # (m/s^2); Blade 1 local flapwise acceleration (absolute) of span station 9; Directed along the local xb1-axis
OutList['Spn9ALyb1']          = False     # (m/s^2); Blade 1 local edgewise acceleration (absolute) of span station 9; Directed along the local yb1-axis
OutList['Spn9ALzb1']          = False     # (m/s^2); Blade 1 local axial acceleration (absolute) of span station 9; Directed along the local zb1-axis
OutList['Spn1TDxb1']          = False     # (m); Blade 1 local flapwise (translational) deflection (relative to the undeflected position) of span station 1; Directed along the xb1-axis
OutList['Spn1TDyb1']          = False     # (m); Blade 1 local edgewise (translational) deflection (relative to the undeflected position) of span station 1; Directed along the yb1-axis
OutList['Spn1TDzb1']          = False     # (m); Blade 1 local axial (translational) deflection (relative to the undeflected position) of span station 1; Directed along the zb1-axis
OutList['Spn2TDxb1']          = False     # (m); Blade 1 local flapwise (translational) deflection (relative to the undeflected position) of span station 2; Directed along the xb1-axis
OutList['Spn2TDyb1']          = False     # (m); Blade 1 local edgewise (translational) deflection (relative to the undeflected position) of span station 2; Directed along the yb1-axis
OutList['Spn2TDzb1']          = False     # (m); Blade 1 local axial (translational) deflection (relative to the undeflected position) of span station 2; Directed along the zb1-axis
OutList['Spn3TDxb1']          = False     # (m); Blade 1 local flapwise (translational) deflection (relative to the undeflected position) of span station 3; Directed along the xb1-axis
OutList['Spn3TDyb1']          = False     # (m); Blade 1 local edgewise (translational) deflection (relative to the undeflected position) of span station 3; Directed along the yb1-axis
OutList['Spn3TDzb1']          = False     # (m); Blade 1 local axial (translational) deflection (relative to the undeflected position) of span station 3; Directed along the zb1-axis
OutList['Spn4TDxb1']          = False     # (m); Blade 1 local flapwise (translational) deflection (relative to the undeflected position) of span station 4; Directed along the xb1-axis
OutList['Spn4TDyb1']          = False     # (m); Blade 1 local edgewise (translational) deflection (relative to the undeflected position) of span station 4; Directed along the yb1-axis
OutList['Spn4TDzb1']          = False     # (m); Blade 1 local axial (translational) deflection (relative to the undeflected position) of span station 4; Directed along the zb1-axis
OutList['Spn5TDxb1']          = False     # (m); Blade 1 local flapwise (translational) deflection (relative to the undeflected position) of span station 5; Directed along the xb1-axis
OutList['Spn5TDyb1']          = False     # (m); Blade 1 local edgewise (translational) deflection (relative to the undeflected position) of span station 5; Directed along the yb1-axis
OutList['Spn5TDzb1']          = False     # (m); Blade 1 local axial (translational) deflection (relative to the undeflected position) of span station 5; Directed along the zb1-axis
OutList['Spn6TDxb1']          = False     # (m); Blade 1 local flapwise (translational) deflection (relative to the undeflected position) of span station 6; Directed along the xb1-axis
OutList['Spn6TDyb1']          = False     # (m); Blade 1 local edgewise (translational) deflection (relative to the undeflected position) of span station 6; Directed along the yb1-axis
OutList['Spn6TDzb1']          = False     # (m); Blade 1 local axial (translational) deflection (relative to the undeflected position) of span station 6; Directed along the zb1-axis
OutList['Spn7TDxb1']          = False     # (m); Blade 1 local flapwise (translational) deflection (relative to the undeflected position) of span station 7; Directed along the xb1-axis
OutList['Spn7TDyb1']          = False     # (m); Blade 1 local edgewise (translational) deflection (relative to the undeflected position) of span station 7; Directed along the yb1-axis
OutList['Spn7TDzb1']          = False     # (m); Blade 1 local axial (translational) deflection (relative to the undeflected position) of span station 7; Directed along the zb1-axis
OutList['Spn8TDxb1']          = False     # (m); Blade 1 local flapwise (translational) deflection (relative to the undeflected position) of span station 8; Directed along the xb1-axis
OutList['Spn8TDyb1']          = False     # (m); Blade 1 local edgewise (translational) deflection (relative to the undeflected position) of span station 8; Directed along the yb1-axis
OutList['Spn8TDzb1']          = False     # (m); Blade 1 local axial (translational) deflection (relative to the undeflected position) of span station 8; Directed along the zb1-axis
OutList['Spn9TDxb1']          = False     # (m); Blade 1 local flapwise (translational) deflection (relative to the undeflected position) of span station 9; Directed along the xb1-axis
OutList['Spn9TDyb1']          = False     # (m); Blade 1 local edgewise (translational) deflection (relative to the undeflected position) of span station 9; Directed along the yb1-axis
OutList['Spn9TDzb1']          = False     # (m); Blade 1 local axial (translational) deflection (relative to the undeflected position) of span station 9; Directed along the zb1-axis
OutList['Spn1RDxb1']          = False     # (deg); Blade 1 local roll (angular/rotational)  deflection (relative to the undeflected position) of span station 1. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local xb1-axis
OutList['Spn1RDyb1']          = False     # (deg); Blade 1 local pitch (angular/rotational) deflection (relative to the undeflected position) of span station 1. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local yb1-axis
OutList['Spn1RDzb1']          = False     # (deg); Blade 1 local torsional (angular/rotational) deflection (relative to the undeflected position) of span station 1. This output will always be zero for FAST simulation results. Use it for examining blade torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.  Please note that this output uses the opposite of the sign convention used for blade pitch angles.; About the local zb1-axis
OutList['Spn2RDxb1']          = False     # (deg); Blade 1 local roll (angular/rotational)  deflection (relative to the undeflected position) of span station 2. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local xb1-axis
OutList['Spn2RDyb1']          = False     # (deg); Blade 1 local pitch (angular/rotational) deflection (relative to the undeflected position) of span station 2. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local yb1-axis
OutList['Spn2RDzb1']          = False     # (deg); Blade 1 local torsional (angular/rotational) deflection (relative to the undeflected position) of span station 2. This output will always be zero for FAST simulation results. Use it for examining blade torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.  Please note that this output uses the opposite of the sign convention used for blade pitch angles.; About the local zb1-axis
OutList['Spn3RDxb1']          = False     # (deg); Blade 1 local roll (angular/rotational)  deflection (relative to the undeflected position) of span station 3. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local xb1-axis
OutList['Spn3RDyb1']          = False     # (deg); Blade 1 local pitch (angular/rotational) deflection (relative to the undeflected position) of span station 3. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local yb1-axis
OutList['Spn3RDzb1']          = False     # (deg); Blade 1 local torsional (angular/rotational) deflection (relative to the undeflected position) of span station 3. This output will always be zero for FAST simulation results. Use it for examining blade torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.  Please note that this output uses the opposite of the sign convention used for blade pitch angles.; About the local zb1-axis
OutList['Spn4RDxb1']          = False     # (deg); Blade 1 local roll (angular/rotational)  deflection (relative to the undeflected position) of span station 4. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local xb1-axis
OutList['Spn4RDyb1']          = False     # (deg); Blade 1 local pitch (angular/rotational) deflection (relative to the undeflected position) of span station 4. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local yb1-axis
OutList['Spn4RDzb1']          = False     # (deg); Blade 1 local torsional (angular/rotational) deflection (relative to the undeflected position) of span station 4. This output will always be zero for FAST simulation results. Use it for examining blade torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.  Please note that this output uses the opposite of the sign convention used for blade pitch angles.; About the local zb1-axis
OutList['Spn5RDxb1']          = False     # (deg); Blade 1 local roll (angular/rotational)  deflection (relative to the undeflected position) of span station 5. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local xb1-axis
OutList['Spn5RDyb1']          = False     # (deg); Blade 1 local pitch (angular/rotational) deflection (relative to the undeflected position) of span station 5. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local yb1-axis
OutList['Spn5RDzb1']          = False     # (deg); Blade 1 local torsional (angular/rotational) deflection (relative to the undeflected position) of span station 5. This output will always be zero for FAST simulation results. Use it for examining blade torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.  Please note that this output uses the opposite of the sign convention used for blade pitch angles.; About the local zb1-axis
OutList['Spn6RDxb1']          = False     # (deg); Blade 1 local roll (angular/rotational)  deflection (relative to the undeflected position) of span station 6. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local xb1-axis
OutList['Spn6RDyb1']          = False     # (deg); Blade 1 local pitch (angular/rotational) deflection (relative to the undeflected position) of span station 6. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local yb1-axis
OutList['Spn6RDzb1']          = False     # (deg); Blade 1 local torsional (angular/rotational) deflection (relative to the undeflected position) of span station 6. This output will always be zero for FAST simulation results. Use it for examining blade torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.  Please note that this output uses the opposite of the sign convention used for blade pitch angles.; About the local zb1-axis
OutList['Spn7RDxb1']          = False     # (deg); Blade 1 local roll (angular/rotational)  deflection (relative to the undeflected position) of span station 7. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local xb1-axis
OutList['Spn7RDyb1']          = False     # (deg); Blade 1 local pitch (angular/rotational) deflection (relative to the undeflected position) of span station 7. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local yb1-axis
OutList['Spn7RDzb1']          = False     # (deg); Blade 1 local torsional (angular/rotational) deflection (relative to the undeflected position) of span station 7. This output will always be zero for FAST simulation results. Use it for examining blade torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.  Please note that this output uses the opposite of the sign convention used for blade pitch angles.; About the local zb1-axis
OutList['Spn8RDxb1']          = False     # (deg); Blade 1 local roll (angular/rotational)  deflection (relative to the undeflected position) of span station 8. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local xb1-axis
OutList['Spn8RDyb1']          = False     # (deg); Blade 1 local pitch (angular/rotational) deflection (relative to the undeflected position) of span station 8. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local yb1-axis
OutList['Spn8RDzb1']          = False     # (deg); Blade 1 local torsional (angular/rotational) deflection (relative to the undeflected position) of span station 8. This output will always be zero for FAST simulation results. Use it for examining blade torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.  Please note that this output uses the opposite of the sign convention used for blade pitch angles.; About the local zb1-axis
OutList['Spn9RDxb1']          = False     # (deg); Blade 1 local roll (angular/rotational)  deflection (relative to the undeflected position) of span station 9. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local xb1-axis
OutList['Spn9RDyb1']          = False     # (deg); Blade 1 local pitch (angular/rotational) deflection (relative to the undeflected position) of span station 9. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local yb1-axis
OutList['Spn9RDzb1']          = False     # (deg); Blade 1 local torsional (angular/rotational) deflection (relative to the undeflected position) of span station 9. This output will always be zero for FAST simulation results. Use it for examining blade torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.  Please note that this output uses the opposite of the sign convention used for blade pitch angles.; About the local zb1-axis

# Blade 2 Local Span Motions
OutList['Spn1ALxb2']          = False     # (m/s^2); Blade 2 local flapwise acceleration (absolute) of span station 1; Directed along the local xb2-axis
OutList['Spn1ALyb2']          = False     # (m/s^2); Blade 2 local edgewise acceleration (absolute) of span station 1; Directed along the local yb2-axis
OutList['Spn1ALzb2']          = False     # (m/s^2); Blade 2 local axial acceleration (absolute) of span station 1; Directed along the local zb2-axis
OutList['Spn2ALxb2']          = False     # (m/s^2); Blade 2 local flapwise acceleration (absolute) of span station 2; Directed along the local xb2-axis
OutList['Spn2ALyb2']          = False     # (m/s^2); Blade 2 local edgewise acceleration (absolute) of span station 2; Directed along the local yb2-axis
OutList['Spn2ALzb2']          = False     # (m/s^2); Blade 2 local axial acceleration (absolute) of span station 2; Directed along the local zb2-axis
OutList['Spn3ALxb2']          = False     # (m/s^2); Blade 2 local flapwise acceleration (absolute) of span station 3; Directed along the local xb2-axis
OutList['Spn3ALyb2']          = False     # (m/s^2); Blade 2 local edgewise acceleration (absolute) of span station 3; Directed along the local yb2-axis
OutList['Spn3ALzb2']          = False     # (m/s^2); Blade 2 local axial acceleration (absolute) of span station 3; Directed along the local zb2-axis
OutList['Spn4ALxb2']          = False     # (m/s^2); Blade 2 local flapwise acceleration (absolute) of span station 4; Directed along the local xb2-axis
OutList['Spn4ALyb2']          = False     # (m/s^2); Blade 2 local edgewise acceleration (absolute) of span station 4; Directed along the local yb2-axis
OutList['Spn4ALzb2']          = False     # (m/s^2); Blade 2 local axial acceleration (absolute) of span station 4; Directed along the local zb2-axis
OutList['Spn5ALxb2']          = False     # (m/s^2); Blade 2 local flapwise acceleration (absolute) of span station 5; Directed along the local xb2-axis
OutList['Spn5ALyb2']          = False     # (m/s^2); Blade 2 local edgewise acceleration (absolute) of span station 5; Directed along the local yb2-axis
OutList['Spn5ALzb2']          = False     # (m/s^2); Blade 2 local axial acceleration (absolute) of span station 5; Directed along the local zb2-axis
OutList['Spn6ALxb2']          = False     # (m/s^2); Blade 2 local flapwise acceleration (absolute) of span station 6; Directed along the local xb2-axis
OutList['Spn6ALyb2']          = False     # (m/s^2); Blade 2 local edgewise acceleration (absolute) of span station 6; Directed along the local yb2-axis
OutList['Spn6ALzb2']          = False     # (m/s^2); Blade 2 local axial acceleration (absolute) of span station 6; Directed along the local zb2-axis
OutList['Spn7ALxb2']          = False     # (m/s^2); Blade 2 local flapwise acceleration (absolute) of span station 7; Directed along the local xb2-axis
OutList['Spn7ALyb2']          = False     # (m/s^2); Blade 2 local edgewise acceleration (absolute) of span station 7; Directed along the local yb2-axis
OutList['Spn7ALzb2']          = False     # (m/s^2); Blade 2 local axial acceleration (absolute) of span station 7; Directed along the local zb2-axis
OutList['Spn8ALxb2']          = False     # (m/s^2); Blade 2 local flapwise acceleration (absolute) of span station 8; Directed along the local xb2-axis
OutList['Spn8ALyb2']          = False     # (m/s^2); Blade 2 local edgewise acceleration (absolute) of span station 8; Directed along the local yb2-axis
OutList['Spn8ALzb2']          = False     # (m/s^2); Blade 2 local axial acceleration (absolute) of span station 8; Directed along the local zb2-axis
OutList['Spn9ALxb2']          = False     # (m/s^2); Blade 2 local flapwise acceleration (absolute) of span station 9; Directed along the local xb2-axis
OutList['Spn9ALyb2']          = False     # (m/s^2); Blade 2 local edgewise acceleration (absolute) of span station 9; Directed along the local yb2-axis
OutList['Spn9ALzb2']          = False     # (m/s^2); Blade 2 local axial acceleration (absolute) of span station 9; Directed along the local zb2-axis
OutList['Spn1TDxb2']          = False     # (m); Blade 2 local flapwise (translational) deflection (relative to the undeflected position) of span station 1; Directed along the xb2-axis
OutList['Spn1TDyb2']          = False     # (m); Blade 2 local edgewise (translational) deflection (relative to the undeflected position) of span station 1; Directed along the yb2-axis
OutList['Spn1TDzb2']          = False     # (m); Blade 2 local axial (translational) deflection (relative to the undeflected position) of span station 1; Directed along the zb2-axis
OutList['Spn2TDxb2']          = False     # (m); Blade 2 local flapwise (translational) deflection (relative to the undeflected position) of span station 2; Directed along the xb2-axis
OutList['Spn2TDyb2']          = False     # (m); Blade 2 local edgewise (translational) deflection (relative to the undeflected position) of span station 2; Directed along the yb2-axis
OutList['Spn2TDzb2']          = False     # (m); Blade 2 local axial (translational) deflection (relative to the undeflected position) of span station 2; Directed along the zb2-axis
OutList['Spn3TDxb2']          = False     # (m); Blade 2 local flapwise (translational) deflection (relative to the undeflected position) of span station 3; Directed along the xb2-axis
OutList['Spn3TDyb2']          = False     # (m); Blade 2 local edgewise (translational) deflection (relative to the undeflected position) of span station 3; Directed along the yb2-axis
OutList['Spn3TDzb2']          = False     # (m); Blade 2 local axial (translational) deflection (relative to the undeflected position) of span station 3; Directed along the zb2-axis
OutList['Spn4TDxb2']          = False     # (m); Blade 2 local flapwise (translational) deflection (relative to the undeflected position) of span station 4; Directed along the xb2-axis
OutList['Spn4TDyb2']          = False     # (m); Blade 2 local edgewise (translational) deflection (relative to the undeflected position) of span station 4; Directed along the yb2-axis
OutList['Spn4TDzb2']          = False     # (m); Blade 2 local axial (translational) deflection (relative to the undeflected position) of span station 4; Directed along the zb2-axis
OutList['Spn5TDxb2']          = False     # (m); Blade 2 local flapwise (translational) deflection (relative to the undeflected position) of span station 5; Directed along the xb2-axis
OutList['Spn5TDyb2']          = False     # (m); Blade 2 local edgewise (translational) deflection (relative to the undeflected position) of span station 5; Directed along the yb2-axis
OutList['Spn5TDzb2']          = False     # (m); Blade 2 local axial (translational) deflection (relative to the undeflected position) of span station 5; Directed along the zb2-axis
OutList['Spn6TDxb2']          = False     # (m); Blade 2 local flapwise (translational) deflection (relative to the undeflected position) of span station 6; Directed along the xb2-axis
OutList['Spn6TDyb2']          = False     # (m); Blade 2 local edgewise (translational) deflection (relative to the undeflected position) of span station 6; Directed along the yb2-axis
OutList['Spn6TDzb2']          = False     # (m); Blade 2 local axial (translational) deflection (relative to the undeflected position) of span station 6; Directed along the zb2-axis
OutList['Spn7TDxb2']          = False     # (m); Blade 2 local flapwise (translational) deflection (relative to the undeflected position) of span station 7; Directed along the xb2-axis
OutList['Spn7TDyb2']          = False     # (m); Blade 2 local edgewise (translational) deflection (relative to the undeflected position) of span station 7; Directed along the yb2-axis
OutList['Spn7TDzb2']          = False     # (m); Blade 2 local axial (translational) deflection (relative to the undeflected position) of span station 7; Directed along the zb2-axis
OutList['Spn8TDxb2']          = False     # (m); Blade 2 local flapwise (translational) deflection (relative to the undeflected position) of span station 8; Directed along the xb2-axis
OutList['Spn8TDyb2']          = False     # (m); Blade 2 local edgewise (translational) deflection (relative to the undeflected position) of span station 8; Directed along the yb2-axis
OutList['Spn8TDzb2']          = False     # (m); Blade 2 local axial (translational) deflection (relative to the undeflected position) of span station 8; Directed along the zb2-axis
OutList['Spn9TDxb2']          = False     # (m); Blade 2 local flapwise (translational) deflection (relative to the undeflected position) of span station 9; Directed along the xb2-axis
OutList['Spn9TDyb2']          = False     # (m); Blade 2 local edgewise (translational) deflection (relative to the undeflected position) of span station 9; Directed along the yb2-axis
OutList['Spn9TDzb2']          = False     # (m); Blade 2 local axial (translational) deflection (relative to the undeflected position) of span station 9; Directed along the zb2-axis
OutList['Spn1RDxb2']          = False     # (deg); Blade 2 local roll (angular/rotational)  deflection (relative to the undeflected position) of span station 1. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local xb2-axis
OutList['Spn1RDyb2']          = False     # (deg); Blade 2 local pitch (angular/rotational) deflection (relative to the undeflected position) of span station 1. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local yb2-axis
OutList['Spn1RDzb2']          = False     # (deg); Blade 2 local torsional (angular/rotational) deflection (relative to the undeflected position) of span station 1. This output will always be zero for FAST simulation results. Use it for examining blade torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.  Please note that this output uses the opposite of the sign convention used for blade pitch angles.; About the local zb2-axis
OutList['Spn2RDxb2']          = False     # (deg); Blade 2 local roll (angular/rotational)  deflection (relative to the undeflected position) of span station 2. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local xb2-axis
OutList['Spn2RDyb2']          = False     # (deg); Blade 2 local pitch (angular/rotational) deflection (relative to the undeflected position) of span station 2. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local yb2-axis
OutList['Spn2RDzb2']          = False     # (deg); Blade 2 local torsional (angular/rotational) deflection (relative to the undeflected position) of span station 2. This output will always be zero for FAST simulation results. Use it for examining blade torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.  Please note that this output uses the opposite of the sign convention used for blade pitch angles.; About the local zb2-axis
OutList['Spn3RDxb2']          = False     # (deg); Blade 2 local roll (angular/rotational)  deflection (relative to the undeflected position) of span station 3. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local xb2-axis
OutList['Spn3RDyb2']          = False     # (deg); Blade 2 local pitch (angular/rotational) deflection (relative to the undeflected position) of span station 3. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local yb2-axis
OutList['Spn3RDzb2']          = False     # (deg); Blade 2 local torsional (angular/rotational) deflection (relative to the undeflected position) of span station 3. This output will always be zero for FAST simulation results. Use it for examining blade torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.  Please note that this output uses the opposite of the sign convention used for blade pitch angles.; About the local zb2-axis
OutList['Spn4RDxb2']          = False     # (deg); Blade 2 local roll (angular/rotational)  deflection (relative to the undeflected position) of span station 4. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local xb2-axis
OutList['Spn4RDyb2']          = False     # (deg); Blade 2 local pitch (angular/rotational) deflection (relative to the undeflected position) of span station 4. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local yb2-axis
OutList['Spn4RDzb2']          = False     # (deg); Blade 2 local torsional (angular/rotational) deflection (relative to the undeflected position) of span station 4. This output will always be zero for FAST simulation results. Use it for examining blade torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.  Please note that this output uses the opposite of the sign convention used for blade pitch angles.; About the local zb2-axis
OutList['Spn5RDxb2']          = False     # (deg); Blade 2 local roll (angular/rotational)  deflection (relative to the undeflected position) of span station 5. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local xb2-axis
OutList['Spn5RDyb2']          = False     # (deg); Blade 2 local pitch (angular/rotational) deflection (relative to the undeflected position) of span station 5. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local yb2-axis
OutList['Spn5RDzb2']          = False     # (deg); Blade 2 local torsional (angular/rotational) deflection (relative to the undeflected position) of span station 5. This output will always be zero for FAST simulation results. Use it for examining blade torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.  Please note that this output uses the opposite of the sign convention used for blade pitch angles.; About the local zb2-axis
OutList['Spn6RDxb2']          = False     # (deg); Blade 2 local roll (angular/rotational)  deflection (relative to the undeflected position) of span station 6. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local xb2-axis
OutList['Spn6RDyb2']          = False     # (deg); Blade 2 local pitch (angular/rotational) deflection (relative to the undeflected position) of span station 6. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local yb2-axis
OutList['Spn6RDzb2']          = False     # (deg); Blade 2 local torsional (angular/rotational) deflection (relative to the undeflected position) of span station 6. This output will always be zero for FAST simulation results. Use it for examining blade torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.  Please note that this output uses the opposite of the sign convention used for blade pitch angles.; About the local zb2-axis
OutList['Spn7RDxb2']          = False     # (deg); Blade 2 local roll (angular/rotational)  deflection (relative to the undeflected position) of span station 7. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local xb2-axis
OutList['Spn7RDyb2']          = False     # (deg); Blade 2 local pitch (angular/rotational) deflection (relative to the undeflected position) of span station 7. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local yb2-axis
OutList['Spn7RDzb2']          = False     # (deg); Blade 2 local torsional (angular/rotational) deflection (relative to the undeflected position) of span station 7. This output will always be zero for FAST simulation results. Use it for examining blade torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.  Please note that this output uses the opposite of the sign convention used for blade pitch angles.; About the local zb2-axis
OutList['Spn8RDxb2']          = False     # (deg); Blade 2 local roll (angular/rotational)  deflection (relative to the undeflected position) of span station 8. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local xb2-axis
OutList['Spn8RDyb2']          = False     # (deg); Blade 2 local pitch (angular/rotational) deflection (relative to the undeflected position) of span station 8. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local yb2-axis
OutList['Spn8RDzb2']          = False     # (deg); Blade 2 local torsional (angular/rotational) deflection (relative to the undeflected position) of span station 8. This output will always be zero for FAST simulation results. Use it for examining blade torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.  Please note that this output uses the opposite of the sign convention used for blade pitch angles.; About the local zb2-axis
OutList['Spn9RDxb2']          = False     # (deg); Blade 2 local roll (angular/rotational)  deflection (relative to the undeflected position) of span station 9. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local xb2-axis
OutList['Spn9RDyb2']          = False     # (deg); Blade 2 local pitch (angular/rotational) deflection (relative to the undeflected position) of span station 9. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local yb2-axis
OutList['Spn9RDzb2']          = False     # (deg); Blade 2 local torsional (angular/rotational) deflection (relative to the undeflected position) of span station 9. This output will always be zero for FAST simulation results. Use it for examining blade torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.  Please note that this output uses the opposite of the sign convention used for blade pitch angles.; About the local zb2-axis

# Blade 3 Local Span Motions
OutList['Spn1ALxb3']          = False     # (m/s^2); Blade 3 local flapwise acceleration (absolute) of span station 1; Directed along the local xb3-axis
OutList['Spn1ALyb3']          = False     # (m/s^2); Blade 3 local edgewise acceleration (absolute) of span station 1; Directed along the local yb3-axis
OutList['Spn1ALzb3']          = False     # (m/s^2); Blade 3 local axial acceleration (absolute) of span station 1; Directed along the local zb3-axis
OutList['Spn2ALxb3']          = False     # (m/s^2); Blade 3 local flapwise acceleration (absolute) of span station 2; Directed along the local xb3-axis
OutList['Spn2ALyb3']          = False     # (m/s^2); Blade 3 local edgewise acceleration (absolute) of span station 2; Directed along the local yb3-axis
OutList['Spn2ALzb3']          = False     # (m/s^2); Blade 3 local axial acceleration (absolute) of span station 2; Directed along the local zb3-axis
OutList['Spn3ALxb3']          = False     # (m/s^2); Blade 3 local flapwise acceleration (absolute) of span station 3; Directed along the local xb3-axis
OutList['Spn3ALyb3']          = False     # (m/s^2); Blade 3 local edgewise acceleration (absolute) of span station 3; Directed along the local yb3-axis
OutList['Spn3ALzb3']          = False     # (m/s^2); Blade 3 local axial acceleration (absolute) of span station 3; Directed along the local zb3-axis
OutList['Spn4ALxb3']          = False     # (m/s^2); Blade 3 local flapwise acceleration (absolute) of span station 4; Directed along the local xb3-axis
OutList['Spn4ALyb3']          = False     # (m/s^2); Blade 3 local edgewise acceleration (absolute) of span station 4; Directed along the local yb3-axis
OutList['Spn4ALzb3']          = False     # (m/s^2); Blade 3 local axial acceleration (absolute) of span station 4; Directed along the local zb3-axis
OutList['Spn5ALxb3']          = False     # (m/s^2); Blade 3 local flapwise acceleration (absolute) of span station 5; Directed along the local xb3-axis
OutList['Spn5ALyb3']          = False     # (m/s^2); Blade 3 local edgewise acceleration (absolute) of span station 5; Directed along the local yb3-axis
OutList['Spn5ALzb3']          = False     # (m/s^2); Blade 3 local axial acceleration (absolute) of span station 5; Directed along the local zb3-axis
OutList['Spn6ALxb3']          = False     # (m/s^2); Blade 3 local flapwise acceleration (absolute) of span station 6; Directed along the local xb3-axis
OutList['Spn6ALyb3']          = False     # (m/s^2); Blade 3 local edgewise acceleration (absolute) of span station 6; Directed along the local yb3-axis
OutList['Spn6ALzb3']          = False     # (m/s^2); Blade 3 local axial acceleration (absolute) of span station 6; Directed along the local zb3-axis
OutList['Spn7ALxb3']          = False     # (m/s^2); Blade 3 local flapwise acceleration (absolute) of span station 7; Directed along the local xb3-axis
OutList['Spn7ALyb3']          = False     # (m/s^2); Blade 3 local edgewise acceleration (absolute) of span station 7; Directed along the local yb3-axis
OutList['Spn7ALzb3']          = False     # (m/s^2); Blade 3 local axial acceleration (absolute) of span station 7; Directed along the local zb3-axis
OutList['Spn8ALxb3']          = False     # (m/s^2); Blade 3 local flapwise acceleration (absolute) of span station 8; Directed along the local xb3-axis
OutList['Spn8ALyb3']          = False     # (m/s^2); Blade 3 local edgewise acceleration (absolute) of span station 8; Directed along the local yb3-axis
OutList['Spn8ALzb3']          = False     # (m/s^2); Blade 3 local axial acceleration (absolute) of span station 8; Directed along the local zb3-axis
OutList['Spn9ALxb3']          = False     # (m/s^2); Blade 3 local flapwise acceleration (absolute) of span station 9; Directed along the local xb3-axis
OutList['Spn9ALyb3']          = False     # (m/s^2); Blade 3 local edgewise acceleration (absolute) of span station 9; Directed along the local yb3-axis
OutList['Spn9ALzb3']          = False     # (m/s^2); Blade 3 local axial acceleration (absolute) of span station 9; Directed along the local zb3-axis
OutList['Spn1TDxb3']          = False     # (m); Blade 3 local flapwise (translational) deflection (relative to the undeflected position) of span station 1; Directed along the xb3-axis
OutList['Spn1TDyb3']          = False     # (m); Blade 3 local edgewise (translational) deflection (relative to the undeflected position) of span station 1; Directed along the yb3-axis
OutList['Spn1TDzb3']          = False     # (m); Blade 3 local axial (translational) deflection (relative to the undeflected position) of span station 1; Directed along the zb3-axis
OutList['Spn2TDxb3']          = False     # (m); Blade 3 local flapwise (translational) deflection (relative to the undeflected position) of span station 2; Directed along the xb3-axis
OutList['Spn2TDyb3']          = False     # (m); Blade 3 local edgewise (translational) deflection (relative to the undeflected position) of span station 2; Directed along the yb3-axis
OutList['Spn2TDzb3']          = False     # (m); Blade 3 local axial (translational) deflection (relative to the undeflected position) of span station 2; Directed along the zb3-axis
OutList['Spn3TDxb3']          = False     # (m); Blade 3 local flapwise (translational) deflection (relative to the undeflected position) of span station 3; Directed along the xb3-axis
OutList['Spn3TDyb3']          = False     # (m); Blade 3 local edgewise (translational) deflection (relative to the undeflected position) of span station 3; Directed along the yb3-axis
OutList['Spn3TDzb3']          = False     # (m); Blade 3 local axial (translational) deflection (relative to the undeflected position) of span station 3; Directed along the zb3-axis
OutList['Spn4TDxb3']          = False     # (m); Blade 3 local flapwise (translational) deflection (relative to the undeflected position) of span station 4; Directed along the xb3-axis
OutList['Spn4TDyb3']          = False     # (m); Blade 3 local edgewise (translational) deflection (relative to the undeflected position) of span station 4; Directed along the yb3-axis
OutList['Spn4TDzb3']          = False     # (m); Blade 3 local axial (translational) deflection (relative to the undeflected position) of span station 4; Directed along the zb3-axis
OutList['Spn5TDxb3']          = False     # (m); Blade 3 local flapwise (translational) deflection (relative to the undeflected position) of span station 5; Directed along the xb3-axis
OutList['Spn5TDyb3']          = False     # (m); Blade 3 local edgewise (translational) deflection (relative to the undeflected position) of span station 5; Directed along the yb3-axis
OutList['Spn5TDzb3']          = False     # (m); Blade 3 local axial (translational) deflection (relative to the undeflected position) of span station 5; Directed along the zb3-axis
OutList['Spn6TDxb3']          = False     # (m); Blade 3 local flapwise (translational) deflection (relative to the undeflected position) of span station 6; Directed along the xb3-axis
OutList['Spn6TDyb3']          = False     # (m); Blade 3 local edgewise (translational) deflection (relative to the undeflected position) of span station 6; Directed along the yb3-axis
OutList['Spn6TDzb3']          = False     # (m); Blade 3 local axial (translational) deflection (relative to the undeflected position) of span station 6; Directed along the zb3-axis
OutList['Spn7TDxb3']          = False     # (m); Blade 3 local flapwise (translational) deflection (relative to the undeflected position) of span station 7; Directed along the xb3-axis
OutList['Spn7TDyb3']          = False     # (m); Blade 3 local edgewise (translational) deflection (relative to the undeflected position) of span station 7; Directed along the yb3-axis
OutList['Spn7TDzb3']          = False     # (m); Blade 3 local axial (translational) deflection (relative to the undeflected position) of span station 7; Directed along the zb3-axis
OutList['Spn8TDxb3']          = False     # (m); Blade 3 local flapwise (translational) deflection (relative to the undeflected position) of span station 8; Directed along the xb3-axis
OutList['Spn8TDyb3']          = False     # (m); Blade 3 local edgewise (translational) deflection (relative to the undeflected position) of span station 8; Directed along the yb3-axis
OutList['Spn8TDzb3']          = False     # (m); Blade 3 local axial (translational) deflection (relative to the undeflected position) of span station 8; Directed along the zb3-axis
OutList['Spn9TDxb3']          = False     # (m); Blade 3 local flapwise (translational) deflection (relative to the undeflected position) of span station 9; Directed along the xb3-axis
OutList['Spn9TDyb3']          = False     # (m); Blade 3 local edgewise (translational) deflection (relative to the undeflected position) of span station 9; Directed along the yb3-axis
OutList['Spn9TDzb3']          = False     # (m); Blade 3 local axial (translational) deflection (relative to the undeflected position) of span station 9; Directed along the zb3-axis
OutList['Spn1RDxb3']          = False     # (deg); Blade 3 local roll (angular/rotational)  deflection (relative to the undeflected position) of span station 1. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local xb3-axis
OutList['Spn1RDyb3']          = False     # (deg); Blade 3 local pitch (angular/rotational) deflection (relative to the undeflected position) of span station 1. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local yb3-axis
OutList['Spn1RDzb3']          = False     # (deg); Blade 3 local torsional (angular/rotational) deflection (relative to the undeflected position) of span station 1. This output will always be zero for FAST simulation results. Use it for examining blade torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.  Please note that this output uses the opposite of the sign convention used for blade pitch angles.; About the local zb3-axis
OutList['Spn2RDxb3']          = False     # (deg); Blade 3 local roll (angular/rotational)  deflection (relative to the undeflected position) of span station 2. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local xb3-axis
OutList['Spn2RDyb3']          = False     # (deg); Blade 3 local pitch (angular/rotational) deflection (relative to the undeflected position) of span station 2. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local yb3-axis
OutList['Spn2RDzb3']          = False     # (deg); Blade 3 local torsional (angular/rotational) deflection (relative to the undeflected position) of span station 2. This output will always be zero for FAST simulation results. Use it for examining blade torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.  Please note that this output uses the opposite of the sign convention used for blade pitch angles.; About the local zb3-axis
OutList['Spn3RDxb3']          = False     # (deg); Blade 3 local roll (angular/rotational)  deflection (relative to the undeflected position) of span station 3. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local xb3-axis
OutList['Spn3RDyb3']          = False     # (deg); Blade 3 local pitch (angular/rotational) deflection (relative to the undeflected position) of span station 3. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local yb3-axis
OutList['Spn3RDzb3']          = False     # (deg); Blade 3 local torsional (angular/rotational) deflection (relative to the undeflected position) of span station 3. This output will always be zero for FAST simulation results. Use it for examining blade torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.  Please note that this output uses the opposite of the sign convention used for blade pitch angles.; About the local zb3-axis
OutList['Spn4RDxb3']          = False     # (deg); Blade 3 local roll (angular/rotational)  deflection (relative to the undeflected position) of span station 4. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local xb3-axis
OutList['Spn4RDyb3']          = False     # (deg); Blade 3 local pitch (angular/rotational) deflection (relative to the undeflected position) of span station 4. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local yb3-axis
OutList['Spn4RDzb3']          = False     # (deg); Blade 3 local torsional (angular/rotational) deflection (relative to the undeflected position) of span station 4. This output will always be zero for FAST simulation results. Use it for examining blade torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.  Please note that this output uses the opposite of the sign convention used for blade pitch angles.; About the local zb3-axis
OutList['Spn5RDxb3']          = False     # (deg); Blade 3 local roll (angular/rotational)  deflection (relative to the undeflected position) of span station 5. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local xb3-axis
OutList['Spn5RDyb3']          = False     # (deg); Blade 3 local pitch (angular/rotational) deflection (relative to the undeflected position) of span station 5. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local yb3-axis
OutList['Spn5RDzb3']          = False     # (deg); Blade 3 local torsional (angular/rotational) deflection (relative to the undeflected position) of span station 5. This output will always be zero for FAST simulation results. Use it for examining blade torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.  Please note that this output uses the opposite of the sign convention used for blade pitch angles.; About the local zb3-axis
OutList['Spn6RDxb3']          = False     # (deg); Blade 3 local roll (angular/rotational)  deflection (relative to the undeflected position) of span station 6. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local xb3-axis
OutList['Spn6RDyb3']          = False     # (deg); Blade 3 local pitch (angular/rotational) deflection (relative to the undeflected position) of span station 6. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local yb3-axis
OutList['Spn6RDzb3']          = False     # (deg); Blade 3 local torsional (angular/rotational) deflection (relative to the undeflected position) of span station 6. This output will always be zero for FAST simulation results. Use it for examining blade torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.  Please note that this output uses the opposite of the sign convention used for blade pitch angles.; About the local zb3-axis
OutList['Spn7RDxb3']          = False     # (deg); Blade 3 local roll (angular/rotational)  deflection (relative to the undeflected position) of span station 7. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local xb3-axis
OutList['Spn7RDyb3']          = False     # (deg); Blade 3 local pitch (angular/rotational) deflection (relative to the undeflected position) of span station 7. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local yb3-axis
OutList['Spn7RDzb3']          = False     # (deg); Blade 3 local torsional (angular/rotational) deflection (relative to the undeflected position) of span station 7. This output will always be zero for FAST simulation results. Use it for examining blade torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.  Please note that this output uses the opposite of the sign convention used for blade pitch angles.; About the local zb3-axis
OutList['Spn8RDxb3']          = False     # (deg); Blade 3 local roll (angular/rotational)  deflection (relative to the undeflected position) of span station 8. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local xb3-axis
OutList['Spn8RDyb3']          = False     # (deg); Blade 3 local pitch (angular/rotational) deflection (relative to the undeflected position) of span station 8. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local yb3-axis
OutList['Spn8RDzb3']          = False     # (deg); Blade 3 local torsional (angular/rotational) deflection (relative to the undeflected position) of span station 8. This output will always be zero for FAST simulation results. Use it for examining blade torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.  Please note that this output uses the opposite of the sign convention used for blade pitch angles.; About the local zb3-axis
OutList['Spn9RDxb3']          = False     # (deg); Blade 3 local roll (angular/rotational)  deflection (relative to the undeflected position) of span station 9. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local xb3-axis
OutList['Spn9RDyb3']          = False     # (deg); Blade 3 local pitch (angular/rotational) deflection (relative to the undeflected position) of span station 9. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small blade deflections, so that the rotation sequence does not matter.; About the local yb3-axis
OutList['Spn9RDzb3']          = False     # (deg); Blade 3 local torsional (angular/rotational) deflection (relative to the undeflected position) of span station 9. This output will always be zero for FAST simulation results. Use it for examining blade torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.  Please note that this output uses the opposite of the sign convention used for blade pitch angles.; About the local zb3-axis

# Blade Pitch Motions
OutList['PtchPMzc1']          = False     # (deg); Blade 1 pitch angle (position); Positive towards feather about the minus zc1- and minus zb1-axes
OutList['PtchPMzb1']          = False     # (deg); Blade 1 pitch angle (position); Positive towards feather about the minus zc1- and minus zb1-axes
OutList['BldPitch1']          = False     # (deg); Blade 1 pitch angle (position); Positive towards feather about the minus zc1- and minus zb1-axes
OutList['BlPitch1']           = False     # (deg); Blade 1 pitch angle (position); Positive towards feather about the minus zc1- and minus zb1-axes
OutList['PtchPMzc2']          = False     # (deg); Blade 2 pitch angle (position); Positive towards feather about the minus zc2- and minus zb2-axes
OutList['PtchPMzb2']          = False     # (deg); Blade 2 pitch angle (position); Positive towards feather about the minus zc2- and minus zb2-axes
OutList['BldPitch2']          = False     # (deg); Blade 2 pitch angle (position); Positive towards feather about the minus zc2- and minus zb2-axes
OutList['BlPitch2']           = False     # (deg); Blade 2 pitch angle (position); Positive towards feather about the minus zc2- and minus zb2-axes
OutList['PtchPMzc3']          = False     # (deg); Blade 3 pitch angle (position); Positive towards feather about the minus zc3- and minus zb3-axes
OutList['PtchPMzb3']          = False     # (deg); Blade 3 pitch angle (position); Positive towards feather about the minus zc3- and minus zb3-axes
OutList['BldPitch3']          = False     # (deg); Blade 3 pitch angle (position); Positive towards feather about the minus zc3- and minus zb3-axes
OutList['BlPitch3']           = False     # (deg); Blade 3 pitch angle (position); Positive towards feather about the minus zc3- and minus zb3-axes

# Teeter Motions
OutList['TeetPya']            = False     # (deg); Rotor teeter angle (position); About the ya-axis
OutList['RotTeetP']           = False     # (deg); Rotor teeter angle (position); About the ya-axis
OutList['TeetDefl']           = False     # (deg); Rotor teeter angle (position); About the ya-axis
OutList['TeetVya']            = False     # (deg/s); Rotor teeter angular velocity; About the ya-axis
OutList['RotTeetV']           = False     # (deg/s); Rotor teeter angular velocity; About the ya-axis
OutList['TeetAya']            = False     # (deg/s^2); Rotor teeter angular acceleration; About the ya-axis
OutList['RotTeetA']           = False     # (deg/s^2); Rotor teeter angular acceleration; About the ya-axis

# Shaft Motions
OutList['LSSTipPxa']          = False     # (deg); Rotor azimuth angle (position); About the xa- and xs-axes
OutList['LSSTipPxs']          = False     # (deg); Rotor azimuth angle (position); About the xa- and xs-axes
OutList['LSSTipP']            = False     # (deg); Rotor azimuth angle (position); About the xa- and xs-axes
OutList['Azimuth']            = False     # (deg); Rotor azimuth angle (position); About the xa- and xs-axes
OutList['LSSTipVxa']          = False     # (rpm); Rotor azimuth angular speed; About the xa- and xs-axes
OutList['LSSTipVxs']          = False     # (rpm); Rotor azimuth angular speed; About the xa- and xs-axes
OutList['LSSTipV']            = False     # (rpm); Rotor azimuth angular speed; About the xa- and xs-axes
OutList['RotSpeed']           = False     # (rpm); Rotor azimuth angular speed; About the xa- and xs-axes
OutList['LSSTipAxa']          = False     # (deg/s^2); Rotor azimuth angular acceleration; About the xa- and xs-axes
OutList['LSSTipAxs']          = False     # (deg/s^2); Rotor azimuth angular acceleration; About the xa- and xs-axes
OutList['LSSTipA']            = False     # (deg/s^2); Rotor azimuth angular acceleration; About the xa- and xs-axes
OutList['RotAccel']           = False     # (deg/s^2); Rotor azimuth angular acceleration; About the xa- and xs-axes
OutList['LSSGagPxa']          = False     # (deg); Low-speed shaft strain gage azimuth angle (position) (on the gearbox side of the low-speed shaft); About the xa- and xs-axes
OutList['LSSGagPxs']          = False     # (deg); Low-speed shaft strain gage azimuth angle (position) (on the gearbox side of the low-speed shaft); About the xa- and xs-axes
OutList['LSSGagP']            = False     # (deg); Low-speed shaft strain gage azimuth angle (position) (on the gearbox side of the low-speed shaft); About the xa- and xs-axes
OutList['LSSGagVxa']          = False     # (rpm); Low-speed shaft strain gage angular speed (on the gearbox side of the low-speed shaft); About the xa- and xs-axes
OutList['LSSGagVxs']          = False     # (rpm); Low-speed shaft strain gage angular speed (on the gearbox side of the low-speed shaft); About the xa- and xs-axes
OutList['LSSGagV']            = False     # (rpm); Low-speed shaft strain gage angular speed (on the gearbox side of the low-speed shaft); About the xa- and xs-axes
OutList['LSSGagAxa']          = False     # (deg/s^2); Low-speed shaft strain gage angular acceleration (on the gearbox side of the low-speed shaft); About the xa- and xs-axes
OutList['LSSGagAxs']          = False     # (deg/s^2); Low-speed shaft strain gage angular acceleration (on the gearbox side of the low-speed shaft); About the xa- and xs-axes
OutList['LSSGagA']            = False     # (deg/s^2); Low-speed shaft strain gage angular acceleration (on the gearbox side of the low-speed shaft); About the xa- and xs-axes
OutList['HSShftV']            = False     # (rpm); Angular speed of the high-speed shaft and generator; Same sign as LSSGagVxa / LSSGagVxs / LSSGagV
OutList['GenSpeed']           = False     # (rpm); Angular speed of the high-speed shaft and generator; Same sign as LSSGagVxa / LSSGagVxs / LSSGagV
OutList['HSShftA']            = False     # (deg/s^2); Angular acceleration of the high-speed shaft and generator; Same sign as LSSGagAxa / LSSGagAxs / LSSGagA
OutList['GenAccel']           = False     # (deg/s^2); Angular acceleration of the high-speed shaft and generator; Same sign as LSSGagAxa / LSSGagAxs / LSSGagA
OutList['TipSpdRat']          = False     # (-); Rotor blade tip speed ratio; N/A
OutList['TSR']                = False     # (-); Rotor blade tip speed ratio; N/A

# Nacelle IMU Motions
OutList['NcIMUTVxs']          = False     # (m/s); Nacelle inertial measurement unit translational velocity (absolute); Directed along the xs-axis
OutList['NcIMUTVys']          = False     # (m/s); Nacelle inertial measurement unit translational velocity (absolute); Directed along the ys-axis
OutList['NcIMUTVzs']          = False     # (m/s); Nacelle inertial measurement unit translational velocity (absolute); Directed along the zs-axis
OutList['NcIMUTAxs']          = False     # (m/s^2); Nacelle inertial measurement unit translational acceleration (absolute); Directed along the xs-axis
OutList['NcIMUTAys']          = False     # (m/s^2); Nacelle inertial measurement unit translational acceleration (absolute); Directed along the ys-axis
OutList['NcIMUTAzs']          = False     # (m/s^2); Nacelle inertial measurement unit translational acceleration (absolute); Directed along the zs-axis
OutList['NcIMURVxs']          = False     # (deg/s); Nacelle inertial measurement unit angular (rotational) velocity (absolute); About the xs-axis
OutList['NcIMURVys']          = False     # (deg/s); Nacelle inertial measurement unit angular (rotational) velocity (absolute); About the ys-axis
OutList['NcIMURVzs']          = False     # (deg/s); Nacelle inertial measurement unit angular (rotational) velocity (absolute); About the zs-axis
OutList['NcIMURAxs']          = False     # (deg/s^2); Nacelle inertial measurement unit angular (rotational) acceleration (absolute); About the xs-axis
OutList['NcIMURAys']          = False     # (deg/s^2); Nacelle inertial measurement unit angular (rotational) acceleration (absolute); About the ys-axis
OutList['NcIMURAzs']          = False     # (deg/s^2); Nacelle inertial measurement unit angular (rotational) acceleration (absolute); About the zs-axis

# Rotor-Furl Motions
OutList['RotFurlP']           = False     # (deg); Rotor-furl angle (position); About the rotor-furl axis
OutList['RotFurl']            = False     # (deg); Rotor-furl angle (position); About the rotor-furl axis
OutList['RotFurlV']           = False     # (deg/s); Rotor-furl angular velocity; About the rotor-furl axis
OutList['RotFurlA']           = False     # (deg/s^2); Rotor-furl angular acceleration; About the rotor-furl axis

# Tail-Furl Motions
OutList['TailFurlP']          = False     # (deg); Tail-furl angle (position); About the tail-furl axis
OutList['TailFurl']           = False     # (deg); Tail-furl angle (position); About the tail-furl axis
OutList['TailFurlV']          = False     # (deg/s); Tail-furl angular velocity; About the tail-furl axis
OutList['TailFurlA']          = False     # (deg/s^2); Tail-furl angular acceleration; About the tail-furl axis

# Nacelle Yaw Motions
OutList['YawPzn']             = False     # (deg); Nacelle yaw angle (position); About the zn- and zp-axes
OutList['YawPzp']             = False     # (deg); Nacelle yaw angle (position); About the zn- and zp-axes
OutList['NacYawP']            = False     # (deg); Nacelle yaw angle (position); About the zn- and zp-axes
OutList['NacYaw']             = False     # (deg); Nacelle yaw angle (position); About the zn- and zp-axes
OutList['YawPos']             = False     # (deg); Nacelle yaw angle (position); About the zn- and zp-axes
OutList['YawVzn']             = False     # (deg/s); Nacelle yaw angular velocity; About the zn- and zp-axes
OutList['YawVzp']             = False     # (deg/s); Nacelle yaw angular velocity; About the zn- and zp-axes
OutList['NacYawV']            = False     # (deg/s); Nacelle yaw angular velocity; About the zn- and zp-axes
OutList['YawRate']            = False     # (deg/s); Nacelle yaw angular velocity; About the zn- and zp-axes
OutList['YawAzn']             = False     # (deg/s^2); Nacelle yaw angular acceleration; About the zn- and zp-axes
OutList['YawAzp']             = False     # (deg/s^2); Nacelle yaw angular acceleration; About the zn- and zp-axes
OutList['NacYawA']            = False     # (deg/s^2); Nacelle yaw angular acceleration; About the zn- and zp-axes
OutList['YawAccel']           = False     # (deg/s^2); Nacelle yaw angular acceleration; About the zn- and zp-axes
OutList['NacYawErr']          = False     # (deg); Nacelle yaw error estimate. This is computed as follows: NacYawErr = HorWndDir - YawPzn - YawBrRDzt - PtfmRDzi. This estimate is not accurate instantaneously in the presence of significant tower deflection or platform angular (rotational) displacement since the angles used in the computation are not all defined about the same axis of rotation. However, the estimate should be useful in a yaw controller if averaged over a time scale long enough to diminish the effects of tower and platform motions (i.e., much longer than the period of oscillation).; About the zi-axis

# Tower-Top / Yaw Bearing Motions
OutList['YawBrTDxp']          = False     # (m); Tower-top / yaw bearing fore-aft (translational) deflection (relative to the undeflected position); Directed along the xp-axis
OutList['YawBrTDyp']          = False     # (m); Tower-top / yaw bearing side-to-side (translational) deflection (relative to the undeflected position); Directed along the yp-axis
OutList['YawBrTDzp']          = False     # (m); Tower-top / yaw bearing axial (translational) deflection (relative to the undeflected position); Directed along the zp-axis
OutList['YawBrTDxt']          = False     # (m); Tower-top / yaw bearing fore-aft (translational) deflection (relative to the undeflected position); Directed along the xt-axis
OutList['TTDspFA']            = False     # (m); Tower-top / yaw bearing fore-aft (translational) deflection (relative to the undeflected position); Directed along the xt-axis
OutList['YawBrTDyt']          = False     # (m); Tower-top / yaw bearing side-to-side (translation) deflection (relative to the undeflected position); Directed along the yt-axis
OutList['TTDspSS']            = False     # (m); Tower-top / yaw bearing side-to-side (translation) deflection (relative to the undeflected position); Directed along the yt-axis
OutList['YawBrTDzt']          = False     # (m); Tower-top / yaw bearing axial (translational) deflection (relative to the undeflected position); Directed along the zt-axis
OutList['TTDspAx']            = False     # (m); Tower-top / yaw bearing axial (translational) deflection (relative to the undeflected position); Directed along the zt-axis
OutList['YawBrTAxp']          = False     # (m/s^2); Tower-top / yaw bearing fore-aft (translational) acceleration (absolute); Directed along the xp-axis
OutList['YawBrTAyp']          = False     # (m/s^2); Tower-top / yaw bearing side-to-side (translational) acceleration (absolute); Directed along the yp-axis
OutList['YawBrTAzp']          = False     # (m/s^2); Tower-top / yaw bearing axial (translational) acceleration (absolute); Directed along the zp-axis
OutList['YawBrRDxt']          = False     # (deg); Tower-top / yaw bearing angular (rotational) roll deflection (relative to the undeflected position). In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower deflections, so that the rotation sequence does not matter.; About the xt-axis
OutList['TTDspRoll']          = False     # (deg); Tower-top / yaw bearing angular (rotational) roll deflection (relative to the undeflected position). In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower deflections, so that the rotation sequence does not matter.; About the xt-axis
OutList['YawBrRDyt']          = False     # (deg); Tower-top / yaw bearing angular (rotational) pitch deflection (relative to the undeflected position). In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower deflections, so that the rotation sequence does not matter.; About the yt-axis
OutList['TTDspPtch']          = False     # (deg); Tower-top / yaw bearing angular (rotational) pitch deflection (relative to the undeflected position). In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower deflections, so that the rotation sequence does not matter.; About the yt-axis
OutList['YawBrRDzt']          = False     # (deg); Tower-top / yaw bearing angular (rotational) torsion deflection (relative to the undeflected position). This output will always be zero for FAST simulation results. Use it for examining tower torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.; About the zt-axis
OutList['TTDspTwst']          = False     # (deg); Tower-top / yaw bearing angular (rotational) torsion deflection (relative to the undeflected position). This output will always be zero for FAST simulation results. Use it for examining tower torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.; About the zt-axis
OutList['YawBrRVxp']          = False     # (deg/s); Tower-top / yaw bearing angular (rotational) roll velocity (absolute); About the xp-axis
OutList['YawBrRVyp']          = False     # (deg/s); Tower-top / yaw bearing angular (rotational) pitch velocity (absolute); About the yp-axis
OutList['YawBrRVzp']          = False     # (deg/s); Tower-top / yaw bearing angular (rotational) torsion velocity. This output will always be very close to zero for FAST simulation results. Use it for examining tower torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. (absolute); About the zp-axis
OutList['YawBrRAxp']          = False     # (deg/s^2); Tower-top / yaw bearing angular (rotational) roll acceleration (absolute); About the xp-axis
OutList['YawBrRAyp']          = False     # (deg/s^2); Tower-top / yaw bearing angular (rotational) pitch acceleration (absolute); About the yp-axis
OutList['YawBrRAzp']          = False     # (deg/s^2); Tower-top / yaw bearing angular (rotational) torsion acceleration. This output will always be very close to zero for FAST simulation results. Use it for examining tower torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. (absolute); About the zp-axis

# Local Tower Motions
OutList['TwHt1ALxt']          = False     # (m/s^2); Local tower fore-aft (translational) acceleration (absolute) of tower gage 1 ; Directed along the local xt-axis
OutList['TwHt1ALyt']          = False     # (m/s^2); Local tower side-to-side (translational) acceleration (absolute) of tower gage 1 ; Directed along the local yt-axis
OutList['TwHt1ALzt']          = False     # (m/s^2); Local tower axial (translational) acceleration (absolute) of tower gage 1 ; Directed along the local zt-axis
OutList['TwHt2ALxt']          = False     # (m/s^2); Local tower fore-aft (translational) acceleration (absolute) of tower gage 2; Directed along the local xt-axis
OutList['TwHt2ALyt']          = False     # (m/s^2); Local tower side-to-side (translational) acceleration (absolute) of tower gage 2; Directed along the local yt-axis
OutList['TwHt2ALzt']          = False     # (m/s^2); Local tower axial (translational) acceleration (absolute) of tower gage 2; Directed along the local zt-axis
OutList['TwHt3ALxt']          = False     # (m/s^2); Local tower fore-aft (translational) acceleration (absolute) of tower gage 3; Directed along the local xt-axis
OutList['TwHt3ALyt']          = False     # (m/s^2); Local tower side-to-side (translational) acceleration (absolute) of tower gage 3; Directed along the local yt-axis
OutList['TwHt3ALzt']          = False     # (m/s^2); Local tower axial (translational) acceleration (absolute) of tower gage 3; Directed along the local zt-axis
OutList['TwHt4ALxt']          = False     # (m/s^2); Local tower fore-aft (translational) acceleration (absolute) of tower gage 4; Directed along the local xt-axis
OutList['TwHt4ALyt']          = False     # (m/s^2); Local tower side-to-side (translational) acceleration (absolute) of tower gage 4; Directed along the local yt-axis
OutList['TwHt4ALzt']          = False     # (m/s^2); Local tower axial (translational) acceleration (absolute) of tower gage 4; Directed along the local zt-axis
OutList['TwHt5ALxt']          = False     # (m/s^2); Local tower fore-aft (translational) acceleration (absolute) of tower gage 5; Directed along the local xt-axis
OutList['TwHt5ALyt']          = False     # (m/s^2); Local tower side-to-side (translational) acceleration (absolute) of tower gage 5; Directed along the local yt-axis
OutList['TwHt5ALzt']          = False     # (m/s^2); Local tower axial (translational) acceleration (absolute) of tower gage 5; Directed along the local zt-axis
OutList['TwHt6ALxt']          = False     # (m/s^2); Local tower fore-aft (translational) acceleration (absolute) of tower gage 6; Directed along the local xt-axis
OutList['TwHt6ALyt']          = False     # (m/s^2); Local tower side-to-side (translational) acceleration (absolute) of tower gage 6; Directed along the local yt-axis
OutList['TwHt6ALzt']          = False     # (m/s^2); Local tower axial (translational) acceleration (absolute) of tower gage 6; Directed along the local zt-axis
OutList['TwHt7ALxt']          = False     # (m/s^2); Local tower fore-aft (translational) acceleration (absolute) of tower gage 7; Directed along the local xt-axis
OutList['TwHt7ALyt']          = False     # (m/s^2); Local tower side-to-side (translational) acceleration (absolute) of tower gage 7; Directed along the local yt-axis
OutList['TwHt7ALzt']          = False     # (m/s^2); Local tower axial (translational) acceleration (absolute) of tower gage 7; Directed along the local zt-axis
OutList['TwHt8ALxt']          = False     # (m/s^2); Local tower fore-aft (translational) acceleration (absolute) of tower gage 8; Directed along the local xt-axis
OutList['TwHt8ALyt']          = False     # (m/s^2); Local tower side-to-side (translational) acceleration (absolute) of tower gage 8; Directed along the local yt-axis
OutList['TwHt8ALzt']          = False     # (m/s^2); Local tower axial (translational) acceleration (absolute) of tower gage 8; Directed along the local zt-axis
OutList['TwHt9ALxt']          = False     # (m/s^2); Local tower fore-aft (translational) acceleration (absolute) of tower gage 9; Directed along the local xt-axis
OutList['TwHt9ALyt']          = False     # (m/s^2); Local tower side-to-side (translational) acceleration (absolute) of tower gage 9; Directed along the local yt-axis
OutList['TwHt9ALzt']          = False     # (m/s^2); Local tower axial (translational) acceleration (absolute) of tower gage 9; Directed along the local zt-axis
OutList['TwHt1TDxt']          = False     # (m); Local tower fore-aft (translational) deflection (relative to the undeflected position) of tower gage 1; Directed along the local xt-axis
OutList['TwHt1TDyt']          = False     # (m); Local tower side-to-side (translational) deflection (relative to the undeflected position) of tower gage 1; Directed along the local yt-axis
OutList['TwHt1TDzt']          = False     # (m); Local tower axial (translational) deflection (relative to the undeflected position) of tower gage 1; Directed along the local zt-axis
OutList['TwHt2TDxt']          = False     # (m); Local tower fore-aft (translational) deflection (relative to the undeflected position) of tower gage 2; Directed along the local xt-axis
OutList['TwHt2TDyt']          = False     # (m); Local tower side-to-side (translational) deflection (relative to the undeflected position) of tower gage 2; Directed along the local yt-axis
OutList['TwHt2TDzt']          = False     # (m); Local tower axial (translational) deflection (relative to the undeflected position) of tower gage 2; Directed along the local zt-axis
OutList['TwHt3TDxt']          = False     # (m); Local tower fore-aft (translational) deflection (relative to the undeflected position) of tower gage 3; Directed along the local xt-axis
OutList['TwHt3TDyt']          = False     # (m); Local tower side-to-side (translational) deflection (relative to the undeflected position) of tower gage 3; Directed along the local yt-axis
OutList['TwHt3TDzt']          = False     # (m); Local tower axial (translational) deflection (relative to the undeflected position) of tower gage 3; Directed along the local zt-axis
OutList['TwHt4TDxt']          = False     # (m); Local tower fore-aft (translational) deflection (relative to the undeflected position) of tower gage 4; Directed along the local xt-axis
OutList['TwHt4TDyt']          = False     # (m); Local tower side-to-side (translational) deflection (relative to the undeflected position) of tower gage 4; Directed along the local yt-axis
OutList['TwHt4TDzt']          = False     # (m); Local tower axial (translational) deflection (relative to the undeflected position) of tower gage 4; Directed along the local zt-axis
OutList['TwHt5TDxt']          = False     # (m); Local tower fore-aft (translational) deflection (relative to the undeflected position) of tower gage 5; Directed along the local xt-axis
OutList['TwHt5TDyt']          = False     # (m); Local tower side-to-side (translational) deflection (relative to the undeflected position) of tower gage 5; Directed along the local yt-axis
OutList['TwHt5TDzt']          = False     # (m); Local tower axial (translational) deflection (relative to the undeflected position) of tower gage 5; Directed along the local zt-axis
OutList['TwHt6TDxt']          = False     # (m); Local tower fore-aft (translational) deflection (relative to the undeflected position) of tower gage 6; Directed along the local xt-axis
OutList['TwHt6TDyt']          = False     # (m); Local tower side-to-side (translational) deflection (relative to the undeflected position) of tower gage 6; Directed along the local yt-axis
OutList['TwHt6TDzt']          = False     # (m); Local tower axial (translational) deflection (relative to the undeflected position) of tower gage 6; Directed along the local zt-axis
OutList['TwHt7TDxt']          = False     # (m); Local tower fore-aft (translational) deflection (relative to the undeflected position) of tower gage 7; Directed along the local xt-axis
OutList['TwHt7TDyt']          = False     # (m); Local tower side-to-side (translational) deflection (relative to the undeflected position) of tower gage 7; Directed along the local yt-axis
OutList['TwHt7TDzt']          = False     # (m); Local tower axial (translational) deflection (relative to the undeflected position) of tower gage 7; Directed along the local zt-axis
OutList['TwHt8TDxt']          = False     # (m); Local tower fore-aft (translational) deflection (relative to the undeflected position) of tower gage 8; Directed along the local xt-axis
OutList['TwHt8TDyt']          = False     # (m); Local tower side-to-side (translational) deflection (relative to the undeflected position) of tower gage 8; Directed along the local yt-axis
OutList['TwHt8TDzt']          = False     # (m); Local tower axial (translational) deflection (relative to the undeflected position) of tower gage 8; Directed along the local zt-axis
OutList['TwHt9TDxt']          = False     # (m); Local tower fore-aft (translational) deflection (relative to the undeflected position) of tower gage 9; Directed along the local xt-axis
OutList['TwHt9TDyt']          = False     # (m); Local tower side-to-side (translational) deflection (relative to the undeflected position) of tower gage 9; Directed along the local yt-axis
OutList['TwHt9TDzt']          = False     # (m); Local tower axial (translational) deflection (relative to the undeflected position) of tower gage 9; Directed along the local zt-axis
OutList['TwHt1RDxt']          = False     # (deg); Local tower angular (rotational) roll deflection (relative to the undeflected position) of tower gage 1. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower deflections, so that the rotation sequence does not matter.; About the local xt-axis
OutList['TwHt1RDyt']          = False     # (deg); Local tower angular (rotational) pitch deflection (relative to the undeflected position) of tower gage 1. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower deflections, so that the rotation sequence does not matter.; About the local yt-axis
OutList['TwHt1RDzt']          = False     # (deg); Local tower angular (rotational) torsion deflection (relative to the undeflected position) of tower gage 1. This output will always be zero for FAST simulation results. Use it for examining tower torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.; About the local zt-axis
OutList['TwHt2RDxt']          = False     # (deg); Local tower angular (rotational) roll deflection (relative to the undeflected position) of tower gage 2. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower deflections, so that the rotation sequence does not matter.; About the local xt-axis
OutList['TwHt2RDyt']          = False     # (deg); Local tower angular (rotational) pitch deflection (relative to the undeflected position) of tower gage 2. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower deflections, so that the rotation sequence does not matter.; About the local yt-axis
OutList['TwHt2RDzt']          = False     # (deg); Local tower angular (rotational) torsion deflection (relative to the undeflected position) of tower gage 2. This output will always be zero for FAST simulation results. Use it for examining tower torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.; About the local zt-axis
OutList['TwHt3RDxt']          = False     # (deg); Local tower angular (rotational) roll deflection (relative to the undeflected position) of tower gage 3. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower deflections, so that the rotation sequence does not matter.; About the local xt-axis
OutList['TwHt3RDyt']          = False     # (deg); Local tower angular (rotational) pitch deflection (relative to the undeflected position) of tower gage 3. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower deflections, so that the rotation sequence does not matter.; About the local yt-axis
OutList['TwHt3RDzt']          = False     # (deg); Local tower angular (rotational) torsion deflection (relative to the undeflected position) of tower gage 3. This output will always be zero for FAST simulation results. Use it for examining tower torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.; About the local zt-axis
OutList['TwHt4RDxt']          = False     # (deg); Local tower angular (rotational) roll deflection (relative to the undeflected position) of tower gage 4. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower deflections, so that the rotation sequence does not matter.; About the local xt-axis
OutList['TwHt4RDyt']          = False     # (deg); Local tower angular (rotational) pitch deflection (relative to the undeflected position) of tower gage 4. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower deflections, so that the rotation sequence does not matter.; About the local yt-axis
OutList['TwHt4RDzt']          = False     # (deg); Local tower angular (rotational) torsion deflection (relative to the undeflected position) of tower gage 4. This output will always be zero for FAST simulation results. Use it for examining tower torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.; About the local zt-axis
OutList['TwHt5RDxt']          = False     # (deg); Local tower angular (rotational) roll deflection (relative to the undeflected position) of tower gage 5. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower deflections, so that the rotation sequence does not matter.; About the local xt-axis
OutList['TwHt5RDyt']          = False     # (deg); Local tower angular (rotational) pitch deflection (relative to the undeflected position) of tower gage 5. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower deflections, so that the rotation sequence does not matter.; About the local yt-axis
OutList['TwHt5RDzt']          = False     # (deg); Local tower angular (rotational) torsion deflection (relative to the undeflected position) of tower gage 5. This output will always be zero for FAST simulation results. Use it for examining tower torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.; About the local zt-axis
OutList['TwHt6RDxt']          = False     # (deg); Local tower angular (rotational) roll deflection (relative to the undeflected position) of tower gage 6. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower deflections, so that the rotation sequence does not matter.; About the local xt-axis
OutList['TwHt6RDyt']          = False     # (deg); Local tower angular (rotational) pitch deflection (relative to the undeflected position) of tower gage 6. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower deflections, so that the rotation sequence does not matter.; About the local yt-axis
OutList['TwHt6RDzt']          = False     # (deg); Local tower angular (rotational) torsion deflection (relative to the undeflected position) of tower gage 6. This output will always be zero for FAST simulation results. Use it for examining tower torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.; About the local zt-axis
OutList['TwHt7RDxt']          = False     # (deg); Local tower angular (rotational) roll deflection (relative to the undeflected position) of tower gage 7. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower deflections, so that the rotation sequence does not matter.; About the local xt-axis
OutList['TwHt7RDyt']          = False     # (deg); Local tower angular (rotational) pitch deflection (relative to the undeflected position) of tower gage 7. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower deflections, so that the rotation sequence does not matter.; About the local yt-axis
OutList['TwHt7RDzt']          = False     # (deg); Local tower angular (rotational) torsion deflection (relative to the undeflected position) of tower gage 7. This output will always be zero for FAST simulation results. Use it for examining tower torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.; About the local zt-axis
OutList['TwHt8RDxt']          = False     # (deg); Local tower angular (rotational) roll deflection (relative to the undeflected position) of tower gage 8. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower deflections, so that the rotation sequence does not matter.; About the local xt-axis
OutList['TwHt8RDyt']          = False     # (deg); Local tower angular (rotational) pitch deflection (relative to the undeflected position) of tower gage 8. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower deflections, so that the rotation sequence does not matter.; About the local yt-axis
OutList['TwHt8RDzt']          = False     # (deg); Local tower angular (rotational) torsion deflection (relative to the undeflected position) of tower gage 8. This output will always be zero for FAST simulation results. Use it for examining tower torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.; About the local zt-axis
OutList['TwHt9RDxt']          = False     # (deg); Local tower angular (rotational) roll deflection (relative to the undeflected position) of tower gage 9. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower deflections, so that the rotation sequence does not matter.; About the local xt-axis
OutList['TwHt9RDyt']          = False     # (deg); Local tower angular (rotational) pitch deflection (relative to the undeflected position) of tower gage 9. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower deflections, so that the rotation sequence does not matter.; About the local yt-axis
OutList['TwHt9RDzt']          = False     # (deg); Local tower angular (rotational) torsion deflection (relative to the undeflected position) of tower gage 9. This output will always be zero for FAST simulation results. Use it for examining tower torsional deflections of ADAMS simulations run using ADAMS datasets created using the FAST-to-ADAMS preprocessor. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence.; About the local zt-axis
OutList['TwHt1TPxi']          = False     # (m); xi-component of the translational position (relative to the inertia frame) of tower gage 1; Directed along the local xi-axis
OutList['TwHt1TPyi']          = False     # (m); yi-component of the translational position (relative to the inertia frame) of tower gage 1; Directed along the local yi-axis
OutList['TwHt1TPzi']          = False     # (m); zi-component of the translational position (relative to ground level [onshore] or MSL [offshore]) of tower gage 1; Directed along the local zi-axis
OutList['TwHt2TPxi']          = False     # (m); xi-component of the translational position (relative to the inertia frame) of tower gage 2; Directed along the local xi-axis
OutList['TwHt2TPyi']          = False     # (m); yi-component of the translational position (relative to the inertia frame) of tower gage 2; Directed along the local yi-axis
OutList['TwHt2TPzi']          = False     # (m); zi-component of the translational position (relative to ground level [onshore] or MSL [offshore]) of tower gage 2; Directed along the local zi-axis
OutList['TwHt3TPxi']          = False     # (m); xi-component of the translational position (relative to the inertia frame) of tower gage 3; Directed along the local xi-axis
OutList['TwHt3TPyi']          = False     # (m); yi-component of the translational position (relative to the inertia frame) of tower gage 3; Directed along the local yi-axis
OutList['TwHt3TPzi']          = False     # (m); zi-component of the translational position (relative to ground level [onshore] or MSL [offshore]) of tower gage 3; Directed along the local zi-axis
OutList['TwHt4TPxi']          = False     # (m); xi-component of the translational position (relative to the inertia frame) of tower gage 4; Directed along the local xi-axis
OutList['TwHt4TPyi']          = False     # (m); yi-component of the translational position (relative to the inertia frame) of tower gage 4; Directed along the local yi-axis
OutList['TwHt4TPzi']          = False     # (m); zi-component of the translational position (relative to ground level [onshore] or MSL [offshore]) of tower gage 4; Directed along the local zi-axis
OutList['TwHt5TPxi']          = False     # (m); xi-component of the translational position (relative to the inertia frame) of tower gage 5; Directed along the local xi-axis
OutList['TwHt5TPyi']          = False     # (m); yi-component of the translational position (relative to the inertia frame) of tower gage 5; Directed along the local yi-axis
OutList['TwHt5TPzi']          = False     # (m); zi-component of the translational position (relative to ground level [onshore] or MSL [offshore]) of tower gage 5; Directed along the local zi-axis
OutList['TwHt6TPxi']          = False     # (m); xi-component of the translational position (relative to the inertia frame) of tower gage 6; Directed along the local xi-axis
OutList['TwHt6TPyi']          = False     # (m); yi-component of the translational position (relative to the inertia frame) of tower gage 6; Directed along the local yi-axis
OutList['TwHt6TPzi']          = False     # (m); zi-component of the translational position (relative to ground level [onshore] or MSL [offshore]) of tower gage 6; Directed along the local zi-axis
OutList['TwHt7TPxi']          = False     # (m); xi-component of the translational position (relative to the inertia frame) of tower gage 7; Directed along the local xi-axis
OutList['TwHt7TPyi']          = False     # (m); yi-component of the translational position (relative to the inertia frame) of tower gage 7; Directed along the local yi-axis
OutList['TwHt7TPzi']          = False     # (m); zi-component of the translational position (relative to ground level [onshore] or MSL [offshore]) of tower gage 7; Directed along the local zi-axis
OutList['TwHt8TPxi']          = False     # (m); xi-component of the translational position (relative to the inertia frame) of tower gage 8; Directed along the local xi-axis
OutList['TwHt8TPyi']          = False     # (m); yi-component of the translational position (relative to the inertia frame) of tower gage 8; Directed along the local yi-axis
OutList['TwHt8TPzi']          = False     # (m); zi-component of the translational position (relative to ground level [onshore] or MSL [offshore]) of tower gage 8; Directed along the local zi-axis
OutList['TwHt9TPxi']          = False     # (m); xi-component of the translational position (relative to the inertia frame) of tower gage 9; Directed along the local xi-axis
OutList['TwHt9TPyi']          = False     # (m); yi-component of the translational position (relative to the inertia frame) of tower gage 9; Directed along the local yi-axis
OutList['TwHt9TPzi']          = False     # (m); zi-component of the translational position (relative to ground level [onshore] or MSL [offshore]) of tower gage 9; Directed along the local zi-axis
OutList['TwHt1RPxi']          = False     # (deg); xi-component of the rotational position (relative to the inertia frame) of tower gage 1. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower and platform rotational deflections, so that the rotation sequence does not matter.; About the local xi-axis
OutList['TwHt1RPyi']          = False     # (deg); yi-component of the rotational position (relative to the inertia frame) of tower gage 1. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower and platform rotational deflections, so that the rotation sequence does not matter.; About the local yi-axis
OutList['TwHt1RPzi']          = False     # (deg); zi-component of the rotational position (relative to the inertia frame) of tower gage 1. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower and platform rotational deflections, so that the rotation sequence does not matter.; About the local zi-axis
OutList['TwHt2RPxi']          = False     # (deg); xi-component of the rotational position (relative to the inertia frame) of tower gage 2. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower and platform rotational deflections, so that the rotation sequence does not matter.; About the local xi-axis
OutList['TwHt2RPyi']          = False     # (deg); yi-component of the rotational position (relative to the inertia frame) of tower gage 2. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower and platform rotational deflections, so that the rotation sequence does not matter.; About the local yi-axis
OutList['TwHt2RPzi']          = False     # (deg); zi-component of the rotational position (relative to the inertia frame) of tower gage 2. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower and platform rotational deflections, so that the rotation sequence does not matter.; About the local zi-axis
OutList['TwHt3RPxi']          = False     # (deg); xi-component of the rotational position (relative to the inertia frame) of tower gage 3. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower and platform rotational deflections, so that the rotation sequence does not matter.; About the local xi-axis
OutList['TwHt3RPyi']          = False     # (deg); yi-component of the rotational position (relative to the inertia frame) of tower gage 3. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower and platform rotational deflections, so that the rotation sequence does not matter.; About the local yi-axis
OutList['TwHt3RPzi']          = False     # (deg); zi-component of the rotational position (relative to the inertia frame) of tower gage 3. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower and platform rotational deflections, so that the rotation sequence does not matter.; About the local zi-axis
OutList['TwHt4RPxi']          = False     # (deg); xi-component of the rotational position (relative to the inertia frame) of tower gage 4. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower and platform rotational deflections, so that the rotation sequence does not matter.; About the local xi-axis
OutList['TwHt4RPyi']          = False     # (deg); yi-component of the rotational position (relative to the inertia frame) of tower gage 4. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower and platform rotational deflections, so that the rotation sequence does not matter.; About the local yi-axis
OutList['TwHt4RPzi']          = False     # (deg); zi-component of the rotational position (relative to the inertia frame) of tower gage 4. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower and platform rotational deflections, so that the rotation sequence does not matter.; About the local zi-axis
OutList['TwHt5RPxi']          = False     # (deg); xi-component of the rotational position (relative to the inertia frame) of tower gage 5. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower and platform rotational deflections, so that the rotation sequence does not matter.; About the local xi-axis
OutList['TwHt5RPyi']          = False     # (deg); yi-component of the rotational position (relative to the inertia frame) of tower gage 5. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower and platform rotational deflections, so that the rotation sequence does not matter.; About the local yi-axis
OutList['TwHt5RPzi']          = False     # (deg); zi-component of the rotational position (relative to the inertia frame) of tower gage 5. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower and platform rotational deflections, so that the rotation sequence does not matter.; About the local zi-axis
OutList['TwHt6RPxi']          = False     # (deg); xi-component of the rotational position (relative to the inertia frame) of tower gage 6. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower and platform rotational deflections, so that the rotation sequence does not matter.; About the local xi-axis
OutList['TwHt6RPyi']          = False     # (deg); yi-component of the rotational position (relative to the inertia frame) of tower gage 6. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower and platform rotational deflections, so that the rotation sequence does not matter.; About the local yi-axis
OutList['TwHt6RPzi']          = False     # (deg); zi-component of the rotational position (relative to the inertia frame) of tower gage 6. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower and platform rotational deflections, so that the rotation sequence does not matter.; About the local zi-axis
OutList['TwHt7RPxi']          = False     # (deg); xi-component of the rotational position (relative to the inertia frame) of tower gage 7. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower and platform rotational deflections, so that the rotation sequence does not matter.; About the local xi-axis
OutList['TwHt7RPyi']          = False     # (deg); yi-component of the rotational position (relative to the inertia frame) of tower gage 7. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower and platform rotational deflections, so that the rotation sequence does not matter.; About the local yi-axis
OutList['TwHt7RPzi']          = False     # (deg); zi-component of the rotational position (relative to the inertia frame) of tower gage 7. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower and platform rotational deflections, so that the rotation sequence does not matter.; About the local zi-axis
OutList['TwHt8RPxi']          = False     # (deg); xi-component of the rotational position (relative to the inertia frame) of tower gage 8. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower and platform rotational deflections, so that the rotation sequence does not matter.; About the local xi-axis
OutList['TwHt8RPyi']          = False     # (deg); yi-component of the rotational position (relative to the inertia frame) of tower gage 8. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower and platform rotational deflections, so that the rotation sequence does not matter.; About the local yi-axis
OutList['TwHt8RPzi']          = False     # (deg); zi-component of the rotational position (relative to the inertia frame) of tower gage 8. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower and platform rotational deflections, so that the rotation sequence does not matter.; About the local zi-axis
OutList['TwHt9RPxi']          = False     # (deg); xi-component of the rotational position (relative to the inertia frame) of tower gage 9. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower and platform rotational deflections, so that the rotation sequence does not matter.; About the local xi-axis
OutList['TwHt9RPyi']          = False     # (deg); yi-component of the rotational position (relative to the inertia frame) of tower gage 9. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower and platform rotational deflections, so that the rotation sequence does not matter.; About the local yi-axis
OutList['TwHt9RPzi']          = False     # (deg); zi-component of the rotational position (relative to the inertia frame) of tower gage 9. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small tower and platform rotational deflections, so that the rotation sequence does not matter.; About the local zi-axis

# Platform Motions
OutList['PtfmTDxt']           = False     # (m); Platform horizontal surge (translational) displacement; Directed along the xt-axis
OutList['PtfmTDyt']           = False     # (m); Platform horizontal sway (translational) displacement; Directed along the yt-axis
OutList['PtfmTDzt']           = False     # (m); Platform vertical heave (translational) displacement; Directed along the zt-axis
OutList['PtfmTDxi']           = False     # (m); Platform horizontal surge (translational) displacement; Directed along the xi-axis
OutList['PtfmSurge']          = False     # (m); Platform horizontal surge (translational) displacement; Directed along the xi-axis
OutList['PtfmTDyi']           = False     # (m); Platform horizontal sway (translational) displacement; Directed along the yi-axis
OutList['PtfmSway']           = False     # (m); Platform horizontal sway (translational) displacement; Directed along the yi-axis
OutList['PtfmTDzi']           = False     # (m); Platform vertical heave (translational) displacement; Directed along the zi-axis
OutList['PtfmHeave']          = False     # (m); Platform vertical heave (translational) displacement; Directed along the zi-axis
OutList['PtfmTVxt']           = False     # (m/s); Platform horizontal surge (translational) velocity; Directed along the xt-axis
OutList['PtfmTVyt']           = False     # (m/s); Platform horizontal sway (translational) velocity; Directed along the yt-axis
OutList['PtfmTVzt']           = False     # (m/s); Platform vertical heave (translational) velocity; Directed along the zt-axis
OutList['PtfmTVxi']           = False     # (m/s); Platform horizontal surge (translational) velocity; Directed along the xi-axis
OutList['PtfmTVyi']           = False     # (m/s); Platform horizontal sway (translational) velocity; Directed along the yi-axis
OutList['PtfmTVzi']           = False     # (m/s); Platform vertical heave (translational) velocity; Directed along the zi-axis
OutList['PtfmTAxt']           = False     # (m/s^2); Platform horizontal surge (translational) acceleration; Directed along the xt-axis
OutList['PtfmTAyt']           = False     # (m/s^2); Platform horizontal sway (translational) acceleration; Directed along the yt-axis
OutList['PtfmTAzt']           = False     # (m/s^2); Platform vertical heave (translational) acceleration; Directed along the zt-axis
OutList['PtfmTAxi']           = False     # (m/s^2); Platform horizontal surge (translational) acceleration; Directed along the xi-axis
OutList['PtfmTAyi']           = False     # (m/s^2); Platform horizontal sway (translational) acceleration; Directed along the yi-axis
OutList['PtfmTAzi']           = False     # (m/s^2); Platform vertical heave (translational) acceleration; Directed along the zi-axis
OutList['PtfmRDxi']           = False     # (deg); Platform roll tilt angular (rotational) displacement. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small rotational platform displacements, so that the rotation sequence does not matter.; About the xi-axis
OutList['PtfmRoll']           = False     # (deg); Platform roll tilt angular (rotational) displacement. In ADAMS, it is output as an Euler angle computed as the 3rd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small rotational platform displacements, so that the rotation sequence does not matter.; About the xi-axis
OutList['PtfmRDyi']           = False     # (deg); Platform pitch tilt angular (rotational) displacement. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small rotational platform displacements, so that the rotation sequence does not matter.; About the yi-axis
OutList['PtfmPitch']          = False     # (deg); Platform pitch tilt angular (rotational) displacement. In ADAMS, it is output as an Euler angle computed as the 2nd rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small rotational platform displacements, so that the rotation sequence does not matter.; About the yi-axis
OutList['PtfmRDzi']           = False     # (deg); Platform yaw angular (rotational) displacement. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small rotational platform displacements, so that the rotation sequence does not matter.; About the zi-axis
OutList['PtfmYaw']            = False     # (deg); Platform yaw angular (rotational) displacement. In ADAMS, it is output as an Euler angle computed as the 1st rotation in the yaw-pitch-roll rotation sequence. It is not output as an Euler angle in FAST, which assumes small rotational platform displacements, so that the rotation sequence does not matter.; About the zi-axis
OutList['PtfmRVxt']           = False     # (deg/s); Platform roll tilt angular (rotational) velocity; About the xt-axis
OutList['PtfmRVyt']           = False     # (deg/s); Platform pitch tilt angular (rotational) velocity; About the yt-axis
OutList['PtfmRVzt']           = False     # (deg/s); Platform yaw angular (rotational) velocity; About the zt-axis
OutList['PtfmRVxi']           = False     # (deg/s); Platform roll tilt angular (rotational) velocity; About the xi-axis
OutList['PtfmRVyi']           = False     # (deg/s); Platform pitch tilt angular (rotational) velocity; About the yi-axis
OutList['PtfmRVzi']           = False     # (deg/s); Platform yaw angular (rotational) velocity; About the zi-axis
OutList['PtfmRAxt']           = False     # (deg/s^2); Platform roll tilt angular (rotational) acceleration; About the xt-axis
OutList['PtfmRAyt']           = False     # (deg/s^2); Platform pitch tilt angular (rotational) acceleration; About the yt-axis
OutList['PtfmRAzt']           = False     # (deg/s^2); Platform yaw angular (rotational) acceleration; About the zt-axis
OutList['PtfmRAxi']           = False     # (deg/s^2); Platform roll tilt angular (rotational) acceleration; About the xi-axis
OutList['PtfmRAyi']           = False     # (deg/s^2); Platform pitch tilt angular (rotational) acceleration; About the yi-axis
OutList['PtfmRAzi']           = False     # (deg/s^2); Platform yaw angular (rotational) acceleration; About the zi-axis

# Blade 1 Root Loads
OutList['RootFxc1']           = False     # (kN); Blade 1 out-of-plane shear force at the blade root; Directed along the xc1-axis
OutList['RootFyc1']           = False     # (kN); Blade 1 in-plane shear force at the blade root; Directed along the yc1-axis
OutList['RootFzc1']           = False     # (kN); Blade 1 axial force at the blade root; Directed along the zc1- and zb1-axes
OutList['RootFzb1']           = False     # (kN); Blade 1 axial force at the blade root; Directed along the zc1- and zb1-axes
OutList['RootFxb1']           = False     # (kN); Blade 1 flapwise shear force at the blade root; Directed along the xb1-axis
OutList['RootFyb1']           = False     # (kN); Blade 1 edgewise shear force at the blade root; Directed along the yb1-axis
OutList['RootMxc1']           = False     # (kN m); Blade 1 in-plane moment (i.e., the moment caused by in-plane forces) at the blade root; About the xc1-axis
OutList['RootMIP1']           = False     # (kN m); Blade 1 in-plane moment (i.e., the moment caused by in-plane forces) at the blade root; About the xc1-axis
OutList['RootMyc1']           = False     # (kN m); Blade 1 out-of-plane moment (i.e., the moment caused by out-of-plane forces) at the blade root; About the yc1-axis
OutList['RootMOoP1']          = False     # (kN m); Blade 1 out-of-plane moment (i.e., the moment caused by out-of-plane forces) at the blade root; About the yc1-axis
OutList['RootMzc1']           = False     # (kN m); Blade 1 pitching moment at the blade root; About the zc1- and zb1-axes
OutList['RootMzb1']           = False     # (kN m); Blade 1 pitching moment at the blade root; About the zc1- and zb1-axes
OutList['RootMxb1']           = False     # (kN m); Blade 1 edgewise moment (i.e., the moment caused by edgewise forces) at the blade root; About the xb1-axis
OutList['RootMEdg1']          = False     # (kN m); Blade 1 edgewise moment (i.e., the moment caused by edgewise forces) at the blade root; About the xb1-axis
OutList['RootMyb1']           = False     # (kN m); Blade 1 flapwise moment (i.e., the moment caused by flapwise forces) at the blade root; About the yb1-axis
OutList['RootMFlp1']          = False     # (kN m); Blade 1 flapwise moment (i.e., the moment caused by flapwise forces) at the blade root; About the yb1-axis

# Blade 2 Root Loads
OutList['RootFxc2']           = False     # (kN); Blade 2 out-of-plane shear force at the blade root; Directed along the xc2-axis
OutList['RootFyc2']           = False     # (kN); Blade 2 in-plane shear force at the blade root; Directed along the yc2-axis
OutList['RootFzc2']           = False     # (kN); Blade 2 axial force at the blade root; Directed along the zc2- and zb2-axes
OutList['RootFzb2']           = False     # (kN); Blade 2 axial force at the blade root; Directed along the zc2- and zb2-axes
OutList['RootFxb2']           = False     # (kN); Blade 2 flapwise shear force at the blade root; Directed along the xb2-axis
OutList['RootFyb2']           = False     # (kN); Blade 2 edgewise shear force at the blade root; Directed along the yb2-axis
OutList['RootMxc2']           = False     # (kN m); Blade 2 in-plane moment (i.e., the moment caused by in-plane forces) at the blade root; About the xc2-axis
OutList['RootMIP2']           = False     # (kN m); Blade 2 in-plane moment (i.e., the moment caused by in-plane forces) at the blade root; About the xc2-axis
OutList['RootMyc2']           = False     # (kN m); Blade 2 out-of-plane moment (i.e., the moment caused by out-of-plane forces) at the blade root; About the yc2-axis
OutList['RootMOoP2']          = False     # (kN m); Blade 2 out-of-plane moment (i.e., the moment caused by out-of-plane forces) at the blade root; About the yc2-axis
OutList['RootMzc2']           = False     # (kN m); Blade 2 pitching moment at the blade root; About the zc2- and zb2-axes
OutList['RootMzb2']           = False     # (kN m); Blade 2 pitching moment at the blade root; About the zc2- and zb2-axes
OutList['RootMxb2']           = False     # (kN m); Blade 2 edgewise moment (i.e., the moment caused by edgewise forces) at the blade root; About the xb2-axis
OutList['RootMEdg2']          = False     # (kN m); Blade 2 edgewise moment (i.e., the moment caused by edgewise forces) at the blade root; About the xb2-axis
OutList['RootMyb2']           = False     # (kN m); Blade 2 flapwise moment (i.e., the moment caused by flapwise forces) at the blade root; About the yb2-axis
OutList['RootMFlp2']          = False     # (kN m); Blade 2 flapwise moment (i.e., the moment caused by flapwise forces) at the blade root; About the yb2-axis

# Blade 3 Root Loads
OutList['RootFxc3']           = False     # (kN); Blade 3 out-of-plane shear force at the blade root; Directed along the xc3-axis
OutList['RootFyc3']           = False     # (kN); Blade 3 in-plane shear force at the blade root; Directed along the yc3-axis
OutList['RootFzc3']           = False     # (kN); Blade 3 axial force at the blade root; Directed along the zc3- and zb3-axes
OutList['RootFzb3']           = False     # (kN); Blade 3 axial force at the blade root; Directed along the zc3- and zb3-axes
OutList['RootFxb3']           = False     # (kN); Blade 3 flapwise shear force at the blade root; Directed along the xb3-axis
OutList['RootFyb3']           = False     # (kN); Blade 3 edgewise shear force at the blade root; Directed along the yb3-axis
OutList['RootMxc3']           = False     # (kN m); Blade 3 in-plane moment (i.e., the moment caused by in-plane forces) at the blade root; About the xc3-axis
OutList['RootMIP3']           = False     # (kN m); Blade 3 in-plane moment (i.e., the moment caused by in-plane forces) at the blade root; About the xc3-axis
OutList['RootMyc3']           = False     # (kN m); Blade 3 out-of-plane moment (i.e., the moment caused by out-of-plane forces) at the blade root; About the yc3-axis
OutList['RootMOoP3']          = False     # (kN m); Blade 3 out-of-plane moment (i.e., the moment caused by out-of-plane forces) at the blade root; About the yc3-axis
OutList['RootMzc3']           = False     # (kN m); Blade 3 pitching moment at the blade root; About the zc3- and zb3-axes
OutList['RootMzb3']           = False     # (kN m); Blade 3 pitching moment at the blade root; About the zc3- and zb3-axes
OutList['RootMxb3']           = False     # (kN m); Blade 3 edgewise moment (i.e., the moment caused by edgewise forces) at the blade root; About the xb3-axis
OutList['RootMEdg3']          = False     # (kN m); Blade 3 edgewise moment (i.e., the moment caused by edgewise forces) at the blade root; About the xb3-axis
OutList['RootMyb3']           = False     # (kN m); Blade 3 flapwise moment (i.e., the moment caused by flapwise forces) at the blade root; About the yb3-axis
OutList['RootMFlp3']          = False     # (kN m); Blade 3 flapwise moment (i.e., the moment caused by flapwise forces) at the blade root; About the yb3-axis

# Blade 1 Local Span Loads
OutList['Spn1MLxb1']          = False     # (kN m); Blade 1 local edgewise moment at span station 1; About the local xb1-axis
OutList['Spn1MLyb1']          = False     # (kN m); Blade 1 local flapwise moment at span station 1; About the local yb1-axis
OutList['Spn1MLzb1']          = False     # (kN m); Blade 1 local pitching moment at span station 1; About the local zb1-axis
OutList['Spn2MLxb1']          = False     # (kN m); Blade 1 local edgewise moment at span station 2; About the local xb1-axis
OutList['Spn2MLyb1']          = False     # (kN m); Blade 1 local flapwise moment at span station 2; About the local yb1-axis
OutList['Spn2MLzb1']          = False     # (kN m); Blade 1 local pitching moment at span station 2; About the local zb1-axis
OutList['Spn3MLxb1']          = False     # (kN m); Blade 1 local edgewise moment at span station 3; About the local xb1-axis
OutList['Spn3MLyb1']          = False     # (kN m); Blade 1 local flapwise moment at span station 3; About the local yb1-axis
OutList['Spn3MLzb1']          = False     # (kN m); Blade 1 local pitching moment at span station 3; About the local zb1-axis
OutList['Spn4MLxb1']          = False     # (kN m); Blade 1 local edgewise moment at span station 4; About the local xb1-axis
OutList['Spn4MLyb1']          = False     # (kN m); Blade 1 local flapwise moment at span station 4; About the local yb1-axis
OutList['Spn4MLzb1']          = False     # (kN m); Blade 1 local pitching moment at span station 4; About the local zb1-axis
OutList['Spn5MLxb1']          = False     # (kN m); Blade 1 local edgewise moment at span station 5; About the local xb1-axis
OutList['Spn5MLyb1']          = False     # (kN m); Blade 1 local flapwise moment at span station 5; About the local yb1-axis
OutList['Spn5MLzb1']          = False     # (kN m); Blade 1 local pitching moment at span station 5; About the local zb1-axis
OutList['Spn6MLxb1']          = False     # (kN m); Blade 1 local edgewise moment at span station 6; About the local xb1-axis
OutList['Spn6MLyb1']          = False     # (kN m); Blade 1 local flapwise moment at span station 6; About the local yb1-axis
OutList['Spn6MLzb1']          = False     # (kN m); Blade 1 local pitching moment at span station 6; About the local zb1-axis
OutList['Spn7MLxb1']          = False     # (kN m); Blade 1 local edgewise moment at span station 7; About the local xb1-axis
OutList['Spn7MLyb1']          = False     # (kN m); Blade 1 local flapwise moment at span station 7; About the local yb1-axis
OutList['Spn7MLzb1']          = False     # (kN m); Blade 1 local pitching moment at span station 7; About the local zb1-axis
OutList['Spn8MLxb1']          = False     # (kN m); Blade 1 local edgewise moment at span station 8; About the local xb1-axis
OutList['Spn8MLyb1']          = False     # (kN m); Blade 1 local flapwise moment at span station 8; About the local yb1-axis
OutList['Spn8MLzb1']          = False     # (kN m); Blade 1 local pitching moment at span station 8; About the local zb1-axis
OutList['Spn9MLxb1']          = False     # (kN m); Blade 1 local edgewise moment at span station 9; About the local xb1-axis
OutList['Spn9MLyb1']          = False     # (kN m); Blade 1 local flapwise moment at span station 9; About the local yb1-axis
OutList['Spn9MLzb1']          = False     # (kN m); Blade 1 local pitching moment at span station 9; About the local zb1-axis
OutList['Spn1FLxb1']          = False     # (kN); Blade 1 local flapwise shear force at span station 1; Directed along the local xb1-axis
OutList['Spn1FLyb1']          = False     # (kN); Blade 1 local edgewise shear force at span station 1; Directed along the local yb1-axis
OutList['Spn1FLzb1']          = False     # (kN); Blade 1 local axial force at span station 1; Directed along the local zb1-axis
OutList['Spn2FLxb1']          = False     # (kN); Blade 1 local flapwise shear force at span station 2; Directed along the local xb1-axis
OutList['Spn2FLyb1']          = False     # (kN); Blade 1 local edgewise shear force at span station 2; Directed along the local yb1-axis
OutList['Spn2FLzb1']          = False     # (kN); Blade 1 local axial force at span station 2; Directed along the local zb1-axis
OutList['Spn3FLxb1']          = False     # (kN); Blade 1 local flapwise shear force at span station 3; Directed along the local xb1-axis
OutList['Spn3FLyb1']          = False     # (kN); Blade 1 local edgewise shear force at span station 3; Directed along the local yb1-axis
OutList['Spn3FLzb1']          = False     # (kN); Blade 1 local axial force at span station 3; Directed along the local zb1-axis
OutList['Spn4FLxb1']          = False     # (kN); Blade 1 local flapwise shear force at span station 4; Directed along the local xb1-axis
OutList['Spn4FLyb1']          = False     # (kN); Blade 1 local edgewise shear force at span station 4; Directed along the local yb1-axis
OutList['Spn4FLzb1']          = False     # (kN); Blade 1 local axial force at span station 4; Directed along the local zb1-axis
OutList['Spn5FLxb1']          = False     # (kN); Blade 1 local flapwise shear force at span station 5; Directed along the local xb1-axis
OutList['Spn5FLyb1']          = False     # (kN); Blade 1 local edgewise shear force at span station 5; Directed along the local yb1-axis
OutList['Spn5FLzb1']          = False     # (kN); Blade 1 local axial force at span station 5; Directed along the local zb1-axis
OutList['Spn6FLxb1']          = False     # (kN); Blade 1 local flapwise shear force at span station 6; Directed along the local xb1-axis
OutList['Spn6FLyb1']          = False     # (kN); Blade 1 local edgewise shear force at span station 6; Directed along the local yb1-axis
OutList['Spn6FLzb1']          = False     # (kN); Blade 1 local axial force at span station 6; Directed along the local zb1-axis
OutList['Spn7FLxb1']          = False     # (kN); Blade 1 local flapwise shear force at span station 7; Directed along the local xb1-axis
OutList['Spn7FLyb1']          = False     # (kN); Blade 1 local edgewise shear force at span station 7; Directed along the local yb1-axis
OutList['Spn7FLzb1']          = False     # (kN); Blade 1 local axial force at span station 7; Directed along the local zb1-axis
OutList['Spn8FLxb1']          = False     # (kN); Blade 1 local flapwise shear force at span station 8; Directed along the local xb1-axis
OutList['Spn8FLyb1']          = False     # (kN); Blade 1 local edgewise shear force at span station 8; Directed along the local yb1-axis
OutList['Spn8FLzb1']          = False     # (kN); Blade 1 local axial force at span station 8; Directed along the local zb1-axis
OutList['Spn9FLxb1']          = False     # (kN); Blade 1 local flapwise shear force at span station 9; Directed along the local xb1-axis
OutList['Spn9FLyb1']          = False     # (kN); Blade 1 local edgewise shear force at span station 9; Directed along the local yb1-axis
OutList['Spn9FLzb1']          = False     # (kN); Blade 1 local axial force at span station 9; Directed along the local zb1-axis

# Blade 2 Local Span Loads
OutList['Spn1MLxb2']          = False     # (kN m); Blade 2 local edgewise moment at span station 1; About the local xb2-axis
OutList['Spn1MLyb2']          = False     # (kN m); Blade 2 local flapwise moment at span station 1; About the local yb2-axis
OutList['Spn1MLzb2']          = False     # (kN m); Blade 2 local pitching moment at span station 1; About the local zb2-axis
OutList['Spn2MLxb2']          = False     # (kN m); Blade 2 local edgewise moment at span station 2; About the local xb2-axis
OutList['Spn2MLyb2']          = False     # (kN m); Blade 2 local flapwise moment at span station 2; About the local yb2-axis
OutList['Spn2MLzb2']          = False     # (kN m); Blade 2 local pitching moment at span station 2; About the local zb2-axis
OutList['Spn3MLxb2']          = False     # (kN m); Blade 2 local edgewise moment at span station 3; About the local xb2-axis
OutList['Spn3MLyb2']          = False     # (kN m); Blade 2 local flapwise moment at span station 3; About the local yb2-axis
OutList['Spn3MLzb2']          = False     # (kN m); Blade 2 local pitching moment at span station 3; About the local zb2-axis
OutList['Spn4MLxb2']          = False     # (kN m); Blade 2 local edgewise moment at span station 4; About the local xb2-axis
OutList['Spn4MLyb2']          = False     # (kN m); Blade 2 local flapwise moment at span station 4; About the local yb2-axis
OutList['Spn4MLzb2']          = False     # (kN m); Blade 2 local pitching moment at span station 4; About the local zb2-axis
OutList['Spn5MLxb2']          = False     # (kN m); Blade 2 local edgewise moment at span station 5; About the local xb2-axis
OutList['Spn5MLyb2']          = False     # (kN m); Blade 2 local flapwise moment at span station 5; About the local yb2-axis
OutList['Spn5MLzb2']          = False     # (kN m); Blade 2 local pitching moment at span station 5; About the local zb2-axis
OutList['Spn6MLxb2']          = False     # (kN m); Blade 2 local edgewise moment at span station 6; About the local xb2-axis
OutList['Spn6MLyb2']          = False     # (kN m); Blade 2 local flapwise moment at span station 6; About the local yb2-axis
OutList['Spn6MLzb2']          = False     # (kN m); Blade 2 local pitching moment at span station 6; About the local zb2-axis
OutList['Spn7MLxb2']          = False     # (kN m); Blade 2 local edgewise moment at span station 7; About the local xb2-axis
OutList['Spn7MLyb2']          = False     # (kN m); Blade 2 local flapwise moment at span station 7; About the local yb2-axis
OutList['Spn7MLzb2']          = False     # (kN m); Blade 2 local pitching moment at span station 7; About the local zb2-axis
OutList['Spn8MLxb2']          = False     # (kN m); Blade 2 local edgewise moment at span station 8; About the local xb2-axis
OutList['Spn8MLyb2']          = False     # (kN m); Blade 2 local flapwise moment at span station 8; About the local yb2-axis
OutList['Spn8MLzb2']          = False     # (kN m); Blade 2 local pitching moment at span station 8; About the local zb2-axis
OutList['Spn9MLxb2']          = False     # (kN m); Blade 2 local edgewise moment at span station 9; About the local xb2-axis
OutList['Spn9MLyb2']          = False     # (kN m); Blade 2 local flapwise moment at span station 9; About the local yb2-axis
OutList['Spn9MLzb2']          = False     # (kN m); Blade 2 local pitching moment at span station 9; About the local zb2-axis
OutList['Spn1FLxb2']          = False     # (kN); Blade 2 local flapwise shear force at span station 1; Directed along the local xb2-axis
OutList['Spn1FLyb2']          = False     # (kN); Blade 2 local edgewise shear force at span station 1; Directed along the local yb2-axis
OutList['Spn1FLzb2']          = False     # (kN); Blade 2 local axial force at span station 1; Directed along the local zb2-axis
OutList['Spn2FLxb2']          = False     # (kN); Blade 2 local flapwise shear force at span station 2; Directed along the local xb2-axis
OutList['Spn2FLyb2']          = False     # (kN); Blade 2 local edgewise shear force at span station 2; Directed along the local yb2-axis
OutList['Spn2FLzb2']          = False     # (kN); Blade 2 local axial force at span station 2; Directed along the local zb2-axis
OutList['Spn3FLxb2']          = False     # (kN); Blade 2 local flapwise shear force at span station 3; Directed along the local xb2-axis
OutList['Spn3FLyb2']          = False     # (kN); Blade 2 local edgewise shear force at span station 3; Directed along the local yb2-axis
OutList['Spn3FLzb2']          = False     # (kN); Blade 2 local axial force at span station 3; Directed along the local zb2-axis
OutList['Spn4FLxb2']          = False     # (kN); Blade 2 local flapwise shear force at span station 4; Directed along the local xb2-axis
OutList['Spn4FLyb2']          = False     # (kN); Blade 2 local edgewise shear force at span station 4; Directed along the local yb2-axis
OutList['Spn4FLzb2']          = False     # (kN); Blade 2 local axial force at span station 4; Directed along the local zb2-axis
OutList['Spn5FLxb2']          = False     # (kN); Blade 2 local flapwise shear force at span station 5; Directed along the local xb2-axis
OutList['Spn5FLyb2']          = False     # (kN); Blade 2 local edgewise shear force at span station 5; Directed along the local yb2-axis
OutList['Spn5FLzb2']          = False     # (kN); Blade 2 local axial force at span station 5; Directed along the local zb2-axis
OutList['Spn6FLxb2']          = False     # (kN); Blade 2 local flapwise shear force at span station 6; Directed along the local xb2-axis
OutList['Spn6FLyb2']          = False     # (kN); Blade 2 local edgewise shear force at span station 6; Directed along the local yb2-axis
OutList['Spn6FLzb2']          = False     # (kN); Blade 2 local axial force at span station 6; Directed along the local zb2-axis
OutList['Spn7FLxb2']          = False     # (kN); Blade 2 local flapwise shear force at span station 7; Directed along the local xb2-axis
OutList['Spn7FLyb2']          = False     # (kN); Blade 2 local edgewise shear force at span station 7; Directed along the local yb2-axis
OutList['Spn7FLzb2']          = False     # (kN); Blade 2 local axial force at span station 7; Directed along the local zb2-axis
OutList['Spn8FLxb2']          = False     # (kN); Blade 2 local flapwise shear force at span station 8; Directed along the local xb2-axis
OutList['Spn8FLyb2']          = False     # (kN); Blade 2 local edgewise shear force at span station 8; Directed along the local yb2-axis
OutList['Spn8FLzb2']          = False     # (kN); Blade 2 local axial force at span station 8; Directed along the local zb2-axis
OutList['Spn9FLxb2']          = False     # (kN); Blade 2 local flapwise shear force at span station 9; Directed along the local xb2-axis
OutList['Spn9FLyb2']          = False     # (kN); Blade 2 local edgewise shear force at span station 9; Directed along the local yb2-axis
OutList['Spn9FLzb2']          = False     # (kN); Blade 2 local axial force at span station 9; Directed along the local zb2-axis

# Blade 3 Local Span Loads
OutList['Spn1MLxb3']          = False     # (kN m); Blade 3 local edgewise moment at span station 1; About the local xb3-axis
OutList['Spn1MLyb3']          = False     # (kN m); Blade 3 local flapwise moment at span station 1; About the local yb3-axis
OutList['Spn1MLzb3']          = False     # (kN m); Blade 3 local pitching moment at span station 1; About the local zb3-axis
OutList['Spn2MLxb3']          = False     # (kN m); Blade 3 local edgewise moment at span station 2; About the local xb3-axis
OutList['Spn2MLyb3']          = False     # (kN m); Blade 3 local flapwise moment at span station 2; About the local yb3-axis
OutList['Spn2MLzb3']          = False     # (kN m); Blade 3 local pitching moment at span station 2; About the local zb3-axis
OutList['Spn3MLxb3']          = False     # (kN m); Blade 3 local edgewise moment at span station 3; About the local xb3-axis
OutList['Spn3MLyb3']          = False     # (kN m); Blade 3 local flapwise moment at span station 3; About the local yb3-axis
OutList['Spn3MLzb3']          = False     # (kN m); Blade 3 local pitching moment at span station 3; About the local zb3-axis
OutList['Spn4MLxb3']          = False     # (kN m); Blade 3 local edgewise moment at span station 4; About the local xb3-axis
OutList['Spn4MLyb3']          = False     # (kN m); Blade 3 local flapwise moment at span station 4; About the local yb3-axis
OutList['Spn4MLzb3']          = False     # (kN m); Blade 3 local pitching moment at span station 4; About the local zb3-axis
OutList['Spn5MLxb3']          = False     # (kN m); Blade 3 local edgewise moment at span station 5; About the local xb3-axis
OutList['Spn5MLyb3']          = False     # (kN m); Blade 3 local flapwise moment at span station 5; About the local yb3-axis
OutList['Spn5MLzb3']          = False     # (kN m); Blade 3 local pitching moment at span station 5; About the local zb3-axis
OutList['Spn6MLxb3']          = False     # (kN m); Blade 3 local edgewise moment at span station 6; About the local xb3-axis
OutList['Spn6MLyb3']          = False     # (kN m); Blade 3 local flapwise moment at span station 6; About the local yb3-axis
OutList['Spn6MLzb3']          = False     # (kN m); Blade 3 local pitching moment at span station 6; About the local zb3-axis
OutList['Spn7MLxb3']          = False     # (kN m); Blade 3 local edgewise moment at span station 7; About the local xb3-axis
OutList['Spn7MLyb3']          = False     # (kN m); Blade 3 local flapwise moment at span station 7; About the local yb3-axis
OutList['Spn7MLzb3']          = False     # (kN m); Blade 3 local pitching moment at span station 7; About the local zb3-axis
OutList['Spn8MLxb3']          = False     # (kN m); Blade 3 local edgewise moment at span station 8; About the local xb3-axis
OutList['Spn8MLyb3']          = False     # (kN m); Blade 3 local flapwise moment at span station 8; About the local yb3-axis
OutList['Spn8MLzb3']          = False     # (kN m); Blade 3 local pitching moment at span station 8; About the local zb3-axis
OutList['Spn9MLxb3']          = False     # (kN m); Blade 3 local edgewise moment at span station 9; About the local xb3-axis
OutList['Spn9MLyb3']          = False     # (kN m); Blade 3 local flapwise moment at span station 9; About the local yb3-axis
OutList['Spn9MLzb3']          = False     # (kN m); Blade 3 local pitching moment at span station 9; About the local zb3-axis
OutList['Spn1FLxb3']          = False     # (kN); Blade 3 local flapwise shear force at span station 1; Directed along the local xb3-axis
OutList['Spn1FLyb3']          = False     # (kN); Blade 3 local edgewise shear force at span station 1; Directed along the local yb3-axis
OutList['Spn1FLzb3']          = False     # (kN); Blade 3 local axial force at span station 1; Directed along the local zb3-axis
OutList['Spn2FLxb3']          = False     # (kN); Blade 3 local flapwise shear force at span station 2; Directed along the local xb3-axis
OutList['Spn2FLyb3']          = False     # (kN); Blade 3 local edgewise shear force at span station 2; Directed along the local yb3-axis
OutList['Spn2FLzb3']          = False     # (kN); Blade 3 local axial force at span station 2; Directed along the local zb3-axis
OutList['Spn3FLxb3']          = False     # (kN); Blade 3 local flapwise shear force at span station 3; Directed along the local xb3-axis
OutList['Spn3FLyb3']          = False     # (kN); Blade 3 local edgewise shear force at span station 3; Directed along the local yb3-axis
OutList['Spn3FLzb3']          = False     # (kN); Blade 3 local axial force at span station 3; Directed along the local zb3-axis
OutList['Spn4FLxb3']          = False     # (kN); Blade 3 local flapwise shear force at span station 4; Directed along the local xb3-axis
OutList['Spn4FLyb3']          = False     # (kN); Blade 3 local edgewise shear force at span station 4; Directed along the local yb3-axis
OutList['Spn4FLzb3']          = False     # (kN); Blade 3 local axial force at span station 4; Directed along the local zb3-axis
OutList['Spn5FLxb3']          = False     # (kN); Blade 3 local flapwise shear force at span station 5; Directed along the local xb3-axis
OutList['Spn5FLyb3']          = False     # (kN); Blade 3 local edgewise shear force at span station 5; Directed along the local yb3-axis
OutList['Spn5FLzb3']          = False     # (kN); Blade 3 local axial force at span station 5; Directed along the local zb3-axis
OutList['Spn6FLxb3']          = False     # (kN); Blade 3 local flapwise shear force at span station 6; Directed along the local xb3-axis
OutList['Spn6FLyb3']          = False     # (kN); Blade 3 local edgewise shear force at span station 6; Directed along the local yb3-axis
OutList['Spn6FLzb3']          = False     # (kN); Blade 3 local axial force at span station 6; Directed along the local zb3-axis
OutList['Spn7FLxb3']          = False     # (kN); Blade 3 local flapwise shear force at span station 7; Directed along the local xb3-axis
OutList['Spn7FLyb3']          = False     # (kN); Blade 3 local edgewise shear force at span station 7; Directed along the local yb3-axis
OutList['Spn7FLzb3']          = False     # (kN); Blade 3 local axial force at span station 7; Directed along the local zb3-axis
OutList['Spn8FLxb3']          = False     # (kN); Blade 3 local flapwise shear force at span station 8; Directed along the local xb3-axis
OutList['Spn8FLyb3']          = False     # (kN); Blade 3 local edgewise shear force at span station 8; Directed along the local yb3-axis
OutList['Spn8FLzb3']          = False     # (kN); Blade 3 local axial force at span station 8; Directed along the local zb3-axis
OutList['Spn9FLxb3']          = False     # (kN); Blade 3 local flapwise shear force at span station 9; Directed along the local xb3-axis
OutList['Spn9FLyb3']          = False     # (kN); Blade 3 local edgewise shear force at span station 9; Directed along the local yb3-axis
OutList['Spn9FLzb3']          = False     # (kN); Blade 3 local axial force at span station 9; Directed along the local zb3-axis

# Hub and Rotor Loads
OutList['LSShftFxa']          = False     # (kN); Low-speed shaft thrust force (this is constant along the shaft and is equivalent to the rotor thrust force); Directed along the xa- and xs-axes
OutList['LSShftFxs']          = False     # (kN); Low-speed shaft thrust force (this is constant along the shaft and is equivalent to the rotor thrust force); Directed along the xa- and xs-axes
OutList['LSSGagFxa']          = False     # (kN); Low-speed shaft thrust force (this is constant along the shaft and is equivalent to the rotor thrust force); Directed along the xa- and xs-axes
OutList['LSSGagFxs']          = False     # (kN); Low-speed shaft thrust force (this is constant along the shaft and is equivalent to the rotor thrust force); Directed along the xa- and xs-axes
OutList['RotThrust']          = False     # (kN); Low-speed shaft thrust force (this is constant along the shaft and is equivalent to the rotor thrust force); Directed along the xa- and xs-axes
OutList['LSShftFya']          = False     # (kN); Rotating low-speed shaft shear force (this is constant along the shaft); Directed along the ya-axis
OutList['LSSGagFya']          = False     # (kN); Rotating low-speed shaft shear force (this is constant along the shaft); Directed along the ya-axis
OutList['LSShftFza']          = False     # (kN); Rotating low-speed shaft shear force (this is constant along the shaft); Directed along the za-axis
OutList['LSSGagFza']          = False     # (kN); Rotating low-speed shaft shear force (this is constant along the shaft); Directed along the za-axis
OutList['LSShftFys']          = False     # (kN); Nonrotating low-speed shaft shear force (this is constant along the shaft); Directed along the ys-axis
OutList['LSSGagFys']          = False     # (kN); Nonrotating low-speed shaft shear force (this is constant along the shaft); Directed along the ys-axis
OutList['LSShftFzs']          = False     # (kN); Nonrotating low-speed shaft shear force (this is constant along the shaft); Directed along the zs-axis
OutList['LSSGagFzs']          = False     # (kN); Nonrotating low-speed shaft shear force (this is constant along the shaft); Directed along the zs-axis
OutList['LSShftMxa']          = False     # (kN m); Low-speed shaft torque (this is constant along the shaft and is equivalent to the rotor torque); About the xa- and xs-axes
OutList['LSShftMxs']          = False     # (kN m); Low-speed shaft torque (this is constant along the shaft and is equivalent to the rotor torque); About the xa- and xs-axes
OutList['LSSGagMxa']          = False     # (kN m); Low-speed shaft torque (this is constant along the shaft and is equivalent to the rotor torque); About the xa- and xs-axes
OutList['LSSGagMxs']          = False     # (kN m); Low-speed shaft torque (this is constant along the shaft and is equivalent to the rotor torque); About the xa- and xs-axes
OutList['RotTorq']            = False     # (kN m); Low-speed shaft torque (this is constant along the shaft and is equivalent to the rotor torque); About the xa- and xs-axes
OutList['LSShftTq']           = False     # (kN m); Low-speed shaft torque (this is constant along the shaft and is equivalent to the rotor torque); About the xa- and xs-axes
OutList['LSSTipMya']          = False     # (kN m); Rotating low-speed shaft bending moment at the shaft tip (teeter pin for 2-blader, apex of rotation for 3-blader); About the ya-axis
OutList['LSSTipMza']          = False     # (kN m); Rotating low-speed shaft bending moment at the shaft tip (teeter pin for 2-blader, apex of rotation for 3-blader); About the za-axis
OutList['LSSTipMys']          = False     # (kN m); Nonrotating low-speed shaft bending moment at the shaft tip (teeter pin for 2-blader, apex of rotation for 3-blader); About the ys-axis
OutList['LSSTipMzs']          = False     # (kN m); Nonrotating low-speed shaft bending moment at the shaft tip (teeter pin for 2-blader, apex of rotation for 3-blader); About the zs-axis
OutList['CThrstAzm']          = False     # (deg); Azimuth location of the center of thrust. This is estimated using values of LSSTipMys, LSSTipMzs, and RotThrust.; About the xa- and xs-axes
OutList['CThrstRad']          = False     # (-); Dimensionless radial (arm) location of the center of thrust. This is estimated using values of LSSTipMys, LSSTipMzs, and RotThrust. (nondimensionalized using the undeflected tip radius normal to the shaft and limited to values between 0 and 1 (inclusive)); Always positive (directed radially outboard at azimuth angle CThrstAzm)
OutList['CThrstArm']          = False     # (-); Dimensionless radial (arm) location of the center of thrust. This is estimated using values of LSSTipMys, LSSTipMzs, and RotThrust. (nondimensionalized using the undeflected tip radius normal to the shaft and limited to values between 0 and 1 (inclusive)); Always positive (directed radially outboard at azimuth angle CThrstAzm)
OutList['RotPwr']             = False     # (kW); Rotor power (this is equivalent to the low-speed shaft power); N/A
OutList['LSShftPwr']          = False     # (kW); Rotor power (this is equivalent to the low-speed shaft power); N/A
OutList['RotCq']              = False     # (-); Rotor torque coefficient (this is equivalent to the low-speed shaft torque coefficient); N/A
OutList['LSShftCq']           = False     # (-); Rotor torque coefficient (this is equivalent to the low-speed shaft torque coefficient); N/A
OutList['RotCp']              = False     # (-); Rotor power coefficient (this is equivalent to the low-speed shaft power coefficient); N/A
OutList['LSShftCp']           = False     # (-); Rotor power coefficient (this is equivalent to the low-speed shaft power coefficient); N/A
OutList['RotCt']              = False     # (-); Rotor thrust coefficient (this is equivalent to the low-speed shaft thrust coefficient); N/A
OutList['LSShftCt']           = False     # (-); Rotor thrust coefficient (this is equivalent to the low-speed shaft thrust coefficient); N/A

# Shaft Strain Gage Loads
OutList['LSSGagMya']          = False     # (kN m); Rotating low-speed shaft bending moment at the shaft's strain gage (shaft strain gage located by input ShftGagL); About the ya-axis
OutList['LSSGagMza']          = False     # (kN m); Rotating low-speed shaft bending moment at the shaft's strain gage (shaft strain gage located by input ShftGagL); About the za-axis
OutList['LSSGagMys']          = False     # (kN m); Nonrotating low-speed shaft bending moment at the shaft's strain gage (shaft strain gage located by input ShftGagL); About the ys-axis
OutList['LSSGagMzs']          = False     # (kN m); Nonrotating low-speed shaft bending moment at the shaft's strain gage (shaft strain gage located by input ShftGagL); About the zs-axis

# Generator and High-Speed Shaft Loads
OutList['HSShftTq']           = False     # (kN m); High-speed shaft torque (this is constant along the shaft); Same sign as LSShftTq / RotTorq / LSShftMxa / LSShftMxs / LSSGagMxa / LSSGagMxs
OutList['HSShftPwr']          = False     # (kW); High-speed shaft power; Same sign as HSShftTq
OutList['HSShftCq']           = False     # (-); High-speed shaft torque coefficient; N/A
OutList['HSShftCp']           = False     # (-); High-speed shaft power coefficient; N/A
OutList['GenTq']              = False     # (kN m); Electrical generator torque; Positive reflects power extracted and negative represents a motoring-up situation (power input)
OutList['GenPwr']             = False     # (kW); Electrical generator power; Same sign as GenTq
OutList['GenCq']              = False     # (-); Electrical generator torque coefficient; N/A
OutList['GenCp']              = False     # (-); Electrical generator power coefficient; N/A
OutList['HSSBrTq']            = False     # (kN m); High-speed shaft brake torque (i.e., the moment applied to the high-speed shaft by the brake); Always positive (indicating dissipation of power)

# Rotor-Furl Bearing Loads
OutList['RFrlBrM']            = False     # (kN m); Rotor-furl bearing moment; About the rotor-furl axis

# Tail-Furl Bearing Loads
OutList['TFrlBrM']            = False     # (kN m); Tail-furl bearing moment; About the tail-furl axis

# Tail Fin Aerodynamic Loads
OutList['TFinAlpha']          = False     # (deg); Tail fin angle of attack. This is the angle between the relative velocity of the wind-inflow at the tail fin center-of-pressure and the tail fin chordline.; About the tail fin z-axis, which is the axis in the tail fin plane normal to the chordline
OutList['TFinCLift']          = False     # (-); Tail fin dimensionless lift coefficient; N/A
OutList['TFinCDrag']          = False     # (-); Tail fin dimensionless drag coefficient; N/A
OutList['TFinDnPrs']          = False     # (Pa); Tail fin dynamic pressure, equal to 1/2*AirDens*Vrel^2 where Vrel is the relative velocity of the wind-inflow at the tail fin center-of-pressure; N/A
OutList['TFinCPFx']           = False     # (kN); Tangential aerodynamic force at the tail fin center-of-pressure; Directed along the tail fin x-axis, which is the axis along the chordline, positive towards the trailing edge
OutList['TFinCPFy']           = False     # (kN); Normal aerodynamic force at the tail fin center-of-pressure; Directed along the tail fin y-axis, which is orthogonal to the tail fin plane

# Tower-Top / Yaw Bearing Loads
OutList['YawBrFxn']           = False     # (kN); Rotating (with nacelle) tower-top / yaw bearing shear force; Directed along the xn-axis
OutList['YawBrFyn']           = False     # (kN); Rotating (with nacelle) tower-top / yaw bearing shear force; Directed along the yn-axis
OutList['YawBrFzn']           = False     # (kN); Tower-top / yaw bearing axial force; Directed along the zn- and zp-axes
OutList['YawBrFzp']           = False     # (kN); Tower-top / yaw bearing axial force; Directed along the zn- and zp-axes
OutList['YawBrFxp']           = False     # (kN); Tower-top / yaw bearing fore-aft (nonrotating) shear force; Directed along the xp-axis
OutList['YawBrFyp']           = False     # (kN); Tower-top / yaw bearing side-to-side (nonrotating) shear force; Directed along the yp-axis
OutList['YawBrMxn']           = False     # (kN m); Rotating (with nacelle) tower-top / yaw bearing roll moment; About the xn-axis
OutList['YawBrMyn']           = False     # (kN m); Rotating (with nacelle) tower-top / yaw bearing pitch moment; About the yn-axis
OutList['YawBrMzn']           = False     # (kN m); Tower-top / yaw bearing yaw moment; About the zn- and zp-axes
OutList['YawBrMzp']           = False     # (kN m); Tower-top / yaw bearing yaw moment; About the zn- and zp-axes
OutList['YawMom']             = False     # (kN m); Tower-top / yaw bearing yaw moment; About the zn- and zp-axes
OutList['YawBrMxp']           = False     # (kN m); Nonrotating tower-top / yaw bearing roll moment; About the xp-axis
OutList['YawBrMyp']           = False     # (kN m); Nonrotating tower-top / yaw bearing pitch moment; About the yp-axis

# Tower Base Loads
OutList['TwrBsFxt']           = False     # (kN); Tower base fore-aft shear force; Directed along the xt-axis
OutList['TwrBsFyt']           = False     # (kN); Tower base side-to-side shear force; Directed along the yt-axis
OutList['TwrBsFzt']           = False     # (kN); Tower base axial force; Directed along the zt-axis
OutList['TwrBsMxt']           = False     # (kN m); Tower base roll (or side-to-side) moment (i.e., the moment caused by side-to-side forces); About the xt-axis
OutList['TwrBsMyt']           = False     # (kN m); Tower base pitching (or fore-aft) moment (i.e., the moment caused by fore-aft forces); About the yt-axis
OutList['TwrBsMzt']           = False     # (kN m); Tower base yaw (or torsional) moment; About the zt-axis

# Local Tower Loads
OutList['TwHt1MLxt']          = False     # (kN m); Local tower roll (or side-to-side) moment of tower gage 1; About the local xt-axis
OutList['TwHt1MLyt']          = False     # (kN m); Local tower pitching (or fore-aft) moment of tower gage 1; About the local yt-axis
OutList['TwHt1MLzt']          = False     # (kN m); Local tower yaw (or torsional) moment of tower gage 1; About the local zt-axis
OutList['TwHt2MLxt']          = False     # (kN m); Local tower roll (or side-to-side) moment of tower gage 2; About the local xt-axis
OutList['TwHt2MLyt']          = False     # (kN m); Local tower pitching (or fore-aft) moment of tower gage 2; About the local yt-axis
OutList['TwHt2MLzt']          = False     # (kN m); Local tower yaw (or torsional) moment of tower gage 2; About the local zt-axis
OutList['TwHt3MLxt']          = False     # (kN m); Local tower roll (or side-to-side) moment of tower gage 3; About the local xt-axis
OutList['TwHt3MLyt']          = False     # (kN m); Local tower pitching (or fore-aft) moment of tower gage 3; About the local yt-axis
OutList['TwHt3MLzt']          = False     # (kN m); Local tower yaw (or torsional) moment of tower gage 3; About the local zt-axis
OutList['TwHt4MLxt']          = False     # (kN m); Local tower roll (or side-to-side) moment of tower gage 4; About the local xt-axis
OutList['TwHt4MLyt']          = False     # (kN m); Local tower pitching (or fore-aft) moment of tower gage 4; About the local yt-axis
OutList['TwHt4MLzt']          = False     # (kN m); Local tower yaw (or torsional) moment of tower gage 4; About the local zt-axis
OutList['TwHt5MLxt']          = False     # (kN m); Local tower roll (or side-to-side) moment of tower gage 5; About the local xt-axis
OutList['TwHt5MLyt']          = False     # (kN m); Local tower pitching (or fore-aft) moment of tower gage 5; About the local yt-axis
OutList['TwHt5MLzt']          = False     # (kN m); Local tower yaw (or torsional) moment of tower gage 5; About the local zt-axis
OutList['TwHt6MLxt']          = False     # (kN m); Local tower roll (or side-to-side) moment of tower gage 6; About the local xt-axis
OutList['TwHt6MLyt']          = False     # (kN m); Local tower pitching (or fore-aft) moment of tower gage 6; About the local yt-axis
OutList['TwHt6MLzt']          = False     # (kN m); Local tower yaw (or torsional) moment of tower gage 6; About the local zt-axis
OutList['TwHt7MLxt']          = False     # (kN m); Local tower roll (or side-to-side) moment of tower gage 7; About the local xt-axis
OutList['TwHt7MLyt']          = False     # (kN m); Local tower pitching (or fore-aft) moment of tower gage 7; About the local yt-axis
OutList['TwHt7MLzt']          = False     # (kN m); Local tower yaw (or torsional) moment of tower gage 7; About the local zt-axis
OutList['TwHt8MLxt']          = False     # (kN m); Local tower roll (or side-to-side) moment of tower gage 8; About the local xt-axis
OutList['TwHt8MLyt']          = False     # (kN m); Local tower pitching (or fore-aft) moment of tower gage 8; About the local yt-axis
OutList['TwHt8MLzt']          = False     # (kN m); Local tower yaw (or torsional) moment of tower gage 8; About the local zt-axis
OutList['TwHt9MLxt']          = False     # (kN m); Local tower roll (or side-to-side) moment of tower gage 9; About the local xt-axis
OutList['TwHt9MLyt']          = False     # (kN m); Local tower pitching (or fore-aft) moment of tower gage 9; About the local yt-axis
OutList['TwHt9MLzt']          = False     # (kN m); Local tower yaw (or torsional) moment of tower gage 9; About the local zt-axis
OutList['TwHt1FLxt']          = False     # (kN); Local tower roll (or side-to-side) force of tower gage 1; About the local xt-axis
OutList['TwHt1FLyt']          = False     # (kN); Local tower pitching (or fore-aft) force of tower gage 1; About the local yt-axis
OutList['TwHt1FLzt']          = False     # (kN); Local tower yaw (or torsional) force of tower gage 1; About the local zt-axis
OutList['TwHt2FLxt']          = False     # (kN); Local tower roll (or side-to-side) force of tower gage 2; About the local xt-axis
OutList['TwHt2FLyt']          = False     # (kN); Local tower pitching (or fore-aft) force of tower gage 2; About the local yt-axis
OutList['TwHt2FLzt']          = False     # (kN); Local tower yaw (or torsional) force of tower gage 2; About the local zt-axis
OutList['TwHt3FLxt']          = False     # (kN); Local tower roll (or side-to-side) force of tower gage 3; About the local xt-axis
OutList['TwHt3FLyt']          = False     # (kN); Local tower pitching (or fore-aft) force of tower gage 3; About the local yt-axis
OutList['TwHt3FLzt']          = False     # (kN); Local tower yaw (or torsional) force of tower gage 3; About the local zt-axis
OutList['TwHt4FLxt']          = False     # (kN); Local tower roll (or side-to-side) force of tower gage 4; About the local xt-axis
OutList['TwHt4FLyt']          = False     # (kN); Local tower pitching (or fore-aft) force of tower gage 4; About the local yt-axis
OutList['TwHt4FLzt']          = False     # (kN); Local tower yaw (or torsional) force of tower gage 4; About the local zt-axis
OutList['TwHt5FLxt']          = False     # (kN); Local tower roll (or side-to-side) force of tower gage 5; About the local xt-axis
OutList['TwHt5FLyt']          = False     # (kN); Local tower pitching (or fore-aft) force of tower gage 5; About the local yt-axis
OutList['TwHt5FLzt']          = False     # (kN); Local tower yaw (or torsional) force of tower gage 5; About the local zt-axis
OutList['TwHt6FLxt']          = False     # (kN); Local tower roll (or side-to-side) force of tower gage 6; About the local xt-axis
OutList['TwHt6FLyt']          = False     # (kN); Local tower pitching (or fore-aft) force of tower gage 6; About the local yt-axis
OutList['TwHt6FLzt']          = False     # (kN); Local tower yaw (or torsional) force of tower gage 6; About the local zt-axis
OutList['TwHt7FLxt']          = False     # (kN); Local tower roll (or side-to-side) force of tower gage 7; About the local xt-axis
OutList['TwHt7FLyt']          = False     # (kN); Local tower pitching (or fore-aft) force of tower gage 7; About the local yt-axis
OutList['TwHt7FLzt']          = False     # (kN); Local tower yaw (or torsional) force of tower gage 7; About the local zt-axis
OutList['TwHt8FLxt']          = False     # (kN); Local tower roll (or side-to-side) force of tower gage 8; About the local xt-axis
OutList['TwHt8FLyt']          = False     # (kN); Local tower pitching (or fore-aft) force of tower gage 8; About the local yt-axis
OutList['TwHt8FLzt']          = False     # (kN); Local tower yaw (or torsional) force of tower gage 8; About the local zt-axis
OutList['TwHt9FLxt']          = False     # (kN); Local tower roll (or side-to-side) force of tower gage 9; About the local xt-axis
OutList['TwHt9FLyt']          = False     # (kN); Local tower pitching (or fore-aft) force of tower gage 9; About the local yt-axis
OutList['TwHt9FLzt']          = False     # (kN); Local tower yaw (or torsional) force of tower gage 9; About the local zt-axis

# Platform Loads
OutList['PtfmFxt']            = False     # (kN); Platform horizontal surge shear force; Directed along the xt-axis
OutList['PtfmFyt']            = False     # (kN); Platform horizontal sway shear force; Directed along the yt-axis
OutList['PtfmFzt']            = False     # (kN); Platform vertical heave force; Directed along the zt-axis
OutList['PtfmFxi']            = False     # (kN); Platform horizontal surge shear force; Directed along the xi-axis
OutList['PtfmFyi']            = False     # (kN); Platform horizontal sway shear force; Directed along the yi-axis
OutList['PtfmFzi']            = False     # (kN); Platform vertical heave force; Directed along the zi-axis
OutList['PtfmMxt']            = False     # (kN m); Platform roll tilt moment; About the xt-axis
OutList['PtfmMyt']            = False     # (kN m); Platform pitch tilt moment; About the yt-axis
OutList['PtfmMzt']            = False     # (kN m); Platform yaw moment; About the zt-axis
OutList['PtfmMxi']            = False     # (kN m); Platform roll tilt moment; About the xi-axis
OutList['PtfmMyi']            = False     # (kN m); Platform pitch tilt moment; About the yi-axis
OutList['PtfmMzi']            = False     # (kN m); Platform yaw moment; About the zi-axis

# Mooring Line Loads
OutList['Fair1Ten']           = False     # (kN); ; 
OutList['Fair1Ang']           = False     # (deg); ; 
OutList['Anch1Ten']           = False     # (kN); ; 
OutList['Anch1Ang']           = False     # (deg); ; 
OutList['Fair2Ten']           = False     # (kN); ; 
OutList['Fair2Ang']           = False     # (deg); ; 
OutList['Anch2Ten']           = False     # (kN); ; 
OutList['Anch2Ang']           = False     # (deg); ; 
OutList['Fair3Ten']           = False     # (kN); ; 
OutList['Fair3Ang']           = False     # (deg); ; 
OutList['Anch3Ten']           = False     # (kN); ; 
OutList['Anch3Ang']           = False     # (deg); ; 
OutList['Fair4Ten']           = False     # (kN); ; 
OutList['Fair4Ang']           = False     # (deg); ; 
OutList['Anch4Ten']           = False     # (kN); ; 
OutList['Anch4Ang']           = False     # (deg); ; 
OutList['Fair5Ten']           = False     # (kN); ; 
OutList['Fair5Ang']           = False     # (deg); ; 
OutList['Anch5Ten']           = False     # (kN); ; 
OutList['Anch5Ang']           = False     # (deg); ; 
OutList['Fair6Ten']           = False     # (kN); ; 
OutList['Fair6Ang']           = False     # (deg); ; 
OutList['Anch6Ten']           = False     # (kN); ; 
OutList['Anch6Ang']           = False     # (deg); ; 
OutList['Fair7Ten']           = False     # (kN); ; 
OutList['Fair7Ang']           = False     # (deg); ; 
OutList['Anch7Ten']           = False     # (kN); ; 
OutList['Anch7Ang']           = False     # (deg); ; 
OutList['Fair8Ten']           = False     # (kN); ; 
OutList['Fair8Ang']           = False     # (deg); ; 
OutList['Anch8Ten']           = False     # (kN); ; 
OutList['Anch8Ang']           = False     # (deg); ; 
OutList['Fair9Ten']           = False     # (kN); ; 
OutList['Fair9Ang']           = False     # (deg); ; 
OutList['Anch9Ten']           = False     # (kN); ; 
OutList['Anch9Ang']           = False     # (deg); ; 

# Wave Motions
OutList['WaveElev']           = False     # (m); ; 
OutList['Wave1Vxi']           = False     # (m/s); ; 
OutList['Wave1Vyi']           = False     # (m/s); ; 
OutList['Wave1Vzi']           = False     # (m/s); ; 
OutList['Wave1Axi']           = False     # (m/s^2); ; 
OutList['Wave1Ayi']           = False     # (m/s^2); ; 
OutList['Wave1Azi']           = False     # (m/s^2); ; 
OutList['Wave2Vxi']           = False     # (m/s); ; 
OutList['Wave2Vyi']           = False     # (m/s); ; 
OutList['Wave2Vzi']           = False     # (m/s); ; 
OutList['Wave2Axi']           = False     # (m/s^2); ; 
OutList['Wave2Ayi']           = False     # (m/s^2); ; 
OutList['Wave2Azi']           = False     # (m/s^2); ; 
OutList['Wave3Vxi']           = False     # (m/s); ; 
OutList['Wave3Vyi']           = False     # (m/s); ; 
OutList['Wave3Vzi']           = False     # (m/s); ; 
OutList['Wave3Axi']           = False     # (m/s^2); ; 
OutList['Wave3Ayi']           = False     # (m/s^2); ; 
OutList['Wave3Azi']           = False     # (m/s^2); ; 
OutList['Wave4Vxi']           = False     # (m/s); ; 
OutList['Wave4Vyi']           = False     # (m/s); ; 
OutList['Wave4Vzi']           = False     # (m/s); ; 
OutList['Wave4Axi']           = False     # (m/s^2); ; 
OutList['Wave4Ayi']           = False     # (m/s^2); ; 
OutList['Wave4Azi']           = False     # (m/s^2); ; 
OutList['Wave5Vxi']           = False     # (m/s); ; 
OutList['Wave5Vyi']           = False     # (m/s); ; 
OutList['Wave5Vzi']           = False     # (m/s); ; 
OutList['Wave5Axi']           = False     # (m/s^2); ; 
OutList['Wave5Ayi']           = False     # (m/s^2); ; 
OutList['Wave5Azi']           = False     # (m/s^2); ; 
OutList['Wave6Vxi']           = False     # (m/s); ; 
OutList['Wave6Vyi']           = False     # (m/s); ; 
OutList['Wave6Vzi']           = False     # (m/s); ; 
OutList['Wave6Axi']           = False     # (m/s^2); ; 
OutList['Wave6Ayi']           = False     # (m/s^2); ; 
OutList['Wave6Azi']           = False     # (m/s^2); ; 
OutList['Wave7Vxi']           = False     # (m/s); ; 
OutList['Wave7Vyi']           = False     # (m/s); ; 
OutList['Wave7Vzi']           = False     # (m/s); ; 
OutList['Wave7Axi']           = False     # (m/s^2); ; 
OutList['Wave7Ayi']           = False     # (m/s^2); ; 
OutList['Wave7Azi']           = False     # (m/s^2); ; 
OutList['Wave8Vxi']           = False     # (m/s); ; 
OutList['Wave8Vyi']           = False     # (m/s); ; 
OutList['Wave8Vzi']           = False     # (m/s); ; 
OutList['Wave8Axi']           = False     # (m/s^2); ; 
OutList['Wave8Ayi']           = False     # (m/s^2); ; 
OutList['Wave8Azi']           = False     # (m/s^2); ; 
OutList['Wave9Vxi']           = False     # (m/s); ; 
OutList['Wave9Vyi']           = False     # (m/s); ; 
OutList['Wave9Vzi']           = False     # (m/s); ; 
OutList['Wave9Axi']           = False     # (m/s^2); ; 
OutList['Wave9Ayi']           = False     # (m/s^2); ; 
OutList['Wave9Azi']           = False     # (m/s^2); ; 

# Internal Degrees of Freedom
OutList['Q_B1E1']             = False     # (m); Displacement of 1st edgewise bending-mode DOF of blade 1; 
OutList['Q_B2E1']             = False     # (m); Displacement of 1st edgewise bending-mode DOF of blade 2; 
OutList['Q_B3E1']             = False     # (m); Displacement of 1st edgewise bending-mode DOF of blade 3; 
OutList['Q_B1F1']             = False     # (m); Displacement of 1st flapwise bending-mode DOF of blade 1; 
OutList['Q_B2F1']             = False     # (m); Displacement of 1st flapwise bending-mode DOF of blade 2; 
OutList['Q_B3F1']             = False     # (m); Displacement of 1st flapwise bending-mode DOF of blade 3; 
OutList['Q_B1F2']             = False     # (m); Displacement of 2nd flapwise bending-mode DOF of blade 1; 
OutList['Q_B2F2']             = False     # (m); Displacement of 2nd flapwise bending-mode DOF of blade 2; 
OutList['Q_B3F2']             = False     # (m); Displacement of 2nd flapwise bending-mode DOF of blade 3; 
OutList['Q_Teet']             = False     # (rad); Displacement of hub teetering DOF; 
OutList['Q_DrTr']             = False     # (rad); Displacement of drivetrain rotational-flexibility DOF; 
OutList['Q_GeAz']             = False     # (rad); Displacement of variable speed generator DOF; 
OutList['Q_RFrl']             = False     # (rad); Displacement of rotor-furl DOF; 
OutList['Q_TFrl']             = False     # (rad); Displacement of tail-furl DOF; 
OutList['Q_Yaw']              = False     # (rad); Displacement of nacelle yaw DOF; 
OutList['Q_TFA1']             = False     # (m); Displacement of 1st tower fore-aft bending mode DOF; 
OutList['Q_TSS1']             = False     # (m); Displacement of 1st tower side-to-side bending mode DOF; 
OutList['Q_TFA2']             = False     # (m); Displacement of 2nd tower fore-aft bending mode DOF; 
OutList['Q_TSS2']             = False     # (m); Displacement of 2nd tower side-to-side bending mode DOF; 
OutList['Q_Sg']               = False     # (m); Displacement of platform horizontal surge translation DOF; 
OutList['Q_Sw']               = False     # (m); Displacement of platform horizontal sway translation DOF; 
OutList['Q_Hv']               = False     # (m); Displacement of platform vertical heave translation DOF; 
OutList['Q_R']                = False     # (rad); Displacement of platform roll tilt rotation DOF; 
OutList['Q_P']                = False     # (rad); Displacement of platform pitch tilt rotation DOF; 
OutList['Q_Y']                = False     # (rad); Displacement of platform yaw rotation DOF; 
OutList['QD_B1E1']            = False     # (m/s); Velocity of 1st edgewise bending-mode DOF of blade 1; 
OutList['QD_B2E1']            = False     # (m/s); Velocity of 1st edgewise bending-mode DOF of blade 2; 
OutList['QD_B3E1']            = False     # (m/s); Velocity of 1st edgewise bending-mode DOF of blade 3; 
OutList['QD_B1F1']            = False     # (m/s); Velocity of 1st flapwise bending-mode DOF of blade 1; 
OutList['QD_B2F1']            = False     # (m/s); Velocity of 1st flapwise bending-mode DOF of blade 2; 
OutList['QD_B3F1']            = False     # (m/s); Velocity of 1st flapwise bending-mode DOF of blade 3; 
OutList['QD_B1F2']            = False     # (m/s); Velocity of 2nd flapwise bending-mode DOF of blade 1; 
OutList['QD_B2F2']            = False     # (m/s); Velocity of 2nd flapwise bending-mode DOF of blade 2; 
OutList['QD_B3F2']            = False     # (m/s); Velocity of 2nd flapwise bending-mode DOF of blade 3; 
OutList['QD_Teet']            = False     # (rad/s); Velocity of hub teetering DOF; 
OutList['QD_DrTr']            = False     # (rad/s); Velocity of drivetrain rotational-flexibility DOF; 
OutList['QD_GeAz']            = False     # (rad/s); Velocity of variable speed generator DOF; 
OutList['QD_RFrl']            = False     # (rad/s); Velocity of rotor-furl DOF; 
OutList['QD_TFrl']            = False     # (rad/s); Velocity of tail-furl DOF; 
OutList['QD_Yaw']             = False     # (rad/s); Velocity of nacelle yaw DOF; 
OutList['QD_TFA1']            = False     # (m/s); Velocity of 1st tower fore-aft bending mode DOF; 
OutList['QD_TSS1']            = False     # (m/s); Velocity of 1st tower side-to-side bending mode DOF; 
OutList['QD_TFA2']            = False     # (m/s); Velocity of 2nd tower fore-aft bending mode DOF; 
OutList['QD_TSS2']            = False     # (m/s); Velocity of 2nd tower side-to-side bending mode DOF; 
OutList['QD_Sg']              = False     # (m/s); Velocity of platform horizontal surge translation DOF; 
OutList['QD_Sw']              = False     # (m/s); Velocity of platform horizontal sway translation DOF; 
OutList['QD_Hv']              = False     # (m/s); Velocity of platform vertical heave translation DOF; 
OutList['QD_R']               = False     # (rad/s); Velocity of platform roll tilt rotation DOF; 
OutList['QD_P']               = False     # (rad/s); Velocity of platform pitch tilt rotation DOF; 
OutList['QD_Y']               = False     # (rad/s); Velocity of platform yaw rotation DOF; 
OutList['QD2_B1E1']           = False     # (m/s^2); Acceleration of 1st edgewise bending-mode DOF of blade 1; 
OutList['QD2_B2E1']           = False     # (m/s^2); Acceleration of 1st edgewise bending-mode DOF of blade 2; 
OutList['QD2_B3E1']           = False     # (m/s^2); Acceleration of 1st edgewise bending-mode DOF of blade 3; 
OutList['QD2_B1F1']           = False     # (m/s^2); Acceleration of 1st flapwise bending-mode DOF of blade 1; 
OutList['QD2_B2F1']           = False     # (m/s^2); Acceleration of 1st flapwise bending-mode DOF of blade 2; 
OutList['QD2_B3F1']           = False     # (m/s^2); Acceleration of 1st flapwise bending-mode DOF of blade 3; 
OutList['QD2_B1F2']           = False     # (m/s^2); Acceleration of 2nd flapwise bending-mode DOF of blade 1; 
OutList['QD2_B2F2']           = False     # (m/s^2); Acceleration of 2nd flapwise bending-mode DOF of blade 2; 
OutList['QD2_B3F2']           = False     # (m/s^2); Acceleration of 2nd flapwise bending-mode DOF of blade 3; 
OutList['QD2_Teet']           = False     # (rad/s^2); Acceleration of hub teetering DOF; 
OutList['QD2_DrTr']           = False     # (rad/s^2); Acceleration of drivetrain rotational-flexibility DOF; 
OutList['QD2_GeAz']           = False     # (rad/s^2); Acceleration of variable speed generator DOF; 
OutList['QD2_RFrl']           = False     # (rad/s^2); Acceleration of rotor-furl DOF; 
OutList['QD2_TFrl']           = False     # (rad/s^2); Acceleration of tail-furl DOF; 
OutList['QD2_Yaw']            = False     # (rad/s^2); Acceleration of nacelle yaw DOF; 
OutList['QD2_TFA1']           = False     # (m/s^2); Acceleration of 1st tower fore-aft bending mode DOF; 
OutList['QD2_TSS1']           = False     # (m/s^2); Acceleration of 1st tower side-to-side bending mode DOF; 
OutList['QD2_TFA2']           = False     # (m/s^2); Acceleration of 2nd tower fore-aft bending mode DOF; 
OutList['QD2_TSS2']           = False     # (m/s^2); Acceleration of 2nd tower side-to-side bending mode DOF; 
OutList['QD2_Sg']             = False     # (m/s^2); Acceleration of platform horizontal surge translation DOF; 
OutList['QD2_Sw']             = False     # (m/s^2); Acceleration of platform horizontal sway translation DOF; 
OutList['QD2_Hv']             = False     # (m/s^2); Acceleration of platform vertical heave translation DOF; 
OutList['QD2_R']              = False     # (rad/s^2); Acceleration of platform roll tilt rotation DOF; 
OutList['QD2_P']              = False     # (rad/s^2); Acceleration of platform pitch tilt rotation DOF; 
OutList['QD2_Y']              = False     # (rad/s^2); Acceleration of platform yaw rotation DOF; 


""" Final Output Dictionary """
Fst7Output = {}
Fst7Output['OutList']         = OutList
